/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <linux/types.h>
#include <linux/config.h>

#include <asm/gt64260_defs.h>
#include "gt64260_i2c.h"
#include "nonstdio.h"

#define I2C_WRITE	0x0
#define I2C_READ	0x1

#ifdef  CONFIG_GT64260_NEW_BASE
static u32      gt64260_base = CONFIG_GT64260_NEW_REG_BASE;
#else
static u32      gt64260_base = CONFIG_GT64260_ORIG_REG_BASE;
#endif

static inline unsigned int
gt_read(unsigned int addr)
{
	unsigned ret;

	addr += gt64260_base;

	__asm__ __volatile__("lwbrx %0,0,%1; eieio" : "=r" (ret) :
			"r" (addr), "m" (*(unsigned int *)addr));
	return ret;
}

static inline void
gt_write(unsigned int addr, unsigned int val)
{
	addr += gt64260_base;

	__asm__ __volatile__("stwbrx %1,0,%2; eieio" :
			"=m" (*(unsigned int *)addr) :
			"r" (val), "r" (addr));
}
static void
gt64260_i2c_clear_iflg(void)
{
	unsigned int data;
	/* Clear IFlg */
	data = gt_read(GT64260_I2C_CONTROL);
	data &= ~GT64260_I2C_IFLG;
	gt_write(GT64260_I2C_CONTROL, data);

}

static void
gt64260_i2c_set_ack(void)
{
	unsigned int data;

	/* Set ACK */
	data = gt_read(GT64260_I2C_CONTROL);
	data |= GT64260_I2C_ACK;
	gt_write(GT64260_I2C_CONTROL, data);
}

static void
gt64260_i2c_clear_ack(void)
{
	unsigned int data;

	/* Clear ACK */
	data = gt_read(GT64260_I2C_CONTROL);
	data &= ~GT64260_I2C_ACK;
	gt_write(GT64260_I2C_CONTROL, data);
}

static void
gt64260_i2c_wait_status(unsigned int check)
{
	while(1)
	{
		if (gt_read(GT64260_I2C_CONTROL) & GT64260_I2C_IFLG)
			if (check == gt_read(GT64260_I2C_STATUS_BAUDE_RATE))
				break;
	}
}
/******************************************************************************
 * START: High -> Low on SDA while SCL is High
 ****************************************************************************/

static void
gt64260_i2c_send_start(void)
{
	unsigned int data;

	data = gt_read(GT64260_I2C_CONTROL);
	data |= GT64260_I2C_STA;
	gt_write(GT64260_I2C_CONTROL, data);

	gt64260_i2c_wait_status(GT64260_I2C_START_TXED);

	/* Note: Leaves with IFLG set */
}

/******************************************************************************
 * START: High -> Low on SDA while SCL is High
 *****************************************************************************/
static void
gt64260_i2c_send_restart(void)
{
	unsigned int data;


	data = gt_read(GT64260_I2C_CONTROL);
	data |= GT64260_I2C_STA;
	gt_write(GT64260_I2C_CONTROL, data);

        /* Clear IFLG to generate a new start */
        gt64260_i2c_clear_iflg();

	gt64260_i2c_wait_status(GT64260_I2C_RESTART_TXED);

	/* Note: Leaves with IFLG set */
}

/******************************************************************************
 * STOP: Low -> High on SDA while SCL is High
 ****************************************************************************/
static void
gt64260_i2c_send_stop(void)
{
	unsigned int data;

	data = gt_read(GT64260_I2C_CONTROL);
	data |= GT64260_I2C_STO;
	gt_write(GT64260_I2C_CONTROL, data);

	/* Clear IFlg */
        gt64260_i2c_clear_iflg();

	/* Poll till STO bit is cleared */
	while (1)
	{
		data = gt_read(GT64260_I2C_CONTROL);
		if (!(data & GT64260_I2C_STO)) break;
	}

}


/******************************************************************************

 *****************************************************************************/
static void
gt64260_i2c_write_addr(unsigned char addr)
{
	unsigned int check;

	/* addr = actual slave address as well as the read/write bit set */
	check = (addr & 0x01) ? GT64260_I2C_ADDR_RD_TXED_ACK_RCVD :
				GT64260_I2C_ADDR_WR_TXED_ACK_RCVD;

	/* Write "addr" */
	gt_write(GT64260_I2C_DATA, addr);

	/* Clear IFlg */
	gt64260_i2c_clear_iflg();

	gt64260_i2c_wait_status(check);

	/* Note: Leaves with IFLG set */
}

static unsigned char
gt64260_i2c_read_data(void)
{
        unsigned int check;

        /* Clear IFLG */
        gt64260_i2c_clear_iflg();

	/* setup expected status */
	check = (gt_read(GT64260_I2C_CONTROL) & GT64260_I2C_ACK) ?
		 GT64260_I2C_MST_RX_DATA_ACK_TXED :
		 GT64260_I2C_MST_RX_DATA_NACK_TXED;

	gt64260_i2c_wait_status(check);

	return gt_read(GT64260_I2C_DATA) & 0xFF;

	/* Note: Leaves with IFLG set */
}

/*-----------------------------------------------------------------------
 */
static void
gt64260_i2c_write_data(unsigned char byte)
{
	/* Write "byte" */
	gt_write(GT64260_I2C_DATA, byte);

	/* Clear IFlg */
	gt64260_i2c_clear_iflg();

	gt64260_i2c_wait_status(GT64260_I2C_MST_TX_DATA_ACK_RCVD);

	/* Note: Leaves with IFLG set */
}

/*
 * Page Write/Read
 */
static void
gt64260_i2c_transfer(unsigned int command, unsigned char address,
		     unsigned int size, unsigned char *bfr)
{
	unsigned int transfered = 0;

	/* Set up target address */
	gt64260_i2c_write_addr(address | command);

	while (transfered++ < size) {
		if (command == I2C_READ) {
			/* issue read command */
			if (transfered < size) {
				/* Set ACK bit */
				gt64260_i2c_set_ack();

				/* Data to be read */
				*bfr++ = gt64260_i2c_read_data();

				/* Clear ACK bit */
				gt64260_i2c_clear_ack();
			}
			else {
				/* Data to be read */
				*bfr++ = gt64260_i2c_read_data();
			}
		}
		else {
			/* Write data */
			gt64260_i2c_write_data(*bfr++);
		}
	}
}

/*
 * address        = i2c_address
 * size_to_expect = size of data
 * datain         = data location
 */
static inline void
gt64260_i2c_receive(unsigned char address,
                unsigned int size, unsigned char *bfr)
{

    gt64260_i2c_transfer(I2C_READ, address, size, bfr);

}

static inline void
gt64260_i2c_send(unsigned char address,
             unsigned int size, unsigned char *bfr)
{

    gt64260_i2c_transfer(I2C_WRITE, address, size, bfr);
}

void
gt64260_i2c_read (unsigned char *addr, unsigned int alen,
		  unsigned char *buffer, unsigned int len)
{
        if (alen != 2) {
		puts("I2C read: addr len ");
		puthex(alen);
		puts(" not supported\n");
                return;
        }

        /* Begin */
        gt64260_i2c_send_start();

        gt64260_i2c_send(*addr << 1, 1, addr+1);

        /* Begin */
        gt64260_i2c_send_restart();
        gt64260_i2c_receive(*addr << 1, len, buffer);

	/* End */
	gt64260_i2c_send_stop();
}

void
gt64260_i2c_init(void)
{

	/*
	 * Soft reset i2c logic to put all registers to defaults.
	 * default divisors (m = 4, n = 4) will produce a I2C clock of
	 * ~42 KHz at 66 MHz, 62.5 KHz at 100 MHz and 83 KHz at 133 MHz
	 * which is fine for what we'll be doing.
	 */
	gt_write(GT64260_I2C_SOFT_RESET, 0x0);

	/* Set Slave address register */
	gt_write(GT64260_I2C_SLAVE_ADDRESS, 0) ;

	/* Set Extended Slave address register */
	gt_write(GT64260_I2C_EXTENDED_SLAVE_ADDRESS, 0) ;

	/* Set clock dividers */
	gt_write(GT64260_I2C_STATUS_BAUDE_RATE, (4 << 3) | 4);

	/* Enable i2c */
	gt_write(GT64260_I2C_CONTROL, GT64260_I2C_ENA) ;

}
