/*
 * arch/ppc/kernel/gt64260_pic.c
 * 
 * Interrupt controller support for Galileo's GT64260.
 *
 * Author: Chris Zankel <chris@mvista.com>
 * Modified by: Mark A. Greer <mgreer@mvista.com>
 *
 * Based on sources from Rabeeh Khoury / Galileo Technology
 *
 * Copyright 2001 MontaVista Software Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

/*
 * This file contains the specific functions to support the GT64260
 * interrupt controller.
 *
 * The GT64260 has two main interrupt registers (high and low) that
 * summarizes the interrupts generated by the units of the GT64260.
 * Each bit is assigned to an interrupt number, where the low register
 * are assigned from IRQ0 to IRQ31 and the high cause register
 * from IRQ32 to IRQ63 
 * The GPP (General Purpose Port) interrupts are assigned from IRQ64 (GPP0) 
 * to IRQ95 (GPP31). 
 * get_irq() returns the lowest interrupt number that is currently asserted.
 *
 * Note: 
 *  - This driver does not initialize the GPP when used as an interrupt 
 *    input.
 *  - *** WARNING *** Only level sensitive GPP interrupts are supported
 */

#include <linux/stddef.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/stddef.h>
#include <linux/delay.h>
#include <linux/irq.h>

#include <asm/io.h>
#include <asm/processor.h>
#include <asm/system.h>
#include <asm/irq.h>
#include <asm/gt64260.h>


#define GPP_LEVEL_INTERRUPTS

/* #define ENABLE_ECC_INT_HANDLER */	/* currently broken.. causes deadlock of some kind */
/* #define ECC_DEBUG */

/* ========================== forward declaration ========================== */

static void gt64260_unmask_irq(unsigned int);
static void gt64260_mask_irq(unsigned int);
static void gt64260_end_irq(unsigned int irq);

/* ========================== local declarations =========================== */

struct hw_interrupt_type gt64260_pic = {
	" GT64260_PIC ",		/* typename */
	NULL,				/* startup */
	NULL,				/* shutdown */
	gt64260_unmask_irq,		/* enable */
	gt64260_mask_irq,		/* disable */
	gt64260_mask_irq,		/* mask and ack */
	gt64260_end_irq,		/* end */
	NULL				/* set_affinity */
};

u32 gt64260_irq_base = 0;      /* GT64260 handles the next 96 IRQs from here */

#ifdef ENABLE_ECC_INT_HANDLER
static u32
gt_ecc_irq_ack(void)
{
    /* NOTE: If ecc_irq_ack is called during mask_and_ack,
     * we lose data_hi and data_lo if we get a new ECC error between
     * the time ack gets called and gt_ecc_irq gets called since reading addr
     * allows new data to get latched in. FIXME!!
     * See GT PIC documentation. */
    /* Currently, we DON'T have a mask_and_ack_irq, so this code is fine */
    static u32 last_addr=0;	/* store local copy since irq_ack overwrites */
    u32 addr=gt_read(GT64260_SDRAM_ERR_ADDR);
    gt_write(GT64260_SDRAM_ERR_ADDR,addr & (~3));

    if(addr&3) last_addr=addr;
    return last_addr;
}

static void
gt_ecc_irq(s32 irq, void *dev_id, struct pt_regs *regs)
{

    u32 addr;
#ifdef ECC_DEBUG
    u32 data_lo, data_hi,type;
    printk("ecc ERR: NIP=%08x LINK=%08x\n", regs->nip, regs->link);

    data_lo=gt_read(GT64260_SDRAM_ERR_DATA_LO);
    data_hi=gt_read(GT64260_SDRAM_ERR_DATA_HI);
#endif

    addr=gt_ecc_irq_ack();

#ifdef ECC_DEBUG
    type=addr&3;

    printk("data:%08x:%08x addr:%08x type:%d\n", data_hi, data_lo, addr&(~3), type);

    printk("rcvd:%08x calc:%08x cntl:%08x cnt:%08x\n",
	    gt_read(GT64260_SDRAM_ERR_ECC_RCVD),
	    gt_read(GT64260_SDRAM_ERR_ECC_CALC),
	    gt_read(GT64260_SDRAM_ERR_ECC_CNTL),
	    gt_read(GT64260_SDRAM_ERR_ECC_ERR_CNT));

    gt_write(GT64260_SDRAM_ERR_ECC_ERR_CNT,0);
#endif /* ECC_DEBUG */

}
#endif /* ENABLE_ECC_INT_HANDLER */


/* gt64260_init_irq()
 *
 *  This function initializes the interrupt controller. It assigns
 *  all interrupts from IRQ0 to IRQ95 to the gt64260 interrupt controller.
 *
 * Input Variable(s):
 *  None.
 *
 * Outpu. Variable(s):
 *  None.
 *
 * Returns:
 *  void
 *
 * Note:
 *  We register all GPP inputs as interrupt source, but disable them.
 */

__init void
gt64260_init_irq(void)
{
	int i;

	if ( ppc_md.progress ) ppc_md.progress("gt64260_init_irq: enter", 0x0);

	ppc_cached_irq_mask[0] = 0;
	ppc_cached_irq_mask[1] = 0x0f000000; /* Enable GPP intrs */
	ppc_cached_irq_mask[2] = 0;

	/* disable all interrupts and clear current interrupts */
	gt_write(GT64260_GPP_INTR_MASK, ppc_cached_irq_mask[2]);
	gt_write(GT64260_GPP_INTR_CAUSE,0);
	gt_write(GT64260_IC_CPU_INTR_MASK_LO, ppc_cached_irq_mask[0]);
	gt_write(GT64260_IC_CPU_INTR_MASK_HI, ppc_cached_irq_mask[1]);

	/* use the gt64260 for all (possible) interrupt sources */
	for( i = gt64260_irq_base;  i < (gt64260_irq_base + 96);  i++ )  {
		irq_desc[i].handler = &gt64260_pic;
	}

#ifdef GPP_LEVEL_INTERRUPTS
        if(gt64260_revision != GT64260)  {
		if ( ppc_md.progress )
			ppc_md.progress("gt64260_init_irq: GPP -> levelint ", 0x0);
                /* #FEr MISC-4 */
                /* Mark GPP intrs level sensitive */
                gt_set_bits(GT64260_COMM_ARBITER_CNTL, (1<<10));

                for(i=gt64260_irq_base+64;i<gt64260_irq_base+64+32;i++)
			irq_desc[i].status |= IRQ_LEVEL;
        }
#endif

#ifdef ENABLE_ECC_INT_HANDLER
	request_irq(gt64260_irq_base + 17, gt_ecc_irq, 0, "GT64260 SDRAM ECC Error", 0);
#endif

	if ( ppc_md.progress ) ppc_md.progress("gt64260_init_irq: exit", 0x0);
	
}


/* gt64260_get_irq()
 *
 *  This function returns the lowest interrupt number of all interrupts that
 *  are currently asserted. 
 *
 * Input Variable(s):
 *  struct pt_regs*	not used
 *
 * Output Variable(s):
 *  None.
 *
 * Returns:
 *  int	<interrupt number> or -2 (bogus interrupt)
 * 
 */
int
gt64260_get_irq(struct pt_regs *regs)
{
	int irq;
	int irq_gpp;
	
	int cpu = smp_processor_id();

	irq = gt_read(GT64260_IC_MAIN_CAUSE_LO);
	irq = __ilog2((irq & 0x3dfffffe) & ppc_cached_irq_mask[0]);

	if (irq == -1) {
		irq = gt_read(GT64260_IC_MAIN_CAUSE_HI);
#ifdef CONFIG_MOT_MVP
		/* Okay, this whole irq mess needs a rewrite bad, but who
		 * am I to argue with a deadline? -- Troy
		 */
		if (cpu == 1){
			irq = __ilog2((irq & 0x08000000) & ppc_cached_irq_mask[1]);
		}else{
			irq = __ilog2((irq & 0x07000db7) & ppc_cached_irq_mask[1]);
		}
#else
		irq = __ilog2((irq & 0x0f000db7) & ppc_cached_irq_mask[1]);
#endif

		if (irq == -1) {
			irq = -2;   /* bogus interrupt, should never happen */
		}
		else if (irq < 24) {
			irq += 32;
		}
		else {
			irq_gpp = gt_read(GT64260_GPP_INTR_CAUSE);
			irq_gpp = __ilog2(irq_gpp &
					  ppc_cached_irq_mask[2]);

			if (irq_gpp == -1) {
				irq = -2;
			} else {
				irq = irq_gpp + 64;
			}
		}
	}

	if( irq < 0 )  {
		return( irq );
	} else  {
		return( gt64260_irq_base + irq );
	}
}

/* gt64260_unmask_irq()
 *
 *  This function enables an interrupt.
 *
 * Input Variable(s):
 *  unsigned int	interrupt number (IRQ0...IRQ95).
 *
 * Output Variable(s):
 *  None.
 *
 * Returns:
 *  void
 */

static void
gt64260_unmask_irq(unsigned int irq)
{
	irq -= gt64260_irq_base;
	if (irq > 31) {
		if (irq > 63) {
			/* unmask GPP irq */
			gt_write(GT64260_GPP_INTR_MASK,  
				     ppc_cached_irq_mask[2] |= (1<<(irq-64)));
		} else {
			/* unmask high interrupt register */
			gt_write(GT64260_IC_CPU_INTR_MASK_HI, 
				     ppc_cached_irq_mask[1] |= (1<<(irq-32)));
		}
	} else {
		/* unmask low interrupt register */
		gt_write(GT64260_IC_CPU_INTR_MASK_LO, 
			     ppc_cached_irq_mask[0] |= (1<<irq));
	}
}


/* gt64260_mask_irq()
 *
 *  This funktion disables the requested interrupt.
 *
 * Input Variable(s):
 *  unsigned int	interrupt number (IRQ0...IRQ95).
 *
 * Output Variable(s):
 *  None.
 *
 * Returns:
 *  void
 */

static void
gt64260_mask_irq(unsigned int irq)
{
	irq -= gt64260_irq_base;
	if (irq > 31) {
		if (irq > 63) {
			/* mask GPP irq */
			gt_write(GT64260_GPP_INTR_MASK,  
				     ppc_cached_irq_mask[2] &= ~(1<<(irq-64)));
		} else {
			/* mask high interrupt register */
			gt_write(GT64260_IC_CPU_INTR_MASK_HI, 
				     ppc_cached_irq_mask[1] &= ~(1<<(irq-32)));
		}
	} else {
		/* mask low interrupt register */
		gt_write(GT64260_IC_CPU_INTR_MASK_LO, 
			     ppc_cached_irq_mask[0] &= ~(1<<irq));
	}

	if (irq == 36) { /* Seems necessary for SDMA interrupts */
		udelay(1);
	}
}


/* gt64260_end_irq()
 *
 * This function unmasks the irq, if its not disabled or in progress
 *
 * Input Variable(s):
 *  unsigned int	interrupt number (IRQ0...IRQ95).
 *
 * Output Variable(s):
 *  None.
 *
 * Returns:
 *  void
 */
static void
gt64260_end_irq(unsigned int irq)
{
	int local_irq = irq - gt64260_irq_base;

	if (!(irq_desc[irq].status & (IRQ_DISABLED|IRQ_INPROGRESS))) {
		/* FEr MISC-4 */
		if ((local_irq > 63) && (irq_desc[irq].status & IRQ_LEVEL)) {
			gt_write(GT64260_GPP_INTR_CAUSE,
				 ~(1<<(local_irq-64)));
		}

		gt64260_unmask_irq(irq);
	}
}
