/*
 * arch/ppc/kernel/ppc4xx_dma.c
 *
 * IBM PPC4xx DMA engine core library
 *
 * Copyright 2000-2003 MontaVista Software Inc.
 *
 * Cleaned by Matt Porter <mporter@mvista.com>
 *
 * Original code by Armin Kuster <akuster@mvista.com>
 * and Pete Popov <ppopov@mvista.com>
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 *
 * You should have received a copy of the  GNU General Public License along
 * with this program; if not, write  to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>
#include <linux/init.h>
#include <linux/module.h>
#include <asm/ibm4xx.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/ppc4xx_dma.h>
#include <asm/ocp.h>

#define NAME "ppc4xx_dma: "

ppc_dma_ch_t *dma_channels[MAX_PPC4xx_DMA_CHANNELS];


unsigned char *dma_stat_str[] = {
	"n/a (bit31)",			/* bit 31 */
	"n/a (bit30)",
	"n/a (bit29)",
	"n/a (bit28)",

	"Ch3: Scatter/gather operation in progress (SG3)",
	"Ch2: Scatter/gather operation in progress (SG2)",
	"Ch1: Scatter/gather operation in progress (SG1)",
	"Ch0: Scatter/gather operation in progress (SG0)",

	"Ch3: Scatter/gather is active, fetching (CB3)",
	"Ch2: Scatter/gather is active, fetching (CB2)",
	"Ch1: Scatter/gather is active, fetching (CB1)",
	"Ch0: Scatter/gather is active, fetching (CB0)",

	"Ch3: External DMA Request pending (ER3)",
	"Ch2: External DMA Request pending (ER2)",
	"Ch1: External DMA Request pending (ER1)",
	"Ch0: External DMA Request pending (ER0)",

	"Ch3: Internal DMA Request pending (IR3)",
	"Ch2: Internal DMA Request pending (IR2)",
	"Ch1: Internal DMA Request pending (IR1)",
	"Ch0: Internal DMA Request pending (IR0)",

	"Ch3: Error occurred (RI3)",
	"Ch2: Error occurred (RI2)",
	"Ch1: Error occurred (RI1)",
	"Ch0: Error occurred (RI0)",

	"Ch3: End Of Transfer (TS3)",
	"Ch2: End Of Transfer (TS2)",
	"Ch1: End Of Transfer (TS1)",
	"Ch0: End Of Transfer (TS0)",

	"Ch3: Terminal Count has been reached (CS3)",
	"Ch2: Terminal Count has been reached (CS2)",
	"Ch1: Terminal Count has been reached (CS1)",
	"Ch0: Terminal Count has been reached (CS0)",	/* bit 0 */
};

void
ppc4xx_show_status(int dmac)
{
	unsigned int stat;

	/* 
	 * check status register.
	 * these bits will be cleared when next enable_dma() is called.
	 */
	stat = ppc4xx_get_dma_status(dmac);

	printk("%s(): DMAC = %d\n", __FUNCTION__, dmac);
	while (ffs(stat)) {
		int bit = ffs(stat);
		printk("%s(): %s\n", __FUNCTION__, dma_stat_str[bit-1]);
		stat &= ~(1<<(bit-1));
	}
	return ;
}

void
ppc4xx_show_dmach(int dmanr)
{
	ppc_dma_ch_t *p_dma_ch = dma_channels[dmanr];
	
	if (!p_dma_ch) {
		printk("%s(): p_dma_ch[%d] is NULL\n",
		       __FUNCTION__, dmanr);
		return;
	}

	printk("channel %d\n" \
	       "in_use        = %d\n" \
	       "int_enable    = %d\n" \
	       "mode          = 0x%x\n" \
	       "pl            = %d\n" \
	       "pwidth        = %d\n" \
	       "dai/sai       = %d/%d\n" \
	       "buffer_enable = %d\n" \
	       "psc/pwc/phc   = %d/%d/%d\n" \
	       "etd_output    = %d\n" \
	       "pf            = 0x%x\n" \
	       "sl            = %d\n"
	       , dmanr,
	       p_dma_ch->in_use,
	       p_dma_ch->int_enable,
	       p_dma_ch->mode,
	       p_dma_ch->pl,
	       p_dma_ch->pwidth,
	       p_dma_ch->dai, p_dma_ch->sai, 
	       p_dma_ch->buffer_enable,
	       p_dma_ch->psc, p_dma_ch->pwc, p_dma_ch->phc,
	       p_dma_ch->etd_output,
	       p_dma_ch->pf,
	       p_dma_ch->sl);
}

/* 
   ppc4xx_request_dma()

   Allocates DMAC0 (or DMAC1 if available) and returns DMA channel
   number.

   Notes:
   - Because it calls kmalloc() inside, it cannot be called within
     interrupt handlers.

   Requires:
      dmac - DMAC Number (0 or 1)
   Returns: 
      Channel Number (0-3 for DMAC0, 4-7 for DMAC1)
 */
int
ppc4xx_request_dma(int dmac)
{
	struct ocp_dev* ocp;
	int ret;
	int dmanr;

	if (!(ocp = ocp_alloc_dev(sizeof(ppc_dma_ch_t)))){
		printk(KERN_CRIT NAME "failed to allocate device data\n");
		return -ENOMEM;
	}

	switch (dmac) {
	case 0:
		ocp->type = DMAC0;
		break;
	case 1:
		ocp->type = DMAC1;
		break;
	default:
		printk(KERN_CRIT NAME "unknown DMAC number (%d)\n", dmac);
		goto fail;
	}
	if ((ret = ocp_register_num(ocp, -1)) < 0){
		printk(KERN_CRIT NAME "no dma channel available.\n");
		goto fail;
	}

	dmanr = ocp->num;
#if NR_DMACS > 1
	if (dmac == 1)
		dmanr += CH_PER_DMAC;
#endif
	dma_channels[dmanr] = (ppc_dma_ch_t*)ocp->ocpdev;

#ifdef DEBUG_4xxDMA
	printk("%s() dmanr = %d, ocp = %p, p_dma_ch = %p\n", __FUNCTION__, 
	       dmanr, ocp, ocp->ocpdev);
	ppc4xx_show_dmach(dmanr);
#endif
	return dmanr;

fail:
	ocp_free_dev(ocp);	
	return ret;
}

/* 
   ppc4xx_request_dma_channel()

   Allocates DMA channel specifing DMA channel number explicitly.

   Notes:
   - Because it calls kmalloc() inside, it cannot be called within
     interrupt handlers.

   Requires:
      Channel Number (0-3 for DMAC0, 4-7 for DMAC1)

   Returns: 
      Channel Number (0-3 for DMAC0, 4-7 for DMAC1)
      Negative value on error.
 */
int
ppc4xx_request_dma_channel(unsigned int dmanr)
{
	struct ocp_dev* ocp;
	int ret;
	int dev_num;

	if (dmanr >= MAX_PPC4xx_DMA_CHANNELS)
		return -EINVAL;

	if (!(ocp = ocp_alloc_dev(sizeof(ppc_dma_ch_t)))){
		printk(KERN_CRIT NAME "failed to allocate device data\n");
		return -ENOMEM;
	}

	switch (CHNUM_TO_DMACNUM(dmanr)) {
	case 0:
		ocp->type = DMAC0;
		break;
	case 1:
		ocp->type = DMAC1;
		break;
	default:
		printk("%s: bad dmanr %d.", __FUNCTION__, dmanr);
		ret = -EINVAL;
		goto fail;
	}

	dev_num = dmanr & 0x3;
	if ((ret = ocp_register_num(ocp, dev_num)) < 0){
		printk(KERN_CRIT NAME "channel %d reserved.\n", dmanr);
		goto fail;
	}

	dma_channels[dmanr] = (ppc_dma_ch_t*)ocp->ocpdev;

#ifdef DEBUG_4xxDMA
	printk("%s() dmanr = %d, ocp = %p, p_dma_ch = %p\n", __FUNCTION__, 
	       dmanr, ocp, ocp->ocpdev);
	ppc4xx_show_dmach(dmanr);
#endif
	return dmanr;

fail:
	ocp_free_dev(ocp);	
	return ret;
}


/* 
   ppc4xx_free_dma()

   Frees DMA channel reserved by ppc4xx_request_dma().

   Requires:
       dmanr - DMA chaneel number (0-3 for DMAC0, 4-7 for DMAC1)
 */
void
ppc4xx_free_dma(int dmanr)
{
	struct ocp_dev* ocp;

	switch (CHNUM_TO_DMACNUM(dmanr)) {
	case 0:
		ocp = ocp_get_dev(DMAC0, dmanr);
		break;
	case 1:
		ocp = ocp_get_dev(DMAC1, dmanr - CH_PER_DMAC);
		break;
	}

#ifdef DEBUG_4xxDMA
	printk("%s() dmanr = %d, ocp=%p\n", __FUNCTION__, dmanr, ocp);
#endif
	if (!ocp) {
		printk(NAME "Trying to free an unreserved DMA CH %d\n", dmanr);
		return;
	}

	ocp_unregister(ocp);
	ocp_free_dev(ocp);

	return;
}

/* 
   ppc4xx_get_dma_irq()

   Returns IRQ number assinged for DMA channel.

   Requires:
       dmanr - DMA chaneel number (0-3 for DMAC0, 4-7 for DMAC1)
   Returns:
       IRQ number.
       < 0: for errors.
 */
int
ppc4xx_get_dma_irq(int dmanr)
{
	int irq = 0;

	switch(CHNUM_TO_DMACNUM(dmanr)) {
	case 0:
		irq = ocp_get_irq(DMAC0, dmanr);
		break;
	case 1:
		irq = ocp_get_irq(DMAC1, dmanr - CH_PER_DMAC);
		break;
	}

#ifdef DEBUG_4xxDMA
	printk("%s() irq = %d\n", __FUNCTION__, irq);
#endif
	return irq;
}


unsigned int
ppc4xx_get_dma_status(int dmac)
{
	switch (dmac) {
	case 0:
		return (mfdcr(DCRN_DMA0_SR));
		break;
#if NR_DMACS > 1
	case 1:
		return (mfdcr(DCRN_DMA1_SR));
		break;
#endif
	default:
		if (dmac >= NR_DMACS)
			printk("%s: bad dmac: %d", __FUNCTION__, dmac);
	}
	return -1;
}

void 
ppc4xx_clear_dma_status(int dmanr)
{
	unsigned int status_bits[] = { DMA_CS0 | DMA_TS0 | DMA_CH0_ERR,
				       DMA_CS1 | DMA_TS1 | DMA_CH1_ERR,
				       DMA_CS2 | DMA_TS2 | DMA_CH2_ERR,
				       DMA_CS3 | DMA_TS3 | DMA_CH3_ERR};

	switch (CHNUM_TO_DMACNUM(dmanr)){
	case 0:
		mtdcr(DCRN_DMA0_SR, status_bits[dmanr]);
		break;
#if NR_DMACS > 1
	case 1:
		mtdcr(DCRN_DMA1_SR, status_bits[dmanr - CH_PER_DMAC]);
		break;
#endif
	default:
		printk("%s: bad dmanr %d.", __FUNCTION__, dmanr);
	}

	return ;
}

void
ppc4xx_set_src_addr(int dmanr, phys_addr_t src_addr)
{
	switch (dmanr) {
		case 0:
#ifdef PPC4xx_DMA_64BIT
			mtdcr(DCRN_DMASAH0, (u32)(src_addr >> 32));
#endif		
			mtdcr(DCRN_DMASA0, (u32)src_addr);
			break;
		case 1:
#ifdef PPC4xx_DMA_64BIT 
			mtdcr(DCRN_DMASAH1, (u32)(src_addr >> 32));
#endif		
			mtdcr(DCRN_DMASA1, (u32)src_addr);
			break;
		case 2:
#ifdef PPC4xx_DMA_64BIT
			mtdcr(DCRN_DMASAH2, (u32)(src_addr >> 32));
#endif			
			mtdcr(DCRN_DMASA2, (u32)src_addr);
			break;
		case 3:
#ifdef PPC4xx_DMA_64BIT
			mtdcr(DCRN_DMASAH3, (u32)(src_addr >> 32));
#endif			
			mtdcr(DCRN_DMASA3, (u32)src_addr);
			break;
#if NR_DMACS > 1
		case 4:
#ifdef PPC4xx_DMA_64BIT
			mtdcr(DCRN_DMASAH4, (u32)(src_addr >> 32));
#endif		
			mtdcr(DCRN_DMASA4, (u32)src_addr);
			break;
		case 5:
#ifdef PPC4xx_DMA_64BIT 
			mtdcr(DCRN_DMASAH5, (u32)(src_addr >> 32));
#endif		
			mtdcr(DCRN_DMASA5, (u32)src_addr);
			break;
		case 6:
#ifdef PPC4xx_DMA_64BIT
			mtdcr(DCRN_DMASAH6, (u32)(src_addr >> 32));
#endif			
			mtdcr(DCRN_DMASA6, (u32)src_addr);
			break;
		case 7:
#ifdef PPC4xx_DMA_64BIT
			mtdcr(DCRN_DMASAH7, (u32)(src_addr >> 32));
#endif			
			mtdcr(DCRN_DMASA7, (u32)src_addr);
			break;
#endif /* NR_DMACS */
		default:
			if (dmanr >= MAX_PPC4xx_DMA_CHANNELS)
				printk("%s: bad channel: %d\n", __FUNCTION__, dmanr);
	}
}

void
ppc4xx_set_dst_addr(int dmanr, phys_addr_t dst_addr)
{
	switch (dmanr) {
		case 0:
#ifdef PPC4xx_DMA_64BIT
			mtdcr(DCRN_DMADAH0, (u32)(dst_addr >> 32));
#endif
			mtdcr(DCRN_DMADA0, (u32)dst_addr);
			break;
		case 1:
#ifdef PPC4xx_DMA_64BIT
			mtdcr(DCRN_DMADAH1, (u32)(dst_addr >> 32));
#endif			
			mtdcr(DCRN_DMADA1, (u32)dst_addr);
			break;
		case 2:
#ifdef PPC4xx_DMA_64BIT
			mtdcr(DCRN_DMADAH2, (u32)(dst_addr >> 32));
#endif
			mtdcr(DCRN_DMADA2, (u32)dst_addr);
			break;
		case 3:
#ifdef PPC4xx_DMA_64BIT
			mtdcr(DCRN_DMADAH3, (u32)(dst_addr >> 32));
#endif
			mtdcr(DCRN_DMADA3, (u32)dst_addr);
			break;
#if NR_DMACS > 1
		case 4:
#ifdef PPC4xx_DMA_64BIT
			mtdcr(DCRN_DMADAH4, (u32)(dst_addr >> 32));
#endif
			mtdcr(DCRN_DMADA4, (u32)dst_addr);
			break;
		case 5:
#ifdef PPC4xx_DMA_64BIT
			mtdcr(DCRN_DMADAH5, (u32)(dst_addr >> 32));
#endif			
			mtdcr(DCRN_DMADA5, (u32)dst_addr);
			break;
		case 6:
#ifdef PPC4xx_DMA_64BIT
			mtdcr(DCRN_DMADAH6, (u32)(dst_addr >> 32));
#endif
			mtdcr(DCRN_DMADA6, (u32)dst_addr);
			break;
		case 7:
#ifdef PPC4xx_DMA_64BIT
			mtdcr(DCRN_DMADAH7, (u32)(dst_addr >> 32));
#endif
			mtdcr(DCRN_DMADA7, (u32)dst_addr);
			break;
#endif /* NR_DMACS */
		default:
			if (dmanr >= MAX_PPC4xx_DMA_CHANNELS)
				printk("%s: bad channel: %d\n", __FUNCTION__, dmanr);
	}
}


void
ppc4xx_enable_dma(unsigned int dmanr)
{
	unsigned int control;
	ppc_dma_ch_t *p_dma_ch = dma_channels[dmanr];

#ifdef DEBUG_4xxDMA
	printk("%s()\n", __FUNCTION__);
#endif

	if (p_dma_ch->in_use) {
		printk("%s: channel %d in use\n", __FUNCTION__, dmanr);
		return;
	}

	if (dmanr >= MAX_PPC4xx_DMA_CHANNELS) {
		printk("%s: bad channel: %d\n", __FUNCTION__, dmanr);
		return;
	}

	if (p_dma_ch->mode == DMA_MODE_READ) {
		/* peripheral to memory */
		ppc4xx_set_src_addr(dmanr, 0);
		ppc4xx_set_dst_addr(dmanr, p_dma_ch->addr);
	} else if (p_dma_ch->mode == DMA_MODE_WRITE) {
		/* memory to peripheral */
		ppc4xx_set_src_addr(dmanr, p_dma_ch->addr);
		ppc4xx_set_dst_addr(dmanr, 0);
	}

	/* for other xfer modes, the addresses are already set */
	switch (dmanr) {
		case 0:
			control = mfdcr(DCRN_DMACR0);
			break;
		case 1:
			control = mfdcr(DCRN_DMACR1);
			break;
		case 2:
			control = mfdcr(DCRN_DMACR2);
			break;
		case 3:
			control = mfdcr(DCRN_DMACR3);
			break;
#if NR_DMACS > 1
		case 4:
			control = mfdcr(DCRN_DMACR4);
			break;
		case 5:
			control = mfdcr(DCRN_DMACR5);
			break;
		case 6:
			control = mfdcr(DCRN_DMACR6);
			break;
		case 7:
			control = mfdcr(DCRN_DMACR7);
			break;
#endif /* NR_DMACS */
		default:
			printk("%s: bad channel: %d\n", __FUNCTION__, dmanr);
	}

	control &= ~(DMA_TM_MASK | DMA_TD);	/* clear all mode bits */
	if (p_dma_ch->mode == DMA_MODE_MM) {
		/* software initiated memory to memory */
		control |= DMA_ETD_OUTPUT | DMA_TCE_ENABLE;
	}

	switch (dmanr) {
		case 0:
			mtdcr(DCRN_DMACR0, control);
			break;
		case 1:
			mtdcr(DCRN_DMACR1, control);
			break;
		case 2:
			mtdcr(DCRN_DMACR2, control);
			break;
		case 3:
			mtdcr(DCRN_DMACR3, control);
			break;
#if NR_DMACS > 1
		case 4:
			mtdcr(DCRN_DMACR4, control);
			break;
		case 5:
			mtdcr(DCRN_DMACR5, control);
			break;
		case 6:
			mtdcr(DCRN_DMACR6, control);
			break;
		case 7:
			mtdcr(DCRN_DMACR7, control);
			break;
#endif /* NR_DMACS */
		default:
			printk("%s: bad channel: %d\n", __FUNCTION__, dmanr);
	}

	/*
	 * Clear the CS, TS, RI bits for the channel from DMASR.  This
	 * has been observed to happen correctly only after the mode and
	 * ETD/DCE bits in DMACRx are set above.  Must do this before
	 * enabling the channel.
	 */
	ppc4xx_clear_dma_status(dmanr);

	/*
	 * For device-paced transfers, Terminal Count Enable apparently
	 * must be on, and this must be turned on after the mode, etc.
	 * bits are cleared above (at least on Redwood-6).
	 */

	if ((p_dma_ch->mode == DMA_MODE_MM_DEVATDST) ||
	    (p_dma_ch->mode == DMA_MODE_MM_DEVATSRC))
		control |= DMA_TCE_ENABLE;

	/*
	 * Now enable the channel.
	 */

	control |= (p_dma_ch->mode | DMA_CE_ENABLE);

	switch (dmanr) {
		case 0:
			mtdcr(DCRN_DMACR0, control);
			break;
		case 1:
			mtdcr(DCRN_DMACR1, control);
			break;
		case 2:
			mtdcr(DCRN_DMACR2, control);
			break;
		case 3:
			mtdcr(DCRN_DMACR3, control);
			break;
#if NR_DMACS > 1
		case 4:
			mtdcr(DCRN_DMACR4, control);
			break;
		case 5:
			mtdcr(DCRN_DMACR5, control);
			break;
		case 6:
			mtdcr(DCRN_DMACR6, control);
			break;
		case 7:
			mtdcr(DCRN_DMACR7, control);
			break;
#endif /* NR_DMACS */
		default:
			printk("%s: bad channel: %d\n", __FUNCTION__, dmanr);
	}

	p_dma_ch->in_use = 1;
}

void
ppc4xx_disable_dma(unsigned int dmanr)
{
	unsigned int control;
	ppc_dma_ch_t *p_dma_ch = dma_channels[dmanr];

	if (!p_dma_ch->in_use) {
		printk("%s: channel %d not in use\n", __FUNCTION__, dmanr);
		return;
	}

	if (dmanr >= MAX_PPC4xx_DMA_CHANNELS) {
		printk("%s: bad channel: %d\n", __FUNCTION__, dmanr);
		return;
	}

	switch (dmanr) {
	case 0:
		control = mfdcr(DCRN_DMACR0);
		control &= ~DMA_CE_ENABLE;
		mtdcr(DCRN_DMACR0, control);
		break;
	case 1:
		control = mfdcr(DCRN_DMACR1);
		control &= ~DMA_CE_ENABLE;
		mtdcr(DCRN_DMACR1, control);
		break;
	case 2:
		control = mfdcr(DCRN_DMACR2);
		control &= ~DMA_CE_ENABLE;
		mtdcr(DCRN_DMACR2, control);
		break;
	case 3:
		control = mfdcr(DCRN_DMACR3);
		control &= ~DMA_CE_ENABLE;
		mtdcr(DCRN_DMACR3, control);
		break;
#if NR_DMACS > 1
	case 4:
		control = mfdcr(DCRN_DMACR4);
		control &= ~DMA_CE_ENABLE;
		mtdcr(DCRN_DMACR4, control);
		break;
	case 5:
		control = mfdcr(DCRN_DMACR5);
		control &= ~DMA_CE_ENABLE;
		mtdcr(DCRN_DMACR5, control);
		break;
	case 6:
		control = mfdcr(DCRN_DMACR6);
		control &= ~DMA_CE_ENABLE;
		mtdcr(DCRN_DMACR6, control);
		break;
	case 7:
		control = mfdcr(DCRN_DMACR7);
		control &= ~DMA_CE_ENABLE;
		mtdcr(DCRN_DMACR7, control);
		break;
#endif /* NR_DMACS */
	default:
		printk("%s: bad channel: %d\n", __FUNCTION__, dmanr);
	}

	p_dma_ch->in_use = 0;
}

/*
 * Sets the dma mode for single DMA transfers only.
 * For scatter/gather transfers, the mode is passed to the
 * alloc_dma_handle() function as one of the parameters.
 *
 * The mode is simply saved and used later.  This allows
 * the driver to call set_dma_mode() and set_dma_addr() in
 * any order.
 *
 * Valid mode values are:
 *
 * DMA_MODE_READ          peripheral to memory 
 * DMA_MODE_WRITE         memory to peripheral
 * DMA_MODE_MM            memory to memory
 * DMA_MODE_MM_DEVATSRC   device-paced memory to memory, device at src
 * DMA_MODE_MM_DEVATDST   device-paced memory to memory, device at dst
 */
int
ppc4xx_set_dma_mode(unsigned int dmanr, unsigned int mode)
{
	ppc_dma_ch_t *p_dma_ch = dma_channels[dmanr];

	if (dmanr >= MAX_PPC4xx_DMA_CHANNELS) {
		printk("%s: bad channel 0x%x\n", __FUNCTION__, dmanr);
		return DMA_STATUS_BAD_CHANNEL;
	}

	p_dma_ch->mode = mode;

	return DMA_STATUS_GOOD;
}

/*
 * Sets the DMA Count register. Note that 'count' is in bytes.
 * However, the DMA Count register counts the number of "transfers",
 * where each transfer is equal to the bus width.  Thus, count
 * MUST be a multiple of the bus width.
 */
void
ppc4xx_set_dma_count(unsigned int dmanr, unsigned int count)
{
	ppc_dma_ch_t *p_dma_ch = dma_channels[dmanr];
	unsigned int ctc = p_dma_ch->ctc;

	if (dmanr >= MAX_PPC4xx_DMA_CHANNELS) {
		printk("%s: bad channel: %d\n", __FUNCTION__, dmanr);
		return;
	}

#ifdef DEBUG_4xxDMA
	{
		int error = 0;
		switch (p_dma_ch->pwidth) {
		case PW_8:
			break;
		case PW_16:
			if (count & 0x1)
				error = 1;
			break;
		case PW_32:
			if (count & 0x3)
				error = 1;
			break;
		case PW_64:
			if (count & 0x7)
				error = 1;
			break;
		case PW_128:
			if (count & 0xf)
				error = 1;
			break;
		default:
			printk("%s: invalid bus width: 0x%x\n",
			       __FUNCTION__, 
			       p_dma_ch->pwidth);
			return;
		}
		if (error)
			printk
			    ("Warning: %s count 0x%x bus width %d\n",
			     __FUNCTION__, 
			     count, p_dma_ch->pwidth);
	}
#endif

	count = count >> p_dma_ch->shift;
	ctc = ctc & ~DMA_CT_TC_MASK;
	ctc |= SET_DMA_CT_TC(count);
	switch (dmanr) {
	case 0:
		mtdcr(DCRN_DMACT0, ctc);
		break;
	case 1:
		mtdcr(DCRN_DMACT1, ctc);
		break;
	case 2:
		mtdcr(DCRN_DMACT2, ctc);
		break;
	case 3:
		mtdcr(DCRN_DMACT3, ctc);
		break;
#if NR_DMACS > 1
	case 4:
		mtdcr(DCRN_DMACT4, ctc);
		break;
	case 5:
		mtdcr(DCRN_DMACT5, ctc);
		break;
	case 6:
		mtdcr(DCRN_DMACT6, ctc);
		break;
	case 7:
		mtdcr(DCRN_DMACT7, ctc);
		break;
#endif /* NR_DMACS */
	default:
		printk("%s: bad channel: %d\n", __FUNCTION__, dmanr);
	}
}

/*
 *   Returns the number of bytes left to be transfered.
 *   After a DMA transfer, this should return zero.
 *   Reading this while a DMA transfer is still in progress will return
 *   unpredictable results.
 */ 
int
ppc4xx_get_dma_residue(unsigned int dmanr)
{
	unsigned int count;
	ppc_dma_ch_t *p_dma_ch = dma_channels[dmanr];

	switch (dmanr) {
	case 0:
		count = mfdcr(DCRN_DMACT0) & DMA_CT_TC_MASK;
		break;
	case 1:
		count = mfdcr(DCRN_DMACT1) & DMA_CT_TC_MASK;
		break;
	case 2:
		count = mfdcr(DCRN_DMACT2) & DMA_CT_TC_MASK;
		break;
	case 3:
		count = mfdcr(DCRN_DMACT3) & DMA_CT_TC_MASK;
		break;
#if NR_DMACS > 1
	case 4:
		count = mfdcr(DCRN_DMACT4) & DMA_CT_TC_MASK;
		break;
	case 5:
		count = mfdcr(DCRN_DMACT5) & DMA_CT_TC_MASK;
		break;
	case 6:
		count = mfdcr(DCRN_DMACT6) & DMA_CT_TC_MASK;
		break;
	case 7:
		count = mfdcr(DCRN_DMACT7) & DMA_CT_TC_MASK;
		break;
#endif /* NR_DMACS */
	default:
		printk("%s: bad channel: %d\n", __FUNCTION__, dmanr);
		return 0;
	}

	return (count << p_dma_ch->shift);
}

/*
 * Sets the DMA address for a memory to peripheral or peripheral
 * to memory transfer.  The address is just saved in the channel
 * structure for now and used later in enable_dma().
 */
void
ppc4xx_set_dma_addr(unsigned int dmanr, phys_addr_t addr)
{
	ppc_dma_ch_t *p_dma_ch = dma_channels[dmanr];

	if (dmanr >= MAX_PPC4xx_DMA_CHANNELS) {
		printk("%s: bad channel: %d\n", __FUNCTION__, dmanr);
		return;
	}

#ifdef DEBUG_4xxDMA
	{
		int error = 0;
		switch (p_dma_ch->pwidth) {
		case PW_8:
			break;
		case PW_16:
			if ((unsigned) addr & 0x1)
				error = 1;
			break;
		case PW_32:
			if ((unsigned) addr & 0x3)
				error = 1;
			break;
		case PW_64:
			if ((unsigned) addr & 0x7)
				error = 1;
			break;
		default:
			printk("%s: invalid bus width: 0x%x\n",
			       __FUNCTION__, 
			       p_dma_ch->pwidth);
			return;
		}
		if (error)
			printk("Warning: %s addr 0x%llxx bus width %d\n",
			       __FUNCTION__, 
			       addr, p_dma_ch->pwidth);
	}
#endif

	/* save dma address and program it later after we know the xfer mode */
	p_dma_ch->addr = addr;
}

/*
 * Sets both DMA addresses for a memory to memory transfer. 
 * For memory to peripheral or peripheral to memory transfers
 * the function set_dma_addr() should be used instead.
 */
void
ppc4xx_set_dma_addr2(unsigned int dmanr, phys_addr_t src_dma_addr,
		     phys_addr_t dst_dma_addr)
{
#ifdef DEBUG_4xxDMA
	{
		ppc_dma_ch_t *p_dma_ch = dma_channels[dmanr];
		int error = 0;
		switch (p_dma_ch->pwidth) {
			case PW_8:
				break;
			case PW_16:
				if (((unsigned) src_dma_addr & 0x1) ||
						((unsigned) dst_dma_addr & 0x1)
				   )
					error = 1;
				break;
			case PW_32:
				if (((unsigned) src_dma_addr & 0x3) ||
						((unsigned) dst_dma_addr & 0x3)
				   )
					error = 1;
				break;
			case PW_64:
				if (((unsigned) src_dma_addr & 0x7) ||
						((unsigned) dst_dma_addr & 0x7)
				   )
					error = 1;
				break;
			case PW_128:
				if (((unsigned) src_dma_addr & 0xf) ||
						((unsigned) dst_dma_addr & 0xf)
				   )
					error = 1;
				break;
			default:
				printk("%s: invalid bus width: 0x%x\n",
				       __FUNCTION__,
				       p_dma_ch->pwidth);
				return;
		}
		if (error)
			printk
				("Warning: %s src 0x%llxx dst 0x%llxx bus width %d\n",
				 __FUNCTION__,
				 src_dma_addr, dst_dma_addr, p_dma_ch->pwidth);
	}
#endif

	if (dmanr >= MAX_PPC4xx_DMA_CHANNELS) {
		printk("%s: bad channel: %d\n", __FUNCTION__, dmanr);
	}
	else {
		ppc4xx_set_src_addr(dmanr, src_dma_addr);
		ppc4xx_set_dst_addr(dmanr, dst_dma_addr);
	}
}

/*
 * Enables the channel interrupt.
 *
 * If performing a scatter/gatter transfer, this function 
 * MUST be called before calling alloc_dma_handle() and building
 * the sgl list.  Otherwise, interrupts will not be enabled, if 
 * they were previously disabled.
 */
int
ppc4xx_enable_dma_interrupt(unsigned int dmanr)
{
	unsigned int control;
	unsigned int ctc;
	ppc_dma_ch_t *p_dma_ch = dma_channels[dmanr];

	p_dma_ch->int_enable = 1;
	p_dma_ch->ctc |= (DMA_CT_TCIE | DMA_CT_ETIE | DMA_CT_EIE);

	/*
	 * Clear the CS, TS, RI bits for the channel from DMASR.
	 */
	ppc4xx_clear_dma_status(dmanr);

	switch (dmanr) {
	case 0:
		ctc = mfdcr(DCRN_DMACT0);
		ctc |= DMA_CT_TCIE | DMA_CT_ETIE | DMA_CT_EIE;
		mtdcr(DCRN_DMACT0, ctc);

		mb();		/* barrier */

		control = mfdcr(DCRN_DMACR0);
		control |= DMA_CIE_ENABLE;	/* Channel Interrupt Enable */
		mtdcr(DCRN_DMACR0, control);

		ctc = mfdcr(DCRN_DMACT0);
		ctc |= DMA_CT_TCIE | DMA_CT_ETIE | DMA_CT_EIE;
		mtdcr(DCRN_DMACT0, ctc);
		break;
	case 1:
		ctc = mfdcr(DCRN_DMACT1);
		ctc |= DMA_CT_TCIE | DMA_CT_ETIE | DMA_CT_EIE;
		mtdcr(DCRN_DMACT1, ctc);

		mb();		/* barrier */

		control = mfdcr(DCRN_DMACR1);
		control |= DMA_CIE_ENABLE;
		mtdcr(DCRN_DMACR1, control);

		ctc = mfdcr(DCRN_DMACT1);
		ctc |= DMA_CT_TCIE | DMA_CT_ETIE | DMA_CT_EIE;
		mtdcr(DCRN_DMACT1, ctc);
		break;
	case 2:
		ctc = mfdcr(DCRN_DMACT2);
		ctc |= DMA_CT_TCIE | DMA_CT_ETIE | DMA_CT_EIE;
		mtdcr(DCRN_DMACT2, ctc);

		mb();		/* barrier */

		control = mfdcr(DCRN_DMACR2);
		control |= DMA_CIE_ENABLE;
		mtdcr(DCRN_DMACR2, control);

		ctc = mfdcr(DCRN_DMACT2);
		ctc |= DMA_CT_TCIE | DMA_CT_ETIE | DMA_CT_EIE;
		mtdcr(DCRN_DMACT2, ctc);
		break;
	case 3:
		ctc = mfdcr(DCRN_DMACT3);
		ctc |= DMA_CT_TCIE | DMA_CT_ETIE | DMA_CT_EIE;
		mtdcr(DCRN_DMACT3, ctc);

		mb();		/* barrier */

		control = mfdcr(DCRN_DMACR3);
		control |= DMA_CIE_ENABLE;
		mtdcr(DCRN_DMACR3, control);

		ctc = mfdcr(DCRN_DMACT3);
		ctc |= DMA_CT_TCIE | DMA_CT_ETIE | DMA_CT_EIE;
		mtdcr(DCRN_DMACT3, ctc);
		break;
#if NR_DMACS > 1
	case 4:
		ctc = mfdcr(DCRN_DMACT4);
		ctc |= DMA_CT_TCIE | DMA_CT_ETIE | DMA_CT_EIE;
		mtdcr(DCRN_DMACT4, ctc);

		mb();		/* barrier */

		control = mfdcr(DCRN_DMACR4);
		control |= DMA_CIE_ENABLE;
		mtdcr(DCRN_DMACR4, control);

		ctc = mfdcr(DCRN_DMACT4);
		ctc |= DMA_CT_TCIE | DMA_CT_ETIE | DMA_CT_EIE;
		mtdcr(DCRN_DMACT4, ctc);
		break;
	case 5:
		ctc = mfdcr(DCRN_DMACT5);
		ctc |= DMA_CT_TCIE | DMA_CT_ETIE | DMA_CT_EIE;
		mtdcr(DCRN_DMACT5, ctc);

		mb();		/* barrier */

		control = mfdcr(DCRN_DMACR5);
		control |= DMA_CIE_ENABLE;
		mtdcr(DCRN_DMACR5, control);

		ctc = mfdcr(DCRN_DMACT5);
		ctc |= DMA_CT_TCIE | DMA_CT_ETIE | DMA_CT_EIE;
		mtdcr(DCRN_DMACT5, ctc);
		break;
	case 6:
		ctc = mfdcr(DCRN_DMACT6);
		ctc |= DMA_CT_TCIE | DMA_CT_ETIE | DMA_CT_EIE;
		mtdcr(DCRN_DMACT6, ctc);

		mb();		/* barrier */

		control = mfdcr(DCRN_DMACR6);
		control |= DMA_CIE_ENABLE;
		mtdcr(DCRN_DMACR6, control);

		ctc = mfdcr(DCRN_DMACT6);
		ctc |= DMA_CT_TCIE | DMA_CT_ETIE | DMA_CT_EIE;
		mtdcr(DCRN_DMACT6, ctc);
		break;
	case 7:
		ctc = mfdcr(DCRN_DMACT7);
		ctc |= DMA_CT_TCIE | DMA_CT_ETIE | DMA_CT_EIE;
		mtdcr(DCRN_DMACT7, ctc);

		mb();		/* barrier */

		control = mfdcr(DCRN_DMACR7);
		control |= DMA_CIE_ENABLE;
		mtdcr(DCRN_DMACR7, control);

		ctc = mfdcr(DCRN_DMACT7);
		ctc |= DMA_CT_TCIE | DMA_CT_ETIE | DMA_CT_EIE;
		mtdcr(DCRN_DMACT7, ctc);
		break;
#endif /* NR_DMACS */
	default:
		printk("%s: bad channel: %d\n", __FUNCTION__, dmanr);
		return DMA_STATUS_BAD_CHANNEL;
	}
	return DMA_STATUS_GOOD;
}

/*
 * Disables the channel interrupt.
 *
 * If performing a scatter/gatter transfer, this function 
 * MUST be called before calling alloc_dma_handle() and building
 * the sgl list.  Otherwise, interrupts will not be disabled, if 
 * they were previously enabled.
 */
int
ppc4xx_disable_dma_interrupt(unsigned int dmanr)
{
	unsigned int control, ctc;
	ppc_dma_ch_t *p_dma_ch = dma_channels[dmanr];

	p_dma_ch->int_enable = 0;
	p_dma_ch->ctc &= ~(DMA_CT_TCIE | DMA_CT_ETIE | DMA_CT_EIE);
	switch (dmanr) {
	case 0:
		control = mfdcr(DCRN_DMACR0);
		control &= ~DMA_CIE_ENABLE;	/* Channel Interrupt Enable */
		mtdcr(DCRN_DMACR0, control);

		ctc = mfdcr(DCRN_DMACT0);
		ctc &= ~(DMA_CT_TCIE | DMA_CT_ETIE | DMA_CT_EIE);
		mtdcr(DCRN_DMACT0, ctc);
		break;
	case 1:
		control = mfdcr(DCRN_DMACR1);
		control &= ~DMA_CIE_ENABLE;
		mtdcr(DCRN_DMACR1, control);

		ctc = mfdcr(DCRN_DMACT1);
		ctc &= ~(DMA_CT_TCIE | DMA_CT_ETIE | DMA_CT_EIE);
		mtdcr(DCRN_DMACT1, ctc);
		break;
	case 2:
		control = mfdcr(DCRN_DMACR2);
		control &= ~DMA_CIE_ENABLE;
		mtdcr(DCRN_DMACR2, control);

		ctc = mfdcr(DCRN_DMACT2);
		ctc &= ~(DMA_CT_TCIE | DMA_CT_ETIE | DMA_CT_EIE);
		mtdcr(DCRN_DMACT2, ctc);
		break;
	case 3:
		control = mfdcr(DCRN_DMACR3);
		control &= ~DMA_CIE_ENABLE;
		mtdcr(DCRN_DMACR3, control);

		ctc = mfdcr(DCRN_DMACT3);
		ctc &= ~(DMA_CT_TCIE | DMA_CT_ETIE | DMA_CT_EIE);
		mtdcr(DCRN_DMACT3, ctc);
		break;
#if NR_DMACS > 1
	case 4:
		control = mfdcr(DCRN_DMACR4);
		control &= ~DMA_CIE_ENABLE;
		mtdcr(DCRN_DMACR4, control);

		ctc = mfdcr(DCRN_DMACT4);
		ctc &= ~(DMA_CT_TCIE | DMA_CT_ETIE | DMA_CT_EIE);
		mtdcr(DCRN_DMACT4, ctc);
		break;
	case 5:
		control = mfdcr(DCRN_DMACR5);
		control &= ~DMA_CIE_ENABLE;
		mtdcr(DCRN_DMACR5, control);

		ctc = mfdcr(DCRN_DMACT5);
		ctc &= ~(DMA_CT_TCIE | DMA_CT_ETIE | DMA_CT_EIE);
		mtdcr(DCRN_DMACT5, ctc);
		break;
	case 6:
		control = mfdcr(DCRN_DMACR6);
		control &= ~DMA_CIE_ENABLE;
		mtdcr(DCRN_DMACR6, control);

		ctc = mfdcr(DCRN_DMACT6);
		ctc &= ~(DMA_CT_TCIE | DMA_CT_ETIE | DMA_CT_EIE);
		mtdcr(DCRN_DMACT6, ctc);
		break;
	case 7:
		control = mfdcr(DCRN_DMACR7);
		control &= ~DMA_CIE_ENABLE;
		mtdcr(DCRN_DMACR7, control);

		ctc = mfdcr(DCRN_DMACT7);
		ctc &= ~(DMA_CT_TCIE | DMA_CT_ETIE | DMA_CT_EIE);
		mtdcr(DCRN_DMACT7, ctc);
		break;
#endif /* NR_DMACS */
	default:
		printk("%s: bad channel: %d\n", __FUNCTION__, dmanr);
		return DMA_STATUS_BAD_CHANNEL;
	}
	return DMA_STATUS_GOOD;
}

/*
 * Configures a DMA channel, including the peripheral bus width, if a
 * peripheral is attached to the channel, the polarity of the DMAReq and
 * DMAAck signals, etc.  This information should really be setup by the boot 
 * code, since most likely the configuration won't change dynamically.
 * If the kernel has to call this function, it's recommended that it's
 * called from platform specific init code.  The driver should not need to 
 * call this function.
 */
int
ppc4xx_init_dma_channel(unsigned int dmanr, ppc_dma_ch_t * p_init)
{
	unsigned int polarity;
	uint32_t control = 0, ctc = 0;
	ppc_dma_ch_t *p_dma_ch = dma_channels[dmanr];

	DMA_MODE_READ = (unsigned long) DMA_TD;	/* Peripheral to Memory */
	DMA_MODE_WRITE = 0;	/* Memory to Peripheral */

	if (!p_init) {
		printk("%s: NULL p_init\n", __FUNCTION__);
		return DMA_STATUS_NULL_POINTER;
	}

	if (dmanr >= MAX_PPC4xx_DMA_CHANNELS) {
		printk("%s: bad channel %d\n", __FUNCTION__, dmanr);
		return DMA_STATUS_BAD_CHANNEL;
	}

	switch (CHNUM_TO_DMACNUM(dmanr)){
	case 0:
#if DCRN_DMA0_POL > 0
		polarity = mfdcr(DCRN_DMA0_POL);
#else
		polarity = 0;
#endif
		break;
#if NR_DMACS > 1
	case 1:
#if DCRN_DMA1_POL > 0
		polarity = mfdcr(DCRN_DMA1_POL);
#else
		polarity = 0;
#endif
		break;
#endif
	}

	/* Setup the control register based on the values passed to
	 * us in p_init.  Then, over-write the control register with this
	 * new value.
	 */
	control |= SET_DMA_CONTROL;

	/* Setup count & control register from p_init */
	ctc |= SET_DMA_CTC;

	switch (dmanr) {
	case 0:
		/* clear all polarity signals and then "or" in new signal levels */
		polarity &= ~GET_DMA_POLARITY(0);
		polarity |= p_dma_ch->polarity;
#if DCRN_DMA0_POL > 0
		mtdcr(DCRN_DMA0_POL, polarity);
#endif
		mtdcr(DCRN_DMACR0, control);
		break;
	case 1:
		polarity &= ~GET_DMA_POLARITY(1);
		polarity |= p_dma_ch->polarity;
#if DCRN_DMA0_POL > 0
		mtdcr(DCRN_DMA0_POL, polarity);
#endif
		mtdcr(DCRN_DMACR1, control);
		break;
	case 2:
		polarity &= ~GET_DMA_POLARITY(2);
		polarity |= p_dma_ch->polarity;
#if DCRN_DMA0_POL > 0
		mtdcr(DCRN_DMA0_POL, polarity);
#endif
		mtdcr(DCRN_DMACR2, control);
		break;
	case 3:
		polarity &= ~GET_DMA_POLARITY(3);
		polarity |= p_dma_ch->polarity;
#if DCRN_DMA0_POL > 0
		mtdcr(DCRN_DMA0_POL, polarity);
#endif
		mtdcr(DCRN_DMACR3, control);
		break;
#if NR_DMACS > 1
	case 4:
		/* clear all polarity signals and then "or" in new signal levels */
		polarity &= ~GET_DMA_POLARITY(0);
		polarity |= p_dma_ch->polarity;
#if DCRN_DMA1_POL > 0
		mtdcr(DCRN_DMA1_POL, polarity);
#endif
		mtdcr(DCRN_DMACR4, control);
		break;
	case 5:
		polarity &= ~GET_DMA_POLARITY(1);
		polarity |= p_dma_ch->polarity;
#if DCRN_DMA1_POL > 0
		mtdcr(DCRN_DMA1_POL, polarity);
#endif
		mtdcr(DCRN_DMACR5, control);
		break;
	case 6:
		polarity &= ~GET_DMA_POLARITY(2);
		polarity |= p_dma_ch->polarity;
#if DCRN_DMA1_POL > 0
		mtdcr(DCRN_DMA1_POL, polarity);
#endif
		mtdcr(DCRN_DMACR6, control);
		break;
	case 7:
		polarity &= ~GET_DMA_POLARITY(3);
		polarity |= p_dma_ch->polarity;
#if DCRN_DMA1_POL > 0
		mtdcr(DCRN_DMA1_POL, polarity);
#endif
		mtdcr(DCRN_DMACR7, control);
		break;
#endif /* NR_DMACS */
	default:
		return DMA_STATUS_BAD_CHANNEL;
	}

	/* save these values in our dma channel structure */
	memcpy(p_dma_ch, p_init, sizeof (ppc_dma_ch_t));
	/*
	 * The peripheral width values written in the control register are:
	 *   PW_8                 0
	 *   PW_16                1
	 *   PW_32                2
	 *   PW_64                3
	 *
	 *   Since the DMA count register takes the number of "transfers",
	 *   we need to divide the count sent to us in certain
	 *   functions by the appropriate number.  It so happens that our
	 *   right shift value is equal to the peripheral width value.
	 */
	p_dma_ch->shift = p_init->pwidth;

	/*
	 * Save the control word for easy access.
	 */
	p_dma_ch->control = control;
	p_dma_ch->ctc = ctc;

	ppc4xx_clear_dma_status(dmanr);

#ifdef DEBUG_4xxDMA
	ppc4xx_show_dmach(dmanr);
#endif

	return DMA_STATUS_GOOD;
}

/*
 * This function returns the channel configuration.
 */
int
ppc4xx_get_channel_config(unsigned int dmanr, ppc_dma_ch_t * p_dma_ch)
{
	unsigned int polarity;
	unsigned int control, ctc;

	if (!p_dma_ch) {
		printk("%s: NULL p_dma_ch\n", __FUNCTION__);
		return DMA_STATUS_NULL_POINTER;
	}

	switch (CHNUM_TO_DMACNUM(dmanr)){
	case 0:
#if DCRN_DMA0_POL > 0
		polarity = mfdcr(DCRN_DMA0_POL);
#else
		polarity = 0;
#endif
		break;
#if NR_DMACS > 1
	case 1:
#if DCRN_DMA1_POL > 0
		polarity = mfdcr(DCRN_DMA1_POL);
#else
		polarity = 0;
#endif
		break;
#endif
	}

	switch (dmanr) {
	case 0:
		p_dma_ch->polarity = polarity & GET_DMA_POLARITY(0);
		control = mfdcr(DCRN_DMACR0);
		ctc = mfdcr(DCRN_DMACT0);
		break;
	case 1:
		p_dma_ch->polarity = polarity & GET_DMA_POLARITY(1);
		control = mfdcr(DCRN_DMACR1);
		ctc = mfdcr(DCRN_DMACT1);
		break;
	case 2:
		p_dma_ch->polarity = polarity & GET_DMA_POLARITY(2);
		control = mfdcr(DCRN_DMACR2);
		ctc = mfdcr(DCRN_DMACT2);
		break;
	case 3:
		p_dma_ch->polarity = polarity & GET_DMA_POLARITY(3);
		control = mfdcr(DCRN_DMACR3);
		ctc = mfdcr(DCRN_DMACT3);
		break;
#if NR_DMACS > 1
	case 4:
		p_dma_ch->polarity = polarity & GET_DMA_POLARITY(0);
		control = mfdcr(DCRN_DMACR4);
		ctc = mfdcr(DCRN_DMACT4);
		break;
	case 5:
		p_dma_ch->polarity = polarity & GET_DMA_POLARITY(1);
		control = mfdcr(DCRN_DMACR5);
		ctc = mfdcr(DCRN_DMACT5);
		break;
	case 6:
		p_dma_ch->polarity = polarity & GET_DMA_POLARITY(2);
		control = mfdcr(DCRN_DMACR6);
		ctc = mfdcr(DCRN_DMACT6);
		break;
	case 7:
		p_dma_ch->polarity = polarity & GET_DMA_POLARITY(3);
		control = mfdcr(DCRN_DMACR7);
		ctc = mfdcr(DCRN_DMACT7);
		break;
#endif /* NR_DMACS */
	default:
		return DMA_STATUS_BAD_CHANNEL;
	}

	p_dma_ch->cp = GET_DMA_PRIORITY(control);
	p_dma_ch->pwidth = GET_DMA_PW(control);
	p_dma_ch->psc = GET_DMA_PSC(control);
	p_dma_ch->pwc = GET_DMA_PWC(control);
	p_dma_ch->phc = GET_DMA_PHC(control);
	p_dma_ch->ce = GET_DMA_CE_ENABLE(control);
	p_dma_ch->int_enable = GET_DMA_CIE_ENABLE(control);
	p_dma_ch->shift = GET_DMA_PW(control);

#ifdef CONFIG_PPC4xx_EDMA
	p_dma_ch->pf = GET_DMA_PREFETCH(control);
#else
	p_dma_ch->ch_enable = GET_DMA_CH(control);
	p_dma_ch->ece_enable = GET_DMA_ECE(control);
	p_dma_ch->tcd_disable = GET_DMA_TCD(control);
#endif

	p_dma_ch->sid = GET_DMA_CT_SID(ctc);
	p_dma_ch->bten = GET_DMA_CT_BTEN(ctc);
	p_dma_ch->bsiz = GET_DMA_CT_BSIZ(ctc);
	p_dma_ch->pce = GET_DMA_CT_PCE(ctc);

	return DMA_STATUS_GOOD;
}

/*
 * Sets the priority for the DMA channel dmanr.
 * Since this is setup by the hardware init function, this function 
 * can be used to dynamically change the priority of a channel.
 *
 * Acceptable priorities:
 *
 * PRIORITY_LOW
 * PRIORITY_MID_LOW
 * PRIORITY_MID_HIGH
 * PRIORITY_HIGH
 *
 */
int
ppc4xx_set_channel_priority(unsigned int dmanr, unsigned int priority)
{
	unsigned int control;

	if ((priority != PRIORITY_LOW) &&
	    (priority != PRIORITY_MID_LOW) &&
	    (priority != PRIORITY_MID_HIGH) && (priority != PRIORITY_HIGH)) {
		printk("%s: bad priority: 0x%x\n", __FUNCTION__, priority);
	}

	switch (dmanr) {
	case 0:
		control = mfdcr(DCRN_DMACR0);
		control &= ~DMA_PRIORITY_MASK;
		control |= SET_DMA_PRIORITY(priority);
		mtdcr(DCRN_DMACR0, control);
		break;
	case 1:
		control = mfdcr(DCRN_DMACR1);
		control &= ~DMA_PRIORITY_MASK;
		control |= SET_DMA_PRIORITY(priority);
		mtdcr(DCRN_DMACR1, control);
		break;
	case 2:
		control = mfdcr(DCRN_DMACR2);
		control &= ~DMA_PRIORITY_MASK;
		control |= SET_DMA_PRIORITY(priority);
		mtdcr(DCRN_DMACR2, control);
		break;
	case 3:
		control = mfdcr(DCRN_DMACR3);
		control &= ~DMA_PRIORITY_MASK;
		control |= SET_DMA_PRIORITY(priority);
		mtdcr(DCRN_DMACR3, control);
		break;
#if NR_DMACS > 1
	case 4:
		control = mfdcr(DCRN_DMACR4);
		control &= ~DMA_PRIORITY_MASK;
		control |= SET_DMA_PRIORITY(priority);
		mtdcr(DCRN_DMACR4, control);
		break;
	case 5:
		control = mfdcr(DCRN_DMACR5);
		control &= ~DMA_PRIORITY_MASK;
		control |= SET_DMA_PRIORITY(priority);
		mtdcr(DCRN_DMACR5, control);
		break;
	case 6:
		control = mfdcr(DCRN_DMACR6);
		control &= ~DMA_PRIORITY_MASK;
		control |= SET_DMA_PRIORITY(priority);
		mtdcr(DCRN_DMACR6, control);
		break;
	case 7:
		control = mfdcr(DCRN_DMACR7);
		control &= ~DMA_PRIORITY_MASK;
		control |= SET_DMA_PRIORITY(priority);
		mtdcr(DCRN_DMACR7, control);
		break;
#endif /* NR_DMACS */
	default:
		printk("%s: bad channel: %d\n", __FUNCTION__, dmanr);
		return DMA_STATUS_BAD_CHANNEL;
	}
	return DMA_STATUS_GOOD;
}

/*
 * Returns the width of the peripheral attached to this channel. This assumes
 * that someone who knows the hardware configuration, boot code or some other 
 * init code, already set the width.
 *
 * The return value is one of:
 *   PW_8
 *   PW_16
 *   PW_32
 *   PW_64
 *
 *   The function returns 0 on error.
 */
unsigned int
ppc4xx_get_peripheral_width(unsigned int dmanr)
{
	unsigned int control;

	switch (dmanr) {
	case 0:
		control = mfdcr(DCRN_DMACR0);
		break;
	case 1:
		control = mfdcr(DCRN_DMACR1);
		break;
	case 2:
		control = mfdcr(DCRN_DMACR2);
		break;
	case 3:
		control = mfdcr(DCRN_DMACR3);
		break;
#if NR_DMACS > 1
	case 4:
		control = mfdcr(DCRN_DMACR4);
		break;
	case 5:
		control = mfdcr(DCRN_DMACR5);
		break;
	case 6:
		control = mfdcr(DCRN_DMACR6);
		break;
	case 7:
		control = mfdcr(DCRN_DMACR7);
		break;
#endif /* NR_DMACS */
	default:
		printk("%s: bad channel: %d\n", __FUNCTION__, dmanr);
		return 0;
	}
	return (GET_DMA_PW(control));
}


EXPORT_SYMBOL(ppc4xx_request_dma);
EXPORT_SYMBOL(ppc4xx_request_dma_channel);
EXPORT_SYMBOL(ppc4xx_free_dma);
EXPORT_SYMBOL(ppc4xx_get_dma_irq);
EXPORT_SYMBOL(ppc4xx_init_dma_channel);
EXPORT_SYMBOL(ppc4xx_get_channel_config);
EXPORT_SYMBOL(ppc4xx_set_channel_priority);
EXPORT_SYMBOL(ppc4xx_get_peripheral_width);
EXPORT_SYMBOL(dma_channels);
EXPORT_SYMBOL(ppc4xx_set_src_addr);
EXPORT_SYMBOL(ppc4xx_set_dst_addr);
EXPORT_SYMBOL(ppc4xx_set_dma_addr);
EXPORT_SYMBOL(ppc4xx_set_dma_addr2);
EXPORT_SYMBOL(ppc4xx_enable_dma);
EXPORT_SYMBOL(ppc4xx_disable_dma);
EXPORT_SYMBOL(ppc4xx_set_dma_mode);
EXPORT_SYMBOL(ppc4xx_set_dma_count);
EXPORT_SYMBOL(ppc4xx_get_dma_residue);
EXPORT_SYMBOL(ppc4xx_enable_dma_interrupt);
EXPORT_SYMBOL(ppc4xx_disable_dma_interrupt);
EXPORT_SYMBOL(ppc4xx_get_dma_status);
EXPORT_SYMBOL(ppc4xx_clear_dma_status);
EXPORT_SYMBOL(ppc4xx_show_status);
EXPORT_SYMBOL(ppc4xx_show_dmach);
