/*
 *  PPC4XX GPIO driver
 *
 *  Copyright 2003 Sony Corporation.
 *
 *
 * With non GPL files, use following license:
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * Otherwise with GPL files, use following license:
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/irq.h>
#include <linux/proc_fs.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/ibm4xx.h>
#include <asm/ppc4xx_pic.h>
#include <asm/ppc4xx_gpio.h>
#include <asm/machdep.h>
#include <platforms/ibm_ocp.h>

#undef PPC4xx_GPIO_DEBUG

#define MODNAME                 "ppc4xx_gpio"
#define MSG_HEAD                "ppc4xx_gpio : "

static struct ppc4xx_gpio_t	gpio[GPIO_NUMS];
static ocp_gpio_t		*gpio_base[GPIO_NUMS];


static inline int
__portbit_to_irq(int port, int bit)
{
	switch (port) {
	case GPIO_PORT_0:
		return PPC4xx_IRQ_GPIO0(bit);
#if GPIO_NUMS > 1
	case GPIO_PORT_1:
		return PPC4xx_IRQ_GPIO1(bit);
#endif
	}
}

#define PORT_BASE(port)         (gpio_base[(port)])

/* Bit set by Big-Endian */
#define BIT(bitnum)		((u_int32_t)(1 << (31 - (bitnum))))

#define FOR_EACH_BIT(bit, bitmask) \
	for (bit = GPIO_BITS_PER_PORT - 1; bit >= 0 ; bit--) \
		if (bitmask & BIT(bit))

static int
check_port(int port)
{
	if ((port < GPIO_PORT_0) || (port > (GPIO_NUMS - 1))) {
		printk(KERN_ERR MSG_HEAD "invalid port %d\n", port);
		return -EINVAL;
	}
	return 0;
}

static int
check_port_bitnum(int port, int bitnum)
{
	if ((port < GPIO_PORT_0) || (port > (GPIO_NUMS - 1))) {
		printk(KERN_ERR MSG_HEAD "invalid port %d\n", port);
		return -EINVAL;
	}

	if ((bitnum < 0) || bitnum > (GPIO_BITS_PER_PORT - 1)) {
		printk(KERN_ERR MSG_HEAD "invalid port %d\n", port);
		return -EINVAL;
	}

	return 0;
}

/* 
 * These routines check specified bitmask at given port are available
 * by GPIO pin multiplex.
 */

static int
check_port_gpio(int port, u_int32_t bitmask)
{
	if (gpio[port].no_gpio & bitmask) {
		printk(KERN_ERR MSG_HEAD 
		       "GPIO %c pins by mask=0x%08x cannot be used for GPIO\n",
		       PORT_NAME(port), bitmask);
		return -EINVAL;
	}
	return 0;
}

static int
check_port_gpio_out(int port, u_int32_t bitmask)
{
	if (gpio[port].no_out & bitmask) {
		printk(KERN_ERR MSG_HEAD 
		       "GPIO %c pins by mask=0x%08x cannot be used for GPIO output\n",
		       PORT_NAME(port), bitmask);
		return -EINVAL;
	}
	return 0;
}

#ifdef CONFIG_CXD4707
static int
check_port_gpio_altout(int port, u_int32_t bitmask, int altfunc)
{
	int i;
	i = (altfunc & 0x3) - 1; /* array offset */
	if (gpio[port].no_altout[i] & bitmask) {
		printk(KERN_ERR MSG_HEAD 
		       "GPIO %c pins by mask=0x%08x cannot be used for GPIO alt_output%d\n",
		       PORT_NAME(port), bitmask, altfunc);
		return -EINVAL;
	}
	return 0;
}
#endif

static int
check_port_gpio_hiz(int port, u_int32_t bitmask)
{
	if (gpio[port].no_hiz & bitmask) {
		printk(KERN_ERR MSG_HEAD 
		       "GPIO %c pins by mask=0x%08x cannot be used for GPIO hi-z\n",
		       PORT_NAME(port), bitmask);
		return -EINVAL;
	}
	return 0;
}

static int
check_port_gpio_in(int port, u_int32_t bitmask)
{
	if (gpio[port].no_in & bitmask) {
		printk(KERN_ERR MSG_HEAD 
		       "GPIO %c pins by mask=0x%08x cannot be used for GPIO input\n",
		       PORT_NAME(port), bitmask);
		return -EINVAL;
	}
	return 0;
}

#ifdef CONFIG_CXD4707
static int
check_port_gpio_altin(int port, u_int32_t bitmask, int altfunc)
{
	int i;
	i = (altfunc & 0x3) - 1; /* array offset */
	if (gpio[port].no_altin[i] & bitmask) {
		printk(KERN_ERR MSG_HEAD 
		       "GPIO %c pins by mask=0x%08x cannot be used for GPIO alt_input%d\n",
		       PORT_NAME(port), bitmask, altfunc);
		return -EINVAL;
	}
	return 0;
}
#endif

static int
check_port_gpio_int(int port, u_int32_t bitmask)
{
	if (gpio[port].no_in & bitmask) {
		printk(KERN_ERR MSG_HEAD 
		       "GPIO %c pins by mask=0x%08x cannot be used for GPIO interrupt\n",
		       PORT_NAME(port), bitmask);
		return -EINVAL;
	}
	return 0;
}

/* register GPIO (per bit) */
int ppc4xx_gpio_register_bit(int port, int bitnum, const char *name)
{
	int ret = 0;
	unsigned long flags;

	if ((ret = check_port_bitnum(port, bitnum)) < 0)
		return ret;

	if ((ret = check_port_gpio(port, BIT(bitnum))) < 0)
		return ret;

	local_irq_save(flags);

	if (gpio[port].used & BIT(bitnum)) {
		printk(KERN_ERR MSG_HEAD 
		       "GPIO %c bit %d is already reserved\n",
		       PORT_NAME(port), bitnum);
		ret = -EBUSY;
	} else {
		gpio[port].used |= BIT(bitnum);
		strncpy(gpio[port].bit[bitnum].name, name, MAX_NAME_SIZE);
	}
	local_irq_restore(flags);

	return ret;
}

/* unregister GPIO (per bit) */
int ppc4xx_gpio_unregister_bit(int port, int bitnum)
{
	int ret = 0;
	unsigned long flags;

	if ((ret = check_port_bitnum(port, bitnum)) < 0)
		return ret;

	if ((ret = check_port_gpio(port, BIT(bitnum))) < 0)
		return ret;

	local_irq_save(flags);

	if (gpio[port].used & BIT(bitnum)) {
		gpio[port].used &= ~BIT(bitnum);
		gpio[port].inout &= ~BIT(bitnum);
		gpio[port].alt &= ~BIT(bitnum);
		ppc4xx_gpio_intr_mask_bit(port, bitnum);
		memset(gpio[port].bit[bitnum].name, 0, MAX_NAME_SIZE);
	} else {
		printk(KERN_ERR MSG_HEAD
		       "GPIO %c bit %d is not reserved\n",
		       PORT_NAME(port), bitnum);
		ret = -EBUSY;
	}

	local_irq_restore(flags);

	return ret;
}

/* register GPIO (via bitmask) */
int ppc4xx_gpio_register(int port, u_int32_t bitmask, const char *name)
{
	int ret = 0;
	int bit;
	unsigned long flags;

	if ((ret = check_port(port)) < 0)
		return ret;

	if ((ret = check_port_gpio(port, bitmask)) < 0)
		return ret;

	local_irq_save(flags);

	if (gpio[port].used & bitmask) {
		printk(KERN_ERR MSG_HEAD
		       "GPIO %c bitmask 0x%08x is reserved\n",
		       PORT_NAME(port), gpio[port].used & bitmask);
		ret = -EBUSY;
	} else {
		gpio[port].used |= bitmask;
		FOR_EACH_BIT(bit, bitmask) {
			strncpy(gpio[port].bit[bit].name,
				name, MAX_NAME_SIZE);
		}
	}

	local_irq_restore(flags);

	return ret;
}
 
int ppc4xx_gpio_unregister(int port, u_int32_t bitmask)
{
	int ret = 0;
	int bit;
	unsigned long flags;

	if ((ret = check_port(port)) < 0)
		return ret;

	if ((ret = check_port_gpio(port, bitmask)) < 0)
		return ret;

	local_irq_save(flags);

	if ((gpio[port].used & bitmask) == bitmask) {
		gpio[port].used &= ~bitmask;
		gpio[port].inout &= ~bitmask;
#ifdef CONFIG_CXD4707
		gpio[port].alt &= ~bitmask;
#endif
		ppc4xx_gpio_intr_mask(port, bitmask);
		FOR_EACH_BIT(bit, bitmask) {
			memset(gpio[port].bit[bit].name, 
			       0, MAX_NAME_SIZE);
		}
	} else {
		printk(KERN_ERR MSG_HEAD 
		       "GPIO %c bitmask 0x%08x is not reserved\n",
		       PORT_NAME(port), 
		       (~gpio[port].used) & bitmask);
		ret = -EBUSY;
	}

	local_irq_restore(flags);

	return ret;
}

/* unregister GPIOs by name */
int ppc4xx_gpio_unregister_by_name(const char *name)
{
	int port, bit;
	int ret = 0;

	for (port = GPIO_PORT_0; port < GPIO_NUMS; port++) {
		for (bit = 0; bit < GPIO_BITS_PER_PORT; bit++) {
			if (strncmp(gpio[port].bit[bit].name, name, MAX_NAME_SIZE) == 0)
				if ((ret = ppc4xx_gpio_unregister_bit(port, bit)) < 0)
					return ret;
		}
	} 

	return 0;
}

/* set input or output (per bit) */
int ppc4xx_gpio_set_mode_bit(int port, int bitnum, u_int32_t mode)
{
	int ret = 0;
	unsigned long flags;
	ocp_gpio_t *pgpio;

	if ((ret = check_port_bitnum(port, bitnum)) < 0)
		return ret;

	pgpio = PORT_BASE(port);

	local_irq_save(flags);

	if (mode & __MODE_HIZ) {
		if ((ret = check_port_gpio_hiz(port, BIT(bitnum))) < 0)
			goto err;

		pgpio->tcr &= ~BIT(bitnum);

		gpio[port].inout |= BIT(bitnum); /* HIZ is output */
	} else if (mode & __MODE_OUT) {
		if ((ret = check_port_gpio_out(port, BIT(bitnum))) < 0)
			goto err;

		pgpio->tcr &= ~BIT(bitnum); /* set hi-z before setting values*/
#ifdef CONFIG_CXD4707
		if (mode & __MODE_ALT_MASK) {
			if ((ret = check_port_gpio_altout(port, BIT(bitnum), 
							  GET_ALT_FUNC(mode))) < 0)
				goto err;

			pgpio->os &= ~((u_int64_t)0x3 << ((31 - bitnum) * 2));
			pgpio->os |= ((u_int64_t)GET_ALT_FUNC(mode) << ((31 - bitnum) * 2));
			gpio[port].alt |= BIT(bitnum);
		} else {
			pgpio->os &= ~((u_int64_t)0x3 << ((31 - bitnum) * 2));
			gpio[port].alt &= ~BIT(bitnum);
		}
#endif
		if (mode & __MODE_HIGH)
			pgpio->or |=  BIT(bitnum);
		else
			pgpio->or &= ~BIT(bitnum);
		pgpio->tcr |= BIT(bitnum); /* set output */
		gpio[port].inout |= BIT(bitnum);
	} else {
		if ((ret = check_port_gpio_in(port, BIT(bitnum))) < 0)
			goto err;

		/* clear output if previously set */
		pgpio->tcr &= ~BIT(bitnum);
#ifdef CONFIG_CXD4707
		if (mode & __MODE_ALT_MASK) {
			if ((ret = check_port_gpio_altin(port, BIT(bitnum),
							 GET_ALT_FUNC(mode))) < 0)
				goto err;

			pgpio->is1 &= ~((u_int64_t)0x3 << ((31 - bitnum) * 2));
			pgpio->is1 |= ((u_int64_t)GET_ALT_FUNC(mode) << ((31 - bitnum) * 2));
			gpio[port].alt |= BIT(bitnum);
		} else {
			pgpio->is1 &= ~((u_int64_t)0x3 << ((31 - bitnum)*2)) ;
			gpio[port].alt &= ~BIT(bitnum);
		}
#endif
		gpio[port].inout &= ~BIT(bitnum);
	}
err:
	local_irq_restore(flags);

	return ret;
}

/* set input or output (via bitmask) */
int ppc4xx_gpio_set_mode(int port, u_int32_t bitmask, __u32 mode)
{
	int ret = 0;
	unsigned long flags;
	ocp_gpio_t *pgpio;
	int bit;

	if ((ret = check_port(port)) < 0)
		return ret;

	pgpio = PORT_BASE(port);

	local_irq_save(flags);

	if (mode & __MODE_HIZ) {
		if ((ret = check_port_gpio_hiz(port, bitmask)) < 0)
			goto err;

		pgpio->tcr &= ~bitmask;
		gpio[port].inout |= bitmask; /* HIZ is output */
	} else if (mode & __MODE_OUT) {
		if ((ret = check_port_gpio_out(port, bitmask)) < 0)
			goto err;

		pgpio->tcr &= ~bitmask;  /* set hi-z before setting values*/

#ifdef CONFIG_CXD4707
		if (mode & __MODE_ALT_MASK) {
			if ((ret = check_port_gpio_altout(port, bitmask, 
							  GET_ALT_FUNC(mode))) < 0)
				goto err;

			FOR_EACH_BIT(bit, bitmask) {
				pgpio->os &= ~((u_int64_t)0x3 << ((31 - bit) * 2));
				pgpio->os |= ((u_int64_t)GET_ALT_FUNC(mode) << ((31 - bit) * 2));
			}
			gpio[port].alt |= bitmask;
		} else {
			FOR_EACH_BIT(bit, bitmask) {
				pgpio->os &= ~((u_int64_t)0x3 << ((31 - bit) * 2)) ;
			}
			gpio[port].alt &= ~bitmask;
		}
#endif

		if (mode & __MODE_HIGH)
			pgpio->or |=   bitmask;
		else
			pgpio->or &=  ~bitmask;
		pgpio->tcr |= bitmask;
		gpio[port].inout |= bitmask;
	} else {
		if ((ret = check_port_gpio_in(port, bitmask)) < 0)
			goto err;

		/* clear output if previously set */
		pgpio->tcr &= ~bitmask;
#ifdef CONFIG_CXD4707
		if (mode & __MODE_ALT_MASK) {
			if ((ret = check_port_gpio_altin(port, bitmask,
							 GET_ALT_FUNC(mode))) < 0)
				goto err;

			/* chose alternate */
			FOR_EACH_BIT(bit, bitmask) {
				pgpio->is1 &= ~((u_int64_t)0x3 << ((31 - bit) * 2));
				pgpio->is1 |= ((u_int64_t)GET_ALT_FUNC(mode) << ((31 - bit) * 2));
			}
			gpio[port].alt |= BIT(bitmask);
		} else {
			FOR_EACH_BIT(bit, bitmask) {
				pgpio->is1 &= ~((u_int64_t)0x3 << ((31 - bit)*2)) ;
			}
			gpio[port].alt &= ~BIT(bitmask);
		}
#endif
		gpio[port].inout &= ~bitmask;
	}
err:
	local_irq_restore(flags);

	return ret;
}

/* get value of GPIO (per bit) */
int ppc4xx_gpio_get_data_bit(int port, int bitnum)
{
	int  ret = 0;
	ocp_gpio_t *pgpio;

	if ((ret = check_port_bitnum(port, bitnum)) < 0)
		return ret;

	if ((ret = check_port_gpio_in(port, BIT(bitnum))) < 0)
		return ret;

	pgpio = PORT_BASE(port);

	return ((pgpio->ir & BIT(bitnum)) ? 1 : 0);
}

/* set value of GPIO (per bit) */
int ppc4xx_gpio_set_data_bit(int port, int bitnum, int value)
{
	int  ret = 0;
	unsigned long flags;
	ocp_gpio_t *pgpio;

	if ((ret = check_port_bitnum(port, bitnum)) < 0)
		return ret;

	if ((ret = check_port_gpio_out(port, BIT(bitnum))) < 0)
		return ret;

	pgpio = PORT_BASE(port);

	local_irq_save(flags);
	if (gpio[port].inout & BIT(bitnum)) {
		if (value)
			pgpio->or |=   BIT(bitnum);
		else
			pgpio->or &=  ~BIT(bitnum);
	}
	local_irq_restore(flags);

	return 0;
}

/* get value of GPIO (via bitmask) */
int ppc4xx_gpio_get_data(int port, u_int32_t bitmask, u_int32_t *value)
{
	int  ret = 0;
	ocp_gpio_t *pgpio;

	if ((ret = check_port(port)) < 0)
		return ret;

	if ((ret = check_port_gpio_in(port, bitmask)) < 0)
		return ret;

	pgpio = PORT_BASE(port);
	*value = pgpio->ir & bitmask;

	return 0;
}

/* set value of GPIO (via bitmask) */
int ppc4xx_gpio_set_data(int port, u_int32_t bitmask, u_int32_t value)
{
	int  ret = 0;
	unsigned long flags;
	ocp_gpio_t *pgpio;

	if ((ret = check_port(port)) < 0)
		return ret;

	if ((ret = check_port_gpio_out(port, bitmask)) < 0)
		return ret;

	pgpio = PORT_BASE(port);

	local_irq_save(flags);
	pgpio->or = ((pgpio->or & gpio[port].inout) & ~bitmask ) | value;
	local_irq_restore(flags);

	return 0;
}

/* set interrupt mode (per bit) */
int ppc4xx_gpio_intr_set_mode_bit(int port, int bitnum, __u32 mode)
{
	int  ret = 0;
	unsigned long flags;
	ocp_gpio_t *pgpio;

	if ((ret = check_port_bitnum(port, bitnum)) < 0)
		return ret;

	if ((ret = check_port_gpio_int(port, BIT(bitnum))) < 0)
		return ret;

	if (gpio[port].inout & BIT(bitnum)) {
		printk(KERN_ERR MSG_HEAD 
		       "GPIO %c bit%d is output\n",
		       PORT_NAME(port), bitnum);
		ret = -EINVAL;
	} else {
		int irq;
		int shift;

		irq = __portbit_to_irq(port, bitnum);
		if (irq < 0)
			ret = -EINVAL;

		pgpio = PORT_BASE(port);

		local_irq_save(flags);
		mask_irq(irq);
#ifdef CONFIG_CXD4707
		shift = (31 - bitnum)*2;
		pgpio->is1 = (pgpio->is1 & ~(0x3ULL << shift)) | 
			     (0x1ULL << shift);
#endif
		ret = ppc4xx_pic_setup(irq, 
				       mode & IBM4xx_IRQ_SNS_MASK, 
				       mode & IBM4xx_IRQ_POL_MASK);
		if (gpio[port].bit[bitnum].irq_handler_set)
			unmask_irq(irq);
		local_irq_restore(flags);
	}

	return ret;
}

/* get interrupt mode (per bit) 
 *
 * Returns
 *   * Logical OR of values below
 *       IBM4xx_IRQ_SNS_LVL or IBM4xx_IRQ_SNS_EDG
 *      and 
 *       IBM4xx_IRQ_POL_POS or IBM4xx_IRQ_POL_NEG
 *   * -EINVAL for non exsistent irq number.
 */
int ppc4xx_gpio_intr_get_mode_bit(int port, int bitnum)
{
	int  ret = 0;

	if ((ret = check_port_bitnum(port, bitnum)) < 0)
		return ret;

	if ((ret = check_port_gpio_int(port, BIT(bitnum))) < 0)
		return ret;

	return ppc4xx_pic_get_setup(__portbit_to_irq(port, bitnum));
}

/* set interrupt mode (via bitmask) */
int ppc4xx_gpio_intr_set_mode(int port, u_int32_t bitmask, __u32 mode)
{
	int ret = 0;
	int bit;
	unsigned long flags;
	ocp_gpio_t *pgpio;

	if ((ret = check_port(port)) < 0)
		return ret;

	if ((ret = check_port_gpio_int(port, bitmask)) < 0)
		return ret;

	pgpio = PORT_BASE(port);

	local_irq_save(flags);
	if (gpio[port].inout & bitmask) {
		printk(KERN_ERR MSG_HEAD 
		       "GPIO %c bitmask 0x%x is output\n",
		       PORT_NAME(port), gpio[port].inout & bitmask);
		ret = -EINVAL;
	} else {
		FOR_EACH_BIT(bit, bitmask) {
			int irq;
			int shift;

			irq = __portbit_to_irq(port, bit);
			if (irq < 0) {
				ret = -EINVAL;
				break;
			}
			mask_irq(irq);
#ifdef CONFIG_CXD4707
			shift = (31 - bit)*2;
			pgpio->is1 = (pgpio->is1 & ~(0x3ULL << shift)) | 
				     (0x1ULL << shift);
#endif
			ret = ppc4xx_pic_setup(irq, 
					       mode & IBM4xx_IRQ_SNS_MASK, 
					       mode & IBM4xx_IRQ_POL_MASK);
			if (gpio[port].bit[bit].irq_handler_set)
				unmask_irq(irq);
			if (ret < 0)
				break;
		}
	}

	local_irq_restore(flags);

	return ret;
}

/* register interrupt handler */
int ppc4xx_gpio_intr_register_handler(
	int port, int bitnum, 
	void (*handler)(int irq, void *dev_id, struct pt_regs *regs),
	void *dev_id)
{
	int ret = 0;
	unsigned long flags;
	int irq;

	if ((ret = check_port_bitnum(port, bitnum)) < 0)
		return ret;

	if ((ret = check_port_gpio_int(port, BIT(bitnum))) < 0)
		return ret;

	local_irq_save(flags);
	if (gpio[port].inout & BIT(bitnum)) {
		printk(KERN_ERR MSG_HEAD 
		       "GPIO %c bit%d is output\n",
		       PORT_NAME(port), bitnum);
		ret = -EINVAL;
	} else {
		irq = __portbit_to_irq(port, bitnum);
		if (irq < 0) {
			ret = -EINVAL;
		} else {
			ret = request_irq(irq, handler, 0, 
					  gpio[port].bit[bitnum].name, dev_id);
			gpio[port].bit[bitnum].irq_handler_set = 1;
		}
	}
	local_irq_restore(flags);

	return ret;
}

/* unregister interrupt handler */
int ppc4xx_gpio_intr_unregister_handler(int port, int bitnum,
					 void *dev_id)
{
	int ret;
	int irq;

	if ((ret = check_port_bitnum(port, bitnum)) < 0)
		return ret;

	if ((ret = check_port_gpio_int(port, BIT(bitnum))) < 0)
		return ret;

	irq = __portbit_to_irq(port, bitnum);
	if (irq < 0)
		return -EINVAL;
	gpio[port].bit[bitnum].irq_handler_set = 0;
	free_irq(irq, dev_id);

	return 0;
}

/* unmask interrupt (per bit) */
int ppc4xx_gpio_intr_unmask_bit(int port, int bitnum)
{
	int ret;
	int irq;
	int sns;

	if ((ret = check_port_bitnum(port, bitnum)) < 0)
		return ret;

	if ((ret = check_port_gpio_int(port, BIT(bitnum))) < 0)
		return ret;

	irq = __portbit_to_irq(port, bitnum);
	if (irq < 0)
		return -EINVAL;

	/* if level, clear interrupt */
	sns = ppc4xx_pic_get_setup(irq) & IBM4xx_IRQ_SNS_MASK;
	if (sns == IBM4xx_IRQ_SNS_LVL)
		ppc4xx_pic_clear_status(irq);

	enable_irq(irq);
	return 0;
}

/* mask interrupt (per bit) */
int ppc4xx_gpio_intr_mask_bit(int port, int bitnum)
{
	int ret;
	int irq;

	if ((ret = check_port_bitnum(port, bitnum)) < 0)
		return ret;

	/*
	 * no sanity check for gpio.int multiplex as unregister_*
	 * should call this function.
	 */

	irq = __portbit_to_irq(port, bitnum);
	if (irq < 0)
		return -EINVAL;

	disable_irq(irq);
	return 0;
}

/* unmask interrupt (via bitmask) */
int ppc4xx_gpio_intr_unmask(int port, u_int32_t bitmask)
{
	int ret;
	int bit;
	int irq;

	if ((ret = check_port(port)) < 0)
		return ret;

	if ((ret = check_port_gpio_int(port, bitmask)) < 0)
		return ret;

	FOR_EACH_BIT(bit, bitmask) {
		irq = __portbit_to_irq(port, bit);
		if (irq < 0)
			return -EINVAL;
		enable_irq(irq);
	}
	return 0;
}
 
/* mask interrupt (via bitmask) */
int ppc4xx_gpio_intr_mask(int port, u_int32_t bitmask)
{
	int ret;
	int bit;
	int irq;

	if ((ret = check_port(port)) < 0)
		return ret;

	/*
	 * no sanity check for gpio.int multiplex as unregister_*
	 * should call this function.
	 */

	FOR_EACH_BIT(bit, bitmask) {
		irq = __portbit_to_irq(port, bit);
		if (irq < 0)
			return -EINVAL;
		disable_irq(irq);
	}
	return 0;
}

/* translate IRQ to port number */
int ppc4xx_gpio_irq_to_port(int irq)
{
	if (ppc_md.irq_to_gpio_port)
		return ppc_md.irq_to_gpio_port(irq);
	else
		return -1;
}

/* translate IRQ to bit number */
int ppc4xx_gpio_irq_to_bit(int irq)
{
	if (ppc_md.irq_to_gpio_bit)
		return ppc_md.irq_to_gpio_bit(irq);
	else
		return -1;
}

#ifdef CONFIG_PPC4xx_GPIO_PROC
/* proc */
static int
ppc4xx_gpio_read_proc(char *buf, char **start, off_t off,
		 int count, int *eof, void *data)
{
	ocp_gpio_t *pgpio;
	int port, len = 0;
	int bit;

	for (port = GPIO_PORT_0; port < GPIO_NUMS; port++) {
		len += sprintf(buf + len, "GPIO %c :\n", PORT_NAME(port));
		for (bit = 0; bit < GPIO_BITS_PER_PORT; bit++) {
			if (gpio[port].used & BIT(bit)) {
				len += sprintf(buf + len, 
					       "    bit %d %s : %s\n", 
					       bit, 
					       (gpio[port].inout & BIT(bit)) ?
					       "[OUT]" : "[IN]", 
					       gpio[port].bit[bit].name);
			}
		}
		len += sprintf(buf + len, "\n");
	}
	len += sprintf(buf + len, "==================================\n");
	for (port = GPIO_PORT_0; port < GPIO_NUMS; port++) {
		pgpio = PORT_BASE(port);
		len += sprintf(buf + len, "GPIO %c :\n", PORT_NAME(port));
		len += sprintf(buf + len, "[TCR :0x%08x]\n", pgpio->tcr);
		len += sprintf(buf + len, "[IR  :0x%08x]\n", pgpio->ir);
		len += sprintf(buf + len, "[OR  :0x%08x]\n", pgpio->or);
		len += sprintf(buf + len, "\n");
	}


	*start = buf + off;
	len -= off;
	if (len <= count)
		*eof = 1;
	if (len < 0)
		len = 0;
	else if (len > count)
		len = count;
	return len;
}
#endif /* CONFIG_PPC4xx_GPIO_PROC */

/*
 *  Initialize PPC4XX GPIO driver
 */
int __init
ppc4xx_gpio_init(void)
{
	static int first = 1;

	if (first) {

		/* GPIO 0 */
		gpio_base[0] = (ocp_gpio_t *)ioremap64(PPC4xx_GPIO0_BASE, 
						       sizeof(ocp_gpio_t));

		memset(&gpio[0], 0, sizeof(struct ppc4xx_gpio_t));
		if (ppc_md.init_gpio)
			ppc_md.init_gpio(GPIO_PORT_0, gpio_base[0], &gpio[0]);

#ifdef PPC4xx_GPIO_DEBUG
		printk("gpio[0](no_gpio    =0x%08x,\n" \
		       "        no_out     =0x%08x,\n" \
		       "        no_hiz     =0x%08x,\n" \
		       "        no_in      =0x%08x,\n" \
		       "        no_irq     =0x%08x,\n" \
		       "        no_altout0 =0x%08x,\n" \
		       "        no_altout1 =0x%08x,\n" \
		       "        no_altout2 =0x%08x,\n" \
		       "        no_altin0  =0x%08x,\n" \
		       "        no_altin1  =0x%08x,\n" \
		       "        no_altin2  =0x%08x,\n",
		       gpio[0].no_gpio,
		       gpio[0].no_out,
		       gpio[0].no_hiz,
		       gpio[0].no_in,
		       gpio[0].no_int,
		       gpio[0].no_altout[0],
		       gpio[0].no_altout[1],
		       gpio[0].no_altout[2],
		       gpio[0].no_altin[0],
		       gpio[0].no_altin[1],
		       gpio[0].no_altin[2]);
#endif

#if GPIO_NUMS > 1
		/* GPIO 1 */
		gpio_base[1] = (ocp_gpio_t *)ioremap64(PPC4xx_GPIO1_BASE,
						       sizeof(ocp_gpio_t));

		memset(&gpio[1], 0, sizeof(struct ppc4xx_gpio_t));
		if (ppc_md.init_gpio)
			ppc_md.init_gpio(GPIO_PORT_1, gpio_base[1], &gpio[1]);
#endif

#ifdef PPC4xx_GPIO_DEBUG
		printk("gpio[1](no_gpio    =0x%08x,\n" \
		       "        no_out     =0x%08x,\n" \
		       "        no_hiz     =0x%08x,\n" \
		       "        no_in      =0x%08x,\n" \
		       "        no_irq     =0x%08x,\n" \
		       "        no_altout0 =0x%08x,\n" \
		       "        no_altout1 =0x%08x,\n" \
		       "        no_altout2 =0x%08x,\n" \
		       "        no_altin0  =0x%08x,\n" \
		       "        no_altin1  =0x%08x,\n" \
		       "        no_altin2  =0x%08x,\n",
		       gpio[1].no_gpio,
		       gpio[1].no_out,
		       gpio[1].no_hiz,
		       gpio[1].no_in,
		       gpio[1].no_int,
		       gpio[1].no_altout[0],
		       gpio[1].no_altout[1],
		       gpio[1].no_altout[2],
		       gpio[1].no_altin[0],
		       gpio[1].no_altin[1],
		       gpio[1].no_altin[2]);
#endif

#ifdef CONFIG_PPC4xx_GPIO_PROC
		/* setup proc entry */
		if (!create_proc_read_entry(MODNAME, 0, NULL,
					    ppc4xx_gpio_read_proc,
					    NULL)) {
			printk(KERN_ERR MODNAME
			       "unable to create proc entry\n");
			return -ENOMEM;
		}
#endif /* CONFIG_PPC4xx_GPIO_PROC */
		first = 0;
	}

	return 0;
}


static void __exit
ppc4xx_gpio_exit(void)
{
#ifdef CONFIG_PPC4xx_GPIO_PROC
	remove_proc_entry(MODNAME, NULL);
#endif

	iounmap(gpio_base[0]);
#if GPIO_NUMS > 1
	iounmap(gpio_base[1]);
#endif
}

module_init(ppc4xx_gpio_init);
module_exit(ppc4xx_gpio_exit);

MODULE_AUTHOR("Sony NSC");
MODULE_LICENSE("GPL");

EXPORT_SYMBOL(ppc4xx_gpio_register_bit);
EXPORT_SYMBOL(ppc4xx_gpio_unregister_bit);
EXPORT_SYMBOL(ppc4xx_gpio_register);
EXPORT_SYMBOL(ppc4xx_gpio_unregister);
EXPORT_SYMBOL(ppc4xx_gpio_unregister_by_name);
EXPORT_SYMBOL(ppc4xx_gpio_set_mode_bit);
EXPORT_SYMBOL(ppc4xx_gpio_get_data_bit);
EXPORT_SYMBOL(ppc4xx_gpio_set_data_bit);
EXPORT_SYMBOL(ppc4xx_gpio_set_mode);
EXPORT_SYMBOL(ppc4xx_gpio_get_data);
EXPORT_SYMBOL(ppc4xx_gpio_set_data);
EXPORT_SYMBOL(ppc4xx_gpio_intr_set_mode_bit);
EXPORT_SYMBOL(ppc4xx_gpio_intr_get_mode_bit);
EXPORT_SYMBOL(ppc4xx_gpio_intr_set_mode);
EXPORT_SYMBOL(ppc4xx_gpio_intr_register_handler);
EXPORT_SYMBOL(ppc4xx_gpio_intr_unregister_handler);
EXPORT_SYMBOL(ppc4xx_gpio_intr_unmask_bit);
EXPORT_SYMBOL(ppc4xx_gpio_intr_mask_bit);
EXPORT_SYMBOL(ppc4xx_gpio_intr_unmask);
EXPORT_SYMBOL(ppc4xx_gpio_intr_mask);
EXPORT_SYMBOL(ppc4xx_gpio_irq_to_port);
EXPORT_SYMBOL(ppc4xx_gpio_irq_to_bit);
