/*
 *  linux/arch/ppc/kernel/traps.c
 *
 *  Copyright (C) 1995-1996  Gary Thomas (gdt@linuxppc.org)
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version
 *  2 of the License, or (at your option) any later version.
 *
 *  Modified by Cort Dougan (cort@cs.nmt.edu)
 *  and Paul Mackerras (paulus@cs.anu.edu.au)
 */

/*
 * This file handles the architecture-dependent parts of hardware exceptions
 */

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/stddef.h>
#include <linux/unistd.h>
#include <linux/ptrace.h>
#include <linux/slab.h>
#include <linux/user.h>
#include <linux/a.out.h>
#include <linux/interrupt.h>
#include <linux/config.h>
#include <linux/init.h>

#include <linux/trace.h>

#include <asm/pgtable.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/processor.h>
#ifdef CONFIG_PMAC_BACKLIGHT
#include <asm/backlight.h>
#endif

extern int fix_alignment(struct pt_regs *);
extern void bad_page_fault(struct pt_regs *, unsigned long, int sig);

#ifdef CONFIG_XMON
extern void xmon(struct pt_regs *regs);
extern int xmon_bpt(struct pt_regs *regs);
extern int xmon_sstep(struct pt_regs *regs);
extern int xmon_iabr_match(struct pt_regs *regs);
extern int xmon_dabr_match(struct pt_regs *regs);
extern void (*xmon_fault_handler)(struct pt_regs *regs);
#endif

#ifdef CONFIG_XMON
void (*debugger)(struct pt_regs *regs) = xmon;
int (*debugger_bpt)(struct pt_regs *regs) = xmon_bpt;
int (*debugger_sstep)(struct pt_regs *regs) = xmon_sstep;
int (*debugger_iabr_match)(struct pt_regs *regs) = xmon_iabr_match;
int (*debugger_dabr_match)(struct pt_regs *regs) = xmon_dabr_match;
void (*debugger_fault_handler)(struct pt_regs *regs);
#else
#ifdef CONFIG_KGDB
void (*debugger)(struct pt_regs *regs);
int (*debugger_bpt)(struct pt_regs *regs);
int (*debugger_sstep)(struct pt_regs *regs);
int (*debugger_iabr_match)(struct pt_regs *regs);
int (*debugger_dabr_match)(struct pt_regs *regs);
void (*debugger_fault_handler)(struct pt_regs *regs);
#else
#define debugger(regs)			do { } while (0)
#define debugger_bpt(regs)		0
#define debugger_sstep(regs)		0
#define debugger_iabr_match(regs)	0
#define debugger_dabr_match(regs)	0
#define debugger_fault_handler		((void (*)(struct pt_regs *))0)
#endif
#endif

#ifdef CONFIG_EXCEPTION_MONITOR
extern int exception_check_mode;
extern void (*exception_check)(int mode, struct pt_regs *regs);
#endif
#ifdef CONFIG_LOG_DRIVER
extern int (*write_log_driver)(char *printbuf);
#endif

/*
 * Trap & Exception support
 */


spinlock_t die_lock = SPIN_LOCK_UNLOCKED;

void die(const char * str, struct pt_regs * fp, long err)
{
#ifdef CONFIG_BOOTX_TEXT
	extern int force_printk_to_btext;
#endif
	console_verbose();
	spin_lock_irq(&die_lock);

#ifdef CONFIG_EXCEPTION_MONITOR
	if (exception_check_mode == 0 || exception_check == NULL) {
#ifdef CONFIG_LOG_DRIVER
		write_log_driver = NULL;
#endif
	}
#endif

#ifdef CONFIG_BOOTX_TEXT
	force_printk_to_btext = 1;
#endif
#ifdef CONFIG_PMAC_BACKLIGHT
	set_backlight_enable(1);
	set_backlight_level(BACKLIGHT_MAX);
#endif
	printk("Oops: %s, sig: %ld\n", str, err);

#ifdef CONFIG_EXCEPTION_MONITOR
	if (exception_check_mode && exception_check) {
		exception_check(exception_check_mode, fp);
	} else {
#endif
		show_regs(fp);
#ifdef CONFIG_EXCEPTION_MONITOR
	}
#endif

#ifdef CONFIG_BOOTX_TEXT
	force_printk_to_btext = 0;
#endif
	spin_unlock_irq(&die_lock);
#ifdef CONFIG_SNSC_PANIC_ON_OOPS
	/* Call panic(), when Oops is occurred */
	panic("Request System Reboot, Because Outbreak of Oops !\n");
#endif	/* CONFIG_SNSC_PANIC_ON_OOPS */
	/* do_exit() should take care of panic'ing from an interrupt
	 * context so we don't handle it here
	 */
	do_exit(err);
}

void
_exception(int signr, struct pt_regs *regs, int code, unsigned long addr)
	{
	siginfo_t info;

	if (!user_mode(regs)) {
		debugger(regs);
		die("Exception in kernel mode", regs, signr);
	}
	TRACE_TRAP_ENTRY(regs->trap, instruction_pointer(regs));
	info.si_signo = signr;
	info.si_errno = 0;
	info.si_code = code;
	info.si_addr = (void *) addr;
	force_sig_info(signr, &info, current);
	TRACE_TRAP_EXIT();
}

	/*
	 * I/O accesses can cause machine checks on powermacs.
	 * Check if the NIP corresponds to the address of a sync
	 * instruction for which there is an entry in the exception
	 * table.
	 * Note that the 601 only takes a machine check on TEA
	 * (transfer error ack) signal assertion, and does not
	 * set any of the top 16 bits of SRR1.
	 *  -- paulus.
	 */
static inline int check_io_access(struct pt_regs *regs)
{
#ifdef CONFIG_ALL_PPC
	unsigned long fixup;
	unsigned long msr = regs->msr;

	if (((msr & 0xffff0000) == 0 || (msr & (0x80000 | 0x40000)))
	    && (fixup = search_exception_table(regs->nip)) != 0) {
		/*
		 * Check that it's a sync instruction, or somewhere
		 * in the twi; isync; nop sequence that inb/inw/inl uses.
		 * As the address is in the exception table
		 * we should be able to read the instr there.
		 * For the debug message, we look at the preceding
		 * load or store.
		 */
		unsigned int *nip = (unsigned int *)regs->nip;
		if (*nip == 0x60000000)		/* nop */
			nip -= 2;
		else if (*nip == 0x4c00012c)	/* isync */
			--nip;
		if (*nip == 0x7c0004ac || (*nip >> 26) == 3) {
			/* sync or twi */
			unsigned int rb;

			--nip;
			rb = (*nip >> 11) & 0x1f;
			printk(KERN_DEBUG "%s bad port %lx at %p\n",
			       (*nip & 0x100)? "OUT to": "IN from",
			       regs->gpr[rb] - _IO_BASE, nip);
			regs->nip = fixup;
			return 1;
		}
	}
#endif /* CONFIG_ALL_PPC */
	return 0;
}

#if defined(CONFIG_4xx) || defined(CONFIG_BOOKE)
/* On 4xx, the reason for the machine check or program exception
   is in the ESR. */
#define get_reason(regs)	((regs)->dsisr)
#define REASON_FP		0
#define REASON_ILLEGAL		ESR_PIL
#define REASON_PRIVILEGED	ESR_PPR
#define REASON_TRAP		ESR_PTR

#else
/* On non-4xx, the reason for the machine check or program
   exception is in the MSR. */
#define get_reason(regs)	((regs)->msr)
#define REASON_FP		0x100000
#define REASON_ILLEGAL		0x80000
#define REASON_PRIVILEGED	0x40000
#define REASON_TRAP		0x20000
#endif

#ifdef CONFIG_CXD4707
#define SPRN_MCSRR0	0x23A	/* Machine Check Save and Restore Register 0 */
#define SPRN_MCSRR1	0x23B	/* Machine Check Save and Restore Register 1 */
#define SPRN_MCSR	0x23C	/* Machine Check Status Register */
#define MCSR_MCS	0x80000000 /* Machine Check Summary */
#define MCSR_IB		0x40000000 /* Instruction PLB Error */
#define MCSR_DRB	0x20000000 /* Data Read PLB Error */
#define MCSR_DWB	0x10000000 /* Data Write PLB Error */
#define MCSR_TLBP	0x08000000 /* TLB Parity Error */
#define MCSR_ICP	0x04000000 /* I-Cache Parity Error */
#define MCSR_DCSP	0x02000000 /* D-Cache Search Parity Error */
#define MCSR_DCFP	0x01000000 /* D-Cache Flush Parity Error */
#define MCSR_IMPE	0x00800000 /* Imprecise Machine Check Exception */
#endif /* CONFIG_CXD4707 */

void
MachineCheckException(struct pt_regs *regs)
{
	unsigned long reason = get_reason(regs);
#ifdef CONFIG_CXD4707
	u32 mcsr;
#endif

	if (user_mode(regs)) {
		_exception(SIGBUS, regs, BUS_ADRERR, regs->nip);
		return;
	}

#if defined(CONFIG_8xx) && defined(CONFIG_PCI)
	/* the qspan pci read routines can cause machine checks -- Cort */
	bad_page_fault(regs, regs->dar, SIGBUS);
	return;
#endif
	if (debugger_fault_handler) {
		debugger_fault_handler(regs);
		return;
	}
	if (check_io_access(regs))
		return;

#ifdef CONFIG_4xx
	if (reason & ESR_IMCP) {
		printk("Instruction");
		mtspr(SPRN_ESR, reason & ~ESR_IMCP);
	} else
		printk("Data");
	printk(" machine check in kernel mode.\n");

#ifdef CONFIG_CXD4707
	/* Dump MCSR */
	printk("MCSRR0: 0x%08x\n", mfspr(SPRN_MCSRR0));
	printk("MCSRR1: 0x%08x\n", mfspr(SPRN_MCSRR1));
	mcsr = mfspr(SPRN_MCSR);
	printk("MCSR: 0x%08x\n", mcsr);
	if (mcsr & MCSR_IB)
		printk("    Instruction Read PLB Error\n");
	if (mcsr & MCSR_DRB)
		printk("    Data Read PLB Error\n");
	if (mcsr & MCSR_DWB)
		printk("    Data Write PLB Error\n");
	if (mcsr & MCSR_TLBP)
		printk("    TLB Parity Error\n");
	if (mcsr & MCSR_ICP)
		printk("    I-Cache Parity Error\n");
	if (mcsr & MCSR_DCSP)
		printk("    D-Cache Search Parity Error\n");
	if (mcsr & MCSR_DCFP)
		printk("    D-Cache Flush Parity Error\n");
	if (mcsr & MCSR_IMPE)
		printk("    Machine Check exception is imprecise\n");
#endif /* CONFIG_CXD4707 */

#else /* !CONFIG_4xx */
	printk("Machine check in kernel mode.\n");
	printk("Caused by (from SRR1=%lx): ", reason);
	switch (reason & 0x601F0000) {
	case 0x80000:
		printk("Machine check signal\n");
		break;
	case 0:		/* for 601 */
	case 0x40000:
	case 0x140000:	/* 7450 MSS error and TEA */
		printk("Transfer error ack signal\n");
		break;
	case 0x20000:
		printk("Data parity error signal\n");
		break;
	case 0x10000:
		printk("Address parity error signal\n");
		break;
	case 0x20000000:
		printk("L1 Data Cache error\n");
		break;
	case 0x40000000:
		printk("L1 Instruction Cache error\n");
		break;
	case 0x00100000:
		printk("L2 data cache parity error\n");
		break;
	default:
		printk("Unknown values in msr\n");
	}
#endif /* CONFIG_4xx */

	debugger(regs);
	die("machine check", regs, SIGBUS);
}

void
SMIException(struct pt_regs *regs)
{
	debugger(regs);
#if !(defined(CONFIG_XMON) || defined(CONFIG_KGDB))
	show_regs(regs);
	panic("System Management Interrupt");
#endif
}

void
UnknownException(struct pt_regs *regs)
{
	printk("Bad trap at PC: %lx, SR: %lx, vector=%lx    %s\n",
	       regs->nip, regs->msr, regs->trap, print_tainted());
	_exception(SIGTRAP, regs, 0, 0);
}

#ifdef CONFIG_SNSC_MPU4xx_WDT
extern int _ppc440_wdt_limit_expired(void);
extern void ppc440_wdt_heartbeat(void);

#undef  WDT_SHOW_REGS
#undef WDT_DEBUG_TIME

#ifdef WDT_DEBUG_TIME
#include <asm/div64.h>
struct u64_time {
	u32	hi;
	u32	lo;
};
#endif

void
WatchdogException(struct pt_regs *regs)
{
#ifdef WDT_SHOW_REGS
	{
		u_int32_t tcr, tsr;

		tcr = mfspr(SPRN_TCR);
		tsr = mfspr(SPRN_TSR);

		printk("%s(): tcr    = 0x%08x (WP = %x, WRC = %x, WIE = %x)\n", 
		       __FUNCTION__, 
		       tcr, (tcr & TCR_WP_MASK)>>30, 
		       (tcr & TCR_WRC_MASK)>>28, (tcr & TCR_WIE) ? 1 : 0);
		printk("%s(): tsr    = 0x%08x (ENW = %x, WIS = %x, WRS = %x)\n",
		       __FUNCTION__, 
		       tsr, (tsr & TSR_ENW) ? 1 : 0, (tsr & TSR_WIS) ? 1 : 0, 
		       (tsr & TSR_WRS(3))>>28);
	}
#endif
#ifdef WDT_DEBUG_TIME
	{
		struct u64_time time;
		u64 *dividend;
		u32 ms;

		/* Time Base counter runs at CPU core clock speed */
		time.lo = mfspr(SPRN_TBRL);
		time.hi = mfspr(SPRN_TBRU);
		dividend = (u64 *)&time;
		do_div(*dividend, (u32)(ibm440_clks.cpu / 1000));
		ms  = time.lo;		/* can count up to 49 days */
		printk("%s() : %5u [ms]\n", __FUNCTION__, ms);
	}
#endif

	if (!_ppc440_wdt_limit_expired()) {
		/* re-set watchdog timer */
		ppc440_wdt_heartbeat();
	} else {

		/* 
		   clear wdt to wait for next printk() to finish
		   printing the cause of this exception.
		*/
		ppc440_wdt_heartbeat();
		printk("Watchdog Timer Expired!!\n");

		/* mask watch dog timer interrupt to give process
		 * chance to reset watchdog timer */
		mtspr(SPRN_TCR, mfspr(SPRN_TCR) & ~TCR_WIE);
	}
}
#endif

void
InstructionBreakpoint(struct pt_regs *regs)
{
	if (debugger_iabr_match(regs))
		return;
	_exception(SIGTRAP, regs, TRAP_BRKPT, 0);
}

void
RunModeException(struct pt_regs *regs)
{
	_exception(SIGTRAP, regs, 0, 0);
}

/* Illegal instruction emulation support.  Originally written to
 * provide the PVR to user applications using the mfspr rd, PVR.
 * Return non-zero if we can't emulate, or EFAULT if the associated
 * memory access caused an access fault.  Return zero on success.
 *
 * There are a couple of ways to do this, either "decode" the instruction
 * or directly match lots of bits.  In this case, matching lots of
 * bits is faster and easier.
 *
 */
#define INST_MFSPR_PVR		0x7c1f42a6
#define INST_MFSPR_PVR_MASK	0xfc1fffff

static int
emulate_instruction(struct pt_regs *regs)
{
	u32 instword;
	u32 rd;
	int retval;

	retval = -EINVAL;

	if (!user_mode(regs))
		return retval;

	if (get_user(instword, (u32 *)(regs->nip)))
		return -EFAULT;

	/* Emulate the mfspr rD, PVR.
	 */
	if ((instword & INST_MFSPR_PVR_MASK) == INST_MFSPR_PVR) {
		rd = (instword >> 21) & 0x1f;
		regs->gpr[rd] = mfspr(PVR);
		retval = 0;
		regs->nip += 4;
	}
	return retval;
}

void
ProgramCheckException(struct pt_regs *regs)
{
	unsigned int reason = get_reason(regs);
	extern int do_mathemu(struct pt_regs *regs);

#ifdef CONFIG_MATH_EMULATION
	/* (reason & REASON_ILLEGAL) would be the obvious thing here,
	 * but there seems to be a hardware bug on the 405GP (RevD)
	 * that means ESR is sometimes set incorrectly - either to
	 * ESR_DST (!?) or 0.  In the process of chasing this with the
	 * hardware people - not sure if it can happen on any illegal
	 * instruction or only on FP instructions, whether there is a
	 * pattern to occurences etc. -dgibson 31/Mar/2003 */
	if (!(reason & REASON_TRAP) && do_mathemu(regs) == 0)
		return;
#endif /* CONFIG_MATH_EMULATION */

	if (reason & REASON_FP) {
		/* IEEE FP exception */
		int code = 0;
		u32 fpscr;

		if (regs->msr & MSR_FP)
			giveup_fpu(current);
		fpscr = current->thread.fpscr;
		fpscr &= fpscr << 22;	/* mask summary bits with enables */
		if (fpscr & FPSCR_VX)
			code = FPE_FLTINV;
		else if (fpscr & FPSCR_OX)
			code = FPE_FLTOVF;
		else if (fpscr & FPSCR_UX)
			code = FPE_FLTUND;
		else if (fpscr & FPSCR_ZX)
			code = FPE_FLTDIV;
		else if (fpscr & FPSCR_XX)
			code = FPE_FLTRES;
		_exception(SIGFPE, regs, code, regs->nip);
		return;
	}

	if (reason & REASON_TRAP) {
		/* trap exception */
#if defined(CONFIG_4xx)
		mtspr(SPRN_DBSR, DBSR_TIE);
#endif
		if (debugger_bpt(regs))
			return;
		_exception(SIGTRAP, regs, TRAP_BRKPT, 0);
		return;
	}

	if (reason & REASON_PRIVILEGED) {
	/* Try to emulate it if we should. */
		if (emulate_instruction(regs) == 0)
			return;
		_exception(SIGILL, regs, ILL_PRVOPC, regs->nip);
		return;
	}

	_exception(SIGILL, regs, ILL_ILLOPC, regs->nip);
}

void
SingleStepException(struct pt_regs *regs)
{
	regs->msr &= ~MSR_SE;  /* Turn off 'trace' bit */
	if (debugger_sstep(regs))
		return;
	_exception(SIGTRAP, regs, TRAP_TRACE, 0);
}

void
AlignmentException(struct pt_regs *regs)
{
	int fixed;

	fixed = fix_alignment(regs);
	if (fixed == 1) {
		regs->nip += 4;	/* skip over emulated instruction */
		return;
	}
	if (fixed == -EFAULT) {
		/* fixed == -EFAULT means the operand address was bad */
		if (user_mode(regs))
			_exception(SIGSEGV, regs, SEGV_ACCERR, regs->dar);
		else
			bad_page_fault(regs, regs->dar, SIGSEGV);
		return;
	}
	_exception(SIGBUS, regs, BUS_ADRALN, regs->dar);
}

void
StackOverflow(struct pt_regs *regs)
{
	printk(KERN_CRIT "Kernel stack overflow in process %p, r1=%lx\n",
	       current, regs->gpr[1]);

#ifdef CONFIG_EXCEPTION_MONITOR
	if (exception_check_mode && exception_check) {
		die("stack overflow", regs, SIGSTKFLT);
	}
#endif

#ifdef CONFIG_LOG_DRIVER
	write_log_driver = NULL;
#endif

	debugger(regs);
	show_regs(regs);
	panic("kernel stack overflow");
}

/* Trace related code */
#if (CONFIG_TRACE || CONFIG_TRACE_MODULE)
asmlinkage void trace_real_syscall_entry(struct pt_regs * regs)
{
        int                 use_depth;
	int                 use_bounds;
	int                 depth = 0;
        int                 seek_depth;
        unsigned long       lower_bound;
        unsigned long       upper_bound;
	unsigned long       addr;
	unsigned long*      stack;
	trace_syscall_entry trace_syscall_event;

        /* Set the syscall ID */
	trace_syscall_event.syscall_id = (uint8_t) regs->gpr[0];

	/* Set the address in any case */
	trace_syscall_event.address  = instruction_pointer(regs);

	/* Are we in the kernel (This is a kernel thread)? */
	if(!user_mode(regs))
	  /* Don't go digining anywhere */
	  goto trace_syscall_end;

	/* Get the trace configuration */
	if(trace_get_config(&use_depth,
			    &use_bounds,
			    &seek_depth,
			    (void*)&lower_bound,
			    (void*)&upper_bound) < 0)
	  goto trace_syscall_end;

	/* Do we have to search for an eip address range */
	if((use_depth == 1) || (use_bounds == 1))
	  {
	  /* Start at the top of the stack (bottom address since stacks grow downward) */
	  stack = (unsigned long*) regs->gpr[1];

	  /* Skip over first stack frame as the return address isn't valid */
	  if(get_user(addr, stack))
	    goto trace_syscall_end;
	  stack = (unsigned long*) addr;

	  /* Keep on going until we reach the end of the process' stack limit (wherever it may be) */
	  while(!get_user(addr, stack + 1)) /* "stack + 1", since this is where the IP is */
	    {
	    /* Does this LOOK LIKE an address in the program */
	    if((addr > current->mm->start_code)
             &&(addr < current->mm->end_code))
	      {
	      /* Does this address fit the description */
	      if(((use_depth == 1) && (depth == seek_depth))
               ||((use_bounds == 1) && (addr > lower_bound) && (addr < upper_bound)))
		{
		/* Set the address */
		trace_syscall_event.address = addr;

		/* We're done */
		goto trace_syscall_end;
		}
	      else
		/* We're one depth more */
		depth++;
	      }
	    /* Go on to the next address */
	    if(get_user(addr, stack))
	      goto trace_syscall_end;
	    stack = (unsigned long*) addr;
	    }
	  }

trace_syscall_end:
	/* Trace the event */
	trace_event(TRACE_EV_SYSCALL_ENTRY, &trace_syscall_event);
}

asmlinkage void trace_real_syscall_exit(void)
{
        trace_event(TRACE_EV_SYSCALL_EXIT, NULL);
}
#endif /* (CONFIG_TRACE || CONFIG_TRACE_MODULE) */

void
trace_syscall(struct pt_regs *regs)
{
	printk("Task: %p(%d), PC: %08lX/%08lX, Syscall: %3ld, Result: %s%ld    %s\n",
	       current, current->pid, regs->nip, regs->link, regs->gpr[0],
	       regs->ccr&0x10000000?"Error=":"", regs->gpr[3], print_tainted());
}

#ifdef CONFIG_8xx
void
SoftwareEmulation(struct pt_regs *regs)
{
	extern int do_mathemu(struct pt_regs *);
	extern int Soft_emulate_8xx(struct pt_regs *);
	int errcode;

	if (!user_mode(regs)) {
		debugger(regs);
		die("Kernel Mode Software FPU Emulation", regs, SIGFPE);
	}

#ifdef CONFIG_MATH_EMULATION
	errcode = do_mathemu(regs);
#else
	errcode = Soft_emulate_8xx(regs);
#endif
	if (errcode) {
		if (errcode > 0)
			_exception(SIGFPE, regs, 0, 0);
		else if (errcode == -EFAULT)
			_exception(SIGSEGV, regs, 0, 0);
		else
			_exception(SIGILL, regs, ILL_ILLOPC, regs->nip);
	}
}
#endif /* CONFIG_8xx */

#if defined(CONFIG_4xx) || defined(CONFIG_BOOKE)

void DebugException(struct pt_regs *regs)
{
	unsigned long debug_status;

	debug_status = mfspr(SPRN_DBSR);

	regs->msr &= ~MSR_DE;  /* Turn off 'debug' bit */
	if (debug_status & DBSR_TIE) {		/* trap instruction*/

		mtspr(SPRN_DBSR, DBSR_TIE);

		if (!user_mode(regs) && debugger_bpt(regs))
			return;
		_exception(SIGTRAP, regs, 0, 0);

	} else if (debug_status & DBSR_IC) {	/* instruction completion */

		mtspr(SPRN_DBSR, DBSR_IC);

		if (user_mode(regs))
			current->thread.dbcr0 &= ~DBCR0_IC;
		else {
#ifdef CONFIG_KGDB
			current->thread.kernel_dbcr0 &= ~DBCR0_IC;
#endif
			if (debugger_sstep(regs))
			return;
		}
		_exception(SIGTRAP, regs, 0, 0);
	}
}
#endif /* CONFIG_4xx || CONFIG_BOOKE */

#if !defined(CONFIG_TAU_INT)
void
TAUException(struct pt_regs *regs)
{
	printk("TAU trap at PC: %lx, SR: %lx, vector=%lx    %s\n",
	       regs->nip, regs->msr, regs->trap, print_tainted());
}
#endif /* CONFIG_INT_TAU */

void __init trap_init(void)
{
}
