/*
 * arch/ppc/platforms/cxd4707_wol.c
 *
 * CXD4707 ES2.x Wake Up On LAN interface.
 *
 *  Copyright 2006 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
#include <linux/config.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include <linux/ethtool.h>

#include <asm/io.h>
#include <asm/system.h>
#include <asm/uaccess.h>

#include <platforms/cxd4707_wol.h>

#define WOL_PROC_NAME "cxd4707_wol"

/* ======================================================================= *
 * Debug print function                                                    *
 *   set permutations of:                                                  *
 *     0 : disables all debug output                                       *
 *     1 : Print output. (Default)                                         *
 *     2 : Verbose, dump registers                                         *
 *     4 : Print entry of functions.                                       *
 * ======================================================================= */

#define DEBUG_NAME "WOL" 
#define DEBUG_FLAG 1

#define ERROR_STRING	"#Error:"
#define PRINTK printk
#define ERROR(format, args...) printk(ERROR_STRING DEBUG_NAME ":%s():" format , __FUNCTION__ , ## args)

#ifdef DEBUG
#undef DEBUG
#endif

#if DEBUG_FLAG & 1
#define PRINT(format, args...) printk(DEBUG_NAME ":%s():" format , __FUNCTION__ , ## args)
#else
#define PRINT(format, args...)
#endif

#if DEBUG_FLAG & 2
#define DEBUG(format, args...) printk(DEBUG_NAME ":%s():%d:" format , __FUNCTION__ , __LINE__ , ## args)
#else
#define DEBUG(format, args...)
#endif

#if DEBUG_FLAG & 4
#define ENTRY() printk(DEBUG_NAME ":%s():entry\n", __FUNCTION__)
#else
#define ENTRY()
#endif

/* ======================================================================= *
 * Global variable                                                         *
 * ======================================================================= */

static unsigned int wol = 0;
static struct proc_dir_entry *proc_wol;

/* Sets wol value. */
int
cxd4707_set_wol(unsigned int val)
{
	ENTRY();
	DEBUG("val=0x%x\n", val);

	/* Sanity check */
	if (val & ~(WOL_ENABLE | WOL_WITH_WOL_MOD)) {
		ERROR("Invalid WOL parameter 0x%x\n", val);
		return -EINVAL;
	}
	wol = val;

	DEBUG("wol=0x%x\n", wol);
	return 0;
}
EXPORT_SYMBOL(cxd4707_set_wol);

/* Gets wol value. Returns 1 if wol is enabled */
int
cxd4707_is_wol_set(void)
{
	ENTRY();
	DEBUG("wol=0x%x\n", wol);

	if (wol & WAKE_MAGIC)
		return 1;
	return 0;
}
EXPORT_SYMBOL(cxd4707_is_wol_set);

/* 
 * Kicks WOL by writing to SDR0_WOLSTBY if global variable wol is set.
 * This function never returns if wol is set, as it resets the CPU. 
 */
void
cxd4707_wol_kick(void)
{
	unsigned int val;

	ENTRY();
	DEBUG("wol=0x%x\n", wol);

	if (wol & WAKE_MAGIC) {
		PRINTK("Enabling Wake-On-LAN.\n");

		val = 0x1;
		if (wol & WOL_WITH_WOL_MOD) 
			val |= WOL_WITH_WOL_MOD;
		else
			val |= WOL_WITH_STANDBY_MODE;

		DEBUG("val=0x%x\n", val);
		mdelay(10);

		mtdcri(DCRN_SDR0, WOLEN, val);
		mb();		/* sync */
	}
}
EXPORT_SYMBOL(cxd4707_wol_kick);

static int
proc_wol_read(char *buf, char **start, off_t off,
	      int count, int *eof, void *data)
{
	ENTRY();

	DEBUG("0x%x\n", wol);
	*eof = 1;
	return sprintf(buf, "%d\n", wol);;
}


#define WOL_MAX_CHAR_INPUT 10
static int
proc_wol_write(struct file *file, const char *buffer,
	       unsigned long count, void *data)
{
	int len;
	char user_input[WOL_MAX_CHAR_INPUT + 1];

	ENTRY();

	len = (count > WOL_MAX_CHAR_INPUT) ? WOL_MAX_CHAR_INPUT : count;
	if (copy_from_user(user_input, buffer, len)) {
		return -EFAULT;
	}
	user_input[len] = '\0';
	cxd4707_set_wol(simple_strtoul(user_input, NULL, 0));
	return len;
}

static int __init
cxd4707_wol_proc_init(void)
{
	ENTRY();

	proc_wol = create_proc_entry(WOL_PROC_NAME, 0666, &proc_root);
	if (proc_wol == NULL) {
		ERROR("failed to create proc entry(/proc/%s) for WakeOnLAN\n",
		      WOL_PROC_NAME);
	} else {
		proc_wol->read_proc = proc_wol_read;
		proc_wol->write_proc = proc_wol_write;
	}
	return 0;
}
__initcall(cxd4707_wol_proc_init);
