/*
 * arch/ppc/platforms/cxd4710_uma.c
 *
 *  CXD-4710 UMA related routines
 *
 *  Copyright 2005 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <linux/config.h>
#include <linux/errno.h>

#ifdef CONFIG_SNSC_UMA

#define DEBUG_UMA_FLAG	1
#define DEBUG_UMA_NAME	"UMA"

#define ERROR_STRING	"#Error:"

#define PRINTK(format, args...) printk(DEBUG_UMA_NAME ":" format, ## args)
#define ERROR(format, args...) printk(ERROR_STRING DEBUG_UMA_NAME ":%s():" format , __FUNCTION__ , ## args)

#ifdef DEBUG_UMA
#undef DEBUG_UMA
#endif

#if DEBUG_UMA_FLAG & 1
#define PRINT(format, args...) printk(DEBUG_UMA_NAME ":%s():" format , __FUNCTION__ , ## args)
#else
#define PRINT(format, args...)
#endif

#if DEBUG_UMA_FLAG & 2
#define DEBUG_UMA(format, args...) printk(DEBUG_UMA_NAME ":%s():%d:" format , __FUNCTION__ , __LINE__ , ## args)
#else
#define DEBUG_UMA(format, args...)
#endif

#if DEBUG_UMA_FLAG & 4
#define ENTRY() printk(DEBUG_UMA_NAME ":%s():entry\n", __FUNCTION__)
#else
#define ENTRY()
#endif

#include <linux/snsc_nblargs.h>
#include <linux/snsc_uma.h>
struct snsc_uma_machdep_calls uma_md;
snsc_uma_desc_t uma_regions[] = {
	{.size	= CONFIG_SNSC_UMA_ALLEGRO_SIZE,
	 .flags	= SNSC_UMA_FLG_FREE_FROM_TAIL
	},
	{.size	= CONFIG_SNSC_UMA_GFX_SIZE,
	 .flags = SNSC_UMA_FLG_FREE_FROM_HEAD
	},
};
#define NUM_UMA_REGIONS sizeof(uma_regions)/sizeof(snsc_uma_desc_t)

/* 
 * Calcurate UMA region start address for the region specified.
 */
static int
cxd4710_uma_init_range(snsc_uma_t region)
{
	extern unsigned long total_memory;
	snsc_uma_desc_t *pd;

	if (region >= NUM_UMA_REGIONS)
		return -EINVAL;

	pd = uma_regions + region;

	if (region == UMA_ALLEGRO) {
		/* setup shared memory region. */
		PRINTK("Set shared memory register to ");
		if (pd->size > SZ_32M) {
			printk("64MB\n");
			pd->size = SZ_64M;
		} else {
			printk("32MB\n");
			pd->size = SZ_32M;
		}
		DEBUG_UMA("total_memory=%ld\n", total_memory);
		/* UMA_ALLEGRO is placed at the end of memory. */
		pd->addr = (u_int32_t)__va(total_memory - pd->size);
				/* UMA uses NBLArgs to reserve memory
				   region.  NBL for DTV uses virtual
				   address mapped the same way as
				   Linux kernel, so we must translate
				   the memory address. */

	} else if (region == UMA_GFX) {
		/* UMA_GFX is placed to the next of UMA_ALLEGRO */
		pd->addr = (u_int32_t)__va(total_memory -
					   uma_regions[UMA_ALLEGRO].size -
					   pd->size);
	}
	{
		int i;
		for (i = 0; i < NUM_UMA_REGIONS; i++) {
			pd = uma_regions + i;
			DEBUG_UMA("addr=0x%x, size=0x%x, flags=0x%x\n",
				  pd->addr, pd->size, pd->flags);
		}
	}
	return 0;
}

static int
cxd4710_uma_get_range(snsc_uma_t region, snsc_uma_desc_t *pd)
{

	DEBUG_UMA("(region=%d)\n", region);
	if (region >= NUM_UMA_REGIONS)
		return -EINVAL;

	pd->addr  = uma_regions[region].addr;
	pd->size  = uma_regions[region].size;
	pd->flags = uma_regions[region].flags;
	return 0;
}

static int
cxd4710_uma_set_range(snsc_uma_t region, snsc_uma_desc_t pd)
{

	DEBUG_UMA("region=%d\n", region);
	if (region >= NUM_UMA_REGIONS)
		return -EINVAL;

	uma_regions[region].addr  = pd.addr;
	uma_regions[region].size  = pd.size;
	uma_regions[region].flags = pd.flags;
	return 0;
}

static void
cxd4710_uma_sync(void *addr, unsigned int size, unsigned int direction)
{
	if (direction == SNSC_UMA_READ) {
		/* Invalidate(dcbi) and sync */
		invalidate_dcache_range((u_int32_t)addr, (u_int32_t)addr + size);
	} else {
		unsigned int tmp;

		/* Flush and invalidate(use dcbf) and sync */
		flush_dcache_range((u_int32_t)addr, (u_int32_t)addr + size);

		/* Dummy read to flush memory controller's buffer */
		asm volatile ("lwz	%0,0(%1)\n"
			      "sync	\n"
			      : "=r" (tmp)
			      : "b" (addr + size - 4)
			      : "memory");
	}
}

void
platform_init_uma(void)
{

	DEBUG_UMA("init");
	uma_md.init_range = cxd4710_uma_init_range;
	uma_md.get_range  = cxd4710_uma_get_range;
	uma_md.set_range  = cxd4710_uma_set_range;
	uma_md.sync       = cxd4710_uma_sync;
}
#endif	/* CONFIG_SNSC_UMA */
