/*
 *  snsc-mpu4xx_dma.c -- PPC440GP PCI DMA controller
 *
 */
/*
 *  Copyright 2004 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/selection.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/spinlock.h>
#include <linux/proc_fs.h>

#include <asm/uaccess.h>
#include <asm/cache.h>
#include <asm/io.h>
#include <asm/semaphore.h>

#include <asm/ppc4xx_dma.h>
#include <asm/snsc-mpu4xx_dma.h>


#define DEBUG_LVL	1
#define NAME "mpu4xx_dma"

#define ERROR(format, args...) printk(NAME "Error:%s() " format , __FUNCTION__ , ## args)

#if DEBUG_LVL & 1
#define PRINT(format, args...) printk(NAME ": %s() " format , __FUNCTION__ , ## args)
#else
#define PRINT(format, args...)
#endif

#if DEBUG_LVL & 2
#define DEBUG2(format, args...) printk(NAME ": %s():%d " format , __FUNCTION__ , __LINE__ , ## args)
#else
#define DEBUG2(format, args...)
#endif

#if DEBUG_LVL & 4
#define DEBUG4(format, args...) printk(NAME ": %s():%d " format , __FUNCTION__ , __LINE__ , ## args)
#else
#define DEBUG4(format, args...)
#endif



#define DEV_NAME            "mpu4xx_dma"
#define BYTES_PER_CHAIN     32
#define CHANNEL_INFO_MAX    MAX_PPC4xx_DMA_CHANNELS /* channel size */
#define DMANR_TO_CHANNEL(x) (1 << (x)) 
#define SUB_CHANNEL_MAX	    2  /* virtual channel size per one channel */

/*
 * chain descriptor buffer size.
 * max chain num. = 4096 (= CHAIN_BUFSIZE / BYTES_PER_CHAIN)
 */
#define CHAIN_BUFSIZE       (PAGE_SIZE * 32)  /* 128KB. order=5 (32 = 2^5) */

/*
 * userspace data <--> buf table size
 * at most, maximum chain size is enough
 */
#define TABLE_SIZE_MAX      (CHAIN_BUFSIZE / BYTES_PER_CHAIN)

/*
 * the meaning of channel_t returned by mpu4xx_dma_request_channel()
 * is as follows;
 *
 *              31            16 15           0
 *               +--------------+-------------+
 *     channel_t |              |             |
 *     (32 bits) +--------------+-------------+
 *                      |              |
 *                      |             one of SNSC_MPU4XX_DMA_CHANNEL_?
 *                      |
 *               sub_channel index
 *               (0-65535, max = (SUB_CHANNEL_MAX - 1))
 *
 * Thus, even if mpu4xx_dma_request_channel() requesting with
 * SNSC_MPU4XX_DMA_CHANNEL_0 successfully returns, it does not
 * guarantee that the returned channel_t is equal to
 * SNSC_MPU4XX_DMA_CHANNEL_0.
 */
#define MPU4XX_CHANNEL_T(sub, dmanr) \
	(channel_t)((((sub) & 0xffff) << 16) | (DMANR_TO_CHANNEL(dmanr) & 0xffff))

#define MPU4XX_SUB_CHANNEL(channel)  (((channel) >> 16) & 0xffff)

/*
 * maximum number of data in physical address associated with
 * _one_ continuous data in virtual address.
 * WARNING: this value restricts the data length.
 */
#define PHYS_CHUNK_MAX      256

/*
 * uncomment this if you want to check illegal
 * function call that the same channel is
 * accessed at the same time.
 */
// #define CHECK_ILLEGAL_CALL   1

#ifdef CHECK_ILLEGAL_CALL

#define LOCK_DMA_INFO(pinfo)  do { \
		if ( down_trylock(&pinfo->lock) ) { \
			printk("%s: illegal function call found in %s.\n", DEV_NAME, __FUNCTION__); \
			return -EAGAIN; \
		} \
	} while (0)

#define UNLOCK_DMA_INFO(pinfo)  do { \
		up(&pinfo->lock); \
	} while (0)

#else /* CHECK_ILLEGAL_CALL */

#define LOCK_DMA_INFO(pinfo)
#define UNLOCK_DMA_INFO(pinfo)

#endif /* CHECK_ILLEGAL_CALL */

/*
 * un-comment this if you want interrupt every
 * time a chain has been transferred.
 */
// #define ONE_INTERRUPT_PER_CHAIN  1

/*
 * buffer size/num for copying userspace's data
 */
#define DEFAULT_BUFSIZE     (PAGE_SIZE * 64) /* 256KB. order=9 (512 = 2^9) */
#define DEFAULT_BUFNUM       1


#ifdef CONFIG_MPU4xx_DMACHAIN_PROC
#define PROC_NAME	"snsc_mpu4xx_dma"
static u_int32_t pw128_cnt[NR_DMACS][CH_PER_DMAC];
static u_int32_t pw64_cnt[NR_DMACS][CH_PER_DMAC];
static u_int32_t pw32_cnt[NR_DMACS][CH_PER_DMAC];
static u_int32_t pw16_cnt[NR_DMACS][CH_PER_DMAC];
static u_int32_t pw8_cnt[NR_DMACS][CH_PER_DMAC];
#endif

/*
 * state transition diagram:
 *
 *                            +------+
 *                            |UNUSED|
 *                            +------+
 * mpu4xx_dma_request_channel() |  A
 *                              |  |
 *                              V  | mpu4xx_dma_release_channel()
 *        (intr.)            +---------+
 *       +-----------------> |AVAILABLE|
 *       |                   +---------+
 *       |                        | mpu4xx_dma_start()
 *       |                        V
 *       |              {interrupt context ?} ------------+ Yes
 *       |                        | No                    |
 *   +-------+    No              V                       |
 *   |RUNNING| <---- {any RUNNING sub_channels ?} <--+    |
 *   +-------+                    | Yes              |    |
 *                                V                  |    |
 *                            +-------+  (tasklet)   |    |
 *                            |PENDING|--------------+    |
 *                            +-------+                   |
 *                                A                       |
 *                                |                       |
 *                                +-----------------------+
 */
typedef enum {
	MPU4XX_DMA_STATE_UNUSED,
	MPU4XX_DMA_STATE_AVAILABLE,
	MPU4XX_DMA_STATE_RUNNING,
	MPU4XX_DMA_STATE_PENDING
} dma_info_state_t;

typedef struct {
	u32 addr_virtual;		/* virtual address for G-Bus data */
	u32 len;			/* data size (unit: bytes) */
} dma_info_buf_t;

typedef struct {
	u32 user;			/* address in user space */
	u32 buf_virtual;		/* virtual address in "buf" */
	u32 len;			/* data size (unit: byes) */
} dma_info_table_t;

typedef struct {
	dma_info_buf_t *buf;		/* temporal area for user data */
	u32 bufsize;
	u32 bufnum;
	dma_info_table_t *table;	/* user <--> buf address table */
	u32 table_size;

	unsigned long chain_virtual;	/*
					 * virtual address of chain buffer
					 */

	unsigned long chain_phys;	/*
					 * physical address of chain buffer
					 */

	u32 chain_num;			/* 
					 * the number of chain.
					 * one chain occupies
					 * BYTES_PER_CHAIN bytes.
					 */

	u8 mem2pci;			/*
					 * transfer direction flag.
					 *
					 *   0: PCI   => G-Bus
					 *   1: G-Bus => PCI
					 *
					 * don't care this if neither src nor
					 * dst are not in userspace.
					 */

	u8 is_user;			/*
					 * 0: physical address
					 * 1: userspace address
					 */

	snsc_dma_info_t info;		/* controller's current info */

	u32 block;			/* blocking mode flag */

					/* callback function */
	void (*handler)(channel_t, snsc_dma_info_t *, void *);
	void *handler_arg;

	struct semaphore lock;		/*
					 * lock for the exclusive function
					 * call.
					 * this protects the data in case
					 * that exported function is called
					 * from more than one context.
					 */

	channel_t channel;		/*
					 * channel number returned via
					 * mpu4xx_dma_request_channel()
					 */

	dma_info_state_t state;		/* state of this virtual channel */

	wait_queue_head_t wait_q;	/*
					 * for waiting previous request's
					 * completion
					 */

} dma_info_t;

typedef struct {
	int irq;			/*
					 * irq returned by
					 * ppc4xx_get_dma_irq().
					 * requires channel_info_lock to
					 * change the value.
					 */

	int dmanr;			/*
					 * channel number returned by
					 * ppc4xx_request_dma().
					 * requires channel_info_lock to
					 * change the value.
					 */

	u32 error_counter;		/*
					 * if a chain transfer fails,
					 * this variable is incremented
					 */

	int running_sub_channel;	/*
					 * index of dma_info[] which is
					 * running now.
					 * this achieves equality for
					 * scheduling the PENDING channels.
					 */

	dma_info_t dma_info[SUB_CHANNEL_MAX];
					/*
					 * each element holds the
					 * sub_channel related resources.
					 */

	spinlock_t channel_info_lock;	/*
					 * mainly used when checking
					 * all states of dma_info[]
					 * at a time
					 */

	char name[16];			/* name for request_irq() */

} channel_info_t;


static channel_info_t *channel_info;

/* struct for userspace/virtual -> physical address conversion */
typedef struct {
	u32 addr;
	u32 len;
} dma_sdram_pair_t;

typedef struct {
	dma_sdram_pair_t virtual;
	dma_sdram_pair_t phys[PHYS_CHUNK_MAX];
	int phys_num;
} dma_sdram_t;


static void mpu4xx_dma_tasklet(unsigned long arg);

static DECLARE_TASKLET(mpu4xx_dma_tasklet_data,
		       mpu4xx_dma_tasklet,
		       0);


#ifdef CONFIG_MPU4xx_DMACHAIN_PROC
static int
mpu4xx_dmachain_read_proc(char *buf, char **start, off_t off,
		 int count, int *eof, void *data)
{
	int len = 0;
	int i, j;
	for (i=0; i<NR_DMACS; i++) {
		len += sprintf(buf + len, "DMAC %d\n", i);
		for (j=0; j<CH_PER_DMAC; j++) {

			if (pw128_cnt[i][j] + pw64_cnt[i][j] +
			    pw32_cnt[i][j] + pw16_cnt[i][j] + pw8_cnt[i][j]) {
				len += sprintf(buf + len, " CH %d\n", j);
				if (pw128_cnt[i][j])
					len += sprintf(buf + len, "  128bit: %8d\n",
						       pw128_cnt[i][j]);
				if (pw64_cnt[i][j])
					len += sprintf(buf + len, "   64bit: %8d\n",
						       pw64_cnt[i][j]);
				if (pw32_cnt[i][j])
					len += sprintf(buf + len, "   32bit: %8d\n",
						       pw32_cnt[i][j]);
				if (pw16_cnt[i][j])
					len += sprintf(buf + len, "   16bit: %8d\n",
						       pw16_cnt[i][j]);
				if (pw8_cnt[i][j])
					len += sprintf(buf + len, "    8bit: %8d\n",
						       pw8_cnt[i][j]);
			}
		}
	}
	*start = buf + off;
	len -= off;
	if (len <= count)
		*eof = 1;
	if (len < 0)
		len = 0;
	else if (len > count)
		len = count;
	return len;
}
#endif /* CONFIG_MPU4xx_DMACHAIN_PROC */

static channel_info_t *get_channel_info(dma_info_t *pinfo)
{
	int j;
	int k;

	for (k=0; k<CHANNEL_INFO_MAX; k++) {
		channel_info_t *cinfo = &(channel_info[k]);
		for (j=0; j<SUB_CHANNEL_MAX; j++) {
			if (pinfo == &(cinfo->dma_info[j])) {
				return cinfo;
			}
		}
	}
	return NULL;
}

static dma_info_t *get_dma_info(channel_t channel)
{
	int sub_channel = MPU4XX_SUB_CHANNEL(channel);
	channel_t ch = channel & 0xffff;

	if ((sub_channel < 0) || (SUB_CHANNEL_MAX <= sub_channel)) {
		ERROR("specified sub_channel is not correct.\n");
		return NULL;
	}

	switch (ch) {
		case SNSC_MPU4XX_DMA_CHANNEL_0:
			return &(channel_info[0].dma_info[sub_channel]);
			break;

		case SNSC_MPU4XX_DMA_CHANNEL_1:
			return &(channel_info[1].dma_info[sub_channel]);
			break;

		case SNSC_MPU4XX_DMA_CHANNEL_2:
			return &(channel_info[2].dma_info[sub_channel]);
			break;

		case SNSC_MPU4XX_DMA_CHANNEL_3:
			return &(channel_info[3].dma_info[sub_channel]);
			break;

#if NR_DMACS > 1
		case SNSC_MPU4XX_DMA_CHANNEL_4:
			return &(channel_info[4].dma_info[sub_channel]);
			break;

		case SNSC_MPU4XX_DMA_CHANNEL_5:
			return &(channel_info[5].dma_info[sub_channel]);
			break;

		case SNSC_MPU4XX_DMA_CHANNEL_6:
			return &(channel_info[6].dma_info[sub_channel]);
			break;

		case SNSC_MPU4XX_DMA_CHANNEL_7:
			return &(channel_info[7].dma_info[sub_channel]);
			break;
#endif /* NR_DMACS */

		default:
			printk("%s invalid channel %08x\n", DEV_NAME, channel);
	}

	return NULL;
}

/*
 * starts DMA transfer.
 */
static void mpu4xx_dma_start_dma(int dmanr)
{
	unsigned int value = 0;
	unsigned int status_bits[] = { DMA_CS0 | DMA_TS0 | DMA_CH0_ERR,
				       DMA_CS1 | DMA_TS1 | DMA_CH1_ERR,
				       DMA_CS2 | DMA_TS2 | DMA_CH2_ERR,
				       DMA_CS3 | DMA_TS3 | DMA_CH3_ERR};

	switch (CHNUM_TO_DMACNUM(dmanr)) {
		case 0:
			value = mfdcr(DCRN_DMA0_ASGC);
			break;
#if NR_DMACS > 1
		case 1:
			value = mfdcr(DCRN_DMA1_ASGC);
			break;
#endif /* NR_DMACS */
	}

	/* avoid restarting the finished channels */
	value &= ~(SSG0_MASK_ENABLE |
		   SSG1_MASK_ENABLE |
		   SSG2_MASK_ENABLE |
		   SSG3_MASK_ENABLE);

	switch (dmanr & 0x3) {
		case 0:
			value |= SSG0_ENABLE | SSG0_MASK_ENABLE;
			break;
		case 1:
			value |= SSG1_ENABLE | SSG1_MASK_ENABLE;
			break;
		case 2:
			value |= SSG2_ENABLE | SSG2_MASK_ENABLE;
			break;
		case 3:
			value |= SSG3_ENABLE | SSG3_MASK_ENABLE;
			break;
		default:
			if (dmanr >= MAX_PPC4xx_DMA_CHANNELS)
				printk("%s: bad channel: %d\n", __FUNCTION__, dmanr);
	}

	switch (CHNUM_TO_DMACNUM(dmanr)){
	case 0:
		mtdcr(DCRN_DMA0_SR, status_bits[dmanr]);
		mtdcr(DCRN_DMA0_ASGC, (u32)value);
		break;
#if NR_DMACS > 1
	case 1:
		mtdcr(DCRN_DMA1_SR, status_bits[dmanr - CH_PER_DMAC]);
		mtdcr(DCRN_DMA1_ASGC, (u32)value);
		break;
#endif /* NR_DMACS */
	}
}

static void mpu4xx_dma_disable_dma(int dmanr)
{
	unsigned int value = 0;

	switch (CHNUM_TO_DMACNUM(dmanr)) {
		case 0:
			value = mfdcr(DCRN_DMA0_ASGC);
			break;
#if NR_DMACS > 1
		case 1:
			value = mfdcr(DCRN_DMA1_ASGC);
			break;
#endif /* NR_DMACS */
	}

	/* avoid restarting the finished channels */
	value &= ~(SSG0_MASK_ENABLE |
		   SSG1_MASK_ENABLE |
		   SSG2_MASK_ENABLE |
		   SSG3_MASK_ENABLE);

	switch (dmanr & 0x3) {
		case 0:
			value &= ~(SSG0_ENABLE | SSG0_MASK_ENABLE);
			break;
		case 1:
			value &= ~(SSG1_ENABLE | SSG1_MASK_ENABLE);
			break;
		case 2:
			value &= ~(SSG2_ENABLE | SSG2_MASK_ENABLE);
			break;
		case 3:
			value &= ~(SSG3_ENABLE | SSG3_MASK_ENABLE);
			break;
		default:
			if (dmanr >= MAX_PPC4xx_DMA_CHANNELS)
				printk("%s: bad channel: %d\n", __FUNCTION__, dmanr);
	}

	switch (CHNUM_TO_DMACNUM(dmanr)) {
		case 0:
			mtdcr(DCRN_DMA0_ASGC, (u32)value);
			break;
#if NR_DMACS > 1
		case 1:
			mtdcr(DCRN_DMA1_ASGC, (u32)value);
			break;
#endif /* NR_DMACS */
	}
}

static void
ppc4xx_set_desc_addr(int dmanr, phys_addr_t desc_addr)
{
	switch (dmanr) {
		case 0:
#ifdef PPC4xx_DMA_64BIT
			mtdcr(DCRN_ASGH0, (u32)(desc_addr >> 32));
#endif		
			mtdcr(DCRN_ASG0, (u32)desc_addr);
			break;
		case 1:
#ifdef PPC4xx_DMA_64BIT 
			mtdcr(DCRN_ASGH1, (u32)(desc_addr >> 32));
#endif		
			mtdcr(DCRN_ASG1, (u32)desc_addr);
			break;
		case 2:
#ifdef PPC4xx_DMA_64BIT
			mtdcr(DCRN_ASGH2, (u32)(desc_addr >> 32));
#endif			
			mtdcr(DCRN_ASG2, (u32)desc_addr);
			break;
		case 3:
#ifdef PPC4xx_DMA_64BIT
			mtdcr(DCRN_ASGH3, (u32)(desc_addr >> 32));
#endif			
			mtdcr(DCRN_ASG3, (u32)desc_addr);
			break;
#if NR_DMACS > 1
		case 4:
#ifdef PPC4xx_DMA_64BIT
			mtdcr(DCRN_ASGH4, (u32)(desc_addr >> 32));
#endif		
			mtdcr(DCRN_ASG4, (u32)desc_addr);
			break;
		case 5:
#ifdef PPC4xx_DMA_64BIT 
			mtdcr(DCRN_ASGH5, (u32)(desc_addr >> 32));
#endif		
			mtdcr(DCRN_ASG5, (u32)desc_addr);
			break;
		case 6:
#ifdef PPC4xx_DMA_64BIT
			mtdcr(DCRN_ASGH6, (u32)(desc_addr >> 32));
#endif			
			mtdcr(DCRN_ASG6, (u32)desc_addr);
			break;
		case 7:
#ifdef PPC4xx_DMA_64BIT
			mtdcr(DCRN_ASGH7, (u32)(desc_addr >> 32));
#endif			
			mtdcr(DCRN_ASG7, (u32)desc_addr);
			break;
#endif /* NR_DMACS */
		default:
			if (dmanr >= MAX_PPC4xx_DMA_CHANNELS)
				printk("%s: bad channel: %d\n", __FUNCTION__, dmanr);
	}
}

static void start_dma(dma_info_t *pinfo)
{
	channel_info_t *cinfo = get_channel_info(pinfo);
	if ( unlikely(NULL == cinfo) ) {
		ERROR("get_channel_info() failed.\n");
		return;
	}

	pinfo->state = MPU4XX_DMA_STATE_RUNNING;
	pinfo->info.state = SNSC_DMA_STATE_BUSY;

	cinfo->error_counter = 0;
	cinfo->running_sub_channel = MPU4XX_SUB_CHANNEL(pinfo->channel);

	ppc4xx_set_desc_addr(cinfo->dmanr, pinfo->chain_phys);
	ppc4xx_enable_dma_interrupt(cinfo->dmanr); /* !! exported by another driver */
	mpu4xx_dma_start_dma(cinfo->dmanr);
}

static void mpu4xx_dma_intr_hdlr(int irq, void *devid, struct pt_regs *regs)
{
	channel_info_t *cinfo = (channel_info_t *)devid;
	dma_info_t *pinfo = &(cinfo->dma_info[cinfo->running_sub_channel]);
	int status;
	int dmanr = cinfo->dmanr;
	int dmac = CHNUM_TO_DMACNUM(dmanr);
	unsigned int status_bits[] = { DMA_CH0_ERR,
				       DMA_CH1_ERR,
				       DMA_CH2_ERR,
				       DMA_CH3_ERR};
	unsigned int busy_bits[] = { DMA_SG0,
				     DMA_SG1,
				     DMA_SG2,
				     DMA_SG3};
	int is_busy = 0;

#if DEBUG_LVL & 4
	DEBUG4("entry\n");
	ppc4xx_show_status(dmac);
#endif
	status = ppc4xx_get_dma_status(dmac);

	switch (dmac){
	case 0:
		cinfo->error_counter += (status & status_bits[dmanr]);
		is_busy = (status & busy_bits[dmanr]);
		break;
#if NR_DMACS > 1
	case 1:
		cinfo->error_counter += (status & status_bits[dmanr - CH_PER_DMAC]);
		is_busy = (status & busy_bits[dmanr - CH_PER_DMAC]);
		break;
#endif /* NR_DMACS */
	}

	ppc4xx_clear_dma_status(dmanr); /* !! exported by another driver */

	if (0 == is_busy) {
		if (MPU4XX_DMA_STATE_RUNNING == pinfo->state) {
			/* keep the value only one time */
			pinfo->state = MPU4XX_DMA_STATE_AVAILABLE;
			pinfo->info.state = (cinfo->error_counter) ?
				(SNSC_DMA_STATE_IDLE | SNSC_DMA_STATE_LAST_NG) :
				(SNSC_DMA_STATE_IDLE | SNSC_DMA_STATE_LAST_OK);
			pinfo->info.error_code = 0;

			/*
			 * TODO: prepare lock for changing pinfo->state.
			 *
			 * currently there are no substantial impacts
			 * without locking, because
			 *
			 *   - in critical section, (i.e., mpu4xx_dma_start(),)
			 *     irq is disabled
			 *   - other codes does not distinguish
			 *     RUNNING from AVAILABLE
			 */

			if (SNSC_DMA_NONBLOCK == pinfo->block) {

				snsc_dma_info_t info = pinfo->info;
				/* struct copy */

				/* call callback function */
				(*pinfo->handler)(pinfo->channel, &info, pinfo->handler_arg);
			}

			/* notify that DMA has been finished */
			wake_up_interruptible(&pinfo->wait_q);
		}

		tasklet_schedule(&mpu4xx_dma_tasklet_data);
	}
}

static void mpu4xx_dma_tasklet(unsigned long arg)
{
	int k;
	DEBUG4("entry");

	for (k=0; k<CHANNEL_INFO_MAX; k++) {
		channel_info_t *cinfo = &(channel_info[k]);

		/* lock cinfo */
		spin_lock_irq(&cinfo->channel_info_lock);

		if (MPU4XX_DMA_STATE_RUNNING != cinfo->dma_info[cinfo->running_sub_channel].state) {
			int j;

			/* check pending and start */
			for (j=0; j<SUB_CHANNEL_MAX; j++) {
				int index = (cinfo->running_sub_channel + j + 1) % SUB_CHANNEL_MAX;
				dma_info_t *pinfo = &(cinfo->dma_info[index]);

				if (MPU4XX_DMA_STATE_PENDING == pinfo->state) {
					start_dma(pinfo);
					break;  /* quit the "for" loop */
				}
			}
		}

		/* unlock cinfo */
		spin_unlock_irq(&cinfo->channel_info_lock);
	}
}

#if DEBUG_LVL & 4
/*
 * for debugging
 */
static void print_chain_descriptor(dma_info_t *pinfo)
{
	int j;
	int m;
	u32 *chain_ptr;

	/* chain buffer */
	chain_ptr = (u32 *)pinfo->chain_virtual;

	for (j=0; j<pinfo->chain_num; j++) {
		printk("CHAIN[%2d]: ", j);
		for (m=0; m<8; m++) {
			printk("%08x/", *chain_ptr);
			chain_ptr ++;
		}
		printk("\n");
	}
}
#else /* DEBUG_LVL & 4 */

#define print_chain_descriptor()

#endif /* !(DEBUG_LVL & 4) */

static void mpu4xx_dma_destroy_buf(dma_info_t *pinfo)
{
	if (0 != pinfo->buf) {
		int j;

		for (j=0; j<pinfo->bufnum; j++) {
			if (0 != pinfo->buf[j].addr_virtual) {
				free_pages(pinfo->buf[j].addr_virtual,
					   get_order(pinfo->bufsize));
				pinfo->buf[j].addr_virtual = 0;
			}
		}

		kfree(pinfo->buf);
		pinfo->buf = 0;
	}
}

static int mpu4xx_dma_create_buf(dma_info_t *pinfo)
{
	int j;
	int error_count;

	if ( in_interrupt() ) {
		pinfo->buf = kmalloc(sizeof(dma_info_buf_t) * pinfo->bufnum, GFP_ATOMIC);
	}
	else {
		pinfo->buf = kmalloc(sizeof(dma_info_buf_t) * pinfo->bufnum, GFP_KERNEL);
	}
	if (0 == pinfo->buf) {
		ERROR("kmalloc() failed\n");
		return -ENOMEM;
	}

	memset(pinfo->buf, 0, sizeof(dma_info_buf_t) * pinfo->bufnum);

	error_count = 0;
	for (j=0; j<pinfo->bufnum; j++) {
		pinfo->buf[j].len = 0;
		if ( in_interrupt() ) {
			pinfo->buf[j].addr_virtual =
				__get_free_pages(GFP_ATOMIC,
						get_order(pinfo->bufsize));
		}
		else {
			pinfo->buf[j].addr_virtual =
				__get_free_pages(GFP_KERNEL,
						get_order(pinfo->bufsize));
		}
		if (0 == pinfo->buf[j].addr_virtual) {
			ERROR("out of memory (%d bytes, %d/%d)\n", pinfo->bufsize, j, pinfo->bufnum);
			error_count ++;
			break;  /* quit the "for" loop */
		}
		DEBUG4("buf[%2d]: 0x%08x (phys)\n", j, (u32)virt_to_bus((void *)pinfo->buf[j].addr_virtual));
	}

	if (error_count) {
		mpu4xx_dma_destroy_buf(pinfo);
		return -ENOMEM;
	}

	return 0;
}

/*
 * release allocated table resources.
 */
static void mpu4xx_dma_destroy_table(dma_info_t *pinfo)
{
	if (pinfo->table) {
		kfree(pinfo->table);
		pinfo->table = 0;
	}
}

static int mpu4xx_dma_create_table(dma_info_t *pinfo)
{
	if ( in_interrupt() ) {
		pinfo->table = kmalloc(sizeof(dma_info_table_t) * TABLE_SIZE_MAX,
				       GFP_ATOMIC);
	}
	else {
		pinfo->table = kmalloc(sizeof(dma_info_table_t) * TABLE_SIZE_MAX,
				       GFP_KERNEL);
	}
	if (NULL == pinfo->table) {
		ERROR("kmalloc() failed.\n");
		return -ENOMEM;
	}

	return 0;
}

/*
 * set table as
 *
 *    addr in userspace addr of buf len
 *   +-----------------+-----------+---+
 *   |                 |           |   |
 *   +-----------------+-----------+---+
 *   |                 |           |   |
 *   .                 .           .   .
 *
 *
 */
static int mpu4xx_dma_set_table(dma_info_t *pinfo, snsc_dma_param_t *param)
{
	int j;
	int buf_index;
	u32 free;
	u32 buf_virtual;


	pinfo->table_size = 0;
	buf_index = 0;
	free = pinfo->bufsize;
	buf_virtual = pinfo->buf[buf_index].addr_virtual;

	for (j=0; j<pinfo->bufnum; j++) {
		pinfo->buf[j].len = 0;
	}

	for (j=0; j<param->data_num; j++) {
		u32 len;
		u32 user_addr;

		if (TABLE_SIZE_MAX <= pinfo->table_size) {
			ERROR("TABLE_SIZE_MAX reached.\n");
			return -ENOMEM;
		}

		if (pinfo->mem2pci) {
			user_addr = param->data[j].src.addr;
		}
		else {
			user_addr = param->data[j].dst.addr;
		}

		len = param->data[j].len;
		if (free < len) {
			/* current buffer is full */

			buf_index ++;
			if (pinfo->bufnum <= buf_index) {
				/* no more buffer */
				ERROR("current buffer is full\n");
				return -EINVAL;
			}

			free = pinfo->bufsize;
			if (free < len) {
				/* data is too long to copy */
				ERROR("data is too long to copy\n");
				return -EINVAL;
			}

			/* set virtual addres as next buffer's top */
			buf_virtual = pinfo->buf[buf_index].addr_virtual;
		}

		/* set value to the table */
		pinfo->table[pinfo->table_size].user = user_addr;
		pinfo->table[pinfo->table_size].buf_virtual = buf_virtual;
		pinfo->table[pinfo->table_size].len = len;

		free -= len;
		pinfo->table_size ++;
		buf_virtual += len;
		pinfo->buf[buf_index].len += len;
	}

	return 0;
}

/*
 * caller have to lock cinfo before calling this function.
 * do not call from interrupt handler, since
 * ppc4xx_request_dma_channel() prohibits doing it.
 */
static int channel_info_init(channel_info_t *cinfo)
{
	int ret = -EFAULT;

	if (0 != cinfo->irq) {
		/* already set */
		ERROR("irq is already set.\n");
		return -EFAULT;
	}

	if ( in_interrupt() ) {
		printk("%s: can't call ppc4xx_request_dma_channel() from interrupt context.\n", DEV_NAME);
		return -EFAULT;
	}

	ret = ppc4xx_request_dma_channel(cinfo->dmanr); /* !! exported by another driver */
	if (ret != cinfo->dmanr) {
		return ret;
	}
	cinfo->irq = ppc4xx_get_dma_irq(cinfo->dmanr); /* !! exported by another driver */

	ret = request_irq(cinfo->irq, mpu4xx_dma_intr_hdlr, SA_INTERRUPT,
			  cinfo->name, cinfo);
	if (0 > ret) {
		ppc4xx_free_dma(cinfo->dmanr); /* !! exported by another driver */
		ERROR("request_irq() failed (err=%d)\n", -ret);
		return ret;
	}

	return 0;
}

static int channel_info_cleanup(channel_info_t *cinfo)
{
	if (0 != cinfo->irq) {
		mpu4xx_dma_disable_dma(cinfo->dmanr);
		free_irq(cinfo->irq, cinfo);
		ppc4xx_free_dma(cinfo->dmanr); /* !! exported by another driver */
		cinfo->irq = 0;
	}

	return 0;
}

static int dma_info_init(dma_info_t *pinfo)
{
	if ( in_interrupt() ) {
		pinfo->chain_virtual =
			__get_free_pages(GFP_ATOMIC, get_order(CHAIN_BUFSIZE));
	}
	else {
		pinfo->chain_virtual =
			__get_free_pages(GFP_KERNEL, get_order(CHAIN_BUFSIZE));
	}
	if (0 == pinfo->chain_virtual) {
		ERROR("get_free_pages() failed\n");
		return 1;
	}

	pinfo->chain_phys = virt_to_bus((void *)pinfo->chain_virtual);
	pinfo->info.state = SNSC_DMA_STATE_IDLE;
	pinfo->state = MPU4XX_DMA_STATE_UNUSED;

	pinfo->buf = 0;
	pinfo->bufsize = DEFAULT_BUFSIZE;
	pinfo->bufnum = DEFAULT_BUFNUM;
	pinfo->table = 0;
	pinfo->table_size = 0;
	pinfo->chain_num = 0;

	/* default blocking mode is "blocking" */
	pinfo->handler = 0;
	pinfo->handler_arg = 0;
	pinfo->block = SNSC_DMA_BLOCK;
	return 0;
}

static void dma_info_cleanup(dma_info_t *pinfo)
{
	mpu4xx_dma_destroy_table(pinfo);
	mpu4xx_dma_destroy_buf(pinfo);

	if (0 != pinfo->chain_virtual) {
		free_pages(pinfo->chain_virtual, get_order(CHAIN_BUFSIZE));
		pinfo->chain_virtual = 0;
	}

	pinfo->state = MPU4XX_DMA_STATE_UNUSED;
}

static void mpu4xx_dma_cleanup_all(void)
{
	if (0 != channel_info) {
		int k;

		for (k=0; k<CHANNEL_INFO_MAX; k++) {
			channel_info_t *cinfo = &(channel_info[k]);
			int j;

			for (j=0; j<SUB_CHANNEL_MAX; j++) {
				dma_info_t *pinfo = &(cinfo->dma_info[j]);
				dma_info_cleanup(pinfo);
			}

			channel_info_cleanup(cinfo);
		}

		kfree(channel_info);
		channel_info = 0;
	}

#ifdef CONFIG_MPU4xx_DMACHAIN_PROC
	remove_proc_entry(PROC_NAME, NULL);
#endif

}

static int channel_info_check_and_init(channel_info_t *cinfo, int any)
{
	unsigned long flags;
	int unused = 0;
	int ret = -EFAULT;
	int j;

	/* lock cinfo */
	spin_lock_irqsave(&cinfo->channel_info_lock, flags);

	/*
	 * make sure that when you lock both "pinfo" and
	 * "cinfo" at the same time, the locking order is
	 * "pinfo" first, then "cinfo" next.
	 * breaking this rule results in the deadlock.
	 */

	for (j=0; j<SUB_CHANNEL_MAX; j++) {
		unused +=
			(cinfo->dma_info[j].state == MPU4XX_DMA_STATE_UNUSED);
	}

	if (SUB_CHANNEL_MAX > unused) {
		/* one of them are in use */
		if (0 == any) {
			goto INFO_EXIT;
		}
	}
	else {
		/* none of the sub_channels are used */
		if (0 == in_interrupt()) {
			ret = channel_info_init(cinfo);
			if (ret) {
				goto INFO_EXIT;
			}
		}
	}

	ret = 0;

INFO_EXIT:
	/* unlock cinfo */
	spin_unlock_irqrestore(&cinfo->channel_info_lock, flags);

	return ret;
}

static int request_sub_channel_locked(dma_info_t *pinfo, int any)
{
	channel_info_t *cinfo = 0;
	int ret;

	if (MPU4XX_DMA_STATE_UNUSED != pinfo->state) {
		return -EFAULT;
	}

	/* release previously allocated resources */
	dma_info_cleanup(pinfo);

	ret = dma_info_init(pinfo);
	if (ret) {
		return ret;
	}

	cinfo = get_channel_info(pinfo);
	if ( unlikely(NULL == cinfo) ) {
		dma_info_cleanup(pinfo);
		ERROR("get_channel_info() failed.\n");
		return -EFAULT;
	}

	ret = channel_info_check_and_init(cinfo, any);
	if (ret) {
		dma_info_cleanup(pinfo);
		return -EFAULT;
	}

	pinfo->state = MPU4XX_DMA_STATE_AVAILABLE;

	return 0;
}

static int request_sub_channel(dma_info_t *pinfo, int any)
{
	int ret;

	/*
	 * can't use LOCK_DMA_INFO() macro here.
	 * there is a risk that the same channel is aquired.
	 */
	if ( in_interrupt() ) {
		/*
		 * we cannot sleep in interrupt context.
		 */
		if ( down_trylock(&pinfo->lock) ) {
			return -EFAULT;
		}
	}
	else {
		if ( down_interruptible(&pinfo->lock) ) {
			return -ERESTARTSYS;
		}
	}

	ret = request_sub_channel_locked(pinfo, any);

	up(&pinfo->lock);

	return ret;
}

int mpu4xx_dma_request_channel(channel_t channel_mask, channel_t *channel)
{
	channel_info_t *cinfo = 0;
	dma_info_t *pinfo = 0;
	int k;
	int ret;

	if (NULL == channel) {
		return -EINVAL;
	}

	/*
	 * first loop.
	 * channel is assigned if _none_ of the sub_channels are in use.
	 *
	 * this search sequence enhances the performance.
	 * for example, if one of channel 0 or 1 is requested
	 * when the channel states are described in the table below,
	 * channel 1 is assigned.
	 *
	 *     +-------+-----------+------+
	 *     |channel|sub_channel|state |
	 *     +-------+-----------+------+
	 *     |       |     0     |unused|
	 *     |   0   |     1     |unused|
	 *     |       |     2     |in use|
	 *     +-------+-----------+------+
	 *     |       |     0     |unused|
	 *     |   1   |     1     |unused|
	 *     |       |     2     |unused|
	 *     +-------+-----------+------+
	 */
	for (k=0; k<CHANNEL_INFO_MAX; k++) {
		cinfo = &(channel_info[k]);
		pinfo = &(cinfo->dma_info[0]);

		if (0 == (channel_mask & DMANR_TO_CHANNEL(cinfo->dmanr))) {
			continue;
		}

		ret = request_sub_channel(pinfo, 0);
		if (0 == ret) {
			pinfo->channel = MPU4XX_CHANNEL_T(0, cinfo->dmanr);
			*channel = pinfo->channel;
			return 0;
		}
	}

	/*
	 * second loop.
	 * any sub_channel which is currently unused is assigned.
	 */
	for (k=0; k<CHANNEL_INFO_MAX; k++) {
		int j;

		cinfo = &(channel_info[k]);

		if (0 == (channel_mask & DMANR_TO_CHANNEL(cinfo->dmanr))) {
			continue;
		}

		for (j=0; j<SUB_CHANNEL_MAX; j++) {
			pinfo = &(cinfo->dma_info[j]);
			ret = request_sub_channel(pinfo, 1);
			if (0 == ret) {
				pinfo->channel = MPU4XX_CHANNEL_T(j, cinfo->dmanr);
				*channel = pinfo->channel;
				return 0;
			}
		}
	}

	/*
	 * TODO:
	 *
	 *   assignment algorithm should be re-considered.
	 *
	 *   with current algorithm, if one of channel 0 or 1 is requested
	 *   when the channel states are described in the table below,
	 *   channel 0 is assigned, though channel 1 might be desirable.
	 *
	 *     +-------+-----------+------+
	 *     |channel|sub_channel|state |
	 *     +-------+-----------+------+
	 *     |       |     0     |in use|
	 *     |       |     1     |in use|
	 *     |   0   |     2     |in use|
	 *     |       |     3     |in use|
	 *     |       |     4     |unused|
	 *     +-------+-----------+------+
	 *     |       |     0     |in use|
	 *     |       |     1     |unused|
	 *     |   1   |     2     |unused|
	 *     |       |     3     |unused|
	 *     |       |     4     |unused|
	 *     +-------+-----------+------+
	 */

	return -EFAULT;
}

int mpu4xx_dma_release_channel_locked(dma_info_t *pinfo)
{
	channel_info_t *cinfo = get_channel_info(pinfo);
	int unused = 0;
	int j;
	unsigned long flags;

	if (NULL == cinfo) {
		return -EFAULT;
	}

	if ( in_interrupt() ) {
		/*
		 * since we cannot wait the pending/running transfer
		 * to be completed, this function succeeds only when
		 * being able to set immediately.
		 */
		if (MPU4XX_DMA_STATE_AVAILABLE != pinfo->state) {
			return -EFAULT;
		}
	}
	else {
		/* wait previous transfer to be completed */
		while (MPU4XX_DMA_STATE_AVAILABLE != pinfo->state) {
			wait_event_interruptible(pinfo->wait_q,
						 (MPU4XX_DMA_STATE_AVAILABLE == pinfo->state));
			/*
			 * ignore any interrupts.
			 *
			 * although it takes too much time to
			 * complete, we can avoid stoking
			 * the channel unreleased because
			 * callers tend to ignore the
			 * return value so that they will never
			 * try calling this function again.
			 */
		}
	}

	dma_info_cleanup(pinfo);

	/* lock cinfo */
	spin_lock_irqsave(&cinfo->channel_info_lock, flags);

	for (j=0; j<SUB_CHANNEL_MAX; j++) {
		unused += (MPU4XX_DMA_STATE_UNUSED == cinfo->dma_info[j].state);
	}

	if (SUB_CHANNEL_MAX == unused) {
		/* no sub_channles are in use */
		channel_info_cleanup(cinfo);
	}

	/* unlock cinfo */
	spin_unlock_irqrestore(&cinfo->channel_info_lock, flags);

	return 0;
}

int mpu4xx_dma_release_channel(channel_t channel)
{
	dma_info_t *pinfo = 0;
	int ret = -EFAULT;

	pinfo = get_dma_info(channel);
	if ( unlikely(NULL == pinfo) ) {
		return -EINVAL;
	}

	LOCK_DMA_INFO(pinfo);

	switch (pinfo->state) {
		case MPU4XX_DMA_STATE_AVAILABLE:
		case MPU4XX_DMA_STATE_RUNNING:
		case MPU4XX_DMA_STATE_PENDING:
			ret = mpu4xx_dma_release_channel_locked(pinfo);
			break;

		case MPU4XX_DMA_STATE_UNUSED:
		default:
			ret = 0;
			break;
	}

	UNLOCK_DMA_INFO(pinfo);

	return ret;
}

int mpu4xx_dma_get_channel_info(channel_t channel, snsc_dma_info_t *info)
{
	dma_info_t *pinfo = 0;
	int ret = -EFAULT;

	if (0 == info) {
		ERROR("info is NULL\n");
		return -EINVAL;
	}

	pinfo = get_dma_info(channel);
	if ( unlikely(NULL == pinfo) ) {
		return -EINVAL;
	}

	LOCK_DMA_INFO(pinfo);

	switch (pinfo->state) {
		case MPU4XX_DMA_STATE_AVAILABLE:
		case MPU4XX_DMA_STATE_RUNNING:
		case MPU4XX_DMA_STATE_PENDING:
			*info = pinfo->info;  /* struct copy */
			ret = 0;
			break;

		case MPU4XX_DMA_STATE_UNUSED:
		default:
			ret = -EINVAL;
			break;
	}

	UNLOCK_DMA_INFO(pinfo);

	return ret;
}

#if DEBUG_LVL & 4
static int mpu4xx_dma_print_chain_info(u_int32_t *ptr)
{
	struct descriptor {
		u_int32_t ctrl;
		u_int32_t count;
		u_int64_t src;
		u_int64_t dst;
		u_int64_t next;
	} *desc = (struct descriptor *)ptr;

	printk("%s()\n", __FUNCTION__);
	printk("  ctrl  = 0x%08x\n", desc->ctrl);
	printk("  count = 0x%08x\n", desc->count);
	printk("  src   = 0x%010llx\n", desc->src);
	printk("  dst   = 0x%010llx\n", desc->dst);
	printk("  next  = 0x%010llx\n", desc->next);

	return 0;
}
#endif

/*
 * adds one chain entry.
 * returns
 *
 *    -x - error
 *     0 - still available
 */
static int mpu4xx_dma_add_chain(dma_info_t *pinfo,
				 phys_addr_t src_addr,
				 phys_addr_t dst_addr,
				 u32 len)
{
	u32 *chain_ptr;
	int transfer_width;
#ifdef CONFIG_MPU4xx_DMACHAIN_PROC
	channel_info_t *cinfo = get_channel_info(pinfo);
#endif

	if ((0 == src_addr) ||
	    (0 == dst_addr) ||
	    (0 == len)) {
		ERROR("src_addr/dest_addr/len is zero.\n");
		return -EINVAL;
	}

#if defined (CONFIG_CXD4707_ES_1_0) || defined (CONFIG_CXD4707_ES_1_1)
	/* These versions only allow 128bit transfer */
	if ( (len & 0xf) || (src_addr & 0xf) || (dst_addr & 0xf) ) {
		ERROR("ES1.x allow only 128 bit transfer.\n  Address and length need to be multiple of 16 byte.\n");
		return -EINVAL;
	}
#endif

	if ( !(len & 0xf) && !(src_addr & 0xf) && !(dst_addr & 0xf) ) {
		DEBUG4("PW_128 src=0x%llx, dst=0x%llx\n", 
		       src_addr, dst_addr);
#ifdef CONFIG_MPU4xx_DMACHAIN_PROC
		pw128_cnt[CHNUM_TO_DMACNUM(cinfo->dmanr)][cinfo->dmanr & 3]++;
#endif
		transfer_width = PW_128;  // 16 bytes
		len >>= 4;
	}
	else if ( !(len & 0x7) && !(src_addr & 0x7) && !(dst_addr & 0x7) ) {
		DEBUG4("PW_64 src=0x%llx, dst=0x%llx\n", 
		       src_addr, dst_addr);
#ifdef CONFIG_MPU4xx_DMACHAIN_PROC
		pw64_cnt[CHNUM_TO_DMACNUM(cinfo->dmanr)][cinfo->dmanr & 3]++;
#endif
		transfer_width = PW_64;  // 8 bytes
		len >>= 3;
	}
	else if ( !(len & 0x3) && !(src_addr & 0x3) && !(dst_addr & 0x3) ) {
		DEBUG4("PW_32 src=0x%llx, dst=0x%llx\n", 
		       src_addr, dst_addr);
#ifdef CONFIG_MPU4xx_DMACHAIN_PROC
		pw32_cnt[CHNUM_TO_DMACNUM(cinfo->dmanr)][cinfo->dmanr & 3]++;
#endif
		transfer_width = PW_32;  // 4 bytes
		len >>= 2;
	}
	else if ( !(len & 0x1) && !(src_addr & 0x1) && !(dst_addr & 0x1)) {
		DEBUG4("PW_16 src=0x%llx, dst=0x%llx\n", 
		       src_addr, dst_addr);
#ifdef CONFIG_MPU4xx_DMACHAIN_PROC
		pw16_cnt[CHNUM_TO_DMACNUM(cinfo->dmanr)][cinfo->dmanr & 3]++;
#endif
		transfer_width = PW_16;  // 2 bytes
		len >>= 1;
	}
	else {
		DEBUG4("PW_8 src=0x%llx, dst=0x%llx\n", 
		       src_addr, dst_addr);
#ifdef CONFIG_MPU4xx_DMACHAIN_PROC
		pw8_cnt[CHNUM_TO_DMACNUM(cinfo->dmanr)][cinfo->dmanr & 3]++;
#endif
		transfer_width = PW_8;  // 1 byte
	}

	if (pinfo->chain_num >= (CHAIN_BUFSIZE / BYTES_PER_CHAIN)) {
		/* chain buffer is full */
		ERROR("chain buffer is full.\n");
		return -ENOMEM;
	}

	if (len > SG_COUNT_MASK) {
		ERROR("specified length exceeds SG_COUNT_MASK\n");
		return -EINVAL;
	}

	chain_ptr = (u32 *)pinfo->chain_virtual;
	chain_ptr += (pinfo->chain_num << 3);
       	/* each chain is u32 * 8 (<< 3) */

	if (0 < pinfo->chain_num) {
		/* set previous chain address */
		*(chain_ptr - 7) |= SG_LINK;
		*(phys_addr_t *)(chain_ptr - 2) = virt_to_bus(chain_ptr);

#ifndef ONE_INTERRUPT_PER_CHAIN
		/* disable previous chain's interrupt */
		*(chain_ptr - 7) &= ~(SG_TCI_ENABLE | SG_ETI_ENABLE);
#endif /* ! ONE_INTERRUPT_PER_CHAIN */
	}

	/* DMA Channel Control Word */
	*chain_ptr = DMA_CE_ENABLE | DMA_CIE_ENABLE |
		     SET_DMA_PW(transfer_width) | DMA_DAI | DMA_SAI | DMA_BEN |
		     DMA_MODE_MM | DMA_ETD_OUTPUT | DMA_TCE_ENABLE;
#ifdef CONFIG_CXD4707_ES_2_0
	if (IS_PCI_REGION(src_addr) || IS_PCI_REGION(dst_addr)) {
		*chain_ptr |= PCI_PRIORITY;
	}
#endif
#if defined (CONFIG_CXD4707_ES_1_0) || defined (CONFIG_CXD4707_ES_1_1)
	if (IS_PCI_REGION(src_addr) || IS_PCI_REGION(dst_addr)) {
		/*
		 * DMA_BEN should not be set in ES1.1.
		 * interrupt is not raised.
		 */
		*chain_ptr &= ~DMA_BEN;
	}
#endif
	chain_ptr ++;

	/* Count */
	*chain_ptr = SG_TCI_ENABLE | SG_ETI_ENABLE | SG_ERI_ENABLE | (len & SG_COUNT_MASK);
	chain_ptr ++;

	/* Source Address */
	*(phys_addr_t *)chain_ptr = src_addr;
	chain_ptr += 2;

	/* Destination Address */
	*(phys_addr_t *)chain_ptr = dst_addr;
	chain_ptr += 2;

	/* next chain */
	*(phys_addr_t *)chain_ptr = 0;

	pinfo->chain_num ++;

#if DEBUG_LVL & 4
	mpu4xx_dma_print_chain_info(chain_ptr-6);
#endif

	return 0;
}

/*
 * get physical address associated with the
 * specified userspace's address.
 *
 * returns
 *
 *    -x - error
 *     0 - still available
 */
static int mpu4xx_dma_get_sdram_phys(dma_info_t *pinfo, dma_sdram_t *sdram)
{
	int j;

	for (j=0; j<pinfo->table_size; j++) {
		if (sdram->virtual.addr == pinfo->table[j].user) {
			sdram->phys[0].addr = virt_to_bus((void *)pinfo->table[j].buf_virtual);
			sdram->phys[0].len = pinfo->table[j].len;
			sdram->phys_num = 1;

			return 0;
			/*
			 * NOTE: this is based on the assumption that
			 * mpu4xx_dma_set_table() has been
			 * executed successfully.
			 *
			 * otherwise, no physical address could be found.
			 */
		}
	}

	/* no one must come here */
	ERROR("physical address for specified user address not found.\n");
	return -EINVAL;

}

/*
 * copy userspace's data => buf
 */
static int mpu4xx_dma_copy_from_user(dma_info_t *pinfo)
{
	int j;
	int ret;

	for (j=0; j<pinfo->table_size; j++) {
		ret = copy_from_user((void *)pinfo->table[j].buf_virtual,
				     (void *)pinfo->table[j].user,
				     pinfo->table[j].len);
		/*
		 * NOTE: this is based on the assumption that
		 * mpu4xx_dma_set_table() has been
		 * executed successfully.
		 */

		if (ret) {
			ERROR("copy_from_user() failed\n");
			return -EINVAL;
		}
	}

	return 0;
}

/*
 * copy buf's data => userspace
 */
static int mpu4xx_dma_copy_to_user(dma_info_t *pinfo)
{
	int j;
	int ret;

	for (j=0; j<pinfo->table_size; j++) {
		ret = copy_to_user((void *)pinfo->table[j].user,
				   (void *)pinfo->table[j].buf_virtual,
				   pinfo->table[j].len);
		/*
		 * NOTE: this is based on the assumption that
		 * mpu4xx_dma_set_table() has been
		 * executed successfully.
		 */

		if (ret) {
			ERROR("copy_to_user() failed\n");
			return -EINVAL;
		}
	}

	return 0;
}

/*
 * set chain data.
 * addresses are assumed to be the physical one.
 */
static int mpu4xx_dma_set_chain(dma_info_t *pinfo, snsc_dma_param_t *param)
{
	int j;

	pinfo->chain_num = 0;

	for (j=0; j<param->data_num; j++) {
		int ret;
		ret = mpu4xx_dma_add_chain(pinfo,
					   param->data[j].src.addr,
					   param->data[j].dst.addr,
					   param->data[j].len);

		if (ret) {
			return ret;
		}
	}

	return 0;
}

/*
 * set chain data.
 * addresses are assumed to be the userspace's one.
 *
 * returns
 *
 *    -x - error
 *     0 - success
 */
static int mpu4xx_dma_set_chain_user(dma_info_t *pinfo, snsc_dma_param_t *param)
{
	int ret;
	int j;
	dma_sdram_t *sdram;

	if ( in_interrupt() ) {
		sdram = kmalloc(sizeof(dma_sdram_t), GFP_ATOMIC);
	}
	else {
		sdram = kmalloc(sizeof(dma_sdram_t), GFP_KERNEL);
	}
	if (NULL == sdram) {
		ERROR("kmalloc() failed.");
		return -ENOMEM;
	}
	/*
	 * the size of this variable may be large so that
	 * it should be allocated in the heap area.
	 */

	pinfo->chain_num = 0;

	for (j=0; j<param->data_num; j++) {
		u64 addr_pci_phys;
		int k;

		if (pinfo->mem2pci) {
			sdram->virtual.addr = param->data[j].src.addr;
			addr_pci_phys = param->data[j].dst.addr;
		}
		else {
			sdram->virtual.addr = param->data[j].dst.addr;
			addr_pci_phys = param->data[j].src.addr;
		}

		sdram->virtual.len = param->data[j].len;
		sdram->phys_num = 0;

		/* get sdram physical buffer address */
		ret = mpu4xx_dma_get_sdram_phys(pinfo, sdram);
		if (ret) {
			kfree(sdram);
			return ret;
		}

		for (k=0; k<sdram->phys_num; k++) {
			phys_addr_t addr_sdram_phys;
			u32 len;

			addr_sdram_phys = sdram->phys[k].addr;
			len = sdram->phys[k].len;

			/* add chain entry */
			if (pinfo->mem2pci) {
				ret = mpu4xx_dma_add_chain(pinfo,
							   addr_sdram_phys,
							   addr_pci_phys,
							   len);
			}
			else {
				ret = mpu4xx_dma_add_chain(pinfo,
							   addr_pci_phys,
							   addr_sdram_phys,
							   len);
			}

			if (ret) {
				kfree(sdram);
				return ret;
			}

			addr_pci_phys += len;
		}
	}

	kfree(sdram);
	return 0;
}

static int mpu4xx_dma_set_param_locked(dma_info_t *pinfo, snsc_dma_param_t *param)
{
	int ret;

	if ( in_interrupt() ) {
		/*
		 * since we cannot wait the pending/running transfer
		 * to be completed, this function succeeds only when
		 * being able to set immediately.
		 */
		if (MPU4XX_DMA_STATE_AVAILABLE != pinfo->state) {
			ERROR("state != MPU4XX_DMA_STATE_AVAILABLE\n");			return -EFAULT;
		}
	}
	else {
		/* wait previous transfer to be completed */
		wait_event_interruptible(pinfo->wait_q,
					 (MPU4XX_DMA_STATE_AVAILABLE == pinfo->state));
		if ( signal_pending(current) ) {
			return -ERESTARTSYS;
		}
	}

	pinfo->mem2pci =
		(param->data[0].dst.flags & SNSC_DMA_ADDR_BUS) ? 1 : 0;

	pinfo->is_user =
		((param->data[0].src.flags & SNSC_DMA_ADDR_USER) ||
		 (param->data[0].dst.flags & SNSC_DMA_ADDR_USER)) ?
		1 : 0;

	if (pinfo->is_user) {
		/* allocate table (first time only) */
		if (0 == pinfo->table) {
			ret = mpu4xx_dma_create_table(pinfo);
			if (ret) {
				return ret;
			}
		}

		/* allocate buffer (first time only) */
		if (0 == pinfo->buf) {
			ret = mpu4xx_dma_create_buf(pinfo);
			if (ret) {
				return ret;
			}
		}

		ret = mpu4xx_dma_set_table(pinfo, param);
		if (ret) {
			return ret;
		}
	}

	/* set chain data */
	if (pinfo->is_user) {
		ret = mpu4xx_dma_set_chain_user(pinfo, param);
	}
	else {
		ret = mpu4xx_dma_set_chain(pinfo, param);
	}

	if (ret) {
		return ret;
	}

	/* flush the chain data's cache */
	dma_cache_wback_inv(pinfo->chain_virtual,
			    pinfo->chain_num * BYTES_PER_CHAIN);

	if (pinfo->is_user) {
		int j;

		/* for G-Bus => PCI */
		if (pinfo->mem2pci) {

			/* copy user's data to buffer */
			ret = mpu4xx_dma_copy_from_user(pinfo);
			if (ret) {
				return ret;
			}

			/* flush the copied data's cache */
			for (j=0; j<pinfo->bufnum; j++) {
				if (pinfo->buf[j].len) {
					dma_cache_wback_inv(pinfo->buf[j].addr_virtual, pinfo->buf[j].len);
				}
			}
		}
		else {
			/* invalidate the cache */
			for (j=0; j<pinfo->bufnum; j++) {
				if (pinfo->buf[j].len) {
					dma_cache_wback_inv(pinfo->buf[j].addr_virtual, pinfo->buf[j].len);
				}
			}
		}
	}

	return 0;
}

int mpu4xx_dma_set_param(channel_t channel, snsc_dma_param_t *param)
{
	int j;
	u_int32_t flags_src;
	u_int32_t flags_dst;
	int ret;
	dma_info_t *pinfo = 0;

	if ((0 == param->data) || (0 == param->data_num)) {
		ERROR("data/data_num is zero.\n");
		return -EINVAL;
	}

	/* currently mixed address type is not supported */
	flags_src = param->data[0].src.flags;
	flags_dst = param->data[0].dst.flags;
	for (j=1; j<param->data_num; j++) {
		if ((flags_src != param->data[j].src.flags) ||
		    (flags_dst != param->data[j].dst.flags)) {
			ERROR("src/dst parameter type mixed.\n");
			return -EINVAL;
		}
	}

	/*
	 * neither userspace-to-userspace DMA
	 * nor phys-to-bus address conversion is not implemented
	 */
	if ((0 == (flags_src & SNSC_DMA_ADDR_BUS)) &&
	    (0 == (flags_dst & SNSC_DMA_ADDR_BUS))) {
		ERROR("no user-user DMA supported.\n");
		return -EINVAL;
	}

	pinfo = get_dma_info(channel);
	if ( unlikely(NULL == pinfo) ) {
		return -EINVAL;
	}

	LOCK_DMA_INFO(pinfo);

	switch (pinfo->state) {
		case MPU4XX_DMA_STATE_AVAILABLE:
		case MPU4XX_DMA_STATE_RUNNING:
		case MPU4XX_DMA_STATE_PENDING:
			ret = mpu4xx_dma_set_param_locked(pinfo, param);
			break;

		case MPU4XX_DMA_STATE_UNUSED:
		default:
			ERROR("erroneous pinfo->state\n");
			ret = -EINVAL;
			break;
	}

	UNLOCK_DMA_INFO(pinfo);

	return ret;
}

int mpu4xx_dma_set_blocking_mode_locked(dma_info_t *pinfo, u32 block,
					void (*handler)(channel_t, snsc_dma_info_t *, void *),
					void *handler_arg)
{
	if ( in_interrupt() ) {
		/*
		 * since we cannot wait the pending/running transfer
		 * to be completed, this function succeeds only when
		 * being able to set immediately.
		 */
		if (MPU4XX_DMA_STATE_AVAILABLE != pinfo->state) {
			ERROR("state != MPU4XX_DMA_STATE_AVAILABLE\n");
			return -EFAULT;
		}
	}
	else {
		/* wait previous transfer to be completed */
		wait_event_interruptible(pinfo->wait_q,
					 (MPU4XX_DMA_STATE_AVAILABLE == pinfo->state));
		if ( signal_pending(current) ) {
			return -ERESTARTSYS;
		}
	}

	pinfo->block = block;
	pinfo->handler = handler;
	pinfo->handler_arg = handler_arg;

	return 0;
}

int mpu4xx_dma_set_blocking_mode(channel_t channel,
				 u32 block,
				 void (*handler)(channel_t, snsc_dma_info_t *, void *),
				 void *handler_arg)
{
	int ret;
	dma_info_t *pinfo = 0;

	if ((SNSC_DMA_NONBLOCK != block) && (SNSC_DMA_BLOCK != block)) {
		ERROR("invaild block parameter.\n");
		return -EINVAL;
	}
	switch (block) {
		case SNSC_DMA_NONBLOCK:
			if (0 == handler) {
				ERROR("specified handler is NULL\n");
				return -EINVAL;
			}
			break;

		case SNSC_DMA_BLOCK:
			break;

		default:
			return -EINVAL;
	}

	pinfo = get_dma_info(channel);
	if ( unlikely(NULL == pinfo) ) {
		return -EINVAL;
	}

	LOCK_DMA_INFO(pinfo);

	switch (pinfo->state) {
		case MPU4XX_DMA_STATE_AVAILABLE:
		case MPU4XX_DMA_STATE_RUNNING:
		case MPU4XX_DMA_STATE_PENDING:
			ret = mpu4xx_dma_set_blocking_mode_locked(pinfo, block, handler,
								  handler_arg);
			break;

		case MPU4XX_DMA_STATE_UNUSED:
		default:
			ERROR("erroneous pinfo->state\n");
			ret = -EINVAL;
			break;
	}

	UNLOCK_DMA_INFO(pinfo);

	return ret;
}

static int mpu4xx_dma_start_locked(dma_info_t *pinfo, int *is_up)
{
	unsigned long flags;
	channel_info_t *cinfo = 0;
	int ret = 0;
	int run;
	int j;

	*is_up = 0;

	if ( in_interrupt() ) {
		/*
		 * since we cannot wait the pending/running transfer
		 * to be completed, this function succeeds only when
		 * being able to set immediately.
		 */
		if (MPU4XX_DMA_STATE_AVAILABLE != pinfo->state) {
			ERROR("state != MPU4XX_DMA_STATE_AVAILABLE\n");
			return -EFAULT;
		}

		if (SNSC_DMA_BLOCK == pinfo->block) {
			ERROR("block == SNSC_DMA_BLOCK mode\n");
			return -EINVAL;
		}

		/*
		 * no matter how many sub_channels are pending/running,
		 * set myself as "pending".
		 */
		pinfo->state = MPU4XX_DMA_STATE_PENDING;

		return 0;
	}

	cinfo = get_channel_info(pinfo);
	if ( unlikely(NULL == cinfo) ) {
		return -EFAULT;
	}

	/* wait previous transfer to be completed */
	wait_event_interruptible(pinfo->wait_q,
				 (MPU4XX_DMA_STATE_AVAILABLE == pinfo->state));
	if ( signal_pending(current) ) {
		return -ERESTARTSYS;
	}

	/* lock cinfo */
	spin_lock_irqsave(&cinfo->channel_info_lock, flags);

	run = 0;
	for (j=0; j<SUB_CHANNEL_MAX; j++) {
		run += (MPU4XX_DMA_STATE_RUNNING == cinfo->dma_info[j].state);
	}
	if (run) {
		pinfo->state = MPU4XX_DMA_STATE_PENDING;
	}
	else {
		start_dma(pinfo);
	}

	if (pinfo->mem2pci && (SNSC_DMA_NONBLOCK == pinfo->block)) {
		/*
		 * do not wait for the DMA to be completed.
		 */

		*is_up = 1;
		UNLOCK_DMA_INFO(pinfo);
		spin_unlock_irqrestore(&cinfo->channel_info_lock, flags);
		/*
		 * "pinfo" have to be unlocked _before_ unlocking "cinfo",
		 * otherwise following case may be the problem.
		 *
		 *   1. unlock "cinfo" first
		 *   2. DMA starts => interrupt handler called
		 *   3. call mpu4xx_dma_start() from the
		 *      interrupt context
		 *       => error (because "pinfo" is locked)
		 *
		 * bad design.  I'm so sorry...
		 */

		return 0;
	}

	/* unlock cinfo */
	spin_unlock_irqrestore(&cinfo->channel_info_lock, flags);

	/* wait for the DMA to be completed */
	wait_event_interruptible(pinfo->wait_q,
				 (MPU4XX_DMA_STATE_AVAILABLE == pinfo->state));
	if ( signal_pending(current) ) {
		/*
		 * ignore interrupts and wait for the completion.
		 * without this, user may free the destination
		 * while being accessed by DMA.
		 */
		while (MPU4XX_DMA_STATE_AVAILABLE != pinfo->state) {
			wait_event_interruptible(pinfo->wait_q,
						 (MPU4XX_DMA_STATE_AVAILABLE == pinfo->state));
		}

		/* TODO: stop DMA transfer currently running */

		return -ERESTARTSYS;
	}

	if (pinfo->info.state & SNSC_DMA_STATE_LAST_NG) {
		/* DMA error */
		ERROR("DMA error\n");
		return -EFAULT;
	}

	if (pinfo->is_user) {
		/* for PCI => G-Bus */
		if (!pinfo->mem2pci) {
			/* copy transferred data to user's buffer */
			ret = mpu4xx_dma_copy_to_user(pinfo);
			if (ret) {
				return ret;
			}
		}
	}

	return 0;
}

int mpu4xx_dma_start(channel_t channel)
{
	int ret;
	dma_info_t *pinfo = 0;
	int is_up;

	pinfo = get_dma_info(channel);
	if ( unlikely(NULL == pinfo) ) {
		return -EINVAL;
	}

	/*
	 * callback function is called when transfer is
	 * finished, but not the time when copy_to_user() is
	 * finished.
	 * that's why prohibited here.
	 */
	if (pinfo->is_user && !pinfo->mem2pci &&
	    (pinfo->block == SNSC_DMA_NONBLOCK)) {
		ERROR("pinfo->is_user && !pinfo->mem2pci && (pinfo->block == SNSC_DMA_NONBLOCK)\n");
		return -EFAULT;
	}

	LOCK_DMA_INFO(pinfo);

	is_up = 0;
	switch (pinfo->state) {
		case MPU4XX_DMA_STATE_AVAILABLE:
		case MPU4XX_DMA_STATE_RUNNING:
		case MPU4XX_DMA_STATE_PENDING:
			ret = mpu4xx_dma_start_locked(pinfo, &is_up);
			break;

		case MPU4XX_DMA_STATE_UNUSED:
		default:
			ERROR("erroneous pinfo->state\n");
			ret = -EINVAL;
			break;
	}

	/* unlock pinfo */
	if (0 == is_up) {
		UNLOCK_DMA_INFO(pinfo);
	}

	return ret;
}

/*
 * set buffer size/num of userspace's data.
 *
 * bufsize - size of one buffer (bytes)
 * bufnum  - number of buffer
 *
 * NOTE: it will be unnecessary if zero-copy is implemented.
 */
int mpu4xx_dma_set_user_buffer(channel_t channel, int bufsize, int bufnum)
{
	int ret = 0;
	dma_info_t *pinfo = 0;

	if ((0 >= bufsize) || (0 >= bufnum)) {
		ERROR("bufsize/bufnum invaild\n");
		return -EINVAL;
	}

	pinfo = get_dma_info(channel);
	if ( unlikely(NULL == pinfo) ) {
		return -EINVAL;
	}

	LOCK_DMA_INFO(pinfo);

	if (pinfo->buf) {
		/* already allocated */
		ERROR("pinfo->buf is already allocated\n");
		ret = -EFAULT;
	}
	else {
		pinfo->bufsize = bufsize;
		pinfo->bufnum = bufnum;
	}

	UNLOCK_DMA_INFO(pinfo);

	return ret;
}

static int __init mpu4xx_dma_init(void)
{
	int k;

	channel_info = kmalloc(sizeof(channel_info_t) * CHANNEL_INFO_MAX,
			       GFP_KERNEL);
	if (0 == channel_info) {
		ERROR("kmalloc() failed\n");
		goto err;
	}

	memset(channel_info, 0, sizeof(channel_info_t) * CHANNEL_INFO_MAX);

	for (k=0; k<CHANNEL_INFO_MAX; k++) {
		channel_info_t *cinfo = &(channel_info[k]);
		int j;

		cinfo->channel_info_lock = SPIN_LOCK_UNLOCKED;
		cinfo->dmanr = k;
		cinfo->irq = 0;
		snprintf(cinfo->name, sizeof(cinfo->name), "%s/%d", DEV_NAME, k);
		cinfo->name[sizeof(cinfo->name) - 1] = '\0';

		for (j=0; j<SUB_CHANNEL_MAX; j++) {
			dma_info_t *pinfo = &(cinfo->dma_info[j]);
			pinfo->state = MPU4XX_DMA_STATE_UNUSED;
			init_MUTEX(&pinfo->lock);
			init_waitqueue_head(&pinfo->wait_q);
		}
	}

#ifdef CONFIG_MPU4xx_DMACHAIN_PROC
	/* setup proc entry */
	if (!create_proc_read_entry(PROC_NAME, 0, NULL,
				    mpu4xx_dmachain_read_proc,
				    NULL)) {
		printk(KERN_ERR PROC_NAME
		       "unable to create proc entry\n");
		goto err;
	}
#endif

	return 0;

err:
	mpu4xx_dma_cleanup_all();
	return 1;
}


#ifdef MODULE

MODULE_AUTHOR("Sony Corporation");
MODULE_DESCRIPTION("PPC440GP PCI DMA Controller");

#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif /* MODULE_LICENSE */


static void __exit mpu4xx_dma_exit(void)
{
	mpu4xx_dma_cleanup_all();
}

module_init(mpu4xx_dma_init);
module_exit(mpu4xx_dma_exit);

#else /* MODULE */

static int mpu4xx_dma_init_boottime(void)
{
	mpu4xx_dma_init();
	return 0;
}

__initcall(mpu4xx_dma_init_boottime);

#endif /* MODULE */


EXPORT_SYMBOL(mpu4xx_dma_request_channel);
EXPORT_SYMBOL(mpu4xx_dma_release_channel);
EXPORT_SYMBOL(mpu4xx_dma_get_channel_info);
EXPORT_SYMBOL(mpu4xx_dma_set_param);
EXPORT_SYMBOL(mpu4xx_dma_set_blocking_mode);
EXPORT_SYMBOL(mpu4xx_dma_start);
EXPORT_SYMBOL(mpu4xx_dma_set_user_buffer);
