/*
 *  snsc_mpu22x_powctrl.c - interface to power control of MPU-4xx
 *
 *  Copyright 2003,2004 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/cxd9790_gpio.h>
#include <linux/snsc_major.h>
#include <linux/snsc_powctrl.h>

#include <asm/processor.h>
#include <asm/string.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <platforms/ibm440.h>

#undef DEBUG

#ifdef CONFIG_SNSC_CXD9790
#include <linux/cxd9790.h>
#endif

#ifdef DEBUG
#define DPR(fmt, args...)	printk(fmt, ## args)
#endif /* DEBUG */

/* ======================================================================= *
 * Constant                                                                *
 * ======================================================================= */

#define MPU4xx_POWCTRL_DEVNAME		"powctrl"

#define MPU4xx_POWCTRL_SHTDWNREQ_PORT	GPIO_PORT_Z
#define MPU4xx_POWCTRL_SHTDWNREQ_BITNUM	3
#define MPU4xx_POWCTRL_SHTDWNACK_PORT	GPIO_PORT_Z
#define MPU4xx_POWCTRL_SHTDWNACK_BITNUM	4

MODULE_AUTHOR("Sony PTC");
MODULE_DESCRIPTION("MPU-4xx power control");
MODULE_LICENSE("GPL");

/* ======================================================================= *
 * Debug print function from dpr package                                   *
 * ======================================================================= */

#define DPL_EMERG	0
#define DPL_ALERT	1
#define DPL_CRIT	2
#define DPL_ERR		3
#define DPL_WARNING	4
#define DPL_NOTICE	5
#define DPL_INFO	6
#define DPL_DEBUG	7

#define DPM_PORPOSE	0x0000000f
#define DPM_ERR		0x00000001
#define DPM_TRACE	0x00000002
#define DPM_LAYER	0x000000f0
#define DPM_APPLICATION	0x00000010
#define DPM_MIDDLEWARE	0x00000020
#define DPM_DRIVER	0x00000040
#define DPM_KERNEL	0x00000080
#define DPM_USERDEF	0xffff0000

#ifdef DEBUG

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL DPL_DEBUG
#endif
#ifndef DEBUG_MASK
#define DEBUG_MASK  DPM_PORPOSE
#endif

#ifndef DPR
#include <stdio.h>
#define DPR(fmt, args...)	fprintf(stderr, fmt, ## args)
#endif

#define DPRL(level, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && DPR(fmt, ## args))
#define DPRM(mask, fmt, args...) \
	(void)(((mask) & DEBUG_MASK) && DPR(fmt, ## args))
#define DPRLM(level, mask, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && ((mask) & DEBUG_MASK) && \
		DPR(fmt, ## args))
#define DPRC(cond, fmt, args...) \
	(void)((cond) && DPR(fmt, ## args))
#define DPRF(fmt, args...)	DPR("%s: " fmt, __FUNCTION__, ## args)
#define DPRFL(level, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && DPRF(fmt, ## args))
#define DPRFM(mask, fmt, args...) \
	(void)(((mask) & DEBUG_MASK) && DPRF(fmt, ## args))
#define DPRFLM(level, mask, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && ((mask) & DEBUG_MASK) && \
		DPRF(fmt, ## args))
#define DPRFC(cond, fmt, args...) \
	(void)((cond) && DPRF(fmt, ## args))

#define	DPRFS	DPRFLM(DPL_DEBUG, DPM_TRACE, "start\n")
#define	DPRFE	DPRFLM(DPL_DEBUG, DPM_TRACE, "end\n")
#define DPRFR	DPRFLM(DPL_DEBUG, DPM_TRACE, "return\n")

#else  /* DEBUG */

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL -1
#endif
#ifndef DEBUG_MASK
#define DEBUG_MASK  0
#endif

#ifndef DPR
#define DPR(fmt, args...)
#endif

#define DPRL(level, fmt, args...)
#define DPRM(mask, fmt, args...)
#define DPRLM(level, mask, fmt, args...)
#define DPRC(cond, fmt, args...)
#define DPRF(fmt, args...)
#define DPRFL(level, fmt, args...)
#define DPRFM(mask, fmt, args...)
#define DPRFLM(level, mask, fmt, args...)
#define DPRFC(cond, fmt, args...)

#define	DPRFS
#define	DPRFE
#define DPRFR

#endif /* DEBUG */

/* ======================================================================= *
 * Type and structure                                                      *
 * ======================================================================= */

/* ======================================================================= *
 * Prototype declaration                                                   *
 * ======================================================================= */

/* initialize/finalize */
static int mpu4xx_powctrl_init(void);
static void mpu4xx_powctrl_exit(void);

/* char driver interface */
static int mpu4xx_powctrl_open(struct inode *inode, struct file *filp);
static int mpu4xx_powctrl_release(struct inode *inode, struct file *filp);
static int mpu4xx_powctrl_ioctl(struct inode *inode, struct file *filp,
				unsigned int cmd, unsigned long arg);

/* interrupt handler */
static void mpu4xx_powctrl_shtdwnreq_intr_hdlr(int irq, void *dev_id,
					       struct pt_regs *regs);

/* ioctl internal */
static int mpu4xx_powctrl_ioctl_shutdown(unsigned long arg);

/* ======================================================================= *
 * Global variable                                                         *
 * ======================================================================= */

static int mpu4xx_powctrl_shtdwnreq_asserted = 0;
static DECLARE_WAIT_QUEUE_HEAD(mpu4xx_powctrl_shtdwnreq_wait);

/* ======================================================================= *
 * Inline function                                                         *
 * ======================================================================= */

/* ======================================================================= *
 * Linux driver interface                                                  *
 * ======================================================================= */

static struct file_operations mpu4xx_powctrl_fops = {
	open:		mpu4xx_powctrl_open,
	release:	mpu4xx_powctrl_release,
	ioctl:		mpu4xx_powctrl_ioctl,
};

/* ------------------------ *
 * initialize/finalize      *
 * ------------------------ */

static int
mpu4xx_powctrl_init(void)
{
	int res;
	DPRFS;

	/* setup SHTDWNREQ pin */
	res = cxd9790_gpio_register_bit(MPU4xx_POWCTRL_SHTDWNREQ_PORT,
					MPU4xx_POWCTRL_SHTDWNREQ_BITNUM,
					MPU4xx_POWCTRL_DEVNAME);
	if (res < 0) {
		goto err1;
	}
	res = cxd9790_gpio_set_mode_bit(MPU4xx_POWCTRL_SHTDWNREQ_PORT,
					MPU4xx_POWCTRL_SHTDWNREQ_BITNUM,
					GPIO_MODE_IN);
	if (res < 0) {
		goto err2;
	}
	res = cxd9790_gpio_intr_set_mode_bit(MPU4xx_POWCTRL_SHTDWNREQ_PORT,
					     MPU4xx_POWCTRL_SHTDWNREQ_BITNUM,
					     GPIO_INTR_EDGE_LOW);
	if (res < 0) {
		goto err2;
	}
	res = cxd9790_gpio_intr_register_handler(
		MPU4xx_POWCTRL_SHTDWNREQ_PORT,
		MPU4xx_POWCTRL_SHTDWNREQ_BITNUM,
		mpu4xx_powctrl_shtdwnreq_intr_hdlr,
		&mpu4xx_powctrl_fops);
	if (res < 0) {
		goto err2;
	}

	/* setup SHTDWNACK pin */
	res = cxd9790_gpio_register_bit(MPU4xx_POWCTRL_SHTDWNACK_PORT,
					MPU4xx_POWCTRL_SHTDWNACK_BITNUM,
					MPU4xx_POWCTRL_DEVNAME);
	if (res < 0) {
		goto err3;
	}
	res = cxd9790_gpio_set_mode_bit(MPU4xx_POWCTRL_SHTDWNACK_PORT,
					MPU4xx_POWCTRL_SHTDWNACK_BITNUM,
					GPIO_MODE_OUT_LOW);
	if (res < 0) {
		goto err4;
	}

	/* register this driver */
	res = register_chrdev(SNSC_POWCTRL_MAJOR, MPU4xx_POWCTRL_DEVNAME,
			      &mpu4xx_powctrl_fops);
	if (res < 0) {
		goto err4;
	}

	DPRFE;
	return 0;

err4:
	cxd9790_gpio_unregister_bit(MPU4xx_POWCTRL_SHTDWNACK_PORT,
				    MPU4xx_POWCTRL_SHTDWNACK_BITNUM);
err3:
	cxd9790_gpio_intr_unregister_handler(MPU4xx_POWCTRL_SHTDWNREQ_PORT,
					     MPU4xx_POWCTRL_SHTDWNREQ_BITNUM,
					     &mpu4xx_powctrl_fops);
err2:
	cxd9790_gpio_unregister_bit(MPU4xx_POWCTRL_SHTDWNREQ_PORT,
				    MPU4xx_POWCTRL_SHTDWNREQ_BITNUM);
err1:
	DPRFR;
	return res;
}

static void
mpu4xx_powctrl_exit(void)
{
	DPRFS;

	unregister_chrdev(SNSC_POWCTRL_MAJOR, MPU4xx_POWCTRL_DEVNAME);
	cxd9790_gpio_unregister_bit(MPU4xx_POWCTRL_SHTDWNACK_PORT,
				    MPU4xx_POWCTRL_SHTDWNACK_BITNUM);
	cxd9790_gpio_intr_unregister_handler(MPU4xx_POWCTRL_SHTDWNREQ_PORT,
					     MPU4xx_POWCTRL_SHTDWNREQ_BITNUM,
					     &mpu4xx_powctrl_fops);
	cxd9790_gpio_unregister_bit(MPU4xx_POWCTRL_SHTDWNREQ_PORT,
				    MPU4xx_POWCTRL_SHTDWNREQ_BITNUM);

	DPRFE;
	return;
}

module_init(mpu4xx_powctrl_init);
module_exit(mpu4xx_powctrl_exit);

/* ------------------------ *
 * char driver interface    *
 * ------------------------ */

static int
mpu4xx_powctrl_open(struct inode *inode, struct file *filp)
{
	DPRFS;

	if (MINOR(inode->i_rdev) != 0) {
		DPRFR;
		return -ENODEV;
	}

	MOD_INC_USE_COUNT;

	DPRFE;
	return 0;
}

static int
mpu4xx_powctrl_release(struct inode *inode, struct file *filp)
{
	DPRFS;

	MOD_DEC_USE_COUNT;

	DPRFE;
	return 0;
}

static int
mpu4xx_powctrl_ioctl(struct inode *inode, struct file *filp,
		     unsigned int cmd, unsigned long arg)
{
	DPRFS;

	switch (cmd) {
	case SNSC_POWCTRL_IOC_SHUTDOWN:
		return mpu4xx_powctrl_ioctl_shutdown(arg);
	}

	DPRFR;
	return -EINVAL;
}

/* ======================================================================= *
 * Interrupt handler                                                       *
 * ======================================================================= */

static void
mpu4xx_powctrl_shtdwnreq_intr_hdlr(int irq, void *dev_id, struct pt_regs *regs)
{
	DPRFS;

	mpu4xx_powctrl_shtdwnreq_asserted = 1;
	wake_up_interruptible(&mpu4xx_powctrl_shtdwnreq_wait);

	DPRFE;
	return;
}

/* ======================================================================= *
 * Internal function                                                       *
 * ======================================================================= */

/* ------------------------ *
 * ioctl internal           *
 * ------------------------ */

static int
mpu4xx_powctrl_ioctl_shutdown(unsigned long arg)
{
	int res;
	DPRFS;

	while (1) {
		if (mpu4xx_powctrl_shtdwnreq_asserted) {
			break;
		}

		res = wait_event_interruptible(
			mpu4xx_powctrl_shtdwnreq_wait,
			mpu4xx_powctrl_shtdwnreq_asserted);
		if (res < 0) {
			DPRFR;
			return res;
		}
	}

	DPRFE;
	return 0;
}
