/*
 *  /proc/ppc4xx_reg to read  4xx series' registers.
 *
 *  Copyright 2003 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <platforms/cxd4707gb.h>


/* ======================================================================= *
 * Constants                                                               *
 * ======================================================================= */

#define PPC4XX_REG_PROC_MODULE          "ppc4xx_reg_proc"
#define PPC4XX_REG_PROC_DIRNAME         "ppc4xx_reg"
#define PPC4XX_REG_PROC_MODULE_FILENAME "all"
#define PPC4XX_REG_PROC_MAX_NAMELEN     32

MODULE_AUTHOR("Sony HENC");
MODULE_DESCRIPTION("IBM PowerPC 4xx register viewer for proc filesystem");
MODULE_LICENSE("GPL");


/* ======================================================================= *
 * Debug print function                                                    *
 *   set permutations of:                                                  *
 *     0 : disables all debug output                                       *
 *     1 : Print output. (Default)                                         *
 *     2 : Verbose, dump registers                                         *
 *     4 : Print entry of functions.                                       *
 * ======================================================================= */


#if REGPROC_DEBUG & 1
#define PRINTK(format, args...)                    \
          printk("%s():%d " format ,            \
		 __FUNCTION__ , __LINE__ , ##args)
#else
#define PRINTK(format, args...) 
#endif

#if REGPROC_DEBUG & 2
#define PRINTK2(format, args...)                   \
          printk("%s():%d " format ,            \
		 __FUNCTION__ , __LINE__ , ##args)
#else
#define PRINTK2(format, args...) 
#endif

#if REGPROC_DEBUG & 4
#define ENTRY()                                    \
          printk("%s():%d\n",                   \
		 __FUNCTION__ , __LINE__ )
#else
#define ENTRY() 
#endif

#define ERROR(format, args...)                     \
          printk(KERN_ERR "%s(): " format ,     \
		 __FUNCTION__ , ##args)



#define reg_mfdcri(addr, data, reg)					\
        ({								\
	        mtdcr(addr, reg);					\
		mfdcr(data);						\
	})

#define reg_mtdcri(addr, data, reg)					\
        do {								\
	        mtdcr(addr, reg);					\
		mtdcr(data, reg);					\
	} while (0)

/* ======================================================================= *
 * Type and structure                                                      *
 * ======================================================================= */

/* Class of registers */
enum reg_type {
	REG_SPR,
	REG_DCR,
	REG_DCRI,
	REG_MBASE,
	REG_MMIO,
};

/* Common register class */
typedef struct ppc4xx_reg_t {
	const char		*name;
	unsigned long		 offset;
	unsigned long		 width;
	unsigned short		 low_ino;
} ppc4xx_reg_t;

/* Module class contains table of registers in on-chip peripheral
 * modules and additional infomations (if any). */
typedef struct ppc4xx_reg_module_t {
	const char		*name;
	enum reg_type		 type;
	ppc4xx_reg_t		*regs;
	unsigned long		 opt1; /* DCRI=addr, MMIO=paddr */
	unsigned long		 opt2; /* DCRI=data, MMIO=maplen */
	unsigned long		 opt3; /* DCRI=n/a,  MMIO=vaddr */
	struct proc_dir_entry		*dir;
	unsigned short			 low_ino;
} ppc4xx_reg_module_t;


/* ======================================================================= *
 * Prototype declaration                                                   *
 * ======================================================================= */

static int ppc4xx_reg_proc_init(void);
static void ppc4xx_reg_proc_exit(void);
static ssize_t ppc4xx_reg_proc_module_read(struct file *filp, char *buf,
					   size_t count, loff_t *f_pos);
static ssize_t ppc4xx_reg_proc_reg_read(struct file *filp, char *buf,
					size_t count, loff_t *f_pos);
static ssize_t ppc4xx_reg_proc_reg_write(struct file * filp, const char * buf,
					 size_t count, loff_t *ppos);


/* ======================================================================= *
 * Global variable                                                         *
 * ======================================================================= */

/* Per-processor definitions */
#ifdef CONFIG_CXD4707
#include "cxd4707_regproc.h"
#else
#error unknwon processor.
#endif

static struct proc_dir_entry  *ppc4xx_reg_proc_root;


/* ======================================================================= *
 * Linux driver interface                                                  *
 * ======================================================================= */

static struct file_operations ppc4xx_reg_proc_module_ops = {
	read:	ppc4xx_reg_proc_module_read,
};

static struct file_operations ppc4xx_reg_proc_reg_ops = {
	read:	ppc4xx_reg_proc_reg_read,
	write:	ppc4xx_reg_proc_reg_write,
};

static int 
ppc4xx_reg_proc_init(void)
{
	unsigned long *vaddr;
	int i, j, res;
	ppc4xx_reg_t *regs;
	struct proc_dir_entry   *dir;
	struct proc_dir_entry   *ent;
	ppc4xx_reg_module_t	*mod;

	ENTRY();

	printk("Registering /proc/%s\n", PPC4XX_REG_PROC_DIRNAME);
	ppc4xx_reg_proc_root = proc_mkdir(PPC4XX_REG_PROC_DIRNAME, NULL);
	if (ppc4xx_reg_proc_root == NULL) {
		ERROR("can't create /proc/%s\n", PPC4XX_REG_PROC_DIRNAME);
		res = -ENOMEM;
		goto err1;
	}

	for (i = 0; ppc4xx_reg_proc_modules[i].name != NULL; i++) {
		mod = ppc4xx_reg_proc_modules+i;

		dir = proc_mkdir(mod->name, ppc4xx_reg_proc_root);

		PRINTK2("Registering /proc/%s/%s/\n", 
			PPC4XX_REG_PROC_DIRNAME, mod->name);
		if (dir == NULL) {
			ERROR("can't create /proc/%s/%s\n",
			      PPC4XX_REG_PROC_DIRNAME, mod->name);
			res = -ENOMEM;
			goto err2;
		}
		mod->dir = dir;
		ent = create_proc_entry(PPC4XX_REG_PROC_MODULE_FILENAME,
					S_IRUSR | S_IRGRP | S_IROTH, dir);

		if (mod->type == REG_MMIO) {
			vaddr = ioremap64(mod->opt1, mod->opt2);
			if (vaddr == NULL) {
				ERROR("failed to map MMIO region for %s\n",
				      mod->name);
				goto err3;
			}
			mod->opt3 = (unsigned long)vaddr;
		}

		PRINTK2("Registering /proc/%s/%s/%s\n", 
			PPC4XX_REG_PROC_DIRNAME, 
			mod->name, PPC4XX_REG_PROC_MODULE_FILENAME);
		if (ent == NULL) {
			ERROR("can't create /proc/%s/%s/%s\n",
			      PPC4XX_REG_PROC_DIRNAME, mod->name,
			      PPC4XX_REG_PROC_MODULE_FILENAME);
			res = -ENOMEM;
			goto err3;
		}
		mod->low_ino = ent->low_ino;
		ent->proc_fops = &ppc4xx_reg_proc_module_ops;

		regs = mod->regs;
		for (j = 0; regs[j].name != NULL; j++) {
			ent = create_proc_entry(regs[j].name,
						S_IWUSR | S_IRUSR | S_IRGRP |
						S_IROTH, dir);

			PRINTK2("Registering /proc/%s/%s/%s (%d)\n", 
				PPC4XX_REG_PROC_DIRNAME, 
				mod->name, regs[j].name, ent->low_ino);

			if (ent == NULL) {
				ERROR("can't create /proc/%s/%s/%s\n",
				      PPC4XX_REG_PROC_DIRNAME,
				      mod->name, regs[j].name);
				res = -ENOMEM;
				goto err4;
			}
			regs[j].low_ino = ent->low_ino;
			ent->proc_fops = &ppc4xx_reg_proc_reg_ops;
		}
	}

	return 0;

	/* handle errors */
	do {
		regs = mod->regs;
		do {
			remove_proc_entry(regs[j].name, mod->dir);
err4:
			j--;
		} while (j >= 0);

		remove_proc_entry(PPC4XX_REG_PROC_MODULE_FILENAME,
				  mod->dir);
err3:
		remove_proc_entry(mod->name,
				  ppc4xx_reg_proc_root);
err2:
		i--;
	} while (i >= 0);
	remove_proc_entry(PPC4XX_REG_PROC_DIRNAME, NULL);
err1:
	return res;
}

static void
ppc4xx_reg_proc_exit(void)
{
	ppc4xx_reg_module_t	*mod;
	ppc4xx_reg_t		*regs;
	int i, j;

	ENTRY();

	for (i = 0; ppc4xx_reg_proc_modules[i].name != NULL; i++) {
		mod = ppc4xx_reg_proc_modules + i;
		regs = ppc4xx_reg_proc_modules[i].regs;

		for (j = 0; regs[j].name != NULL; j++) 
			remove_proc_entry(regs[j].name, mod->dir);

		remove_proc_entry(PPC4XX_REG_PROC_MODULE_FILENAME, mod->dir);
		remove_proc_entry(mod->name,  ppc4xx_reg_proc_root);

		if (mod->type == REG_MMIO) 
			iounmap((unsigned int *)mod->opt3);
	}
	remove_proc_entry(PPC4XX_REG_PROC_DIRNAME, NULL);
	printk("Unregistering /proc/%s\n", PPC4XX_REG_PROC_DIRNAME);
	return ;
}

module_init(ppc4xx_reg_proc_init);
module_exit(ppc4xx_reg_proc_exit);


static ssize_t
ppc4xx_reg_proc_module_read(struct file *filp, char *buf,
			      size_t count, loff_t *f_pos)
{
	ppc4xx_reg_module_t	*mod;
	ppc4xx_reg_t	*regs;
	ssize_t		nwritten;
	loff_t		tmp_pos;
	unsigned int	val;
	int		i, j, res, tmp_len, ncopy;
	unsigned short	low_ino;
	char		tmp_buf[PPC4XX_REG_PROC_MAX_NAMELEN + 22];

	res = access_ok(VERIFY_WRITE, buf, count);
	if (!res) 
		return -EFAULT;

	if (count == 0) 
		return 0;

	low_ino = filp->f_dentry->d_inode->i_ino;
	for (i = 0; ppc4xx_reg_proc_modules[i].name != NULL; i++) {
		if (ppc4xx_reg_proc_modules[i].low_ino == low_ino) {
			goto module_found;
		}
	}
	return -ENXIO;

module_found:
	mod = ppc4xx_reg_proc_modules + i;
	regs = ppc4xx_reg_proc_modules[i].regs;

	nwritten = 0;
	tmp_pos = 0;
	val = 0xdeadbeef;
	for (j = 0; regs[j].name != NULL; j++) {
		switch(mod->type) {
		case REG_SPR:
			val = mfspr(regs[j].offset);
			break;
		case REG_DCR:
			val = mfdcr(regs[j].offset);
			break;
		case REG_DCRI:
			val = reg_mfdcri(mod->opt1, mod->opt2, regs[j].offset);
			break;
		case REG_MMIO:
			if ((regs[j].offset & 0x3) == 0x1) {
				res = sprintf(tmp_buf, "%-26s = 0x%02x\n", 
					regs[j].name, 
					*(unsigned char *)(mod->opt3 + regs[j].offset));
			} else if ((regs[j].offset & 0x3) == 0x2) {
				res = sprintf(tmp_buf, "%-26s = 0x%04x\n", 
					regs[j].name, 
					*(unsigned short *)(mod->opt3 + regs[j].offset));
			} else {
				res = sprintf(tmp_buf, "%-26s = 0x%08x\n", 
					regs[j].name, 
					*(unsigned int *)(mod->opt3 + regs[j].offset));
			}
			break;
		default:
			return -EINVAL;
		}

		if (mod->type != REG_MMIO)
			res = sprintf(tmp_buf, "%-26s = 0x%08x\n", regs[j].name, val);

		tmp_len = res;
		if (tmp_pos == *f_pos) {
			ncopy = tmp_len;
			if (ncopy > count) {
				ncopy = count;
			}
			res = copy_to_user(buf + nwritten, tmp_buf, ncopy);
			if (res != 0) {
				return -EFAULT;
			}
			*f_pos += ncopy;
			nwritten += ncopy;
			count -= ncopy;
			if (count == 0) {
				break;
			}
		} else if (tmp_pos < *f_pos && tmp_pos + tmp_len > *f_pos) {
			ncopy = tmp_pos + tmp_len - *f_pos;
			if (ncopy > count) {
				ncopy = count;
			}
			res = copy_to_user(buf + nwritten,
					   tmp_buf + (*f_pos - tmp_pos),
					   ncopy);
			if (res != 0) {
				return -EFAULT;
			}
			*f_pos += ncopy;
			nwritten += ncopy;
			count -= ncopy;
			if (count == 0) {
				break;
			}
		}
		tmp_pos += tmp_len;
	}
	return nwritten;
}


static ssize_t
ppc4xx_reg_proc_reg_read(struct file *filp, char *buf,
			   size_t count, loff_t *f_pos)
{
	ppc4xx_reg_module_t	*mod;
	ppc4xx_reg_t	*regs;
	ssize_t		nwritten;
	unsigned int	val;
	int		i, j, res, tmp_len, ncopy;
	unsigned short	low_ino;
	char		tmp_buf[PPC4XX_REG_PROC_MAX_NAMELEN + 22];

	ENTRY();

	res = access_ok(VERIFY_WRITE, buf, count);
	if (!res)
		return -EFAULT;

	if (count == 0)
		return 0;

	low_ino = filp->f_dentry->d_inode->i_ino;
	PRINTK2("Searching %d\n", low_ino);
	for (i = 0; ppc4xx_reg_proc_modules[i].name != NULL; i++) {
		mod = ppc4xx_reg_proc_modules + i;
		regs = ppc4xx_reg_proc_modules[i].regs;
		for (j = 0; regs[j].name != NULL; j++) {
			PRINTK2("%s:%s (%d)\n", mod->name, regs[j].name,regs[j].low_ino);
			if (regs[j].low_ino == low_ino) {
				goto found;
			}
		}
	}
	PRINTK("Not registered in table\n");
	return -ENXIO;

found:
	nwritten = 0;
	PRINTK2("%s(%ld) type=%d\n", 
	       regs[j].name, regs[j].offset, mod->type);

	switch(mod->type) {
	case REG_SPR:
		val = mfspr(regs[j].offset);
		break;
	case REG_DCR:
		val = mfdcr(regs[j].offset);
		break;
	case REG_DCRI:
		val = reg_mfdcri(mod->opt1, mod->opt2, regs[j].offset);
		break;
	case REG_MMIO:
		if ((regs[j].offset & 0x3) == 0x1) {
			res = sprintf(tmp_buf, "%-26s = 0x%02x\n", 
				      regs[j].name, 
				      *(unsigned char *)(mod->opt3 + regs[j].offset));
		} else if ((regs[j].offset & 0x3) == 0x2) {
			res = sprintf(tmp_buf, "%-26s = 0x%04x\n", 
				      regs[j].name, 
				      *(unsigned short *)(mod->opt3 + regs[j].offset));
		} else {
			res = sprintf(tmp_buf, "%-26s = 0x%08x\n", 
				      regs[j].name, 
				      *(unsigned int *)(mod->opt3 + regs[j].offset));
		}
		break;
	default:
		return -EINVAL;
	}
	if (mod->type != REG_MMIO)
		res = sprintf(tmp_buf, "%-26s = 0x%08x\n", regs[j].name, val);

	tmp_len = res;
	if (*f_pos == 0) {
		ncopy = tmp_len;
		if (ncopy > count) {
			ncopy = count;
		}
		res = copy_to_user(buf, tmp_buf, ncopy);
		if (res != 0) {
			ERROR("copy_to_user() failed\n");
			return -EFAULT;
		}
		*f_pos += ncopy;
		nwritten += ncopy;
		count -= ncopy;
	} else if (0 < *f_pos && tmp_len > *f_pos) {
		ncopy = tmp_len - *f_pos;
		if (ncopy > count) {
			ncopy = count;
		}
		res = copy_to_user(buf, tmp_buf + *f_pos, ncopy);
		if (res != 0) {
			ERROR("copy_to_user() failed\n");
			return -EFAULT;
		}
		*f_pos += ncopy;
		nwritten += ncopy;
		count -= ncopy;
	}

	return nwritten;
}


static ssize_t
ppc4xx_reg_proc_reg_write(struct file * filp, const char * buf,
			  size_t count, loff_t *ppos)
{
	ppc4xx_reg_module_t	*mod;
	ppc4xx_reg_t	*regs;
	unsigned long	newRegValue;
	int		i, j, res;
	unsigned short	low_ino;
	char *endp;

	ENTRY();

	res = access_ok(VERIFY_WRITE, buf, count);
	if (!res)
		return -EFAULT;

	if (count == 0)
		return 0;

	low_ino = filp->f_dentry->d_inode->i_ino;
	for (i = 0; ppc4xx_reg_proc_modules[i].name != NULL; i++) {
		mod = ppc4xx_reg_proc_modules + i;
		regs = ppc4xx_reg_proc_modules[i].regs;
		for (j = 0; regs[j].name != NULL; j++) {
			if (regs[j].low_ino == low_ino) {
				goto found;
			}
		}
	}
	PRINTK("Not registered in table\n");
	return -ENXIO;

found:
	newRegValue = simple_strtoul(buf,&endp,0);

	PRINTK2("%s(%ld) type=%d <= %lx\n", 
	       regs[j].name, regs[j].offset, mod->type, newRegValue);


	switch(mod->type) {
	case REG_SPR:
		mtspr(regs[j].offset, newRegValue);
		break;
	case REG_DCR:
		mtdcr(regs[j].offset, newRegValue);
		break;
	case REG_DCRI:
		reg_mtdcri(mod->opt1, newRegValue, regs[j].offset);
		break;
	case REG_MMIO:
		*(unsigned int *)(mod->opt3 + regs[j].offset) = newRegValue;
		break;
	default:
		return -EINVAL;
	}
	return (count+endp-buf);


}

