/*
 *  Sony NSC Watchdog Timer platform dependent part for MPU-4xx.
 *
 *  Copyright 2003 Sony Corporation
 *
 *  This program is free software; you can redistribute	 it and/or modify it
 *  under  the terms of	 the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the	License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED	  ``AS	IS'' AND   ANY	EXPRESS OR IMPLIED
 *  WARRANTIES,	  INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO	EVENT  SHALL   THE AUTHOR  BE	 LIABLE FOR ANY	  DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED	  TO, PROCUREMENT OF  SUBSTITUTE GOODS	OR SERVICES; LOSS OF
 *  USE, DATA,	OR PROFITS; OR	BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN	 CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/config.h>

#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/module.h>
#include <linux/seq_file.h>
#include <linux/init.h>
#include <linux/snsc_watchdog.h>

#include <asm/processor.h>
#include <asm/io.h>
#include <asm/string.h>
#include <asm/uaccess.h>
#include <asm/div64.h>
#include <platforms/ibm440.h>

#undef DEBUG

/* ======================================================================= *
 * Constants                                                                *
 * ======================================================================= */

#define NAME	"SNSC MPU-4xx WDT"

/* ======================================================================= *
 * Types and structures                                                     *
 * ======================================================================= */

/* because it is big-endian machine, hi-32bit must come first */
struct ibm440gp_wdt_limit {
	u_int32_t hi;
	u_int32_t lo;
};


/* ======================================================================= *
 * Prototype declaration                                                   *
 * ======================================================================= */

static u_int32_t ppc440_wdt_get_reset_cause(void);
static void ppc440_wdt_clear_reset_cause(void);
void ppc440_wdt_heartbeat(void);
static u_int32_t ppc440_wdt_get_count(void);
static void ppc440_wdt_set_count(u_int32_t sec);
static void ppc440_wdt_enable(void);
static void ppc440_wdt_disable(void);
static int ppc440_wdt_is_enabled(void);
int _ppc440_wdt_limit_expired(void);
static u_int32_t _ppc440_wdt_get_sec_to_deadline(void);
static void _ppc440_wdt_set_deadline(u_int32_t sec);

struct watchdog_md *mach_watchdog_init(void);

/* ======================================================================= *
 * Global variables                                                        *
 * ======================================================================= */

extern struct ibm440_clocks ibm440_clks;
static struct ibm440gp_wdt_limit ppc440_wdt_limit;
static struct watchdog_md ppc440_wdt_ops = {
	disable:		ppc440_wdt_disable,
	enable:			ppc440_wdt_enable,
	is_enabled:		ppc440_wdt_is_enabled,
	set_count:		ppc440_wdt_set_count,
	get_count:		ppc440_wdt_get_count,
	heartbeat:		ppc440_wdt_heartbeat,
	get_reset_cause:	ppc440_wdt_get_reset_cause,
	clear_reset_cause:	ppc440_wdt_clear_reset_cause,
};


/* ======================================================================= *
 * SNSC Watchdog Timer interface functions                                 *
 * ======================================================================= */

#ifdef DEBUG
static int check_regs(void)
{
	u_int32_t msr, ivpr, ivor_12; 	/* system related */
	u_int32_t tcr, tsr;		/* watchdog related */

	msr = mfmsr();
	printk("msr    = 0x%08x, (MSG_CE = %x)\n", msr, (msr & MSR_CE) ? 1 : 0);

	ivpr = mfspr(SPRN_IVPR);
	ivor_12 = mfspr(SPRN_IVOR12);
	printk("WDTvec = 0x%08x ", (ivpr & 0xffff0000) | (ivor_12 & 0x0000fff0));
	printk("(ivpr   = 0x%08x, ivor_12 = 0x%08x)\n", ivpr, ivor_12);

	tcr = mfspr(SPRN_TCR);
	tsr = mfspr(SPRN_TSR);
	printk("tcr    = 0x%08x (WP = %x, WRC = %x, WIE = %x)\n", tcr, 
	       (tcr & TCR_WP_MASK)>>30, (tcr & TCR_WRC_MASK)>>28, (tcr & TCR_WIE) ? 1 : 0);
	printk("tsr    = 0x%08x (ENW = %x, WIS = %x, WRS = %x)\n",
	       tsr, (tsr & TSR_ENW) ? 1 : 0, (tsr & TSR_WIS) ? 1 : 0, 
	       (tsr & TSR_WRS(3))>>28);

	return 0;
}
#endif

static void
ppc440_wdt_disable(void)
{
	if ((mfspr(SPRN_TCR) & TCR_WRC_MASK) == 0x0) {
#ifdef DEBUG
		printk("No watchdog is set\n");
		check_regs();
#endif
		return;
	}

	/* 
	   We do not disable watchdog timer interrupt to occur, as wdt
	   module seems like it cannot be stopped once it is set.

	   To avoid this, we make timer period longer(21.4s) and keep
	   clearing watchdog timer exceptions silently in background.
	*/
	mtspr(SPRN_TSR, TSR_WIS);
	mtspr(SPRN_TCR, (mfspr(SPRN_TCR) & ~TCR_WP_MASK) | 
	      TCR_WP(WP_2_29) | 	/* wdt period = 21.4s  */
	      TCR_WIE); 		/* enable wdt interrupt */

	_ppc440_wdt_set_deadline(0);

#ifdef DEBUG
	check_regs();
#endif
}

static void
ppc440_wdt_enable(void)
{
	/* clear WIS */
	mtspr(SPRN_TSR, TSR_WIS);

	/* set watchdog timer */
	mtspr(SPRN_TCR, (mfspr(SPRN_TCR) & ~TCR_WP_MASK) | 
	      TCR_WP(WP_2_25) | 	/* wdt period = 1.3s  */
	      TCR_WRC(WRC_SYSTEM) | 	/* system reset at final wdt */
	      TCR_WIE); 		/* enable wdt interrupt */

#ifdef DEBUG
	check_regs();
#endif
}

static int
ppc440_wdt_is_enabled(void)
{
	u_int32_t tcr;
	tcr = mfspr(SPRN_TCR);
	return (tcr & TCR_WIE);
}

static void
ppc440_wdt_set_count(u_int32_t sec)
{

	/* 
	 * Watchdog periid of 1 sec cannot be set, as 440GP's watchdog
	 * period set at ppc440_wdt_enable() is 1.3 sec. 
	 * Round it to 2 secound.
	 */
	if (sec == 1) sec = 2;

	_ppc440_wdt_set_deadline(sec);
}

static u_int32_t
ppc440_wdt_get_count(void)
{
	return _ppc440_wdt_get_sec_to_deadline();
}

void
ppc440_wdt_heartbeat(void)
{
	mtspr(SPRN_TSR, TSR_WIS);
}

static void
ppc440_wdt_clear_reset_cause(void)
{
	mtspr(SPRN_TSR, TSR_WRS(3));
}

static u_int32_t
ppc440_wdt_get_reset_cause(void)
{
	return (mfspr(SPRN_TSR)>>28) & 0x3;
}


struct watchdog_md *mach_watchdog_init(void)
{
	return &ppc440_wdt_ops;
}

/* ======================================================================= *
 * Utility functions                                                       *
 * ======================================================================= */

/* 
 * PPC440 series have only a fixed length watchdog timer.
 *
 * Thus we estimate watchdog expriration time and compare them with
 * Time Base registers at each occurence of watchdog timer exception.
 */
static void
_ppc440_wdt_set_deadline(u_int32_t sec)
{
	u_int64_t current, limit;
	u_int64_t *wdt_limit = (u_int64_t *)&ppc440_wdt_limit;
	u_int32_t current_lo;

	/* stop watchdog timer */
	if (sec == 0) {
		*wdt_limit = 0;
		return ;
	}

	/* get current time base value */
	current    = mfspr(SPRN_TBRU);
	current_lo = mfspr(SPRN_TBRL);
	current    = (current << 32) | current_lo;

	/* get wdt expire limit */
	limit = sec;
	limit *= (ibm440_clks.cpu != 0) ? ibm440_clks.cpu : 400000000;
	limit += current;

	/*
	 * This is the rare case (once in 1462 yrs at 400Mhz) where
	 * u_int64_t addition has overflowed.  
	 *
	 * Here we reset Time Base counters to zero.
	 */
	if (limit < current) {
		printk(NAME ": WDT limit overflow, Time Base counter reset done to fix the problem");

		mtspr(SPRN_TBWU, 0x0);
		mtspr(SPRN_TBWL, 0x0);

		limit = sec;
		limit *= ibm440_clks.cpu;
	}

	*wdt_limit = limit;

#ifdef DEBUG
	printk("%s(): current(0x%llx), limit(0x%llx)\n", 
	       __FUNCTION__, 
	       current, *wdt_limit);
#endif
}

static u_int32_t
_ppc440_wdt_get_sec_to_deadline(void)
{
	u_int64_t current, timeleft_64;
	u_int32_t current_lo;
	u_int64_t *wdt_limit = (u_int64_t *)&ppc440_wdt_limit;
	u_int32_t timeleft_32;

	if (*wdt_limit == 0)
		return 0;

	/* get current time base value */
	current    = mfspr(SPRN_TBRU);
	current_lo = mfspr(SPRN_TBRL);
	current = current << 32 | current_lo;

	timeleft_64 = *wdt_limit - current;

#ifdef DEBUG
	printk("%s(): current(0x%llx), limit(0x%llx), timeleft(0x%llx)\n",
	       __FUNCTION__, 
	       current, *wdt_limit, timeleft_64);
#endif

	do_div(timeleft_64, ibm440_clks.cpu); /* to second */

	timeleft_32 = (u_int32_t)(timeleft_64 & 0xffffffff); /* lower half */

	return timeleft_32;
}

/* 
 * This function is called from watchdog timer trap handler.
 *
 * Returns:
 *  1 : Expired. current time base  > time base limit
 *  0 : Not yet. current time base <= time base limit
 */
int
_ppc440_wdt_limit_expired(void)
{
	u_int64_t current;
	u_int32_t current_lo;
	u_int64_t *wdt_limit = (u_int64_t *)&ppc440_wdt_limit;

#ifdef DEBUG
	printk("%s(): ppc440_wdt enabled?=0x%08x, *wdt_limit=0x%llx\n", 
	       __FUNCTION__, 
	       ppc440_wdt_is_enabled(), *wdt_limit);
#endif
	/* if wdt is not enabled, it never expires */
	if (ppc440_wdt_is_enabled() == 0)
		return 0;

	if (*wdt_limit == 0)
		return 0;

	/* get current time base value */
	current    = mfspr(SPRN_TBRU);
	current_lo = mfspr(SPRN_TBRL);
	current = current << 32 | current_lo;

#ifdef DEBUG
	printk("%s(): current(0x%llx), limit(0x%llx)\n", 
	       __FUNCTION__, 
	       current, *wdt_limit);
#endif

	if (current > *wdt_limit)
		return 1;
	else
		return 0;
}
