/*
 *  cryptowrap-cipher.h - glue for cipher function of CryptoAPI
 *
 *  Copyright 2003 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef _CRYPTOWRAP_CIPHER_H_
#define _CRYPTOWRAP_CIPHER_H_

typedef struct CryptoWrap_cryptoapi_cipher {
	struct cipher_implementation *ci;
	struct cipher_context *cx;
	int mode;
} cryptoapi_cipher_object;

static inline int
CIPHER_new(cryptoapi_cipher_object **cop, char *name, unsigned int mode)
{
	char ci_name[16];
	cryptoapi_cipher_object *co;
	struct cipher_implementation *ci;
	struct cipher_context *cx;

	strcpy(ci_name, name);

	switch (mode) {
	case CryptoWrap_CIPHER_MODE_ECB:
		strcat(ci_name,"-ecb");
		break;
	case CryptoWrap_CIPHER_MODE_CBC:
		strcat(ci_name, "-cbc");
		break;
	case CryptoWrap_CIPHER_MODE_CFB:
		strcat(ci_name, "-cfb8");
		break;
	case CryptoWrap_CIPHER_MODE_OFB:
		strcat(ci_name, "-ofb8");
		break;
	default:
		return CryptoWrap_E_NOTSUPPORTED;
	}

	ci = find_cipher_by_name(ci_name, 1);
	if (ci == NULL)
		return CryptoWrap_E_NOTSUPPORTED;

	ci->lock();

	cx = ci->realloc_context(NULL, ci, 0);
	if (cx == NULL) {
		ci->unlock();
		return CryptoWrap_E_NOMEM;
	}

	co = kmalloc(sizeof(cryptoapi_cipher_object), GFP_KERNEL);
	if (co == NULL) {
		ci->wipe_context(cx);
		ci->free_context(cx);
		ci->unlock();
		return CryptoWrap_E_NOMEM;
	}

	co->ci = ci;
	co->cx = cx;
	co->mode = mode;
	*cop = co;
	MOD_INC_USE_COUNT;
	return CryptoWrap_SUCCESS;
}

static inline int
CIPHER_destroy(cryptoapi_cipher_object *co)
{
	co->ci->wipe_context(co->cx);
	co->ci->free_context(co->cx);
	co->ci->unlock();
	kfree(co);
	MOD_DEC_USE_COUNT;
	return CryptoWrap_SUCCESS;
}

static inline int
CIPHER_set_key(cryptoapi_cipher_object *co, unsigned char *key, int key_len)
{
	int error;

	error = co->ci->set_key(co->cx, key, key_len);
	if (error == -1)
		return CryptoWrap_E_BADPARITY;
	if (error == -2)
		return CryptoWrap_E_WEAKKEY;
	if (error < 0)
		return CryptoWrap_E_INTERNAL;

	return CryptoWrap_SUCCESS;
}

static inline int
CIPHER_encrypt(cryptoapi_cipher_object *co, unsigned char *from, int flen, unsigned char *to, int *tlen, unsigned char *iv)
{
	int error;

	error = co->ci->_encrypt(co->cx, from, to, flen, 1, iv);
	if (error < 0)
		return CryptoWrap_E_INTERNAL;

	if ((co->mode == CryptoWrap_CIPHER_MODE_ECB
		|| co->mode == CryptoWrap_CIPHER_MODE_CBC)
	    && flen % 8)
		*tlen = flen - flen % 8 + 8;
	else
		*tlen = flen;

	return CryptoWrap_SUCCESS;
}

static inline int
CIPHER_decrypt(cryptoapi_cipher_object *co, unsigned char *from, int flen, unsigned char *to, int *tlen, unsigned char *iv)
{
	int error;

	error = co->ci->_decrypt(co->cx, from, to, flen, 1, iv);
	if (error < 0)
		return CryptoWrap_E_INTERNAL;

	if ((co->mode == CryptoWrap_CIPHER_MODE_ECB
		|| co->mode == CryptoWrap_CIPHER_MODE_CBC)
	    && flen % 8)
		*tlen = flen - flen % 8 + 8;
	else
		*tlen = flen;

	return CryptoWrap_SUCCESS;
}

#endif /* _CRYPTOWRAP_CIPHER_H_ */
