/*
 *  cryptowrap-hmac.c - glue for HMAC function of CryptoAPI
 *
 *  Copyright 2003 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/module.h>
#include <linux/version.h>

#include <linux/init.h>
#include <linux/config.h>
#include <linux/string.h>
#include <linux/crypto.h>

#include <linux/cryptowrap.h>

/*
 *  CryptoWrap for HMAC in CryptoAPI
 */

typedef struct CryptoWrap_cryptoapi_hmac {
	struct digest_implementation *di;
	struct digest_context *idx;
	struct digest_context *odx;
} cryptoapi_hmac_object;

int
CryptoWrap_HMAC_new(CryptoWrap_HMAC_object *objp, int type)
{
	char *name;
	cryptoapi_hmac_object *co;
	struct digest_implementation *di;
	struct digest_context *idx, *odx;

	switch (type) {
	case CryptoWrap_HMAC_DIGEST_SHA1:
		name = "sha1";
		break;
	case CryptoWrap_HMAC_DIGEST_MD5:
		name = "md5";
		break;
	default:
		return CryptoWrap_E_NOTSUPPORTED;
	}

	di = find_digest_by_name(name, 1);
	if (di == NULL)
		return CryptoWrap_E_NOTSUPPORTED;

	di->lock();

	idx = di->realloc_context(NULL, di);
	if (idx == NULL) {
		di->unlock();
		return CryptoWrap_E_NOMEM;
	}

	odx = di->realloc_context(NULL, di);
	if (odx == NULL) {
		di->free_context(idx);
		di->unlock();
		return CryptoWrap_E_NOMEM;
	}

	co = kmalloc(sizeof(cryptoapi_hmac_object), GFP_KERNEL);
	if (co == NULL) {
		di->free_context(idx);
		di->free_context(odx);
		di->unlock();
		return CryptoWrap_E_NOMEM;
	}
	co->di = di;
	co->idx = idx;
	co->odx = odx;
	*objp = (CryptoWrap_HMAC_object)co;
	MOD_INC_USE_COUNT;
	return CryptoWrap_SUCCESS;
}

int
CryptoWrap_HMAC_destroy(CryptoWrap_HMAC_object obj)
{
	cryptoapi_hmac_object *co = (cryptoapi_hmac_object *)obj;

	co->di->free_context(co->idx);
	co->di->free_context(co->odx);
	co->di->unlock();
	kfree(co);
	MOD_DEC_USE_COUNT;
	return CryptoWrap_SUCCESS;
}

void
CryptoWrap_HMAC_Init(CryptoWrap_HMAC_object obj, const unsigned char *key, int keylen)
{
	cryptoapi_hmac_object *co = (cryptoapi_hmac_object *)obj;
	unsigned char intkey[CryptoWrap_HMAC_BLOCKSIZE];
	unsigned char outkey[CryptoWrap_HMAC_BLOCKSIZE];
	int i;

	memset(intkey, 0, CryptoWrap_HMAC_BLOCKSIZE);
	memset(outkey, 0, CryptoWrap_HMAC_BLOCKSIZE);

	if (keylen > CryptoWrap_HMAC_BLOCKSIZE) {
		co->di->_open(co->idx, 1);
		co->di->_update(co->idx, key, keylen, 1);
		co->di->_close(co->idx, intkey, 1);
	}
	else {
		memcpy(intkey, key, keylen);
	}

	memcpy(outkey, intkey, CryptoWrap_HMAC_BLOCKSIZE);

	for (i = 0; i < CryptoWrap_HMAC_BLOCKSIZE; i++) {
		*(intkey + i) ^= 0x36;
		*(outkey + i) ^= 0x5c;
	}

	co->di->_open(co->idx, 1);
	co->di->_update(co->idx, intkey, CryptoWrap_HMAC_BLOCKSIZE, 1);

	co->di->_open(co->odx, 1);
	co->di->_update(co->odx, outkey, CryptoWrap_HMAC_BLOCKSIZE, 1);
}

void
CryptoWrap_HMAC_Update(CryptoWrap_HMAC_object obj, const unsigned char *data, int len)
{
	cryptoapi_hmac_object *co = (cryptoapi_hmac_object *)obj;

	co->di->_update(co->idx, data, len, 1);
}

void
CryptoWrap_HMAC_Final(CryptoWrap_HMAC_object obj, unsigned char *md, int *len)
{
	cryptoapi_hmac_object *co = (cryptoapi_hmac_object *)obj;
	unsigned char intmd[20];	/* XXX */

	memset(intmd, 0, 20);

	co->di->_close(co->idx, intmd, 1);
	co->di->_update(co->odx, intmd, co->di->blocksize, 1);
	co->di->_close(co->odx, md, 1);
	if(len)
	{
		*len = co->di->blocksize;
	}
}

static int CryptoWrap_HMAC_module_init(void);
static void CryptoWrap_HMAC_module_exit(void);

static int
CryptoWrap_HMAC_module_init()
{
	printk(KERN_INFO "CryptoWrap_cryptoapi: loaded\n");
	return 0;
}

static void
CryptoWrap_HMAC_module_exit()
{
	printk(KERN_INFO "CryptoWrap_cryptoapi: unloaded\n");
}

module_init(CryptoWrap_HMAC_module_init);
module_exit(CryptoWrap_HMAC_module_exit);

EXPORT_SYMBOL(CryptoWrap_HMAC_new);
EXPORT_SYMBOL(CryptoWrap_HMAC_destroy);
EXPORT_SYMBOL(CryptoWrap_HMAC_Init);
EXPORT_SYMBOL(CryptoWrap_HMAC_Update);
EXPORT_SYMBOL(CryptoWrap_HMAC_Final);
