/*
 *  cryptowrap-rsa-glue.c - glue for RSA function of GnuPG
 *
 *  Copyright 2003 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/module.h>
#include <linux/version.h>

#include <linux/kernel.h>
#include <linux/random.h>
#include <linux/slab.h>
#include <linux/string.h>

#include <linux/cryptowrap.h>

#include "mpi.h"
#include "rsa.h"

#define mpi_get_nbytes(a)	((mpi_get_nbits((a)) + 7) / 8)

typedef struct {
	MPI n;
	MPI e;
} RSA_public_key;

typedef struct {
	MPI n;
	MPI e;
	MPI d;
	MPI p;
	MPI q;
	MPI u;
} RSA_secret_key;

typedef struct CryptoWrap_gnupg_rsa {
	RSA_public_key pubkey;
	RSA_secret_key privkey;
} gnupg_rsa_object;

/* for test */
#include "cryptowrap-rsa-test.c"

int
CryptoWrap_RSA_new(CryptoWrap_RSA_object *objp)
{
	gnupg_rsa_object *rsa_obj;

	rsa_obj = kmalloc(sizeof(gnupg_rsa_object), GFP_KERNEL);
	if (rsa_obj == NULL)
                return CryptoWrap_E_NOMEM;

	memset(rsa_obj, 0, sizeof(gnupg_rsa_object));
        *objp = (CryptoWrap_RSA_object)rsa_obj;
        MOD_INC_USE_COUNT;
	return CryptoWrap_SUCCESS;
}

int
CryptoWrap_RSA_destroy(CryptoWrap_RSA_object obj)
{
	gnupg_rsa_object *rsa_obj = (gnupg_rsa_object *)obj;


	if (rsa_obj->pubkey.e)
		mpi_free(rsa_obj->pubkey.e);
	if (rsa_obj->pubkey.n)
		mpi_free(rsa_obj->pubkey.n);
	if (rsa_obj->privkey.d)
		mpi_free(rsa_obj->privkey.d);
	if (rsa_obj->privkey.n)
		mpi_free(rsa_obj->privkey.n);
	kfree(obj);
	MOD_DEC_USE_COUNT;
	return CryptoWrap_SUCCESS;
}

int
CryptoWrap_RSA_set_public_key(CryptoWrap_RSA_object obj, unsigned char *e, int elen, unsigned char *n, int nlen)
{
	gnupg_rsa_object *rsa_obj = (gnupg_rsa_object *)obj;

	if (rsa_obj->pubkey.e)
		mpi_free(rsa_obj->pubkey.e);
	if (rsa_obj->pubkey.n)
		mpi_free(rsa_obj->pubkey.n);

	if ((rsa_obj->pubkey.e = mpi_bin2mpi(e, elen, 1)) == NULL)
		return CryptoWrap_E_NOMEM;
	if ((rsa_obj->pubkey.n = mpi_bin2mpi(n, nlen, 1)) == NULL) {
		mpi_free(rsa_obj->pubkey.e);
		return CryptoWrap_E_NOMEM;
	}

	return CryptoWrap_SUCCESS;
}




int
CryptoWrap_RSA_get_public_key(CryptoWrap_RSA_object obj, unsigned char *e, int *elen, unsigned char *n, int *nlen)
{
	gnupg_rsa_object *rsa_obj = (gnupg_rsa_object *)obj;
 	int e1, n1;	

	if (rsa_obj->pubkey.e) {
	 	e1 =mpi_get_nbytes(rsa_obj->pubkey.e); 
		mpi_mpi2bin(rsa_obj->pubkey.e, e);
		(*elen) = e1;
	}

	if (rsa_obj->pubkey.n){
		n1 = mpi_get_nbytes(rsa_obj->pubkey.n);
		mpi_mpi2bin(rsa_obj->pubkey.n, n);
		(*nlen) = n1;
	}
	return CryptoWrap_SUCCESS;
}


int
CryptoWrap_RSA_set_private_key(CryptoWrap_RSA_object obj, unsigned char *d, int dlen, unsigned char *n, int nlen)
{
	gnupg_rsa_object *rsa_obj = (gnupg_rsa_object *)obj;

	if (rsa_obj->privkey.d)
		mpi_free(rsa_obj->privkey.d);
	if (rsa_obj->privkey.n)
		mpi_free(rsa_obj->privkey.n);

	if ((rsa_obj->privkey.d = mpi_bin2mpi(d, dlen, 1)) == NULL)
		return CryptoWrap_E_NOMEM;
	if ((rsa_obj->privkey.n = mpi_bin2mpi(n, nlen, 1)) == NULL) {
		mpi_free(rsa_obj->pubkey.e);
		return CryptoWrap_E_NOMEM;
	}

	return CryptoWrap_SUCCESS;
}

int
CryptoWrap_RSA_get_private_key(CryptoWrap_RSA_object obj, unsigned char *d, int *dlen, unsigned char *n, int *nlen)
{
	gnupg_rsa_object *rsa_obj = (gnupg_rsa_object *)obj;
	int d1, n1;

	if (rsa_obj->privkey.d) {
	    d1 =mpi_get_nbytes(rsa_obj->privkey.d);

	    mpi_mpi2bin(rsa_obj->privkey.d, d);
	    (*dlen) = d1;
	}
	if (rsa_obj->privkey.n){
	    n1 = mpi_get_nbytes(rsa_obj->privkey.n);
	    mpi_mpi2bin(rsa_obj->privkey.n, n);
	    (*nlen) = n1;
	}
	return CryptoWrap_SUCCESS;
}


static int
MGF(unsigned char *mgfSeed, int mgfSeedLen, unsigned char *mask, int maskLen)
{
	int ret;
	unsigned int i;
	unsigned char counter[4], md[CryptoWrap_SHA1_DIGEST_LEN];
	CryptoWrap_SHA1_object sha1_obj;

	ret = CryptoWrap_SHA1_new(&sha1_obj);
	if (ret != CryptoWrap_SUCCESS)
		return ret;

	for (i = 0;
		maskLen > 0;
		i++,
		maskLen -= CryptoWrap_SHA1_DIGEST_LEN,
		mask += CryptoWrap_SHA1_DIGEST_LEN) {
		counter[0] = (unsigned char)((i >> 24) & 0xff);
		counter[1] = (unsigned char)((i >> 16) & 0xff);
		counter[2] = (unsigned char)((i >> 8) & 0xff);
		counter[3] = (unsigned char)(i & 0xff);
		CryptoWrap_SHA1_Init(sha1_obj);
		CryptoWrap_SHA1_Update(sha1_obj, mgfSeed, mgfSeedLen);
		CryptoWrap_SHA1_Update(sha1_obj, counter, 4);
		CryptoWrap_SHA1_Final(sha1_obj, md);
		if (maskLen >= CryptoWrap_SHA1_DIGEST_LEN)
			memcpy(mask, md, CryptoWrap_SHA1_DIGEST_LEN);
		else
			memcpy(mask, md, maskLen);
	}
	CryptoWrap_SHA1_destroy(sha1_obj);

	return CryptoWrap_SUCCESS;
}

/*
 * PKCS #1 v2.1 - 7.2.1
 *     RSAES-PKCS1-v1_5-ENCRYPT
 *     EM = 0x00||0x02||PS||0x00||M
 */ 
static int
RSA_padding_pkcs1(unsigned char *msg, int mlen, unsigned char *em, int emlen)
{
	int i;

	if (mlen > emlen - 11)
		return CryptoWrap_E_MSG_TOO_LONG;

	*em = 0x00;
	*(em + 1) = 0x02;

	/*
	 * PS - consisting of pseudo randomly geenrated nonzero octets
	 *      The length will be at least eight octets.
	 */
	for (i = 2; i < emlen - mlen - 1; i++) {
again:
		get_random_bytes(em + i, 1);
		if (*(em + i) == 0x00)
			goto again;
	}
	*(em + i) = 0x00;

	/* M */
	memcpy(em + (emlen - mlen), msg, mlen);

	return CryptoWrap_SUCCESS;
}

/*
 * PKCS #1 v2.1 - 7.2.2
 *     RSAES-PKCS1-v1_5-DECRYPT
 *     EM = 0x00||0x02||PS||0x00||M
 */ 
static int
RSA_ipadding_pkcs1(unsigned char *em, int emlen, unsigned char *msg, int *mlen)
{
	int i;

	if (*em != 0x00 || *(em + 1) != 0x02)
		return CryptoWrap_E_MSG_INVALID;

	/* search 0x00 - boundary of PS and M */
	for (i = 2; i < emlen; i++) {
		if (*(em + i) == 0x00) {
			i++;
			break;
		}
	}
	if (i == emlen)
		return CryptoWrap_E_MSG_INVALID;

	*mlen = emlen - i;
	memcpy(msg, em + i, emlen - i);
	return CryptoWrap_SUCCESS;
}

/*
 * PKCS #1 v2.1 - 7.1.1
 *     RSAES-OAEP-ENCRYPT
 *     EM = 0x00||maskedSeed||maskedDB
 */
static const unsigned char lHash[] = { 0xda, 0x39, 0xa3, 0xee, 0x5e, 0x6b, 0x4b, 0x0d, 0x32, 0x55, 0xbf, 0xef, 0x95, 0x60, 0x18, 0x90, 0xaf, 0xd8, 0x07, 0x09 };

static int
RSA_padding_oaep(unsigned char *msg, int mlen, unsigned char *em, int emlen)
{
	int i, ret, DBlen, PSlen;
	unsigned char *DB, *dbMask;
	unsigned char *seed, seedMask[CryptoWrap_SHA1_DIGEST_LEN];

	if (mlen > emlen - 2 * CryptoWrap_SHA1_DIGEST_LEN - 2)
		return CryptoWrap_E_MSG_TOO_LONG;

	DB = em + 1 + CryptoWrap_SHA1_DIGEST_LEN;
	DBlen = emlen - 1 - CryptoWrap_SHA1_DIGEST_LEN;
	dbMask = kmalloc(DBlen, GFP_KERNEL);
	if (dbMask == NULL)
		return CryptoWrap_E_NOMEM;

	/* DB = lHash||PS||0x01||M */
	memcpy(DB, lHash, CryptoWrap_SHA1_DIGEST_LEN);
	/* PS - consisting of zero octets */
	if ((PSlen = emlen - mlen - 2 * CryptoWrap_SHA1_DIGEST_LEN - 2) > 0)
		memset(DB + CryptoWrap_SHA1_DIGEST_LEN, 0, PSlen);
	else
		PSlen = 0;
	*(DB + CryptoWrap_SHA1_DIGEST_LEN + PSlen) = 0x01;
	memcpy(DB + CryptoWrap_SHA1_DIGEST_LEN + PSlen + 1, msg, mlen);

	/* seed - randomly octets */
	seed = em + 1;
	get_random_bytes(seed, CryptoWrap_SHA1_DIGEST_LEN);

	/* dbMask = MGF(seed, k - hLen - 1) */
	ret = MGF(seed, CryptoWrap_SHA1_DIGEST_LEN, dbMask, DBlen);
	if (ret != CryptoWrap_SUCCESS)
		goto bad;
	/* maskedDB = DB (+) dbMask */
	for (i = 0; i < DBlen; i++)
		*(DB+i) ^=  *(dbMask+i);

	/* seedMask = MGF(maskedDB, hLen) */
	ret = MGF(DB, DBlen, seedMask, CryptoWrap_SHA1_DIGEST_LEN);
	if (ret != CryptoWrap_SUCCESS)
		goto bad;
	/* maskedSeed = seed (+) seedMask */
	for (i = 0; i < CryptoWrap_SHA1_DIGEST_LEN; i++)
		*(seed+i) ^= seedMask[i];

	*em = 0x00;
	ret = CryptoWrap_SUCCESS;
bad:
	kfree(dbMask);
	return ret;
}

/*
 * PKCS #1 v2.1 - 7.1.2
 *     RSAES-OAEP-DECRYPT
 *     EM = 0x00||maskedSeed||maskedDB
 */
static int
RSA_ipadding_oaep(unsigned char *em, int emlen, unsigned char *msg, int *mlen)
{
	int i, ret, DBlen;
	unsigned char *DB, *dbMask;
	unsigned char *seed, seedMask[CryptoWrap_SHA1_DIGEST_LEN];

	if (*em != 0x00)
		return CryptoWrap_E_MSG_INVALID;

	DB = em + CryptoWrap_SHA1_DIGEST_LEN + 1;
	DBlen = emlen - CryptoWrap_SHA1_DIGEST_LEN - 1;
	dbMask = kmalloc(DBlen, GFP_KERNEL);
	if (dbMask == NULL)
		return CryptoWrap_E_NOMEM;

	/* seedMask = MGF(maskedDB, hLen) */
	ret = MGF(DB, DBlen, seedMask, CryptoWrap_SHA1_DIGEST_LEN);
	if (ret != CryptoWrap_SUCCESS)
		goto bad;
	/* seed = maskedSeed (+) seedMask */
	seed = em + 1;
	for (i = 0; i < CryptoWrap_SHA1_DIGEST_LEN; i++)
		*(seed+i) ^= seedMask[i];

	/* dbMask = MGF(seed, k - hLen - 1) */
	ret = MGF(seed, CryptoWrap_SHA1_DIGEST_LEN, dbMask, DBlen);
	if (ret != CryptoWrap_SUCCESS)
		goto bad;
	/* DB = maskedDB (+) dbMask */
	for (i = 0; i < DBlen; i++)
		*(DB+i) ^= *(dbMask+i);

	/* DB = lHash'||PS||0x01||M */
	if (memcmp(lHash, DB, CryptoWrap_SHA1_DIGEST_LEN)) {
		ret = CryptoWrap_E_MSG_INVALID;
		goto bad;
	}
	/* search 0x01 - boundary of PS and M */
	for (i =  CryptoWrap_SHA1_DIGEST_LEN; i < DBlen; i++) {
		if (*(DB + i) == 0x01) {
			i++;
			break;
		}
	}
	if (i == DBlen) {
		ret = CryptoWrap_E_MSG_INVALID;
		goto bad;
	}

	*mlen = DBlen - i;
	memcpy(msg, DB + i, DBlen - i);
	ret = CryptoWrap_SUCCESS;
bad:
	kfree(dbMask);
	return ret;
}

static inline int
RSA_padding_no(unsigned char *msg, int mlen, unsigned char *em, int emlen)
{
	if (mlen > emlen)
		return CryptoWrap_E_MSG_TOO_LONG;
	if (mlen < emlen)
		return CryptoWrap_E_MSG_TOO_SHORT;
	memcpy(em, msg, emlen);
	return CryptoWrap_SUCCESS;
}

static inline int
RSA_ipadding_no(unsigned char *em, int emlen, unsigned char *msg, int *mlen)
{
	*mlen = emlen;
	memcpy(msg, em, emlen);
	return CryptoWrap_SUCCESS;
}

int
CryptoWrap_RSA_public_encrypt(CryptoWrap_RSA_object obj, unsigned char *from, int flen, unsigned char *to, int *tlen, int padding)
{
	gnupg_rsa_object *rsa_obj = (gnupg_rsa_object *)obj;
	unsigned char *em;
	int ret, k, emlen;
	MPI result, data;

	k = mpi_get_nbytes(rsa_obj->pubkey.n);
	emlen = k;
	em = kmalloc(emlen, GFP_KERNEL);
	if (em == NULL)
		return CryptoWrap_E_NOMEM;

	switch (padding) {
	case CryptoWrap_RSA_PADDING_PKCS1:
		ret = RSA_padding_pkcs1(from, flen, em, emlen);
		break;
	case CryptoWrap_RSA_PADDING_OAEP:
		ret = RSA_padding_oaep(from, flen, em, emlen);
		break;
	case CryptoWrap_RSA_PADDING_NO:
		ret = RSA_padding_no(from, flen, em, emlen);
		break;
	default:
		ret = CryptoWrap_E_NOTSUPPORTED;
		break;
	}

	if (ret != CryptoWrap_SUCCESS)
		goto bad0;

	data = mpi_bin2mpi(em, emlen, 1);
	if (data == NULL) {
		ret = CryptoWrap_E_NOMEM;
		goto bad0;
	}

	if (padding == CryptoWrap_RSA_PADDING_NO
		&& mpi_cmp(data, rsa_obj->pubkey.n) > 0) {
		ret = CryptoWrap_E_MSG_TOO_LONG;
		goto bad0;
	}

	if (rsa_encrypt(&result, data, &rsa_obj->pubkey.n) != 0) {
		ret = CryptoWrap_E_INTERNAL;
		goto bad1;
	}

	*tlen = k;
	mpi_mpi2buf(result, to, k);
	ret = CryptoWrap_SUCCESS;

	mpi_free(result);
bad1:
	mpi_free(data);
bad0:
	kfree(em);
	return ret;
}

int
CryptoWrap_RSA_private_decrypt(CryptoWrap_RSA_object obj, unsigned char *from, int flen, unsigned char *to, int *tlen, int padding)
{
	gnupg_rsa_object *rsa_obj = (gnupg_rsa_object *)obj;
	unsigned char *em;
	int ret, k, emlen;
	MPI data, result;

	k = mpi_get_nbytes(rsa_obj->privkey.n);
	if (flen != k)
		return CryptoWrap_E_MSG_INVALID;

	switch (padding) {
	case CryptoWrap_RSA_PADDING_PKCS1:
		if (k < 11)
			return CryptoWrap_E_KEY_TOO_SHORT;
		break;
	case CryptoWrap_RSA_PADDING_OAEP:
		if (k < 2 * CryptoWrap_SHA1_DIGEST_LEN + 2)
			return CryptoWrap_E_KEY_TOO_SHORT;
		break;
	case CryptoWrap_RSA_PADDING_NO:
		break;
	}

	data = mpi_bin2mpi(from, flen, 1);
	if (data == NULL)
		return CryptoWrap_E_NOMEM;

	if (rsa_decrypt(&result, data, &rsa_obj->privkey.n) != 0) {
		ret = CryptoWrap_E_INTERNAL;
		goto bad0;
	}

	emlen = k;
	em = kmalloc(emlen, GFP_KERNEL);
	if (em == NULL) {
		ret = CryptoWrap_E_NOMEM;
		goto bad1;
	}
	mpi_mpi2buf(result, em, emlen);

	switch (padding) {
	case CryptoWrap_RSA_PADDING_PKCS1:
		ret = RSA_ipadding_pkcs1(em, emlen, to, tlen);
		break;
	case CryptoWrap_RSA_PADDING_OAEP:
		ret = RSA_ipadding_oaep(em, emlen, to, tlen);
		break;
	case CryptoWrap_RSA_PADDING_NO:
		ret = RSA_ipadding_no(em, emlen, to, tlen);
		break;
	default:
		ret = CryptoWrap_E_NOTSUPPORTED;
		break;
	}

	kfree(em);
bad1:
	mpi_free(result);
bad0:
	mpi_free(data);
	return ret;
}

/*
 * PKCS #1 v2.1 - 9.1.1
 *     EMSA-PSS-ENCODE
 *     EM = maskedDB||H||0xbc
 */
static const unsigned char zeroMask[] = { 0xff, 0x7f, 0x3f, 0x1f, 0x0f, 0x07, 0x03, 0x01 };

static int
RSA_pss_get_hash(unsigned char *msg, int mlen, unsigned char *H)
{
	int ret;
	CryptoWrap_SHA1_object sha1_obj;
	unsigned char mHash[CryptoWrap_SHA1_DIGEST_LEN], zero[8];

	ret = CryptoWrap_SHA1_new(&sha1_obj);
	if (ret != CryptoWrap_SUCCESS)
		return ret;

	/* mHash = Hash(M) */
	CryptoWrap_SHA1_Init(sha1_obj);
	CryptoWrap_SHA1_Update(sha1_obj, msg, mlen);
	CryptoWrap_SHA1_Final(sha1_obj, mHash);

	/* M' = (0x)00 00 00 00 00 00 00 00||mHash||salt */
	memset(zero, 0, 8);
	CryptoWrap_SHA1_Init(sha1_obj);
	CryptoWrap_SHA1_Update(sha1_obj, zero, 8);
	CryptoWrap_SHA1_Update(sha1_obj, mHash, CryptoWrap_SHA1_DIGEST_LEN);
	CryptoWrap_SHA1_Final(sha1_obj, H);

	CryptoWrap_SHA1_destroy(sha1_obj);

	return CryptoWrap_SUCCESS;
}

static int
RSA_encoding_pss(unsigned char *msg, int mlen, unsigned char *em, int emlen, int embits)
{
	unsigned char *DB, *H, *dbMask;
	int i, ret, DBlen;

	if (emlen < CryptoWrap_SHA1_DIGEST_LEN + 2)
		return CryptoWrap_E_KEY_TOO_SHORT;

	DB = em;
	DBlen = emlen - CryptoWrap_SHA1_DIGEST_LEN - 1;
	dbMask = kmalloc(DBlen, GFP_KERNEL);
	if (dbMask == NULL)
		return CryptoWrap_E_NOMEM;

	H = em + DBlen;
	ret = RSA_pss_get_hash(msg, mlen, H);
	if (ret != CryptoWrap_SUCCESS)
		goto bad;

	/* DB = PS||0x01||salt */

	/* PS - cosisting of zero octets */
	memset(DB, 0, DBlen - 1);
	*(DB + DBlen - 1) = 0x01;

	/* dbMask = MGF(H, emLen - hLen - 1); */
	ret = MGF(H, CryptoWrap_SHA1_DIGEST_LEN, dbMask, DBlen);
	if (ret != CryptoWrap_SUCCESS)
		goto bad;
	/* maskedDB = DB (+) dbMask */
	for (i = 0; i < DBlen; i++)
		*(DB + i) ^= *(dbMask + i);

	*(em) &= zeroMask[emlen * 8 - embits];
	*(em + emlen - 1) = 0xbc;
	ret = CryptoWrap_SUCCESS;
bad:
	kfree(dbMask);
	return ret;
}
/*
 * PKCS #1 v2.1 - 9.1.2
 *     EMSA-PSS-VERIFY
 *     EM = maskedDB||H||0xbc
 */
static int
RSA_verify_pss(unsigned char *msg, int mlen, unsigned char *em, int emlen, int embits)
{
	unsigned char hDash[CryptoWrap_SHA1_DIGEST_LEN], *DB, *H, *dbMask;
	int i, ret, DBlen;

	if (emlen < CryptoWrap_SHA1_DIGEST_LEN +2)
		return CryptoWrap_E_VERIFY;

	if (*(em + emlen - 1) != 0xbc)
		return CryptoWrap_E_VERIFY;

	if (*(em) & ~zeroMask[emlen * 8 - embits])
		return CryptoWrap_E_VERIFY;

	DB = em;
	DBlen = emlen - CryptoWrap_SHA1_DIGEST_LEN - 1;
	dbMask = kmalloc(DBlen, GFP_KERNEL);
	if (dbMask == NULL)
		return CryptoWrap_E_NOMEM;
	H = em + DBlen;

	ret = MGF(H, CryptoWrap_SHA1_DIGEST_LEN, dbMask, DBlen);
	for (i = 0; i < DBlen; i++)
		*(DB + i) ^= *(dbMask + i);

	*(DB) &= zeroMask[emlen * 8 - embits];
	for (i = 0; i < DBlen - 1; i++) {
		if (*(DB + i) != 0x00) {
			ret = CryptoWrap_E_VERIFY;
			goto bad;
		}
	}
	if (*(DB + i) != 0x01) {
		ret = CryptoWrap_E_VERIFY;
		goto bad;
	}

	ret = RSA_pss_get_hash(msg, mlen, hDash);
	if (ret != CryptoWrap_SUCCESS)
		goto bad;

	if (memcmp(H, hDash, CryptoWrap_SHA1_DIGEST_LEN)) {
		ret = CryptoWrap_E_VERIFY;
		goto bad;
	}

	ret = CryptoWrap_SUCCESS;
bad:
	kfree(dbMask);
	return ret;
}

/*
 * PKCS #1 v2.1 - 9.2.1
 *     EMSA-PKCS1-v1_5-ENCODE
 *     EM = 0x00||0x01||PS||0x00||T
 */
static const unsigned char SHA1DigestInfo[] = { 0x30, 0x21, 0x30, 0x09, 0x06, 0x05, 0x2b, 0x0e, 0x03, 0x02, 0x1a, 0x05, 0x00, 0x04, 0x14 };

static int
RSA_encoding_pkcs1(unsigned char *msg, int mlen, unsigned char *em, int emlen)
{
	CryptoWrap_SHA1_object sha1_obj;
	int i, ret;

	if (emlen < sizeof(SHA1DigestInfo) + CryptoWrap_SHA1_DIGEST_LEN + 11)
		return CryptoWrap_E_KEY_TOO_SHORT;

	*(em) = 0x00;
	*(em + 1) = 0x01;

	/* PS - consisting of value 0xff octests */
	for (i = 2; i < emlen - 1 - sizeof(SHA1DigestInfo) - CryptoWrap_SHA1_DIGEST_LEN; i++)
		*(em + i) = 0xff;

	*(em + i) = 0x00;
	memcpy(em + emlen - sizeof(SHA1DigestInfo) - CryptoWrap_SHA1_DIGEST_LEN, SHA1DigestInfo, sizeof(SHA1DigestInfo));

	ret = CryptoWrap_SHA1_new(&sha1_obj);
	if (ret != CryptoWrap_SUCCESS)
		return ret;

	CryptoWrap_SHA1_Init(sha1_obj);
	CryptoWrap_SHA1_Update(sha1_obj, msg, mlen);
	CryptoWrap_SHA1_Final(sha1_obj, em + emlen - CryptoWrap_SHA1_DIGEST_LEN);
	CryptoWrap_SHA1_destroy(sha1_obj);

	return CryptoWrap_SUCCESS;
}

static int
RSA_verify_pkcs1(unsigned char *msg, int mlen, unsigned char *em, int emlen)
{
	int ret;
	unsigned char *emDash;

	emDash = kmalloc(emlen, GFP_KERNEL);
	ret = RSA_encoding_pkcs1(msg, mlen, emDash, emlen);
	if (ret != CryptoWrap_SUCCESS)
		goto bad;

	if (memcmp(em, emDash, emlen)) {
		ret = CryptoWrap_E_VERIFY;
		goto bad;
	}

	ret = CryptoWrap_SUCCESS;
bad:
	kfree(emDash);
	return ret;
}

static inline int
RSA_encoding_no(unsigned char *msg, int mlen, unsigned char *em, int emlen)
{
	if (mlen > emlen)
		return CryptoWrap_E_MSG_TOO_LONG;
	if (mlen < emlen)
		return CryptoWrap_E_MSG_TOO_SHORT;
	memcpy(em, msg, emlen);
	return CryptoWrap_SUCCESS;
}

int
CryptoWrap_RSA_private_encrypt(CryptoWrap_RSA_object obj, unsigned char *from, int flen, unsigned char *to, int *tlen)
{
	return CryptoWrap_RSA_sign(obj, from, flen, to, tlen, CryptoWrap_RSA_SIGN_RAW);
}

int
CryptoWrap_RSA_public_decrypt(CryptoWrap_RSA_object obj, unsigned char *from, int flen, unsigned char *to, int *tlen)
{
	int ret, k;
	gnupg_rsa_object *rsa_obj = (gnupg_rsa_object *)obj;
	MPI result, data;

	k = mpi_get_nbytes(rsa_obj->privkey.n);
	if (flen != k)
		return CryptoWrap_E_MSG_INVALID;

	data = mpi_bin2mpi(from, flen, 1);
	if (data == NULL)
		return CryptoWrap_E_NOMEM;

	if (rsa_verify(&result, data, &rsa_obj->pubkey.n) != 0) {
		ret = CryptoWrap_E_INTERNAL;
		goto bad0;
	}

	*tlen = k;
	mpi_mpi2buf(result, to, *tlen);
	ret = CryptoWrap_SUCCESS;

	mpi_free(result);
bad0:
	mpi_free(data);
	return ret;
}

int
CryptoWrap_RSA_sign(CryptoWrap_RSA_object obj, unsigned char *msg, int mlen, unsigned char *sig, int *slen, int type)
{
	gnupg_rsa_object *rsa_obj = (gnupg_rsa_object *)obj;
	MPI result, data;
	unsigned char *em;
	int ret, k, emlen, embits;

	k = mpi_get_nbytes(rsa_obj->privkey.n);
	switch (type) {
	case CryptoWrap_RSA_SIGN_SHA1_PSS:
		embits = mpi_get_nbits(rsa_obj->privkey.n) - 1;
		emlen = (embits + 7) / 8;
		em = kmalloc(emlen, GFP_KERNEL);
		if (em == NULL)
			return CryptoWrap_E_NOMEM;
		ret = RSA_encoding_pss(msg, mlen, em, emlen, embits);
		break;
	case CryptoWrap_RSA_SIGN_SHA1_PKCS1:
		emlen = k;
		em = kmalloc(emlen, GFP_KERNEL);
		if (em == NULL)
			return CryptoWrap_E_NOMEM;
		ret = RSA_encoding_pkcs1(msg, mlen, em, emlen);
		break;
	case CryptoWrap_RSA_SIGN_RAW:
		emlen = k;
		em = kmalloc(emlen, GFP_KERNEL);
		if (em  == NULL)
			return CryptoWrap_E_NOMEM;
		ret = RSA_encoding_no(msg, mlen, em, emlen);
		break;
	default:
		return CryptoWrap_E_NOTSUPPORTED;
	}

	if (ret != CryptoWrap_SUCCESS)
		goto bad0;

	data = mpi_bin2mpi(em, emlen, 1);
	if (data == NULL)
		return CryptoWrap_E_NOMEM;

	if (type == CryptoWrap_RSA_SIGN_RAW
		&& mpi_cmp(data, rsa_obj->pubkey.n) > 0) {
		ret = CryptoWrap_E_MSG_TOO_LONG;
		goto bad0;
	}

	if (rsa_sign(&result, data, &rsa_obj->privkey.n) != 0) {
		ret = CryptoWrap_E_INTERNAL;
		goto bad1;
	}

	*slen = k;
	mpi_mpi2buf(result, sig, *slen);
	ret = CryptoWrap_SUCCESS;

	mpi_free(result);
bad1:
	mpi_free(data);
bad0:
	kfree(em);
	return ret;
}

int
CryptoWrap_RSA_verify(CryptoWrap_RSA_object obj, unsigned char *msg, int mlen, unsigned char *sig, int slen, int type)
{
	gnupg_rsa_object *rsa_obj = (gnupg_rsa_object *)obj;
	unsigned char *em;
	int ret, k, emlen, embits;
	MPI result, data;

	k = mpi_get_nbytes(rsa_obj->pubkey.n);
	if (slen != k)
		return CryptoWrap_E_VERIFY;

	data = mpi_bin2mpi(sig, slen, 1);
	if (data == NULL)
		return CryptoWrap_E_NOMEM;

	if (rsa_verify(&result, data, &rsa_obj->pubkey.n) != 0) {
		ret = CryptoWrap_E_INTERNAL;
		goto bad0;
	}

	emlen = k;
	em = kmalloc(emlen, GFP_KERNEL);
	if (em == NULL) {
		ret = CryptoWrap_E_NOMEM;
		goto bad1;
	}
	mpi_mpi2buf(result, em, emlen);

	switch (type) {
	case CryptoWrap_RSA_SIGN_SHA1_PSS:
		embits = mpi_get_nbits(result);
		ret = RSA_verify_pss(msg, mlen, em, emlen, embits);
		break;
	case CryptoWrap_RSA_SIGN_SHA1_PKCS1:
		ret = RSA_verify_pkcs1(msg, mlen, em, emlen);
		break;
	default:
		ret = CryptoWrap_E_NOTSUPPORTED;
		break;
	}

	kfree(em);
bad1:
	mpi_free(result);
bad0:
	mpi_free(data);
	return ret;
}

// Need to optimize
int
CryptoWrap_RSA_init(CryptoWrap_RSA_object obj)
{
	int nbits = 1024;
	RSA_secret_key skey;
	MPI *factors;
	unsigned char *e, *d, *n;
	int elen, dlen, nlen;
	int i, ret = 0;


	printk(KERN_INFO "Begin of CryptoWrap_RSA_init()\n");
	rsa_generate(nbits, (MPI *)&skey, &factors);
	elen = mpi_get_nbytes(skey.e);
	dlen = mpi_get_nbytes(skey.d);
	nlen = mpi_get_nbytes(skey.n);
	e = kmalloc(elen, GFP_KERNEL);
	d = kmalloc(dlen, GFP_KERNEL);
	n = kmalloc(nlen, GFP_KERNEL);

	mpi_mpi2bin(skey.e, e);
	mpi_mpi2bin(skey.d, d);
	mpi_mpi2bin(skey.n, n);
	ret = CryptoWrap_RSA_set_public_key(obj, e, elen, n, nlen);
	if (ret != CryptoWrap_SUCCESS) {
		printk(KERN_INFO "CryptoWrap_RSA_set_public_key() failed.(%d)\n", ret);
		ret = -1;
	goto bad;
	}
	ret = CryptoWrap_RSA_set_private_key(obj, d, dlen, n, nlen);
	if (ret != CryptoWrap_SUCCESS) {
		printk(KERN_INFO "CryptoWrap_RSA_set_private_key() failed.(%d)\n", ret);
		ret = -1;
		goto bad;
	}
	printk(KERN_INFO"\nCreated: key e[%d]=\n", elen);
	for (i = 0; i<elen; i++) printk(KERN_INFO"%u", e[i]);
	printk(KERN_INFO"\nCreated: public key n[%d]=\n", nlen);
	for (i = 0; i<nlen; i++) printk(KERN_INFO"%u", n[i]);
	printk(KERN_INFO"\nCreated: private key d[%d]=\n", dlen);
	for (i = 0; i<dlen; i++) printk(KERN_INFO"%u", d[i]);

bad:
	/*
	kfree(e);
	kfree(d);
	kfree(n);
	*/
	mpi_free(skey.n);
	mpi_free(skey.e);
	mpi_free(skey.d);
	mpi_free(skey.p);
	mpi_free(skey.q);
	mpi_free(skey.u);

	printk(KERN_INFO "End of CryptoWrap_RSA_init()\n");
	return ret;
}


static int CryptoWrap_RSA_module_init(void);
static void CryptoWrap_RSA_module_exit(void);

static int
CryptoWrap_RSA_module_init()
{
        printk(KERN_INFO "CryptoWrap_RSA: loaded\n");
        return 0;
}

static void
CryptoWrap_RSA_module_exit()
{
        printk(KERN_INFO "CryptoWrap_RSA: unloaded\n");
}

module_init(CryptoWrap_RSA_module_init);
module_exit(CryptoWrap_RSA_module_exit);

EXPORT_SYMBOL(CryptoWrap_RSA_new);
EXPORT_SYMBOL(CryptoWrap_RSA_init);
EXPORT_SYMBOL(CryptoWrap_RSA_destroy);
EXPORT_SYMBOL(CryptoWrap_RSA_set_public_key);
EXPORT_SYMBOL(CryptoWrap_RSA_get_public_key);
EXPORT_SYMBOL(CryptoWrap_RSA_set_private_key);
EXPORT_SYMBOL(CryptoWrap_RSA_get_private_key);
EXPORT_SYMBOL(CryptoWrap_RSA_public_encrypt);
EXPORT_SYMBOL(CryptoWrap_RSA_private_decrypt);
EXPORT_SYMBOL(CryptoWrap_RSA_private_encrypt);
EXPORT_SYMBOL(CryptoWrap_RSA_public_decrypt);
EXPORT_SYMBOL(CryptoWrap_RSA_sign);
EXPORT_SYMBOL(CryptoWrap_RSA_verify);
EXPORT_SYMBOL(CryptoWrap_RSA_test);
