/*
 *  cryptowrap-rsa-test.c - test for RSA function of GnuPG
 *
 *  Copyright 2003 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* included from cryptowrap-rsa-glue.c */

static void dump_bytes(unsigned char *label, unsigned char *p, int len);
static void dump_msg(unsigned char *label, unsigned char *msg);

static void
dump_bytes(unsigned char *label, unsigned char *p, int len)
{
#if 0
	int i;

	printk("%s : ", label);
	for (i = 0; i < len; i++)
		printk("%02x", *(p + i));
	printk("\n");
#endif
}

static void
dump_msg(unsigned char *label, unsigned char *msg)
{
#if 0
	printk("%s: %s\n", label, msg);
#endif
}

int
CryptoWrap_RSA_test(CryptoWrap_RSA_object obj)
{
	int nbits = 1024;
	RSA_secret_key skey;
	MPI *factors;
	int ret, count = 0;
	unsigned char *e, *d, *n, *msg, *blob, *dmsg;
	unsigned char text[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
	int elen, dlen, nlen, msglen, bloblen, dmsglen;

	printk(KERN_INFO "*** CryptoWrap_RSA_test() start ***\n");

	printk(KERN_INFO "RSA key generating...\n");
	rsa_generate(nbits, (MPI *)&skey, &factors);
#if 0
	g10_log_mpidump("[skey.n]", skey.n);
	g10_log_mpidump("[skey.e]", skey.e);
	g10_log_mpidump("[skey.d]", skey.d);
	g10_log_mpidump("[skey.p]", skey.p);
	g10_log_mpidump("[skey.q]", skey.q);
	g10_log_mpidump("[skey.u]", skey.u);
#endif

	elen = mpi_get_nbytes(skey.e);
	dlen = mpi_get_nbytes(skey.d);
	nlen = mpi_get_nbytes(skey.n);
	e = kmalloc(elen, GFP_KERNEL);
	d = kmalloc(dlen, GFP_KERNEL);
	n = kmalloc(nlen, GFP_KERNEL);

	mpi_mpi2bin(skey.e, e);
	mpi_mpi2bin(skey.d, d);
	mpi_mpi2bin(skey.n, n);

	msglen = bloblen = dmsglen = nlen;
	msg = kmalloc(msglen, GFP_KERNEL);
	blob = kmalloc(bloblen, GFP_KERNEL);
	dmsg = kmalloc(dmsglen, GFP_KERNEL);

	printk(KERN_INFO "RSA set key pair...\n");
	ret = CryptoWrap_RSA_set_public_key(obj, e, elen, n, nlen);
	if (ret != CryptoWrap_SUCCESS) {
		printk(KERN_INFO "CryptoWrap_RSA_set_public_key() failed.(%d)\n", ret);
		count = ret;
		goto bad;
	}
	ret = CryptoWrap_RSA_set_private_key(obj, d, dlen, n, nlen);
	if (ret != CryptoWrap_SUCCESS) {
		printk(KERN_INFO "CryptoWrap_RSA_set_private_key() failed.(%d)\n", ret);
		count = ret;
		goto bad;
	}

	printk(KERN_INFO "*** public encrypt -> private decrypt with no padding ***\n");

	msglen = bloblen = dmsglen = nlen;
	get_random_bytes(msg, msglen);
	*msg &= 0x01;
	memset(blob, 0, bloblen);
	memset(dmsg, 0, dmsglen);

	dump_bytes("msg", msg, msglen);

	ret = CryptoWrap_RSA_public_encrypt(obj, msg, msglen, blob, &bloblen, CryptoWrap_RSA_PADDING_NO);
	if (ret != CryptoWrap_SUCCESS) {
		printk(KERN_INFO "CryptoWrap_RSA_public_encrypt() failed.(%d)\n", ret);
		count = ret;
		goto bad;
	}

	printk(KERN_INFO "CryptoWrap_RSA_public_encrypt() failed.(%d)\n", ret);
	dump_bytes("blob", blob, bloblen);

	ret = CryptoWrap_RSA_private_decrypt(obj, blob, bloblen, dmsg, &dmsglen, CryptoWrap_RSA_PADDING_NO);
	if (ret != CryptoWrap_SUCCESS) {
		printk(KERN_INFO "CryptoWrap_RSA_private_decrypt() failed.(%d)\n", ret);
		count = ret;
		goto bad;
	}

	dump_bytes("dmsg", dmsg, dmsglen);

	if (msglen == dmsglen
		&& (ret = memcmp(msg, dmsg, msglen)) == 0)
		printk(KERN_INFO "Succeeded!\n");
	else {
		printk(KERN_INFO "Failed.(%d)\n", ret);
		count++;
	}

	printk(KERN_INFO "*** public encrypt -> private decrypt with PKCS1 padding ***\n");
	memset(msg, 0, nlen);
	memset(blob, 0, nlen);
	memset(dmsg, 0, nlen);
	msglen = strlen(text);
	bloblen = dmsglen = nlen;
	memcpy(msg, text, msglen);

	dump_msg("msg", msg);

	ret = CryptoWrap_RSA_public_encrypt(obj, msg, msglen, blob, &bloblen, CryptoWrap_RSA_PADDING_PKCS1);
	if (ret != CryptoWrap_SUCCESS) {
		printk(KERN_INFO "CryptoWrap_RSA_public_encrypt() failed.(%d)\n", ret);
		count = ret;
		goto bad;
	}

	dump_bytes("blob", blob, bloblen);

	ret = CryptoWrap_RSA_private_decrypt(obj, blob, bloblen, dmsg, &dmsglen, CryptoWrap_RSA_PADDING_PKCS1);
	if (ret != CryptoWrap_SUCCESS) {
		printk(KERN_INFO "CryptoWrap_RSA_private_decrypt() failed.(%d)\n", ret);
		count = ret;
		goto bad;
	}

	dump_msg("dmsg", dmsg);

	if (msglen == dmsglen
		&& (ret = memcmp(msg, dmsg, msglen)) == 0)
		printk(KERN_INFO "Succeeded!\n");
	else {
		printk(KERN_INFO "Failed.(%d)\n", ret);
		count++;
	}

	printk(KERN_INFO "*** public encrypt -> private decrypt with OAEP padding ***\n");
	memset(msg, 0, nlen);
	memset(blob, 0, nlen);
	memset(dmsg, 0, nlen);
	msglen = strlen(text);
	bloblen = dmsglen = nlen;
	memcpy(msg, text, msglen);

	dump_msg("msg", msg);

	ret = CryptoWrap_RSA_public_encrypt(obj, msg, msglen, blob, &bloblen, CryptoWrap_RSA_PADDING_OAEP);
	if (ret != CryptoWrap_SUCCESS) {
		printk(KERN_INFO "CryptoWrap_RSA_public_encrypt() failed.(%d)\n", ret);
		count = ret;
		goto bad;
	}

	dump_bytes("blob", blob, bloblen);

	ret = CryptoWrap_RSA_private_decrypt(obj, blob, bloblen, dmsg, &dmsglen, CryptoWrap_RSA_PADDING_OAEP);
	if (ret != CryptoWrap_SUCCESS) {
		printk(KERN_INFO "CryptoWrap_RSA_private_decrypt() failed.(%d)\n", ret);
		count = ret;
		goto bad;
	}

	dump_msg("dmsg", dmsg);

	if (msglen == dmsglen
		&& (ret = memcmp(msg, dmsg, msglen)) == 0)
		printk(KERN_INFO "Succeeded!\n");
	else {
		printk(KERN_INFO "Failed.(%d)\n", ret);
		count++;
	}

	printk(KERN_INFO "*** private encrypt -> public decrypt with no padding ***\n");

	msglen = bloblen = dmsglen = nlen;
	get_random_bytes(msg, msglen);
	*msg &= 0x01;
	memset(blob, 0, bloblen);
	memset(dmsg, 0, dmsglen);

	dump_bytes("msg", msg, msglen);

	ret = CryptoWrap_RSA_private_encrypt(obj, msg, msglen, blob, &bloblen);
	if (ret != CryptoWrap_SUCCESS) {
		printk(KERN_INFO "CryptoWrap_RSA_private_encrypt() failed.(%d)\n", ret);
		count = ret;
		goto bad;
	}

	dump_bytes("blob", blob, bloblen);

	ret = CryptoWrap_RSA_public_decrypt(obj, blob, bloblen, dmsg, &dmsglen);
	if (ret != CryptoWrap_SUCCESS) {
		printk(KERN_INFO "CryptoWrap_RSA_public_decrypt() failed.(%d)\n", ret);
		count = ret;
		goto bad;
	}

	dump_bytes("dmsg", dmsg, dmsglen);

	if (msglen == dmsglen
		&& (ret = memcmp(msg, dmsg, msglen)) == 0)
		printk("Succeeded!\n");
	else {
		printk("Failed.(%d)\n", ret);
		count++;
	}

	printk(KERN_INFO "*** sign -> verify SHA1 PKCS1 ***\n");

	memset(msg, 0, nlen);
	memset(blob, 0, nlen);
	memset(dmsg, 0, nlen);
	msglen = strlen(text);
	bloblen = dmsglen = nlen;
	memcpy(msg, text, msglen);

	dump_msg("msg", msg);

	ret = CryptoWrap_RSA_sign(obj, msg, msglen, blob, &bloblen, CryptoWrap_RSA_SIGN_SHA1_PKCS1);
	if (ret != CryptoWrap_SUCCESS) {
		printk(KERN_INFO "CryptoWrap_RSA_sign() failed.(%d)\n", ret);
		count = ret;
		goto bad;
	}

	dump_bytes("blob", blob, bloblen);

	ret = CryptoWrap_RSA_verify(obj, msg, msglen, blob, bloblen, CryptoWrap_RSA_SIGN_SHA1_PKCS1);
	if (ret != CryptoWrap_SUCCESS) {
		printk(KERN_INFO "CryptoWrap_RSA_verify() failed.(%d)\n", ret);
		count = ret;
		goto bad;
	}
	else
		printk("Suceeded!\n");

	printk(KERN_INFO "*** sign -> verify SHA1 PSS ***\n");

	memset(msg, 0, nlen);
	memset(blob, 0, nlen);
	memset(dmsg, 0, nlen);
	msglen = strlen(text);
	bloblen = dmsglen = nlen;
	memcpy(msg, text, msglen);

	dump_msg("msg", msg);

	ret = CryptoWrap_RSA_sign(obj, msg, msglen, blob, &bloblen, CryptoWrap_RSA_SIGN_SHA1_PSS);
	if (ret != CryptoWrap_SUCCESS) {
		printk(KERN_INFO "CryptoWrap_RSA_sign() failed.(%d)\n", ret);
		count = ret;
		goto bad;
	}

	dump_bytes("blob", blob, bloblen);

	ret = CryptoWrap_RSA_verify(obj, msg, msglen, blob, bloblen, CryptoWrap_RSA_SIGN_SHA1_PSS);
	if (ret != CryptoWrap_SUCCESS) {
		printk(KERN_INFO "CryptoWrap_RSA_verify() failed.(%d)\n", ret);
		count = ret;
		goto bad;
	}
	else
		printk("Suceeded!\n");

bad:
	kfree(e);
	kfree(d);
	kfree(n);

	kfree(msg);
	kfree(blob);
	kfree(dmsg);

	mpi_free(skey.n);
	mpi_free(skey.e);
	mpi_free(skey.d);
	mpi_free(skey.p);
	mpi_free(skey.q);
	mpi_free(skey.u);

	printk(KERN_INFO "*** CryptoWrap_RSA_test() finished. ***\n");

	return count;
}
