/*
 *  memory.c - in-kernel memory allocation function for MPI
 *
 *  Copyright 2003 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/kernel.h>
#include <linux/slab.h>

struct m_head {
	size_t size;
	unsigned int secure;
};

int
m_size(void *p)

{
	return ((struct m_head *)(p - sizeof(struct m_head)))->size;
}

int
m_is_secure(void *p)
{
	return ((struct m_head *)(p - sizeof(struct m_head)))->secure;
}

static inline void*
m_ialloc(size_t n, int secure)
{
	void *p;
	struct m_head *mh;
	if (n == 0)
		n = 1;
	p = kmalloc(n + sizeof(struct m_head), GFP_KERNEL);
	if (p == NULL)
		panic("m_ialloc: Failed to allocate memory.\n");

	mh = (struct m_head *)p;
	mh->size = n;
	mh->secure = secure;
	return (p + sizeof(struct m_head));
}

void*
m_alloc(size_t n)
{
	return m_ialloc(n, 0);
}

void*
m_alloc_secure(size_t n)
{
	return m_ialloc(n, 1);
}

void*
m_alloc_clear(size_t n)
{
	void *p;
	p = m_alloc(n);
	memset(p, 0, n);
	return p;
}

void
m_free(void *p)
{
	kfree(p - sizeof(struct m_head));
}

void*
m_realloc(void *p, size_t n)
{
	void *q;
	int osize;
	if (p) {
		osize = m_size(p);
		if (n <= osize)
			return p;
		q = m_ialloc(n, m_is_secure(p));
		memcpy(q, p, osize);
		m_free(p);
	}
	else {
		q = m_ialloc(n, 0);
	}
	return q;
}
