/*
 *  cryptowrap-ioctl.c - ioctl interface for CryptoWrapperAPI
 *
 *  Copyright 2003 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/module.h>
#include <linux/version.h>

#include <linux/init.h>
#include <linux/config.h>
#include <linux/string.h>
#include <linux/ioctl.h>
#include <linux/slab.h>
#include <linux/fs.h>

#include <asm/uaccess.h>

#include <linux/cryptowrap.h>
#include <linux/cryptowrap_ioctl.h>

#define ALGORITHM(dev) (MINOR(dev) >> 4)
#define MODE(dev) (MINOR(dev) & 0xf)

static int cryptowrap_ioctl_open(struct inode *inode, struct file *filp);

struct file_operations cryptowrap_ioctl_fops = {
	open:		cryptowrap_ioctl_open
};

static int cryptowrap_des_open(struct inode *inode, struct file *filp);
static int cryptowrap_des_release(struct inode *inode, struct file *filp);
static int cryptowrap_des_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg);

struct file_operations cryptowrap_des_fops = {
	open:		cryptowrap_des_open,
	release:	cryptowrap_des_release,
	ioctl:		cryptowrap_des_ioctl
};

static int cryptowrap_3des_open(struct inode *inode, struct file *filp);
static int cryptowrap_3des_release(struct inode *inode, struct file *filp);
static int cryptowrap_3des_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg);

struct file_operations cryptowrap_3des_fops = {
	open:		cryptowrap_3des_open,
	release:	cryptowrap_3des_release,
	ioctl:		cryptowrap_3des_ioctl
};

static int cryptowrap_aes_open(struct inode *inode, struct file *filp);
static int cryptowrap_aes_release(struct inode *inode, struct file *filp);
static int cryptowrap_aes_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg);

struct file_operations cryptowrap_aes_fops = {
	open:		cryptowrap_aes_open,
	release:	cryptowrap_aes_release,
	ioctl:		cryptowrap_aes_ioctl
};

static int cryptowrap_rsa_open(struct inode *inode, struct file *filp);
static int cryptowrap_rsa_release(struct inode *inode, struct file *filp);
static int cryptowrap_rsa_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg);

struct file_operations cryptowrap_rsa_fops = {
	open:		cryptowrap_rsa_open,
	release:	cryptowrap_rsa_release,
	ioctl:		cryptowrap_rsa_ioctl
};

static int cryptowrap_sha1_open(struct inode *inode, struct file *filp);
static int cryptowrap_sha1_release(struct inode *inode, struct file *filp);
static int cryptowrap_sha1_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg);

struct file_operations cryptowrap_sha1_fops = {
	open:		cryptowrap_sha1_open,
	release:	cryptowrap_sha1_release,
	ioctl:		cryptowrap_sha1_ioctl
};

static int cryptowrap_md5_open(struct inode *inode, struct file *filp);
static int cryptowrap_md5_release(struct inode *inode, struct file *filp);
static int cryptowrap_md5_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg);

struct file_operations cryptowrap_md5_fops = {
	open:		cryptowrap_md5_open,
	release:	cryptowrap_md5_release,
	ioctl:		cryptowrap_md5_ioctl
};

static int cryptowrap_hmac_open(struct inode *inode, struct file *filp);
static int cryptowrap_hmac_release(struct inode *inode, struct file *filp);
static int cryptowrap_hmac_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg);

struct file_operations cryptowrap_hmac_fops = {
	open:		cryptowrap_hmac_open,
	release:	cryptowrap_hmac_release,
	ioctl:		cryptowrap_hmac_ioctl
};

struct file_operations *cryptowrap_ioctl_fop_array[] = {
	&cryptowrap_des_fops,	/* DES  = 0 */
	&cryptowrap_3des_fops,	/* 3DES = 1 */
	&cryptowrap_rsa_fops,	/* RSA  = 2 */
	&cryptowrap_sha1_fops,	/* SHA1 = 3 */
	&cryptowrap_md5_fops,	/* MD5  = 4 */
	&cryptowrap_hmac_fops,	/* HMAC = 5 */
	&cryptowrap_aes_fops,	/* AES = 6 */
};

#define CRYPTOWRAP_IOCTL_ALGORITHM_MAX 6

static int cryptowrap_ioctl_open(struct inode *inode, struct file *filp)
{
	int type = ALGORITHM(inode->i_rdev);

	if (type > CRYPTOWRAP_IOCTL_ALGORITHM_MAX)
		return -ENODEV;

	/* dispatch fops */
	filp->f_op = cryptowrap_ioctl_fop_array[type];
	return filp->f_op->open(inode, filp);
}

static int cryptowrap_des_open(struct inode *inode, struct file *filp)
{
	int mode = MODE(inode->i_rdev);
	int error;

	error = CryptoWrap_DES_new((CryptoWrap_DES_object *)&filp->private_data, mode);

	switch (error) {
	case CryptoWrap_E_NOTSUPPORTED:
		return -EINVAL;
	case CryptoWrap_E_NOMEM:
		return -ENOMEM;
	}

	return 0;
}

static int cryptowrap_des_release(struct inode *inode, struct file *filp)
{
	CryptoWrap_DES_destroy((CryptoWrap_DES_object)filp->private_data);

	return 0;
}

static int cryptowrap_des_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	CryptoWrap_DES_object obj = (CryptoWrap_DES_object)filp->private_data;
	CryptoWrap_Cipher_Data *data;
	CryptoWrap_Cipher_Key *key;
	int error = 0;

	switch (cmd) {

	case CRYPTOWRAP_IOCENCRYPT:
		data = kmalloc(sizeof(*data), GFP_KERNEL);
		if (data == NULL)
			return -ENOMEM;
		if (copy_from_user((void *)data, (void *)arg, sizeof(*data))) {
			kfree(data);
			return -EFAULT;
		}
		error = CryptoWrap_DES_encrypt(obj,
						data->des.from,
						data->des.flen,
						data->des.to,
						&data->des.tlen,
						data->des.iv);
		if (copy_to_user((void *)arg, (void *)data, sizeof(*data))) {
			kfree(data);
			return -EFAULT;
		}
		kfree(data);
		break;
	case CRYPTOWRAP_IOCDECRYPT:
		data = kmalloc(sizeof(*data), GFP_KERNEL);
		if (data == NULL)
			return -ENOMEM;
		if (copy_from_user((void *)data, (void *)arg, sizeof(*data))) {
			kfree(data);
			return -EFAULT;
		}
		error = CryptoWrap_DES_decrypt(obj,
						data->des.from,
						data->des.flen,
						data->des.to,
						&data->des.tlen,
						data->des.iv);
		if (copy_to_user((void *)arg, (void *)data, sizeof(*data))) {
			kfree(data);
			return -EFAULT;
		}
		kfree(data);
		break;
	case CRYPTOWRAP_IOCSETKEY:
		key = kmalloc(sizeof(*key), GFP_KERNEL);
		if (copy_from_user((void *)key, (void *)arg, sizeof(*key))) {
			kfree(key);
			return -EFAULT;
		}
		error = CryptoWrap_DES_set_key(obj, key->des.key);
		kfree(key);
		break;
	default:
		return -EINVAL;
	}

	if (error)
		return -EINVAL;	/* XXX */

	return 0;
}

static int cryptowrap_3des_open(struct inode *inode, struct file *filp)
{
	int mode = MODE(inode->i_rdev);
	int error;

	error = CryptoWrap_3DES_new((CryptoWrap_3DES_object *)&filp->private_data, mode);

	switch (error) {
	case CryptoWrap_E_NOTSUPPORTED:
		return -EINVAL;
	case CryptoWrap_E_NOMEM:
		return -ENOMEM;
	}

	return 0;
}

static int cryptowrap_3des_release(struct inode *inode, struct file *filp)
{
	CryptoWrap_3DES_destroy((CryptoWrap_3DES_object)filp->private_data);

	return 0;
}

static int cryptowrap_3des_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	CryptoWrap_3DES_object obj = (CryptoWrap_3DES_object)filp->private_data;
	CryptoWrap_Cipher_Data *data;
	CryptoWrap_Cipher_Key *key;
	int error = 0;

	switch (cmd) {

	case CRYPTOWRAP_IOCENCRYPT:
		data = kmalloc(sizeof(*data), GFP_KERNEL);
		if (data == NULL)
			return -ENOMEM;
		if (copy_from_user((void *)data, (void *)arg, sizeof(*data))) {
			kfree(data);
			return -EFAULT;
		}
		error = CryptoWrap_3DES_encrypt(obj,
						data->triple_des.from,
						data->triple_des.flen,
						data->triple_des.to,
						&data->triple_des.tlen,
						data->triple_des.iv);
		if (copy_to_user((void *)arg, (void *)data, sizeof(*data))) {
			kfree(data);
			return -EFAULT;
		}
		kfree(data);
		break;
	case CRYPTOWRAP_IOCDECRYPT:
		data = kmalloc(sizeof(*data), GFP_KERNEL);
		if (data == NULL)
			return -ENOMEM;
		if (copy_from_user((void *)data, (void *)arg, sizeof(*data))) {
			kfree(data);
			return -EFAULT;
		}
		error = CryptoWrap_3DES_decrypt(obj,
						data->triple_des.from,
						data->triple_des.flen,
						data->triple_des.to,
						&data->triple_des.tlen,
						data->triple_des.iv);
		if (copy_to_user((void *)arg, (void *)data, sizeof(*data))) {
			kfree(data);
			return -EFAULT;
		}
		kfree(data);
		break;
	case CRYPTOWRAP_IOCSETKEY:
		key = kmalloc(sizeof(*key), GFP_KERNEL);
		if (key == NULL)
			return -ENOMEM;
		if (copy_from_user((void *)key, (void *)arg, sizeof(*key))) {
			kfree(key);
			return -EFAULT;
		}
		error = CryptoWrap_3DES_set_key(obj,
						key->triple_des.key1,
						key->triple_des.key2,
						key->triple_des.key3);
		kfree(key);
		break;
	default:
		return -EINVAL;
	}

	if (error)
		return -EINVAL;	/* XXX */

	return 0;
}

static int cryptowrap_aes_open(struct inode *inode, struct file *filp)
{
	int mode = MODE(inode->i_rdev);
	int error;

	error = CryptoWrap_AES_new((CryptoWrap_AES_object *)&filp->private_data, mode);

	switch (error) {
	case CryptoWrap_E_NOTSUPPORTED:
		return -EINVAL;
	case CryptoWrap_E_NOMEM:
		return -ENOMEM;
	}

	return 0;
}

static int cryptowrap_aes_release(struct inode *inode, struct file *filp)
{
	CryptoWrap_AES_destroy((CryptoWrap_AES_object)filp->private_data);

	return 0;
}

static int cryptowrap_aes_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	CryptoWrap_AES_object obj = (CryptoWrap_AES_object)filp->private_data;
	CryptoWrap_Cipher_Data *data;
	CryptoWrap_Cipher_Key *key;
	int error = 0;

	switch (cmd) {

	case CRYPTOWRAP_IOCENCRYPT:
		data = kmalloc(sizeof(*data), GFP_KERNEL);
		if (data == NULL)
			return -ENOMEM;
		if (copy_from_user((void *)data, (void *)arg, sizeof(*data))) {
			kfree(data);
			return -EFAULT;
		}
		error = CryptoWrap_AES_encrypt(obj,
						data->aes.from,
						data->aes.flen,
						data->aes.to,
						&data->aes.tlen,
						data->aes.iv);
		if (copy_to_user((void *)arg, (void *)data, sizeof(*data))) {
			kfree(data);
			return -EFAULT;
		}
		kfree(data);
		break;
	case CRYPTOWRAP_IOCDECRYPT:
		data = kmalloc(sizeof(*data), GFP_KERNEL);
		if (data == NULL)
			return -ENOMEM;
		if (copy_from_user((void *)data, (void *)arg, sizeof(*data))) {
			kfree(data);
			return -EFAULT;
		}
		error = CryptoWrap_AES_decrypt(obj,
						data->aes.from,
						data->aes.flen,
						data->aes.to,
						&data->aes.tlen,
						data->aes.iv);
		if (copy_to_user((void *)arg, (void *)data, sizeof(*data))) {
			kfree(data);
			return -EFAULT;
		}
		kfree(data);
		break;
	case CRYPTOWRAP_IOCSETKEY:
		key = kmalloc(sizeof(*key), GFP_KERNEL);
		if (copy_from_user((void *)key, (void *)arg, sizeof(*key))) {
			kfree(key);
			return -EFAULT;
		}
		error = CryptoWrap_AES_set_key(obj, key->aes.key, key->aes.keylen);
		kfree(key);
		break;
	default:
		return -EINVAL;
	}

	if (error)
		return -EINVAL;	/* XXX */

	return 0;
}

static int cryptowrap_rsa_open(struct inode *inode, struct file *filp)
{
	int error;

	error = CryptoWrap_RSA_new((CryptoWrap_RSA_object *)&filp->private_data);

	switch (error) {
	case CryptoWrap_E_NOTSUPPORTED:
		return -EINVAL;
	case CryptoWrap_E_NOMEM:
		return -ENOMEM;
	}

	return 0;
}

static int cryptowrap_rsa_release(struct inode *inode, struct file *filp)
{
	CryptoWrap_RSA_destroy((CryptoWrap_RSA_object)filp->private_data);

	return 0;
}

static int cryptowrap_rsa_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	CryptoWrap_RSA_object obj = (CryptoWrap_RSA_object)filp->private_data;

	switch (cmd) {

	case CRYPTOWRAP_IOCINIT:
		if (CryptoWrap_RSA_test(obj) != 0)
			return -EINVAL;
		break;
	default:
		return -EINVAL;
	}
	return 0;
}

static int cryptowrap_sha1_open(struct inode *inode, struct file *filp)
{
	int error;

	error = CryptoWrap_SHA1_new((CryptoWrap_SHA1_object *)&filp->private_data);

	switch (error) {
	case CryptoWrap_E_NOTSUPPORTED:
		return -EINVAL;
	case CryptoWrap_E_NOMEM:
		return -ENOMEM;
	}

	return 0;
}

static int cryptowrap_sha1_release(struct inode *inode, struct file *filp)
{
	CryptoWrap_SHA1_destroy((CryptoWrap_SHA1_object)filp->private_data);

	return 0;
}

static int cryptowrap_sha1_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	CryptoWrap_SHA1_object obj = (CryptoWrap_SHA1_object)filp->private_data;
	CryptoWrap_Digest_Data *data;
	CryptoWrap_Digest_Value *value;

	switch (cmd) {

	case CRYPTOWRAP_IOCINIT:
		CryptoWrap_SHA1_Init(obj);
		break;
	case CRYPTOWRAP_IOCUPDATE:
		data = kmalloc(sizeof(*data), GFP_KERNEL);
		if (data == NULL)
			return -ENOMEM;
		if (copy_from_user((void *)data, (void *)arg, sizeof(*data))) {
			kfree(data);
			return -EFAULT;
		}
		CryptoWrap_SHA1_Update(obj, data->data, data->len);
		kfree(data);
		break;
	case CRYPTOWRAP_IOCFINAL:
		value = kmalloc(sizeof(*value), GFP_KERNEL);
		if (value == NULL)
			return -ENOMEM;
		memset((void *)value, 0, sizeof(*value));
		value->len = 20;
		CryptoWrap_SHA1_Final(obj, value->md);
		if (copy_to_user((void *)arg, (void *)value, sizeof(*value))) {
			kfree(value);
			return -EFAULT;
		}
		kfree(value);
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

static int cryptowrap_md5_open(struct inode *inode, struct file *filp)
{
	int error;

	error = CryptoWrap_MD5_new((CryptoWrap_MD5_object *)&filp->private_data);

	switch (error) {
	case CryptoWrap_E_NOTSUPPORTED:
		return -EINVAL;
	case CryptoWrap_E_NOMEM:
		return -ENOMEM;
	}

	return 0;
}

static int cryptowrap_md5_release(struct inode *inode, struct file *filp)
{
	CryptoWrap_MD5_destroy((CryptoWrap_MD5_object)filp->private_data);

	return 0;
}

static int cryptowrap_md5_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	CryptoWrap_MD5_object obj = (CryptoWrap_MD5_object)filp->private_data;
	CryptoWrap_Digest_Data *data;
	CryptoWrap_Digest_Value *value;

	switch (cmd) {

	case CRYPTOWRAP_IOCINIT:
		CryptoWrap_MD5_Init(obj);
		break;
	case CRYPTOWRAP_IOCUPDATE:
		data = kmalloc(sizeof(*data), GFP_KERNEL);
		if (data == NULL)
			return -ENOMEM;
		if (copy_from_user((void *)data, (void *)arg, sizeof(*data))) {
			kfree(data);
			return -EFAULT;
		}
		CryptoWrap_MD5_Update(obj, data->data, data->len);
		kfree(data);
		break;
	case CRYPTOWRAP_IOCFINAL:
		value = kmalloc(sizeof(*value), GFP_KERNEL);
		if (value == NULL)
			return -ENOMEM;
		memset((void *)value, 0, sizeof(*value));
		value->len = 16;
		CryptoWrap_MD5_Final(obj, value->md);
		if (copy_to_user((void *)arg, (void *)value, sizeof(*value))) {
			kfree(value);
			return -EFAULT;
		}
		kfree(value);
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

static int cryptowrap_hmac_open(struct inode *inode, struct file *filp)
{
	int mode = MODE(inode->i_rdev);
	int error;

	error = CryptoWrap_HMAC_new((CryptoWrap_MD5_object *)&filp->private_data, mode);

	switch (error) {
	case CryptoWrap_E_NOTSUPPORTED:
		return -EINVAL;
	case CryptoWrap_E_NOMEM:
		return -ENOMEM;
	}

	return 0;
}

static int cryptowrap_hmac_release(struct inode *inode, struct file *filp)
{
	CryptoWrap_HMAC_destroy((CryptoWrap_HMAC_object)filp->private_data);

	return 0;
}

static int cryptowrap_hmac_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	CryptoWrap_HMAC_object obj = (CryptoWrap_HMAC_object)filp->private_data;
	CryptoWrap_Digest_Data *data;
	CryptoWrap_Digest_Value *value;
	CryptoWrap_Auth_Key *key;

	switch (cmd) {

	case CRYPTOWRAP_IOCHINIT:
		key = kmalloc(sizeof(*key), GFP_KERNEL);
		if (key == NULL)
			return -ENOMEM;
		if (copy_from_user((void *)key, (void *)arg, sizeof(*key))) {
			kfree(key);
			return -EFAULT;
		}
		CryptoWrap_HMAC_Init(obj, key->key, key->len);
		kfree(key);
		break;
	case CRYPTOWRAP_IOCUPDATE:
		data = kmalloc(sizeof(*data), GFP_KERNEL);
		if (data == NULL)
			return -ENOMEM;
		if (copy_from_user((void *)data, (void *)arg, sizeof(*data))) {
			kfree(data);
			return -EFAULT;
		}
		CryptoWrap_HMAC_Update(obj, data->data, data->len);
		kfree(data);
		break;
	case CRYPTOWRAP_IOCFINAL:
		value = kmalloc(sizeof(*value), GFP_KERNEL);
		if (value == NULL)
			return -ENOMEM;
		memset((void *)value, 0, sizeof(*value));
		CryptoWrap_HMAC_Final(obj, value->md, &value->len);
		if (copy_to_user((void *)arg, (void *)value, sizeof(*value))) {
			kfree(value);
			return -EFAULT;
		}
		kfree(value);
		break;
	default:
		return -EINVAL;
	}

	return 0;
}


static int cryptowrap_ioctl_module_init(void);
static void cryptowrap_ioctl_module_exit(void);

static int cryptowrap_ioctl_major = CRYPTOWRAP_IOCTL_MAJOR;

static int cryptowrap_ioctl_module_init(void)
{
	int result;

	result = register_chrdev(cryptowrap_ioctl_major, "cryptowrap-ioctl", &cryptowrap_ioctl_fops);
	if (result < 0) {
		printk(KERN_WARNING "cryptowrap-ioctl: can't get major %d\n", cryptowrap_ioctl_major);
		return result;
	}
	if (cryptowrap_ioctl_major == 0)
		cryptowrap_ioctl_major = result;

	printk(KERN_INFO "cryptowrap-ioctl: loaded(major = %d)\n", cryptowrap_ioctl_major);

	return 0;
}

static void cryptowrap_ioctl_module_exit(void)
{
	int error;

	error = unregister_chrdev(cryptowrap_ioctl_major, "cryptowrap-ioctl");
	if (error < 0)
		printk(KERN_WARNING "cryptowrap-ioctl: can't release major %d\n", cryptowrap_ioctl_major);

	printk(KERN_INFO "cryptowrap-ioctl: unloaded\n");
}

module_init(cryptowrap_ioctl_module_init);
module_exit(cryptowrap_ioctl_module_exit);
