/*
 *  CXD9790 core module
 *
 *  Copyright 2003 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/irq.h>
#include <linux/proc_fs.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <linux/cxd9790.h>

#define MODNAME              "CXD9790"

/* proc */
#define CXD9790_PROC_NAME    "cxd9790"

static spinlock_t cxd9790_lock = SPIN_LOCK_UNLOCKED;
static int cxd9790_initialized = 0;
static int cxd9790_probed = 0;

MODULE_AUTHOR("Sony NSC");
MODULE_DESCRIPTION("CXD9790 core");
MODULE_LICENSE("GPL");


/*
 *  CXD9790 is present or not
 */
int
cxd9790_present(void)
{
	spin_lock(&cxd9790_lock);
	if (!cxd9790_initialized) {
		/* check CXD9790 CID */
		if (CXD9790_READ(CXD9790_CID) == CXD9790_ID) {
			printk(MODNAME ": chip revision 0x%x\n",
			       CXD9790_READ(CXD9790_CREV));
			cxd9790_probed = 1;
		} else {
			printk(MODNAME ": not found\n");
		}

		cxd9790_initialized = 1;
	}
	spin_unlock(&cxd9790_lock);

	return cxd9790_probed;
}

#ifdef CONFIG_SNSC_CXD9790_ES1_WORKAROUND
/*
 *  CXD9790 is ES1?
 */
int
cxd9790_is_es1(void)
{
	static int is_es1 = -1;

	if (is_es1 >= 0)
		return is_es1;

	if (CXD9790_READ(CXD9790_CREV) == CXD9790_ES1)
		is_es1 = 1;
	else
		is_es1 = 0;
	return is_es1;
}
#endif /* CONFIG_SNSC_CXD9790_ES1_WORKAROUND */

#ifdef CONFIG_SNSC_CXD9790_PROC
/* proc */
static int
cx9790_read_proc(char *buf, char **start, off_t off,
		 int count, int *eof, void *data)
{
	int i,j, len = 0;
	u_int8_t reg, base;
	const char *irc_str[] = {"UART0", "UART1", "UART2", "UART3",
				 " I2C ", " GPCD", " GPAB", " GPZ "};
	const char gpio_str[] = "ZABCD";
	const char *trig_str[] = {"LL", "HL", "FE", "RE"};

	/* CID, CREV */
	len += sprintf(buf + len, "Chip ID  : 0x%02x", 
		       CXD9790_READ(CXD9790_CID));
	len += sprintf(buf + len, " (Rev : 0x%02x)\n",
		       CXD9790_READ(CXD9790_CREV));

	/* GPI */
	len += sprintf(buf + len, "GPI      : 0x%02x\n", 
		       CXD9790_READ(CXD9790_CPI));

	/* CKEN */
	reg = CXD9790_READ(CXD9790_CKEN);
	len += sprintf(buf + len, "CKEN     :");
	for (i = 4; i >= 0; i--) {
		if (reg & (1 << i))
			len += sprintf(buf + len, " %s", irc_str[i]);
	}
	len += sprintf(buf + len, "\n\n");

	/* Interrupts */
	len += sprintf(buf + len, "[Common Interrupts]\n");
	len += sprintf(buf + len, "          ");
	for (i = 7; i >= 0; i--) {
		len += sprintf(buf + len, "%s ", irc_str[i]);
	}
	len += sprintf(buf + len, "\n");
	reg = CXD9790_READ(CXD9790_IROUT);
	len += sprintf(buf + len, "INTO0    :");
	for (i = 7; i >= 0; i--) {
		if (!(reg & (1 << i)))
			len += sprintf(buf + len, "  *   ");
		else
			len += sprintf(buf + len, "      ");
	}
	len += sprintf(buf + len, "\n");
	len += sprintf(buf + len, "INTO1    :");
	for (i = 7; i >= 0; i--) {
		if (reg & (1 << i))
			len += sprintf(buf + len, "  *   ");
		else
			len += sprintf(buf + len, "      ");
	}
	len += sprintf(buf + len, "\n");
	len += sprintf(buf + len, "IRCS0    : 0x%02x\n", 
		       CXD9790_READ(CXD9790_IRCS0));
	len += sprintf(buf + len, "IRCS1    : 0x%02x\n", 
		       CXD9790_READ(CXD9790_IRCS1));
#ifdef CONFIG_SNSC_CXD9790_ES1_WORKAROUND
	if (!cxd9790_is_es1())
#endif
	len += sprintf(buf + len, "IRCMSK   : 0x%02x\n", 
		       CXD9790_READ(CXD9790_IRCMSK));

	/* GPIO */
	len += sprintf(buf + len, "\n[GPIO]\n");
	len += sprintf(buf + len, "(LL:Low Level, HL:High Level,"
		       " FE: Falling Edge, RE: Rising Edge)\n");
	for (i = 0; i < 5; i++) {
		base = CXD9790_GPIOZ + (i << 4);
		len += sprintf(buf + len, "<%c>", gpio_str[i]);
		len += sprintf(buf + len, " IRDM    :");
		reg = CXD9790_READ(base + CXD9790_IRDM1);
		for (j = 3; j >= 0; j--) {
			len += sprintf(buf + len, " %s",
				       trig_str[(reg >> (j << 1)) & 0x3]);
		}
		reg = CXD9790_READ(base + CXD9790_IRDM0);
		for (j = 3; j >= 0; j--) {
			len += sprintf(buf + len, " %s",
				       trig_str[(reg >> (j << 1)) & 0x3]);
		}
		len += sprintf(buf + len, "\n");

		len += sprintf(buf + len, "    IRMSK   : 0x%02x", 
			       CXD9790_READ(base + CXD9790_IRMSK));
		len += sprintf(buf + len, "            IRSTS   : 0x%02x\n", 
			       CXD9790_READ(base + CXD9790_IRSTS));
		len += sprintf(buf + len, "    POEN    : 0x%02x", 
			       CXD9790_READ(base + CXD9790_POEN));
		len += sprintf(buf + len, "            PI      : 0x%02x\n", 
			       CXD9790_READ(base + CXD9790_PI));
		len += sprintf(buf + len, "\n");
	}

	*start = buf + off;
	len -= off;
	if (len <= count)
		*eof = 1;
	if (len < 0)
		len = 0;
	else if (len > count)
		len = count;
	return len;
}
#endif /* CONFIG_SNSC_CXD9790_PROC */


/*
 *  initalized CXD9790 module
 */
static int
cxd9790_init(void)
{
	if (cxd9790_present()) {
#ifdef CONFIG_SNSC_CXD9790_PROC
		/* setup proc entry */
		if (!create_proc_read_entry(CXD9790_PROC_NAME, 0, NULL,
					    cx9790_read_proc, NULL)) {
			printk(MODNAME ": unable to create proc entry %s\n",
			       CXD9790_PROC_NAME);
			return -ENOMEM;
		}
#endif /* CONFIG_SNSC_CXD9790_PROC */
	} else
		return -ENODEV;

	return 0;
}

static void __exit
cxd9790_exit(void)
{
#ifdef CONFIG_SNSC_CXD9790_PROC
	remove_proc_entry(CXD9790_PROC_NAME, NULL);
#endif /* CONFIG_SNSC_CXD9790_PROC */
}

module_init(cxd9790_init);
module_exit(cxd9790_exit);
EXPORT_SYMBOL(cxd9790_present);
#ifdef CONFIG_SNSC_CXD9790_ES1_WORKAROUND
EXPORT_SYMBOL(cxd9790_is_es1);
#endif
