/*
 *  CXD9790 GPIO driver
 *
 *  Copyright 2003 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/irq.h>
#include <linux/proc_fs.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>

#include <linux/cxd9790_gpio.h>

#define MODNAME                 "cxd9790_gpio"
#define MSG_HEAD                "cxd9790_gpio : "

#define MAX_NAME_SIZE           15

struct gpio_t {
	__u8 used;
	__u8 inout;
	__u8 intr;
	struct {
		char  name[MAX_NAME_SIZE + 1];
	} bit[8];
};

static struct gpio_t gpio[GPIO_PORT_D + 1];
static spinlock_t gpio_lock = SPIN_LOCK_UNLOCKED;

#define PORTBIT_TO_IRQ(port, bit) \
	(CXD9790_IRQ_GPIOZ(bit) + ((port) << 3))

#define PORT_NAME(port)         ("ZABCD"[port])

#define PORT_BASE(port)         (((port) + 1) << 4)

#define FOR_EACH_BIT(bit, bitmask) \
	for (bit = 0; bit < 8 && bitmask; bit++, bitmask >>= 1) \
		if (bitmask & 1)

static inline void
__cxd9790_set_bits(int port, int reg, __u8 bitmask)
{
	__u8 val;

	val = CXD9790_READ(PORT_BASE(port) + reg);
	CXD9790_WRITE(PORT_BASE(port) + reg, val | bitmask);
}

static inline void
__cxd9790_clr_bits(int port, int reg, __u8 bitmask)
{
	__u8 val;

	val = CXD9790_READ(PORT_BASE(port) + reg);
	CXD9790_WRITE(PORT_BASE(port) + reg, val & ~bitmask);
}

static inline void
__cxd9790_mod_bits(int port, int reg, __u8 bitmask)
{
	__u8 val;

	val = CXD9790_READ(PORT_BASE(port) + reg);
	CXD9790_WRITE(PORT_BASE(port) + reg, val & ~bitmask);
}

static int
check_port(int port)
{
	if ((port < GPIO_PORT_Z) || (port > GPIO_PORT_D)) {
		printk(KERN_ERR MSG_HEAD "invalid port %d\n", port);
		return -EINVAL;
	}
	return 0;
}

static int
check_port_bitnum(int port, int bitnum)
{
	if ((port < GPIO_PORT_Z) || (port > GPIO_PORT_D)) {
		printk(KERN_ERR MSG_HEAD "invalid port %d\n", port);
		return -EINVAL;
	}

	if ((bitnum < 0) || bitnum >= 8) {
		printk(KERN_ERR MSG_HEAD "invalid port %d\n", port);
		return -EINVAL;
	}

	return 0;
}

/* register GPIO (per bit) */
int cxd9790_gpio_register_bit(int port, int bitnum, const char *name)
{
	int ret = 0;
	unsigned long flags;

	if ((ret = check_port_bitnum(port, bitnum)) < 0)
		return ret;

	spin_lock_irqsave(&gpio_lock, flags);
	if (gpio[port].used & (1 << bitnum)) {
		printk(KERN_ERR MSG_HEAD 
		       "GPIO %c bit%d is already reserved\n",
		       PORT_NAME(port), bitnum);
		ret = -EBUSY;
	} else {
		gpio[port].used |= (1 << bitnum);
		strncpy(gpio[port].bit[bitnum].name, name, MAX_NAME_SIZE);
	}
	spin_unlock_irqrestore(&gpio_lock, flags);

	return ret;
}

/* unregister GPIO (per bit) */
int cxd9790_gpio_unregister_bit(int port, int bitnum)
{
	int ret = 0;
	unsigned long flags;

	if ((ret = check_port_bitnum(port, bitnum)) < 0)
		return ret;

	spin_lock_irqsave(&gpio_lock, flags);
	if (gpio[port].used & (1 << bitnum)) {
		gpio[port].used &= ~(1 << bitnum);
		gpio[port].inout &= ~(1 << bitnum);
		cxd9790_gpio_intr_mask_bit(port, bitnum);
		memset(gpio[port].bit[bitnum].name, 0, MAX_NAME_SIZE);
	} else {
		printk(KERN_ERR MSG_HEAD
		       "GPIO %c bit%d is not reserved\n",
		       PORT_NAME(port), bitnum);
		ret = -EBUSY;
	}

	spin_unlock_irqrestore(&gpio_lock, flags);

	return ret;
}

/* register GPIO (via bitmask) */
int cxd9790_gpio_register(int port, __u8 bitmask, const char *name)
{
	int ret = 0;
	int bit;
	unsigned long flags;

	if ((ret = check_port(port)) < 0)
		return ret;

	spin_lock_irqsave(&gpio_lock, flags);
	if (gpio[port].used & bitmask) {
		printk(KERN_ERR MSG_HEAD
		       "GPIO %c bitmask 0x%x is reserved\n",
		       PORT_NAME(port), gpio[port].used & bitmask);
		ret = -EBUSY;
	} else {
		gpio[port].used |= bitmask;
		FOR_EACH_BIT(bit, bitmask) {
			strncpy(gpio[port].bit[bit].name,
				name, MAX_NAME_SIZE);
		}
	}
	spin_unlock_irqrestore(&gpio_lock, flags);

	return ret;
}
 
int cxd9790_gpio_unregister(int port, __u8 bitmask)
{
	int ret = 0;
	int bit;
	unsigned long flags;

	if ((ret = check_port(port)) < 0)
		return ret;

	spin_lock_irqsave(&gpio_lock, flags);
	if ((gpio[port].used & bitmask) == bitmask) {
		gpio[port].used &= ~bitmask;
		gpio[port].inout &= ~bitmask;
		cxd9790_gpio_intr_mask(port, bitmask);
		FOR_EACH_BIT(bit, bitmask) {
			memset(gpio[port].bit[bit].name, 
			       0, MAX_NAME_SIZE);
		}
	} else {
		printk(KERN_ERR MSG_HEAD 
		       "GPIO %c bitmask 0x%x is not reserved\n",
		       PORT_NAME(port), 
		       (~gpio[port].used) & bitmask);
		ret = -EBUSY;
	}

	spin_unlock_irqrestore(&gpio_lock, flags);

	return ret;
}

/* unregister GPIOs by name */
int cxd9790_gpio_unregister_by_name(const char *name)
{
	int port, bit;
	int ret = 0;

	for (port = GPIO_PORT_Z; port <= GPIO_PORT_D; port++) {
		for (bit = 0; bit < 8; bit++) {
			if (strncmp(gpio[port].bit[bit].name, name, MAX_NAME_SIZE) == 0)
				if ((ret = cxd9790_gpio_unregister_bit(port, bit)) < 0)
					return ret;
		}
	} 

	return 0;
}

/* set input or output (per bit) */
int cxd9790_gpio_set_mode_bit(int port, int bitnum, __u32 mode)
{
	int ret = 0;
	unsigned long flags;

	if ((ret = check_port_bitnum(port, bitnum)) < 0)
		return ret;

	spin_lock_irqsave(&gpio_lock, flags);
	if (mode & __MODE_OUT) {
		if (mode & __MODE_HIGH)
			__cxd9790_set_bits(port, CXD9790_PO, (1 << bitnum));
		else
			__cxd9790_clr_bits(port, CXD9790_PO, (1 << bitnum));
		__cxd9790_set_bits(port, CXD9790_POEN, (1 << bitnum));
		gpio[port].inout |= (1 << bitnum);
	} else {
		__cxd9790_clr_bits(port, CXD9790_POEN, (1 << bitnum));
		gpio[port].inout &= ~(1 << bitnum);
	}
	spin_unlock_irqrestore(&gpio_lock, flags);

	return 0;
}

/* set input or output (via bitmask) */
int cxd9790_gpio_set_mode(int port, __u8 bitmask, __u32 mode)
{
	int ret = 0;
	unsigned long flags;

	if ((ret = check_port(port)) < 0)
		return ret;

	spin_lock_irqsave(&gpio_lock, flags);
	if (mode & __MODE_OUT) {
		if (mode & __MODE_HIGH)
			__cxd9790_set_bits(port, CXD9790_PO, bitmask);
		else
			__cxd9790_clr_bits(port, CXD9790_PO, bitmask);
		__cxd9790_set_bits(port, CXD9790_POEN, bitmask);
		gpio[port].inout |= bitmask;
	} else {
		__cxd9790_clr_bits(port, CXD9790_POEN, bitmask);
		gpio[port].inout &= ~bitmask;
	}

	spin_unlock_irqrestore(&gpio_lock, flags);

	return 0;
}

/* get value of GPIO (per bit) */
int cxd9790_gpio_get_data_bit(int port, int bitnum)
{
	int  ret = 0;

	if ((ret = check_port_bitnum(port, bitnum)) < 0)
		return ret;

	return (CXD9790_READ(PORT_BASE(port) + CXD9790_PI) 
		& (1 << bitnum) ? 1 : 0);
}

/* set value of GPIO (per bit) */
int cxd9790_gpio_set_data_bit(int port, int bitnum, int value)
{
	int  ret = 0;
	unsigned long flags;

	if ((ret = check_port_bitnum(port, bitnum)) < 0)
		return ret;

	spin_lock_irqsave(&gpio_lock, flags);
	if (gpio[port].inout & (1 << bitnum)) {
		if (value)
			__cxd9790_set_bits(port, CXD9790_PO, (1 << bitnum));
		else
			__cxd9790_clr_bits(port, CXD9790_PO, (1 << bitnum));
	}
	spin_unlock_irqrestore(&gpio_lock, flags);

	return 0;
}

/* get value of GPIO (via bitmask) */
int cxd9790_gpio_get_data(int port, __u8 bitmask, __u8 *value)
{
	int  ret = 0;

	if ((ret = check_port(port)) < 0)
		return ret;

	*value = CXD9790_READ(PORT_BASE(port) + CXD9790_PI) & bitmask;

	return 0;
}

/* set value of GPIO (via bitmask) */
int cxd9790_gpio_set_data(int port, __u8 bitmask, __u8 value)
{
	int  ret = 0;
	__u8 val;
	unsigned long flags;

	if ((ret = check_port(port)) < 0)
		return ret;

	spin_lock_irqsave(&gpio_lock, flags);
	val = CXD9790_READ(PORT_BASE(port) + CXD9790_PI);
	CXD9790_WRITE(PORT_BASE(port) + CXD9790_PO, 
		      ((val & gpio[port].inout) & ~bitmask ) | value);
	spin_unlock_irqrestore(&gpio_lock, flags);

	return 0;
}

/* set interrupt mode (per bit) */
int cxd9790_gpio_intr_set_mode_bit(int port, int bitnum, __u32 mode)
{
	int  ret = 0;
	unsigned long flags;

	if ((ret = check_port_bitnum(port, bitnum)) < 0)
		return ret;

	spin_lock_irqsave(&gpio_lock, flags);
	if (gpio[port].inout & (1 << bitnum)) {
		printk(KERN_ERR MSG_HEAD 
		       "GPIO %c bit%d is output\n",
		       PORT_NAME(port), bitnum);
		ret = -EINVAL;
	} else {
		ret = cxd9790_set_trigger_mode(
			PORTBIT_TO_IRQ(port, bitnum), mode);
	}
	spin_unlock_irqrestore(&gpio_lock, flags);

	return ret;
}

/* set interrupt mode (via bitmask) */
int cxd9790_gpio_intr_set_mode(int port, __u8 bitmask, __u32 mode)
{
	int ret = 0;
	int bit;
	unsigned long flags;

	if ((ret = check_port(port)) < 0)
		return ret;

	spin_lock_irqsave(&gpio_lock, flags);
	if (gpio[port].inout & bitmask) {
		printk(KERN_ERR MSG_HEAD 
		       "GPIO %c bitmask 0x%x is output\n",
		       PORT_NAME(port), gpio[port].inout & bitmask);
		ret = -EINVAL;
	} else {
		FOR_EACH_BIT(bit, bitmask) {
			ret = cxd9790_set_trigger_mode(
				PORTBIT_TO_IRQ(port, bit), mode);
			if (ret < 0)
				break;
		}
	}

	spin_unlock_irqrestore(&gpio_lock, flags);

	return ret;
}

/* register interrupt handler */
int cxd9790_gpio_intr_register_handler(
	int port, int bitnum, 
	void (*handler)(int irq, void *dev_id, struct pt_regs *regs),
	void *dev_id)
{
	int ret = 0;
	unsigned long flags;

	if ((ret = check_port_bitnum(port, bitnum)) < 0)
		return ret;

	spin_lock_irqsave(&gpio_lock, flags);
	if (gpio[port].inout & (1 << bitnum)) {
		printk(KERN_ERR MSG_HEAD 
		       "GPIO %c bit%d is output\n",
		       PORT_NAME(port), bitnum);
		ret = -EINVAL;
	} else {
		ret = request_irq(PORTBIT_TO_IRQ(port, bitnum), handler, 0, 
				  MODNAME, dev_id);
	}
	spin_unlock_irqrestore(&gpio_lock, flags);

	return ret;
}

/* unregister interrupt handler */
int cxd9790_gpio_intr_unregister_handler(int port, int bitnum,
					 void *dev_id)
{
	int ret;

	if ((ret = check_port_bitnum(port, bitnum)) < 0)
		return ret;

	free_irq(PORTBIT_TO_IRQ(port, bitnum), dev_id);

	return 0;
}

/* unmask interrupt (per bit) */
int cxd9790_gpio_intr_unmask_bit(int port, int bitnum)
{
	int ret;

	if ((ret = check_port_bitnum(port, bitnum)) < 0)
		return ret;

	enable_irq(PORTBIT_TO_IRQ(port, bitnum));
	return 0;
}

/* mask interrupt (per bit) */
int cxd9790_gpio_intr_mask_bit(int port, int bitnum)
{
	int ret;

	if ((ret = check_port_bitnum(port, bitnum)) < 0)
		return ret;

	disable_irq(PORTBIT_TO_IRQ(port, bitnum));
	return 0;
}

/* unmask interrupt (via bitmask) */
int cxd9790_gpio_intr_unmask(int port, __u8 bitmask)
{
	int ret;
	int bit;

	if ((ret = check_port(port)) < 0)
		return ret;

	FOR_EACH_BIT(bit, bitmask) {
		enable_irq(PORTBIT_TO_IRQ(port, bit));
	}
	return 0;
}
 
/* mask interrupt (via bitmask) */
int cxd9790_gpio_intr_mask(int port, __u8 bitmask)
{
	int ret;
	int bit;

	if ((ret = check_port(port)) < 0)
		return ret;

	FOR_EACH_BIT(bit, bitmask) {
		disable_irq(PORTBIT_TO_IRQ(port, bit));
	}
	return 0;
}

/* translate IRQ to port number */
int cxd9790_gpio_irq_to_port(int irq)
{
	return (irq - CXD9790_IRQ_GPIOZ(0)) >> 3;
}

/* translate IRQ to bit number */
int cxd9790_gpio_irq_to_bit(int irq)
{
	return (irq - CXD9790_IRQ_GPIOZ(0)) & 7;
}

#ifdef CONFIG_SNSC_CXD9790_GPIO_PROC
/* proc */
static int
cx9790_gpio_read_proc(char *buf, char **start, off_t off,
		 int count, int *eof, void *data)
{
	int port, len = 0;
	int bit;

	for (port = GPIO_PORT_Z; port <= GPIO_PORT_D; port++) {
		len += sprintf(buf + len, "GPIO %c :\n", PORT_NAME(port));
		for (bit = 0; bit < 8; bit++) {
			if (gpio[port].used & (1 << bit)) {
				len += sprintf(buf + len, 
					       "    bit %d : %s\n", 
					       bit, gpio[port].bit[bit].name);
			}
		}
		len += sprintf(buf + len, "\n");
	}

	*start = buf + off;
	len -= off;
	if (len <= count)
		*eof = 1;
	if (len < 0)
		len = 0;
	else if (len > count)
		len = count;
	return len;
}
#endif /* CONFIG_SNSC_CXD9790_GPIO_PROC */

/*
 *  Initialize CXD9790 GPIO driver
 */
int __init
cxd9790_gpio_init(void)
{
	static int first = 1;

	if (first) {
		if (cxd9790_present()) {
			memset(gpio, 0, sizeof(gpio));
#ifdef CONFIG_SNSC_CXD9790_GPIO_PROC
			/* setup proc entry */
			if (!create_proc_read_entry(MODNAME, 0, NULL,
						    cx9790_gpio_read_proc,
						    NULL)) {
				printk(KERN_ERR MODNAME
				       "unable to create proc entry\n");
				return -ENOMEM;
			}
#endif /* CONFIG_SNSC_CXD9790_GPIO_PROC */
		} else
			return -ENODEV;

		first = 0;
	}

	return 0;
}


static void __exit
cxd9790_gpio_exit(void)
{
#ifdef CONFIG_SNSC_CXD9790_GPIO_PROC
	remove_proc_entry(MODNAME, NULL);
#endif
}

module_init(cxd9790_gpio_init);
module_exit(cxd9790_gpio_exit);

MODULE_AUTHOR("Sony NSC");
MODULE_LICENSE("GPL");

EXPORT_SYMBOL(cxd9790_gpio_register_bit);
EXPORT_SYMBOL(cxd9790_gpio_unregister_bit);
EXPORT_SYMBOL(cxd9790_gpio_register);
EXPORT_SYMBOL(cxd9790_gpio_unregister);
EXPORT_SYMBOL(cxd9790_gpio_unregister_by_name);
EXPORT_SYMBOL(cxd9790_gpio_set_mode_bit);
EXPORT_SYMBOL(cxd9790_gpio_set_mode);
EXPORT_SYMBOL(cxd9790_gpio_get_data_bit);
EXPORT_SYMBOL(cxd9790_gpio_set_data_bit);
EXPORT_SYMBOL(cxd9790_gpio_get_data);
EXPORT_SYMBOL(cxd9790_gpio_set_data);
EXPORT_SYMBOL(cxd9790_gpio_intr_set_mode_bit);
EXPORT_SYMBOL(cxd9790_gpio_intr_set_mode);
EXPORT_SYMBOL(cxd9790_gpio_intr_register_handler);
EXPORT_SYMBOL(cxd9790_gpio_intr_unregister_handler);
EXPORT_SYMBOL(cxd9790_gpio_intr_unmask_bit);
EXPORT_SYMBOL(cxd9790_gpio_intr_mask_bit);
EXPORT_SYMBOL(cxd9790_gpio_intr_unmask);
EXPORT_SYMBOL(cxd9790_gpio_intr_mask);
EXPORT_SYMBOL(cxd9790_gpio_irq_to_port);
EXPORT_SYMBOL(cxd9790_gpio_irq_to_bit);
