/*
 *  CXD9790 interrupt module
 *
 *  Copyright 2003 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/irq.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>

#include <linux/cxd9790.h>

/* IRQ <-> local IRQ */
#define __IRQ_TO_LOCALIRQ(irq)    ((irq) - CXD9790_IRQ_OFFSET)
#define __LOCALIRQ_TO_IRQ(irq)    ((irq) + CXD9790_IRQ_OFFSET)

/* IRQ -> GPIO register address and bit */
#define __IRQ_TO_GPIOREG(irq)     ((__IRQ_TO_LOCALIRQ(irq) & 0x38) << 1)
#define __IRQ_TO_GPIOBIT(irq)     ((__IRQ_TO_LOCALIRQ(irq) & 0x7))

#define __IS_IRCS(irq)                                           \
	(((irq) >= CXD9790_IRQ_OFFSET)                           \
	 && (irq < (CXD9790_IRQ_OFFSET + CXD9790_NR_IRQS_IRCS)))

#define __IS_GPIO(irq)                                           \
	(((irq) >= (CXD9790_IRQ_OFFSET + CXD9790_NR_IRQS_IRCS))  \
	 && ((irq) < (CXD9790_IRQ_OFFSET + CXD9790_NR_IRQS)))

#define __IS_EDGE_TRIG(trig)      (((trig) & IRDM_EDGE) == IRDM_EDGE)

static spinlock_t cxd9790int_lock = SPIN_LOCK_UNLOCKED;


/* debug */
#ifdef CXD9790_DEBUG
#define DPRINTK(fmt, args...)    printk("CXD9790: " fmt, ## args)
#else
#define DPRINTK(fmt, args...)
#endif

MODULE_AUTHOR("Sony NSC");
MODULE_LICENSE("GPL");

/*
 *  "CXD9790 IRCS"
 *      interrupt controller type for CXD9790 IRCSx registers
 */
static void
enable_cxd9790_ircs_irq(unsigned int irq)
{
	u_int8_t      mask;
	unsigned long flags;

	spin_lock_irqsave(&cxd9790int_lock, flags);
	mask = CXD9790_READ(CXD9790_IRCMSK);
	mask &= ~(1 << __IRQ_TO_LOCALIRQ(irq));
	CXD9790_WRITE(CXD9790_IRCMSK, mask);
	spin_unlock_irqrestore(&cxd9790int_lock, flags);
}

static void
disable_cxd9790_ircs_irq(unsigned int irq)
{
	u_int8_t      mask;
	unsigned long flags;

	spin_lock_irqsave(&cxd9790int_lock, flags);
	mask = CXD9790_READ(CXD9790_IRCMSK);
	mask |= (1 << __IRQ_TO_LOCALIRQ(irq));
	CXD9790_WRITE(CXD9790_IRCMSK, mask);
	spin_unlock_irqrestore(&cxd9790int_lock, flags);
}

#define mask_and_ack_cxd9790_ircs_irq           disable_cxd9790_ircs_irq

#ifdef CXD9790_IRQ_TYPE1
#define shutdown_cxd9790_ircs_irq               disable_cxd9790_ircs_irq

static unsigned int
startup_cxd9790_ircs_irq(unsigned int irq)
{
	enable_cxd9790_ircs_irq(irq);
	return 0;
}

static void
end_cxd9790_ircs_irq(unsigned int irq)
{
	if (!(irq_desc[irq].status & (IRQ_DISABLED | IRQ_INPROGRESS))) {
		enable_cxd9790_ircs_irq(irq);
	}
}

struct hw_interrupt_type cxd9790_ircs_irq_type = {
	"CXD9790 IRCS",
	startup_cxd9790_ircs_irq,
	shutdown_cxd9790_ircs_irq,
	enable_cxd9790_ircs_irq,
	disable_cxd9790_ircs_irq,
	mask_and_ack_cxd9790_ircs_irq,
	end_cxd9790_ircs_irq,
	NULL
};
#endif /* CXD9790_IRQ_TYPE1 */


/*
 *  intterupt type for ES1
 */
#ifdef CONFIG_SNSC_CXD9790_ES1_WORKAROUND
static void
_dummy_cxd9790_es1_ircs_irq(unsigned int irq) { /* nothing to do */ }

#define enable_cxd9790_es1_ircs_irq       _dummy_cxd9790_es1_ircs_irq
#define disable_cxd9790_es1_ircs_irq      _dummy_cxd9790_es1_ircs_irq
#define mask_and_ack_cxd9790_es1_ircs_irq _dummy_cxd9790_es1_ircs_irq

#ifdef CXD9790_IRQ_TYPE1
static unsigned int
startup_cxd9790_es1_ircs_irq(unsigned int irq) { return 0; }

#define shutdown_cxd9790_es1_ircs_irq     _dummy_cxd9790_es1_ircs_irq
#define end_cxd9790_es1_ircs_irq          _dummy_cxd9790_es1_ircs_irq

struct hw_interrupt_type cxd9790_es1_ircs_irq_type = {
	"CXD9790 IRCS",
	startup_cxd9790_es1_ircs_irq,
	shutdown_cxd9790_es1_ircs_irq,
	enable_cxd9790_es1_ircs_irq,
	disable_cxd9790_es1_ircs_irq,
	mask_and_ack_cxd9790_es1_ircs_irq,
	end_cxd9790_es1_ircs_irq,
	NULL
};
#endif /* CXD9790_IRQ_TYPE1 */
#endif /* CONFIG_SNSC_CXD9790_ES1_WORKAROUND */


#ifdef CXD9790_IRQ_TYPE1
struct irqaction cxd9790_ircs_irq_action = {
	no_action, 0, 0, "CXD9790 IRCS", NULL, NULL 
};
#endif /* CXD9790_IRQ_TYPE1 */


/*
 *  "CXD9790 EDGE" and "CXD9790 LEVEL"
 *      interrupt controller type for CXD9790 interrupts
 */

static inline void
__clear_cxd9790_edge_irq(unsigned int irq)
{
	u_int32_t     addr;

	addr = __IRQ_TO_GPIOREG(irq) + CXD9790_IREDC;
	CXD9790_WRITE(addr, (1 << __IRQ_TO_GPIOBIT(irq)));
}

static inline void
__enable_cxd9790_irq(unsigned int irq)
{
	u_int8_t      mask;
	u_int32_t     addr;

	addr = __IRQ_TO_GPIOREG(irq) + CXD9790_IRMSK;
	mask = CXD9790_READ(addr);
	mask &= ~(1 << __IRQ_TO_GPIOBIT(irq));
	CXD9790_WRITE(addr, mask);
}

static inline void
__disable_cxd9790_irq(unsigned int irq)
{
	u_int8_t      mask;
	u_int32_t     addr;

	addr = __IRQ_TO_GPIOREG(irq) + CXD9790_IRMSK;
	mask = CXD9790_READ(addr);
	mask |= (1 << __IRQ_TO_GPIOBIT(irq));
	CXD9790_WRITE(addr, mask);
}

static void
enable_cxd9790_level_irq(unsigned int irq)
{
	unsigned long flags;

	spin_lock_irqsave(&cxd9790int_lock, flags);
	__enable_cxd9790_irq(irq);
	spin_unlock_irqrestore(&cxd9790int_lock, flags);
}

static void
enable_cxd9790_edge_irq(unsigned int irq)
{
	unsigned long flags;

	spin_lock_irqsave(&cxd9790int_lock, flags);
	__clear_cxd9790_edge_irq(irq);
	__enable_cxd9790_irq(irq);
	spin_unlock_irqrestore(&cxd9790int_lock, flags);
}

static void
disable_cxd9790_irq(unsigned int irq)
{
	unsigned long flags;

	spin_lock_irqsave(&cxd9790int_lock, flags);
	__disable_cxd9790_irq(irq);
	spin_unlock_irqrestore(&cxd9790int_lock, flags);
}

static void
mask_and_ack_cxd9790_edge_irq(unsigned int irq)
{
	unsigned long flags;

	spin_lock_irqsave(&cxd9790int_lock, flags);
	__clear_cxd9790_edge_irq(irq);
	__disable_cxd9790_irq(irq);
	spin_unlock_irqrestore(&cxd9790int_lock, flags);
}

static void
mask_and_ack_cxd9790_level_irq(unsigned int irq)
{
	unsigned long flags;

	spin_lock_irqsave(&cxd9790int_lock, flags);
	__disable_cxd9790_irq(irq);
	spin_unlock_irqrestore(&cxd9790int_lock, flags);
}

#ifdef CXD9790_IRQ_TYPE1
static unsigned int
startup_cxd9790_level_irq(unsigned int irq)
{
	enable_cxd9790_level_irq(irq);
	return 0;
}

static unsigned int
startup_cxd9790_edge_irq(unsigned int irq)
{
	enable_cxd9790_edge_irq(irq);
	return 0;
}

static void
shutdown_cxd9790_irq(unsigned int irq)
{
	disable_cxd9790_irq(irq);
}

static void
end_cxd9790_level_irq(unsigned int irq)
{
	if (!(irq_desc[irq].status & (IRQ_DISABLED | IRQ_INPROGRESS))) {
		enable_cxd9790_level_irq(irq);
	}
}

static void
end_cxd9790_edge_irq(unsigned int irq)
{
	if (!(irq_desc[irq].status & (IRQ_DISABLED | IRQ_INPROGRESS))) {
		enable_cxd9790_edge_irq(irq);
	}
}

struct hw_interrupt_type cxd9790_edge_irq_type = {
	"CXD9790 Edge",
	startup_cxd9790_edge_irq,
	shutdown_cxd9790_irq,
	enable_cxd9790_edge_irq,
	disable_cxd9790_irq,
	mask_and_ack_cxd9790_edge_irq,
	end_cxd9790_edge_irq,
	NULL
};

struct hw_interrupt_type cxd9790_level_irq_type = {
	"CXD9790 Level",
	startup_cxd9790_level_irq,
	shutdown_cxd9790_irq,
	enable_cxd9790_level_irq,
	disable_cxd9790_irq,
	mask_and_ack_cxd9790_level_irq,
	end_cxd9790_level_irq,
	NULL
};

struct irqaction cxd9790_gpioz_irq_action = {
	no_action, 0, 0, "CXD9790 GPIOZ", NULL, NULL 
};

struct irqaction cxd9790_gpioab_irq_action = {
	no_action, 0, 0, "CXD9790 GPIOA/B", NULL, NULL 
};

struct irqaction cxd9790_gpiocd_irq_action = {
	no_action, 0, 0, "CXD9790 GPIOC/D", NULL, NULL 
};
#endif /* CXD9790_IRQ_TYPE1 */


/* get/set(modify) IRDM */
static int
__get_irdm(int irq)
{
	u_int32_t addr, bit;
	u_int8_t  mode;

	addr = __IRQ_TO_GPIOREG(irq) + CXD9790_IRDM0;
	bit  = __IRQ_TO_GPIOBIT(irq) << 1;
	if (bit >= 8) {
		/* IRDM1 is used */
		addr++; 
		bit -= 8;
	}

	mode = CXD9790_READ(addr);

	return (mode >> bit) & 3;
}

static void
__set_irdm(int irq, int trig)
{
	u_int32_t addr, bit;
	u_int8_t  mode;

	addr = __IRQ_TO_GPIOREG(irq) + CXD9790_IRDM0;
	bit  = __IRQ_TO_GPIOBIT(irq) << 1;
	if (bit >= 8) {
		/* IRDM1 is used */
		addr++; 
		bit -= 8;
	}

	mode = CXD9790_READ(addr);
	mode &= ~(0x3 << bit);
	mode |= ((trig & 0x3) << bit);
	CXD9790_WRITE(addr, mode);
}

/* set irq_desc[].handler or irq_desc[].mask_ack */
static void
__set_irq_desc_handler(int irq, int trig)
{
#ifdef CXD9790_IRQ_TYPE1
	if (__IS_EDGE_TRIG(trig))
		irq_desc[irq].handler = 
			&cxd9790_edge_irq_type;
	else
		irq_desc[irq].handler = 
			&cxd9790_level_irq_type;
#else
	if (__IS_EDGE_TRIG(trig))
		irq_desc[irq].mask_ack =
			mask_and_ack_cxd9790_edge_irq;
	else
		irq_desc[irq].mask_ack =
			mask_and_ack_cxd9790_level_irq;
#endif /* CXD9790_IRQ_TYPE1 */	
}

/* mask and clear all interrupts */
static int
__mask_and_clearall(void)
{
	int i;

	for (i = CXD9790_GPIOZ; i <= CXD9790_GPIOD; i += 0x10) {
		CXD9790_WRITE(i + CXD9790_IRMSK, 0xff);
		CXD9790_WRITE(i + CXD9790_IREDC, 0xff);
	}

	return 0;
}


/*
 *  initialize CXD9790 interrupts
 */
int
cxd9790_irq_init(void)
{
	int i;

	if (!cxd9790_present())
		return -ENODEV;

	__mask_and_clearall();

	for (i = CXD9790_IRQ_OFFSET;
	     i < CXD9790_IRQ_OFFSET + CXD9790_NR_IRQS_IRCS;
	     i++) {
#ifdef CXD9790_IRQ_TYPE1
		irq_desc[i].status  = IRQ_DISABLED;
		irq_desc[i].action  = 0;
		irq_desc[i].depth   = 1;
#ifdef CONFIG_SNSC_CXD9790_ES1_WORKAROUND
		if (cxd9790_is_es1())
			irq_desc[i].handler = &cxd9790_es1_ircs_irq_type;
		else
#endif /* CONFIG_SNSC_CXD9790_ES1_WORKAROUND */
		irq_desc[i].handler = &cxd9790_ircs_irq_type;

#else /* !CXD9790_IRQ_TYPE1 */
		irq_desc[i].valid    = 1;
		irq_desc[i].probe_ok = 1;
#ifdef CONFIG_SNSC_CXD9790_ES1_WORKAROUND
		if (cxd9790_is_es1()) {
			irq_desc[i].mask     = disable_cxd9790_es1_ircs_irq;
			irq_desc[i].unmask   = enable_cxd9790_es1_ircs_irq;
			irq_desc[i].mask_ack = mask_and_ack_cxd9790_es1_ircs_irq;
		} else {
#endif /* CONFIG_SNSC_CXD9790_ES1_WORKAROUND */
		irq_desc[i].mask     = disable_cxd9790_ircs_irq;
		irq_desc[i].unmask   = enable_cxd9790_ircs_irq;
		irq_desc[i].mask_ack = mask_and_ack_cxd9790_ircs_irq;
#ifdef CONFIG_SNSC_CXD9790_ES1_WORKAROUND
		}
#endif /* CONFIG_SNSC_CXD9790_ES1_WORKAROUND */
#endif /* CXD9790_IRQ_TYPE1 */
	}

	for (; i < CXD9790_IRQ_OFFSET + CXD9790_NR_IRQS; i++) {
#ifdef CXD9790_IRQ_TYPE1
		irq_desc[i].status  = IRQ_DISABLED;
		irq_desc[i].action  = 0;
		irq_desc[i].depth   = 1;
#else
		irq_desc[i].valid    = 1;
		irq_desc[i].probe_ok = 1;
		irq_desc[i].mask     = disable_cxd9790_irq;
		irq_desc[i].unmask   = enable_cxd9790_irq;
#endif /* CXD9790_IRQ_TYPE1 */
		__set_irq_desc_handler(i, __get_irdm(i));
	}

#ifdef CXD9790_IRQ_TYPE1
	setup_irq(CXD9790_IRQ_GPZ, &cxd9790_gpioz_irq_action);
	setup_irq(CXD9790_IRQ_GPAB, &cxd9790_gpioab_irq_action);
	setup_irq(CXD9790_IRQ_GPCD, &cxd9790_gpiocd_irq_action);
#endif /* CXD9790_IRQ_TYPE1 */

	return 0;
}


/*
 *  set interrupt trigger mode
 */
int
cxd9790_set_trigger_mode(int irq, int trig)
{
	unsigned long flags;
	int ret =0;

	if (__IS_GPIO(irq)) {
		spin_lock_irqsave(&cxd9790int_lock, flags);

		if ((CXD9790_READ(__IRQ_TO_GPIOREG(irq) + CXD9790_IRMSK)
		     & (1 << __IRQ_TO_GPIOBIT(irq))) == 0) {
			printk("CXD9790: cannot change trigger mode for "
			       "irq %d because it is enabled\n", irq);
			ret = -EBUSY;
		} else {
			__set_irdm(irq, trig);
			__set_irq_desc_handler(irq, trig);
		}
		spin_unlock_irqrestore(&cxd9790int_lock, flags);
	} else {
		ret = -EINVAL;
	}

	return ret;
}


/*
 *  select IROUT
 */
int
cxd9790_set_irout(int irq, int irout)
{
	int      bit;
	u_int8_t iro;
	unsigned long flags;

	if (__IS_IRCS(irq)) {
		bit = __IRQ_TO_LOCALIRQ(irq);
		
		spin_lock_irqsave(&cxd9790int_lock, flags);
		iro = CXD9790_READ(CXD9790_IROUT);
		iro &= ~(1 << bit);
		iro |= (irout & 1) << bit;
		CXD9790_WRITE(CXD9790_IROUT, iro);
		spin_unlock_irqrestore(&cxd9790int_lock, flags);
	} else
		return -EINVAL;

	return 0;
}


static inline int
sts2num(u_int8_t sts)
{
	int i;

	for (i = 0; i < 8; i++) {
		if (sts & (1 << i))
			return i;
	}

	return -1;
}

/*
 *  function for getting IRQ number
 */
int
cxd9790_get_irq(int irout)
{
	u_int8_t sts;
	int      irq;

	if (irout == IROUT0)
		sts = CXD9790_READ(CXD9790_IRCS0);
	else if (irout == IROUT1)
		sts = CXD9790_READ(CXD9790_IRCS1);
	else {
		printk("CXD9790: irregal irout %d\n", irout);
		return -EINVAL;
	}
	DPRINTK("IRCS = 0x%x\n", sts);

	if ((irq = sts2num(sts)) < 0)
		goto spurious;

	DPRINTK("IRCS local irq =%d\n", irq);

	if (irq < 5) {
		irq = __LOCALIRQ_TO_IRQ(irq);
	} else if (irq == 7) {
		/* GPIO Z */
		sts = CXD9790_READ(CXD9790_GPIOZ + CXD9790_IRSTS);
		DPRINTK("IRSTS_Z = 0x%x\n", sts);
		if ((irq = sts2num(sts)) < 0)
			goto spurious;
		irq = CXD9790_IRQ_GPIOZ(irq);
	} else if (irq == 6) {
		/* GPIO A */
		sts = CXD9790_READ(CXD9790_GPIOA + CXD9790_IRSTS);
		DPRINTK("IRSTS_A = 0x%x\n", sts);
		if ((irq = sts2num(sts)) >= 0)
			irq = CXD9790_IRQ_GPIOA(irq);
		else {
			/* GPIO B */
			sts = CXD9790_READ(CXD9790_GPIOB + CXD9790_IRSTS);
			DPRINTK("IRSTS_B = 0x%x\n", sts);
			if ((irq = sts2num(sts)) < 0)
				goto spurious;
			irq = CXD9790_IRQ_GPIOB(irq);
		}
	} else if (irq == 5) {
		/* GPIO C */
		sts = CXD9790_READ(CXD9790_GPIOC + CXD9790_IRSTS);
		DPRINTK("IRSTS_C = 0x%x\n", sts);
		if ((irq = sts2num(sts)) >= 0)
			irq = CXD9790_IRQ_GPIOC(irq);
		else {
			/* GPIO D */
			sts = CXD9790_READ(CXD9790_GPIOD + CXD9790_IRSTS);
			DPRINTK("IRSTS_D = 0x%x\n", sts);
			if ((irq = sts2num(sts)) < 0)
				goto spurious;
			irq = CXD9790_IRQ_GPIOD(irq);
		}
	}

	DPRINTK("irq = %d\n", irq);
	return irq;

spurious:
	printk("CXD9790: spurious interupt\n");
	return -EINVAL;
}

EXPORT_SYMBOL(cxd9790_set_trigger_mode);
EXPORT_SYMBOL(cxd9790_set_irout);
