/*
 *  cxd9802_spi.c -- CXD9802 SPI driver
 *
 */
/*
 *  Copyright 2004 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/module.h>
#include <linux/version.h>

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/poll.h>
#include <linux/wrapper.h>
#include <linux/sched.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>
#include <asm/softirq.h>
#include <linux/init.h>
#include <linux/config.h>
#include <linux/list.h>
#include <linux/interrupt.h>

#include <linux/snsc_spi.h>

#include <asm/cxd9802/cxd9802.h>

#define DEBUG
#define DEBUG_LEVEL DPL_NOTICE

#ifdef DEBUG
#define DPR(fmt , args...)	printk(fmt, ## args)
#endif /* DEBUG */

MODULE_AUTHOR("Sony CESP");
MODULE_DESCRIPTION("CXD9802 SPI driver");
MODULE_LICENSE("GPL");

/* ======================================================================= *
 * Constant                                                                *
 * ======================================================================= */

#define CXD9802_SPI_DEVNAME		"CXD9802 SPI"
#define CXD9802_SPI_DEFAULT_FREQ	(cxd9802_imbusclk_freq / 4) /* max */
#define CXD9802_SPI_DEFAULT_IFDELAY	0
#define CXD9802_SPI_DEFAULT_SIZE	8
#define CXD9802_SPI_DEFAULT_MODE	0

/* ======================================================================= *
 * Debug print function from dpr package                                   *
 * ======================================================================= */

#define DPL_EMERG	0
#define DPL_ALERT	1
#define DPL_CRIT	2
#define DPL_ERR		3
#define DPL_WARNING	4
#define DPL_NOTICE	5
#define DPL_INFO	6
#define DPL_DEBUG	7

#define DPM_PURPOSE	0x0000000f
#define DPM_ERR		0x00000001
#define DPM_TRACE	0x00000002
#define DPM_LAYER	0x000000f0
#define DPM_APPLICATION	0x00000010
#define DPM_MIDDLEWARE	0x00000020
#define DPM_DRIVER	0x00000040
#define DPM_KERNEL	0x00000080
#define DPM_USERDEF	0xffff0000

#ifdef DEBUG

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL DPL_DEBUG
#endif
#ifndef DEBUG_MASK
#define DEBUG_MASK  DPM_PURPOSE
#endif

#ifndef DPR
#include <stdio.h>
#define DPR(fmt, args...)	fprintf(stderr, fmt , ## args)
#endif

#define DPRL(level, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && DPR(fmt , ## args))
#define DPRM(mask, fmt, args...) \
	(void)(((mask) & DEBUG_MASK) && DPR(fmt , ## args))
#define DPRLM(level, mask, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && ((mask) & DEBUG_MASK) && \
		DPR(fmt , ## args))
#define DPRC(cond, fmt, args...) \
	(void)((cond) && DPR(fmt , ## args))
#define DPRF(fmt, args...)	DPR("%s: " fmt, __FUNCTION__ , ## args)
#define DPRFL(level, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && DPRF(fmt , ## args))
#define DPRFM(mask, fmt, args...) \
	(void)(((mask) & DEBUG_MASK) && DPRF(fmt , ## args))
#define DPRFLM(level, mask, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && ((mask) & DEBUG_MASK) && \
		DPRF(fmt , ## args))
#define DPRFC(cond, fmt, args...) \
	(void)((cond) && DPRF(fmt , ## args))

#define	DPRFS	DPRFLM(DPL_DEBUG, DPM_TRACE, "start\n")
#define	DPRFE	DPRFLM(DPL_DEBUG, DPM_TRACE, "end\n")
#define DPRFR	DPRFLM(DPL_DEBUG, DPM_TRACE, "return\n")

#else  /* DEBUG */

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL -1
#endif
#ifndef DEBUG_MASK
#define DEBUG_MASK  0
#endif

#ifndef DPR
#define DPR(fmt, args...)
#endif

#define DPRL(level, fmt, args...)
#define DPRM(mask, fmt, args...)
#define DPRLM(level, mask, fmt, args...)
#define DPRC(cond, fmt, args...)
#define DPRF(fmt, args...)
#define DPRFL(level, fmt, args...)
#define DPRFM(mask, fmt, args...)
#define DPRFLM(level, mask, fmt, args...)
#define DPRFC(cond, fmt, args...)

#define	DPRFS
#define	DPRFE
#define DPRFR

#endif /* DEBUG */

/* ======================================================================= *
 * SPI register                                                            *
 * ======================================================================= */

#define REG_SPMCR		0x00
#define  REGM_OPMODE		(0x3 << 6)
#define  REGC_OPMODE_CONFIG	(1 << 6)
#define  REGC_OPMODE_ACTIVE	(2 << 6)
#define  REGM_SPICSENB		(1 << 4)
#define  REGM_SPICSOUT		(1 << 3)
#define  REGM_SPSTP		(1 << 1)
#define  REGM_BCLR		(1 << 0)
#define REG_SPCR0		0x04
#define  REGM_TXIFL		(0x3 << 14)
#define  REGS_TXIFL(x)		((x) << 14)
#define  REGM_RXIFL		(0x3 << 12)
#define  REGS_RXIFL(x)		((x) << 12)
#define  REGM_SILIE		(1 << 11)
#define  REGM_SOEIE		(1 << 10)
#define  REGM_RBSIE		(1 << 9)
#define  REGM_TBSIE		(1 << 8)
#define  REGM_IFSPSE		(1 << 4)
#define  REGM_SBOS		(1 << 2)
#define  REGM_SPHA		(1 << 1)
#define  REGM_SPOL		(1 << 0)
#define REG_SPCR1		0x08
#define  REGM_SER		(0xff << 8)
#define  REGS_SER(x)		((x) << 8)
#define  REGG_SER(r)		(((r) >> 8) & 0xff)
#define  REGM_SSZ		(0x1f << 0)
#define  REGG_SSZ(r)		(((r) >> 0) & 0x1f)
#define  REGC_SSZ_8BITS		(0x08 << 0)
#define  REGC_SSZ_16BITS	(0x10 << 0)
#define REG_SPFS		0x0c
#define  REGM_IFS		(0x3ff << 0)
#define REG_SPSR		0x14
#define  REGM_TBSI		(1 << 15)
#define  REGM_RBSI		(1 << 14)
#define  REGM_TBS		(0x7 << 11)
#define  REGG_TBS(r)		(((r) >> 11) & 0x7)
#define  REGM_RBS		(0x7 << 8)
#define  REGG_RBS(r)		(((r) >> 8) & 0x7)
#define  REGM_SPOE		(1 << 7)
#define  REGM_IFSD		(1 << 3)
#define  REGM_SIDLE		(1 << 2)
#define  REGM_STRDY		(1 << 1)
#define  REGM_SRRDY		(1 << 0)
#define REG_SPDR		0x18

/* ======================================================================= *
 * type and structure                                                      *
 * ======================================================================= */

/* ======================================================================= *
 * prototype declaration                                                   *
 * ======================================================================= */

/* SPI driver interface */
static int cxd9802_spi_open(snscspi_bus_t *bus);
static int cxd9802_spi_release(snscspi_bus_t *bus);
static int cxd9802_spi_ioctl(snscspi_bus_t *bus,
			     unsigned int cmd, void *arg, int argtype);

static int cxd9802_spi_init(void);
static void cxd9802_spi_final(void);

/* interrupt handler */
static void cxd9802_spi_intr_handler(int irq, void *devid,
				     struct pt_regs *regs);
/* ioctl internal */
static int cxd9802_spi_ioctl_transaction(snscspi_bus_t *bus,
					 void *arg, int argtype);
static int cxd9802_spi_ioctl_set_freq(snscspi_bus_t *bus,
				      void *arg, int argtype);
static int cxd9802_spi_ioctl_get_freq(snscspi_bus_t *bus,
				      void *arg, int argtype);
static int cxd9802_spi_ioctl_set_ifdelay(snscspi_bus_t *bus,
					 void *arg, int argtype);
static int cxd9802_spi_ioctl_get_ifdelay(snscspi_bus_t *bus,
					 void *arg, int argtype);
static int cxd9802_spi_ioctl_set_size(snscspi_bus_t *bus,
				      void *arg, int argtype);
static int cxd9802_spi_ioctl_get_size(snscspi_bus_t *bus,
				      void *arg, int argtype);
static int cxd9802_spi_ioctl_set_mode(snscspi_bus_t *bus,
				      void *arg, int argtype);
static int cxd9802_spi_ioctl_get_mode(snscspi_bus_t *bus,
				      void *arg, int argtype);

/* internal function */
static int cxd9802_spi_do_msg(snscspi_msg_t *msg, int buf_is_kernel_addr);
static void cxd9802_spi_tx_bytes(void);
static void cxd9802_spi_rx_bytes(void);
static int cxd9802_spi_chip_select(int cs);
static void cxd9802_spi_set_freq(u_int32_t freq);
static u_int32_t cxd9802_spi_get_freq(void);
static void cxd9802_spi_set_ifdelay(u_int32_t ifdelay);
static u_int32_t cxd9802_spi_get_ifdelay(void);
static int cxd9802_spi_set_size(u_int32_t nbits);
static u_int32_t cxd9802_spi_get_size(void);
static void cxd9802_spi_set_mode(int mode);
static int cxd9802_spi_get_mode(void);

/* ======================================================================= *
 * global variable                                                         *
 * ======================================================================= */

static spinlock_t cxd9802_spi_irqlock = SPIN_LOCK_UNLOCKED;
static wait_queue_head_t cxd9802_spi_txrx_wait;

static int cxd9802_spi_in_config_mode = 0;

#define CXD9802_SPI_TXRX_TX	(1 << 0)
#define CXD9802_SPI_TXRX_RX	(1 << 1)
static int	 cxd9802_spi_txrx_task = 0;
static int	 cxd9802_spi_txrx_error = 0;

static char	*cxd9802_spi_txrx_buf = NULL;
static size_t	 cxd9802_spi_txrx_len = 0;
static u_int32_t cxd9802_spi_txrx_txpos = 0;
static u_int32_t cxd9802_spi_txrx_rxpos = 0;
static u_int32_t cxd9802_spi_txrx_size;

/* ======================================================================= *
 * hardware interface                                                      *
 * ======================================================================= */

static inline void
cxd9802_spi_reg_write(int offset, u32 value)
{
	DPRFL(DPL_DEBUG, "[%02x]=%08x\n", offset, value);
	cxd9802_reg_write32(CXD9802_SPIC_BASE + offset, value);
}

static inline u32
cxd9802_spi_reg_read(int offset)
{
	u32 value;

	value = cxd9802_reg_read32(CXD9802_SPIC_BASE + offset);
	DPRFL(DPL_DEBUG, "[%02x]=%08x\n", offset, value);
	return value;
}

static inline void
cxd9802_spi_reg_changebits(int offset, u32 bits, u32 value)
{
	cxd9802_spi_reg_write(offset,
			      (cxd9802_spi_reg_read(offset) & ~bits) | value);
}

static inline void
cxd9802_spi_reg_setbits(int offset, u32 bits)
{
	cxd9802_spi_reg_write(offset, cxd9802_spi_reg_read(offset) | bits);
}

static inline void
cxd9802_spi_reg_clearbits(int offset, u32 bits)
{
	cxd9802_spi_reg_changebits(offset, bits, 0);
}

/* ======================================================================= *
 * inline function                                                         *
 * ======================================================================= */

static inline void
cxd9802_spi_enter_config_mode(void)
{
	DPRFS;

	if (!cxd9802_spi_in_config_mode) {
		cxd9802_spi_reg_changebits(REG_SPMCR, REGM_OPMODE,
					   REGC_OPMODE_CONFIG);
		cxd9802_spi_in_config_mode = 1;
	}

	DPRFE;
	return;
}

static inline void
cxd9802_spi_enter_active_mode(void)
{
	DPRFS;

	cxd9802_spi_reg_changebits(REG_SPMCR, REGM_OPMODE,
				   REGC_OPMODE_ACTIVE);
	cxd9802_spi_in_config_mode = 0;

	DPRFE;
	return;
}

/* ======================================================================= *
 * SPI driver interface                                                    *
 * ======================================================================= */

static snscspi_bus_t cxd9802_spi_bus = {
	name:		CXD9802_SPI_DEVNAME,
	open:		cxd9802_spi_open,
	release:	cxd9802_spi_release,
	ioctl:		cxd9802_spi_ioctl,
};

static int
cxd9802_spi_open(snscspi_bus_t *bus)
{
	DPRFS;

	MOD_INC_USE_COUNT;

	DPRFE;
	return 0;
}

static int
cxd9802_spi_release(snscspi_bus_t *bus)
{
	DPRFS;

	MOD_DEC_USE_COUNT;

	DPRFE;
	return 0;
}

static int
cxd9802_spi_ioctl(snscspi_bus_t *bus,
		  unsigned int cmd, void *arg, int argtype)
{
	int res;
	DPRFS;

	switch (cmd) {
	case SNSCSPI_IOC_TRANSACTION:
		res = cxd9802_spi_ioctl_transaction(bus, arg, argtype);
		break;
	case SNSCSPI_IOC_SET_FREQ:
		res = cxd9802_spi_ioctl_set_freq(bus, arg, argtype);
		break;
	case SNSCSPI_IOC_GET_FREQ:
		res = cxd9802_spi_ioctl_get_freq(bus, arg, argtype);
		break;
	case SNSCSPI_IOC_SET_IFDELAY:
		res = cxd9802_spi_ioctl_set_ifdelay(bus, arg, argtype);
		break;
	case SNSCSPI_IOC_GET_IFDELAY:
		res = cxd9802_spi_ioctl_get_ifdelay(bus, arg, argtype);
		break;
	case SNSCSPI_IOC_SET_SIZE:
		res = cxd9802_spi_ioctl_set_size(bus, arg, argtype);
		break;
	case SNSCSPI_IOC_GET_SIZE:
		res = cxd9802_spi_ioctl_get_size(bus, arg, argtype);
		break;
	case SNSCSPI_IOC_SET_MODE:
		res = cxd9802_spi_ioctl_set_mode(bus, arg, argtype);
		break;
	case SNSCSPI_IOC_GET_MODE:
		res = cxd9802_spi_ioctl_get_mode(bus, arg, argtype);
		break;
	default:
		res = -EINVAL;
	}

	DPRFE;
	return res;
}

int __init
cxd9802_spi_init(void)
{
	int res;
	static int initialized = 0;
	DPRFS;

	if (initialized) {
		DPRFR;
		return 0;
	}

	init_waitqueue_head(&cxd9802_spi_txrx_wait);

	/* reset */
	cxd9802_spi_enter_config_mode();

	/* set default parameters */
	cxd9802_spi_set_freq(CXD9802_SPI_DEFAULT_FREQ);
	cxd9802_spi_set_size(CXD9802_SPI_DEFAULT_SIZE);
	cxd9802_spi_set_mode(CXD9802_SPI_DEFAULT_MODE);
	cxd9802_spi_set_ifdelay(CXD9802_SPI_DEFAULT_IFDELAY);

	/* disable interrupts */
	cxd9802_spi_reg_clearbits(REG_SPCR0,
				  REGM_SILIE | REGM_SOEIE |
				  REGM_RBSIE | REGM_TBSIE);

	/* request irq */
	res = request_irq(CXD9802_IRQ_SPI, cxd9802_spi_intr_handler,
			  SA_SHIRQ | SA_INTERRUPT, CXD9802_SPI_DEVNAME,
			  &cxd9802_spi_bus);
	if (res < 0) {
		goto err1;
	}

	/* register this driver */
	res = snscspi_register(&cxd9802_spi_bus);
	if (res < 0) {
		goto err2;
	}

	initialized = 1;

	DPRFE;
	return 0;

err2:
	free_irq(CXD9802_IRQ_SPI, &cxd9802_spi_bus);
err1:
	DPRFR;
	return res;
}

static void
cxd9802_spi_final(void)
{
	DPRFS;

	snscspi_unregister(&cxd9802_spi_bus);

	free_irq(CXD9802_IRQ_SPI, &cxd9802_spi_bus);

	/* disable interrupts */
	cxd9802_spi_enter_config_mode();
	cxd9802_spi_reg_clearbits(REG_SPCR0,
				  REGM_SILIE | REGM_SOEIE |
				  REGM_RBSIE | REGM_TBSIE);

	DPRFE;
	return;
}

module_init(cxd9802_spi_init);
module_exit(cxd9802_spi_final);

/* ======================================================================= *
 * Interrupt handler                                                       *
 * ======================================================================= */

static void
cxd9802_spi_intr_handler(int irq, void *devid, struct pt_regs *regs)
{
	u32 spsr;
	DPRFS;

	spsr = cxd9802_spi_reg_read(REG_SPSR);
	DPRFL(DPL_DEBUG, "SPSR=%08x\n", spsr);
	if (spsr & REGM_SRRDY) {
		cxd9802_spi_rx_bytes();
	}
	if ((spsr & REGM_STRDY) && cxd9802_spi_txrx_task != 0) {
		cxd9802_spi_tx_bytes();
	}

	DPRFE;
	return;
}

/* ======================================================================= *
 * ioctl internal                                                          *
 * ======================================================================= */

static int
cxd9802_spi_ioctl_transaction(snscspi_bus_t *bus, void *arg, int argtype)
{
	int res;
	snscspi_transaction_t *trans = (snscspi_transaction_t *)arg;
	u_int32_t msg_i;
	snscspi_msg_t *msg;
	snscspi_msg_t tmp_msg;
	DPRFS;

	for (msg_i = 0; msg_i < trans->nmsgs; msg_i++) {
		if (argtype & SNSCSPI_T_ARG_KERNEL) {
			msg = &(trans->msgs[msg_i]);
		} else {
			res = copy_from_user(&tmp_msg, &(trans->msgs[msg_i]),
					     sizeof(tmp_msg));
			if (res != 0) {
				DPRFR;
				return -EFAULT;
			}
			msg = &tmp_msg;
		}

		res = cxd9802_spi_do_msg(msg,
					 (argtype & SNSCSPI_T_ARG_KERNEL) &&
					 !(argtype & SNSCSPI_T_BUF_USER));
		if (res < 0) {
			DPRFR;
			return res;
		}
	}

	DPRFE;
	return 0;
}

static int
cxd9802_spi_ioctl_set_freq(snscspi_bus_t *bus, void *arg, int argtype)
{
	u_int32_t freq = (u_int32_t)arg;
	DPRFS;

	cxd9802_spi_set_freq(freq);

	DPRFE;
	return 0;
}

static int
cxd9802_spi_ioctl_get_freq(snscspi_bus_t *bus, void *arg, int argtype)
{
	int res;
	u_int32_t freq;
	DPRFS;

	freq = cxd9802_spi_get_freq();

	if (argtype & SNSCSPI_T_ARG_KERNEL) {
		*((u_int32_t *)arg) = freq;
	} else {
		res = copy_to_user(arg, &freq, sizeof(freq));
		if (res != 0) {
			DPRFR;
			return -EFAULT;
		}
	}

	DPRFE;
	return 0;
}

static int
cxd9802_spi_ioctl_set_ifdelay(snscspi_bus_t *bus, void *arg, int argtype)
{
	u_int32_t ifdelay = (u_int32_t)arg;
	DPRFS;

	cxd9802_spi_set_ifdelay(ifdelay);

	DPRFE;
	return 0;
}

static int
cxd9802_spi_ioctl_get_ifdelay(snscspi_bus_t *bus, void *arg, int argtype)
{
	int res;
	u_int32_t ifdelay;
	DPRFS;

	ifdelay = cxd9802_spi_get_ifdelay();

	if (argtype & SNSCSPI_T_ARG_KERNEL) {
		*((u_int32_t *)arg) = ifdelay;
	} else {
		res = copy_to_user(arg, &ifdelay, sizeof(ifdelay));
		if (res != 0) {
			DPRFR;
			return -EFAULT;
		}
	}

	DPRFE;
	return 0;
}

static int
cxd9802_spi_ioctl_set_size(snscspi_bus_t *bus, void *arg, int argtype)
{
	int res;
	u_int32_t size = (u_int32_t)arg;
	DPRFS;

	res = cxd9802_spi_set_size(size);

	DPRFE;
	return res;
}

static int
cxd9802_spi_ioctl_get_size(snscspi_bus_t *bus, void *arg, int argtype)
{
	int res;
	u_int32_t size;
	DPRFS;

	size = cxd9802_spi_get_size();

	if (argtype & SNSCSPI_T_ARG_KERNEL) {
		*((u_int32_t *)arg) = size;
	} else {
		res = copy_to_user(arg, &size, sizeof(size));
		if (res != 0) {
			DPRFR;
			return -EFAULT;
		}
	}

	DPRFE;
	return 0;
}

static int
cxd9802_spi_ioctl_set_mode(snscspi_bus_t *bus, void *arg, int argtype)
{
	int mode = (int)arg;
	DPRFS;

	cxd9802_spi_set_mode(mode);

	DPRFE;
	return 0;
}

static int
cxd9802_spi_ioctl_get_mode(snscspi_bus_t *bus, void *arg, int argtype)
{
	int res;
	int mode;

	mode = cxd9802_spi_get_mode();

	if (argtype & SNSCSPI_T_ARG_KERNEL) {
		*((u_int32_t *)arg) = mode;
	} else {
		res = copy_to_user(arg, &mode, sizeof(mode));
		if (res != 0) {
			DPRFR;
			return -EFAULT;
		}
	}

	DPRFE;
	return 0;
}

/* ======================================================================= *
 * Internal function                                                       *
 * ======================================================================= */

static int
cxd9802_spi_do_msg(snscspi_msg_t *msg, int buf_is_kernel_addr)
{
	static char tmp_buf[256];
	int res;
	u_int32_t msgbuf_pos;
	DPRFS;

	if (cxd9802_spi_get_size() == 16 && msg->len % 2 != 0) {
		DPRFR;
		return -EINVAL;
	}

	if (msg->flags & SNSCSPI_M_CS) {
		res = cxd9802_spi_chip_select(msg->cs);
		if (res < 0) {
			DPRFR;
			return res;
		}
	}
	if (msg->len == 0) {
		DPRFE;
		return 0;
	}

	cxd9802_spi_enter_config_mode();
	cxd9802_spi_enter_active_mode();

	cxd9802_spi_txrx_size = cxd9802_spi_get_size();
	if (cxd9802_spi_txrx_size != 8 && cxd9802_spi_txrx_size != 16) {
		DPRFL(DPL_WARNING, "something wrong size(%d)\n",
		      cxd9802_spi_txrx_size);
	}

	msgbuf_pos = 0;
	while (msgbuf_pos < msg->len) {
		if (buf_is_kernel_addr) {
			cxd9802_spi_txrx_len = msg->len;
			cxd9802_spi_txrx_buf = msg->buf;
		} else {
			cxd9802_spi_txrx_len = msg->len - msgbuf_pos;
			if (cxd9802_spi_txrx_len > sizeof(tmp_buf)) {
				cxd9802_spi_txrx_len = sizeof(tmp_buf);
			}
			if (msg->flags & SNSCSPI_M_WR) {
				res = copy_from_user(tmp_buf,
						     (char *)msg->buf +
						     msgbuf_pos,
						     cxd9802_spi_txrx_len);
				if (res != 0) {
					DPRFR;
					return -EFAULT;
				}
			}
			cxd9802_spi_txrx_buf = tmp_buf;
		}
		cxd9802_spi_txrx_txpos = 0;
		cxd9802_spi_txrx_rxpos = 0;
		cxd9802_spi_txrx_task = 0;
		if (msg->flags & SNSCSPI_M_WR) {
			cxd9802_spi_txrx_task |= CXD9802_SPI_TXRX_TX;
		}
		if (msg->flags & SNSCSPI_M_RD) {
			cxd9802_spi_txrx_task |= CXD9802_SPI_TXRX_RX;
		}
		if (cxd9802_spi_txrx_task == 0) {
			DPRFR;
			return -EINVAL;
		}

		cxd9802_spi_tx_bytes();

		cxd9802_spi_reg_changebits(REG_SPCR0,
					   REGM_TXIFL | REGM_RXIFL |
					   REGM_SILIE | REGM_SOEIE |
					   REGM_RBSIE | REGM_TBSIE,
					   REGS_TXIFL(2) | REGS_RXIFL(2) |
					   REGM_SILIE |
					   REGM_RBSIE);

		res = wait_event_interruptible(cxd9802_spi_txrx_wait,
					       cxd9802_spi_txrx_task == 0);
		cxd9802_spi_txrx_task = 0;
		if (res < 0) {
			cxd9802_spi_enter_config_mode();
			DPRFR;
			return res;
		}

		if (!buf_is_kernel_addr && (msg->flags & SNSCSPI_M_RD)) {
			res = copy_to_user((char *)msg->buf + msgbuf_pos,
					   tmp_buf, cxd9802_spi_txrx_len);
			if (res != 0) {
				DPRFR;
				return -EFAULT;
			}
		}

		msgbuf_pos += cxd9802_spi_txrx_len;
	}

	DPRFE;
	return 0;
}

static void
cxd9802_spi_tx_bytes(void)
{
	int tbs;
	int ndata_free_in_txbuf;
	int i;
	void *buf_pos;
	u_int32_t tx_data;
	DPRFS;

	tbs = REGG_TBS(cxd9802_spi_reg_read(REG_SPSR));
	if (tbs > 4) {
		DPRFL(DPL_WARNING, "something wrong TBS(%d)\n", tbs);
		tbs = 4;
	}
	ndata_free_in_txbuf = 4 - tbs;

	for (i = 0; i < ndata_free_in_txbuf; i++) {
		if (cxd9802_spi_txrx_txpos == cxd9802_spi_txrx_len) {
			break;
		}
		tx_data = 0;
		if (cxd9802_spi_txrx_task & CXD9802_SPI_TXRX_TX) {
			buf_pos = cxd9802_spi_txrx_buf +
				cxd9802_spi_txrx_txpos;
			if (cxd9802_spi_txrx_size == 8) {
				tx_data = *((u_int8_t *)buf_pos);
			} else if (cxd9802_spi_txrx_size == 16) {
				tx_data = *((u_int16_t *)buf_pos);
			}
		}
		cxd9802_spi_reg_write(REG_SPDR, tx_data);
		if (cxd9802_spi_txrx_size == 8) {
			cxd9802_spi_txrx_txpos++;
		} else if (cxd9802_spi_txrx_size == 16) {
			cxd9802_spi_txrx_txpos += 2;
		}
	}

	DPRFE;
	return;
}

static void
cxd9802_spi_rx_bytes(void)
{
	int rbs;
	int i;
	u_int32_t rx_data;
	void *buf_pos;
	DPRFS;

	rbs = REGG_RBS(cxd9802_spi_reg_read(REG_SPSR));
	if (rbs > 4) {
		DPRFL(DPL_WARNING, "something wrong RBS(%d)\n", rbs);
		rbs = 4;
	}

	for (i = 0; i < rbs; i++) {
		rx_data = cxd9802_spi_reg_read(REG_SPDR);
		if (cxd9802_spi_txrx_task & CXD9802_SPI_TXRX_RX) {
			buf_pos = cxd9802_spi_txrx_buf +
				cxd9802_spi_txrx_rxpos;
			if (cxd9802_spi_txrx_size == 8) {
				*((u_int8_t *)buf_pos) = rx_data;
			} else if (cxd9802_spi_txrx_size == 16) {
				*((u_int16_t *)buf_pos) = rx_data;
			}
		}
		if (cxd9802_spi_txrx_size == 8) {
			cxd9802_spi_txrx_rxpos++;
		} else if (cxd9802_spi_txrx_size == 16) {
			cxd9802_spi_txrx_rxpos += 2;
		}
		if (cxd9802_spi_txrx_rxpos == cxd9802_spi_txrx_len) {
			DPRFL(DPL_DEBUG, "task finished\n");
			cxd9802_spi_reg_clearbits(REG_SPCR0, REGM_SILIE);
			cxd9802_spi_txrx_task = 0;
			cxd9802_spi_txrx_error = 0;
			wake_up_interruptible(&cxd9802_spi_txrx_wait);
		}
	}

	DPRFE;
	return;
}

static int
cxd9802_spi_chip_select(int cs)
{
	unsigned long flags;
	DPRFS;

	switch (cs) {
	case 0:
		spin_lock_irqsave(&cxd9802_spi_irqlock, flags);
		cxd9802_spi_reg_changebits(REG_SPMCR,
					   REGM_SPICSENB | REGM_SPICSOUT,
					   0);
		spin_unlock_irqrestore(&cxd9802_spi_irqlock, flags);
		break;
	case 1:
		spin_lock_irqsave(&cxd9802_spi_irqlock, flags);
		cxd9802_spi_reg_changebits(REG_SPMCR,
					   REGM_SPICSENB | REGM_SPICSOUT,
					   REGM_SPICSOUT);
		spin_unlock_irqrestore(&cxd9802_spi_irqlock, flags);
		break;
	case -1:
		spin_lock_irqsave(&cxd9802_spi_irqlock, flags);
		cxd9802_spi_reg_setbits(REG_SPMCR, REGM_SPICSENB);
		spin_unlock_irqrestore(&cxd9802_spi_irqlock, flags);
		break;
	default:
		DPRFR;
		return -EINVAL;
	}

	DPRFE;
	return 0;
}

static void
cxd9802_spi_set_freq(u_int32_t freq)
{
	int ser;
	unsigned long flags;
	DPRFS;

	ser = (cxd9802_imbusclk_freq / 2 / freq) - 1;
	if (ser < 1) {
		ser = 1;
	} else if (ser > 0xff) {
		ser = 0xff;
	}

	spin_lock_irqsave(&cxd9802_spi_irqlock, flags);
	cxd9802_spi_enter_config_mode();
	cxd9802_spi_reg_changebits(REG_SPCR1, REGM_SER, REGS_SER(ser));
	spin_unlock_irqrestore(&cxd9802_spi_irqlock, flags);

	DPRFE;
	return;
}

static u_int32_t
cxd9802_spi_get_freq(void)
{
	u32 regval;
	int ser;
	u_int32_t freq;
	DPRFS;

	regval = cxd9802_spi_reg_read(REG_SPCR1);
	ser = REGG_SER(regval);
	freq = cxd9802_imbusclk_freq / 2 / (ser + 1);

	DPRFE;
	return freq;
}

static void
cxd9802_spi_set_ifdelay(u_int32_t ifdelay)
{
	int ifs;
	int ifspse;
	unsigned long flags;
	DPRFS;

	if (ifdelay > 0x3ff) {
		ifs = ifdelay / 32;
		ifspse = REGM_IFSPSE;
	} else {
		ifs = ifdelay;
		ifspse = 0;
	}
	if (ifs > 0x3ff) {
		ifs = 0x3ff;
	}

	spin_lock_irqsave(&cxd9802_spi_irqlock, flags);
	cxd9802_spi_enter_config_mode();
	cxd9802_spi_reg_changebits(REG_SPCR0, REGM_IFSPSE, ifspse);
	cxd9802_spi_reg_write(REG_SPFS, ifs);
	spin_unlock_irqrestore(&cxd9802_spi_irqlock, flags);

	DPRFE;
	return;
}

static u_int32_t
cxd9802_spi_get_ifdelay(void)
{
	int ifs;
	u_int32_t ifdelay;
	DPRFS;

	ifs = (cxd9802_spi_reg_read(REG_SPFS) & REGM_IFS);
	if (cxd9802_spi_reg_read(REG_SPCR0) & REGM_IFSPSE) {
		ifdelay = ifs * 32;
	} else {
		ifdelay = ifs;
	}

	DPRFE;
	return ifdelay;
}

static int
cxd9802_spi_set_size(u_int32_t nbits)
{
	int ssz;
	unsigned long flags;
	DPRFS;

	switch (nbits) {
	case 8:
		ssz = REGC_SSZ_8BITS;
		break;
	case 16:
		ssz = REGC_SSZ_16BITS;
		break;
	default:
		DPRFS;
		return -EINVAL;
	}

	spin_lock_irqsave(&cxd9802_spi_irqlock, flags);
	cxd9802_spi_enter_config_mode();
	cxd9802_spi_reg_changebits(REG_SPCR1, REGM_SSZ, ssz);
	spin_unlock_irqrestore(&cxd9802_spi_irqlock, flags);

	DPRFE;
	return 0;
}

static u_int32_t
cxd9802_spi_get_size(void)
{
	int ssz;
	DPRFS;

	ssz = REGG_SSZ(cxd9802_spi_reg_read(REG_SPCR1));

	DPRFE;
	return ssz;
}

static void
cxd9802_spi_set_mode(int mode)
{
	int sbos;
	int spol;
	int spha;
	unsigned long flags;
	DPRFS;

	/* default */
	spol = 0;		/* high active */
	spha = 0;		/* PHA_1 */
	sbos = REGM_SBOS;	/* MSB first */

	if (mode & SNSCSPI_M_POL_HIGH) {
		spol = 0;
	} else if (mode & SNSCSPI_M_POL_LOW) {
		spol = REGM_SPOL;
	}
	if (mode & SNSCSPI_M_PHA_1) {
		spha = 0;
	} else if (mode & SNSCSPI_M_PHA_2) {
		spha = REGM_SPHA;
	}
	if (mode & SNSCSPI_M_LSB) {
		sbos = 0;
	} else if (mode & SNSCSPI_M_MSB) {
		sbos = REGM_SBOS;
	}

	spin_lock_irqsave(&cxd9802_spi_irqlock, flags);
	cxd9802_spi_enter_config_mode();
	cxd9802_spi_reg_changebits(REG_SPCR0,
				   REGM_SBOS | REGM_SPOL | REGM_SPHA,
				   sbos | spol | spha);
	spin_unlock_irqrestore(&cxd9802_spi_irqlock, flags);

	DPRFE;
	return;
}

static int
cxd9802_spi_get_mode(void)
{
	u32 spcr0;
	int mode;
	DPRFS;

	spcr0 = cxd9802_spi_reg_read(REG_SPCR0);
	
	mode = 0;
	if (spcr0 & REGM_SPOL) {
		mode |= SNSCSPI_M_POL_LOW;
	} else {
		mode |= SNSCSPI_M_POL_HIGH;
	}
	if (spcr0 & REGM_SPHA) {
		mode |= SNSCSPI_M_PHA_2;
	} else {
		mode |= SNSCSPI_M_PHA_1;
	}
	if (spcr0 & REGM_SBOS) {
		mode |= SNSCSPI_M_MSB;
	} else {
		mode |= SNSCSPI_M_LSB;
	}

	DPRFE;
	return mode;
}
