/*
 *  RTC driver for RTC-7301
 *
 *  Copyright 2003 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* #define DEBUG 1 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/miscdevice.h>
#include <linux/poll.h>
#include <linux/spinlock.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <asm/delay.h>
#include <asm/io.h>

#include <linux/rtc.h>

#ifdef DEBUG
#define DPR(fmt, args...)	printk(fmt, ## args)
#endif /* DEBUG */

/* -------------------------- *
 * platform dependent         *
 * -------------------------- */

#ifdef CONFIG_SNSC_MPU22X
#include <linux/cxd9790.h>
#include <linux/cxd9790_reg.h>
#include <asm-mips/addrspace.h>
#include <asm-mips/tx4927/snsc_mpu22x/snsc_mpu22x.h>
#include <asm-mips/tx4927/snsc_mpu22x/mpu22x-memmap.h>
#endif /* CONFIG_SNSC_MPU22X */

#if (defined CONFIG_KAISER) || (defined CONFIG_SNSC_MPU410)
#include <linux/cxd9790.h>
#include <linux/cxd9790_reg.h>
#include <asm/io.h>
#include <asm/rtc7301.h>
#endif

/* ======================================================================= *
 * Constant                                                                *
 * ======================================================================= */

#define RTC7301_RTC_DEVNAME	"rtc"
#define RTC7301_RTC_NAME_STR	"RTC-7301 RTC driver"

#define RTC7301_RTC_WAIT_TIME_TO_STOP_CLOCK	130	/* >= 122 (us) */
#define RTC7301_RTC_WAIT_TIME_TO_NONBUSY	130
#define RTC7301_RTC_MAX_TIMES_TO_WAIT_NONBUSY	100

#define RTC7301_RTC_EVENT_ALARM		0xf0
#define RTC7301_RTC_EVENT_SEC_UPDATED	0xd0
#define RTC7301_RTC_EVENT_PERIODIC	0xc0

#define RTC7301_RTC_UIE_DEFAULT_POLLING_PERIOD	50	/* ms */

MODULE_AUTHOR("Sony PTC");
MODULE_DESCRIPTION(RTC7301_RTC_NAME_STR);
MODULE_LICENSE("GPL");

/* -------------------------- *
 * platform dependent         *
 * -------------------------- */

#ifdef CONFIG_SNSC_MPU22X
#define RTC7301_BASE_ADDR	(KSEG1 + RTC_BASE)
#define RTC7301_IRQ_NUM		CXD9790_IRQ_GPIOA(6)
#endif

#if (defined CONFIG_KAISER) || (defined CONFIG_SNSC_MPU410)
#define RTC7301_BASE_ADDR	PPC_RTC7301_BASE_VADDR
#define RTC7301_IRQ_NUM		PPC_RTC7301_IRQ_NUM
#endif

/* -------------------------- *
 * EPSON RTC-7301's register  *
 * -------------------------- */
/*
 * RTC7301R_xxx: configuration/status register offset
 * RTC7301M_xxx: mask to locate subfield in register
 * RTC7301C_xxx: constant to locate subfield in register
 * RTC7301S_xxx: set a value to subfield in register
 */

/* Bank 0 registers */
#define RTC7301R_CLK_SEC1		0x00
#define RTC7301R_CLK_SEC10		0x01
#define  RTC7301M_FOS			 (1 << 3)
#define  RTC7301M_CLK_SEC10		 (0x7 << 0)
#define  RTC7301S_CLK_SEC10(x)		 ((x) << 0)
#define RTC7301R_CLK_MIN1		0x02
#define RTC7301R_CLK_MIN10		0x03
#define RTC7301R_CLK_HOUR1		0x04
#define RTC7301R_CLK_HOUR10		0x05
#define RTC7301R_CLK_WDAY		0x06
#define RTC7301R_CLK_MDAY1		0x07
#define RTC7301R_CLK_MDAY10		0x08
#define RTC7301R_CLK_MON1		0x09
#define RTC7301R_CLK_MON10		0x0a
#define RTC7301R_CLK_YEAR1		0x0b
#define RTC7301R_CLK_YEAR10		0x0c
#define RTC7301R_CLK_YEAR100		0x0d
#define RTC7301R_CLK_YEAR1000		0x0e
#define  RTC7301M_TEST			 (1 << 3)
#define  RTC7301M_TEMP			 (1 << 2)
#define  RTC7301M_CLK_YEAR1000		 (0x3 << 0)
#define  RTC7301S_CLK_YEAR1000(x)	 ((x) << 0)
#define RTC7301R_CTRL			0x0f
#define  RTC7301M_BANK_SEL		 (0x3 << 2)
#define  RTC7301S_BANK_SEL(x)		 ((x) << 2)
#define  RTC7301M_STOP			 (1 << 1)
#define  RTC7301M_BUSY_ADJ		 (1 << 0)

/* Bank 1 */
#define RTC7301R_ALM_SEC1		0x00
#define RTC7301R_ALM_SEC10		0x01
#define  RTC7301M_ALM_SEC_AE		 (1 << 3)
#define  RTC7301M_ALM_SEC10		 (0x7 << 0)
#define  RTC7301S_ALM_SEC10(x)		 ((x) << 0)
#define RTC7301R_ALM_MIN1		0x02
#define RTC7301R_ALM_MIN10		0x03
#define  RTC7301M_ALM_MIN_AE		 (1 << 3)
#define  RTC7301M_ALM_MIN10		 (0x7 << 0)
#define  RTC7301S_ALM_MIN10(x)		 ((x) << 0)
#define RTC7301R_ALM_HOUR1		0x04
#define RTC7301R_ALM_HOUR10		0x05
#define  RTC7301M_ALM_HOUR_AE		 (1 << 3)
#define  RTC7301M_ALM_HOUR10		 (0x3 << 0)
#define  RTC7301S_ALM_HOUR10(x)		 ((x) << 0)
#define RTC7301R_ALM_WDAY		0x06
#define  RTC7301M_ALM_WDAY_AE		 (1 << 3)
#define  RTC7301M_ALM_WDAY		 (0x7 << 0)
#define  RTC7301S_ALM_WDAY(x)		 ((x) << 0)
#define RTC7301R_ALM_MDAY1		0x07
#define RTC7301R_ALM_MDAY10		0x08
#define  RTC7301M_ALM_MDAY_AE		 (1 << 3)
#define  RTC7301M_ALM_MDAY10		 (0x3 << 0)
#define  RTC7301S_ALM_MDAY10(x)		 ((x) << 0)
#define RTC7301R_CS1_CTRL		0x0b
#define  RTC7301M_CTEMP			 (1 << 3)
#define  RTC7301M_CDT_ON		 (1 << 2)
#define RTC7301R_FOUT_DIV		0x0c
#define  RTC7301M_FD210			 (0x7 << 0)
#define RTC7301R_FOUT_FREQ		0x0d
#define  RTC7301M_FE			 (1 << 3)
#define  RTC7301M_FD43			 (0x3 << 0)
#define RTC7301R_ALM_CTRL		0x0e
#define  RTC7301M_AF			 (1 << 1)
#define  RTC7301M_AIE			 (1 << 0)

/* Bank 2 */
#define RTC7301R_DT_L			0x00
#define RTC7301R_DT_H			0x01
#define  RTC7301M_DT_ON			 (1 << 3)
#define  RTC7301M_DT_H			 (0x7 << 0)
#define  RTC7301S_DT_H(x)		 ((x) << 0)
#define RTC7301R_TIMER_PRESET_L		0x04
#define RTC7301R_TIMER_PRESET_H		0x05
#define RTC7301R_TIMER_DATA_L		0x06
#define RTC7301R_TIMER_DATA_H		0x07
#define RTC7301R_TIMER_CONF		0x08
#define  RTC7301M_TE			 (1 << 3)
#define  RTC7301M_TI_TP			 (1 << 2)
#define  RTC7301M_TD			 (0x3 << 0)
#define  RTC7301S_TD(x)			 ((x) << 0)
#define  RTC7301C_TD_4096HZ		 RTC7301S_TD(0)
#define  RTC7301C_TD_64HZ		 RTC7301S_TD(1)
#define  RTC7301C_TD_1HZ		 RTC7301S_TD(2)
#define  RTC7301C_TD_MIN		 RTC7301S_TD(3)
#define RTC7301R_TIMER_CTRL		0x0e
#define  RTC7301M_TF			 (1 << 1)
#define  RTC7301M_TIE			 (1 << 0)

/* -------------------------- *
 * Sony CXD9790's register    *
 * -------------------------- */

#define CXD9790R_IRDM1_A		0x20
#define  CXD9790M_IC(n)			 (0x3 << (((n) & 0x3) * 2))
#define  CXD9790S_IC(n, x)		 ((x) << (((n) & 0x3) * 2))
#define  CXD9790C_IC_LEVEL_LOW		 0x0
#define  CXD9790C_IC_LEVEL_HIGH		 0x1
#define  CXD9790C_IC_EDGE_DOWN		 0x2
#define  CXD9790C_IC_EDGE_UP		 0x3

/* ======================================================================= *
 * Debug print function from dpr package                                   *
 * ======================================================================= */

#define DPL_EMERG	0
#define DPL_ALERT	1
#define DPL_CRIT	2
#define DPL_ERR		3
#define DPL_WARNING	4
#define DPL_NOTICE	5
#define DPL_INFO	6
#define DPL_DEBUG	7

#define DPM_PORPOSE	0x0000000f
#define DPM_ERR		0x00000001
#define DPM_TRACE	0x00000002
#define DPM_LAYER	0x000000f0
#define DPM_APPLICATION	0x00000010
#define DPM_MIDDLEWARE	0x00000020
#define DPM_DRIVER	0x00000040
#define DPM_KERNEL	0x00000080
#define DPM_USERDEF	0xffff0000

#ifdef DEBUG

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL DPL_DEBUG
#endif
#ifndef DEBUG_MASK
#define DEBUG_MASK  DPM_PORPOSE
#endif

#ifndef DPR
#include <stdio.h>
#define DPR(fmt, args...)	fprintf(stderr, fmt, ## args)
#endif

#define DPRL(level, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && DPR(fmt, ## args))
#define DPRM(mask, fmt, args...) \
	(void)(((mask) & DEBUG_MASK) && DPR(fmt, ## args))
#define DPRLM(level, mask, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && ((mask) & DEBUG_MASK) && \
		DPR(fmt, ## args))
#define DPRC(cond, fmt, args...) \
	(void)((cond) && DPR(fmt, ## args))
#define DPRF(fmt, args...)	DPR("%s: " fmt, __FUNCTION__, ## args)
#define DPRFL(level, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && DPRF(fmt, ## args))
#define DPRFM(mask, fmt, args...) \
	(void)(((mask) & DEBUG_MASK) && DPRF(fmt, ## args))
#define DPRFLM(level, mask, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && ((mask) & DEBUG_MASK) && \
		DPRF(fmt, ## args))
#define DPRFC(cond, fmt, args...) \
	(void)((cond) && DPRF(fmt, ## args))

#define	DPRFS	DPRFLM(DPL_DEBUG, DPM_TRACE, "start\n")
#define	DPRFE	DPRFLM(DPL_DEBUG, DPM_TRACE, "end\n")
#define DPRFR	DPRFLM(DPL_DEBUG, DPM_TRACE, "return\n")

#else  /* DEBUG */

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL -1
#endif
#ifndef DEBUG_MASK
#define DEBUG_MASK  0
#endif

#ifndef DPR
#define DPR(fmt, args...)
#endif

#define DPRL(level, fmt, args...)
#define DPRM(mask, fmt, args...)
#define DPRLM(level, mask, fmt, args...)
#define DPRC(cond, fmt, args...)
#define DPRF(fmt, args...)
#define DPRFL(level, fmt, args...)
#define DPRFM(mask, fmt, args...)
#define DPRFLM(level, mask, fmt, args...)
#define DPRFC(cond, fmt, args...)

#define	DPRFS
#define	DPRFE
#define DPRFR

#endif /* DEBUG */

/* ======================================================================= *
 * Type and structure                                                      *
 * ======================================================================= */

/* ======================================================================= *
 * Prototype declaration                                                   *
 * ======================================================================= */

/* initialize/finalize */
static int rtc7301_rtc_init(void);
static void rtc7301_rtc_exit(void);

/* RTC driver interface */
static int rtc7301_rtc_open(struct inode *inode, struct file *filp);
static int rtc7301_rtc_release(struct inode *inode, struct file *filp);
static ssize_t rtc7301_rtc_read(struct file *filp, char *buf,
				size_t count, loff_t *f_ops);
static int rtc7301_rtc_ioctl(struct inode *inode, struct file *filp,
			     unsigned int cmd, unsigned long arg);
static unsigned int rtc7301_rtc_poll(struct file *filp, poll_table *p_tab);
static int rtc7301_rtc_fasync(int fd, struct file *filp, int mode);

/* rtc_get_time() routine (export) */
unsigned long rtc7301_rtc_rtc_get_time(void);

/* interrupt handler */
static void rtc7301_rtc_intr_hdlr(int irq, void *dev_id, struct pt_regs *regs);

/* timer */
static void rtc7301_rtc_uie_polling(unsigned long arg);

/* device initialization */
static void rtc7301_rtc_dev_init(void);

/* ioctl internal */
static int rtc7301_rtc_ioctl_set_time(unsigned long arg);
static int rtc7301_rtc_ioctl_rd_time(unsigned long arg);
static int rtc7301_rtc_ioctl_epoch_set(unsigned long arg);
static int rtc7301_rtc_ioctl_epoch_read(unsigned long arg);
static int rtc7301_rtc_ioctl_alm_set(unsigned long arg);
static int rtc7301_rtc_ioctl_alm_read(unsigned long arg);
static int rtc7301_rtc_ioctl_aie_on(unsigned long arg);
static int rtc7301_rtc_ioctl_aie_off(unsigned long arg);
static int rtc7301_rtc_ioctl_uie_on(unsigned long arg);
static int rtc7301_rtc_ioctl_uie_off(unsigned long arg);
static int rtc7301_rtc_ioctl_irqp_set(unsigned long arg);
static int rtc7301_rtc_ioctl_irqp_read(unsigned long arg);
static int rtc7301_rtc_ioctl_pie_on(unsigned long arg);
static int rtc7301_rtc_ioctl_pie_off(unsigned long arg);

/* misc */
static void rtc7301_rtc_bank_select(int bank);
static int rtc7301_rtc_is_leap_year(int year);
static int rtc7301_rtc_days_in_month(int year, int month);
static int rtc7301_rtc_is_valid_time(const struct rtc_time *rtc_tm);
static int rtc7301_rtc_is_valid_alarm_time(const struct rtc_time *rtc_tm);
static void rtc7301_rtc_set_time_bank0_selected(const struct rtc_time *rtc_tm);
static void rtc7301_rtc_get_time_bank0_selected(struct rtc_time *rtc_tm);
static void rtc7301_rtc_set_alarm_time_bank1_selected(
	const struct rtc_time *rtc_tm);
static void rtc7301_rtc_get_alarm_time_bank1_selected(struct rtc_time *rtc_tm);
static int rtc7301_rtc_read_time(struct rtc_time *rtc_tm, int use_udelay);

/* platform dependent */
static void rtc7301_rtc_init_irq_pin(void);

/* ======================================================================= *
 * Global variable                                                         *
 * ======================================================================= */

static int rtc7301_rtc_uie_polling_period =
	RTC7301_RTC_UIE_DEFAULT_POLLING_PERIOD;
MODULE_PARM(rtc7301_rtc_uie_polling_period, "i");
MODULE_PARM_DESC(rtc7301_rtc_uie_polling_period,
		 "polling period for UIE (ms)");

static spinlock_t rtc7301_rtc_irqlock = SPIN_LOCK_UNLOCKED;
static DECLARE_WAIT_QUEUE_HEAD(rtc7301_rtc_event_wait);
static struct fasync_struct *rtc7301_rtc_fasync_queue = NULL;
static struct timer_list rtc7301_rtc_uie_timer;

static unsigned char rtc7301_rtc_event = 0;
static unsigned int rtc7301_rtc_nevents = 0;

static int rtc7301_rtc_previous_sec = -1;

/* ======================================================================= *
 * Inline function                                                         *
 * ======================================================================= */

/* ------------------------ *
 * register access          *
 * ------------------------ */

static inline u8
rtc7301_rtc_inb(int offset)
{
	/* mask by 0x0f needed because it only use lower 4-bits and
	 * upper 4-bits are somtimes stained by previous
	 * transactions. */
	return (readb(RTC7301_BASE_ADDR + offset) & 0xf); 
}

static inline void
rtc7301_rtc_outb(int offset, u8 value)
{
	writeb(value, RTC7301_BASE_ADDR + offset);
}

static inline void
rtc7301_rtc_changebitsb(int offset, u8 bits, u8 value)
{
	rtc7301_rtc_outb(offset, (rtc7301_rtc_inb(offset) & ~bits) | value);
}

static inline void
rtc7301_rtc_setbitsb(int offset, u8 bits)
{
	rtc7301_rtc_outb(offset, rtc7301_rtc_inb(offset) | bits);
}

static inline void
rtc7301_rtc_clearbitsb(int offset, u8 bits)
{
	rtc7301_rtc_changebitsb(offset, bits, 0);
}

static inline void
rtc7301_rtc_changebitsb_atomic(int offset, u8 bits, u8 value)
{
	unsigned long flags;

	spin_lock_irqsave(&rtc7301_rtc_irqlock, flags);
	rtc7301_rtc_outb(offset, (rtc7301_rtc_inb(offset) & ~bits) | value);
	spin_unlock_irqrestore(&rtc7301_rtc_irqlock, flags);
}

static inline void
rtc7301_rtc_setbitsb_atomic(int offset, u8 bits)
{
	unsigned long flags;

	spin_lock_irqsave(&rtc7301_rtc_irqlock, flags);
	rtc7301_rtc_outb(offset, rtc7301_rtc_inb(offset) | bits);
	spin_unlock_irqrestore(&rtc7301_rtc_irqlock, flags);
}

static inline void
rtc7301_rtc_clearbitsb_atomic(int offset, u8 bits)
{
	rtc7301_rtc_changebitsb_atomic(offset, bits, 0);
}

/* ======================================================================= *
 * Linux driver interface                                                  *
 * ======================================================================= */

static struct file_operations rtc7301_rtc_fops = {
	owner:		THIS_MODULE,
	open:		rtc7301_rtc_open,
	release:	rtc7301_rtc_release,
	read:		rtc7301_rtc_read,
	ioctl:		rtc7301_rtc_ioctl,
	poll:		rtc7301_rtc_poll,
	fasync:		rtc7301_rtc_fasync,
};

static struct miscdevice rtc7301_rtc_misc_dev = {
	minor:		RTC_MINOR,
	name:		RTC7301_RTC_DEVNAME,
	fops:		&rtc7301_rtc_fops,
};

/* ------------------------ *
 * initialize/finalize      *
 * ------------------------ */

static const struct rtc_time rtc7301_rtc_initial_time = {
	tm_sec:		0,
	tm_min:		0,
	tm_hour:	0,
	tm_mday:	1,
	tm_mon:		0,
	tm_year:	70,
	tm_wday:	4,	/* 1970/1/1 is Thursday */
};

static int
rtc7301_rtc_init(void)
{
	int res;
	unsigned long flags;
	DPRFS;

	init_timer(&rtc7301_rtc_uie_timer);

	/* check the FOS flags */
	rtc7301_rtc_bank_select(0);
	if (rtc7301_rtc_inb(RTC7301R_CLK_SEC10) & RTC7301M_FOS) {
		/* initialize the RTC device */
		printk(KERN_NOTICE
		       "rtc7301_rtc: RTC is not initialized or backuped.\n"
		       "rtc7301_rtc: Alarm and timer interrupt will be "
		       "disabled.\n"
		       "rtc7301_rtc: Please set the current time to RTC.\n");
		rtc7301_rtc_dev_init();
	}

	/* disable alarm and periodic interrupts */
	spin_lock_irqsave(&rtc7301_rtc_irqlock, flags);
	rtc7301_rtc_bank_select(1);
	rtc7301_rtc_clearbitsb(RTC7301R_ALM_CTRL, RTC7301M_AF | RTC7301M_AIE);
	rtc7301_rtc_bank_select(2);
	rtc7301_rtc_clearbitsb(RTC7301R_TIMER_CONF,
			       RTC7301M_TE | RTC7301M_TI_TP);
	spin_unlock_irqrestore(&rtc7301_rtc_irqlock, flags);

	rtc7301_rtc_init_irq_pin();

	res = request_irq(RTC7301_IRQ_NUM, rtc7301_rtc_intr_hdlr,
			  SA_INTERRUPT, RTC7301_RTC_DEVNAME,
			  &rtc7301_rtc_fops);
	if (res < 0) {
		goto err1;
	}

	res = misc_register(&rtc7301_rtc_misc_dev);
	if (res < 0) {
		goto err2;
	}

	printk(KERN_NOTICE "RTC: %s\n", RTC7301_RTC_NAME_STR);
	DPRFE;
	return 0;

err2:
	free_irq(RTC7301_IRQ_NUM, &rtc7301_rtc_fops);
err1:
	DPRFR;
	return res;
}

static void
rtc7301_rtc_exit(void)
{
	unsigned long flags;
	DPRFS;

	/* disable alarm and periodic interrupts */
	spin_lock_irqsave(&rtc7301_rtc_irqlock, flags);
	rtc7301_rtc_bank_select(1);
	rtc7301_rtc_clearbitsb(RTC7301R_ALM_CTRL, RTC7301M_AF | RTC7301M_AIE);
	rtc7301_rtc_bank_select(2);
	rtc7301_rtc_clearbitsb(RTC7301R_TIMER_CONF,
			       RTC7301M_TE | RTC7301M_TI_TP);
	spin_unlock_irqrestore(&rtc7301_rtc_irqlock, flags);

	misc_deregister(&rtc7301_rtc_misc_dev);

	free_irq(RTC7301_IRQ_NUM, &rtc7301_rtc_fops);

	del_timer(&rtc7301_rtc_uie_timer);

	DPRFE;
	return;
}

module_init(rtc7301_rtc_init);
module_exit(rtc7301_rtc_exit);

/* ------------------------ *
 * RTC driver interface     *
 * ------------------------ */

static int
rtc7301_rtc_open(struct inode *inode, struct file *filp)
{
	DPRFS;

	MOD_INC_USE_COUNT;

	DPRFE;
	return 0;
}

static int
rtc7301_rtc_release(struct inode *inode, struct file *filp)
{
	DPRFS;

	MOD_DEC_USE_COUNT;

	DPRFE;
	return 0;
}

static ssize_t
rtc7301_rtc_read(struct file *filp, char *buf, size_t count, loff_t *f_ops)
{
	int res;
	unsigned long flags;
	unsigned long event;
	DPRFS;

	if (count < sizeof(event)) {
		DPRFR;
		return 0;
	}
	res = access_ok(VERIFY_WRITE, buf, count);
	if (!res) {
		DPRFR;
		return -EFAULT;
	}

	while (1) {
		spin_lock_irqsave(&rtc7301_rtc_irqlock, flags);
		if (rtc7301_rtc_nevents > 0) {
			break;
		}
		spin_unlock_irqrestore(&rtc7301_rtc_irqlock, flags);

		if (filp->f_flags & O_NONBLOCK) {
			DPRFR;
			return -EAGAIN;
		}

		res = wait_event_interruptible(rtc7301_rtc_event_wait,
					       rtc7301_rtc_nevents > 0);
		if (res < 0) {
			DPRFR;
			return res;
		}
	}

	/* TRICK: still holding irqlock here */
	if (rtc7301_rtc_nevents > 0xffffff) {
		rtc7301_rtc_nevents = 0xffffff;
	}
	event = (rtc7301_rtc_event | (rtc7301_rtc_nevents << 8));
	rtc7301_rtc_nevents = 0;

	spin_unlock_irqrestore(&rtc7301_rtc_irqlock, flags);

	put_user(event, (unsigned long *)buf);

	DPRFE;
	return sizeof(event);
}

static int
rtc7301_rtc_ioctl(struct inode *inode, struct file *filp,
		 unsigned int cmd, unsigned long arg)
{
	DPRFS;

	switch (cmd) {
	case RTC_SET_TIME:	return rtc7301_rtc_ioctl_set_time(arg);
	case RTC_RD_TIME:	return rtc7301_rtc_ioctl_rd_time(arg);
	case RTC_EPOCH_SET:	return rtc7301_rtc_ioctl_epoch_set(arg);
	case RTC_EPOCH_READ:	return rtc7301_rtc_ioctl_epoch_read(arg);
	case RTC_ALM_SET:	return rtc7301_rtc_ioctl_alm_set(arg);
	case RTC_ALM_READ:	return rtc7301_rtc_ioctl_alm_read(arg);
	case RTC_AIE_ON:	return rtc7301_rtc_ioctl_aie_on(arg);
	case RTC_AIE_OFF:	return rtc7301_rtc_ioctl_aie_off(arg);
	case RTC_UIE_ON:	return rtc7301_rtc_ioctl_uie_on(arg);
	case RTC_UIE_OFF:	return rtc7301_rtc_ioctl_uie_off(arg);
	case RTC_IRQP_SET:	return rtc7301_rtc_ioctl_irqp_set(arg);
	case RTC_IRQP_READ:	return rtc7301_rtc_ioctl_irqp_read(arg);
	case RTC_PIE_ON:	return rtc7301_rtc_ioctl_pie_on(arg);
	case RTC_PIE_OFF:	return rtc7301_rtc_ioctl_pie_off(arg);
	}

	DPRFR;
	return -EINVAL;
}

static unsigned int
rtc7301_rtc_poll(struct file *filp, poll_table *p_tab)
{
	unsigned int mask;
	DPRFS;

	poll_wait(filp, &rtc7301_rtc_event_wait, p_tab);

	mask = 0;
	if (rtc7301_rtc_nevents > 0) {
		mask |= POLLIN | POLLRDNORM;
	}

	DPRFE;
	return mask;
}

static int
rtc7301_rtc_fasync(int fd, struct file *filp, int mode)
{
	int res;
	DPRFS;

	res = fasync_helper(fd, filp, mode, &rtc7301_rtc_fasync_queue);

	DPRFE;
	return res;
}

/* ------------------------ *
 * rtc_get_time() routine   *
 * ------------------------ */

unsigned long
rtc7301_rtc_rtc_get_time(void)
{
	int res;
	struct rtc_time rtc_tm;
	long rtc_time;
	DPRFS;

	res = rtc7301_rtc_read_time(&rtc_tm, 1);
	if (res < 0) {
		printk(KERN_WARNING "rtc7301_rtc: Cannot get RTC time!\n");
		DPRFR;
		return mktime(2000, 1, 1, 0, 0, 0);
	}
	rtc_time = mktime(rtc_tm.tm_year + 1900, rtc_tm.tm_mon + 1,
			  rtc_tm.tm_mday,
			  rtc_tm.tm_hour, rtc_tm.tm_min, rtc_tm.tm_sec);
	if (rtc_time < 0) {
		rtc_time = 0;
	}

	DPRFE;
	return rtc_time;
}

/* ======================================================================= *
 * Interrupt handler                                                       *
 * ======================================================================= */

static void
rtc7301_rtc_intr_hdlr(int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned long flags;
	DPRFS;

	spin_lock_irqsave(&rtc7301_rtc_irqlock, flags);

	/* check timer interrupt */
	rtc7301_rtc_bank_select(2);
	if (rtc7301_rtc_inb(RTC7301R_TIMER_CTRL) & RTC7301M_TF) {
		rtc7301_rtc_clearbitsb(RTC7301R_TIMER_CTRL, RTC7301M_TF);
		rtc7301_rtc_event = RTC7301_RTC_EVENT_PERIODIC;
		rtc7301_rtc_nevents++;
	}

	/* check alarm interrupt */
	rtc7301_rtc_bank_select(1);
	if (rtc7301_rtc_inb(RTC7301R_ALM_CTRL) & RTC7301M_AF) {
		rtc7301_rtc_clearbitsb(RTC7301R_ALM_CTRL, RTC7301M_AF);
		rtc7301_rtc_event = RTC7301_RTC_EVENT_ALARM;
		rtc7301_rtc_nevents++;
	}

	spin_unlock_irqrestore(&rtc7301_rtc_irqlock, flags);

	if (rtc7301_rtc_nevents > 0) {
		wake_up_interruptible(&rtc7301_rtc_event_wait);
	}
	if (rtc7301_rtc_nevents > 0 && rtc7301_rtc_fasync_queue != NULL) {
		kill_fasync(&rtc7301_rtc_fasync_queue, SIGIO, POLL_IN);
	}

	DPRFE;
	return;
}

/* ======================================================================= *
 * Timer                                                                   *
 * ======================================================================= */

static void
rtc7301_rtc_uie_polling(unsigned long arg)
{
	unsigned long flags;
	int sec;
	DPRFS;

	spin_lock_irqsave(&rtc7301_rtc_irqlock, flags);

	rtc7301_rtc_bank_select(0);
	sec = (rtc7301_rtc_inb(RTC7301R_CLK_SEC1) +
	       (rtc7301_rtc_inb(RTC7301R_CLK_SEC10) &
		RTC7301M_CLK_SEC10) * 10);
	if (sec != rtc7301_rtc_previous_sec) {
		rtc7301_rtc_event = RTC7301_RTC_EVENT_SEC_UPDATED;
		rtc7301_rtc_nevents++;
	}
	rtc7301_rtc_previous_sec = sec;

	spin_unlock_irqrestore(&rtc7301_rtc_irqlock, flags);

	if (rtc7301_rtc_nevents > 0) {
		wake_up_interruptible(&rtc7301_rtc_event_wait);
	}
	if (rtc7301_rtc_nevents > 0 && rtc7301_rtc_fasync_queue != NULL) {
		kill_fasync(&rtc7301_rtc_fasync_queue, SIGIO, POLL_IN);
	}

	/* set timer for polling */
	if (!timer_pending(&rtc7301_rtc_uie_timer)) {
		rtc7301_rtc_uie_timer.expires = jiffies +
			(rtc7301_rtc_uie_polling_period * HZ) / 1000;
		if (rtc7301_rtc_uie_timer.expires == jiffies) {
			rtc7301_rtc_uie_timer.expires++;
		}
		rtc7301_rtc_uie_timer.function = rtc7301_rtc_uie_polling;
		add_timer(&rtc7301_rtc_uie_timer);
	}

	DPRFE;
	return;
}

/* ======================================================================= *
 * Internal function                                                       *
 * ======================================================================= */

/* ------------------------ *
 * device initialization    *
 * ------------------------ */

static void
rtc7301_rtc_dev_init(void)
{
	DPRFS;

	/* TEST <- 0, TEMP <- 0 */
	rtc7301_rtc_bank_select(0);
	rtc7301_rtc_clearbitsb(RTC7301R_CLK_YEAR1000,
			       RTC7301M_TEST | RTC7301M_TEMP);

	/* stop the clock */
	rtc7301_rtc_outb(RTC7301R_CTRL, RTC7301M_STOP);
		
	/* 122us or more wait */
	udelay(RTC7301_RTC_WAIT_TIME_TO_STOP_CLOCK);

	/* disable alarm */
	rtc7301_rtc_bank_select(1);
	rtc7301_rtc_clearbitsb(RTC7301R_ALM_CTRL, RTC7301M_AF | RTC7301M_AIE);

	/* initialize CS1 (but not used) */
	rtc7301_rtc_setbitsb(RTC7301R_CS1_CTRL,
			     RTC7301M_CTEMP | RTC7301M_CDT_ON);

	/* disable FOUT */
	rtc7301_rtc_clearbitsb(RTC7301R_FOUT_FREQ, RTC7301M_FE);

	/* disable digital correction */
	rtc7301_rtc_bank_select(2);
	rtc7301_rtc_clearbitsb(RTC7301R_DT_H, RTC7301M_DT_ON);

	/* disable timer interrupt */
	rtc7301_rtc_clearbitsb(RTC7301R_TIMER_CONF,
			       RTC7301M_TE | RTC7301M_TI_TP);
	rtc7301_rtc_clearbitsb(RTC7301R_TIMER_CTRL,
			       RTC7301M_TF | RTC7301M_TIE);

	/* set time */
	rtc7301_rtc_bank_select(0);
	rtc7301_rtc_set_time_bank0_selected(&rtc7301_rtc_initial_time);

	/* start the clock */
	rtc7301_rtc_outb(RTC7301R_CTRL, 0);

	DPRFE;
	return;
}

/* ------------------------ *
 * ioctl internal           *
 * ------------------------ */

static int
rtc7301_rtc_ioctl_set_time(unsigned long arg)
{
	int res;
	struct rtc_time rtc_tm;
	unsigned long flags;
	DPRFS;

	if (!capable(CAP_SYS_TIME)) {
		DPRFR;
		return -EACCES;
	}

	res = copy_from_user(&rtc_tm, (void *)arg, sizeof(rtc_tm));
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}

	if (!rtc7301_rtc_is_valid_time(&rtc_tm)) {
		DPRFR;
		return -EINVAL;
	}

	/* stop the clock */
	spin_lock_irqsave(&rtc7301_rtc_irqlock, flags);
	rtc7301_rtc_bank_select(0);
	rtc7301_rtc_outb(RTC7301R_CTRL, RTC7301M_STOP);
	spin_unlock_irqrestore(&rtc7301_rtc_irqlock, flags);

	/* 122us wait */
	udelay(RTC7301_RTC_WAIT_TIME_TO_STOP_CLOCK);

	spin_lock_irqsave(&rtc7301_rtc_irqlock, flags);

	/* set time */
	rtc7301_rtc_bank_select(0);
	rtc7301_rtc_set_time_bank0_selected(&rtc_tm);

	/* restart the clock */
	rtc7301_rtc_outb(RTC7301R_CTRL, 0);

	spin_unlock_irqrestore(&rtc7301_rtc_irqlock, flags);

	DPRFE;
	return 0;
}

static int
rtc7301_rtc_ioctl_rd_time(unsigned long arg)
{
	int res;
	struct rtc_time rtc_tm;
	DPRFS;

	res = rtc7301_rtc_read_time(&rtc_tm, 0);
	if (res < 0) {
		DPRFR;
		return res;
	}

	res = copy_to_user((void *)arg, &rtc_tm, sizeof(rtc_tm));
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}

	if (!rtc7301_rtc_is_valid_time(&rtc_tm)) {
		printk(KERN_WARNING
		       "rtc7301_rtc: RTC holds invalid time!\n"
		       "rtc7301_rtc: Please reset the time.\n");
	}

	DPRFE;
	return 0;
}

static int
rtc7301_rtc_ioctl_epoch_set(unsigned long arg)
{
	DPRFS;

	/* NOTE: always fail */

	if (!capable(CAP_SYS_TIME)) {
		DPRFR;
		return -EACCES;
	}

	DPRFE;
	return -ENODEV;
}

static int
rtc7301_rtc_ioctl_epoch_read(unsigned long arg)
{
	int res;
	unsigned long epoch;
	DPRFS;

	epoch = 0;
	
	res = copy_to_user((void *)arg, &epoch, sizeof(epoch));
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}

	DPRFE;
	return 0;
}

static int
rtc7301_rtc_ioctl_alm_set(unsigned long arg)
{
	int res;
	struct rtc_time rtc_tm;
	unsigned long flags;
	int aie_save;
	DPRFS;

	res = copy_from_user(&rtc_tm, (void *)arg, sizeof(rtc_tm));
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}

	if (!rtc7301_rtc_is_valid_alarm_time(&rtc_tm)) {
		DPRFR;
		return -EINVAL;
	}

	spin_lock_irqsave(&rtc7301_rtc_irqlock, flags);

	/* disable alarm */
	rtc7301_rtc_bank_select(1);
	aie_save = (rtc7301_rtc_inb(RTC7301R_ALM_CTRL) & RTC7301M_AIE);
	rtc7301_rtc_clearbitsb(RTC7301R_ALM_CTRL, RTC7301M_AF | RTC7301M_AIE);

	/* set alarm time */
	rtc7301_rtc_set_alarm_time_bank1_selected(&rtc_tm);

	/* re-enable alarm if necessary */
	rtc7301_rtc_changebitsb(RTC7301R_ALM_CTRL, RTC7301M_AF | RTC7301M_AIE,
				aie_save);

	spin_unlock_irqrestore(&rtc7301_rtc_irqlock, flags);

	DPRFE;
	return 0;
}

static int
rtc7301_rtc_ioctl_alm_read(unsigned long arg)
{
	int res;
	struct rtc_time rtc_tm;
	unsigned long flags;
	DPRFS;

	spin_lock_irqsave(&rtc7301_rtc_irqlock, flags);

	rtc7301_rtc_bank_select(1);
	rtc7301_rtc_get_alarm_time_bank1_selected(&rtc_tm);

	spin_unlock_irqrestore(&rtc7301_rtc_irqlock, flags);

	res = copy_to_user((void *)arg, &rtc_tm, sizeof(rtc_tm));
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}

	if (!rtc7301_rtc_is_valid_alarm_time(&rtc_tm)) {
		printk(KERN_WARNING
		       "rtc7301_rtc: RTC holds invalid alarm time!\n");
	}

	DPRFE;
	return 0;
}

static int
rtc7301_rtc_ioctl_aie_on(unsigned long arg)
{
	unsigned long flags;
	DPRFS;

	spin_lock_irqsave(&rtc7301_rtc_irqlock, flags);
	rtc7301_rtc_bank_select(1);
	rtc7301_rtc_changebitsb(RTC7301R_ALM_CTRL,
				RTC7301M_AF | RTC7301M_AIE, RTC7301M_AIE);
	spin_unlock_irqrestore(&rtc7301_rtc_irqlock, flags);

	DPRFE;
	return 0;
}

static int
rtc7301_rtc_ioctl_aie_off(unsigned long arg)
{
	unsigned long flags;
	DPRFS;

	spin_lock_irqsave(&rtc7301_rtc_irqlock, flags);
	rtc7301_rtc_bank_select(1);
	rtc7301_rtc_clearbitsb(RTC7301R_ALM_CTRL, RTC7301M_AF | RTC7301M_AIE);
	spin_unlock_irqrestore(&rtc7301_rtc_irqlock, flags);

	DPRFE;
	return 0;
}

static int
rtc7301_rtc_ioctl_uie_on(unsigned long arg)
{
	unsigned long flags;
	DPRFS;

	spin_lock_irqsave(&rtc7301_rtc_irqlock, flags);

	if (timer_pending(&rtc7301_rtc_uie_timer)) {
		spin_unlock_irqrestore(&rtc7301_rtc_irqlock, flags);
		/* UIE timer is already enabled */
		DPRFR;
		return 0;
	}

	rtc7301_rtc_bank_select(0);
	rtc7301_rtc_previous_sec = (rtc7301_rtc_inb(RTC7301R_CLK_SEC1) +
				    (rtc7301_rtc_inb(RTC7301R_CLK_SEC10) &
				     RTC7301M_CLK_SEC10) * 10);

	rtc7301_rtc_uie_timer.expires = jiffies +
		(rtc7301_rtc_uie_polling_period * HZ) / 1000;
	if (rtc7301_rtc_uie_timer.expires == jiffies) {
		rtc7301_rtc_uie_timer.expires++;
	}
	rtc7301_rtc_uie_timer.function = rtc7301_rtc_uie_polling;
	add_timer(&rtc7301_rtc_uie_timer);

	spin_unlock_irqrestore(&rtc7301_rtc_irqlock, flags);

	DPRFE;
	return 0;
}

static int
rtc7301_rtc_ioctl_uie_off(unsigned long arg)
{
	DPRFS;

	del_timer(&rtc7301_rtc_uie_timer);

	DPRFE;
	return 0;
}

static int
rtc7301_rtc_ioctl_irqp_set(unsigned long arg)
{
	unsigned long flags;
	unsigned long hz;
	int td_bits;
	unsigned char preset_value;
	DPRFS;

	hz = arg;

	if (hz < 1) {
		DPRFR;
		return -EINVAL;
	}
	if (hz > 16) {
		td_bits = RTC7301C_TD_4096HZ;
		preset_value = 4096 / hz;
	} else {
		td_bits = RTC7301C_TD_64HZ;
		preset_value = 64 / hz;
	}
	if (preset_value == 0) {
		preset_value = 1;
	}

	spin_lock_irqsave(&rtc7301_rtc_irqlock, flags);

	rtc7301_rtc_bank_select(2);

	/* disable timer */
	rtc7301_rtc_clearbitsb(RTC7301R_TIMER_CONF,
			       RTC7301M_TE | RTC7301M_TI_TP);
	rtc7301_rtc_clearbitsb(RTC7301R_TIMER_CTRL,
			       RTC7301M_TF | RTC7301M_TIE);

	/* set preset value */
	rtc7301_rtc_outb(RTC7301R_TIMER_DATA_L, 0);
	rtc7301_rtc_outb(RTC7301R_TIMER_DATA_H, 0);
	rtc7301_rtc_outb(RTC7301R_TIMER_PRESET_L, preset_value & 0x0f);
	rtc7301_rtc_outb(RTC7301R_TIMER_PRESET_H, preset_value >> 4);

	rtc7301_rtc_changebitsb(RTC7301R_TIMER_CONF,
				RTC7301M_TI_TP | RTC7301M_TD,
				RTC7301M_TI_TP | td_bits);

	spin_unlock_irqrestore(&rtc7301_rtc_irqlock, flags);

	DPRFE;
	return 0;
}

static int
rtc7301_rtc_ioctl_irqp_read(unsigned long arg)
{
	int res;
	unsigned long flags;
	unsigned long hz;
	unsigned char preset_value;
	DPRFS;

	spin_lock_irqsave(&rtc7301_rtc_irqlock, flags);

	rtc7301_rtc_bank_select(2);

	preset_value = (rtc7301_rtc_inb(RTC7301R_TIMER_PRESET_L) |
			(rtc7301_rtc_inb(RTC7301R_TIMER_PRESET_H) << 4));
	if (preset_value == 0) {
		spin_unlock_irqrestore(&rtc7301_rtc_irqlock, flags);
		DPRFR;
		return -ERANGE;
	}

	switch (rtc7301_rtc_inb(RTC7301R_TIMER_CONF) & RTC7301M_TD) {
	case RTC7301C_TD_4096HZ:
		hz = 4096 / preset_value;
		break;
	case RTC7301C_TD_64HZ:
		hz = 64 / preset_value;
		break;
	default:
		spin_unlock_irqrestore(&rtc7301_rtc_irqlock, flags);
		DPRFR;
		return -ERANGE;
	}

	spin_unlock_irqrestore(&rtc7301_rtc_irqlock, flags);

	res = copy_to_user((void *)arg, &hz, sizeof(hz));
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}

	DPRFE;
	return 0;
}

static int
rtc7301_rtc_ioctl_pie_on(unsigned long arg)
{
	unsigned long flags;
	unsigned char preset_value;
	DPRFS;

	spin_lock_irqsave(&rtc7301_rtc_irqlock, flags);

	rtc7301_rtc_bank_select(2);

	preset_value = (rtc7301_rtc_inb(RTC7301R_TIMER_PRESET_L) |
			(rtc7301_rtc_inb(RTC7301R_TIMER_PRESET_H) << 4));
	if (preset_value == 0) {
		spin_unlock_irqrestore(&rtc7301_rtc_irqlock, flags);
		DPRFR;
		return -EPERM;
	}

	/* enable timer */
	rtc7301_rtc_changebitsb(RTC7301R_TIMER_CTRL,
				RTC7301M_TF | RTC7301M_TIE,
				RTC7301M_TIE);
	rtc7301_rtc_setbitsb(RTC7301R_TIMER_CONF,
			     RTC7301M_TE | RTC7301M_TI_TP);

	spin_unlock_irqrestore(&rtc7301_rtc_irqlock, flags);

	DPRFE;
	return 0;
}

static int
rtc7301_rtc_ioctl_pie_off(unsigned long arg)
{
	unsigned long flags;
	DPRFS;

	spin_lock_irqsave(&rtc7301_rtc_irqlock, flags);
	rtc7301_rtc_bank_select(2);
	rtc7301_rtc_clearbitsb(RTC7301R_TIMER_CONF,
			       RTC7301M_TE | RTC7301M_TI_TP);
	spin_unlock_irqrestore(&rtc7301_rtc_irqlock, flags);

	DPRFE;
	return 0;
}

/* ------------------------ *
 * misc                     *
 * ------------------------ */

static void
rtc7301_rtc_bank_select(int bank)
{
	u8 regval;
	unsigned long flags;
	DPRFS;

	spin_lock_irqsave(&rtc7301_rtc_irqlock, flags);
	regval = rtc7301_rtc_inb(RTC7301R_CTRL);
	regval &= ~RTC7301M_BANK_SEL;
	regval |= RTC7301S_BANK_SEL(bank);
	regval &= ~RTC7301M_BUSY_ADJ;
	rtc7301_rtc_outb(RTC7301R_CTRL, regval);
	spin_unlock_irqrestore(&rtc7301_rtc_irqlock, flags);

	DPRFE;
	return;
}

static int
rtc7301_rtc_is_leap_year(int year)
{
	return (year % 4 == 0 && (year % 100 != 0 || year % 400 == 0));
}


static int
rtc7301_rtc_days_in_month(int year, int month)
{
	/* month = 0, 1, ..., 11 */
	static const int days_in_month[] = {
		31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
	};

	if (month == 1 && rtc7301_rtc_is_leap_year(year)) {
		return 29;
	}

	return days_in_month[month];
};

static int
rtc7301_rtc_is_valid_time(const struct rtc_time *rtc_tm)
{
	int is_valid;
	DPRFS;

	is_valid = 0;
	if (rtc_tm->tm_wday >= 0 && rtc_tm->tm_wday < 7 &&
	    rtc_tm->tm_year >= 0 && rtc_tm->tm_year + 1900 < 3999 &&
	    rtc_tm->tm_mon >= 0 && rtc_tm->tm_mon < 12 &&
	    rtc_tm->tm_mday >= 1 &&
	    rtc_tm->tm_mday <= rtc7301_rtc_days_in_month(rtc_tm->tm_year + 1900,
							rtc_tm->tm_mon) &&
	    rtc_tm->tm_hour >= 0 && rtc_tm->tm_hour < 24 &&
	    rtc_tm->tm_min >= 0 && rtc_tm->tm_min < 60 &&
	    rtc_tm->tm_sec >= 0 && rtc_tm->tm_sec < 60) {
		is_valid = 1;
	}

	DPRFE;
	return is_valid;
}

static int
rtc7301_rtc_is_valid_alarm_time(const struct rtc_time *rtc_tm)
{
	int is_valid;
	DPRFS;

	is_valid = 0;
	if ((rtc_tm->tm_wday == -1 ||
	     (rtc_tm->tm_wday >= 0 && rtc_tm->tm_wday < 7)) &&
	    (rtc_tm->tm_mday == -1 ||
	     (rtc_tm->tm_mday >= 1 && rtc_tm->tm_mday <= 31)) &&
	    (rtc_tm->tm_hour == -1 ||
	     (rtc_tm->tm_hour >= 0 && rtc_tm->tm_hour < 24)) &&
	    (rtc_tm->tm_min == -1 ||
	     (rtc_tm->tm_min >= 0 && rtc_tm->tm_min < 60)) &&
	    (rtc_tm->tm_min == -1 ||
	     (rtc_tm->tm_sec >= 0 && rtc_tm->tm_sec < 60))) {
		is_valid = 1;
	}

	DPRFE;
	return is_valid;
}

static void
rtc7301_rtc_set_time_bank0_selected(const struct rtc_time *rtc_tm)
{
	int year;
	DPRFS;

	/* NOTE: FOS flag will be cleared */

	year = rtc_tm->tm_year + 1900;

	rtc7301_rtc_outb(RTC7301R_CLK_SEC1,	rtc_tm->tm_sec % 10);
	rtc7301_rtc_outb(RTC7301R_CLK_SEC10,	rtc_tm->tm_sec / 10);
	rtc7301_rtc_outb(RTC7301R_CLK_MIN1,	rtc_tm->tm_min % 10);
	rtc7301_rtc_outb(RTC7301R_CLK_MIN10,	rtc_tm->tm_min / 10);
	rtc7301_rtc_outb(RTC7301R_CLK_HOUR1,	rtc_tm->tm_hour % 10);
	rtc7301_rtc_outb(RTC7301R_CLK_HOUR10,	rtc_tm->tm_hour / 10);
	rtc7301_rtc_outb(RTC7301R_CLK_WDAY,	rtc_tm->tm_wday);
	rtc7301_rtc_outb(RTC7301R_CLK_MDAY1,	rtc_tm->tm_mday % 10);
	rtc7301_rtc_outb(RTC7301R_CLK_MDAY10,	rtc_tm->tm_mday / 10);
	rtc7301_rtc_outb(RTC7301R_CLK_MON1,	(rtc_tm->tm_mon + 1) % 10);
	rtc7301_rtc_outb(RTC7301R_CLK_MON10,	(rtc_tm->tm_mon + 1) / 10);
	rtc7301_rtc_outb(RTC7301R_CLK_YEAR1,	year % 10);
	rtc7301_rtc_outb(RTC7301R_CLK_YEAR10,	(year / 10) % 10);
	rtc7301_rtc_outb(RTC7301R_CLK_YEAR100,	(year / 100) % 10);
	rtc7301_rtc_outb(RTC7301R_CLK_YEAR1000,	year / 1000);

	DPRFE;
	return;
}

static void
rtc7301_rtc_get_time_bank0_selected(struct rtc_time *rtc_tm)
{
	DPRFS;

	memset(rtc_tm, 0, sizeof(*rtc_tm));

	rtc_tm->tm_sec = (rtc7301_rtc_inb(RTC7301R_CLK_SEC1) +
			  (rtc7301_rtc_inb(RTC7301R_CLK_SEC10) &
			   RTC7301M_CLK_SEC10) * 10);
	rtc_tm->tm_min = (rtc7301_rtc_inb(RTC7301R_CLK_MIN1) +
			  rtc7301_rtc_inb(RTC7301R_CLK_MIN10) * 10);
	rtc_tm->tm_hour = (rtc7301_rtc_inb(RTC7301R_CLK_HOUR1) +
			   rtc7301_rtc_inb(RTC7301R_CLK_HOUR10) * 10);
	rtc_tm->tm_wday = rtc7301_rtc_inb(RTC7301R_CLK_WDAY);
	rtc_tm->tm_mday = (rtc7301_rtc_inb(RTC7301R_CLK_MDAY1) +
			   rtc7301_rtc_inb(RTC7301R_CLK_MDAY10) * 10);
	rtc_tm->tm_mon = (rtc7301_rtc_inb(RTC7301R_CLK_MON1) +
			  rtc7301_rtc_inb(RTC7301R_CLK_MON10) * 10) - 1;
	rtc_tm->tm_year = (rtc7301_rtc_inb(RTC7301R_CLK_YEAR1) +
			   rtc7301_rtc_inb(RTC7301R_CLK_YEAR10) * 10 +
			   rtc7301_rtc_inb(RTC7301R_CLK_YEAR100) * 100 +
			   (rtc7301_rtc_inb(RTC7301R_CLK_YEAR1000) &
			    RTC7301M_CLK_YEAR1000) * 1000) - 1900;

	DPRFE;
	return;
}

static void
rtc7301_rtc_set_alarm_time_bank1_selected(const struct rtc_time *rtc_tm)
{
	DPRFS;

	if (rtc_tm->tm_sec == -1) {
		rtc7301_rtc_setbitsb(RTC7301R_ALM_SEC10, RTC7301M_ALM_SEC_AE);
	} else {
		rtc7301_rtc_outb(RTC7301R_ALM_SEC1,	rtc_tm->tm_sec % 10);
		rtc7301_rtc_outb(RTC7301R_ALM_SEC10,
				 RTC7301S_ALM_SEC10(rtc_tm->tm_sec / 10));
	}
	if (rtc_tm->tm_min == -1) {
		rtc7301_rtc_setbitsb(RTC7301R_ALM_MIN10, RTC7301M_ALM_MIN_AE);
	} else {
		rtc7301_rtc_outb(RTC7301R_ALM_MIN1,	rtc_tm->tm_min % 10);
		rtc7301_rtc_outb(RTC7301R_ALM_MIN10,
				 RTC7301S_ALM_MIN10(rtc_tm->tm_min / 10));
	}
	if (rtc_tm->tm_hour == -1) {
		rtc7301_rtc_setbitsb(RTC7301R_ALM_HOUR10,
				     RTC7301M_ALM_HOUR_AE);
	} else {
		rtc7301_rtc_outb(RTC7301R_ALM_HOUR1,	rtc_tm->tm_hour % 10);
		rtc7301_rtc_changebitsb(RTC7301R_ALM_HOUR10,
					RTC7301M_ALM_HOUR_AE |
					RTC7301M_ALM_HOUR10,
					RTC7301S_ALM_HOUR10(rtc_tm->tm_hour /
							    10));
	}
	if (rtc_tm->tm_wday == -1) {
		rtc7301_rtc_setbitsb(RTC7301R_ALM_WDAY, RTC7301M_ALM_WDAY_AE);
	} else {
		rtc7301_rtc_outb(RTC7301R_ALM_WDAY,
				 RTC7301S_ALM_WDAY(rtc_tm->tm_wday));
	}
	if (rtc_tm->tm_mday == -1) {
		rtc7301_rtc_setbitsb(RTC7301R_ALM_MDAY10,
				     RTC7301M_ALM_MDAY_AE);
	} else {
		rtc7301_rtc_outb(RTC7301R_ALM_MDAY1,	rtc_tm->tm_mday % 10);
		rtc7301_rtc_changebitsb(RTC7301R_ALM_MDAY10,
					RTC7301M_ALM_MDAY_AE |
					RTC7301M_ALM_MDAY10,
					RTC7301S_ALM_MDAY10(rtc_tm->tm_mday /
							    10));
	}

	DPRFE;
	return;
}

static void
rtc7301_rtc_get_alarm_time_bank1_selected(struct rtc_time *rtc_tm)
{
	unsigned char tmp_regval;
	DPRFS;

	memset(rtc_tm, 0, sizeof(*rtc_tm));

	tmp_regval = rtc7301_rtc_inb(RTC7301R_ALM_SEC10);
	if (tmp_regval & RTC7301M_ALM_SEC_AE) {
		rtc_tm->tm_sec = -1;
	} else {
		rtc_tm->tm_sec = (rtc7301_rtc_inb(RTC7301R_ALM_SEC1) +
				  (tmp_regval & RTC7301M_ALM_SEC10) * 10);
	}
	tmp_regval = rtc7301_rtc_inb(RTC7301R_ALM_MIN10);
	if (tmp_regval & RTC7301M_ALM_MIN_AE) {
		rtc_tm->tm_min = -1;
	} else {
		rtc_tm->tm_min = (rtc7301_rtc_inb(RTC7301R_ALM_MIN1) +
				  (tmp_regval & RTC7301M_ALM_MIN10) * 10);
	}
	tmp_regval = rtc7301_rtc_inb(RTC7301R_ALM_HOUR10);
	if (tmp_regval & RTC7301M_ALM_HOUR_AE) {
		rtc_tm->tm_hour = -1;
	} else {
		rtc_tm->tm_hour = (rtc7301_rtc_inb(RTC7301R_ALM_HOUR1) +
				   (tmp_regval & RTC7301M_ALM_HOUR10) * 10);
	}
	tmp_regval = rtc7301_rtc_inb(RTC7301R_ALM_WDAY);
	if (tmp_regval & RTC7301M_ALM_WDAY_AE) {
		rtc_tm->tm_wday = -1;
	} else {
		rtc_tm->tm_wday = (tmp_regval & RTC7301M_ALM_WDAY);
	}
	tmp_regval = rtc7301_rtc_inb(RTC7301R_ALM_MDAY10);
	if (tmp_regval & RTC7301M_ALM_MDAY_AE) {
		rtc_tm->tm_mday = -1;
	} else {
		rtc_tm->tm_mday = (rtc7301_rtc_inb(RTC7301R_ALM_MDAY1) +
				   (tmp_regval & RTC7301M_ALM_MDAY10) * 10);
	}

	DPRFE;
	return;
}

static int
rtc7301_rtc_read_time(struct rtc_time *rtc_tm, int use_udelay)
{
	int i;
	unsigned long flags;
	DPRFS;

	/* wait for busy = 0 */
	for (i = 0; i < RTC7301_RTC_MAX_TIMES_TO_WAIT_NONBUSY; i++) {
		spin_lock_irqsave(&rtc7301_rtc_irqlock, flags);
		if (!(rtc7301_rtc_inb(RTC7301R_CTRL) & RTC7301M_BUSY_ADJ)) {
			break;
		}
		spin_unlock_irqrestore(&rtc7301_rtc_irqlock, flags);

		if (use_udelay) {
			udelay(RTC7301_RTC_WAIT_TIME_TO_NONBUSY);
		} else {
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(1);
			if (signal_pending(current)) {
				DPRFR;
				return -ERESTARTSYS;
			}
		}
	}
	/* TRICK: still holding irqlock here */
	if (i == RTC7301_RTC_MAX_TIMES_TO_WAIT_NONBUSY) {
		spin_unlock_irqrestore(&rtc7301_rtc_irqlock, flags);
		DPRFR;
		return -EIO;
	}

	rtc7301_rtc_bank_select(0);
	rtc7301_rtc_get_time_bank0_selected(rtc_tm);

	spin_unlock_irqrestore(&rtc7301_rtc_irqlock, flags);

	DPRFE;
	return 0;
}

/* ======================================================================= *
 * Platform dependent                                                      *
 * ======================================================================= */

#if defined (CONFIG_SNSC_MPU22X) || defined (CONFIG_KAISER)
static void
rtc7301_rtc_init_irq_pin(void)
{
	DPRFS;

	/* set CXD9790's GPIO interrupt mode */
	cxd9790_set_trigger_mode(RTC7301_IRQ_NUM, TRIG_FALLING_EDGE);

	DPRFE;
	return;
}
#endif /* CONFIG_SNSC_MPU22X */
