/*
 *  snsc_cxd9790_kpad.c
 *
 *  Keypad driver for IOC-9 connected to CXD9790 GPIO
 *
 *  Copyright 2001-2003 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/interrupt.h>

#include <linux/snsc_kpad.h>
#include <linux/cxd9790_gpio.h>

MODULE_AUTHOR("Sony NSC");
MODULE_LICENSE("GPL");

/* 8x8 matrix */
#define KEYCODE(row, col)  ( ((col)<< 4) + (row) + 1 )
#define KBUP               0x80

#define KPAD_ROWS          8
#define KPAD_COLS          8

/* various delay */
#define INTVL_TIMER        3
#define CHAT_DELAY         50
#define SCAN_DELAY         10   /* usec */

/* port & bitmask used for keypad */
#define IN_PORT            GPIO_PORT_D
#define OUT_PORT           GPIO_PORT_C
#define IN_MASK            0xff
#define OUT_MASK           0xff

#define MODNAME            "CXD9790 GPIO keypad"

typedef struct _scan_info {
	unsigned char lastkey[KPAD_ROWS];
} SCANINFO;

static SCANINFO scaninfo;

static struct timer_list kpad_timer = {
	function:       NULL
};

static int minor = -1;   /* dynamic allocation */

/* prototypes */
static void cxd9790_kpad_key_scan(unsigned long arg);
static void cxd9790_kpad_tasklet(unsigned long arg);
static int  cxd9790_kpad_open( struct keypad_dev *dev );
static int  cxd9790_kpad_release( struct keypad_dev *dev );
static void cxd9790_kpad_intr_handler(int irq, void *dev_id, struct pt_regs *regs);

static DECLARE_TASKLET(cxd9790_kpad_tasklet_data,
		       cxd9790_kpad_tasklet,
		       (unsigned long)&scaninfo);

static struct keypad_dev cxd9790_kpad_dev = {
	name:           "cxd9790_kpad",
	minor:          -1,
	open:           cxd9790_kpad_open,
	release:        cxd9790_kpad_release,
	ioctl:          NULL,
};

/*
 *  Initialize & Cleanup
 */
static int
cxd9790_kpad_init( void )
{
	int res;
	int i;
	SCANINFO *sinfo = &scaninfo;

	for (i = 0; i < KPAD_ROWS; i++){
		sinfo->lastkey[i] = 0;
	}

	/* register as keypad driver */
	cxd9790_kpad_dev.minor = minor;
	res = register_keypad(&cxd9790_kpad_dev);
	if (res < 0){
		printk(KERN_WARNING MODNAME 
		       ": register_keypad failed(%d)\n", res);
		return res;
	}

	/* initialize GPIO */
	res = cxd9790_gpio_register(IN_PORT, IN_MASK, "Keypad In");
	if (res < 0)
		return res;
	res = cxd9790_gpio_register(OUT_PORT, OUT_MASK, "Keypad Out");
	if (res < 0)
		goto err1;
	res = cxd9790_gpio_set_mode(IN_PORT, IN_MASK, GPIO_MODE_IN);
	if (res < 0)
		goto err1;
	res = cxd9790_gpio_set_mode(OUT_PORT, OUT_MASK, GPIO_MODE_OUT_LOW);
	if (res < 0)
		goto err1;
	res = cxd9790_gpio_intr_set_mode(IN_PORT, IN_MASK, GPIO_INTR_EDGE_LOW);
	if (res < 0)
		goto err1;

	for (i = 0; i < 8; i++) {
		res = cxd9790_gpio_intr_register_handler(
			IN_PORT, i, cxd9790_kpad_intr_handler, NULL);
		if (res < 0)
			goto err2;
	}

	res = cxd9790_gpio_intr_mask(IN_PORT, IN_MASK);
	if (res < 0)
		goto err2;

	res = cxd9790_gpio_intr_unmask(IN_PORT, IN_MASK);
	if (res < 0)
		goto err2;

	return 0;

  err2:
	cxd9790_gpio_intr_mask(IN_PORT, IN_MASK);
	for (i = 0; i < 8; i++) {
		cxd9790_gpio_intr_unregister_handler(IN_PORT, i, NULL);
	}

  err1:
	cxd9790_gpio_unregister(IN_PORT, IN_MASK);
	cxd9790_gpio_unregister(OUT_PORT, OUT_MASK);

	unregister_keypad(&cxd9790_kpad_dev);

	return res;
}

static void
cxd9790_kpad_final( void )
{
	int i;

	cxd9790_gpio_intr_mask(IN_PORT, IN_MASK);
	for (i = 0; i < 8; i++) {
		cxd9790_gpio_intr_unregister_handler(IN_PORT, i, NULL);
	}
	cxd9790_gpio_unregister(IN_PORT, IN_MASK);
	cxd9790_gpio_unregister(OUT_PORT, OUT_MASK);

	if (timer_pending(&kpad_timer)) {
		del_timer(&kpad_timer);
	}

	unregister_keypad(&cxd9790_kpad_dev);
}

module_init(cxd9790_kpad_init);
module_exit(cxd9790_kpad_final);


/*
 * SNSC keypad driver interface
 */
static int
cxd9790_kpad_open( struct keypad_dev *dev )
{
	MOD_INC_USE_COUNT;

	return 0;
}

static int
cxd9790_kpad_release( struct keypad_dev *dev )
{
	MOD_DEC_USE_COUNT;

	return 0;
}

/*
 *  Interrupt handler
 */
static void
cxd9790_kpad_intr_handler(int irq, void *dev_id, 
			  struct pt_regs *regs)
{
	int res;
	int port, bit;

	port = cxd9790_gpio_irq_to_port(irq);
	if (port != IN_PORT) {
		printk(KERN_WARNING MODNAME 
		       ": invalid interrupt port(%d)\n",port);
		return;
	}

	bit = cxd9790_gpio_irq_to_bit(irq);
	if (bit < 0) {
		printk(KERN_WARNING MODNAME
		       ": invalid interrupt bit\n");
		return;
	}

	res = cxd9790_gpio_intr_mask(IN_PORT, IN_MASK);
	if (res < 0) {
		return;
	}

	tasklet_schedule(&cxd9790_kpad_tasklet_data);
}

static void
cxd9790_kpad_tasklet(unsigned long arg)
{
	udelay(CHAT_DELAY);
	cxd9790_kpad_key_scan(arg);
}


/*
 *  Key scan routine (called by interrupt handler and/or timer)
 */
static void
cxd9790_kpad_key_scan(unsigned long arg)
{
	SCANINFO *sinfo = (SCANINFO *)arg;
	int i;
	int push_flg = 0;

	for (i = 0; i < KPAD_ROWS; i++){
		unsigned char diffkey;
		unsigned char key;

		cxd9790_gpio_set_data(OUT_PORT, OUT_MASK, ~(1 << i));
		udelay(SCAN_DELAY);
		cxd9790_gpio_get_data(IN_PORT, IN_MASK, &key);
		key = ~key;

		if (key)
			push_flg = 1;

		diffkey = key ^ sinfo->lastkey[i];
		if ( diffkey ){
			unsigned char downkey;
			unsigned char upkey;
			int j;

			downkey = key & diffkey;
			upkey = ~key & diffkey;

			for ( j = 0; j < KPAD_COLS; j++){
				if ( (1 << j) & downkey ){
					keypad_event(&cxd9790_kpad_dev,
						     KEYCODE(i, j));
				} else if ( (1 << j) & upkey ){
					keypad_event(&cxd9790_kpad_dev,
						     KEYCODE(i, j) | KBUP);
				}
			}
		}
		sinfo->lastkey[i] = key;
	}

	if (push_flg){ /* keep polling mode */
		kpad_timer.expires = jiffies + INTVL_TIMER;
		kpad_timer.data = arg;
		kpad_timer.function = cxd9790_kpad_key_scan;
		if (!(timer_pending(&kpad_timer))){
			add_timer(&kpad_timer);
		}
	} else { /* enter interrupt mode again */
		cxd9790_gpio_set_data(OUT_PORT, OUT_MASK, 0);
		cxd9790_gpio_intr_unmask(IN_PORT, IN_MASK);
	}

	return;
}
