/*
 *  linux/driver/char/snsc_ich_gpio.c
 *
 *  driver for ICH2 GPIO
 *
 *  Copyright 2001-2003 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/module.h>
#include <linux/config.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include <linux/spinlock.h>
#include <linux/smp_lock.h>
#include <linux/miscdevice.h>
#include <linux/snsc_major.h>
#include <linux/pci.h>
#include <linux/pci_ids.h>
#include <linux/snsc_ich_gpio.h>
#include <linux/errno.h>
#include <asm/io.h>
#include <asm/uaccess.h>

/* ICH2 Device ID */
#define PCI_DEVICE_ID_ICH2_LPC      0x2440
#define PCI_DEVICE_ID_ICH2_PCIBRG   0x244e

/* LPC interface PCI Configuration Register */
#define PMBASE                      0x40
#define ACPI_CNTL                   0x44   /* for SCI IRQ */
#define GPIOBASE                    0x58
#define GPI_ROUT                    0xB8

/* ICH2 GPIO I/O registers*/
#define  GPIO_USE_SEL    0x00
#define  GP_IO_SEL       0x04
#define  GP_LVL          0x0c
#define  GPI_INV         0x2c

/* ICH2 PM I/O registers*/
#define  GPE1_STS        0x2c
#define  GPE1_EN         0x2e

#define NUM_GPIO         32
#define OWNER_NAME_LEN   16

/* pins which can cause interrupt: 0, 1, 3, 4, 6, 7, 8, 11, 12, 13 */
#define GPIO_MASK_INTERRUPTIBLE ((1 << 0) | (1 << 1) | (1 << 3) |  \
				 (1 << 4) | (1 << 6) | (1 << 7) |  \
				 (1 << 8) | (1 << 11) | (1 << 12) | (1 << 13))

//#define DEBUG

#ifdef DEBUG
#define GPIO_DEBUG(args...) printk(KERN_NOTICE args);
#else
#define GPIO_DEBUG(args...)
#endif

/* for internal use */
static void ichgpio_use_sel(__u32 bitmask, __u32 sel);
static void ichgpio_io_sel(__u32 bitmask, __u32 sel);
static void ichgpio_sig_inv(__u32 bitmask, __u32 level);
static void ichgpio_ctrl_rout(int id, __u32 rout);
static void ichgpio_enable_intr(__u16 bitmask);
static void ichgpio_disable_intr(__u16 bitmask);
static void ichgpio_clear_intr(__u16 bitmask);
static __u16 ichgpio_int_status(void);
static __u32 ichgpio_in(void);
static void ichgpio_out(__u32 v, __u32 mask);
static int ichgpio_irq(void);

static void ich_gpio_intr_handler(int irq, void *dev_id, struct pt_regs *regs);

struct ichgpio_info {
        struct pci_dev  *lpc;
        __u32           pm_base;
        __u32           gpio_base;
        int             sci_irq;
        spinlock_t      lock;
	__u32           registered;
	__u32           intr_registered;
	struct _gpio_pin {
		char            owner[OWNER_NAME_LEN];
		void            (*intr_handler)(int irq, void *dev_id,
						struct pt_regs *regs);
		void            *intr_dev_id;
	} gpio[NUM_GPIO];
	spinlock_t      reg_lock;
};

static struct ichgpio_info  ichgpio_info;

#define MSG_HEAD  "snsc_ich_gpio.o: "

/* Use Select */
static void ichgpio_use_sel(__u32 bitmask, __u32 sel)
{
        struct ichgpio_info *info = &ichgpio_info;
        unsigned long flags;
        __u32  v32;

        spin_lock_irqsave(&info->lock ,flags);
        v32 = inl(info->gpio_base + GPIO_USE_SEL);
        v32 &= ~bitmask;
        outl(v32 | ((sel & 1) * bitmask), info->gpio_base + GPIO_USE_SEL);
        spin_unlock_irqrestore(&info->lock, flags);
}

/* I/O Select */
static void ichgpio_io_sel(__u32 bitmask, __u32 sel)
{
        struct ichgpio_info *info = &ichgpio_info;
        unsigned long flags;
        __u32  v32;

        spin_lock_irqsave(&info->lock ,flags);
        v32 = inl(info->gpio_base + GP_IO_SEL);
        v32 &= ~bitmask;
        outl(v32 | ((sel & 1) * bitmask), info->gpio_base + GP_IO_SEL);
        spin_unlock_irqrestore(&info->lock, flags);
}

/* Signal invert */
static void ichgpio_sig_inv(__u32 bitmask, __u32 level)
{
        struct ichgpio_info *info = &ichgpio_info;
        unsigned long flags;
        __u32  v32;

        spin_lock_irqsave(&info->lock ,flags);
        v32 = inl(info->gpio_base + GPI_INV);
        v32 &= ~bitmask;
        outl(v32 | ((level & 1) * bitmask), info->gpio_base + GPI_INV);
        spin_unlock_irqrestore(&info->lock, flags);
}

/* GPI Routing Control */
static void ichgpio_ctrl_rout(int id, __u32 rout)
{
        struct ichgpio_info *info = &ichgpio_info;
        unsigned long flags;
        __u32  v32;

        spin_lock_irqsave(&info->lock ,flags);
        pci_read_config_dword(info->lpc, GPI_ROUT, &v32);
        v32 &= ~(0x3 << (id << 1));
        pci_write_config_dword(info->lpc, GPI_ROUT, v32 | ((rout & 0x3) << (id << 1)));
        spin_unlock_irqrestore(&info->lock, flags);
}

/* enable interrupt */
static void ichgpio_enable_intr(__u16 bitmask)
{
        struct ichgpio_info *info = &ichgpio_info;
        unsigned long flags;
        __u16  v16;

        spin_lock_irqsave(&info->lock ,flags);
        v16 = inw(info->pm_base + GPE1_EN);
        outw(v16 | bitmask, info->pm_base + GPE1_EN);
        spin_unlock_irqrestore(&info->lock, flags);
}

/* disable interrupt */
static void ichgpio_disable_intr(__u16 bitmask)
{
        struct ichgpio_info *info = &ichgpio_info;
        unsigned long flags;
        __u16  v16;

        spin_lock_irqsave(&info->lock ,flags);
        v16 = inw(info->pm_base + GPE1_EN);
        outw(v16 & ~bitmask, info->pm_base + GPE1_EN);
        spin_unlock_irqrestore(&info->lock, flags);
}

/* clear status */
static void ichgpio_clear_intr(__u16 bitmask)
{
        struct ichgpio_info *info = &ichgpio_info;
        unsigned long flags;
        __u16  v16;

        spin_lock_irqsave(&info->lock ,flags);
        v16 = inw(info->pm_base + GPE1_STS);
        outw(bitmask, info->pm_base + GPE1_STS);
        spin_unlock_irqrestore(&info->lock, flags);
}

/* read status */
static __u16 ichgpio_int_status(void)
{
        struct ichgpio_info *info = &ichgpio_info;

        return inw(info->pm_base + GPE1_STS);
}

/* in */
static __u32 ichgpio_in(void)
{
        struct ichgpio_info *info = &ichgpio_info;
	unsigned long flags;
	__u32 val;

	spin_lock_irqsave(&info->lock, flags);

	val = inl(info->gpio_base + GP_LVL);
	val ^= (inl(info->gpio_base + GPI_INV) & GPIO_MASK_INTERRUPTIBLE);

	spin_unlock_irqrestore(&info->lock, flags);

        return val;
}

/* out */
static void ichgpio_out(__u32 v, __u32 mask)
{
        struct ichgpio_info *info = &ichgpio_info;
	unsigned long flags;
	__u32 v32;

	spin_lock_irqsave(&info->lock, flags);

	v32 = ichgpio_in();
	v32 &= ~mask;
	v32 |= (v & mask);
        outl(v32, info->gpio_base + GP_LVL);

	spin_unlock_irqrestore(&info->lock, flags);
}

/* get irq */
static int ichgpio_irq(void)
{
        struct ichgpio_info *info = &ichgpio_info;
        return info->sci_irq;
}


static int ichgpio_open(struct inode *inode, struct file *file)
{
        unsigned int minor = MINOR(inode->i_rdev);
        if (minor != ICH_GPIO_MINOR)
                return -ENODEV;

        MOD_INC_USE_COUNT;
        return 0;
}

static int ichgpio_release(struct inode *inode, struct file *file)
{
        MOD_DEC_USE_COUNT;
        return 0;
}

static int ichgpio_ioctl(struct inode *inode, struct file *file,
                            unsigned int cmd, unsigned long arg)
{
        unsigned int minor = MINOR(inode->i_rdev);
        struct ichgpio_iocdata iocdata;

        if (minor != ICH_GPIO_MINOR)
                return -ENODEV;

        switch (cmd) {
        case ICHGPIO_IOC_USESEL:
                copy_from_user((void *)&iocdata, (void *)arg, sizeof(struct ichgpio_iocdata));
                ichgpio_use_sel((1 << iocdata.id), iocdata.val);
                break;

        case ICHGPIO_IOC_IOSEL:
                copy_from_user((void *)&iocdata, (void *)arg, sizeof(struct ichgpio_iocdata));
                ichgpio_io_sel((1 << iocdata.id), iocdata.val);
                break;

        case ICHGPIO_IOC_SIGINV:
                copy_from_user((void *)&iocdata, (void *)arg, sizeof(struct ichgpio_iocdata));
                ichgpio_sig_inv((1 << iocdata.id), iocdata.val);
                break;

        case ICHGPIO_IOC_CLRSTAT:
                ichgpio_clear_intr((1 << (int)arg));
                break;

        case ICHGPIO_IOC_READSTAT:
                return (ichgpio_int_status() & (1 << (int)arg)) ? 1 : 0;

        case ICHGPIO_IOC_IN:
                return ichgpio_in();

        case ICHGPIO_IOC_OUT:
                ichgpio_out((__u32)arg, 0xffffffff);
                break;

        default:
                return -ENOIOCTLCMD;
        }

        return 0;
}

static struct file_operations ichgpio_fops =
{
	owner:		THIS_MODULE,
	ioctl:		ichgpio_ioctl,
	open:		ichgpio_open,
	release:	ichgpio_release,
};

static struct miscdevice ichgpio_miscdev =
{
	ICH_GPIO_MINOR,
	"ichgpio",
	&ichgpio_fops
};

int __init ichgpio_init(void)
{
        struct ichgpio_info *info = &ichgpio_info;
        static int ichgpio_initialized = 0;
        int    rval;

        if (ichgpio_initialized)
                return 0;
        ichgpio_initialized = 1;

        printk("ICH GPIO driver: $Revision: 1.14 $\n");

        /* find LPC */
        info->lpc = pci_find_device (PCI_VENDOR_ID_INTEL,
				     PCI_DEVICE_ID_ICH2_LPC, NULL);
        if (!info->lpc) {
                printk("%s : cannot find %x\n",
		       __FUNCTION__, PCI_DEVICE_ID_ICH2_LPC);
                return 1;
        }

        /* get PM base and GPIO base */
        pci_read_config_dword(info->lpc, PMBASE, &info->pm_base);
        info->pm_base &= PCI_BASE_ADDRESS_IO_MASK;
        pci_read_config_dword(info->lpc, GPIOBASE, &info->gpio_base);
        info->gpio_base &= PCI_BASE_ADDRESS_IO_MASK;

        /* SCI IRQ */
        pci_read_config_dword(info->lpc, ACPI_CNTL, &info->sci_irq);
        info->sci_irq = (info->sci_irq & 0x7) + 9;
        if (info->sci_irq >= 13) {
                info->sci_irq += 7;
        }

        /* initialize spinlock */
        spin_lock_init(&info->lock);
	spin_lock_init(&info->reg_lock);

        /* register as a misc driver */
        if ((rval = misc_register(&ichgpio_miscdev)) < 0) {
                printk(MSG_HEAD "cannot register ichgpio as a misc driver\n");
                return rval;
        }

	if (request_irq(ichgpio_irq(), ich_gpio_intr_handler,
			SA_SHIRQ, "ich_gpio", info)) {
		printk(KERN_NOTICE "ich_gpio: failed to request_irq\n");
		misc_deregister(&ichgpio_miscdev);
		return -EBUSY;
	}

        return 0;
}

static void __exit ichgpio_cleanup(void)
{
        misc_deregister(&ichgpio_miscdev);
}



/*
 * SNSC common GPIO driver interfaces
 */

#define PORT_CHECK(port) do {                                                 \
                              if ((port) != 0) {                              \
                                  GPIO_DEBUG("%s: port %d is not supported\n",\
                                             __FUNCTION__, (port));           \
                                  return -EINVAL;                             \
                              }                                               \
                         } while (0);

#define BITNUM_CHECK(bitnum) do {                                             \
                                 if ((bitnum) < 0 || (bitnum) >= NUM_GPIO) {  \
                                     GPIO_DEBUG("%s: invalid pin number %d\n",\
                                                __FUNCTION__, (bitnum));      \
                                     return -EINVAL;                          \
                                 }                                            \
                             } while (0);

/*
 * interrupt handler
 */
static void
ich_gpio_intr_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	struct ichgpio_info *info = &ichgpio_info;
	__u32 status;
	int i;

	ich_gpio_intr_get_status(0, &status);
	status &= (__u32)inw(info->pm_base + GPE1_EN);

	status &= GPIO_MASK_INTERRUPTIBLE;

	for (i = 0; i < NUM_GPIO; i++) {
		if (status & (1 << i)) {
			if (((info->intr_registered & (1 << i)) == 0) ||
			    (info->gpio[i].intr_handler == NULL)) {
				GPIO_DEBUG("%s: warning, gpio %d assserts "
					   "interrupt, but handler is not "
					   "registered.\n", __FUNCTION__, i);
				continue;
			}
			/* call registered interrupt handler */
			(*info->gpio[i].intr_handler)
				(irq, info->gpio[i].intr_dev_id, regs);

			ich_gpio_intr_clear_bit(0, i);
		}
	}
}


/* initialize */
int
ich_gpio_register_bit(int port, int bitnum, const char *name)
{
	BITNUM_CHECK(bitnum);
	return ich_gpio_register(port, (1 << bitnum), name);
}

int
ich_gpio_unregister_bit(int port, int bitnum)
{
	BITNUM_CHECK(bitnum);
	return ich_gpio_unregister(port, (1 << bitnum));
}

int
ich_gpio_register(int port, __u32 bitmask, const char *name)
{
        struct ichgpio_info *info = &ichgpio_info;
	unsigned long flags;
	int i;

	PORT_CHECK(port);

	if (name == NULL) {
		GPIO_DEBUG("%s: name must be specified\n", __FUNCTION__);
		return -EINVAL;
	}

	spin_lock_irqsave(&info->reg_lock, flags);

	if (bitmask & info->registered) {
		spin_unlock_irqrestore(&info->reg_lock, flags);
		GPIO_DEBUG("%s: bit(s) are already registered(0x%08x)\n",
			   __FUNCTION__, bitmask);
		return -EBUSY;
	}

	info->registered |= bitmask;

	for (i = 0; i < NUM_GPIO; i++) {
		if (bitmask & (1 << i)) {
			strncpy(info->gpio[i].owner, name, OWNER_NAME_LEN);
			info->gpio[i].owner[OWNER_NAME_LEN] = '\0';
		}
	}

	spin_unlock_irqrestore(&info->reg_lock, flags);

	GPIO_DEBUG("%s: registered pin 0x%08x by %s\n",
		   __FUNCTION__, bitmask, name);

	return 0;
}

int
ich_gpio_unregister(int port, __u32 bitmask)
{
        struct ichgpio_info *info = &ichgpio_info;
	unsigned long flags;
	int i;

	PORT_CHECK(port);

	spin_lock_irqsave(&info->reg_lock, flags);

	if ((bitmask & info->registered) != bitmask) {
		GPIO_DEBUG("%s: attempt to unregister 0x%08x, but "
			   "registered pin(s) are 0x%08x\n",
			   __FUNCTION__, bitmask, info->registered);
		spin_unlock_irqrestore(&info->reg_lock, flags);
		return -EINVAL;
	}

	info->registered &= ~bitmask;

	for (i = 0; i < NUM_GPIO; i++) {
		if (bitmask & (1 << i))
			memset(info->gpio[i].owner, '\0', OWNER_NAME_LEN);
	}

	ichgpio_use_sel(bitmask, ICHGPIO_USESEL_NATIVE);

	spin_unlock_irqrestore(&info->reg_lock, flags);

	GPIO_DEBUG("%s: unregistered pin 0x%08x\n",
		   __FUNCTION__, bitmask);

	return 0;
}

int
ich_gpio_unregister_by_name(const char *name)
{
        struct ichgpio_info *info = &ichgpio_info;
	unsigned long flags;
	__u32 bitmask = 0;
	int i;

	if (name == NULL) {
		GPIO_DEBUG("%s: name must be specified\n", __FUNCTION__);
		return -EINVAL;
	}

	spin_lock_irqsave(&info->reg_lock, flags);

	for (i = 0; i < NUM_GPIO; i++) {
		if (strncmp(info->gpio[i].owner, name, OWNER_NAME_LEN) == 0) {
			bitmask |= (1 << i);
		}
	}

	spin_unlock_irqrestore(&info->reg_lock, flags);

	if (bitmask == 0) {
		GPIO_DEBUG("%s: there's no pin registered by %s!\n",
			   __FUNCTION__, name);
		return -EINVAL;
	}

	return ich_gpio_unregister(0, bitmask);
}

int
ich_gpio_set_mode_bit(int port, int bitnum, __u32 mode)
{
	BITNUM_CHECK(bitnum);
	return ich_gpio_set_mode(port, (1 << bitnum), mode);
}

int
ich_gpio_set_mode(int port, __u32 bitmask , __u32 mode)
{
	PORT_CHECK(port);

	if (mode == GPIO_MODE_IN) {
		ichgpio_io_sel(bitmask, ICHGPIO_IOSEL_IN);
		ichgpio_use_sel(bitmask, ICHGPIO_USESEL_GPIO);
		GPIO_DEBUG("%s: set mode for pin 0x%08x to 0x%08x\n",
			   __FUNCTION__, bitmask, mode);
		return 0;
	}

	if (mode == GPIO_MODE_OUT_HIGH) {
		ichgpio_out(bitmask, bitmask);
	} else if (mode == GPIO_MODE_OUT_LOW) {
		ichgpio_out(0, bitmask);
	} else {
		GPIO_DEBUG("%s: invalid mode!\n", __FUNCTION__);
		return -EINVAL;
	}

	ichgpio_io_sel(bitmask, ICHGPIO_IOSEL_OUT);
	ichgpio_use_sel(bitmask, ICHGPIO_USESEL_GPIO);

	GPIO_DEBUG("%s: set mode for pin 0x%08x to 0x%08x\n",
		   __FUNCTION__, bitmask, mode);

	return 0;
}

/* data register access */
int
ich_gpio_get_data_bit(int port, int bitnum)
{
	__u32 value;
	int ret;

	BITNUM_CHECK(bitnum);

	ret = ich_gpio_get_data(port, (1 << bitnum), &value);

	if (ret == 0) {
		return (value != 0);
	} else {
		return ret;
	}
}

int
ich_gpio_set_data_bit(int port, int bitnum, int value)
{
	BITNUM_CHECK(bitnum);
	return ich_gpio_set_data(port, (1 << bitnum), ((value & 1) << bitnum));
}

int
ich_gpio_get_data(int port, __u32 bitmask, __u32 *value)
{
	PORT_CHECK(port);

	*value = (ichgpio_in() & bitmask);

	GPIO_DEBUG("%s: read 0x%08x (mask 0x%08x)\n",
		   __FUNCTION__, *value, bitmask);

	return 0;
}

int
ich_gpio_set_data(int port, __u32 bitmask , __u32 value)
{
	PORT_CHECK(port);

	ichgpio_out(value, bitmask);

	GPIO_DEBUG("%s: write 0x%08x (mask 0x%08x)\n",
		   __FUNCTION__, value, bitmask);

	return 0;
}

/* interrupt services */

/*
 * optional: we don't provide these function
 */
/*
void ich_gpio_intr_enable( void );
void ich_gpio_intr_disable( void );
*/

int
ich_gpio_intr_set_mode_bit(int port, int bitnum, __u32 mode)
{
	BITNUM_CHECK(bitnum);
	return ich_gpio_intr_set_mode(port, (1 << bitnum), mode);
}

int
ich_gpio_intr_set_mode(int port, __u32 bitmask, __u32 mode)
{
	__u32 rout;
	int i;

	PORT_CHECK(port);

	if (bitmask & ~GPIO_MASK_INTERRUPTIBLE) {
		GPIO_DEBUG("%s: specified pin(s) cannot cause interrupt",
			   __FUNCTION__);
		return -EINVAL;
	}

	ich_gpio_intr_mask(port, bitmask);

	switch (mode & GPIO_INTR_LEVEL_MASK) {
	case GPIO_INTR_LEVEL_LOW:
		ichgpio_sig_inv(bitmask, ICHGPIO_SIGINV_LOW);
		break;
	case GPIO_INTR_LEVEL_HIGH:
	default:
		ichgpio_sig_inv(bitmask, ICHGPIO_SIGINV_HIGH);
		break;
	}

	switch(mode & GPIO_INTR_ROUT_MASK) {
	case GPIO_INTR_ROUT_NO:
		rout = ICHGPIO_CTRLROUT_NO;
		break;
	case GPIO_INTR_ROUT_SMI:
		rout = ICHGPIO_CTRLROUT_SMI;
		break;
	case GPIO_INTR_ROUT_SCI:
	default:
		rout = ICHGPIO_CTRLROUT_SCI;
		break;
	}

	for (i = 0; i < NUM_GPIO; i++) {
		if (bitmask & (1 << i)) {
			ichgpio_ctrl_rout(i, rout);
		}
	}

	GPIO_DEBUG("%s: set intr mode for pin 0x%08x to 0x%08x\n",
		   __FUNCTION__, bitmask, mode);

	return 0;
}

int
ich_gpio_intr_register_handler(int port, int bitnum,
			       void (*handler)(int irq, void *dev_id,
					       struct pt_regs *regs),
			       void *dev_id)
{
        struct ichgpio_info *info = &ichgpio_info;
	unsigned long flags;

	PORT_CHECK(port);
	BITNUM_CHECK(bitnum);

	if (dev_id == NULL) {
		GPIO_DEBUG("%s: dev_id must be specified, cannot be NULL",
			   __FUNCTION__);
		return -EINVAL;
	}

	if (!(GPIO_MASK_INTERRUPTIBLE & (1 << bitnum))) {
		GPIO_DEBUG("%s: specified pin cannot cause interrupt",
			   __FUNCTION__);
		return -EINVAL;
	}

	spin_lock_irqsave(&info->reg_lock, flags);

	if (info->intr_registered & (1 << bitnum)) {
		GPIO_DEBUG("%s: attempt to register intr handler which has"
			   " already another handler registered",
			   __FUNCTION__);
		spin_unlock_irqrestore(&info->reg_lock, flags);
		return -EBUSY;
	}

	info->gpio[bitnum].intr_handler = handler;
	info->gpio[bitnum].intr_dev_id = dev_id;
	info->intr_registered |= (1 << bitnum);

	spin_unlock_irqrestore(&info->reg_lock, flags);

	GPIO_DEBUG("%s: registered handler(0x%08x) for pin %d, "
		   "dev_id = 0x%08x\n", __FUNCTION__, (unsigned int)handler,
		   bitnum, (unsigned int)dev_id);
	
	return 0;
}

int
ich_gpio_intr_unregister_handler(int port, int bitnum, void *dev_id)
{
        struct ichgpio_info *info = &ichgpio_info;
	unsigned long flags;

	PORT_CHECK(port);
	BITNUM_CHECK(bitnum);

	spin_lock_irqsave(&info->reg_lock, flags);

	if (((info->intr_registered & (1 << bitnum)) == 0) ||
	    (info->gpio[bitnum].intr_dev_id != dev_id)) {
		GPIO_DEBUG("%s: specified intr_handler is not registered",
			   __FUNCTION__);
		spin_unlock_irqrestore(&info->reg_lock, flags);
		return -EINVAL;
	}

	ich_gpio_intr_mask_bit(port, bitnum);

	info->intr_registered &= ~(1 << bitnum);
	info->gpio[bitnum].intr_dev_id = NULL;

	spin_unlock_irqrestore(&info->reg_lock, flags);

	GPIO_DEBUG("%s: unregistered handler for pin %d (dev_id = 0x%08x)\n",
		   __FUNCTION__, bitnum, (unsigned int)dev_id);

	return 0;
}

int
ich_gpio_intr_mask_bit(int port, int bitnum)
{
	BITNUM_CHECK(bitnum);
	return ich_gpio_intr_mask(port, (1 << bitnum));
}

int
ich_gpio_intr_unmask_bit(int port, int bitnum)
{
	BITNUM_CHECK(bitnum);
	return ich_gpio_intr_unmask(port, (1 << bitnum));
}

int
ich_gpio_intr_mask(int port, __u32 bitmask)
{
	PORT_CHECK(port);

	GPIO_DEBUG("%s: intr mask for 0x%08x\n", __FUNCTION__, bitmask);

	ichgpio_disable_intr((__u16)bitmask);

	return 0;
}

int
ich_gpio_intr_unmask(int port, __u32 bitmask)
{
#if 0  /* forbid unmask pin(s) which is not registered */ 
        struct ichgpio_info *info = &ichgpio_info;
	unsigned long flags;

	spin_lock_irqsave(&info->reg_lock, flags);

	bitmask &= info->intr_registered;

	spin_unlock_irqrestore(&info->reg_lock, flags);
#endif

	PORT_CHECK(port);

	if (bitmask == 0) {
		GPIO_DEBUG("%s: there's no pin that can unmask intr\n",
			   __FUNCTION__);
		return -EINVAL;
	}

	GPIO_DEBUG("%s: intr unmask for 0x%08x\n", __FUNCTION__, bitmask);

	ichgpio_enable_intr((__u16)bitmask);

	return 0;
}

int
ich_gpio_intr_get_status_bit(int port, int bitnum)
{
	__u32 status;
	int ret;

	BITNUM_CHECK(bitnum);
	
	ret = ich_gpio_intr_get_status(port, &status);

	if (ret == 0) {
		return ((status & (1 << bitnum)) != 0);
	} else {
		return ret;
	}
}

int
ich_gpio_intr_get_status(int port, __u32 *status)
{
	PORT_CHECK(port);

	*status = (__u32)ichgpio_int_status();

	return 0;
}

int
ich_gpio_intr_clear_bit(int port, int bitnum)
{
	BITNUM_CHECK(bitnum);
	return ich_gpio_intr_clear(port, (1 << bitnum));
}

int
ich_gpio_intr_clear(int port, __u32 bitmask)
{
	PORT_CHECK(port);

	ichgpio_clear_intr((__u16)bitmask);

	GPIO_DEBUG("%s: intr clear for 0x%08x\n", __FUNCTION__, bitmask);

	return 0;
}


EXPORT_SYMBOL(ich_gpio_register_bit);
EXPORT_SYMBOL(ich_gpio_unregister_bit);
EXPORT_SYMBOL(ich_gpio_register);
EXPORT_SYMBOL(ich_gpio_unregister);
EXPORT_SYMBOL(ich_gpio_unregister_by_name);
EXPORT_SYMBOL(ich_gpio_set_mode_bit);
EXPORT_SYMBOL(ich_gpio_set_mode);

EXPORT_SYMBOL(ich_gpio_get_data_bit);
EXPORT_SYMBOL(ich_gpio_set_data_bit);
EXPORT_SYMBOL(ich_gpio_get_data);
EXPORT_SYMBOL(ich_gpio_set_data);

//EXPORT_SYMBOL(ich_gpio_intr_enable);
//EXPORT_SYMBOL(ich_gpio_intr_disable);
EXPORT_SYMBOL(ich_gpio_intr_set_mode_bit);
EXPORT_SYMBOL(ich_gpio_intr_set_mode);
EXPORT_SYMBOL(ich_gpio_intr_register_handler);
EXPORT_SYMBOL(ich_gpio_intr_unregister_handler);
EXPORT_SYMBOL(ich_gpio_intr_mask_bit);
EXPORT_SYMBOL(ich_gpio_intr_unmask_bit);
EXPORT_SYMBOL(ich_gpio_intr_mask);
EXPORT_SYMBOL(ich_gpio_intr_unmask);
EXPORT_SYMBOL(ich_gpio_intr_get_status_bit);
EXPORT_SYMBOL(ich_gpio_intr_get_status);
EXPORT_SYMBOL(ich_gpio_intr_clear_bit);
EXPORT_SYMBOL(ich_gpio_intr_clear);

module_init(ichgpio_init);
module_exit(ichgpio_cleanup);

EXPORT_SYMBOL(ichgpio_init);
