/*
 *  snsc_mpu210_kpad.c -- MPU-210 keypad driver
 *
 */
/*
 *  Copyright 2002,2003 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/module.h>
#include <linux/version.h>

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/poll.h>
#include <linux/wrapper.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/ioport.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>
#include <linux/init.h>
#include <asm/delay.h>
#include <linux/spinlock.h>

#include <linux/snsc_kpad.h>
#include <linux/snsc_mpu210_siogpio.h>


#define KEYCODE( row, col)	( ((col)<<4) + (row) + 1)
#define KBUP			0x80
#define INTVL_TIMER 		3
#define INTVL_CHAT_DELAY 	50

#define GPIO_INT_BITNUM         0
#define GPIO_ROW_MASK           0xff
#define GPIO_COL_MASK           0xff
#define GPIO_NAME               "keypad"

#define MPU210_KPAD_SCAN_DELAY	10	/* us */
#define MPU210_KPAD_MINOR	-1	/* dynamic allocation */
#define MPU210_KPAD_DEVNAME	"mpu210_kpad"
#define MPU210_KPAD_ROWS	8
#define MPU210_KPAD_COLS	8


typedef struct _SCANINFO {
	char	lastkey[MPU210_KPAD_ROWS];
} SCANINFO;

static SCANINFO scaninfo; 

static struct timer_list kpad_timer = {
	function:	NULL
};

static int minor = MPU210_KPAD_MINOR;
static int cdelay = INTVL_CHAT_DELAY;

#ifdef MODULE
MODULE_AUTHOR("Sony NSC");
MODULE_DESCRIPTION("MPU-210 keypad driver");

#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif /* MODULE_LICENSE */

MODULE_PARM(cdelay, "i");
MODULE_PARM_DESC(cdelay, "interval to avoid chattering");
#endif /* MODULE */


/* ======================================================================= *
 * prototype declaration                                                   *
 * ======================================================================= */

static int  mpu210_kpad_init(void);
static void mpu210_kpad_final(void);
static int  mpu210_kpad_init_module(void);
static void mpu210_kpad_cleanup_module(void);
static int  mpu210_kpad_open(struct keypad_dev *dev);
static int  mpu210_kpad_release(struct keypad_dev *dev);
static void mpu210_kpad_intr_hdlr(int irq, void *devid, struct pt_regs *regs);
static void mpu210_kpad_tasklet(unsigned long arg);
static void mpu210_kpad_key_scan(unsigned long arg);

/* ======================================================================= *
 * Sony NSC keypad driver interface                                        *
 * ======================================================================= */

static DECLARE_TASKLET(mpu210_kpad_tasklet_data,
		       mpu210_kpad_tasklet,
		       (unsigned long)&scaninfo);

static struct keypad_dev mpu210_kpad_dev = {
	name:		MPU210_KPAD_DEVNAME,
	minor:		-1,
	open:		mpu210_kpad_open,
	release:	mpu210_kpad_release,
	ioctl:		NULL,
};

static int mpu210_kpad_init(void)
{
	int  res;
	int i;
	SCANINFO *sinfo = &scaninfo;

        for (i = 0; i < 8; i++) {
                sinfo->lastkey[i]  = 0x00;
        }

	/* register myself as keypad driver */
	mpu210_kpad_dev.minor = minor;
	res = register_keypad(&mpu210_kpad_dev);
	if (res < 0) {
		printk(KERN_WARNING "register_keypad() failed (err=%d)\n",
		      -res);
		goto err1;
	}

	mpu210_siogpio_register(MPU210_SIOGPIO_PORT_PIN2, GPIO_ROW_MASK,
				GPIO_NAME);

	mpu210_siogpio_set_mode(MPU210_SIOGPIO_PORT_PIN2, GPIO_ROW_MASK,
				MPU210_SIOGPIO_MODE_OUT_LOW |
				MPU210_SIOGPIO_MODE_NOINVERT |
				MPU210_SIOGPIO_MODE_NOPULLUP);

	mpu210_siogpio_register(MPU210_SIOGPIO_PORT_PIN3, GPIO_COL_MASK,
				GPIO_NAME);

	mpu210_siogpio_set_mode(MPU210_SIOGPIO_PORT_PIN3, GPIO_COL_MASK,
				MPU210_SIOGPIO_MODE_IN |
				MPU210_SIOGPIO_MODE_NOINVERT |
				MPU210_SIOGPIO_MODE_NOPULLUP);

	mpu210_siogpio_register_bit(MPU210_SIOGPIO_PORT_PIN1, GPIO_INT_BITNUM,
				    GPIO_NAME);

	mpu210_siogpio_set_mode_bit(MPU210_SIOGPIO_PORT_PIN1, GPIO_INT_BITNUM,
				    MPU210_SIOGPIO_MODE_IN |
				    MPU210_SIOGPIO_MODE_NOINVERT |
				    MPU210_SIOGPIO_MODE_NOPULLUP);

	/* register interrupt handler */
	res = mpu210_siogpio_intr_register_handler(MPU210_SIOGPIO_PORT_PIN1,
						   GPIO_INT_BITNUM,
						   mpu210_kpad_intr_hdlr,
						   &scaninfo);

	if (res < 0) {
		goto err2;
	}

	mpu210_siogpio_intr_set_mode_bit(MPU210_SIOGPIO_PORT_PIN1,
					 GPIO_INT_BITNUM,
					 MPU210_SIOGPIO_INTR_EDGE_HIGH);

	mpu210_siogpio_intr_unmask_bit(MPU210_SIOGPIO_PORT_PIN1,
				       GPIO_INT_BITNUM);

	return 0;

err2:
	unregister_keypad(&mpu210_kpad_dev);
err1:
	return res;
}

static void mpu210_kpad_final(void)
{
	mpu210_siogpio_intr_mask_bit(MPU210_SIOGPIO_PORT_PIN1, GPIO_INT_BITNUM);

	mpu210_siogpio_intr_unregister_handler(MPU210_SIOGPIO_PORT_PIN1,
					       GPIO_INT_BITNUM,
					       &scaninfo);

	mpu210_siogpio_unregister_by_name(GPIO_NAME);

	unregister_keypad(&mpu210_kpad_dev);
}

/* ------------------------ *
 * module                   *
 * ------------------------ */

static int mpu210_kpad_init_module(void)
{
	int res;

	res = mpu210_kpad_init();

	return res;
}

static void mpu210_kpad_cleanup_module(void)
{
	mpu210_kpad_final();
}

module_init(mpu210_kpad_init_module);
module_exit(mpu210_kpad_cleanup_module);

/* ------------------------ *
 * file operation           *
 * ------------------------ */

static int mpu210_kpad_open(struct keypad_dev *dev)
{
	MOD_INC_USE_COUNT;

	return 0;
}

static int mpu210_kpad_release(struct keypad_dev *dev)
{
	MOD_DEC_USE_COUNT;

	return 0;
}

/* ======================================================================= *
 * interrupt handler                                                       *
 * ======================================================================= */

static void mpu210_kpad_intr_hdlr(int irq, void *devid, struct pt_regs *regs)
{
	/* disable interrupt */
	mpu210_siogpio_intr_mask_bit(MPU210_SIOGPIO_PORT_PIN1, GPIO_INT_BITNUM);

	tasklet_schedule(&mpu210_kpad_tasklet_data);
}

static void mpu210_kpad_tasklet(unsigned long arg)
{
	udelay(cdelay);
	mpu210_kpad_key_scan(arg);
}

/* ======================================================================= *
 * internal functions                                                      *
 * ======================================================================= */

/* ------------------------ *
 * key detection            *
 * ------------------------ */

static void mpu210_kpad_key_scan(unsigned long arg)
{
	SCANINFO *sinfo = (SCANINFO *)arg;
	int  i;
	int push_flg;

	push_flg = 0;
	for (i = 0; i < MPU210_KPAD_ROWS; i++) {
		char diffkey;
		char key[MPU210_KPAD_ROWS];

		mpu210_siogpio_set_data(MPU210_SIOGPIO_PORT_PIN2,
					GPIO_ROW_MASK,
					~(1 << i));
		udelay(MPU210_KPAD_SCAN_DELAY);
		mpu210_siogpio_get_data(MPU210_SIOGPIO_PORT_PIN3,
					GPIO_COL_MASK,
					&(key[i]));
		key[i] = ~key[i];

		if ( key[i] ) {
			push_flg = 1;
		}

		diffkey = key[i] ^ sinfo->lastkey[i];
		if ( diffkey ) {
			unsigned char downkey;
			unsigned char upkey;
			int j;

			downkey = key[i] & diffkey;
			upkey = ~key[i] & diffkey;
			for (j = 0; j < MPU210_KPAD_COLS; j++) {
				if ( (1 << j) & downkey ) {
					keypad_event(&mpu210_kpad_dev,
						     KEYCODE(i, j));
				}
				else if ( (1 << j) & upkey ) {
					keypad_event(&mpu210_kpad_dev,
						     KEYCODE(i, j) | KBUP);
				}
			}
		}
		sinfo->lastkey[i]  = key[i];
	}

        if( push_flg ){

                /* Timer Set */
                kpad_timer.expires = jiffies + INTVL_TIMER;
                kpad_timer.data = arg;
                kpad_timer.function = mpu210_kpad_key_scan;
                if(!(timer_pending(&kpad_timer))){
                        add_timer(&kpad_timer);
                }

        } else {
		mpu210_siogpio_set_data(MPU210_SIOGPIO_PORT_PIN2,
					GPIO_ROW_MASK, 0);

		/* KeyPad Interrput Enable */
		mpu210_siogpio_intr_unmask_bit(MPU210_SIOGPIO_PORT_PIN1,
					       GPIO_INT_BITNUM);
	}
}
