/*
 *  snsc_mpu210_siogpio.c : MPU-210 SuperIO GPIO driver
 *
 *  Copyright 2002,2003 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifndef __KERNEL__
#error "this module must be compiled with -D__KERNEL__"
#endif

#include <linux/module.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/spinlock.h>
#include <linux/irq.h>
#ifdef DEBUG
#include <linux/proc_fs.h>
#endif /* DEBUG */
#include <asm/errno.h>

#include <asm/io.h>
#include <asm/it8712.h>
#include <asm-mips/it8172/it8172.h>
#include <asm-mips/it8172/it8172_int.h>
#include <linux/snsc_mpu210_siogpio.h>


/*
 * port / bitnum assignments:
 *
 *    +------+-------------------------------------------------------+
 *    |      |                        bitnum                         |
 *    |      +------+------+------+------+------+------+------+------+
 *    |      |   7  |   6  |   5  |   4  |   3  |   2  |   1  |   0  |
 *    +----+-+------+------+------+------+------+------+------+------+
 *    |    |1|GPIO17|GPIO16|GPIO15|GPIO14|GPIO13|GPIO12|GPIO11|GPIO10|
 *    |    +-+------+------+------+------+------+------+------+------+
 *    |    |2|GPIO27|GPIO26|GPIO25|GPIO24|GPIO23|GPIO22|GPIO21|GPIO20|
 *    |    +-+------+------+------+------+------+------+------+------+
 *    |Port|3|GPIO37|GPIO36|GPIO35|GPIO34|GPIO33|GPIO32|GPIO31|GPIO30|
 *    |    +-+------+------+------+------+------+------+------+------+
 *    |    |4|GPIO47|GPIO46|GPIO45|GPIO44|GPIO43|GPIO42|GPIO41|GPIO40|
 *    |    +-+------+------+------+------+------+------+------+------+
 *    |    |5|      |      |GPIO55|GPIO54|GPIO53|GPIO52|GPIO51|GPIO50|
 *    +----+-+------+------+------+------+------+------+------+------+
 *
 *
 * GPIO control API implementation:
 *
 *   There are five states that each pin should take:
 *
 *         STATE_INITIAL - initial state (being not used)
 *
 *             + transitted from any states to here, by calling
 *               gpio_unregister() or gpio_unregister_bit()
 *             + the only acceptable function calls at this state are
 *               gpio_register() or gpio_register_bit(), which
 *               transits its state to the STATE_REGISTERED
 *
 *         STATE_REGISTERED - registered _BUT_ mode has not been set yet
 *
 *             + transitted only from STATE_INITIAL, by calling
 *               gpio_register() or gpio_register_bit()
 *             + the only acceptable function calls at this state are
 *               gpio_set_mode() or gpio_set_mode_bit(), which
 *               transits its state to the STATE_MODE_SET
 *             + returns to STATE_INITIAL when gpio_unregister() or
 *               gpio_unregister_bit() is called
 *
 *         STATE_MODE_SET - registered _AND_ the mode is set
 *
 *             + get/set data is available here
 *             + transitted only from STATE_REGISTERED, by calling
 *               gpio_set_mode() or gpio_set_mode_bit()
 *             + transitted to STATE_INTR_REGISTERED, by calling
 *               gpio_intr_register_handler()
 *             + returns to STATE_INITIAL when gpio_unregister() or
 *               gpio_unregister_bit() is called
 *
 *         STATE_INTR_REGISTERED - interrupt handler is set _BUT_
 *                                 interrupt mode has not been set yet
 *
 *             + transitted only from STATE_MODE_SET, by calling
 *               gpio_intr_register_hanlder()
 *             + transitted to STATE_INTR_MODE_SET, by calling
 *               gpio_intr_set_mode() or gpio_intr_set_mode_bit()
 *             + returns to STATE_INITIAL when gpio_unregister() or
 *               gpio_unregister_bit() is called
 *
 *         STATE_INTR_MODE_SET - interrupt handler is set _AND_
 *                               interrupt mode is set
 *
 *             + mask/unmask or clear interrupt is available
 *             + transitted only from STATE_INTR_REGISTERED, by calling
 *               gpio_intr_set_mode() or gpio_intr_set_mode_bit()
 *             + returns to STATE_INITIAL when gpio_unregister() or
 *               gpio_unregister_bit() is called
 *
 *
 *   The advantage of such state management is that it decreases
 *  ambiguous behavior, especially in case of illegal function call.
 *
 *
 *
 * (state transition diagram)
 *
 *         +-------+             unregister()
 *         |INITIAL| <----+---------+------------+------------+
 *         |   0x01|      |         |            |            |
 *         +-------+      |         |            |            |
 *             |    +----------+    |            |            |
 *             |    |REGISTERED|    |            |            |
 *  register() +--> |      0x02|    |            |            |
 *                  +----------+    |            |            |
 *                        |    +--------+        |            |
 *                        |    |MODE_SET|   intr_unregister_handler()
 *             set_mode() +--> |    0x04| <------+------------+
 *                             +--------+        |            |
 *                                  |    +---------------+    |
 *                                  |    |INTR_REGISTERED|    |
 *          intr_register_handler() +--> |           0x08|    |
 *                                       +---------------+    |
 *                                               |     +-------------+
 *                                               |     |INTR_MODE_SET|
 *                               intr_set_mode() +---> |         0x10|
 *                                                     +-------------+
 *
 * (valid state for function call)
 *
 *      function name                  01 02 04 08 10 change state
 *      ------------------------------ -- -- -- -- -- ------------
 *      gpio_intr_enable()              *  *  *  *  *
 *      gpio_intr_disable()             *  *  *  *  *
 *      gpio_register_bit()             *                 01 => 02
 *      gpio_register()                 *                 01 => 02
 *      gpio_unregister_bit()              *  *  *  *     xx => 01
 *      gpio_unregister()                  *  *  *  *     xx => 01
 *      gpio_unregister_by_name()          *  *  *  *     xx => 01
 *      gpio_set_mode_bit()                *  *  *  *     02 => 04
 *      gpio_set_mode()                    *  *  *  *     02 => 04
 *      gpio_get_data_bit()                   *  *  *
 *      gpio_set_data_bit()                   *  *  *
 *      gpio_get_data()                       *  *  *
 *      gpio_set_data()                       *  *  *
 *      gpio_intr_register_handler()          *           04 => 08
 *      gpio_intr_unregister_handler()           *  *  08,10 => 04
 *      gpio_intr_set_mode_bit()                 *  *     08 => 10
 *      gpio_intr_set_mode()                     *  *     08 => 10
 *      gpio_intr_mask_bit()                        *
 *      gpio_intr_unmask_bit()                      *
 *      gpio_intr_mask()                            *
 *      gpio_intr_unmask()                          *
 *      gpio_intr_get_status_bit()                  *
 *      gpio_intr_get_status()                      *
 *      gpio_intr_clear_bit()                       *
 *      gpio_intr_clear()                           *
 *  
 *  when state is not valid, error is returned.
 */

#define MPU210_SIOGPIO_PORT_NUM   5
#define MPU210_SIOGPIO_BIT_NUM    8

#define CAP_NONE                  0x0
#define CAP_READABLE              0x1
#define CAP_WRITABLE              0x2

#define STATE_INITIAL             0x01
#define STATE_REGISTERED          0x02
#define STATE_MODE_SET            0x04
#define STATE_INTR_REGISTERED     0x08
#define STATE_INTR_MODE_SET       0x10

#define OWNER_NAME_LEN            16

#define MPU210_SIOGPIO_SIMPLE_IO  0x810 /* 0x14000810 - 0x14000814 */
#define MPU210_SIOGPIO_DEBOUNCE   0x800 /* 0x14000800 - 0x14000804 */

#ifdef DEBUG

#define SIOGPIO_LOCK()	do {                                                   \
				printk("+%s: %d\n", __FUNCTION__, __LINE__);   \
				spin_lock_irqsave(&siogpio_lock, flags);       \
			} while (0);

#define SIOGPIO_UNLOCK() do {                                                  \
				printk("-%s: %d\n", __FUNCTION__, __LINE__);   \
				spin_unlock_irqrestore(&siogpio_lock, flags);  \
			 } while (0);

#else /* DEBUG */

#define SIOGPIO_LOCK()            spin_lock_irqsave(&siogpio_lock, flags)
#define SIOGPIO_UNLOCK()          spin_unlock_irqrestore(&siogpio_lock, flags)

#endif /* DEBUG */


static spinlock_t siogpio_lock;

static struct mpu210_siogpio_t {
	struct {
		__u8 capability;
			/*
			 * logical OR of GPIO_CAP_XXX
			 */

		__u8 state;
			/*
			 * current state of the pin.
			 * one of STATE_XXX
			 */

		__u8 org_value;
			/*
			 * buffer for saving original value
			 *
			 *   bit 0: multi-function pin selection
			 *   bit 1: polarity
			 *   bit 2: internal pull-up enable
			 *   bit 3: simple I/O enable
			 *   bit 4: output enable
			 *   bit 5: panel button de-bounce enable
			 *   bit 6: interrupt trigger mode (edge/level)
			 *   bit 7: interrupt active level (positive/negative)
			 */

		char name[OWNER_NAME_LEN];
			/*
			 * specified GPIO pin name
			 */

		void (*handler)(int irq, void *dev_id, struct pt_regs *regs);
			/*
			 * user's interrupt handler
			 */

		void *dev_id;
			/*
			 * argument for user's interrupt handler
			 */

		int irq;
			/*
			 * IRQ number for the pin
			 */

	} bit[MPU210_SIOGPIO_BIT_NUM]; 

} gpio[MPU210_SIOGPIO_PORT_NUM];



static inline int get_gpio_register(char LdnNumber, char Index, int bitnum)
{
	char data = LPCGetConfig(LdnNumber, Index);
	return ((1 << bitnum) & data) ? 1 : 0;
}

static void set_gpio_register(char LdnNumber, char Index, int bitnum, int yes)
{
	char data;

	data = LPCGetConfig(LdnNumber, Index);

	if (yes) {
	    data |= (1 << bitnum);
	}
	else {
	    data &= ~(1 << bitnum);
	}
	LPCSetConfig(LdnNumber, Index, data);
}

static inline int get_multi_function(__u8 port, int bitnum)
{
	return get_gpio_register(0x07, 0x25 + port, bitnum);
}

static inline int get_polarity(__u8 port, int bitnum)
{
	return get_gpio_register(0x07, 0xb0 + port, bitnum);
}

static inline int get_pullup(__u8 port, int bitnum)
{
	return get_gpio_register(0x07, 0xb8 + port, bitnum);
}

static inline int get_function(__u8 port, int bitnum)
{
	return get_gpio_register(0x07, 0xc0 + port, bitnum);
}

static inline int get_io(__u8 port, int bitnum)
{
	return get_gpio_register(0x07, 0xc8 + port, bitnum);
}

static inline int get_debounce(__u8 port, int bitnum)
{
	return get_gpio_register(0x07, 0xd1 + port, bitnum);
}

static inline void set_multi_function(__u8 port, int bitnum, int is_gpio)
{
	set_gpio_register(0x07, 0x25 + port, bitnum, is_gpio);
}

static inline void set_polarity(__u8 port, int bitnum, int is_invert)
{
	set_gpio_register(0x07, 0xb0 + port, bitnum, is_invert);
}

static inline void set_pullup(__u8 port, int bitnum, int is_pullup)
{
	set_gpio_register(0x07, 0xb8 + port, bitnum, is_pullup);
}

static inline void set_function(__u8 port, int bitnum, int is_simpleio)
{
	set_gpio_register(0x07, 0xc0 + port, bitnum, is_simpleio);
}

static inline void set_io(__u8 port, int bitnum, int is_output)
{
	set_gpio_register(0x07, 0xc8 + port, bitnum, is_output);
}

static inline void set_debounce(__u8 port, int bitnum, int enable)
{
	set_gpio_register(0x07, 0xd1 + port, bitnum, enable);
}

static inline void set_data(__u16 base, __u8 port, __u8 value)
{
	outb(value, IT8172_PCI_IO_BASE + base + port);
}

static inline __u8 get_data(__u16 base, __u8 port)
{
	return (__u8) inb(IT8172_PCI_IO_BASE + base + port);
}

static void set_io_base(__u16 base)
{
	char LdnNumber;
	char Index;

	/* set Base Address Register */
	LdnNumber = 0x07;
	Index = 0x62;
	LPCSetConfig(LdnNumber, Index, (base >> 8) & 0xff);
	Index = 0x63;
	LPCSetConfig(LdnNumber, Index, base & 0xff);
}

static void set_debounce_base(__u16 base)
{
	char LdnNumber;
	char Index;

	/* set Base Address Register */
	LdnNumber = 0x07;
	Index = 0x64;
	LPCSetConfig(LdnNumber, Index, (base >> 8) & 0xff);
	Index = 0x65;
	LPCSetConfig(LdnNumber, Index, base & 0xff);
}

static inline int is_available(int port, int bitnum)
{
	return (CAP_NONE == gpio[port].bit[bitnum].capability) ? 0 : 1;
}

static inline int is_in_state(int port, int bitnum, int state)
{
	return (state & gpio[port].bit[bitnum].state) ? 1 : 0;
}

static inline int is_out(int port, int bitnum)
{
	return  ( get_io(port, bitnum) ) ? 1 : 0;
}

static inline int is_port_valid(int port)
{
	return ((0 > port) || (MPU210_SIOGPIO_PORT_NUM <= port)) ? 0 : 1;
}

static inline int is_bitnum_valid(int bitnum)
{
	return ((0 > bitnum) || (MPU210_SIOGPIO_BIT_NUM <= bitnum)) ? 0 : 1;
}

static int is_port_bitnum_valid(int port, int bitnum)
{
	if ((0 > port) || (MPU210_SIOGPIO_PORT_NUM <= port)) {
		return 0;
	}
	if ((0 > bitnum) || (MPU210_SIOGPIO_BIT_NUM <= bitnum)) {
		return 0;
	}
	if (0 == is_available(port, bitnum)) {
		return 0;
	}
	return 1;
}

static int is_port_bitmask_valid(int port, __u8 bitmask)
{
	int j;

	if ((0 > port) || (MPU210_SIOGPIO_PORT_NUM <= port)) {
		return 0;
	}
	for (j=0; j<MPU210_SIOGPIO_BIT_NUM; j++) {
		if ((bitmask >> j) & 1) {
			if (0 == is_available(port, j)) {
				return 0;
			}
		}
	}
	return 1;
}


/*
 * interrupt related functions
 */

static spinlock_t irq_table_lock;

static struct irq_table_t {
	int irq;	/* must be 3-7, 9-12, 14, 15 (see p.47) */
	int use;	/* 0: free     1: in use */

} irq_table[] = {
	{IT8172_SERIRQ_6,  0},
	{IT8172_SERIRQ_7,  0},
	{IT8172_SERIRQ_11, 0},
	{-1, 0}			/* terminator */
};

static int siogpio_acquire_irq(void)
{
	int index = 0;
	unsigned long flags;

	spin_lock_irqsave(&irq_table_lock, flags);

	while (-1 != irq_table[index].irq) {
		if (0 == irq_table[index].use) {
			irq_table[index].use = 1;
			spin_unlock_irqrestore(&irq_table_lock, flags);
			return irq_table[index].irq;
		}
		index ++;
	}

	spin_unlock_irqrestore(&irq_table_lock, flags);
	return -1;
}

static void siogpio_release_irq(int irq)
{
	int index = 0;
	unsigned long flags;

	spin_lock_irqsave(&irq_table_lock, flags);

	while (-1 != irq_table[index].irq) {
		if (irq == irq_table[index].irq) {
			irq_table[index].use = 0;
			spin_unlock_irqrestore(&irq_table_lock, flags);
			return;
		}
		index ++;
	}

	spin_unlock_irqrestore(&irq_table_lock, flags);
}


#ifdef DEBUG
/*
 * proc entry
 */

static int siogpio_read_proc(char *buf, char **start, off_t fpos,
			     int length, int *eof, void *data)
{
	int j;
	int k;
	int len = 0;

	len += sprintf(buf + len, "port bit state name            irq\n");

	for (k=0; k<MPU210_SIOGPIO_PORT_NUM; k++) {
		len += sprintf(buf + len, "---- --- ----- --------------- ----\n");
		for (j=0; j<MPU210_SIOGPIO_BIT_NUM; j++) {
			int irq;

			if (0 == is_available(k, j)) {
				continue;
			}

			irq = gpio[k].bit[j].irq;

			if (0 == j) {
				len += sprintf(buf + len, "%4d", (k + 1));
			}
			else {
				len += sprintf(buf + len, "    ");
			}

			len += sprintf(buf + len, " %3d  0x%02x %-15s ",
					j,
					gpio[k].bit[j].state,
					gpio[k].bit[j].name);

			if (0 < irq) {
				len += sprintf(buf + len, "%d", irq);
			}

			len += sprintf(buf + len, "\n");
		}
	}

	if (fpos >= len) {
		*start = buf;
		*eof = 1;
		return 0;
	}
	*start = buf + fpos;
	if ((len -= fpos) > length)
		return length;
	*eof = 1;
	return len;
}
#endif /* DEBUG */


/*
 * SNSC common GPIO driver interfaces
 */

int mpu210_siogpio_register_bit(int port, int bitnum, const char *name)
{
	if (0 == is_bitnum_valid(bitnum)) {
		return -EINVAL;
	}
	return mpu210_siogpio_register(port, (1 << bitnum), name);
}

int mpu210_siogpio_unregister_bit(int port, int bitnum)
{
	if (0 == is_bitnum_valid(bitnum)) {
		return -EINVAL;
	}
	return mpu210_siogpio_unregister(port, (1 << bitnum));
}

int mpu210_siogpio_register(int port, __u8 bitmask, const char *name)
{
	int j;
	unsigned long flags;

	if (0 == is_port_bitmask_valid(port, bitmask)) {
		return -EINVAL;
	}
	if (NULL == name) {
		return -EINVAL;
	}

	SIOGPIO_LOCK();

	for (j=0; j<MPU210_SIOGPIO_BIT_NUM; j++) {
		if ((bitmask >> j) & 1) {
			/* state check */
			if (0 == is_in_state(port, j, STATE_INITIAL)) {
				/* has already been registered */
				SIOGPIO_UNLOCK();
				return -EBUSY;
			}
		}
	}

	for (j=0; j<MPU210_SIOGPIO_BIT_NUM; j++) {
		if ((bitmask >> j) & 1) {
			/* change state: INITIAL => REGISTERED */
			gpio[port].bit[j].state = STATE_REGISTERED;

			/* copy name */
			memset(gpio[port].bit[j].name, '\0', OWNER_NAME_LEN);
			strncpy(gpio[port].bit[j].name, name, OWNER_NAME_LEN);
			gpio[port].bit[j].name[OWNER_NAME_LEN - 1] = '\0';

			/* save original values */
			gpio[port].bit[j].org_value = 0;
			gpio[port].bit[j].org_value |= get_debounce(port, j);
			gpio[port].bit[j].org_value <<= 1;
			gpio[port].bit[j].org_value |= get_io(port, j);
			gpio[port].bit[j].org_value <<= 1;
			gpio[port].bit[j].org_value |= get_function(port, j);
			gpio[port].bit[j].org_value <<= 1;
			gpio[port].bit[j].org_value |= get_pullup(port, j);
			gpio[port].bit[j].org_value <<= 1;
			gpio[port].bit[j].org_value |= get_polarity(port, j);
			gpio[port].bit[j].org_value <<= 1;
			gpio[port].bit[j].org_value |= get_multi_function(port, j);
		}
	}

	SIOGPIO_UNLOCK();
	return 0;
}

int mpu210_siogpio_unregister(int port, __u8 bitmask)
{
	int j;
	unsigned long flags;

	if (0 == is_port_bitmask_valid(port, bitmask)) {
		return -EINVAL;
	}

	SIOGPIO_LOCK();

	for (j=0; j<MPU210_SIOGPIO_BIT_NUM; j++) {
		if ((bitmask >> j) & 1) {
			/* state check */
			if ( is_in_state(port, j, STATE_INITIAL) ) {
				/* not registered */
				SIOGPIO_UNLOCK();
				return -EINVAL;
			}
		}
	}

	for (j=0; j<MPU210_SIOGPIO_BIT_NUM; j++) {
		if ((bitmask >> j) & 1) {
			int value;

			if ( is_in_state(port, j, STATE_INTR_REGISTERED |
						  STATE_INTR_MODE_SET) ) {
				/* call intr_unregister_handler */
				mpu210_siogpio_intr_unregister_handler(port, j, gpio[port].bit[j].dev_id);
			}

			/* change state: => INITIAL */
			gpio[port].bit[j].state = STATE_INITIAL;

			/* reset name */
			memset(gpio[port].bit[j].name, '\0', OWNER_NAME_LEN);

			/* restore original values */
			value = (gpio[port].bit[j].org_value >> 0) & 1;
			set_multi_function(port, j, value);
			value = (gpio[port].bit[j].org_value >> 1) & 1;
			set_polarity(port, j, value);
			value = (gpio[port].bit[j].org_value >> 2) & 1;
			set_pullup(port, j, value);
			value = (gpio[port].bit[j].org_value >> 3) & 1;
			set_function(port, j, value);
			value = (gpio[port].bit[j].org_value >> 4) & 1;
			set_io(port, j, value);
			value = (gpio[port].bit[j].org_value >> 5) & 1;
			set_debounce(port, j, value);

			gpio[port].bit[j].org_value = 0;
		}
	}

	SIOGPIO_UNLOCK();
	return 0;
}

int mpu210_siogpio_unregister_by_name(const char *name)
{
	int j;
	unsigned long flags;
	int k;
	int counter;

	if (NULL == name) {
		return -EINVAL;
	}

	SIOGPIO_LOCK();

	/* first loop: check consistency */
	counter = 0;
	for (k=0; k<MPU210_SIOGPIO_PORT_NUM; k++) {
		for (j=0; j<MPU210_SIOGPIO_BIT_NUM; j++) {
			if (0 == strncmp(gpio[k].bit[j].name, name,
						OWNER_NAME_LEN)) {

				/* state check */
				if ( is_in_state(k, j, STATE_INITIAL) ) {
					/* not registered */
					SIOGPIO_UNLOCK();
					return -EINVAL;
				}
			}
			counter ++;
		}
	}

	if (0 == counter) {
		/* no entry found */
		SIOGPIO_UNLOCK();
		return -EINVAL;
	}

	/* second loop: unregister */
	for (k=0; k<MPU210_SIOGPIO_PORT_NUM; k++) {
		for (j=0; j<MPU210_SIOGPIO_BIT_NUM; j++) {
			if (0 == strncmp(gpio[k].bit[j].name, name,
						OWNER_NAME_LEN)) {
				mpu210_siogpio_unregister_bit(k, j);
			}
		}
	}

	SIOGPIO_UNLOCK();
	return 0;
}

int mpu210_siogpio_set_mode_bit(int port, int bitnum, __u32 mode)
{
	if (0 == is_bitnum_valid(bitnum)) {
		return -EINVAL;
	}
	return mpu210_siogpio_set_mode(port, (1 << bitnum), mode);
}

int mpu210_siogpio_set_mode(int port, __u8 bitmask, __u32 mode)
{
	int j;
	unsigned long flags;

	if (0 == is_port_bitmask_valid(port, bitmask)) {
		return -EINVAL;
	}

	if ( (mode & MPU210_SIOGPIO_MODE_NOINVERT) &&
	     (mode & MPU210_SIOGPIO_MODE_INVERT) ) {
		return -EINVAL;
	}

	if ( !(mode & MPU210_SIOGPIO_MODE_NOINVERT) &&
	     !(mode & MPU210_SIOGPIO_MODE_INVERT) ) {
		return -EINVAL;
	}

	if ( (mode & MPU210_SIOGPIO_MODE_NOPULLUP) &&
	     (mode & MPU210_SIOGPIO_MODE_PULLUP) ) {
		return -EINVAL;
	}

	if ( !(mode & MPU210_SIOGPIO_MODE_NOPULLUP) &&
	     !(mode & MPU210_SIOGPIO_MODE_PULLUP) ) {
		return -EINVAL;
	}

	if ( ((mode & MPU210_SIOGPIO_MODE_OUT_HIGH) ||
	      (mode & MPU210_SIOGPIO_MODE_OUT_LOW)) &&
	     (mode & MPU210_SIOGPIO_MODE_IN) ) {
		return -EINVAL;
	}

	if ( !((mode & MPU210_SIOGPIO_MODE_OUT_HIGH) ||
	       (mode & MPU210_SIOGPIO_MODE_OUT_LOW)) &&
	     !(mode & MPU210_SIOGPIO_MODE_IN) ) {
		return -EINVAL;
	}

	SIOGPIO_LOCK();

	for (j=0; j<MPU210_SIOGPIO_BIT_NUM; j++) {
		if ((bitmask >> j) & 1) {
			int capability;

			/* state check */
			if ( is_in_state(port, j, STATE_INITIAL) ) {
				/* not registered */
				SIOGPIO_UNLOCK();
				return -EINVAL;
			}

			/* capability check */
			capability = gpio[port].bit[j].capability;

			if (MPU210_SIOGPIO_MODE_IN & mode) {
				if (0 == (CAP_READABLE & capability)) {
					SIOGPIO_UNLOCK();
					return -EINVAL;
				}
			}
			else {
				if (0 == (CAP_WRITABLE & capability)) {
					SIOGPIO_UNLOCK();
					return -EINVAL;
				}
			}
		}
	}

	for (j=0; j<MPU210_SIOGPIO_BIT_NUM; j++) {
		if ((bitmask >> j) & 1) {
			int value;

			/* change state: REGISTERED => MODE_SET */
			if (STATE_REGISTERED == gpio[port].bit[j].state) {
				gpio[port].bit[j].state = STATE_MODE_SET;
			}

			/* select multi-function as GPIO */
			value = 1;
			set_multi_function(port, j, value);

			value = (mode & MPU210_SIOGPIO_MODE_INVERT) ? 1 : 0;
			set_polarity(port, j, value);

			value = (mode & MPU210_SIOGPIO_MODE_PULLUP) ? 1 : 0;
			set_pullup(port, j, value);

			/* select Simple I/O function */
			value = 1;
			set_function(port, j, value);

			value = (mode & MPU210_SIOGPIO_MODE_IN) ? 0 : 1;
			set_io(port, j, value);

			/* disable panel button de-bounce */
			value = 0;
			set_debounce(port, j, value);

			/* initialize default OUT */
			if (mode & MPU210_SIOGPIO_MODE_OUT_HIGH) {
				mpu210_siogpio_set_data_bit(port, j, 1);
			}
			else if (mode & MPU210_SIOGPIO_MODE_OUT_LOW) {
				mpu210_siogpio_set_data_bit(port, j, 0);
			}
		}
	}

	SIOGPIO_UNLOCK();
	return 0;
}

int mpu210_siogpio_get_data_bit(int port, int bitnum)
{
	int ret;
	unsigned long flags;

	if (0 == is_port_bitnum_valid(port, bitnum)) {
		return -EINVAL;
	}
	if ( is_out(port, bitnum) ) {
		return -EINVAL;
	}

	SIOGPIO_LOCK();

	/* state check */
	if ( is_in_state(port, bitnum, STATE_INITIAL | STATE_REGISTERED) ) {
		/* no registered, or mode has not been set yet */
		return -EINVAL;
	}

	ret = ((get_data(MPU210_SIOGPIO_SIMPLE_IO, port) >> bitnum) & 1) ?
		1 : 0;

	SIOGPIO_LOCK();
	return ret;
}

int mpu210_siogpio_set_data_bit(int port, int bitnum, int value)
{
	__u8 data;
	unsigned long flags;

	if (0 == is_port_bitnum_valid(port, bitnum)) {
		return -EINVAL;
	}

	SIOGPIO_LOCK();

	/* state check */
	if ( is_in_state(port, bitnum, STATE_INITIAL | STATE_REGISTERED) ) {
		/* no registered, or mode has not been set yet */
		SIOGPIO_UNLOCK();
		return -EINVAL;
	}

	if (0 == is_out(port, bitnum)) {
		/* do nothing */
		SIOGPIO_UNLOCK();
		return 0;
	}

	data = get_data(MPU210_SIOGPIO_SIMPLE_IO, port);
	if (value) {
		data |= (1 << bitnum);
	}
	else {
		data &= ~(1 << bitnum);
	}
	set_data(MPU210_SIOGPIO_SIMPLE_IO, port, data);

	SIOGPIO_UNLOCK();
	return 0;
}

int mpu210_siogpio_get_data(int port, __u8 bitmask, __u8 *value)
{
	int j;
	unsigned long flags;

	if (0 == is_port_bitmask_valid(port, bitmask)) {
		return -EINVAL;
	}
	if (NULL == value) {
		return -EINVAL;
	}

	SIOGPIO_LOCK();

	for (j=0; j<MPU210_SIOGPIO_BIT_NUM; j++) {
		if ((bitmask >> j) & 1) {
			if ( is_out(port, j) ) {
				SIOGPIO_UNLOCK();
				return -EINVAL;
			}

			/* state check */
			if ( is_in_state(port, j, STATE_INITIAL |
						  STATE_REGISTERED)) {
				/* no registered, or mode not set */
				SIOGPIO_UNLOCK();
				return -EINVAL;
			}
		}
	}

	*value = get_data(MPU210_SIOGPIO_SIMPLE_IO, port) & bitmask;

	SIOGPIO_UNLOCK();
	return 0;
}

int mpu210_siogpio_set_data(int port, __u8 bitmask, __u8 value)
{
	int j;
	unsigned long flags;
	__u8 data;

	if (0 == is_port_bitmask_valid(port, bitmask)) {
		return -EINVAL;
	}

	SIOGPIO_LOCK();

	for (j=0; j<MPU210_SIOGPIO_BIT_NUM; j++) {
		if ((bitmask >> j) & 1) {

			/* state check */
			if ( is_in_state(port, j, STATE_INITIAL |
						  STATE_REGISTERED)) {
				/* no registered, or mode not set */
				SIOGPIO_UNLOCK();
				return -EINVAL;
			}

			if (0 == is_out(port, j)) {
				/* clear mask for doing nothing */
				bitmask &= ~(1 << j);
			}
		}
	}

	data = get_data(MPU210_SIOGPIO_SIMPLE_IO, port);
	data &= ~bitmask;
	data |= value;
	set_data(MPU210_SIOGPIO_SIMPLE_IO, port, data);

	SIOGPIO_UNLOCK();
	return 0;
}


void mpu210_siogpio_intr_enable(void) {}
void mpu210_siogpio_intr_disable(void) {}

int mpu210_siogpio_intr_set_mode_bit(int port, int bitnum, __u32 mode)
{
	if (0 == is_bitnum_valid(bitnum)) {
		return -EINVAL;
	}
	return mpu210_siogpio_intr_set_mode(port, (1 << bitnum), mode);
}

int mpu210_siogpio_intr_set_mode(int port, __u8 bitmask, __u32 mode)
{
	int j;
	unsigned long flags;
	int is_positive;
	int is_edge;

	if (0 == is_port_bitmask_valid(port, bitmask)) {
		return -EINVAL;
	}

	switch (mode) {
		case MPU210_SIOGPIO_INTR_EDGE_HIGH:
			is_positive = 1;
			is_edge = 1;
			break;

		case MPU210_SIOGPIO_INTR_EDGE_LOW:
			is_positive = 0;
			is_edge = 1;
			break;

		case MPU210_SIOGPIO_INTR_LEVEL_HIGH:
			is_positive = 1;
			is_edge = 0;
			break;

		case MPU210_SIOGPIO_INTR_LEVEL_LOW:
			is_positive = 0;
			is_edge = 0;
			break;

		default:
			return -EINVAL;
	}

	SIOGPIO_LOCK();

	for (j=0; j<MPU210_SIOGPIO_BIT_NUM; j++) {
		if ((bitmask >> j) & 1) {

			/* state check */
			if (0 == is_in_state(port, j, STATE_INTR_REGISTERED |
						      STATE_INTR_MODE_SET)) {
				/* no handler registered */
				SIOGPIO_UNLOCK();
				return -EINVAL;
			}

			if ( is_out(port, j) ) {
				SIOGPIO_UNLOCK();
				return -EINVAL;
			}
		}
	}

	for (j=0; j<MPU210_SIOGPIO_BIT_NUM; j++) {
		if ((bitmask >> j) & 1) {
			int ret;
			int irq;

			irq = gpio[port].bit[j].irq;

			/* set interrupt mode */
			set_lpc_irq_mode(irq, is_positive, is_edge);

			/* route input to the specified IRQ */
			LPCSetConfig(0x07, 0xe0 + irq, ((port + 1) << 3) | j);

			/*
			 * activate interrupt.
			 * SA_SHIRQ should not be set here, because
			 * IRQ itself is disabled when
			 * mpu210_siogpio_intr_mask() is called.
			 */
			ret = request_irq(irq,
					  gpio[port].bit[j].handler,
					  SA_INTERRUPT,
					  gpio[port].bit[j].name,
					  gpio[port].bit[j].dev_id);
			if (ret < 0) {
				printk(KERN_NOTICE "siogpio: request_irq(%d) failed (err=%d)\n", irq, -ret);
				/* TODO: return error */
			}

			/*
			 * disable interrupt explicitly, because
			 * interrupt is supposed to be disabled by default.
			 */
			disable_irq(irq);

			/* change state: INTR_REGISTERED => INTR_MODE_SET */
			if (STATE_INTR_REGISTERED == gpio[port].bit[j].state) {
				gpio[port].bit[j].state = STATE_INTR_MODE_SET;
			}
		}
	}

	SIOGPIO_UNLOCK();
	return 0;
}

int mpu210_siogpio_intr_register_handler(int port, int bitnum,
                                          void (*handler)(int irq, void *dev_id, struct pt_regs *regs),
                                          void *dev_id)
{
	int irq;
	unsigned long flags;

	if (0 == is_port_bitnum_valid(port, bitnum)) {
		return -EINVAL;
	}
	if (NULL == handler) {
		return -EINVAL;
	}

	SIOGPIO_LOCK();

	/* state check */
	switch (gpio[port].bit[bitnum].state) {

		case STATE_INITIAL:
		case STATE_REGISTERED:
			/* not registered */
			SIOGPIO_UNLOCK();
			return -EINVAL;

		case STATE_INTR_REGISTERED:
		case STATE_INTR_MODE_SET:
			/* already handler registered */
			SIOGPIO_UNLOCK();
			return -EBUSY;

		default:
			break;
	}

	if ( is_out(port, bitnum) ) {
		/* set as "output" */
		SIOGPIO_UNLOCK();
		return -EINVAL;
	}

	irq = siogpio_acquire_irq();
	if ( irq < 0 ) {
		/* no IRQ available */
		SIOGPIO_UNLOCK();
		return -EBUSY;
	}

	/* change state: MODE_SET => INTR_REGISTERED */
	gpio[port].bit[bitnum].state = STATE_INTR_REGISTERED;

	/* select Simple I/O function as "alternate" */
	set_function(port, bitnum, 0);

	/* save original modes */
	{
		int is_positive;
		int is_edge;

		get_lpc_irq_mode(irq, &is_positive, &is_edge);
		gpio[port].bit[bitnum].org_value |= (is_edge     << 6);
		gpio[port].bit[bitnum].org_value |= (is_positive << 7);
	}

	gpio[port].bit[bitnum].handler = handler;
	gpio[port].bit[bitnum].dev_id = dev_id;
	gpio[port].bit[bitnum].irq = irq;

	/*
	 * not call request_irq() here, because
	 * trigger mode has not been specified yet.
	 */

	SIOGPIO_UNLOCK();
	return 0;
}

int mpu210_siogpio_intr_unregister_handler(int port, int bitnum, void *dev_id)
{
	int irq;
	unsigned long flags;

	if (0 == is_port_bitnum_valid(port, bitnum)) {
		return -EINVAL;
	}

	if (dev_id != gpio[port].bit[bitnum].dev_id) {
		return -EINVAL;
	}

	SIOGPIO_LOCK();

	/* state check */
	if (0 == is_in_state(port, bitnum, STATE_INTR_REGISTERED |
					   STATE_INTR_MODE_SET)) {
		/* not registered */
		SIOGPIO_UNLOCK();
		return -EINVAL;
	}

	irq = gpio[port].bit[bitnum].irq;

	if ( is_in_state(port, bitnum, STATE_INTR_MODE_SET) ) {
		mpu210_siogpio_intr_mask_bit(port, bitnum);
		free_irq(irq, dev_id);
	}

	siogpio_release_irq(irq);

	/* set function as "simple I/O" */
	set_function(port, bitnum, 1);

	/* change state: INTR_MODE_SET => MODE_SET */
	gpio[port].bit[bitnum].state = STATE_MODE_SET;

	/* restore original modes */
	{
		int is_positive;
		int is_edge;

		is_positive = (gpio[port].bit[bitnum].org_value >> 6) & 1;
		is_edge     = (gpio[port].bit[bitnum].org_value >> 7) & 1;
		set_lpc_irq_mode(irq, is_positive, is_edge);
	}

	gpio[port].bit[bitnum].handler = NULL;
	gpio[port].bit[bitnum].dev_id = NULL;
	gpio[port].bit[bitnum].irq = -1;

	SIOGPIO_UNLOCK();
	return 0;
}

int mpu210_siogpio_intr_mask_bit(int port, int bitnum)
{
	if (0 == is_port_bitnum_valid(port, bitnum)) {
		return -EINVAL;
	}
	return mpu210_siogpio_intr_mask(port, (1 << bitnum));
}

int mpu210_siogpio_intr_unmask_bit(int port, int bitnum)
{
	if (0 == is_port_bitnum_valid(port, bitnum)) {
		return -EINVAL;
	}
	return mpu210_siogpio_intr_unmask(port, (1 << bitnum));
}

int mpu210_siogpio_intr_mask(int port, __u8 bitmask)
{
	int j;
	unsigned long flags;

	if (0 == is_port_bitmask_valid(port, bitmask)) {
		return -EINVAL;
	}

	SIOGPIO_LOCK();

	for (j=0; j<MPU210_SIOGPIO_BIT_NUM; j++) {
		if ((bitmask >> j) & 1) {
			/* state check */
			if (0 == is_in_state(port, j, STATE_INTR_MODE_SET)) {
				/* no intr mode set */
				SIOGPIO_UNLOCK();
				return -EINVAL;
			}
		}
	}

	for (j=0; j<MPU210_SIOGPIO_BIT_NUM; j++) {
		if ((bitmask >> j) & 1) {
			disable_irq_nosync( gpio[port].bit[j].irq );
			/*
			 * call disable_irq_nosync(), because
			 * this function may be called from interrupt handler.
			 */
		}
	}

	SIOGPIO_UNLOCK();
	return 0;
}

int mpu210_siogpio_intr_unmask(int port, __u8 bitmask)
{
	int j;
	unsigned long flags;

	if (0 == is_port_bitmask_valid(port, bitmask)) {
		return -EINVAL;
	}

	SIOGPIO_LOCK();

	for (j=0; j<MPU210_SIOGPIO_BIT_NUM; j++) {
		if ((bitmask >> j) & 1) {
			/* state check */
			if (0 == is_in_state(port, j, STATE_INTR_MODE_SET)) {
				/* no intr mode set */
				SIOGPIO_UNLOCK();
				return -EINVAL;
			}
		}
	}

	for (j=0; j<MPU210_SIOGPIO_BIT_NUM; j++) {
		if ((bitmask >> j) & 1) {
			enable_irq( gpio[port].bit[j].irq );
		}
	}

	SIOGPIO_UNLOCK();
	return 0;
}

int mpu210_siogpio_intr_get_status_bit(int port, int bitnum)
{
	return 0;
}

int mpu210_siogpio_intr_get_status(int port, __u8 *status)
{
	return 0;
}

int mpu210_siogpio_intr_clear_bit(int port, int bitnum)
{
	return 0;
}

int mpu210_siogpio_intr_clear(int port, __u8 bitmask)
{
	return 0;
}


/*
 * setup / cleanup functions
 */

static int __init mpu210_siogpio_init(void)
{
	int j;
	int k;

	set_io_base(MPU210_SIOGPIO_SIMPLE_IO);
	set_debounce_base(MPU210_SIOGPIO_DEBOUNCE);
	spin_lock_init(&siogpio_lock);
	spin_lock_init(&irq_table_lock);

	/* initialize pin info. */
	for (k=0; k<MPU210_SIOGPIO_PORT_NUM; k++) {
		for (j=0; j<MPU210_SIOGPIO_BIT_NUM; j++) {
			gpio[k].bit[j].capability = CAP_READABLE | CAP_WRITABLE;
			gpio[k].bit[j].state = STATE_INITIAL;
			gpio[k].bit[j].org_value = 0;
			gpio[k].bit[j].name[0] = '\0';
			gpio[k].bit[j].handler = NULL;
			gpio[k].bit[j].dev_id = NULL;
			gpio[k].bit[j].irq = -1;
		}
	}

	/* only PIN 5 has invalid bits */
	gpio[MPU210_SIOGPIO_PORT_PIN5].bit[6].capability = CAP_NONE;
	gpio[MPU210_SIOGPIO_PORT_PIN5].bit[7].capability = CAP_NONE;

#ifdef DEBUG
	create_proc_read_entry("siogpio", 0, 0, siogpio_read_proc, NULL);
#endif /* DEBUG */

	return 0;
}

static int mpu210_siogpio_init_boottime(void)
{
	mpu210_siogpio_init();
	return 0;
}

__initcall(mpu210_siogpio_init_boottime);


EXPORT_SYMBOL(mpu210_siogpio_register_bit);
EXPORT_SYMBOL(mpu210_siogpio_unregister_bit);
EXPORT_SYMBOL(mpu210_siogpio_register);
EXPORT_SYMBOL(mpu210_siogpio_unregister);
EXPORT_SYMBOL(mpu210_siogpio_unregister_by_name);
EXPORT_SYMBOL(mpu210_siogpio_set_mode_bit);
EXPORT_SYMBOL(mpu210_siogpio_set_mode);

EXPORT_SYMBOL(mpu210_siogpio_get_data_bit);
EXPORT_SYMBOL(mpu210_siogpio_set_data_bit);
EXPORT_SYMBOL(mpu210_siogpio_get_data);
EXPORT_SYMBOL(mpu210_siogpio_set_data);

EXPORT_SYMBOL(mpu210_siogpio_intr_enable);
EXPORT_SYMBOL(mpu210_siogpio_intr_disable);
EXPORT_SYMBOL(mpu210_siogpio_intr_set_mode_bit);
EXPORT_SYMBOL(mpu210_siogpio_intr_set_mode);
EXPORT_SYMBOL(mpu210_siogpio_intr_register_handler);
EXPORT_SYMBOL(mpu210_siogpio_intr_unregister_handler);
EXPORT_SYMBOL(mpu210_siogpio_intr_unmask_bit);
EXPORT_SYMBOL(mpu210_siogpio_intr_mask_bit);
EXPORT_SYMBOL(mpu210_siogpio_intr_unmask);
EXPORT_SYMBOL(mpu210_siogpio_intr_mask);
EXPORT_SYMBOL(mpu210_siogpio_intr_get_status_bit);
EXPORT_SYMBOL(mpu210_siogpio_intr_get_status);
EXPORT_SYMBOL(mpu210_siogpio_intr_clear_bit);
EXPORT_SYMBOL(mpu210_siogpio_intr_clear);

