/*
 *  snsc_mpu30x_siogpio.c : MPU-30X SuperIO GPIO driver
 *
 *  Copyright 2002,2003 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __KERNEL__
#error "this module must be compiled with -D__KERNEL__"
#endif

#include <linux/version.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/delay.h>
#include <linux/spinlock.h>
#include <linux/smp_lock.h>
#include <linux/snsc_ich_gpio.h>
#include <linux/snsc_mpu30x_siogpio.h>
#include <linux/errno.h>
#include <asm/io.h>


#ifdef MODULE
MODULE_AUTHOR("Sony NSC");
MODULE_DESCRIPTION("MPU-300 SuperIO GPIO driver");

#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif /* MODULE_LICENSE */
#endif /* MODULE */

//#define DEBUG

#ifdef DEBUG
#define GPIO_DEBUG(args...)  printk(KERN_NOTICE args)
#define GPIO_DEBUG_INTR_STATUS    __print_intr_status()
static void __print_intr_status( void );
#else
#define GPIO_DEBUG(args...)
#define GPIO_DEBUG_INTR_STATUS
#endif

#define PME_BASE  0x800

/* status registers */
#define SPIO_PME_STS     (PME_BASE + 0x00)
#define SPIO_PME_EN      (PME_BASE + 0x02)
#define SPIO_PME_STS1    (PME_BASE + 0x04)
#define SPIO_PME_STS2    (PME_BASE + 0x05)
#define SPIO_PME_STS3    (PME_BASE + 0x06)
#define SPIO_PME_STS4    (PME_BASE + 0x07)
#define SPIO_PME_STS5    (PME_BASE + 0x08)
#define SPIO_PME_EN1     (PME_BASE + 0x0a)
#define SPIO_PME_EN2     (PME_BASE + 0x0b)
#define SPIO_PME_EN3     (PME_BASE + 0x0c)
#define SPIO_PME_EN4     (PME_BASE + 0x0d)
#define SPIO_PME_EN5     (PME_BASE + 0x0e)
#define SPIO_SMI_STS1    (PME_BASE + 0x10)
#define SPIO_SMI_STS2    (PME_BASE + 0x11)
#define SPIO_SMI_STS3    (PME_BASE + 0x12)
#define SPIO_SMI_STS4    (PME_BASE + 0x13)
#define SPIO_SMI_STS5    (PME_BASE + 0x14)
#define SPIO_SMI_EN1     (PME_BASE + 0x16)
#define SPIO_SMI_EN2     (PME_BASE + 0x17)
#define SPIO_SMI_EN3     (PME_BASE + 0x18)
#define SPIO_SMI_EN4     (PME_BASE + 0x19)
#define SPIO_SMI_EN5     (PME_BASE + 0x1a)
#define SPIO_MSC_STS     (PME_BASE + 0x1c)

/* gpio mode & function registers */
#define SPIO_GP20        (PME_BASE + 0x2b)
#define SPIO_GP21        (PME_BASE + 0x2c)
#define SPIO_GP22        (PME_BASE + 0x2d)
#define SPIO_GP24        (PME_BASE + 0x2f)
#define SPIO_GP25        (PME_BASE + 0x30)
#define SPIO_GP26        (PME_BASE + 0x31)
#define SPIO_GP27        (PME_BASE + 0x32)

#define SPIO_GP34        (PME_BASE + 0x37)
#define SPIO_GP35        (PME_BASE + 0x38)
#define SPIO_GP36        (PME_BASE + 0x39)
#define SPIO_GP37        (PME_BASE + 0x3a)

#define SPIO_GP40        (PME_BASE + 0x3b)
#define SPIO_GP41        (PME_BASE + 0x3c)
#define SPIO_GP42        (PME_BASE + 0x3d)
#define SPIO_GP43        (PME_BASE + 0x3e)

#define SPIO_GP60        (PME_BASE + 0x47)
#define SPIO_GP61        (PME_BASE + 0x48)

/* gpio data registers */
#define SPIO_GP2         (PME_BASE + 0x4c)
#define SPIO_GP3         (PME_BASE + 0x4d)
#define SPIO_GP4         (PME_BASE + 0x4e)
#define SPIO_GP6         (PME_BASE + 0x50)


/* ICH2 GPIO[8] is used after MPU-301-11 */
#define ICHGPIO_LPC_PME  8


/* GPIO mode bitmask */
#define MODE_INOUT    (1 << 0)
#define MODE_POLARITY (1 << 1)
#define MODE_OUTTYPE  (1 << 7)
#define MODE_EETI     (2 << 2)
#define MODE_EETI_43  (3 << 2)


#define NUM_GPIO  16
#define OWNER_NAME_LEN  16

static const char driver_name[] = "snsc_mpu30x_siogpio";

static void
mpu30x_siogpio_intr_handler(int irq, void *dev_id, struct pt_regs *regs);

struct siogpio_info {
	__u16 registered;
	__u16 intr_registered;
	struct _gpio_pin {
		char owner[OWNER_NAME_LEN];
		void (*intr_handler)(int irq, void *dev_id,
				     struct pt_regs *regs);
		void *intr_dev_id;
	} gpio[NUM_GPIO];
	spinlock_t write_lock;
	spinlock_t conf_lock;
	spinlock_t reg_lock;
	spinlock_t pme_lock;
};

static struct siogpio_info siogpio_info;

#define GPIO_MASK_INTERRUPTIBLE      0x03ff
#define GPIO_MASK_EETINTERRUPTIBLE   0x003f

static struct pin_info {
	int name;
	unsigned int mode_reg;
	unsigned int data_reg;
	__u8 data_mask;
	unsigned int intr_reg;
	unsigned int status_reg;
	__u8 pme_mask;
} pin_info[NUM_GPIO] = {
  { 21, SPIO_GP21, SPIO_GP2, 0x02, SPIO_PME_EN3, SPIO_PME_STS3, 0x02 },
  { 22, SPIO_GP22, SPIO_GP2, 0x04, SPIO_PME_EN3, SPIO_PME_STS3, 0x04 },
  { 41, SPIO_GP41, SPIO_GP4, 0x02, SPIO_PME_EN4, SPIO_PME_STS4, 0x10 },
  { 43, SPIO_GP43, SPIO_GP4, 0x08, SPIO_PME_EN4, SPIO_PME_STS4, 0x20 },
  { 60, SPIO_GP60, SPIO_GP6, 0x01, SPIO_PME_EN4, SPIO_PME_STS4, 0x40 },
  { 61, SPIO_GP61, SPIO_GP6, 0x02, SPIO_PME_EN4, SPIO_PME_STS4, 0x80 },
  { 20, SPIO_GP20, SPIO_GP2, 0x01, SPIO_PME_EN3, SPIO_PME_STS3, 0x01 }, 
  { 25, SPIO_GP25, SPIO_GP2, 0x20, SPIO_PME_EN3, SPIO_PME_STS3, 0x20 },
  { 26, SPIO_GP26, SPIO_GP2, 0x40, SPIO_PME_EN3, SPIO_PME_STS3, 0x40 },
  { 27, SPIO_GP27, SPIO_GP2, 0x80, SPIO_PME_EN3, SPIO_PME_STS3, 0x80 },
  { 34, SPIO_GP34, SPIO_GP3, 0x10 },
  { 35, SPIO_GP35, SPIO_GP3, 0x20 },
  { 36, SPIO_GP36, SPIO_GP3, 0x40 },
  { 37, SPIO_GP37, SPIO_GP3, 0x80 },
  { 40, SPIO_GP40, SPIO_GP4, 0x01 },
  { 24, SPIO_GP24, SPIO_GP2, 0x10 }
};


/*
 * setup / cleanup functions
 */
static int mpu30x_siogpio_init( void )
{
	struct siogpio_info *info = &siogpio_info;
	int res;

	spin_lock_init(&info->write_lock);
	spin_lock_init(&info->conf_lock);
	spin_lock_init(&info->reg_lock);
	spin_lock_init(&info->pme_lock);

	/* ICH2 GPIO setup */
	ich_gpio_register_bit(0, ICHGPIO_LPC_PME, driver_name);
	ich_gpio_intr_mask_bit(0, ICHGPIO_LPC_PME);
	ich_gpio_set_mode_bit(0, ICHGPIO_LPC_PME, GPIO_MODE_IN);
	ich_gpio_intr_set_mode_bit(0, ICHGPIO_LPC_PME,
				   GPIO_INTR_LEVEL_LOW |
				   GPIO_INTR_ROUT_SCI);

	/* Disable interrupt and mask all bit */
	mpu30x_siogpio_intr_disable();
	mpu30x_siogpio_intr_mask(0, SIOGPIO_BITS_ALL);

	/* Interrupt status bit clear */
	mpu30x_siogpio_intr_clear(0, SIOGPIO_BITS_ALL);

	/* enable nIO_PME output */
	outb_p(0x04, SPIO_GP42);

	/* clear PME_STS and ICH2 GPE_STS */
	outb_p(1, SPIO_PME_STS);
	ich_gpio_intr_clear_bit(0, ICHGPIO_LPC_PME);

	/* register inerrupt handler */
	res = ich_gpio_intr_register_handler(0, ICHGPIO_LPC_PME,
					     mpu30x_siogpio_intr_handler,
					     info);
	if (res != 0){
		printk(KERN_INFO
		       "%s: can't install interrupt handler, res=%d\n",
		       driver_name, res);
		return res;
	}

	printk(KERN_INFO "%s: installed interrupt handler\n", driver_name);

	GPIO_DEBUG_INTR_STATUS;

	ich_gpio_intr_unmask_bit(0, ICHGPIO_LPC_PME);

	/* enable nIO_PME signal assertion */
	mpu30x_siogpio_intr_enable();

	return 0;
}

static void mpu30x_siogpio_cleanup( void )
{
	struct siogpio_info *info = &siogpio_info;

	ich_gpio_intr_mask_bit(0, ICHGPIO_LPC_PME);
	mpu30x_siogpio_intr_disable();

	mpu30x_siogpio_intr_clear(0, SIOGPIO_BITS_ALL);
	outb_p(1, SPIO_PME_STS);
	ich_gpio_intr_clear_bit(0, ICHGPIO_LPC_PME);

	ich_gpio_intr_unregister_handler(0, ICHGPIO_LPC_PME, info);

	ich_gpio_unregister_bit(0, ICHGPIO_LPC_PME);
}

#ifdef MODULE
static int mpu30x_siogpio_init_module( void )
{
	printk(KERN_INFO "%s: SuperIO GPIO driver loaded.\n", driver_name);
	return mpu30x_siogpio_init();
}

static void mpu30x_siogpio_cleanup_module( void )
{
	printk(KERN_INFO "%s: SuperIO GPIO driver unloaded.\n", driver_name);
	mpu30x_siogpio_cleanup();
}
#endif /* MODULE */


/*
 * SNSC common GPIO driver interfaces
 */

#define IS_BIT_SET(bitmask, x)  ((bitmask) & (1 << (x)))

#define PORT_CHECK(port) do {                                                 \
                              if ((port) != 0) {                              \
                                  GPIO_DEBUG("%s: port %d is not supported\n",\
                                             __FUNCTION__, (port));           \
                                  return -EINVAL;                             \
                              }                                               \
                         } while (0);

#define BITNUM_CHECK(bitnum) do {                                             \
                                 if ((bitnum) < 0 || (bitnum) >= NUM_GPIO) {  \
                                     GPIO_DEBUG("%s: invalid pin number %d\n",\
                                                __FUNCTION__, (bitnum));      \
                                     return -EINVAL;                          \
                                 }                                            \
                             } while (0);

/*
 * interrupt handler
 */
static void
mpu30x_siogpio_intr_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	struct siogpio_info *info = &siogpio_info;
	__u16 status;
	__u16 en3, en4;
	int i;

#ifdef DEBUG
	GPIO_DEBUG("%s: Interrupt occured,\n", driver_name);
	GPIO_DEBUG_INTR_STATUS;

	if (!ich_gpio_intr_get_status_bit(0, ICHGPIO_LPC_PME)){
		GPIO_DEBUG("this interrupt is not caused by SuperIO\n");
		return;
	}
#endif /* DEBUG */

	mpu30x_siogpio_intr_get_status(0, &status);

	en3 = (__u16)inb(SPIO_PME_EN3) & 0xff;
	en4 = (__u16)inb(SPIO_PME_EN4) & 0xff;

	status &= ( ((en3 & 0x06) >> 1) |
		    ((en4 & 0xf0) >> 4) << 2 |
		    ((en3 & 0x01) >> 0) << 6 |
		    ((en3 & 0xe0) >> 5) << 7 );

	for (i = 0; i < NUM_GPIO; i++) {
		if (IS_BIT_SET(status, i)) {
			if (!(IS_BIT_SET(info->intr_registered, i)) ||
			    (info->gpio[i].intr_handler == NULL)) {
				GPIO_DEBUG("%s: warning, gpio %d asserts "
					   "interrupt, but handler is not "
					   "registered.\n", __FUNCTION__, i);
				continue;
			}
			/* call registered interrupt handler */
			(*info->gpio[i].intr_handler)
				(irq, info->gpio[i].intr_dev_id, regs);

			mpu30x_siogpio_intr_clear_bit(0, i);
		}
	}

	outb_p(1, SPIO_PME_STS);

#ifdef DEBUG
	/* ich gpio driver will clear the status bit, so 
	   we don't need to clear status here */
	ich_gpio_intr_clear_bit(0, ICHGPIO_LPC_PME);

	GPIO_DEBUG("%s: Interrupt handler end,\n", driver_name);
	GPIO_DEBUG_INTR_STATUS;
#endif /* DEBUG */
}

#ifdef DEBUG
static void __print_intr_status( void )
{
	__u16 status;

	mpu30x_siogpio_intr_get_status(0, &status);

	GPIO_DEBUG("PME_STS1=%02x STS2=%02x STS3=%02x STS4=%02x STS5=%02x, "
		   "status=%02x\n",
		   inb(SPIO_PME_STS1), inb(SPIO_PME_STS2), inb(SPIO_PME_STS3),
		   inb(SPIO_PME_STS4), inb(SPIO_PME_STS5), status);
	GPIO_DEBUG(" PME_EN1=%02x  EN2=%02x  EN3=%02x  EN4=%02x  EN5=%02x\n",
		   inb(SPIO_PME_EN1), inb(SPIO_PME_EN2), inb(SPIO_PME_EN3),
		   inb(SPIO_PME_EN4), inb(SPIO_PME_EN5));
	GPIO_DEBUG("PME_STS=%02x PME_EN=%02x ichgpio=%0x\n",
		   inb(SPIO_PME_STS), inb(SPIO_PME_EN),
		   ich_gpio_intr_get_status_bit(0, ICHGPIO_LPC_PME));
}
#endif


int
mpu30x_siogpio_register_bit(int port, int bitnum, const char *name)
{
	BITNUM_CHECK(bitnum);
	return mpu30x_siogpio_register(port, (1 << bitnum), name);
}

int
mpu30x_siogpio_unregister_bit(int port, int bitnum)
{
	BITNUM_CHECK(bitnum);
	return mpu30x_siogpio_unregister(port, (1 << bitnum));
}

int
mpu30x_siogpio_register(int port, __u16 bitmask, const char *name)
{
        struct siogpio_info *info = &siogpio_info;
	unsigned long flags;
	int i;

	PORT_CHECK(port);

	if (name == NULL) {
		GPIO_DEBUG("%s: name must be specified\n", __FUNCTION__);
		return -EINVAL;
	}

	spin_lock_irqsave(&info->reg_lock, flags);

	if (bitmask & info->registered) {
		spin_unlock_irqrestore(&info->reg_lock, flags);
		GPIO_DEBUG("%s: bit(s) are already registered(0x%08x)\n",
			   __FUNCTION__, bitmask);
		return -EBUSY;
	}

	info->registered |= bitmask;

	for (i = 0; i < NUM_GPIO; i++) {
		if (IS_BIT_SET(bitmask, i)) {
			strncpy(info->gpio[i].owner, name, OWNER_NAME_LEN);
			info->gpio[i].owner[OWNER_NAME_LEN] = '\0';
#ifdef MODULE
			MOD_INC_USE_COUNT;
#endif
		}
	}

	spin_unlock_irqrestore(&info->reg_lock, flags);

	GPIO_DEBUG("%s: registered pin 0x%08x by %s\n",
		   __FUNCTION__, bitmask, name);

	return 0;
}

int
mpu30x_siogpio_unregister(int port, __u16 bitmask)
{
        struct siogpio_info *info = &siogpio_info;
	unsigned long flags;
	int i;

	PORT_CHECK(port);

	spin_lock_irqsave(&info->reg_lock, flags);

	if ((bitmask & info->registered) != bitmask) {
		GPIO_DEBUG("%s: attempt to unregister 0x%08x, but "
			   "registered pin(s) are 0x%08x\n",
			   __FUNCTION__, bitmask, info->registered);
		spin_unlock_irqrestore(&info->reg_lock, flags);
		return -EINVAL;
	}

	info->registered &= ~bitmask;

	for (i = 0; i < NUM_GPIO; i++) {
		if (IS_BIT_SET(bitmask, i)) {
			memset(info->gpio[i].owner, '\0', OWNER_NAME_LEN);
#ifdef MODULE
			MOD_DEC_USE_COUNT;
#endif
		}
	}

	spin_unlock_irqrestore(&info->reg_lock, flags);

	GPIO_DEBUG("%s: unregistered pin 0x%08x\n",
		   __FUNCTION__, bitmask);

	return 0;
}

int
mpu30x_siogpio_unregister_by_name(const char *name)
{
        struct siogpio_info *info = &siogpio_info;
	unsigned long flags;
	__u16 bitmask = 0;
	int i;

	if (name == NULL) {
		GPIO_DEBUG("%s: name must be specified\n", __FUNCTION__);
		return -EINVAL;
	}

	spin_lock_irqsave(&info->reg_lock, flags);

	for (i = 0; i < NUM_GPIO; i++) {
		if (strncmp(info->gpio[i].owner, name, OWNER_NAME_LEN) == 0) {
			bitmask |= (1 << i);
		}
	}

	spin_unlock_irqrestore(&info->reg_lock, flags);

	if (bitmask == 0) {
		GPIO_DEBUG("%s: there's no pin registered by %s!\n",
			   __FUNCTION__, name);
		return -EINVAL;
	}

	return mpu30x_siogpio_unregister(0, bitmask);
}

int
mpu30x_siogpio_set_mode_bit(int port, int bitnum, __u32 mode)
{
	BITNUM_CHECK(bitnum);
	return mpu30x_siogpio_set_mode(port, (1 << bitnum), mode);
}

int
mpu30x_siogpio_set_mode(int port, __u16 bitmask , __u32 mode)
{
	struct siogpio_info *info = &siogpio_info;
	unsigned long flags;
	int i;
	__u8 conf = 0;

	PORT_CHECK(port);

	if ((mode & SIOGPIO_MODE_INVMASK) == SIOGPIO_MODE_INVERT) {
		conf |= MODE_POLARITY;
	}

	if ((mode & SIOGPIO_MODE_OUTDRVMASK) == SIOGPIO_MODE_OPENDRAIN) {
		conf |= MODE_OUTTYPE;
	}

	/* bit 15 cannot be used for input */
	switch (mode & SIOGPIO_MODE_IOMASK) {
	case SIOGPIO_MODE_IN:
		if (IS_BIT_SET(bitmask, 15)) {
			return -EINVAL;
		}
		conf |= MODE_INOUT;
		break;
	case SIOGPIO_MODE_OUT_HIGH:
		mpu30x_siogpio_set_data(port, bitmask, bitmask);
		break;
	case SIOGPIO_MODE_OUT_LOW:
		mpu30x_siogpio_set_data(port, bitmask, 0);
		break;
	}

	spin_lock_irqsave(&info->conf_lock, flags);

	for ( i = 0 ; i < NUM_GPIO; i++){
		if (IS_BIT_SET(bitmask, i)){
			outb_p(conf, pin_info[i].mode_reg);
		}
	}

	spin_unlock_irqrestore(&info->conf_lock, flags);

	GPIO_DEBUG("%s: set mode for pin 0x%08x to 0x%08x\n",
		   __FUNCTION__, bitmask, mode);

	return 0;
}

/* data register access */
int
mpu30x_siogpio_get_data_bit(int port, int bitnum)
{
	__u16 value;

	PORT_CHECK(port);
	BITNUM_CHECK(bitnum);

	value = inb(pin_info[bitnum].data_reg);

	return ((value & pin_info[bitnum].data_mask) != 0) ? 1 : 0;
}

int
mpu30x_siogpio_set_data_bit(int port, int bitnum, int value)
{
        struct siogpio_info *info = &siogpio_info;
	unsigned long flags;
	__u8 data;

	PORT_CHECK(port);
	BITNUM_CHECK(bitnum);

	spin_lock_irqsave(&info->write_lock, flags);
 
	data = inb(pin_info[bitnum].data_reg);

	if (value & 1)
		data |= pin_info[bitnum].data_mask;
	else
		data &= ~pin_info[bitnum].data_mask;

	outb_p(data, pin_info[bitnum].data_reg);

	spin_unlock_irqrestore(&info->write_lock, flags);

	return 0;
}

int
mpu30x_siogpio_get_data(int port, __u16 bitmask, __u16 *value)
{
	__u8 gp2, gp3, gp4, gp6;

	PORT_CHECK(port);

	gp2 = inb(SPIO_GP2);
	gp3 = inb(SPIO_GP3);
	gp4 = inb(SPIO_GP4);
	gp6 = inb(SPIO_GP6);

	*value = ( ((gp2 & 0x06) >> 1) |
		   ((gp4 & 0x02) >> 1) <<  2 |
		   ((gp4 & 0x08) >> 3) <<  3 |
		   ((gp6 & 0x03) >> 0) <<  4 |
		   ((gp2 & 0x01) >> 0) <<  6 |
		   ((gp2 & 0xe0) >> 5) <<  7 |
		   ((gp3 & 0xf0) >> 4) << 10 |
		   ((gp4 & 0x01) >> 0) << 14 |
		   ((gp2 & 0x10) >> 4) << 15 );

	*value &= bitmask;

	/* general (but slightly inefficient) code */
	/*
	int i;
	*value = 0;

	for (i = 0; i < NUM_GPIO; i++){
		if (IS_BIT_SET(bitmask, i)){
			*value |= mpu30x_siogpio_get_data_bit(0, i) << i;
		}
	}
	*/

	GPIO_DEBUG("%s: read 0x%08x (mask 0x%08x)\n",
		   __FUNCTION__, *value, bitmask);

	return 0;
}

int
mpu30x_siogpio_set_data(int port, __u16 bitmask , __u16 value)
{
        struct siogpio_info *info = &siogpio_info;
	unsigned long flags;
	__u16 old;
	__u8 gp2, gp3, gp4, gp6;

	PORT_CHECK(port);

	spin_lock_irqsave(&info->write_lock ,flags);

	mpu30x_siogpio_get_data(0, SIOGPIO_BITS_ALL, &old);

	value = (value & bitmask) | (old & ~bitmask);

	gp2 = ( ((value & 0x0040) >>  6) |
		((value & 0x0003) >>  0) << 1 |
		((value & 0x8000) >> 15) << 4 |
		((value & 0x0380) >>  7) << 5 );

	gp3 = ( ((value & 0x3c00) >> 10 ) << 4 );

	gp4 = ( ((value & 0x4000) >> 14) |
		((value & 0x0004) >>  2) << 1 |
		((value & 0x0008) >>  3) << 3 );

	gp6 = (	((value & 0x0030) >> 4) );

	outb_p(gp2, SPIO_GP2);
	outb_p(gp3, SPIO_GP3);
	outb_p(gp4, SPIO_GP4);
	outb_p(gp6, SPIO_GP6);

	spin_unlock_irqrestore(&info->write_lock, flags);

	/* general (but slightly inefficient) code */
	/*
	int i;

	spin_locl_irqsave(&indo->write_lock, flags);

	for (i = 0;  i < NUM_GPIO; i++){
		if (IS_BIT_SET(bitmask, i)){
			mpu30x_siogpio_set_data_bit(0, i,
						    IS_BIT_SET(value, i));
		}
	}

	spin_unlock_irqrestore(&info->write_lock, flags);
	*/

	GPIO_DEBUG("%s: write 0x%08x (mask 0x%08x)\n",
		   __FUNCTION__, value, bitmask);

	return 0;
}

/*
 * Interrupt handling functions
 */
void
mpu30x_siogpio_intr_enable( void )
{
	outb_p(inb(SPIO_PME_EN) | 0x01, SPIO_PME_EN);
}


void
mpu30x_siogpio_intr_disable( void )
{
	outb_p(inb(SPIO_PME_EN) & ~0x01, SPIO_PME_EN);
}

int
mpu30x_siogpio_intr_set_mode_bit(int port, int bitnum, __u32 mode)
{
	BITNUM_CHECK(bitnum);
	return mpu30x_siogpio_intr_set_mode(port, (1 << bitnum), mode);
}

int
mpu30x_siogpio_intr_set_mode(int port, __u16 bitmask, __u32 mode)
{
        struct siogpio_info *info = &siogpio_info;
	unsigned long flags;
	int i;

	PORT_CHECK(port);

	if (bitmask & ~GPIO_MASK_INTERRUPTIBLE) {
		GPIO_DEBUG("%s: specified pin(s) cannot cause interrupt",
			   __FUNCTION__);
		return -EINVAL;
	}

	if ((mode == SIOGPIO_INTR_EDGE_BOTH) &&
	    (bitmask & ~GPIO_MASK_EETINTERRUPTIBLE)) {
		GPIO_DEBUG("%s: specified pin(s) cannot cause "
			   "interrupt on both edge", __FUNCTION__);
		return -EINVAL;
	}

	mpu30x_siogpio_intr_mask(port, bitmask);

	spin_lock_irqsave(&info->conf_lock, flags);

	for (i = 0; i < NUM_GPIO; i++){
		if (IS_BIT_SET(bitmask, i)){
			__u8 conf;
			conf = inb(pin_info[i].mode_reg);
			switch (mode){
			case SIOGPIO_INTR_EDGE_HIGH:
				conf &= ~MODE_POLARITY;
				break;
			case SIOGPIO_INTR_EDGE_LOW:
				conf |= MODE_POLARITY;
				break;
			case SIOGPIO_INTR_EDGE_BOTH:
				conf |= (pin_info[i].name == 43) ? 
					MODE_EETI_43 : MODE_EETI;
				break;
			}
			outb_p(conf, pin_info[i].mode_reg);
		}
	}

	spin_unlock_irqrestore(&info->conf_lock, flags);

	GPIO_DEBUG("%s: set intr mode for pin 0x%08x to 0x%08x\n",
		   __FUNCTION__, bitmask, mode);

	return 0;
}

int
mpu30x_siogpio_intr_register_handler(int port, int bitnum,
				     void (*handler)(int irq, void *dev_id,
						     struct pt_regs *regs),
				     void *dev_id)
{
        struct siogpio_info *info = &siogpio_info;
	unsigned long flags;

	PORT_CHECK(port);
	BITNUM_CHECK(bitnum);

	if (dev_id == NULL) {
		GPIO_DEBUG("%s: dev_id must be specified, cannot be NULL\n",
			   __FUNCTION__);
		return -EINVAL;
	}

	if (!(GPIO_MASK_INTERRUPTIBLE & (1 << bitnum))) {
		GPIO_DEBUG("%s: specified pin cannot cause interrupt",
			   __FUNCTION__);
		return -EINVAL;
	}

	spin_lock_irqsave(&info->reg_lock, flags);

	if (info->intr_registered & (1 << bitnum)) {
		GPIO_DEBUG("%s: attempt to register intr handler which has"
			   " already another handler registered\n",
			   __FUNCTION__);
		spin_unlock_irqrestore(&info->reg_lock, flags);
		return -EBUSY;
	}

	info->gpio[bitnum].intr_handler = handler;
	info->gpio[bitnum].intr_dev_id = dev_id;
	info->intr_registered |= (1 << bitnum);

#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif

	spin_unlock_irqrestore(&info->reg_lock, flags);

	GPIO_DEBUG("%s: registered handler(0x%08x) for pin %d, "
		   "dev_id = 0x%08x\n", __FUNCTION__, (unsigned int)handler,
		   bitnum, (unsigned int)dev_id);
	
	return 0;
}

int
mpu30x_siogpio_intr_unregister_handler(int port, int bitnum, void *dev_id)
{
        struct siogpio_info *info = &siogpio_info;
	unsigned long flags;

	PORT_CHECK(port);
	BITNUM_CHECK(bitnum);

	spin_lock_irqsave(&info->reg_lock, flags);

	if (!(IS_BIT_SET(info->intr_registered, bitnum)) ||
	    (info->gpio[bitnum].intr_dev_id != dev_id)) {
		GPIO_DEBUG("%s: specified intr_handler is not registered\n",
			   __FUNCTION__);
		spin_unlock_irqrestore(&info->reg_lock, flags);
		return -EINVAL;
	}

	mpu30x_siogpio_intr_mask_bit(port, bitnum);

	info->intr_registered &= ~(1 << bitnum);
	info->gpio[bitnum].intr_dev_id = NULL;

#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif

	spin_unlock_irqrestore(&info->reg_lock, flags);

	GPIO_DEBUG("%s: unregistered handler for pin %d (dev_id = 0x%08x)\n",
		   __FUNCTION__, bitnum, (unsigned int)dev_id);

	return 0;
}

int
mpu30x_siogpio_intr_mask_bit(int port, int bitnum)
{
	struct siogpio_info *info = &siogpio_info;
	unsigned long flags;
	__u8 val;

	PORT_CHECK(port);
	BITNUM_CHECK(bitnum);

	if (!(GPIO_MASK_INTERRUPTIBLE & (1 << bitnum))) {
		GPIO_DEBUG("%s: specified pin(s) cannot cause interrupt,"
			   " so cannot mask interrupt\n", __FUNCTION__);
		return -EINVAL;
	}

	spin_lock_irqsave(&info->pme_lock, flags);

	val = inb(pin_info[bitnum].intr_reg);
	val &= ~pin_info[bitnum].pme_mask;
	outb_p(val, pin_info[bitnum].intr_reg);

	spin_unlock_irqrestore(&info->pme_lock, flags);

	return 0;
}

int
mpu30x_siogpio_intr_unmask_bit(int port, int bitnum)
{
	struct siogpio_info *info = &siogpio_info;
	unsigned long flags;
	__u8 val;

	PORT_CHECK(port);
	BITNUM_CHECK(bitnum);

	if (!(GPIO_MASK_INTERRUPTIBLE & (1 << bitnum))) {
		GPIO_DEBUG("%s: specified pin(s) cannot cause interrupt,"
			   " so cannot unmask interrupt\n", __FUNCTION__);
		return -EINVAL;
	}

	spin_lock_irqsave(&info->pme_lock, flags);

	val = inb(pin_info[bitnum].intr_reg);
	val |= pin_info[bitnum].pme_mask;
	outb_p(val, pin_info[bitnum].intr_reg);

	spin_unlock_irqrestore(&info->pme_lock, flags);

	return 0;
}

#define INTR_MASK   0
#define INTR_UNMASK 1

static void
__intr_change_mask(__u16 bitmask, int enable)
{
	struct siogpio_info *info = &siogpio_info;
	unsigned long flags;
	__u8 mask_lb, mask_hb;
	__u8 en3, en4;

	spin_lock_irqsave(&info->pme_lock, flags);

	mask_lb = ( ((bitmask & 0x0003) >> 0) << 1 |
		    ((bitmask & 0x0040) >> 6) |
		    ((bitmask & 0x0380) >> 7) << 5 );

	mask_hb = ( ((bitmask & 0x003c) >> 2) << 4 );

	en3 = inb(SPIO_PME_EN3);
	en4 = inb(SPIO_PME_EN4);

	if (enable == INTR_UNMASK){
		en3 |= mask_lb;
		en4 |= mask_hb;
	} else {
		en3 &= ~mask_lb;
		en4 &= ~mask_hb;
	}

	outb_p(en3, SPIO_PME_EN3);
	outb_p(en4, SPIO_PME_EN4);

	spin_unlock_irqrestore(&info->pme_lock, flags);  
}

int
mpu30x_siogpio_intr_mask(int port, __u16 bitmask)
{
	PORT_CHECK(port);

	__intr_change_mask(bitmask, INTR_MASK);

	return 0;
}

int
mpu30x_siogpio_intr_unmask(int port, __u16 bitmask)
{
	PORT_CHECK(port);

	__intr_change_mask(bitmask, INTR_UNMASK);

	return 0;
}

int
mpu30x_siogpio_intr_get_status_bit(int port, int bitnum)
{
	__u8 status;

	PORT_CHECK(port);
	BITNUM_CHECK(bitnum);

	if (!(GPIO_MASK_INTERRUPTIBLE & (1 << bitnum))) {
		return -EINVAL;
	}

	status = inb(pin_info[bitnum].status_reg);

	return ((status & pin_info[bitnum].pme_mask) != 0) ? 1 : 0;
}

int
mpu30x_siogpio_intr_get_status(int port, __u16 *status)
{
	__u16 sts3, sts4;

	PORT_CHECK(port);

	sts3 = (__u16)inb(SPIO_PME_STS3) & 0xff;
	sts4 = (__u16)inb(SPIO_PME_STS4) & 0xff;

	*status = ( ((sts3 & 0x06) >> 1) |
		    ((sts4 & 0xf0) >> 4) << 2 |
		    ((sts3 & 0x01) >> 0) << 6 |
		    ((sts3 & 0xe0) >> 5) << 7 );

	/* general (but slightly inefficient) code */
	/*
	int i;

	*status = 0;

	for (i = 0; i < NUM_GPIO; i++){
		if (IS_BIT_SET(bits, i)){
			*status |= mpu30x_siogpio_get_status_bit(port, i) << i;
		}
	}
	*/

	return 0;
}

int
mpu30x_siogpio_intr_clear_bit(int port, int bitnum)
{
	PORT_CHECK(port);
	BITNUM_CHECK(bitnum);

	if (!(GPIO_MASK_INTERRUPTIBLE & (1 << bitnum))) {
		return -EINVAL;
	}

	outb_p(pin_info[bitnum].pme_mask, pin_info[bitnum].status_reg);

	return 0;
}

int
mpu30x_siogpio_intr_clear(int port, __u16 bitmask)
{
	__u8 mask_lb, mask_hb;

	PORT_CHECK(port);

	mask_lb = ( ((bitmask & 0x0003) >> 0) << 1 |
		    ((bitmask & 0x0040) >> 6) |
		    ((bitmask & 0x0380) >> 7) << 5 );

	mask_hb = ( ((bitmask & 0x003c) >> 2) << 4 );

	outb_p(mask_lb, SPIO_PME_STS3);
	outb_p(mask_hb, SPIO_PME_STS4);

	return 0;
}



EXPORT_SYMBOL(mpu30x_siogpio_register_bit);
EXPORT_SYMBOL(mpu30x_siogpio_unregister_bit);
EXPORT_SYMBOL(mpu30x_siogpio_register);
EXPORT_SYMBOL(mpu30x_siogpio_unregister);
EXPORT_SYMBOL(mpu30x_siogpio_unregister_by_name);
EXPORT_SYMBOL(mpu30x_siogpio_set_mode_bit);
EXPORT_SYMBOL(mpu30x_siogpio_set_mode);

EXPORT_SYMBOL(mpu30x_siogpio_get_data_bit);
EXPORT_SYMBOL(mpu30x_siogpio_set_data_bit);
EXPORT_SYMBOL(mpu30x_siogpio_get_data);
EXPORT_SYMBOL(mpu30x_siogpio_set_data);

EXPORT_SYMBOL(mpu30x_siogpio_intr_enable);
EXPORT_SYMBOL(mpu30x_siogpio_intr_disable);
EXPORT_SYMBOL(mpu30x_siogpio_intr_set_mode_bit);
EXPORT_SYMBOL(mpu30x_siogpio_intr_set_mode);
EXPORT_SYMBOL(mpu30x_siogpio_intr_register_handler);
EXPORT_SYMBOL(mpu30x_siogpio_intr_unregister_handler);
EXPORT_SYMBOL(mpu30x_siogpio_intr_unmask_bit);
EXPORT_SYMBOL(mpu30x_siogpio_intr_mask_bit);
EXPORT_SYMBOL(mpu30x_siogpio_intr_unmask);
EXPORT_SYMBOL(mpu30x_siogpio_intr_mask);
EXPORT_SYMBOL(mpu30x_siogpio_intr_get_status_bit);
EXPORT_SYMBOL(mpu30x_siogpio_intr_get_status);
EXPORT_SYMBOL(mpu30x_siogpio_intr_clear_bit);
EXPORT_SYMBOL(mpu30x_siogpio_intr_clear);

#ifndef MODULE
__initcall(mpu30x_siogpio_init);
__exitcall(mpu30x_siogpio_cleanup);
#else /* MODULE */
module_init(mpu30x_siogpio_init_module);
module_exit(mpu30x_siogpio_cleanup_module);
#endif /* MODULE */
