/*
 *  snsc_spi.c -- Sony NSC SPI core module
 *
 */
/*
 *  Copyright 2002,2004 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/module.h>
#include <linux/version.h>

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/poll.h>
#include <linux/wrapper.h>
#include <linux/sched.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>
#include <asm/softirq.h>
#include <linux/init.h>
#include <linux/config.h>
#include <linux/list.h>
#include <linux/interrupt.h>

#include <linux/snsc_spi.h>

#ifdef DEBUG
#define DPR(fmt , args...)	printk(fmt, ## args)
#endif /* DEBUG */

MODULE_AUTHOR("Sony CESP");
MODULE_DESCRIPTION("Sony NSC SPI interface");
MODULE_LICENSE("GPL");

/* ======================================================================= *
 * Constant                                                                *
 * ======================================================================= */

#define SNSCSPI_MAJOR		SNSC_SPI_MAJOR	/* major device number */
#define SNSCSPI_DEVNAME		"snscspi"

#define SNSCSPI_DYNAMIC_MAX	127

/* ======================================================================= *
 * Debug print function from dpr package                                   *
 * ======================================================================= */

#define DPL_EMERG	0
#define DPL_ALERT	1
#define DPL_CRIT	2
#define DPL_ERR		3
#define DPL_WARNING	4
#define DPL_NOTICE	5
#define DPL_INFO	6
#define DPL_DEBUG	7

#define DPM_PURPOSE	0x0000000f
#define DPM_ERR		0x00000001
#define DPM_TRACE	0x00000002
#define DPM_LAYER	0x000000f0
#define DPM_APPLICATION	0x00000010
#define DPM_MIDDLEWARE	0x00000020
#define DPM_DRIVER	0x00000040
#define DPM_KERNEL	0x00000080
#define DPM_USERDEF	0xffff0000

#ifdef DEBUG

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL DPL_DEBUG
#endif
#ifndef DEBUG_MASK
#define DEBUG_MASK  DPM_PURPOSE
#endif

#ifndef DPR
#include <stdio.h>
#define DPR(fmt, args...)	fprintf(stderr, fmt , ## args)
#endif

#define DPRL(level, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && DPR(fmt , ## args))
#define DPRM(mask, fmt, args...) \
	(void)(((mask) & DEBUG_MASK) && DPR(fmt , ## args))
#define DPRLM(level, mask, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && ((mask) & DEBUG_MASK) && \
		DPR(fmt , ## args))
#define DPRC(cond, fmt, args...) \
	(void)((cond) && DPR(fmt , ## args))
#define DPRF(fmt, args...)	DPR("%s: " fmt, __FUNCTION__ , ## args)
#define DPRFL(level, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && DPRF(fmt , ## args))
#define DPRFM(mask, fmt, args...) \
	(void)(((mask) & DEBUG_MASK) && DPRF(fmt , ## args))
#define DPRFLM(level, mask, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && ((mask) & DEBUG_MASK) && \
		DPRF(fmt , ## args))
#define DPRFC(cond, fmt, args...) \
	(void)((cond) && DPRF(fmt , ## args))

#define	DPRFS	DPRFLM(DPL_DEBUG, DPM_TRACE, "start\n")
#define	DPRFE	DPRFLM(DPL_DEBUG, DPM_TRACE, "end\n")
#define DPRFR	DPRFLM(DPL_DEBUG, DPM_TRACE, "return\n")

#else  /* DEBUG */

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL -1
#endif
#ifndef DEBUG_MASK
#define DEBUG_MASK  0
#endif

#ifndef DPR
#define DPR(fmt, args...)
#endif

#define DPRL(level, fmt, args...)
#define DPRM(mask, fmt, args...)
#define DPRLM(level, mask, fmt, args...)
#define DPRC(cond, fmt, args...)
#define DPRF(fmt, args...)
#define DPRFL(level, fmt, args...)
#define DPRFM(mask, fmt, args...)
#define DPRFLM(level, mask, fmt, args...)
#define DPRFC(cond, fmt, args...)

#define	DPRFS
#define	DPRFE
#define DPRFR

#endif /* DEBUG */

/* ======================================================================= *
 * type and structure                                                      *
 * ======================================================================= */

typedef struct snscspi_reg_t {
	struct list_head	 list;
	snscspi_bus_t		*bus;
	u_int32_t		 nopen;
	struct semaphore	 semlock;
} snscspi_reg_t;

/* ======================================================================= *
 * prototype declaration                                                   *
 * ======================================================================= */

static int snscspi_init(void);
static void snscspi_final(void);

static int snscspi_open(struct inode *inode, struct file *filp);
static int snscspi_release(struct inode *inode, struct file *filp);
static int snscspi_ioctl(struct inode *inode, struct file *filp,
			  unsigned int cmd, unsigned long arg);
static ssize_t snscspi_read(struct file *filp, char *buf, size_t count,
			    loff_t *f_pos);
static ssize_t snscspi_write(struct file *filp, const char *buf, size_t count,
			     loff_t *f_pos);

static snscspi_reg_t *snscspi_reg_find(int busno);
static void snscspi_reg_add(snscspi_reg_t *entry);
static void snscspi_reg_del(snscspi_reg_t *entry);
static int snscspi_reg_alloc_busno(void);

/* ======================================================================= *
 * global variable                                                         *
 * ======================================================================= */

static int major = SNSCSPI_MAJOR;

static struct list_head	snscspi_reg_head = LIST_HEAD_INIT(snscspi_reg_head);
static rwlock_t snscspi_reg_lock = RW_LOCK_UNLOCKED;

/* ======================================================================= *
 * Linux character driver interface                                        *
 * ======================================================================= */

static struct file_operations snscspi_fops = {
	owner:		THIS_MODULE,
	open:		snscspi_open,
	release:	snscspi_release,
	read:		snscspi_read,
	write:		snscspi_write,
	ioctl:		snscspi_ioctl,
};

static int
snscspi_init(void)
{
	int res;
	DPRFS;

	res = register_chrdev(major, SNSCSPI_DEVNAME, &snscspi_fops);
	if (res < 0) {
		DPRFL(DPL_WARNING,
		      "register_chrdev() failed (err=%d)\n", -res);
		return res;
	}
	if (major == 0) {
		major = res;
	}

	DPRFE;
	return 0;
}

static void
snscspi_final(void)
{
	DPRFS;

	unregister_chrdev(major, SNSCSPI_DEVNAME);

	DPRFE;
	return;
}

module_init(snscspi_init);
module_exit(snscspi_final);

/* ------------------------ *
 * file operation           *
 * ------------------------ */

static int
snscspi_open(struct inode *inode, struct file *filp)
{
	int res;
	snscspi_reg_t *entry;
	snscspi_bus_t *bus;
	DPRFS;

	bus = snscspi_get_bus(MINOR(inode->i_rdev));
	if (bus == NULL) {
		DPRFR;
		return -ENXIO;
	}
	entry = bus->reg;

	filp->private_data = bus;

	if (bus->open != NULL) {
		res = down_interruptible(&(entry->semlock));
		if (res < 0) {
			DPRFR;
			return res;
		}
		res = bus->open(bus);
		up(&(entry->semlock));
		if (res < 0) {
			DPRFR;
			return res;
		}
	}

	entry->nopen++;

	DPRFE;
	return 0;
}

static int
snscspi_release(struct inode *inode, struct file *filp)
{
	int res;
	snscspi_reg_t *entry;
	snscspi_bus_t *bus;
	DPRFS;

	bus = filp->private_data;
	entry = bus->reg;

	if (entry->nopen == 0) {
		DPRFR;
		return -EIO;	/* something wrong */
	}

	if (bus->release != NULL) {
		res = down_interruptible(&(entry->semlock));
		if (res < 0) {
			DPRFR;
			return res;
		}
		res = bus->release(bus);
		up(&(entry->semlock));
		if (res < 0) {
			DPRFR;
			return res;
		}
	}

	entry->nopen--;

	DPRFE;
	return 0;
}

static int
snscspi_ioctl(struct inode *inode, struct file *filp,
	      unsigned int cmd, unsigned long arg)
{
	int res;
	snscspi_reg_t *entry;
	snscspi_bus_t *bus;
	DPRFS;

	bus = filp->private_data;
	entry = bus->reg;

	res = down_interruptible(&(entry->semlock));
	if (res < 0) {
		DPRFR;
		return res;
	}
	res = bus->ioctl(bus, cmd, (void *)arg, 0);
	up(&(entry->semlock));

	DPRFE;
	return res;
}

static ssize_t
snscspi_read(struct file *filp, char *buf, size_t count, loff_t *f_pos)
{
	int res;
	snscspi_reg_t *entry;
	snscspi_bus_t *bus;
	snscspi_transaction_t trans;
	snscspi_msg_t msg;
	DPRFS;

	bus = filp->private_data;
	entry = bus->reg;

	memset(&msg, 0, sizeof(msg));
	msg.flags = SNSCSPI_M_RD;
	msg.len = count;
	msg.buf = buf;

	trans.nmsgs = 1;
	trans.msgs = &msg;

	res = down_interruptible(&(entry->semlock));
	if (res < 0) {
		DPRFR;
		return res;
	}
	res = bus->ioctl(bus, SNSCSPI_IOC_TRANSACTION, &trans,
			 SNSCSPI_T_ARG_KERNEL | SNSCSPI_T_BUF_USER);
	up(&(entry->semlock));

	DPRFE;
	return count;
}

static ssize_t
snscspi_write(struct file *filp, const char *buf, size_t count, loff_t *f_pos)
{
	int res;
	snscspi_reg_t *entry;
	snscspi_bus_t *bus;
	snscspi_transaction_t trans;
	snscspi_msg_t msg;
	DPRFS;

	bus = filp->private_data;
	entry = bus->reg;

	memset(&msg, 0, sizeof(msg));
	msg.flags = SNSCSPI_M_WR;
	msg.len = count;
	msg.buf = (void *)buf;

	trans.nmsgs = 1;
	trans.msgs = &msg;

	res = down_interruptible(&(entry->semlock));
	if (res < 0) {
		DPRFR;
		return res;
	}
	res = bus->ioctl(bus, SNSCSPI_IOC_TRANSACTION, &trans,
			 SNSCSPI_T_ARG_KERNEL | SNSCSPI_T_BUF_USER);
	up(&(entry->semlock));

	DPRFE;
	return count;
}

/* ======================================================================= *
 * kernel interface                                                        *
 * ======================================================================= */

int
snscspi_do_ioctl(snscspi_bus_t *bus, unsigned int cmd, void *arg)
{
	int res;
	snscspi_reg_t *entry;

	entry = bus->reg;

	res = down_interruptible(&(entry->semlock));
	if (res < 0) {
		DPRFR;
		return res;
	}
	res = bus->ioctl(bus, cmd, arg, SNSCSPI_T_ARG_KERNEL);
	up(&(entry->semlock));

	DPRFE;
	return res;
}

snscspi_bus_t *
snscspi_get_bus(int busno)
{
	snscspi_reg_t *entry;
	snscspi_bus_t *bus;
	DPRFS;

	read_lock(&snscspi_reg_lock);
	entry = snscspi_reg_find(busno);
	if (entry == NULL) {
		read_unlock(&snscspi_reg_lock);
		DPRFR;
		return NULL;
	}
	bus = entry->bus;
	read_unlock(&snscspi_reg_lock);

	DPRFE;
	return bus;
}

EXPORT_SYMBOL(snscspi_do_ioctl);
EXPORT_SYMBOL(snscspi_get_bus);

/* ======================================================================= *
 * Interface for low-level driver                                          *
 * ======================================================================= */

int
snscspi_register(snscspi_bus_t *bus)
{
	int res;
	snscspi_reg_t *entry;
	DPRFS;

	if (bus == NULL) {
		res = -EINVAL;
		goto err;
	}
	entry = kmalloc(sizeof(*entry), GFP_KERNEL);
	if (entry == NULL) {
		res = -ENOMEM;
		goto err;
	}

	write_lock(&snscspi_reg_lock);
	res = snscspi_reg_alloc_busno();
	if (res < 0) {
		goto err_kfree_entry;
	}
	bus->busno = res;

	entry->bus = bus;
	entry->nopen = 0;
	init_MUTEX(&(entry->semlock));

	bus->reg = entry;

	snscspi_reg_add(entry);
	write_unlock(&snscspi_reg_lock);

	MOD_INC_USE_COUNT;

	DPRFE;
	return bus->busno;

err_kfree_entry:
	write_unlock(&snscspi_reg_lock);
	kfree(entry);
err:
	DPRFR;
	return res;
}

int
snscspi_unregister(snscspi_bus_t *bus)
{
	int res;
	snscspi_reg_t *entry;
	DPRFS;

	entry = bus->reg;
	if (entry == NULL) {
		DPRFR;
		return -EINVAL;
	}
	bus->reg = NULL;

	res = down_interruptible(&(entry->semlock));
	if (res < 0) {
		DPRFR;
		return res;
	}
	write_lock(&snscspi_reg_lock);
	snscspi_reg_del(entry);
	write_unlock(&snscspi_reg_lock);
	up(&(entry->semlock));

	kfree(entry);

	MOD_DEC_USE_COUNT;

	DPRFE;
	return 0;
}

EXPORT_SYMBOL(snscspi_register);
EXPORT_SYMBOL(snscspi_unregister);

/* ======================================================================= *
 * Internal function                                                       *
 * ======================================================================= */

/* should be called with read_lock of snscspi_reg_lock */
static snscspi_reg_t *
snscspi_reg_find(int busno)
{
	struct list_head *p;
	snscspi_reg_t    *entry;

	list_for_each(p, &snscspi_reg_head) {
		entry = list_entry(p, snscspi_reg_t, list);
		if (entry->bus->busno == busno) {
			return entry;
		}
	}
	return NULL;
}

/* should be called with write_lock of snscspi_reg_lock */
static void
snscspi_reg_add(snscspi_reg_t *entry)
{
	list_add(&entry->list, &snscspi_reg_head);
}

/* should be called with write_lock of snscspi_reg_lock */
static void
snscspi_reg_del(snscspi_reg_t *entry)
{
	list_del(&entry->list);
}

/* should be called with read_lock of snscspi_reg_lock */
static int
snscspi_reg_alloc_busno(void)
{
	int i;
	snscspi_reg_t *entry;

	/**** TODO: fix me */
	for (i = 0; i <= SNSCSPI_DYNAMIC_MAX; i++) {
		entry = snscspi_reg_find(i);
		if (entry == NULL) {
			return i;
		}
	}

	return -ENOSPC;
}
