/*
 *  Sony NSC UMA
 *
 *  Copyright 2005 Sony Corporation
 *
 *  This program is free software; you can redistribute	 it and/or modify it
 *  under  the terms of	 the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the	License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED	  ``AS	IS'' AND   ANY	EXPRESS OR IMPLIED
 *  WARRANTIES,	  INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO	EVENT  SHALL   THE AUTHOR  BE	 LIABLE FOR ANY	  DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED	  TO, PROCUREMENT OF  SUBSTITUTE GOODS	OR SERVICES; LOSS OF
 *  USE, DATA,	OR PROFITS; OR	BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN	 CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/module.h>
#include <linux/seq_file.h>
#include <linux/kernel.h>
#include <linux/bootmem.h>
#include <asm/io.h>
#include <asm/string.h>
#include <asm/uaccess.h>
#include <linux/snsc_nblargs.h>
#include <linux/snsc_uma.h>


#define DEBUG_FLAG	1
#define DEBUG_NAME	"UMA"

#define ERROR_STRING	"#Error:"

#define PRINTK(format, args...) printk(DEBUG_NAME ":" format, ## args)
#define ERROR(format, args...) printk(ERROR_STRING DEBUG_NAME ":%s():" format , __FUNCTION__ , ## args)

#ifdef DEBUG
#undef DEBUG
#endif

#if DEBUG_FLAG & 1
#define PRINT(format, args...) printk(DEBUG_NAME ":%s():" format , __FUNCTION__ , ## args)
#else
#define PRINT(format, args...)
#endif

#if DEBUG_FLAG & 2
#define DEBUG(format, args...) printk(DEBUG_NAME ":%s():%d:" format , __FUNCTION__ , __LINE__ , ## args)
#else
#define DEBUG(format, args...)
#endif

#if DEBUG_FLAG & 4
#define ENTRY() printk(DEBUG_NAME ":%s():entry\n", __FUNCTION__)
#else
#define ENTRY()
#endif

extern struct snsc_uma_machdep_calls uma_md;


#define CHECK_UMA_FUNC(uma_md)				\
	if ((uma_md.init_range == NULL) ||		\
	    (uma_md.get_range == NULL)) {		\
		ERROR("Target dependent UMA operators not implemented!!" \
		      " Disabling UMA.\n");		\
		return -ENOENT;				\
	}

#define CHECK_UMA_OCCUPIED(pd)				\
	if ((pd.flags & SNSC_UMA_FLG_OCCUPIED)		\
	    == SNSC_UMA_FLG_OCCUPIED) {			\
		ERROR("specified region %d is occupied.\n", region); \
		retval = -EBUSY;			\
		goto exit;				\
	}


/* 
   Initializes UMA by adding entries to NBLArgs.
   Should only be called from nblargs_init().
 */
int
snsc_uma_init(void)
{
	int i;
	struct nblargs_entry e;
	snsc_uma_desc_t pd;

	ENTRY();
	CHECK_UMA_FUNC(uma_md);

	i = 0;
	while (uma_md.init_range(i++) >= 0);

	i = 0;
	memset(&e, 0, sizeof(struct nblargs_entry));
	while (uma_md.get_range(i, &pd) >= 0) {
		e.keylen = snprintf(e.key, NBLARGS_KEY_LEN - 1,
				    "uma%d", i) + 1;
		e.size = pd.size;
		e.addr = pd.addr;
		DEBUG("name=%s keylen=%d addr=0x%x size=0x%x\n",
		      e.key, e.keylen, e.addr, e.size);
		nblargs_add(&e, 1);
		i++;
	}

	return 0;
}

int
snsc_uma_getsize(snsc_uma_t region, unsigned int *size)
{
	snsc_uma_desc_t pd;
	unsigned long flags;
	int retval = 0;

	ENTRY();
	CHECK_UMA_FUNC(uma_md);

	if (size == NULL) {
		ERROR("*size is NULL.\n");
		return -EFAULT;
	}

	local_irq_save(flags);
	if (uma_md.get_range(region, &pd) < 0) {
		ERROR("specified region %d does not exist.\n", region);
		retval = -EINVAL;
		goto exit;
	}

	*size = pd.size;

	CHECK_UMA_OCCUPIED(pd);
exit:
	local_irq_restore(flags);
	return retval;
}
EXPORT_SYMBOL(snsc_uma_getsize);

int
snsc_uma_request(snsc_uma_t region, unsigned int pages,
		 void *vaddr, dma_addr_t *paddr)
{
	snsc_uma_desc_t pd;
	unsigned long flags;
	unsigned int free_start, free_end, free_size;;
	int retval = 0;

	ENTRY();
	CHECK_UMA_FUNC(uma_md);

	if ((vaddr == NULL) || paddr == NULL) {
		ERROR("vaddr or paddr is NULL.\n");
		return -EFAULT;
	}

	local_irq_save(flags);
	memset(&pd, 0, sizeof(pd));
	if (uma_md.get_range(region, &pd) < 0) {
		ERROR("specified region %d does not exist.\n", region);
		retval = -EINVAL;
		goto exit;
	}
	DEBUG("pd.addr =0x%08x\n", pd.addr);
	DEBUG("pd.size =0x%08x\n", pd.size);
	DEBUG("pd.flags=0x%08x\n", pd.flags);
	CHECK_UMA_OCCUPIED(pd);

	if (PAGE_SIZE * pages > pd.size) {
		ERROR("requested size (%d pages) too large.\n", pages);
		retval = -ENOMEM;
		goto exit;
	}

	free_size = pd.size - PAGE_SIZE * pages;
	DEBUG("free_size  = 0x%x\n", free_size);
	if ((pd.flags & SNSC_UMA_FLG_FREE_DIR) == 
	    SNSC_UMA_FLG_FREE_FROM_TAIL) {
		free_end   = nbl_to_va(pd.addr) + pd.size;
		free_start = free_end - free_size;
	} else {
		free_start = nbl_to_va(pd.addr);
		free_end   = free_start + free_size;
		pd.addr   += free_size;
	}
	DEBUG("free_start = 0x%x\n", free_start);
	DEBUG("free_end   = 0x%x\n", free_end);
	PRINTK("%dk freed\n", (free_end - free_start) >> 10);
	for (; free_start < free_end; free_start += PAGE_SIZE) {
		extern unsigned long totalram_pages;
		ClearPageReserved(virt_to_page(free_start));
		set_page_count(virt_to_page(free_start), 1);
		free_page(free_start);
		totalram_pages++;
	}

	pd.size  -= free_size;
	pd.flags |= SNSC_UMA_FLG_OCCUPIED;
	uma_md.set_range(region, pd);

	*(unsigned long *)vaddr = nbl_to_va(pd.addr);
	*paddr = nbl_to_pa(pd.addr);

exit:
	local_irq_restore(flags);
	return retval;
}
EXPORT_SYMBOL(snsc_uma_request);

int
snsc_uma_free(snsc_uma_t region)
{
	snsc_uma_desc_t pd;
	unsigned long flags;
	int retval = 0;

	ENTRY();
	CHECK_UMA_FUNC(uma_md);

	local_irq_save(flags);
	if (uma_md.get_range(region, &pd) < 0) {
		ERROR("specified region %d does not exist.\n", region);
		retval = -EINVAL;
		goto exit;
	}

	pd.flags &= ~SNSC_UMA_FLG_OCCUPIED;
	uma_md.set_range(region, pd);

exit:
	local_irq_restore(flags);
	return retval;
}
EXPORT_SYMBOL(snsc_uma_free);

void
snsc_uma_sync(void *addr, unsigned int size, unsigned int direction)
{
	if (uma_md.sync)
		uma_md.sync(addr, size, direction);
}
EXPORT_SYMBOL(snsc_uma_sync);

int
snsc_uma_sync_all(snsc_uma_t region)
{
	snsc_uma_desc_t pd;
	int retval = 0;

	ENTRY();
	CHECK_UMA_FUNC(uma_md);

	if (uma_md.get_range(region, &pd) < 0) {
		ERROR("specified region %d does not exist.\n", region);
		retval = -EINVAL;
		goto exit;
	}

	if (uma_md.sync)
		uma_md.sync((void *)pd.addr, pd.size, SNSC_UMA_WRITE);

exit:
	return retval;
}
EXPORT_SYMBOL(snsc_uma_sync_all);
