/*
 *  WatchDog Timer driver for SNSC MPU-xxx series
 *
 *  Copyright 2003 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

//#define DEBUG 1

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/snsc_watchdog.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/processor.h>


#ifdef DEBUG
#define DPR(fmt, args...)	printk(fmt, ## args)
#endif /* DEBUG */

/* ======================================================================= *
 * Constants                                                               *
 * ======================================================================= */

#define WDT_DEVNAME	"wdt"

#define WDT_DIRNAME		"wdt"
#define	WDT_COUNTER_FILENAME	"counter"
#define	WDT_RESET_CAUSE_FILENAME	"reset_cause"

MODULE_AUTHOR("Sony PTC");
MODULE_DESCRIPTION("SNSC WatchDog Timer driver");
MODULE_LICENSE("GPL");

/* ======================================================================= *
 * Debug print function from dpr package                                   *
 * ======================================================================= */

#define DPL_EMERG	0
#define DPL_ALERT	1
#define DPL_CRIT	2
#define DPL_ERR		3
#define DPL_WARNING	4
#define DPL_NOTICE	5
#define DPL_INFO	6
#define DPL_DEBUG	7

#define DPM_PORPOSE	0x0000000f
#define DPM_ERR		0x00000001
#define DPM_TRACE	0x00000002
#define DPM_LAYER	0x000000f0
#define DPM_APPLICATION	0x00000010
#define DPM_MIDDLEWARE	0x00000020
#define DPM_DRIVER	0x00000040
#define DPM_KERNEL	0x00000080
#define DPM_USERDEF	0xffff0000

#ifdef DEBUG

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL DPL_DEBUG
#endif
#ifndef DEBUG_MASK
#define DEBUG_MASK  DPM_PORPOSE
#endif

#ifndef DPR
#include <stdio.h>
#define DPR(fmt, args...)	fprintf(stderr, fmt, ## args)
#endif

#define DPRL(level, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && DPR(fmt, ## args))
#define DPRM(mask, fmt, args...) \
	(void)(((mask) & DEBUG_MASK) && DPR(fmt, ## args))
#define DPRLM(level, mask, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && ((mask) & DEBUG_MASK) && \
		DPR(fmt, ## args))
#define DPRC(cond, fmt, args...) \
	(void)((cond) && DPR(fmt, ## args))
#define DPRF(fmt, args...)	DPR("%s: " fmt, __FUNCTION__ , ## args)
#define DPRFL(level, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && DPRF(fmt, ## args))
#define DPRFM(mask, fmt, args...) \
	(void)(((mask) & DEBUG_MASK) && DPRF(fmt, ## args))
#define DPRFLM(level, mask, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && ((mask) & DEBUG_MASK) && \
		DPRF(fmt, ## args))
#define DPRFC(cond, fmt, args...) \
	(void)((cond) && DPRF(fmt, ## args))

#define	DPRFS	DPRFLM(DPL_DEBUG, DPM_TRACE, "start\n")
#define	DPRFE	DPRFLM(DPL_DEBUG, DPM_TRACE, "end\n")
#define DPRFR	DPRFLM(DPL_DEBUG, DPM_TRACE, "return\n")

#else  /* DEBUG */

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL -1
#endif
#ifndef DEBUG_MASK
#define DEBUG_MASK  0
#endif

#ifndef DPR
#define DPR(fmt, args...)
#endif

#define DPRL(level, fmt, args...)
#define DPRM(mask, fmt, args...)
#define DPRLM(level, mask, fmt, args...)
#define DPRC(cond, fmt, args...)
#define DPRF(fmt, args...)
#define DPRFL(level, fmt, args...)
#define DPRFM(mask, fmt, args...)
#define DPRFLM(level, mask, fmt, args...)
#define DPRFC(cond, fmt, args...)

#define	DPRFS
#define	DPRFE
#define DPRFR

#endif /* DEBUG */

/* ======================================================================= *
 * Types and structures                                                    *
 * ======================================================================= */

/* ======================================================================= *
 * Prototype declarations                                                  *
 * ======================================================================= */

static int wdt_init(void);
static void wdt_exit(void);

static ssize_t wdt_counter_read(struct file *filp, char *buf,
				       size_t count, loff_t *f_pos);
static ssize_t wdt_counter_write(struct file *filp, const char *buf,
					size_t count, loff_t *f_pos);

static ssize_t wdt_reset_cause_read(struct file *filp, char *buf,
					   size_t count, loff_t *f_pos);

/* internal function */
static int wdt_skip_whitespaces_user(const char *user_buf_pos,
					    size_t count);
static int wdt_skip_to_lineend_user(const char *user_buf_pos,
					   size_t count);
static int wdt_find_next_whitespace(const char *str);

/* ======================================================================= *
 * Global variables                                                        *
 * ======================================================================= */

static struct watchdog_md *wdt;

static struct proc_dir_entry *wdt_proc_root;
static int wdt_reset_cause_is_wdt = 0;

/* ======================================================================= *
 * Inline functions                                                        *
 * ======================================================================= */

static inline int
wdt_is_whitespace(unsigned char c)
{
	return (c == 0x20 || c == 0x09 || c == 0x0d || c == 0x0a);
}


/* ======================================================================= *
 * Linux driver interfaces                                                 *
 * ======================================================================= */

static struct file_operations wdt_counter_ops = {
	owner:		THIS_MODULE,
	read:		wdt_counter_read,
	write:		wdt_counter_write,
};

static struct file_operations wdt_reset_cause_ops = {
	owner:		THIS_MODULE,
	read:		wdt_reset_cause_read,
};

/* ------------------------ *
 * initialize/finalize      *
 * ------------------------ */

static int
wdt_init(void)
{
	int res;
	struct proc_dir_entry *ent;
	DPRFS;

	/* initialize */
	wdt = mach_watchdog_init();

	/* check reset cause */
	wdt_reset_cause_is_wdt = wdt->get_reset_cause();

	/* clear reset cause */
	wdt->clear_reset_cause();

	/* make proc directory */
	wdt_proc_root = proc_mkdir(WDT_DIRNAME,
					  proc_root_driver);
	if (wdt_proc_root == NULL) {
		DPRFL(DPL_WARNING, "cannot create proc directory\n");
		res = -ENOMEM;
		goto err1;
	}

	/* make counter file */
	ent = create_proc_entry(WDT_COUNTER_FILENAME,
				S_IRUGO | S_IWUSR, wdt_proc_root);
	if (ent == NULL) {
		DPRFL(DPL_WARNING, "cannot create counter file.\n");
		res = -ENOMEM;
		goto err2;
	}
	ent->proc_fops = &wdt_counter_ops;

	/* make reset_cause file */
	ent = create_proc_entry(WDT_RESET_CAUSE_FILENAME,
				S_IRUGO, wdt_proc_root);
	if (ent == NULL) {
		DPRFL(DPL_WARNING, "cannot create reset_cause file.\n");
		res = -ENOMEM;
		goto err3;
	}
	ent->proc_fops = &wdt_reset_cause_ops;

	DPRFE;
	return 0;

err3:
	remove_proc_entry(WDT_COUNTER_FILENAME, wdt_proc_root);
err2:
	remove_proc_entry(WDT_DIRNAME, proc_root_driver);
err1:
	DPRFR;
	return res;
}

static void
wdt_exit(void)
{
	DPRFS;

	remove_proc_entry(WDT_RESET_CAUSE_FILENAME,
			  wdt_proc_root);
	remove_proc_entry(WDT_COUNTER_FILENAME, wdt_proc_root);
	remove_proc_entry(WDT_DIRNAME, proc_root_driver);

	DPRFE;
	return;
}

module_init(wdt_init);
module_exit(wdt_exit);

/* ------------------------ *
 * counter                  *
 * ------------------------ */

static ssize_t
wdt_counter_read(struct file *filp, char *buf,
			size_t count, loff_t *f_pos)
{
	int res;
	unsigned long rest_sec;
	char str_to_write[16];
	int str_len;
	DPRFS;

	if (count == 0 || *f_pos != 0) {
		DPRFR;
		return 0;
	}

	rest_sec = 0;

	if (wdt->is_enabled()) {
		rest_sec = wdt->get_count();
	}

	str_len = snprintf(str_to_write, sizeof(str_to_write),
			   "%ld\n", rest_sec);

	if (count < str_len) {
		DPRFR;
		return -EINVAL;
	}
	res = copy_to_user(buf, str_to_write, str_len);
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}
	*f_pos += str_len;

	DPRFE;
	return str_len;
}

static ssize_t
wdt_counter_write(struct file *filp, const char *buf,
			 size_t count, loff_t *f_pos)
{
	int res;
	char tmp_buf[16];
	ssize_t nread;
	int ncopy;
	unsigned long counter_sec;
	unsigned long long tmp64;
	char *endp;
	unsigned int flags;
	DPRFS;

	if (count == 0) {
		DPRFR;
		return 0;
	}

	/* parse number */
	nread = 0;
	res = wdt_skip_whitespaces_user(buf, count);
	if (res < 0) {
		DPRFR;
		return res;
	}
	nread += res;
	buf += res;
	count -= res;

	ncopy = sizeof(tmp_buf) - 1;
	if (count < ncopy) {
		ncopy = count;
	}
	res = copy_from_user(tmp_buf, buf, ncopy);
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}
	tmp_buf[ncopy] = 0;
	res = wdt_find_next_whitespace(tmp_buf);
	if (res >= 0) {
		tmp_buf[res] = 0;
	}

	tmp64 = simple_strtoull(tmp_buf, &endp, 0);
	if (*endp != 0) {
		DPRFR;
		return -EFAULT;
	}
	if (tmp64 > ULONG_MAX) {
		DPRFR;
		return -EFAULT;
	}
	counter_sec = tmp64;
	if (counter_sec  > SNSC_WATCHDOG_MAX) {
		DPRFR;
		return -EINVAL;
	}

	res = wdt_skip_to_lineend_user(buf, count);
	if (res < 0) {
		DPRFR;
		return res;
	}
	nread += res;
	buf += res;
	count -= res;

	DPRFL(DPL_DEBUG, "counter_sec=%ld\n", counter_sec);

	/* set watchdog registers */
	if (counter_sec == 0) {
		/* disable watchdog timer */
		DPRFL(DPL_DEBUG, "disable WDT\n");

		local_irq_save(flags);
		wdt->disable();
		local_irq_restore(flags);
	} else {
		/* set and enable watchdog timer */
		DPRFL(DPL_DEBUG, "enable WDT\n");

		local_irq_save(flags);

		wdt->heartbeat();
		wdt->set_count(counter_sec);
		wdt->enable();

		local_irq_restore(flags);
	}
	*f_pos += nread;

	DPRFE;
	return nread;
}

/* ------------------------ *
 * reset_cause              *
 * ------------------------ */

static ssize_t
wdt_reset_cause_read(struct file *filp, char *buf,
			    size_t count, loff_t *f_pos)
{
	int res;
	const char *str_to_write;
	ssize_t nwrite;
	DPRFS;

	if (count == 0 || *f_pos != 0) {
		DPRFR;
		return 0;
	}

	if (wdt_reset_cause_is_wdt) {
		str_to_write = "1\n";
	} else {
		str_to_write = "0\n";
	}
	if (count == 1) {
		nwrite = 1;
	} else {
		nwrite = 2;
	}

	res = copy_to_user(buf, str_to_write, nwrite);
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}
	*f_pos += nwrite;

	DPRFE;
	return nwrite;
}

/* ======================================================================= *
 * Internal function                                                       *
 * ======================================================================= */

static int
wdt_skip_whitespaces_user(const char *user_buf_pos, size_t count)
{
	int res;
	unsigned char c;
	const char *p;
	DPRFS;

	p = user_buf_pos;
	while (count > 0) {
		res = get_user(c, p);
		if (res != 0) {
			DPRFR;
			return -EFAULT;
		}
		if (!wdt_is_whitespace(c)) {
			break;
		}
		p++;
		count--;
	}

	DPRFE;
	return p - user_buf_pos;
}

static int
wdt_skip_to_lineend_user(const char *user_buf_pos, size_t count)
{
	int res;
	unsigned char c;
	const char *p;
	DPRFS;

	p = user_buf_pos;
	while (count > 0) {
		res = get_user(c, p);
		if (res != 0) {
			DPRFR;
			return -EFAULT;
		}
		p++;
		count--;
		if (c == 0x0d) {
			/* LF */
			break;
		}
		if (c == 0x0a) {
			/* CR */
			if (count > 0) {
				res = get_user(c, p);
				if (res != 0) {
					DPRFR;
					return -EFAULT;
				}
				if (c == 0x0d) {
					/* LF */
					p++;
					count--;
				}
			}
			break;
		}
	}

	DPRFE;
	return p - user_buf_pos;
}

static int
wdt_find_next_whitespace(const char *str)
{
	int res;
	int pos;
	DPRFS;

	res = -ENOENT;
	pos = 0;
	while (*str != 0) {
		if (wdt_is_whitespace(*str)) {
			res = pos;
			break;
		}
		str++;
		pos++;
	}

	DPRFE;
	return res;
}

