/*
 *  WatchDog Timer driver for TX4927 series
 *
 *  Copyright 2003 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* #define DEBUG 1 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>

#include <asm/tx4927/tx4927.h>

#ifdef DEBUG
#define DPR(fmt, args...)	printk(fmt, ## args)
#endif /* DEBUG */

/* ======================================================================= *
 * Constant                                                                *
 * ======================================================================= */

#define TX4927_WDT_DEVNAME	"wdt"

#define TX4927_WDT_DIRNAME		"wdt"
#define	TX4927_WDT_COUNTER_FILENAME	"counter"
#define	TX4927_WDT_CONTROL_FILENAME	"control"
#define	TX4927_WDT_RESET_CAUSE_FILENAME	"reset_cause"

MODULE_AUTHOR("Sony PTC");
MODULE_DESCRIPTION("TX4927 WatchDog Timer driver");
MODULE_LICENSE("GPL");

/* -------------------------- *
 * TX4927's register          *
 * -------------------------- */
/*
 * TX4927R_xxx: configuration/status register offset
 * TX4927M_xxx: mask to locate subfield in register
 * TX4927C_xxx: constant to locate subfield in register
 * TX4927S_xxx: set a value to subfield in register
 */

/* configuration registers */

#define TX4927R_CONFIG_CCFG		TX4927_CONFIG_CCFG
#define  TX4927M_CONFIG_WDRST		 (1LL << 41)
#define  TX4927M_CONFIG_WDREXEN		 (1LL << 40)
#define  TX4927M_CONFIG_BCFG		 (0xffLL << 32)
#define  TX4927M_CONFIG_GTOT		 (0x3 << 25)
#define  TX4927S_CONFIG_GTOT(x)		 ((x) << 25)
#define  TX4927M_CONFIG_TINTDIS		 (1 << 24)
#define  TX4927M_CONFIG_PCI66		 (1 << 23)
#define  TX4927M_CONFIG_PCIMODE		 (1 << 22)
#define  TX4927M_CONFIG_DIVMODE		 (0xf << 17)
#define  TX4927S_CONFIG_DIVMODE(x)	 ((x) << 17)
#define  TX4927M_CONFIG_BEOW		 (1 << 16)
#define  TX4927M_CONFIG_WR		 (1 << 15)
#define  TX4927M_CONFIG_TOE		 (1 << 14)
#define  TX4927M_CONFIG_PCIARB		 (1 << 13)
#define  TX4927M_CONFIG_PCIDIVMODE	 (0x7 << 10)
#define  TX4927S_CONFIG_PCIDIVMODE(x)	 ((x) << 10)
#define  TX4927M_CONFIG_SYSSP		 (0x3 << 6)
#define  TX4927S_CONFIG_SYSSP(x)	 ((x) << 6)
#define  TX4927M_CONFIG_ENDIAN		 (1 << 2)
#define  TX4927M_CONFIG_ARMODE		 (1 << 1)
#define  TX4927M_CONFIG_ACEHOLD		 (1 << 0)
#define TX4927R_CONFIG_REVID		TX4927_CONFIG_REVID
#define  TX4927M_CONFIG_PCODE		 (0xffff << 16)
#define  TX4927S_CONFIG_PCODE(x)	 ((x) << 16)
#define  TX4927M_CONFIG_MJERREV		 (0xf << 12)
#define  TX4927S_CONFIG_MJERREV(x)	 ((x) << 12)
#define  TX4927M_CONFIG_MINEREV		 (0xf << 8)
#define  TX4927S_CONFIG_MINEREV(x)	 ((x) << 8)
#define  TX4927M_CONFIG_MJREV		 (0xf << 4)
#define  TX4927S_CONFIG_MJREV(x)	 ((x) << 4)
#define  TX4927M_CONFIG_MINREV		 (0xf << 0)
#define  TX4927S_CONFIG_MINREV(x)	 ((x) << 0)

/* timer registers */
#define TX4927R_TIMER_TMTCR(ch)		(0xf000 + 0x100 * (ch))
#define  TX4927M_TIMER_TCE		 (1 << 7)
#define  TX4927M_TIMER_CCDE		 (1 << 6)
#define  TX4927M_TIMER_CRE		 (1 << 5)
#define  TX4927M_TIMER_ECES		 (1 << 3)
#define  TX4927M_TIMER_CCS		 (1 << 2)
#define  TX4927M_TIMER_TMODE		 (0x3 << 0)
#define  TX4927S_TIMER_TMODE(x)		 ((x) << 0)
#define  TX4927C_TIMER_TMODE_INTERVAL	 TX4927S_TIMER_TMODE(0)
#define  TX4927C_TIMER_TMODE_PULSE	 TX4927S_TIMER_TMODE(1)
#define  TX4927C_TIMER_TMODE_WATCHDOG	 TX4927S_TIMER_TMODE(2)
#define TX4927R_TIMER_TMTISR(ch)	(0xf004 + 0x100 * (ch))
#define  TX4927M_TIMER_TWIS		 (1 << 3)
#define  TX4927M_TIMER_TPIBS		 (1 << 2)
#define  TX4927M_TIMER_TPIAS		 (1 << 1)
#define  TX4927M_TIMER_TIIS		 (1 << 0)
#define TX4927R_TIMER_TMCPRA(ch)	(0xf008 + 0x100 * (ch))
#define TX4927R_TIMER_TMCPRB(ch)	(0xf00c + 0x100 * (ch))
#define TX4927R_TIMER_TMITMR(ch)	(0xf010 + 0x100 * (ch))
#define  TX4927M_TIMER_TIIE		 (1 << 15)
#define  TX4927M_TIMER_TZCE		 (1 << 0)
#define TX4927R_TIMER_TMCCDR(ch)	(0xf020 + 0x100 * (ch))
#define  TX4927M_TIMER_CCD		 (0x7 << 0)
#define  TX4927S_TIMER_CCD(x)		 ((x) << 0)
#define TX4927R_TIMER_TMPGMR(ch)	(0xf030 + 0x100 * (ch))
#define  TX4927M_TIMER_TPIBE		 (1 << 15)
#define  TX4927M_TIMER_TPIAE		 (1 << 14)
#define  TX4927M_TIMER_FFI		 (1 << 0)
#define TX4927R_TIMER_TMWTMR(ch)	(0xf040 + 0x100 * (ch))
#define  TX4927M_TIMER_TWIE		 (1 << 15)
#define  TX4927M_TIMER_WDIS		 (1 << 7)
#define  TX4927M_TIMER_TWC		 (1 << 0)
#define TX4927R_TIMER_TMTRR(ch)		(0xf0f0 + 0x100 * (ch))

/* ======================================================================= *
 * Debug print function from dpr package                                   *
 * ======================================================================= */

#define DPL_EMERG	0
#define DPL_ALERT	1
#define DPL_CRIT	2
#define DPL_ERR		3
#define DPL_WARNING	4
#define DPL_NOTICE	5
#define DPL_INFO	6
#define DPL_DEBUG	7

#define DPM_PORPOSE	0x0000000f
#define DPM_ERR		0x00000001
#define DPM_TRACE	0x00000002
#define DPM_LAYER	0x000000f0
#define DPM_APPLICATION	0x00000010
#define DPM_MIDDLEWARE	0x00000020
#define DPM_DRIVER	0x00000040
#define DPM_KERNEL	0x00000080
#define DPM_USERDEF	0xffff0000

#ifdef DEBUG

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL DPL_DEBUG
#endif
#ifndef DEBUG_MASK
#define DEBUG_MASK  DPM_PORPOSE
#endif

#ifndef DPR
#include <stdio.h>
#define DPR(fmt, args...)	fprintf(stderr, fmt, ## args)
#endif

#define DPRL(level, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && DPR(fmt, ## args))
#define DPRM(mask, fmt, args...) \
	(void)(((mask) & DEBUG_MASK) && DPR(fmt, ## args))
#define DPRLM(level, mask, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && ((mask) & DEBUG_MASK) && \
		DPR(fmt, ## args))
#define DPRC(cond, fmt, args...) \
	(void)((cond) && DPR(fmt, ## args))
#define DPRF(fmt, args...)	DPR("%s: " fmt, __FUNCTION__, ## args)
#define DPRFL(level, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && DPRF(fmt, ## args))
#define DPRFM(mask, fmt, args...) \
	(void)(((mask) & DEBUG_MASK) && DPRF(fmt, ## args))
#define DPRFLM(level, mask, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && ((mask) & DEBUG_MASK) && \
		DPRF(fmt, ## args))
#define DPRFC(cond, fmt, args...) \
	(void)((cond) && DPRF(fmt, ## args))

#define	DPRFS	DPRFLM(DPL_DEBUG, DPM_TRACE, "start\n")
#define	DPRFE	DPRFLM(DPL_DEBUG, DPM_TRACE, "end\n")
#define DPRFR	DPRFLM(DPL_DEBUG, DPM_TRACE, "return\n")

#else  /* DEBUG */

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL -1
#endif
#ifndef DEBUG_MASK
#define DEBUG_MASK  0
#endif

#ifndef DPR
#define DPR(fmt, args...)
#endif

#define DPRL(level, fmt, args...)
#define DPRM(mask, fmt, args...)
#define DPRLM(level, mask, fmt, args...)
#define DPRC(cond, fmt, args...)
#define DPRF(fmt, args...)
#define DPRFL(level, fmt, args...)
#define DPRFM(mask, fmt, args...)
#define DPRFLM(level, mask, fmt, args...)
#define DPRFC(cond, fmt, args...)

#define	DPRFS
#define	DPRFE
#define DPRFR

#endif /* DEBUG */

/* ======================================================================= *
 * Type and structure                                                      *
 * ======================================================================= */

/* ======================================================================= *
 * Prototype declaration                                                   *
 * ======================================================================= */

static int tx4927_wdt_init(void);
static void tx4927_wdt_exit(void);

static ssize_t tx4927_wdt_counter_read(struct file *filp, char *buf,
				       size_t count, loff_t *f_pos);
static ssize_t tx4927_wdt_counter_write(struct file *filp, const char *buf,
					size_t count, loff_t *f_pos);

static ssize_t tx4927_wdt_control_write(struct file *filp, const char *buf,
					size_t count, loff_t *f_pos);

static ssize_t tx4927_wdt_reset_cause_read(struct file *filp, char *buf,
					   size_t count, loff_t *f_pos);

/* internal function */
static int tx4927_wdt_skip_whitespaces_user(const char *user_buf_pos,
					    size_t count);
static int tx4927_wdt_skip_to_lineend_user(const char *user_buf_pos,
					   size_t count);
static int tx4927_wdt_find_next_whitespace(const char *str);
static void tx4927_wdt_set_wdt_mode(int do_reset);

/* ======================================================================= *
 * Global variable                                                         *
 * ======================================================================= */

static struct proc_dir_entry *tx4927_wdt_proc_root;

static int tx4927_wdt_reset_cause_is_wdt = 0;

/* ======================================================================= *
 * Inline function                                                         *
 * ======================================================================= */

static inline int
tx4927_wdt_is_whitespace(unsigned char c)
{
	return (c == 0x20 || c == 0x09 || c == 0x0d || c == 0x0a);
}

/* ======================================================================= *
 * Linux driver interface                                                  *
 * ======================================================================= */

static struct file_operations tx4927_wdt_counter_ops = {
	owner:		THIS_MODULE,
	read:		tx4927_wdt_counter_read,
	write:		tx4927_wdt_counter_write,
};

static struct file_operations tx4927_wdt_control_ops = {
	owner:		THIS_MODULE,
	write:		tx4927_wdt_control_write,
};

static struct file_operations tx4927_wdt_reset_cause_ops = {
	owner:		THIS_MODULE,
	read:		tx4927_wdt_reset_cause_read,
};

/* ------------------------ *
 * initialize/finalize      *
 * ------------------------ */

static int
tx4927_wdt_init(void)
{
	int res;
	unsigned long flags;
	u64 ccfg_value;
	struct proc_dir_entry *ent;
	DPRFS;

	/* check reset cause */
	if (tx4927_wdrst_flag) {
		tx4927_wdt_reset_cause_is_wdt = 1;
	} else {
		local_irq_save(flags);
		ccfg_value = tx4927_reg_read64(TX4927R_CONFIG_CCFG);
		tx4927_wdt_reset_cause_is_wdt =
			((ccfg_value & TX4927M_CONFIG_WDRST) != 0);
		/* reset WDRST flag by writting 1 */
		tx4927_reg_write64(TX4927R_CONFIG_CCFG, ccfg_value);
		local_irq_restore(flags);
	}

	/* make proc directory */
	tx4927_wdt_proc_root = proc_mkdir(TX4927_WDT_DIRNAME,
					  proc_root_driver);
	if (tx4927_wdt_proc_root == NULL) {
		DPRFL(DPL_WARNING, "cannot create proc directory\n");
		res = -ENOMEM;
		goto err1;
	}

	/* make counter file */
	ent = create_proc_entry(TX4927_WDT_COUNTER_FILENAME,
				S_IRUGO | S_IWUSR, tx4927_wdt_proc_root);
	if (ent == NULL) {
		DPRFL(DPL_WARNING, "cannot create counter file.\n");
		res = -ENOMEM;
		goto err2;
	}
	ent->proc_fops = &tx4927_wdt_counter_ops;

	/* make control file */
	ent = create_proc_entry(TX4927_WDT_CONTROL_FILENAME,
				S_IWUSR, tx4927_wdt_proc_root);
	if (ent == NULL) {
		DPRFL(DPL_WARNING, "cannot create control file.\n");
		res = -ENOMEM;
		goto err3;
	}
	ent->proc_fops = &tx4927_wdt_control_ops;

	/* make reset_cause file */
	ent = create_proc_entry(TX4927_WDT_RESET_CAUSE_FILENAME,
				S_IRUGO, tx4927_wdt_proc_root);
	if (ent == NULL) {
		DPRFL(DPL_WARNING, "cannot create reset_cause file.\n");
		res = -ENOMEM;
		goto err4;
	}
	ent->proc_fops = &tx4927_wdt_reset_cause_ops;

	DPRFE;
	return 0;

err4:
	remove_proc_entry(TX4927_WDT_CONTROL_FILENAME, tx4927_wdt_proc_root);
err3:
	remove_proc_entry(TX4927_WDT_COUNTER_FILENAME, tx4927_wdt_proc_root);
err2:
	remove_proc_entry(TX4927_WDT_DIRNAME, proc_root_driver);
err1:
	DPRFR;
	return res;
}

static void
tx4927_wdt_exit(void)
{
	DPRFS;

	remove_proc_entry(TX4927_WDT_RESET_CAUSE_FILENAME,
			  tx4927_wdt_proc_root);
	remove_proc_entry(TX4927_WDT_CONTROL_FILENAME, tx4927_wdt_proc_root);
	remove_proc_entry(TX4927_WDT_COUNTER_FILENAME, tx4927_wdt_proc_root);
	remove_proc_entry(TX4927_WDT_DIRNAME, proc_root_driver);

	DPRFE;
	return;
}

module_init(tx4927_wdt_init);
module_exit(tx4927_wdt_exit);

/* ------------------------ *
 * counter                  *
 * ------------------------ */

static ssize_t
tx4927_wdt_counter_read(struct file *filp, char *buf,
			size_t count, loff_t *f_pos)
{
	int res;
	u32 tmwtmr2_value;
	u32 tmtcr2_value;
	unsigned long rest_sec;
	char str_to_write[16];
	int str_len;
	DPRFS;

	if (count == 0 || *f_pos != 0) {
		DPRFR;
		return 0;
	}

	rest_sec = 0;

	tmtcr2_value = tx4927_reg_read32(TX4927R_TIMER_TMTCR(2));
	tmwtmr2_value = tx4927_reg_read32(TX4927R_TIMER_TMWTMR(2));
	if ((tmtcr2_value & TX4927M_TIMER_CCDE) &&
	    ((tmtcr2_value & TX4927M_TIMER_TMODE) ==
	     TX4927C_TIMER_TMODE_WATCHDOG) &&
	    !(tmwtmr2_value & TX4927M_TIMER_WDIS) &&
	    (tmwtmr2_value & TX4927M_TIMER_TWIE)) {
		/* watchdog timer is enabled */
		u32 counter_rest;
		int ccd_value;
		u32 counter_freq;

		counter_rest = (tx4927_reg_read32(TX4927R_TIMER_TMCPRA(2)) -
				tx4927_reg_read32(TX4927R_TIMER_TMTRR(2)));
		ccd_value = (tx4927_reg_read32(TX4927R_TIMER_TMCCDR(2)) &
			     TX4927M_TIMER_CCD);
		counter_freq =
			tx4927_gbusclk_freq / 2 / (1 << (ccd_value + 1));
			/* NOTE: may cause mathematical error */
		rest_sec = counter_rest / counter_freq;
	}

	str_len = snprintf(str_to_write, sizeof(str_to_write),
			   "%ld\n", rest_sec);

	if (count < str_len) {
		DPRFR;
		return -EINVAL;
	}
	res = copy_to_user(buf, str_to_write, str_len);
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}
	*f_pos += str_len;

	DPRFE;
	return str_len;
}

static ssize_t
tx4927_wdt_counter_write(struct file *filp, const char *buf,
			 size_t count, loff_t *f_pos)
{
	int res;
	char tmp_buf[16];
	ssize_t nread;
	int ncopy;
	unsigned long counter_sec;
	unsigned long long tmp64;
	char *endp;
	DPRFS;

	if (count == 0) {
		DPRFR;
		return 0;
	}

	/* parse number */
	nread = 0;
	res = tx4927_wdt_skip_whitespaces_user(buf, count);
	if (res < 0) {
		DPRFR;
		return res;
	}
	nread += res;
	buf += res;
	count -= res;

	ncopy = sizeof(tmp_buf) - 1;
	if (count < ncopy) {
		ncopy = count;
	}
	res = copy_from_user(tmp_buf, buf, ncopy);
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}
	tmp_buf[ncopy] = 0;
	res = tx4927_wdt_find_next_whitespace(tmp_buf);
	if (res >= 0) {
		tmp_buf[res] = 0;
	}

	tmp64 = simple_strtoull(tmp_buf, &endp, 0);
	if (*endp != 0) {
		DPRFR;
		return -EFAULT;
	}
	if (tmp64 > ULONG_MAX) {
		DPRFR;
		return -EFAULT;
	}
	counter_sec = tmp64;
	if (counter_sec >
	    (unsigned long long)ULONG_MAX * tx4927_gbusclk_freq / 2 / 256) {
		DPRFR;
		return -EINVAL;
	}

	res = tx4927_wdt_skip_to_lineend_user(buf, count);
	if (res < 0) {
		DPRFR;
		return res;
	}
	nread += res;
	buf += res;
	count -= res;

	DPRFL(DPL_DEBUG, "counter_sec=%ld\n", counter_sec);

	/* set watchdog registers */
	if (counter_sec == 0) {
		/* disable watchdog timer */
		unsigned long flags;

		DPRFL(DPL_DEBUG, "disable WDT\n");

		local_irq_save(flags);
		tx4927_reg_setbits32(TX4927R_TIMER_TMWTMR(2),
				     TX4927M_TIMER_WDIS);
		tx4927_reg_clearbits32(TX4927R_TIMER_TMTCR(2),
				       TX4927M_TIMER_TCE | TX4927M_TIMER_CCDE);
		tx4927_reg_setbits32(TX4927R_TIMER_TMWTMR(2),
				     TX4927M_TIMER_WDIS);
		tx4927_reg_write32(TX4927R_TIMER_TMWTMR(2),
				   TX4927M_TIMER_TWC);
		local_irq_restore(flags);
	} else {
		/* set and enable watchdog timer */
		unsigned long flags;
		u32 comp_value;

		comp_value = ((unsigned long long)counter_sec *
			      tx4927_gbusclk_freq / 2 / 256);

		local_irq_save(flags);

		/* temporarily disable timer */
		tx4927_reg_setbits32(TX4927R_TIMER_TMWTMR(2),
				     TX4927M_TIMER_WDIS);
		tx4927_reg_clearbits32(TX4927R_TIMER_TMTCR(2),
				       TX4927M_TIMER_TCE | TX4927M_TIMER_CCDE);

		/* set timer mode */
		tx4927_reg_changebits32(TX4927R_TIMER_TMTCR(2),
					TX4927M_TIMER_TMODE,
					TX4927C_TIMER_TMODE_WATCHDOG);

		/* set divider */
		tx4927_reg_write32(TX4927R_TIMER_TMCCDR(2),
				   TX4927S_TIMER_CCD(7));

		/* set compare register */
		tx4927_reg_write32(TX4927R_TIMER_TMCPRA(2), comp_value);

		/* clear interrupt status */
		tx4927_reg_write32(TX4927R_TIMER_TMTISR(2), 0);

		/* enable interrupt and clear counter */
		tx4927_reg_write32(TX4927R_TIMER_TMWTMR(2),
				   TX4927M_TIMER_TWIE | TX4927M_TIMER_TWC);

		/* start timer */
		tx4927_reg_write32(TX4927R_TIMER_TMTCR(2),
				   TX4927M_TIMER_TCE | TX4927M_TIMER_CCDE |
				   TX4927C_TIMER_TMODE_WATCHDOG);

		local_irq_restore(flags);

		DPRFL(DPL_DEBUG, "enable WDT\n");
	}
	*f_pos += nread;

	DPRFE;
	return nread;
}

/* ------------------------ *
 * control                  *
 * ------------------------ */

#define TX4927_WDT_COMMAND_SET_RESET	1
#define TX4927_WDT_COMMAND_SET_NMI	2
#define TX4927_WDT_COMMAND_INVALID	-1

static ssize_t
tx4927_wdt_control_write(struct file *filp, const char *buf,
			 size_t count, loff_t *f_pos)
{
	int res;
	char tmp_buf[7];
	ssize_t nread;
	int ncopy;
	int command;
	DPRFS;

	if (count == 0) {
		DPRFR;
		return 0;
	}

	/* parse command */
	command = TX4927_WDT_COMMAND_INVALID;
	nread = 0;
	res = tx4927_wdt_skip_whitespaces_user(buf, count);
	if (res < 0) {
		DPRFR;
		return res;
	}
	nread += res;
	buf += res;
	count -= res;

	ncopy = sizeof(tmp_buf) - 1;
	if (count < ncopy) {
		ncopy = count;
	}
	res = copy_from_user(tmp_buf, buf, ncopy);
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}
	tmp_buf[ncopy] = 0;
	res = tx4927_wdt_find_next_whitespace(tmp_buf);
	if (res >= 0) {
		tmp_buf[res] = 0;
	}

	if (strcmp(tmp_buf, "reset") == 0) {
		command = TX4927_WDT_COMMAND_SET_RESET;
	} else if (strcmp(tmp_buf, "NMI") == 0) {
		command = TX4927_WDT_COMMAND_SET_NMI;
	}

	res = tx4927_wdt_skip_to_lineend_user(buf, count);
	if (res < 0) {
		DPRFR;
		return res;
	}
	nread += res;
	buf += res;
	count -= res;

	/* do command */
	switch (command) {
	case TX4927_WDT_COMMAND_SET_RESET:
		DPRFL(DPL_DEBUG, "Watchdog timer will cause reset.\n");
		tx4927_wdt_set_wdt_mode(1);
		break;
	case TX4927_WDT_COMMAND_SET_NMI:
		DPRFL(DPL_DEBUG, "Watchdog timer will cause NMI.\n");
		tx4927_wdt_set_wdt_mode(0);
		break;
	default:
		DPRFR;
		return -EINVAL;
	}
	*f_pos += nread;

	DPRFE;
	return nread;
}

/* ------------------------ *
 * reset_cause              *
 * ------------------------ */

static ssize_t
tx4927_wdt_reset_cause_read(struct file *filp, char *buf,
			    size_t count, loff_t *f_pos)
{
	int res;
	const char *str_to_write;
	ssize_t nwrite;
	DPRFS;

	if (count == 0 || *f_pos != 0) {
		DPRFR;
		return 0;
	}

	if (tx4927_wdt_reset_cause_is_wdt) {
		str_to_write = "1\n";
	} else {
		str_to_write = "0\n";
	}
	if (count == 1) {
		nwrite = 1;
	} else {
		nwrite = 2;
	}

	res = copy_to_user(buf, str_to_write, nwrite);
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}
	*f_pos += nwrite;

	DPRFE;
	return nwrite;
}

/* ======================================================================= *
 * Internal function                                                       *
 * ======================================================================= */

static int
tx4927_wdt_skip_whitespaces_user(const char *user_buf_pos, size_t count)
{
	int res;
	unsigned char c;
	const char *p;
	DPRFS;

	p = user_buf_pos;
	while (count > 0) {
		res = get_user(c, p);
		if (res != 0) {
			DPRFR;
			return -EFAULT;
		}
		if (!tx4927_wdt_is_whitespace(c)) {
			break;
		}
		p++;
		count--;
	}

	DPRFE;
	return p - user_buf_pos;
}

static int
tx4927_wdt_skip_to_lineend_user(const char *user_buf_pos, size_t count)
{
	int res;
	unsigned char c;
	const char *p;
	DPRFS;

	p = user_buf_pos;
	while (count > 0) {
		res = get_user(c, p);
		if (res != 0) {
			DPRFR;
			return -EFAULT;
		}
		p++;
		count--;
		if (c == 0x0d) {
			/* LF */
			break;
		}
		if (c == 0x0a) {
			/* CR */
			if (count > 0) {
				res = get_user(c, p);
				if (res != 0) {
					DPRFR;
					return -EFAULT;
				}
				if (c == 0x0d) {
					/* LF */
					p++;
					count--;
				}
			}
			break;
		}
	}

	DPRFE;
	return p - user_buf_pos;
}

static int
tx4927_wdt_find_next_whitespace(const char *str)
{
	int res;
	int pos;
	DPRFS;

	res = -ENOENT;
	pos = 0;
	while (*str != 0) {
		if (tx4927_wdt_is_whitespace(*str)) {
			res = pos;
			break;
		}
		str++;
		pos++;
	}

	DPRFE;
	return res;
}

static void
tx4927_wdt_set_wdt_mode(int do_reset)
{
	DPRFS;

	if (do_reset) {
		tx4927_reg_setbits64_atomic(TX4927R_CONFIG_CCFG,
					    TX4927M_CONFIG_WR);
	} else {
		tx4927_reg_clearbits64_atomic(TX4927R_CONFIG_CCFG,
					      TX4927M_CONFIG_WR);
	}

	DPRFE;
	return;
}
