/************************************/
/* GPIO Register read/write driver  */
/************************************/
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <linux/config.h>
#include <linux/delay.h>
#include <linux/spinlock.h>
#include <linux/init.h>
#include <asm/uaccess.h>

/* Driver DEBUG flag */
#define GPIO_DEBUG		0

/* Driver MajorNo Define		*/
#define GPIO_CHAR_MAJOR		231

/* FlagBit using in GPIO Accese address table */
#define GPIO_SIZE_BYTE		1		/* RegisterAccessSize is ByteAccsess	*/
#define GPIO_SIZE_WORD		2		/* RegisterAccessSize is WordAccsess	*/
#define GPIO_SIZE_LONG		4		/* RegisterAccessSize is LongAccsess	*/
#define GPIO_READ_OK		(1)		/* RegisterAcceseRead/WriteFlag READ OK	*/
#define GPIO_WRITE_OK		(1<<1)		/* RegisterAcceseRead/WriteFlag WRITE OK*/
#define GPIO_R_W_OK		(1 | (1<<1))	/* RegisterAcceseRead/WriteFlag WRITE OK*/

/* Register Access MACRO */
#define RD08( reg      )   (*(char*)(reg))
#define WR08( reg, val )  ((*(char*)(reg))=(val))
#define RD16( reg      )   (*(short*)(reg))
#define WR16( reg, val )  ((*(short*)(reg))=(val))
#define RD32( reg      )   (*(long*)(reg))
#define WR32( reg, val )  ((*(long*)(reg))=(val))

/* GPIO Access address table */
struct gpio_access_addr {
	long	addr;			/* Access address	*/
	char	size;			/* register access size ( 1=byte, 2=word, 3=long)	*/
	char	r_w;			/* Read/Write Check ( bit0 = Read , bit1 = write )	*/
} gpio_access_addr[] = {
	{0xb4000009, GPIO_SIZE_BYTE, GPIO_READ_OK },			/* MPU-220 FPGA GPIO PI0*/
	{0xb400000a, GPIO_SIZE_BYTE, GPIO_READ_OK },			/* MPU-220 FPGA GPIO PI1*/
	{0xb400000b, GPIO_SIZE_BYTE, GPIO_READ_OK },			/* MPU-220 FPGA GPIO PI2*/
	{0xb400000c, GPIO_SIZE_BYTE, GPIO_R_W_OK },			/* MPU-220 FPGA GPIO PO0*/
	{0xb400000d, GPIO_SIZE_BYTE, GPIO_R_W_OK },			/* MPU-220 FPGA GPIO PO1*/
	{0xb400000e, GPIO_SIZE_BYTE, GPIO_R_W_OK },			/* MPU-220 FPGA GPIO PO2*/
	{0xb400000f, GPIO_SIZE_BYTE, GPIO_R_W_OK },			/* MPU-220 FPGA GPIO PO3*/
 {0xffffffff, 0, 0x00 }						/* stopper		*/
};

/* prototypes */
static int tx_gpio_open(struct inode * inode, struct file * file);
static int tx_gpio_close(struct inode * inode, struct file * file);
static ssize_t tx_gpio_read(struct file * file, char * buf, size_t count, loff_t *ppos);
static ssize_t tx_gpio_write(struct file * file, const char * buf, size_t count, loff_t *ppos);

/********************************************************/
/*	GPIO Open					*/
/********************************************************/
static int tx_gpio_open(struct inode * inode, struct file * file)
{
	ushort val = 0;
	int minor  = MINOR(inode->i_rdev);

#if GPIO_DEBUG
	printk("GPIO: open() is called \n");
#endif

	file->private_data = (void *)minor;

	MOD_INC_USE_COUNT;

	return val;
}

/********************************************************/
/*	GPIO Close					*/
/********************************************************/
static int tx_gpio_close(struct inode * inode, struct file * file)
{
	ushort val = 0;

#if GPIO_DEBUG
	printk("GPIO: close() is called \n");
#endif

	MOD_DEC_USE_COUNT;
	return val;
}

/********************************************************/
/*	GPIO Read Module				*/
/********************************************************/
static ssize_t tx_gpio_read(struct file * file, char * buf, size_t count,
			loff_t *ppos)
{
	long	*addr;
	long	get_gpio_l;
	short	get_gpio_w;
	char	get_gpio_b;
	int	minor   = (int)file->private_data;

	if ( (gpio_access_addr[minor].r_w & GPIO_READ_OK) ==  0x00  ){		/* Check this register can reading */
#if GPIO_DEBUG
		printk("GPIO: This Register(%d) is not authorized read \b");
#endif
		return -1;
	}
	if ( count < gpio_access_addr[minor].size ){	/* Check read(=buffer) size */
		return 0;
	}

	/* Read Data */
	addr = (long *)gpio_access_addr[minor].addr;

	switch ( gpio_access_addr[minor].size){
		case GPIO_SIZE_BYTE:
			get_gpio_b = RD08(addr);
			copy_to_user( buf, &get_gpio_b, sizeof(get_gpio_b) );
			break;
		case GPIO_SIZE_WORD:
			get_gpio_w = RD16(addr);
			copy_to_user( buf, &get_gpio_w, sizeof(get_gpio_w) );
			break;
		case GPIO_SIZE_LONG:
			get_gpio_l = RD32(addr);
			copy_to_user( buf, &get_gpio_l, sizeof(get_gpio_l) );
			break;
	}

	return gpio_access_addr[minor].size;

}

/********************************************************/
/*	GPIO Write Module				*/
/********************************************************/
static ssize_t tx_gpio_write(struct file * file, const char * buf, size_t count,
			 loff_t *ppos)
{
	long	*addr;
	long	set_gpio_l;
	short	set_gpio_w;
	char	set_gpio_b;
	int	minor	   = (int)file->private_data;

	if ( (gpio_access_addr[minor].r_w & GPIO_WRITE_OK) ==  0x00  ){		/* Check this register can reading */
#if GPIO_DEBUG
		printk("GPIO: This Register(%d) is not authorized write \b");
#endif
		return -1;
	}

	if ( count != gpio_access_addr[minor].size ){	/* Check read(=buffer) size */
		return 0;
	}

	addr = (long *)gpio_access_addr[minor].addr;

	/* write Data */
	switch ( gpio_access_addr[minor].size){
		case GPIO_SIZE_BYTE:
			copy_from_user( &set_gpio_b, buf, sizeof(set_gpio_b) );
			WR08(addr, set_gpio_b);
			break;
		case GPIO_SIZE_WORD:
			copy_from_user( &set_gpio_w, buf, sizeof(set_gpio_w) );
			WR08(addr, set_gpio_w);
			break;
		case GPIO_SIZE_LONG:
			copy_from_user( &set_gpio_l, buf, sizeof(set_gpio_l) );
			WR08(addr, set_gpio_l);
			break;
	}

	/* set buf and return */
	return gpio_access_addr[minor].size;

}

/* GPIO Control Drivers	*/
struct file_operations gpio_fopes = {
	open:		tx_gpio_open,
	write:		tx_gpio_write,
	read:		tx_gpio_read,
	release:	tx_gpio_close,
};

/********************************************************/
/*	GPIO Init Module				*/
/********************************************************/
static int __init tx_gpio_init_module(void)
{
	if (register_chrdev(GPIO_CHAR_MAJOR, "gpio", &gpio_fopes))
	{
		printk(KERN_NOTICE "Can't allocate major number %d for GPIO Driver.\n",
		       GPIO_CHAR_MAJOR);
		return -EAGAIN;
	}
	printk("GPIO Register Order read/write Driver\n");
	return 0;
}


/********************************************************/
/*	GPIO Cleanup Module				*/
/********************************************************/
void __exit tx_gpio_cleanup_module(void)
{
	unregister_chrdev(GPIO_CHAR_MAJOR, "gpio");
}

module_init(tx_gpio_init_module);
module_exit(tx_gpio_cleanup_module);

