/*
   -------------------------------------------------------------------------
   i2c-algo-goku_s.c i2c driver algorithms for GOKU-S adapters

   Copyright 2002 TOSHIBA CORPRATION SEMICONDUCTOR COMPANY.

   ---------------------------------------------------------------------------
   This file was highly leveraged from i2c-algo-pcf.c, which was created
   by Simon G. Vogl and Hans Berglund:


     Copyright (C) 1995-1997 Simon G. Vogl
                   1998-2000 Hans Berglund

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.		     */
/* ------------------------------------------------------------------------- */


#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <linux/ioport.h>
#include <linux/errno.h>
#include <linux/sched.h>

#include <linux/i2c.h>
#include <linux/i2c-algo-goku-s.h>
#include "i2c-goku-s.h"


/* ----- global defines ----------------------------------------------- */
#define DEB(x) if (i2c_debug>=1) x
#define DEB2(x) if (i2c_debug>=2) x
#define DEB3(x) if (i2c_debug>=3) x /* print several statistical values*/
#define DEBPROTO(x) if (i2c_debug>=9) x;
 	/* debug the protocol by showing transferred bits */

/*#define	GOKU_I2C_DEBUG*/
#if defined(GOKU_I2C_DEBUG)
#define DEB_START	printk(__FUNCTION__": start\n");
#define DEB_END		printk(__FUNCTION__": end\n");
#define DEB_SNAP(x...)	printk(__FUNCTION__": "##x);
#else
#define DEB_START
#define DEB_END
#define DEB_SNAP(x...)
#endif

#define DEF_TIMEOUT 16

/* debugging - slow down transfer to have a look at the data .. 	*/
/* I use this with two leds&resistors, each one connected to sda,scl 	*/
/* respectively. This makes sure that the algorithm works. Some chips   */
/* might not like this, as they have an internal timeout of some mils	*/


/* ----- global variables ---------------------------------------------	*/


/* module parameters:
 */
static int i2c_debug=0;
static int i2c_test=0;	/* see if the line-setting functions work	*/
/* static int i2c_scan=0;	*//* have a look at what's hanging 'round		*/

/* --- setting states on the bus with the right timing: ---------------	*/

#define GOKU_I2C_DATA(adap)			((struct goku_s_i2c *)(adap->data))

#define get_clock(adap)				adap->getclock(adap->data)
#define i2c_outb(adap, reg, val)	adap->setiic(adap->data, reg, val)
#define i2c_inb(adap, reg)			adap->getiic(adap->data, reg)
#define i2c_getclk(adap)			adap->getclock(adap->data)
#define i2c_getown(adap)			adap->getown(adap->data)
#define gpio_outb(adap, reg, val)	adap->setgpio(adap->data, reg, val)

#define i2c_status(adap)			(GOKU_I2C_DATA(adap)->i2c_sts)

#define i2c_start(adap)				adap->start()
#define i2c_stop(adap)				adap->stop()
#define i2c_reset(adap)				adap->reset()

#define i2c_txlock(adap)			spin_lock(&(GOKU_I2C_DATA(adap)->tx_lock))
#define i2c_txunlock(adap)			spin_unlock(&(GOKU_I2C_DATA(adap)->tx_lock))

#define i2c_txset(adap, buf, len)	i2c_txlock(adap);					\
									GOKU_I2C_DATA(adap)->tx_buf = buf; \
									GOKU_I2C_DATA(adap)->tx_size = len;	\
									GOKU_I2C_DATA(adap)->tx_count = 0;	\
									i2c_txunlock(adap);

#define i2c_txreset(adap)			i2c_txlock(adap);					\
									GOKU_I2C_DATA(adap)->tx_buf = NULL; \
									GOKU_I2C_DATA(adap)->tx_size = 0;	\
									i2c_txunlock(adap);
#define i2c_txget(adap, len)		len = GOKU_I2C_DATA(adap)->tx_count;

/*#define i2c_txwait(adap, tmout)		interruptible_sleep_on_timeout(&(GOKU_I2C_DATA(adap)->tx_wait), tmout)*/
#define i2c_txwait(adap, tmout)		wait_for_completion(&(GOKU_I2C_DATA(adap)->tx_comp))

#define i2c_rxlock(adap)			spin_lock(&(GOKU_I2C_DATA(adap)->rx_lock))
#define i2c_rxunlock(adap)			spin_unlock(&(GOKU_I2C_DATA(adap)->rx_lock))
#define i2c_rxset(adap, buf, len)	i2c_rxlock(adap);					\
									GOKU_I2C_DATA(adap)->rx_buf = buf; \
									GOKU_I2C_DATA(adap)->rx_size = len;	\
									GOKU_I2C_DATA(adap)->rx_count = 0;	\
									i2c_rxunlock(adap);

#define i2c_rxreset(adap)			GOKU_I2C_DATA(adap)->rx_buf = NULL; \
									GOKU_I2C_DATA(adap)->rx_size = 0;
#define i2c_rxget(adap, len)		len = GOKU_I2C_DATA(adap)->rx_count;
/*#define i2c_rxwait(adap, tmout)		interruptible_sleep_on_timeout(&(GOKU_I2C_DATA(adap)->rx_wait), tmout)*/
#define i2c_rxwait(adap, tmout)		wait_for_completion(&(GOKU_I2C_DATA(adap)->rx_comp))

#define i2c_slavelock(adap)			spin_lock(&(GOKU_I2C_DATA(adap)->slave_lock))
#define i2c_slaveunlock(adap)		spin_unlock(&(GOKU_I2C_DATA(adap)->slave_lock))
#define i2c_slavebuf(adap)			(GOKU_I2C_DATA(adap)->slave_buf)


/* --- other auxiliary functions --------------------------------------	*/

/*
 * Wait for Bus busy : return zero when successfully
 */
static int wait_for_bb(struct goku_s_i2c_algo_data *adap)
{
	int timeout = DEF_TIMEOUT;
	__u8 status;

DEB_START;
	status = i2c_inb(adap, GOKU_S_I2C_SBICR2);
#ifndef STUB_I2C
	while (timeout-- && (status & GOKU_S_I2C_BB)) {
		schedule_timeout(HZ / 1000);
/*		udelay(1000); *//* How much is this? */
		status = i2c_inb(adap, GOKU_S_I2C_SBICR2);
	}
#endif
	if (timeout <= 0) {
		printk(KERN_ERR "Timeout, host is busy\n");
		i2c_reset(adap);
	}
	return(timeout <= 0);
}

/*
 * Puts this process to sleep for a period equal to timeout
 */
static inline void goku_s_i2c_sleep(unsigned long timeout)
{
DEB_START;
	schedule_timeout( timeout * HZ);
}

/* After we issue a transaction on the IIC bus, this function
 * is called.  It puts this process to sleep until we get an interrupt from
 * from the controller telling us that the transaction we requested in complete.
 */
/*
 * wait for ack of transaction
 */
static int wait_for_ack(struct goku_s_i2c_algo_data *adap, __u8 trx) {

	int timeout = DEF_TIMEOUT;
	__u8 status;

DEB_START;
#if 0
	timeout = wait_for_bb(adap);
	if (timeout) {
  		DEB2(printk("Timeout waiting for host not busy\n");)
  		return -EIO;
	}
	timeout = DEF_TIMEOUT;
#endif

#if 0
	status = i2c_status(adap);
#else
	status = ~(trx << 6);
#endif
#ifndef STUB_I2C
	while (timeout-- && ((status & GOKU_S_I2C_TRX) != (trx << 6))) {
	   adap->waitforpin(2);
	   status = i2c_status(adap);
	}
#endif
	if (timeout <= 0) {
		return(-1);
	} else {
		return(0);
	}
}

#if 0	/* un used!! */
/*
 * wait for completion of transaction
 */
static int wait_for_pin(struct goku_s_i2c_algo_data *adap, __u8 *status) {

	int timeout = DEF_TIMEOUT;

DEB_START;
#if 0
	timeout = wait_for_bb(adap);
	if (timeout) {
  		DEB2(printk("Timeout waiting for host not busy\n");)
  		return -EIO;
	}                           
	timeout = DEF_TIMEOUT;
#endif

	*status = i2c_inb(adap, GOKU_S_I2C_SBICR2);
#ifndef STUB_I2C
	while (timeout-- && !(*status & GOKU_S_I2C_TRX)) {
		adap->waitforpin(2);
		*status = i2c_inb(adap, GOKU_S_I2C_SBICR2);
	}
#endif
	if (timeout <= 0) {
		return(-1);
	} else {
		return(0);
	}
}
#endif

static int goku_s_i2c_init (struct goku_s_i2c_algo_data *adap)
{
	__u8 val;

DEB_START;
	/* Set ACK and SCL to zero in SBICR1 */
#if 0
	val = i2c_inb(adap, GOKU_S_I2C_SBICR1);
	val &= ~(GOKU_S_I2C_ACK | GOKU_S_I2C_SCK);
	val |= i2c_getclk(adap);
#else
	val = (GOKU_S_I2C_ACK | i2c_getclk(adap));
#endif
	i2c_outb(adap, GOKU_S_I2C_SBICR1, val);

	/* Clear Slabe address in I2CAR */
#if 1
	val = (i2c_getown(adap) << 1) & 0xfe;
#else
	val = (i2c_getown(adap) << 1) | GOKU_S_I2C_ALS;
#endif
	i2c_outb(adap, GOKU_S_I2C_I2CAR, val);
	
	/* Set to slave receiver mode */
	i2c_outb(adap, GOKU_S_I2C_SBICR2, GOKU_S_I2C_PIN);

#if 0
	DEB2(printk("goku_s_i2c_init: Initialized I2C on GOKU-S 0x%x\n",
		i2c_inw(adap, ITE_I2CHSR)));
#endif

	return 0;
}


/*
 * Sanity check for the adapter hardware - check the reaction of
 * the bus lines only if it seems to be idle.
 */
static int test_bus(struct goku_s_i2c_algo_data *adap, char *name)
{
DEB_START;
	return (0);
}

/* ----- Utility functions
 */


/* Verify the device we want to talk to on the IIC bus really exists. */
static inline int try_address(struct goku_s_i2c_algo_data *adap,
		       unsigned int addr, int retries)
{
	int i, ret = -1;

DEB_START;
	for (i = 0; i < retries; i++) {
		/* wait for bus free */
		if (wait_for_bb(adap)) {
			/* error */
			DEB_SNAP("wait_for_bb is error\n");
			break;
		}

		i2c_txreset(adap);

		/* set to ack mode */
		i2c_outb(adap, GOKU_S_I2C_SBICR1,
			(i2c_getclk(adap) | GOKU_S_I2C_ACK));
		/* set slave address */
		i2c_outb(adap, GOKU_S_I2C_SBIDBR, addr);
		i2c_start(adap);
		if (wait_for_ack(adap, 0) == 0) {
			i2c_stop(adap);
			DEB_SNAP("found device 0x%x\n", (addr >> 1) & 0x7f);
			ret = 1;
			break;	/* success! */
		}
		i2c_stop(adap);
		udelay(adap->udelay);
		DEB_SNAP(".");
	}

	DEB2(if (i) printk(": needed %d retries for 0x%x\n",
						i, addr));

	return ret;
}


#if 0
static int i2c_sendbytes(struct i2c_adapter *i2c_adap,const char *buf,
                         int count)
{
	struct goku_s_i2c_algo_data *adap = i2c_adap->algo_data;
	int wrcount=0, timeout, i;
	__u8 val;

DEB_START;
	for (i = 0; i < count; i++) {
		/* check MST */
		val = i2c_inb(adap, GOKU_S_I2C_SBICR2);
		if (!(val & GOKU_S_I2C_MST)){
			DEB_SNAP("Device is not Master\n");
			return (-1);
		}
		if (!(val & GOKU_S_I2C_TRX)) {
			DEB_SNAP("Device is not Transmitter\n");
			return (-1);
		}
		if (val & GOKU_S_I2C_LRB) {
			i2c_stop(adap);
			printk(": %s write error - no ack.\n", i2c_adap->name);
			return -EREMOTEIO; /* got a better one ?? */
		}
 
		/* write 1 word data */
		i2c_outb(adap, GOKU_S_I2C_SBIDBR, buf[wrcount++]);

		/* wait for next ack */
		timeout = wait_for_ack(adap, 0);
		if(timeout) {
			i2c_stop(adap);
			printk(": %s write timeout.\n", i2c_adap->name);
			return -EREMOTEIO;
		}
	}
	i2c_stop(adap);
	return wrcount;
}
#else
static int i2c_sendbytes(struct i2c_adapter *i2c_adap, struct i2c_msg *pmsg)
{
	struct goku_s_i2c_algo_data *adap = i2c_adap->algo_data;
	__u8 addr = ((pmsg->addr & 0x7F) << 1);
	int wrcount = 0;
   
DEB_START;
	i2c_txset(adap, pmsg->buf, pmsg->len);

	/* set slave address */
	i2c_outb(adap, GOKU_S_I2C_SBICR1, (GOKU_S_I2C_ACK | get_clock(adap)));
	i2c_outb(adap, GOKU_S_I2C_SBIDBR, addr);

	/* start condition */
	if (wait_for_bb(adap)) {
  		DEB2(printk("Timeout waiting for host not busy\n");)
  		return wrcount;
	}                           
	i2c_start(adap);

	/* wait for completion */
/*	adap->waitforpin(pmsg->len);*/
	if (pmsg->len) {
		wrcount = pmsg->len * HZ;
	} else {
		wrcount = HZ;
	}
#if 1
	if (wait_for_ack(adap, 1)) {
		i2c_stop(adap);
		DEB_SNAP("Can't found device 0x%x\n", (addr >> 1) & 0x7f);
  		return wrcount;
	}
#endif
	DEB_SNAP("start sleep: %d\n", wrcount);
#if 0
	if(i2c_txwait(adap, (pmsg->len) * HZ) == 0) {
			DEB_SNAP("timeout\n");
	}
#else
	i2c_txwait(adap, wrcount);
#endif
DEB_END;
	i2c_txget(adap, wrcount);
	
	return wrcount;
}
#endif

#if 0
static int i2c_readbytes(struct i2c_adapter *i2c_adap, char *buf, int count)
{
	struct goku_s_i2c_algo_data *adap = i2c_adap->algo_data;
	int rdcount=0, i, timeout;
	__u8 val;

DEB_START;
	if (count == 1) {
		/* set no ACK mode */
		i2c_outb(adap, GOKU_S_I2C_SBICR1,
			 (i2c_inb(adap, GOKU_S_I2C_SBICR1) & ~(GOKU_S_I2C_ACK)));
	}
	for (i = 0; i < (count - 1); i++) {
		/* check MST */
		val = i2c_inb(adap, GOKU_S_I2C_SBICR2);
		if (!(val & GOKU_S_I2C_MST)){
			DEB_SNAP("Device is not Master\n");
			return (-1);
		}
		if (val & GOKU_S_I2C_TRX) {
			DEB_SNAP("Device is not receiver\n");
			return (-1);
		}
 
		/* read 1 word data */
		buf[rdcount++] = i2c_inb(adap, GOKU_S_I2C_SBIDBR);

		/* wait for next ack */
		timeout = wait_for_ack(adap, 0);
		if(timeout) {
			i2c_stop(adap);
			printk(":  %s read timeout.\n", i2c_adap->name);
			return (-1);
		}
	}

	/* check MST */
	val = i2c_inb(adap, GOKU_S_I2C_SBICR2);
	if (!(val & GOKU_S_I2C_MST)){
		DEB_SNAP("Device is not Master\n");
		return (-1);
	}
	if (val & GOKU_S_I2C_TRX) {
		DEB_SNAP("Device is not Receiver\n");
		return (-1);
	}

	/* set 1 bit trans */
	i2c_outb(adap, GOKU_S_I2C_SBICR1,
		(i2c_inb(adap, GOKU_S_I2C_SBICR1) & GOKU_S_I2C_SCK) | 0x20);

	/* write 1 word data */
	buf[rdcount++] = i2c_inb(adap, GOKU_S_I2C_SBIDBR);

	i2c_stop(adap);
	return rdcount;
}
#else

static int i2c_readbytes(struct i2c_adapter *i2c_adap, struct i2c_msg *pmsg)
{
	struct goku_s_i2c_algo_data *adap = i2c_adap->algo_data;
	__u8 addr = ((pmsg->addr & 0x7F) << 1) | 0x01;
	int rdcount = 0;
/*	__u8 off = 0x00;	*//* read offset */

DEB_START;
	i2c_rxset(adap, pmsg->buf, pmsg->len);

	/* set slave address */
	i2c_outb(adap, GOKU_S_I2C_SBICR1, (GOKU_S_I2C_ACK | get_clock(adap)));
	i2c_outb(adap, GOKU_S_I2C_SBIDBR, addr);

	/* start condition */
	if (wait_for_bb(adap)) {
  		DEB2(printk("Timeout waiting for host not busy\n");)
  		return rdcount;
	}
	i2c_start(adap);

	/* wait for completion */
/*	adap->waitforpin(pmsg->len);*/
	if (pmsg->len) {
		rdcount = pmsg->len * HZ;
	} else {
		rdcount = HZ;
	}
#if 1
	if (wait_for_ack(adap, 0)) {
		i2c_stop(adap);
		printk(": Can't found device 0x%x\n", (addr >> 1) & 0x7f);
  		return rdcount;
	}
#endif
	DEB_SNAP("start sleep: %d\n", rdcount);
#if 0
	if (i2c_rxwait(adap, (pmsg->len) * HZ) == 0) {
			DEB_SNAP("timeout\n");
	}
#else
	i2c_rxwait(adap, rdcount);
#endif
DEB_END;
	i2c_rxget(adap, rdcount);

	return rdcount;
}
#endif

/* This function implements combined transactions.  Combined
 * transactions consist of combinations of reading and writing blocks of data.
 * Each transfer (i.e. a read or a write) is separated by a repeated start
 * condition.
 */
#if 0
static int i2c_combined_transaction(struct i2c_adapter *i2c_adap, struct i2c_msg msgs[], int num)
{
   int i;
   struct i2c_msg *pmsg;
   int ret;

   DEB2(printk("Beginning combined transaction\n"));

   for(i=0; i<(num-1); i++) {
      pmsg = &msgs[i];
      if(pmsg->flags & I2C_M_RD) {
         DEB2(printk("  This one is a read\n"));
         ret = i2c_readbytes(i2c_adap, pmsg->buf, pmsg->len, IIC_COMBINED_XFER);
      }
      else if(!(pmsg->flags & I2C_M_RD)) {
         DEB2(printk("This one is a write\n"));
         ret = i2c_sendbytes(i2c_adap, pmsg->buf, pmsg->len, IIC_COMBINED_XFER);
      }
   }
   /* Last read or write segment needs to be terminated with a stop */
   pmsg = &msgs[i];

   if(pmsg->flags & I2C_M_RD) {
      DEB2(printk("Doing the last read\n"));
      ret = i2c_readbytes(i2c_adap, pmsg->buf, pmsg->len, IIC_SINGLE_XFER);
   }
   else if(!(pmsg->flags & I2C_M_RD)) {
      DEB2(printk("Doing the last write\n"));
      ret = i2c_sendbytes(i2c_adap, pmsg->buf, pmsg->len, IIC_SINGLE_XFER);
   }

   return ret;
}
#endif


/* Whenever we initiate a transaction, the first byte clocked
 * onto the bus after the start condition is the address (7 bit) of the
 * device we want to talk to.  This function manipulates the address specified
 * so that it makes sense to the hardware when written to the IIC peripheral.
 *
 * Note: 10 bit addresses are not supported in this driver, although they are
 * supported by the hardware.  This functionality needs to be implemented.
 */
static inline int i2c_doAddress(struct goku_s_i2c_algo_data *adap,
                                struct i2c_msg *msg, int retries)
{
	unsigned short flags = msg->flags;
	unsigned int addr;
	int ret;

DEB_START;
/* Ten bit addresses not supported right now */
	if ( (flags & I2C_M_TEN)  ) {
#if 0	/* not support 10bit address */
#endif
		return -1;
	} else {
		addr = (msg->addr & 0x7F) << 1;
		if (flags & I2C_M_RD ) {
			addr |= 0x01;
		}
#if 0
		if (flags & I2C_M_RD )
			addr |= 1;
		if (flags & I2C_M_REV_DIR_ADDR )
			addr ^= 1;
#endif

		ret = try_address(adap, addr, retries);
		if (ret != 1) {
			printk(": died at address code (0x%x).\n", addr);
			return -EREMOTEIO;
		}
	}

	return 0;
}


/* Description: Prepares the controller for a transaction (clearing status
 * registers, data buffers, etc), and then calls either i2c_readbytes or
 * i2c_sendbytes to do the actual transaction.
 *
 * still to be done: Before we issue a transaction, we should
 * verify that the bus is not busy or in some unknown state.
 */
static int goku_s_i2c_xfer(struct i2c_adapter *i2c_adap,
		    struct i2c_msg msgs[],
		    int num)
{
	struct goku_s_i2c_algo_data *adap = i2c_adap->algo_data;
	struct i2c_msg *pmsg;
	int i = 0;
	int ret = 0, timeout;

DEB_START;
	for (i = 0; i < num; i++) {
		pmsg = &msgs[i];

		if(!pmsg->len) {
			DEB2(printk(": read/write length is 0\n");)
			return -EIO;
		}
#if 0
		if(!(pmsg->flags & I2C_M_RD) && (!(pmsg->len)%2) ) {
			DEB2(printk(": write buffer length is not odd\n");)
			return -EIO;
		}
#endif

		/* Wait for any pending transfers to complete */
		timeout = wait_for_bb(adap);
		if (timeout) {
			DEB2(printk(": Timeout waiting for host not busy\n");)
			return -EIO;
		}

		/* Load address */
#if 0
		ret = i2c_doAddress(adap, pmsg, i2c_adap->retries);
		if (ret) {
			printk(": Device is not found\n");
			return -EIO;
		}
#endif

		DEB3(printk(": Msg %d, addr=0x%x, flags=0x%x, len=%d\n",
			i, msgs[i].addr, msgs[i].flags, msgs[i].len);)

		if(pmsg->flags & I2C_M_RD) { 		/* Read */
			ret = i2c_readbytes(i2c_adap, pmsg);
/*			ret = i2c_readbytes(i2c_adap, pmsg->buf, pmsg->len);*/
		} else {													/* Write */
			udelay(1000);
			ret = i2c_sendbytes(i2c_adap, pmsg);
/*			ret = i2c_sendbytes(i2c_adap, pmsg->buf, pmsg->len);*/
		}

		if (ret != pmsg->len) {
			DEB3(printk(": error or fail on read/write %d bytes.\n",ret));
		} else {
			DEB3(printk(": read/write %d bytes.\n",ret));
		}
	}

	return ret;
}


/* Implements device specific ioctls.  Higher level ioctls can
 * be found in i2c-core.c and are typical of any i2c controller (specifying
 * slave address, timeouts, etc).  These ioctls take advantage of any hardware
 * features built into the controller for which this algorithm-adapter set
 * was written.  These ioctls allow you to take control of the data and clock
 * lines and set the either high or low,
 * similar to a GPIO pin.
 */
static int algo_control(struct i2c_adapter *adapter, 
	unsigned int cmd, unsigned long arg)
{
#if 0
	struct goku_s_i2c_algo_data *adap = adapter->algo_data;
	struct i2c_iic_msg s_msg;
	char *buf;
	int ret;

DEB_START;
	if (cmd == I2C_SREAD) {
		if(copy_from_user(&s_msg, (struct i2c_iic_msg *)arg, 
				sizeof(struct i2c_iic_msg)))
			return -EFAULT;
		buf = kmalloc(s_msg.len, GFP_KERNEL);
		if (buf== NULL)
			return -ENOMEM;

		/* Flush FIFO */
		i2c_outw(adap, ITE_I2CFCR, ITE_I2CFCR_FLUSH);

		/* Load address */
		i2c_outw(adap, ITE_I2CSAR,s_msg.addr<<1);
		i2c_outw(adap, ITE_I2CSSAR,s_msg.waddr & 0xff);

		ret = i2c_readbytes(adapter, buf, s_msg.len, 1);
		if (ret>=0) {
			if(copy_to_user( s_msg.buf, buf, s_msg.len) ) 
				ret = -EFAULT;
		}
		kfree(buf);
	}
#else
	struct goku_s_i2c_algo_data *adap = adapter->algo_data;
	struct i2c_slavebuf_data bufdata;

DEB_START;
	switch (cmd) {
	case I2C_RESET:
		i2c_reset(adap);
		goku_s_i2c_init(adap);
		break;
	case I2C_SETSLAVEBUF:
		/* get parameters */
		if(copy_from_user(&bufdata, (struct i2c_slavebuf_data *)arg, 
				sizeof(struct i2c_slavebuf_data))) {
			return -EFAULT;
		}
		/* check paramters */
		if ((bufdata.offset >= GOKU_S_SLAVE_BUFSIZE)
			|| ((bufdata.offset + bufdata.size) > GOKU_S_SLAVE_BUFSIZE)) {
			return -EFAULT;
		}
		/* copy data */
		if(copy_from_user(&(i2c_slavebuf(adap)[bufdata.offset]), bufdata.buf, 
				bufdata.size)) {
			return -EFAULT;
		}
		break;
	case I2C_GETSLAVEBUF:
		/* get parameters */
		if(copy_from_user(&bufdata, (struct i2c_slavebuf_data *)arg, 
				sizeof(struct i2c_slavebuf_data))) {
			return -EFAULT;
		}
		/* check paramters */
		if ((bufdata.offset >= GOKU_S_SLAVE_BUFSIZE)
			|| ((bufdata.offset + bufdata.size) > GOKU_S_SLAVE_BUFSIZE)) {
			return -EFAULT;
		}
		/* copy data */
		if(copy_to_user(bufdata.buf, &(i2c_slavebuf(adap)[bufdata.offset]), 
				bufdata.size)) {
			return -EFAULT;
		}
		break;
	}
#endif
	return 0;
}


static u32 goku_s_i2c_func(struct i2c_adapter *adap)
{
DEB_START;
	return I2C_FUNC_SMBUS_EMUL |
	       I2C_FUNC_PROTOCOL_MANGLING;
}

/* -----exported algorithm data: -------------------------------------	*/

static struct i2c_algorithm goku_s_i2c_algo = {
	"GOKU-S I2C algorithm",
	I2C_ALGO_GOKU_S,
	goku_s_i2c_xfer,		/* master_xfer	*/
	NULL,				/* smbus_xfer	*/
	NULL,				/* slave_xmit		*/
	NULL,				/* slave_recv		*/
	algo_control,			/* ioctl		*/
	goku_s_i2c_func,		/* functionality	*/
};


/*
 * registering functions to load algorithms at runtime
 */
int i2c_iic_add_bus(struct i2c_adapter *adap)
{
	struct goku_s_i2c_algo_data *i2c_adap = adap->algo_data;

DEB_START;
	if (i2c_test) {
		int ret = test_bus(i2c_adap, adap->name);
		if (ret < 0) {
			printk(": test_bus is error\n");
			return -ENODEV;
		}
	}

	DEB2(printk("i2c-algo-goku-s: hw routines for %s registered.\n",
	            adap->name));

	/* register new adapter to i2c module... */

	adap->id |= goku_s_i2c_algo.id;
	adap->algo = &goku_s_i2c_algo;

	adap->timeout = 100;	/* default values, should	*/
	adap->retries = 3;		/* be replaced by defines	*/
	adap->flags = 0;

#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif

	i2c_add_adapter(adap);
	goku_s_i2c_init(i2c_adap);

#if 0
	/* scan bus */
	/* By default scanning the bus is turned off. */
	if (i2c_scan) {
		printk(KERN_INFO " i2c-algo-goku-s: scanning bus %s.\n",
		       adap->name);
		for (i = 0x80; i < 0xff; i++) { /* Slave address | Read Direction */
			/* check bus free */
			if (wait_for_bb(i2c_adap)) {
				DEB_SNAP("bus is busy\n");
				continue;
			}
			/* set to ack mode */
			i2c_outb(i2c_adap, GOKU_S_I2C_SBICR1,
				(i2c_inb(i2c_adap, GOKU_S_I2C_SBICR1) | GOKU_S_I2C_ACK));
			/* set slave address */
			i2c_outb(i2c_adap, GOKU_S_I2C_SBIDBR, i);
			/* start */
			i2c_start(i2c_adap);
			/* wait for ACK interrupt (TRX bit change to zero) */
			if (wait_for_ack(i2c_adap, 0) == 0) { 
				printk(KERN_INFO "\n(%02x)\n", i & 0x7F); 
			} else {
				printk(KERN_INFO ".");
				i2c_reset(i2c_adap);
			}
			udelay(i2c_adap->udelay);
		}
	}
#endif
	return 0;
}


int i2c_iic_del_bus(struct i2c_adapter *adap)
{
	int res;
DEB_START;
	if ((res = i2c_del_adapter(adap)) < 0) {
		return res;
	}
	DEB2(printk("i2c-algo-goku-s: adapter unregistered: %s\n",adap->name));

#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif
	return 0;
}


int __init i2c_algo_goku_s_init (void)
{
	printk(KERN_INFO "GOKU-S i2c algorithm module\n");
	return 0;
}


void i2c_algo_goku_s_exit(void)
{
	return;
}


EXPORT_SYMBOL(i2c_iic_add_bus);
EXPORT_SYMBOL(i2c_iic_del_bus);

#if defined(MODULES)
/* The MODULE_* macros resolve to nothing if MODULES is not defined
 * when this file is compiled.
 */
MODULE_AUTHOR("TOSHIBA CORPRATION SEMICONDUCTOR COMPANY");
MODULE_DESCRIPTION("GOKU-S i2c algorithm");
MODULE_LICENSE("GPL");

MODULE_PARM(i2c_test, "i");
/* MODULE_PARM(i2c_scan, "i");*/
MODULE_PARM(i2c_debug,"i");

MODULE_PARM_DESC(i2c_test, "Test if the I2C bus is available");
/* MODULE_PARM_DESC(i2c_scan, "Scan for active chips on the bus");*/
MODULE_PARM_DESC(i2c_debug,
        "debug level - 0 off; 1 normal; 2,3 more verbose; 9 iic-protocol");


/* This function resolves to init_module (the function invoked when a module
 * is loaded via insmod) when this file is compiled with MODULES defined.
 * Otherwise (i.e. if you want this driver statically linked to the kernel),
 * a pointer to this function is stored in a table and called
 * during the intialization of the kernel (in do_basic_setup in /init/main.c) 
 *
 * All this functionality is complements of the macros defined in linux/init.h
 */
module_init(i2c_algo_goku_s_init);


/* If MODULES is defined when this file is compiled, then this function will
 * resolved to cleanup_module.
 */
module_exit(i2c_algo_goku_s_exit);

#endif	/* defined(MODULES) */
