/*
 *  i2c-cxd9802.c -- CXD9802 I2C driver
 *
 */
/*
 *  Copyright 2004 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/module.h>
#include <linux/version.h>

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/types.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <linux/ioport.h>
#include <linux/errno.h>
#include <linux/sched.h>

#include <linux/i2c.h>

#include <asm-mips/cxd9802/cxd9802.h>

#define CXD9802_I2C_DEVNAME		"CXD9802 I2C"
#define CXD9802_I2C_DEFAULT_SCL_FREQ	100000	/* 100kHz */
#define CXD9802_I2C_WAIT_IDLE_TIMEOUT	100	/* 100msec */

#ifdef DEBUG
#define DPR(fmt, args...)	printk(fmt , ## args)
#endif /* DEBUG */

/* ======================================================================= *
 * Debug print function from dpr package                                   *
 * ======================================================================= */

#define DPL_EMERG	0
#define DPL_ALERT	1
#define DPL_CRIT	2
#define DPL_ERR		3
#define DPL_WARNING	4
#define DPL_NOTICE	5
#define DPL_INFO	6
#define DPL_DEBUG	7

#define DPM_PORPOSE	0x0000000f
#define DPM_ERR		0x00000001
#define DPM_TRACE	0x00000002
#define DPM_LAYER	0x000000f0
#define DPM_APPLICATION	0x00000010
#define DPM_MIDDLEWARE	0x00000020
#define DPM_DRIVER	0x00000040
#define DPM_KERNEL	0x00000080
#define DPM_USERDEF	0xffff0000

#ifdef DEBUG

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL DPL_DEBUG
#endif
#ifndef DEBUG_MASK
#define DEBUG_MASK  DPM_PORPOSE
#endif

#ifndef DPR
#include <stdio.h>
#define DPR(fmt, args...)	fprintf(stderr, fmt , ## args)
#endif

#define DPRL(level, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && DPR(fmt , ## args))
#define DPRM(mask, fmt, args...) \
	(void)(((mask) & DEBUG_MASK) && DPR(fmt , ## args))
#define DPRLM(level, mask, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && ((mask) & DEBUG_MASK) && \
		DPR(fmt , ## args))
#define DPRC(cond, fmt, args...) \
	(void)((cond) && DPR(fmt , ## args))
#define DPRF(fmt, args...)	DPR("%s: " fmt, __FUNCTION__ , ## args)
#define DPRFL(level, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && DPRF(fmt , ## args))
#define DPRFM(mask, fmt, args...) \
	(void)(((mask) & DEBUG_MASK) && DPRF(fmt , ## args))
#define DPRFLM(level, mask, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && ((mask) & DEBUG_MASK) && \
		DPRF(fmt , ## args))
#define DPRFC(cond, fmt, args...) \
	(void)((cond) && DPRF(fmt , ## args))

#define	DPRFS	DPRFLM(DPL_DEBUG, DPM_TRACE, "start\n")
#define	DPRFE	DPRFLM(DPL_DEBUG, DPM_TRACE, "end\n")
#define DPRFR	DPRFLM(DPL_DEBUG, DPM_TRACE, "return\n")

#else  /* DEBUG */

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL -1
#endif
#ifndef DEBUG_MASK
#define DEBUG_MASK  0
#endif

#ifndef DPR
#define DPR(fmt, args...)
#endif

#define DPRL(level, fmt, args...)
#define DPRM(mask, fmt, args...)
#define DPRLM(level, mask, fmt, args...)
#define DPRC(cond, fmt, args...)
#define DPRF(fmt, args...)
#define DPRFL(level, fmt, args...)
#define DPRFM(mask, fmt, args...)
#define DPRFLM(level, mask, fmt, args...)
#define DPRFC(cond, fmt, args...)

#define	DPRFS
#define	DPRFE
#define DPRFR

#endif /* DEBUG */

/* ======================================================================= *
 * I2C register                                                            *
 * ======================================================================= */

#define REG_I2C_ICTSR		0x00
#define  REGM_I2C_CY		(1 << 24)
#define  REGM_I2C_CNT		(0x1f << 16)
#define  REGM_I2C_ASTS		(1 << 7)
#define  REGM_I2C_DSTS		(1 << 6)
#define  REGM_I2C_AINTFL	(1 << 5)
#define  REGM_I2C_DINTFL	(1 << 4)
#define  REGM_I2C_AEN		(1 << 1)
#define  REGM_I2C_DEN		(1 << 0)
#define REG_I2C_TXRR		0x04
#define REG_I2C_CR		0x08
#define  REGM_STA		(1 << 7)
#define  REGM_STO		(1 << 6)
#define  REGM_RD		(1 << 5)
#define  REGM_WR		(1 << 4)
#define  REGM_ACK		(1 << 3)
#define REG_I2C_PRE		0x0c
#define REG_I2C_CTR		0x20
#define  REGM_I2Creset		(1 << 16)
#define  REGM_GIOSDAi		(1 << 11)
#define  REGM_GIOSCLi		(1 << 10)
#define  REGM_GIOSDAC		(1 << 9)
#define  REGM_GIOSCLC		(1 << 8)
#define  REGM_GPIOenable	(1 << 2)
#define  REGM_I2CModeSelect	(1 << 1)
#define  REGM_I2Cenable		(1 << 0)
#define REG_I2C_BTXRR		0x34
#define REG_I2C_BCR		0x38
#define  REGM_I2C_StopEN	(1 << 1)
#define  REGM_I2C_start		(1 << 0)
#define REG_I2C_BRBCNT		0x3c

/* ======================================================================= *
 * global variable                                                         *
 * ======================================================================= */

static spinlock_t cxd9802_i2c_irqlock = SPIN_LOCK_UNLOCKED;
static wait_queue_head_t cxd9802_i2c_txrx_wait;

#define CXD9802_I2C_TXRX_IDLE	0
#define CXD9802_I2C_TXRX_WRITE	1
#define CXD9802_I2C_TXRX_READ	2
static int     cxd9802_i2c_txrx_state = CXD9802_I2C_TXRX_IDLE;
static int     cxd9802_i2c_txrx_error = 0;

static char   *cxd9802_i2c_txrx_buf = NULL;
static size_t  cxd9802_i2c_txrx_len = 0;
static int     cxd9802_i2c_txrx_pos = 0;

/* ======================================================================= *
 * module parameter                                                        *
 * ======================================================================= */

MODULE_AUTHOR("Sony Corporation");
MODULE_DESCRIPTION("CXD9802 I2C driver");
MODULE_LICENSE("GPL");

static int scl_freq = CXD9802_I2C_DEFAULT_SCL_FREQ;
MODULE_PARM(scl_freq, "i");
MODULE_PARM_DESC(scl_ferq, "SCL frequency (Hz)");

/* ======================================================================= *
 * hardware interface (adapter)                                            *
 * ======================================================================= */

static void
cxd9802_i2c_reg_write(int offset, u32 value)
{
	cxd9802_reg_write32(CXD9802_I2C_BASE + offset, value);
}

static u32
cxd9802_i2c_reg_read(int offset)
{
	return cxd9802_reg_read32(CXD9802_I2C_BASE + offset);
}

static inline void
cxd9802_i2c_reg_changebits(int offset, u32 bits, u32 value)
{
	cxd9802_i2c_reg_write(offset,
			      (cxd9802_i2c_reg_read(offset) & ~bits) | value);
}

static inline void
cxd9802_i2c_reg_setbits(int offset, u32 bits)
{
	cxd9802_i2c_reg_write(offset, cxd9802_i2c_reg_read(offset) | bits);
}

static inline void
cxd9802_i2c_reg_clearbits(int offset, u32 bits)
{
	cxd9802_i2c_reg_changebits(offset, bits, 0);
}

/* ======================================================================= *
 * txrx byte (algorithm)                                                   *
 * ======================================================================= */

/* should be called with irqlock */
static void
cxd9802_i2c_tx_byte(void)
{
	u32 cr;
	DPRFS;

	if (cxd9802_i2c_txrx_pos == cxd9802_i2c_txrx_len) {
		/* finished */
		cxd9802_i2c_txrx_state = CXD9802_I2C_TXRX_IDLE;
		cxd9802_i2c_txrx_error = 0;
		wake_up_interruptible(&cxd9802_i2c_txrx_wait);
		DPRFR;
		return;
	}

	DPRFL(DPL_DEBUG, "txchar[%d]=%02x\n", cxd9802_i2c_txrx_pos,
	      cxd9802_i2c_txrx_buf[cxd9802_i2c_txrx_pos]);
	cxd9802_i2c_reg_write(REG_I2C_TXRR,
			      cxd9802_i2c_txrx_buf[cxd9802_i2c_txrx_pos]);
	cxd9802_i2c_txrx_pos++;

	cr = REGM_WR;
	if (cxd9802_i2c_txrx_pos == cxd9802_i2c_txrx_len) {
		/* last byte */
		cr |= REGM_STO;
	}
	cxd9802_i2c_reg_write(REG_I2C_CR, cr);

	DPRFE;
	return;
}

/* should be called with irqlock */
static void
cxd9802_i2c_rx_byte(void)
{
	u32 cr;
	DPRFS;

	if (cxd9802_i2c_txrx_pos >= 0) {
		cxd9802_i2c_txrx_buf[cxd9802_i2c_txrx_pos] =
			cxd9802_i2c_reg_read(REG_I2C_TXRR);
		DPRFL(DPL_DEBUG, "rxchar[%d]=%02x\n", cxd9802_i2c_txrx_pos,
		      (unsigned char)
		      cxd9802_i2c_txrx_buf[cxd9802_i2c_txrx_pos]);
	}
	cxd9802_i2c_txrx_pos++;
	if (cxd9802_i2c_txrx_pos == cxd9802_i2c_txrx_len) {
		/* finished */
		cxd9802_i2c_txrx_state = CXD9802_I2C_TXRX_IDLE;
		cxd9802_i2c_txrx_error = 0;
		wake_up_interruptible(&cxd9802_i2c_txrx_wait);
		DPRFR;
		return;
	}

	cr = REGM_RD;
	if (cxd9802_i2c_txrx_pos + 1 == cxd9802_i2c_txrx_len) {
		/* last byte */
		cr |= (REGM_ACK | REGM_STO);
	}
	cxd9802_i2c_reg_write(REG_I2C_CR, cr);

	DPRFE;
	return;
}

/* ======================================================================= *
 * interrupt handler                                                       *
 * ======================================================================= */

static void
cxd9802_i2c_intr_handler(int irq, void *devid, struct pt_regs *regs)
{
	u32 ictsr;
	DPRFS;

	ictsr = cxd9802_i2c_reg_read(REG_I2C_ICTSR);

	if ((ictsr & REGM_I2C_AINTFL) || (ictsr & REGM_I2C_DINTFL)) {
		/* clear the interrupts */
		cxd9802_i2c_reg_write(REG_I2C_ICTSR, ictsr);
	}

	DPRFL(DPL_DEBUG, "ictsr=%08x\n", ictsr);
	if ((ictsr & REGM_I2C_AINTFL) && (ictsr & REGM_I2C_AEN)) {
		/**** TODO ****/
	}
	if ((ictsr & REGM_I2C_DINTFL) && (ictsr & REGM_I2C_DEN)) {
		if (cxd9802_i2c_txrx_state == CXD9802_I2C_TXRX_WRITE) {
			if (ictsr & REGM_I2C_ASTS) {
				/* NACK received */
				cxd9802_i2c_txrx_state = CXD9802_I2C_TXRX_IDLE;
				cxd9802_i2c_txrx_error = -EIO;
				wake_up_interruptible(&cxd9802_i2c_txrx_wait);
			}
			cxd9802_i2c_tx_byte();
		} else if (cxd9802_i2c_txrx_state == CXD9802_I2C_TXRX_READ) {
			cxd9802_i2c_rx_byte();
		}
	}

	DPRFE;
	return;
}

/* ======================================================================= *
 * I2C adapter interface                                                   *
 * ======================================================================= */

static void
cxd9802_i2c_inc_use(struct i2c_adapter *adap)
{
	DPRFS;

	MOD_INC_USE_COUNT;

	DPRFE;
	return;
}

static void
cxd9802_i2c_dec_use(struct i2c_adapter *adap)
{
	DPRFS;

	MOD_DEC_USE_COUNT;

	DPRFE;
	return;
}

static int
cxd9802_i2c_client_register(struct i2c_client *client)
{
	DPRFS;

	DPRFE;
	return 0;
}

static int
cxd9802_i2c_client_unregister(struct i2c_client *client)
{
	DPRFS;

	DPRFE;
	return 0;
}

/* this variable is exported */
struct i2c_adapter cxd9802_i2c_adap = {
	name:			"CXD9802 I2C adapter",
	id:			0,
	algo:			NULL,
	algo_data:		NULL,
	inc_use:		cxd9802_i2c_inc_use,
	dec_use:		cxd9802_i2c_dec_use,
	client_register:	cxd9802_i2c_client_register,
	client_unregister:	cxd9802_i2c_client_unregister,
	data:			NULL,
};

/* ======================================================================= *
 * I2C algorithm interface                                                 *
 * ======================================================================= */

static int
cxd9802_i2c_wait_idle(int timeout_msec)
{
	unsigned long limit_jiffies;
	DPRFS;

	limit_jiffies = jiffies + (timeout_msec * HZ) / 1000;
	while (cxd9802_i2c_reg_read(REG_I2C_ICTSR) & REGM_I2C_DSTS) {
		schedule_timeout(1);
		if (jiffies > limit_jiffies) {
			DPRFR;
			return -EBUSY;
		}
	}

	DPRFE;
	return 0;
}

static int
cxd9802_i2c_read_bytes(struct i2c_msg *msg)
{
	int res;
	unsigned long flags;
	u32 cr;
	DPRFS;

	spin_lock_irqsave(&cxd9802_i2c_irqlock, flags);
	cxd9802_i2c_txrx_buf = msg->buf;
	cxd9802_i2c_txrx_len = msg->len;
	cxd9802_i2c_txrx_pos = -1;
	cxd9802_i2c_txrx_state = CXD9802_I2C_TXRX_READ;
	spin_unlock_irqrestore(&cxd9802_i2c_irqlock, flags);

	DPRFL(DPL_DEBUG, "rxaddr=%02x\n", ((msg->addr & 0x7f) << 1) | 1);
	cxd9802_i2c_reg_write(REG_I2C_TXRR, ((msg->addr & 0x7f) << 1) | 1);
	cr = REGM_WR;
	if (!(msg->flags & I2C_M_NOSTART)) {
		cr |= REGM_STA;
	}
	cxd9802_i2c_reg_write(REG_I2C_CR, cr);

	res = wait_event_interruptible(cxd9802_i2c_txrx_wait,
				       cxd9802_i2c_txrx_state ==
				       CXD9802_I2C_TXRX_IDLE);
	cxd9802_i2c_txrx_state = CXD9802_I2C_TXRX_IDLE;
	if (res < 0) {
		DPRFR;
		return res;
	}

	DPRFE;
	return cxd9802_i2c_txrx_error;
}

static int
cxd9802_i2c_write_bytes(struct i2c_msg *msg)
{
	int res;
	unsigned long flags;
	u32 cr;
	DPRFS;

	spin_lock_irqsave(&cxd9802_i2c_irqlock, flags);
	cxd9802_i2c_txrx_buf = msg->buf;
	cxd9802_i2c_txrx_len = msg->len;
	cxd9802_i2c_txrx_pos = 0;
	cxd9802_i2c_txrx_state = CXD9802_I2C_TXRX_WRITE;
	spin_unlock_irqrestore(&cxd9802_i2c_irqlock, flags);

	DPRFL(DPL_DEBUG, "txaddr=%02x\n", (msg->addr & 0x7f) << 1);
	cxd9802_i2c_reg_write(REG_I2C_TXRR, (msg->addr & 0x7f) << 1);
	cr = REGM_WR;
	if (!(msg->flags & I2C_M_NOSTART)) {
		cr |= REGM_STA;
	}
	cxd9802_i2c_reg_write(REG_I2C_CR, cr);

	res = wait_event_interruptible(cxd9802_i2c_txrx_wait,
				       cxd9802_i2c_txrx_state ==
				       CXD9802_I2C_TXRX_IDLE);
	cxd9802_i2c_txrx_state = CXD9802_I2C_TXRX_IDLE;
	if (res < 0) {
		DPRFR;
		return res;
	}

	DPRFE;
	return cxd9802_i2c_txrx_error;
}

/* this function is exported */
int
cxd9802_i2c_master_xfer(struct i2c_adapter *adap,
			struct i2c_msg msgs[], int num)
{
	int res;
	int i;
	DPRFS;

	for (i = 0; i < num; i++) {
		if (msgs[i].flags & I2C_M_TEN) {
			/* TODO: ten bit address is not supported */
			DPRFR;
			return -ENODEV;
		}
		if (msgs[i].len <= 0) {
			DPRFR;
			return -EINVAL;
		}

		res = cxd9802_i2c_wait_idle(CXD9802_I2C_WAIT_IDLE_TIMEOUT);
		if (res < 0) {
			DPRFR;
			return res;
		}

		if (msgs[i].flags & I2C_M_RD) {
			res = cxd9802_i2c_read_bytes(&(msgs[i]));
			if (res < 0) {
				DPRFR;
				return res;
			}
		} else {
			res = cxd9802_i2c_write_bytes(&(msgs[i]));
			if (res < 0) {
				DPRFR;
				return res;
			}
		}
	}

	DPRFE;
	return 0;
}

static int
cxd9802_i2c_algo_control(struct i2c_adapter *adap,
			 unsigned int cmd, unsigned long arg)
{
	DPRFS;

	DPRFE;
	return -EINVAL;
}

static u32
cxd9802_i2c_functionality(struct i2c_adapter *adap)
{
	DPRFS;

	DPRFE;
	return I2C_FUNC_SMBUS_EMUL;	/**** TODO ****/
}

static struct i2c_algorithm cxd9802_i2c_algo = {
	name:		"CXD9802 I2C algorithm",
	id:		I2C_ALGO_CXD9802,
	master_xfer:	cxd9802_i2c_master_xfer,
	smbus_xfer:	NULL,
	slave_send:	NULL,
	slave_recv:	NULL,
	algo_control:	cxd9802_i2c_algo_control,
	functionality:	cxd9802_i2c_functionality,
};

/* ======================================================================= *
 * resigter/unregister (algorithm)                                         *
 * ======================================================================= */

static int
i2c_cxd9802_add_bus(struct i2c_adapter *adap)
{
	int res;
	DPRFS;

	adap->id |= cxd9802_i2c_algo.id;
	adap->algo = &cxd9802_i2c_algo;

	res = i2c_add_adapter(adap);
	if (res < 0) {
		DPRFR;
		return res;
	}

	MOD_INC_USE_COUNT;

	DPRFE;
	return 0;
}

static int
i2c_cxd9802_del_bus(struct i2c_adapter *adap)
{
	int res;
	DPRFS;

	res = i2c_del_adapter(adap);
	if (res < 0) {
		DPRFR;
		return res;
	}

	MOD_DEC_USE_COUNT;

	DPRFE;
	return 0;
}

/* ======================================================================= *
 * initialize/finalize (adapter)                                           *
 * ======================================================================= */

/* this function is exported */
int __init
i2c_cxd9802_init(void)
{
	int res;
	static int initialized = 0;
	DPRFS;

	if (initialized) {
		DPRFR;
		return 0;
	}

	init_waitqueue_head(&cxd9802_i2c_txrx_wait);

	/* reset */
	cxd9802_i2c_reg_write(REG_I2C_CTR, REGM_I2Creset);
	cxd9802_i2c_reg_write(REG_I2C_CTR, 0);

	/* clear and disable interrupts */
	cxd9802_i2c_reg_write(REG_I2C_ICTSR,
			      REGM_I2C_AINTFL | REGM_I2C_DINTFL);

	/* set prescale value */
	cxd9802_i2c_reg_write(REG_I2C_PRE,
			      cxd9802_imbusclk_freq / (5 * scl_freq) - 1);

	/* enable I2C */
	cxd9802_i2c_reg_write(REG_I2C_CTR, REGM_I2Cenable);

	/* request irq */
	res = request_irq(CXD9802_IRQ_I2C, cxd9802_i2c_intr_handler,
			  SA_SHIRQ | SA_INTERRUPT, CXD9802_I2C_DEVNAME,
			  cxd9802_i2c_intr_handler);
	if (res < 0) {
		goto err1;
	}

	/* enable interrupts */
	cxd9802_i2c_reg_write(REG_I2C_ICTSR,
			      REGM_I2C_AEN | REGM_I2C_DEN);

	/* register adapter */
	res = i2c_cxd9802_add_bus(&cxd9802_i2c_adap);
	if (res < 0) {
		goto err2;
	}

	initialized = 1;

	DPRFE;
	return 0;

err2:
	free_irq(CXD9802_IRQ_I2C, cxd9802_i2c_intr_handler);
err1:
	DPRFR;
	return res;
}

static void
i2c_cxd9802_exit(void)
{
	DPRFS;

	i2c_cxd9802_del_bus(&cxd9802_i2c_adap);

	free_irq(CXD9802_IRQ_I2C, cxd9802_i2c_intr_handler);

	/* clear and disable interrupts */
	cxd9802_i2c_reg_write(REG_I2C_ICTSR,
			      REGM_I2C_AINTFL | REGM_I2C_DINTFL);

	/* disable device */
	cxd9802_i2c_reg_write(REG_I2C_CTR, 0);

	DPRFE;
	return;
}

module_init(i2c_cxd9802_init);
module_exit(i2c_cxd9802_exit);
