/*
 *  i2c-dbmx1.c
 *
 *  i2c algorithm & adapter for MPU110
 *
 *  Copyright 2001-2003 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <linux/ioport.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/proc_fs.h>
#include <linux/hfs_sysdep.h>

#include <linux/i2c.h>
#include <linux/i2c-algo-bit.h>
#include <linux/i2c-id.h>
#include <linux/i2c-dev.h>
#include <linux/delay.h>
#include <asm/arch/gpio.h>
#include <asm/irq.h>
#include "i2c-dbmx1.h"

#define I2C_DEBUG


#ifdef I2C_DEBUG
#define M_LEVEL 5
#define H_LEVEL 3
static int debug = 2;
MODULE_PARM(debug, "i");
MODULE_PARM_DESC(debug, "Debug enabled or not");
#define PDEBUG(level, format, arg...) \
          if (level <= debug) \
                 printk(KERN_DEBUG "%s: %s: " format , __FILE__ , __FUNCTION__ , ##arg)
#else
#define PDEBUG(level, format, arg...)
#endif 

static unsigned int status;
static wait_queue_head_t i2c_dbmx1_wait;
static struct proc_dir_entry *i2c_dbmx1_proc,*i2c_dbmx1_dir;

static int i2c_dbmx1_xfer(struct i2c_adapter *i2c_adap, struct i2c_msg msgs[], int num);
static int i2c_dbmx1_transfer(struct i2c_msg msgs[], int num);
static int i2c_dbmx1_ioctl(struct i2c_adapter *adapter,unsigned int cmd, unsigned long arg);

static u32 i2c_dbmx1_func(struct i2c_adapter *adap)
{
	return I2C_FUNC_PROTOCOL_MANGLING;
}
 

static struct i2c_algorithm i2c_dbmx1_algorithm = {
	/* name */ "DBMX1 I2C A/A",
	/* id */ I2C_ALGO_BIT,
	/* master_xfer */ i2c_dbmx1_xfer,
	/*  smbus_access */NULL,
	/* slave_send */ NULL,
	/* slave_rcv */ NULL,
	/* algo_control */ i2c_dbmx1_ioctl,
	/* functionality */i2c_dbmx1_func, 
};

static void i2c_dbmx1_inc(struct i2c_adapter *adap);
static void i2c_dbmx1_dec(struct i2c_adapter *adap);

static struct i2c_adapter i2c_csi_adapter = {
	"DBMX1 I2C adapter",
	I2C_ALGO_BIT | I2C_HW_B_SER,
	&i2c_dbmx1_algorithm,
	NULL,
	i2c_dbmx1_inc,
	i2c_dbmx1_dec,
	NULL,
	NULL
};


static void i2c_set(unsigned long data, unsigned long addr)
{
	unsigned long temp;
	temp = inl(addr);
	temp |= data;
	outl(temp, addr);
	return;
}	

static void i2c_unset(unsigned long data, unsigned long addr)
{
	unsigned long temp;
	temp = inl(addr);
	temp &= ~data;
	outl(temp, addr);
	return;
}

static void i2c_dbmx1_start(void)
{
	unsigned long temp;
	PDEBUG(M_LEVEL, "start \n");
	temp = inl(I2C_DBMX1_I2CR);
	temp |= I2C_DBMX1_I2CR_MSTA;
	outl(temp, I2C_DBMX1_I2CR);
	PDEBUG(M_LEVEL, "end \n");
	return;
}

static void i2c_dbmx1_stop(void)
{
	i2c_unset(I2C_DBMX1_I2CR_MSTA, I2C_DBMX1_I2CR);
	return;
}



static int i2c_dbmx1_bus_busy(void)
{
	unsigned int val;
	PDEBUG(H_LEVEL, "start \n");
	while (!(val = inl(I2C_DBMX1_I2SR) & I2C_DBMX1_I2SR_IBB )) {
		if (val & I2C_DBMX1_I2SR_IAL) {
			i2c_unset(I2C_DBMX1_I2SR_IAL, I2C_DBMX1_I2SR);
			return 1;
		}
	}
	PDEBUG(H_LEVEL, "end \n");
	return 0;
}

/* return when a stop signal is detect */
static void i2c_dbmx1_bus_release(void)
{
  	PDEBUG(M_LEVEL, "start \n");
	while (inl(I2C_DBMX1_I2SR) & (unsigned long) I2C_DBMX1_I2SR_IBB) ;
	PDEBUG(M_LEVEL, "end \n");
	return;
}

static void i2c_dbmx1_transfer_complete(void)
{
	PDEBUG(M_LEVEL, "complete 1\n");
	if(inl(I2C_DBMX1_I2SR) & I2C_DBMX1_I2SR_IIF){
	  i2c_unset(I2C_DBMX1_I2SR_IIF,I2C_DBMX1_I2SR);
	  return;
	}
	i2c_set(I2C_DBMX1_I2CR_IIEN, I2C_DBMX1_I2CR);
	wait_event_interruptible(i2c_dbmx1_wait,(inl(I2C_DBMX1_I2SR) & I2C_DBMX1_I2SR_IIF));
	//	interruptible_sleep_on(&i2c_dbmx1_wait);
	i2c_unset(I2C_DBMX1_I2SR_IIF,I2C_DBMX1_I2SR);
	PDEBUG(M_LEVEL, "complete 2\n");
	return;
}

static void i2c_dbmx1_int(int irq,void *dev_id,struct pt_regs *regs)
{
  	PDEBUG(M_LEVEL, "int 1\n");
	i2c_unset(I2C_DBMX1_I2CR_IIEN, I2C_DBMX1_I2CR);
	wake_up_interruptible(&i2c_dbmx1_wait);
	
	PDEBUG(M_LEVEL, "int 3\n");
	return;
}

static void i2c_dbmx1_inc(struct i2c_adapter *adap)
{
	MOD_INC_USE_COUNT;
	return;
}

static void i2c_dbmx1_dec(struct i2c_adapter *adap)
{
	MOD_DEC_USE_COUNT;
	return;
}

#ifdef CONFIG_DRAGONBALL_SNSC_MPU110
static int proc_read_i2c_dbmx1(char *page, char **start,
			      off_t off, int count, 
			      int *eof, void *data)
{
	short i;
	int len = 0;
	
	struct i2c_msg my_msg;
	char addr_data[2];
	char i2c_dbmx1_proc_data[70];

	addr_data[0] = 0;
	addr_data[1] = 0;

	my_msg.addr = MPU110_EEPROM_ADDR;
	my_msg.flags = I2C_M_WR;
	my_msg.buf = addr_data;
	my_msg.len = 2;
	i2c_dbmx1_transfer(&my_msg,1);
	my_msg.flags = I2C_M_RD;
	my_msg.buf = i2c_dbmx1_proc_data;
	my_msg.len = 64;
	i2c_dbmx1_transfer(&my_msg,1);
	
	len += sprintf(page+len, "EEPRPM TOP ADDR DATA\n");
	for(i=0;i < 4;i++){
		len += sprintf(page+len, "%c %c %c %c %c %c %c %c %c %c %c %c %c %c %c %c\n"
			       ,i2c_dbmx1_proc_data[(i*16)],i2c_dbmx1_proc_data[(i*16)+1]
			       ,i2c_dbmx1_proc_data[(i*16)+2],i2c_dbmx1_proc_data[(i*16)+3],
			       i2c_dbmx1_proc_data[(i*16)+4],i2c_dbmx1_proc_data[(i*16)+5],
			       i2c_dbmx1_proc_data[(i*16)+6],i2c_dbmx1_proc_data[(i*16)+7],
			       i2c_dbmx1_proc_data[(i*16)+8],i2c_dbmx1_proc_data[(i*16)+9],
			       i2c_dbmx1_proc_data[(i*16)+10],i2c_dbmx1_proc_data[(i*16)+11],
			       i2c_dbmx1_proc_data[(i*16)+12],i2c_dbmx1_proc_data[(i*16)+13],
			       i2c_dbmx1_proc_data[(i*16)+14],i2c_dbmx1_proc_data[(i*16)+15]);
	}
	return(len);
}

static int proc_write_i2c_dbmx1(struct file *file,
			     const char *buffer,
			     unsigned long count, 
			     void *data)
{
	unsigned long n;
	char *start;
	struct i2c_msg my_msg;
	char i2c_dbmx1_proc_data[100];
       
	n = count;
	if(n > 65)
	    n = 65;
	
        start = &i2c_dbmx1_proc_data[2];
	memset(i2c_dbmx1_proc_data,'0',100);
	if(copy_from_user(start,buffer,n-1))
	  return(-EFAULT);
	
	i2c_dbmx1_proc_data[0] = 0;
	i2c_dbmx1_proc_data[1] = 0;
	my_msg.addr = MPU110_EEPROM_ADDR;
	my_msg.flags = I2C_M_WR;
	my_msg.buf = i2c_dbmx1_proc_data;
	my_msg.len = 66;
	i2c_dbmx1_transfer(&my_msg,1);

	return n;
}
#endif


static int i2c_dbmx1_ioctl(struct i2c_adapter *adapter, unsigned int cmd, unsigned long arg)
{
	switch (cmd) {
	case I2C_DBMX1_CHANGE_FREQ:
		outl((unsigned long) (arg & 0x003F), I2C_DBMX1_IFDR);
		break;
	case I2C_DBMX1_GET_STATUS:
		arg = inl(I2C_DBMX1_I2SR);
		break;
	default:
		break;
	}
	return 0;
}

static int i2c_dbmx1_xfer(struct i2c_adapter *i2c_adap, struct i2c_msg msgs[], int num)
{
  return (i2c_dbmx1_transfer(msgs,num));
}

static int i2c_dbmx1_transfer(struct i2c_msg msgs[], int num)
{
	short i, j, count = 0;
	PDEBUG(M_LEVEL, "start \n");
	for (i = 0; i < num; i++) 
	  {  
	  start:
	    PDEBUG(H_LEVEL, "roop start \n");
	    outl(0x0, I2C_DBMX1_I2CR);/* reset */
	    outl(DEF_FREQ, I2C_DBMX1_IFDR);
	    i2c_set(I2C_DBMX1_I2CR_IEN, I2C_DBMX1_I2CR);
	    i2c_set(I2C_DBMX1_I2CR_TXAK, I2C_DBMX1_I2CR);
	    i2c_unset(I2C_DBMX1_I2CR_IIEN, I2C_DBMX1_I2CR);
	    PDEBUG(M_LEVEL, "addr = 0x%x  flgs = 0x%x len = %d buf = %p \n", msgs[i].addr, msgs[i].flags,  msgs[i].len, &msgs[i].buf[0]);
	    if(!(msgs[i].flags & I2C_M_NOSTART))
	      {
		i2c_dbmx1_start();
		//	i2c_set(I2C_DBMX1_I2CR_MTX, I2C_DBMX1_I2CR);
		if (i2c_dbmx1_bus_busy()) {
			i2c_dbmx1_stop();
			i2c_dbmx1_bus_release();
			goto start;
		}
		i2c_set(I2C_DBMX1_I2CR_MTX, I2C_DBMX1_I2CR);
		if (((msgs[i].flags) & (unsigned short) I2C_M_RD))
		  msgs[i].addr |= I2C_PROTCOL_READ;
		if (((msgs[i].flags) & (unsigned short) I2C_M_REV_DIR_ADDR))
		  msgs[i].addr ^= I2C_PROTCOL_READ;
		outl(msgs[i].addr, I2C_DBMX1_I2DR);
		PDEBUG(M_LEVEL, "target addr = 0x%d\n", msgs[i].addr);
		i2c_dbmx1_transfer_complete();
		PDEBUG(M_LEVEL, "flags = 0x%d\n", msgs[i].flags);
		if (((msgs[i].flags) & (unsigned short) I2C_M_RD)) {
			PDEBUG(M_LEVEL, "read\n");
			i2c_unset(I2C_DBMX1_I2CR_TXAK, I2C_DBMX1_I2CR);
			i2c_unset(I2C_DBMX1_I2CR_MTX, I2C_DBMX1_I2CR);
			msgs[i].buf[0] = inl(I2C_DBMX1_I2DR);
			PDEBUG(M_LEVEL, "complete2\n");
			i2c_dbmx1_transfer_complete();
			for (j = 0; j < msgs[i].len -2;j++) {
				msgs[i].buf[j] = inl(I2C_DBMX1_I2DR);
				//	PDEBUG(M_LEVEL, "complete3\n");
				i2c_dbmx1_transfer_complete();
				PDEBUG(M_LEVEL, "read data = %c\n",msgs[i].buf[j]);
			}
			PDEBUG(M_LEVEL, "read end\n");
			i2c_set(I2C_DBMX1_I2CR_TXAK, I2C_DBMX1_I2CR);
			msgs[i].buf[j] = inl(I2C_DBMX1_I2DR);
			i2c_dbmx1_transfer_complete();
			i2c_dbmx1_stop();
			msgs[i].buf[j+1] = inl(I2C_DBMX1_I2DR);
			i2c_dbmx1_bus_release();
		
		} 
		else {	
			PDEBUG(M_LEVEL, "write\n");
			for (j = 0; j < msgs[i].len ; j++) {
				PDEBUG(M_LEVEL, "write data = %c\n",msgs[i].buf[j]);
				outl(msgs[i].buf[j], I2C_DBMX1_I2DR);
				PDEBUG(M_LEVEL, "complete4\n");
				i2c_dbmx1_transfer_complete();
			}
			i2c_dbmx1_stop();
			i2c_dbmx1_bus_release();	

		}	
		count += msgs[i].len;
		//	i2c_unset(I2C_DBMX1_I2CR_RSTA, I2C_DBMX1_I2CR);
	      }
	    outl(0x0, I2C_DBMX1_I2CR);
	    PDEBUG(M_LEVEL, "END \n");
	  }		
	return count;
}

static int __init i2c_dbmx1_init(void)
{
	int result = 0;
	result = dragonball_register_gpios(PORT_A, (3 << 15), PRIMARY, "i2c");
	if (result < 0)
		goto gpio_fault;	
	init_waitqueue_head (&i2c_dbmx1_wait); 

	/* add dbmx1 I2C driver to i2c */
	if (i2c_add_adapter(&i2c_csi_adapter)) {
		goto i2c_adap_fault;
	}

	if (request_irq(IRQ_I2C_INT, i2c_dbmx1_int, SA_SHIRQ, "i2c_dbmx1", &status) < 0) { 
		printk("Failed request_irq!!\n");
		result = -EFAULT;
		goto irq_fault;
	}

#ifdef CONFIG_DRAGONBALL_SNSC_MPU110
	i2c_dbmx1_proc = create_proc_entry("dragonball_mx1/i2c", 0,i2c_dbmx1_dir );
	if(i2c_dbmx1_proc == NULL) {
		result = -ENOMEM;
		goto proc_ent;
	}

	i2c_dbmx1_proc->read_proc = proc_read_i2c_dbmx1;
	i2c_dbmx1_proc->write_proc = proc_write_i2c_dbmx1;
	i2c_dbmx1_proc->owner = THIS_MODULE;
#endif
	return (result);

#ifdef CONFIG_DRAGONBALL_SNSC_MPU110
 proc_ent:
	remove_proc_entry("dragonball_mx1/i2c",i2c_dbmx1_dir);
#endif

 irq_fault:
	free_irq(IRQ_I2C_INT, &status);

 i2c_adap_fault:
	i2c_del_adapter(&i2c_csi_adapter);
 gpio_fault:
	dragonball_unregister_gpios(PORT_A, (3 << 15));
	return (result);
}

static void __exit i2c_dbmx1_cleanup(void)
{
	i2c_del_adapter(&i2c_csi_adapter);
	dragonball_unregister_gpios(PORT_A, (3 << 15));
	remove_proc_entry("eeprom",i2c_dbmx1_dir);
	remove_proc_entry("i2c_dbmx1", NULL);
	free_irq(IRQ_I2C_INT, &status);
}

#ifdef MODULE
int init_module(void)
{
	printk(KERN_INFO "I2C driver for MPU-110\n");
	return i2c_dbmx1_init();
}

void cleanup_module(void)
{
	i2c_dbmx1_cleanup();
}
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif
#endif
