/*
 *  NOR/NAND Flash driver for Sony NSC MPU-110 IOA-3
 *
 * [NOR]
 *
 * This code is based on snsc_mpu110-flash.c,
 * which was based on physmap.c.
 *
 * [NAND]
 *
 * This code is based on snsc_mpu110-nand.c
 *
 * Copyright 2002,2003 Sony Corporation.
 * Copyright (C) 2000 Steven J. Hill (sjhill@cotw.com) 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>

#include <linux/slab.h>

#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/nand.h>
#include <asm/io.h>

#include <asm/arch/ioa_3.h>
#include <asm/arch/ioa_3_func.h>


#if !defined(CONFIG_MPU110_IOA_3_CS)
#error "CONFIG_MPU110_IOA_3_CS is not defined."
#endif


__u8  snsc_ioa3_read8(struct map_info *map, unsigned long ofs);
__u16 snsc_ioa3_read16(struct map_info *map, unsigned long ofs);
__u32 snsc_ioa3_read32(struct map_info *map, unsigned long ofs);
void  snsc_ioa3_copy_from(struct map_info *map, void *to,
			  unsigned long from, ssize_t len);
void  snsc_ioa3_write8(struct map_info *map, __u8 d, unsigned long adr);
void  snsc_ioa3_write16(struct map_info *map, __u16 d, unsigned long adr);
void  snsc_ioa3_write32(struct map_info *map, __u32 d, unsigned long adr);
void  snsc_ioa3_copy_to(struct map_info *map, unsigned long to,
			const void *from, ssize_t len);

void  snsc_ioa3_nand_control0( int cmd );
void  snsc_ioa3_nand_control1( int cmd );
int   snsc_ioa3_nand_ready( void );


#define IOA_3_MTD_DEVICES   (IOA_3_NOR_DEVICES + IOA_3_NAND_DEVICES)

static struct mtd_info *mtdinfo[IOA_3_MTD_DEVICES];

static u_int32_t nand_gpio_base = IOA_3_GPIO4_BASE;
static u_int32_t nand_fio_base = IOA_3_NAND_BASE;


struct map_info nor0_map = {
	name:      "IOA-3 NOR flash 0",
	buswidth:  2,
	read8:     snsc_ioa3_read8,
	read16:    snsc_ioa3_read16,
	read32:    snsc_ioa3_read32,
	copy_from: snsc_ioa3_copy_from,
	write8:    snsc_ioa3_write8,
	write16:   snsc_ioa3_write16,
	write32:   snsc_ioa3_write32,
	copy_to:   snsc_ioa3_copy_to
};

struct map_info nor1_map = {
	name:      "IOA-3 NOR flash 1",
	buswidth:  2,
	read8:     snsc_ioa3_read8,
	read16:    snsc_ioa3_read16,
	read32:    snsc_ioa3_read32,
	copy_from: snsc_ioa3_copy_from,
	write8:    snsc_ioa3_write8,
	write16:   snsc_ioa3_write16,
	write32:   snsc_ioa3_write32,
	copy_to:   snsc_ioa3_copy_to
};

/*
 * NOR raw access functions
 */
__u8 snsc_ioa3_read8(struct map_info *map, unsigned long ofs)
{
	return __raw_readb(map->map_priv_1 + ofs);
}

__u16 snsc_ioa3_read16(struct map_info *map, unsigned long ofs)
{
	return __raw_readw(map->map_priv_1 + ofs);
}

__u32 snsc_ioa3_read32(struct map_info *map, unsigned long ofs)
{
	return __raw_readl(map->map_priv_1 + ofs);
}

void snsc_ioa3_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	/* not much optimized in ARM */
	memcpy_fromio(to, map->map_priv_1 + from, len);
}

void snsc_ioa3_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	__raw_writeb(d, map->map_priv_1 + adr);
	mb();
}

void snsc_ioa3_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	__raw_writew(d, map->map_priv_1 + adr);
	mb();
}

void snsc_ioa3_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	__raw_writel(d, map->map_priv_1 + adr);
	mb();
}

void snsc_ioa3_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	/* not much optimized in ARM */
	memcpy_toio(map->map_priv_1 + to, from, len);
}


/*
 * NAND raw access funcions
 */
void snsc_ioa3_nand_control0( int cmd )
{
	static struct {
		u_int16_t bitmask;
		int value;
	} tab[] = {
		[NAND_CTL_SETCLE] = { IOA_3_NAND_CLE, 0 },
		[NAND_CTL_CLRCLE] = { IOA_3_NAND_CLE, 1 },
		[NAND_CTL_SETALE] = { IOA_3_NAND_ALE, 0 },
		[NAND_CTL_CLRALE] = { IOA_3_NAND_ALE, 1 },
		[NAND_CTL_SETNCE] = { IOA_3_NAND_CS2, 0 },
		[NAND_CTL_CLRNCE] = { IOA_3_NAND_CS2, 1 }
	};

	u_int16_t val;

	val = inw(nand_gpio_base);
	if (tab[cmd].value) {
		val |= tab[cmd].bitmask;
	} else {
		val &= ~tab[cmd].bitmask;
	}
	outw(val, nand_gpio_base);
}

void snsc_ioa3_nand_control1( int cmd )
{
	static struct {
		u_int16_t bitmask;
		int value;
	} tab[] = {
		[NAND_CTL_SETCLE] = { IOA_3_NAND_CLE, 0 },
		[NAND_CTL_CLRCLE] = { IOA_3_NAND_CLE, 1 },
		[NAND_CTL_SETALE] = { IOA_3_NAND_ALE, 0 },
		[NAND_CTL_CLRALE] = { IOA_3_NAND_ALE, 1 },
		[NAND_CTL_SETNCE] = { IOA_3_NAND_CS3, 0 },
		[NAND_CTL_CLRNCE] = { IOA_3_NAND_CS3, 1 }
	};

	u_int16_t val;

	val = inw(nand_gpio_base);
	if (tab[cmd].value) {
		val |= tab[cmd].bitmask;
	} else {
		val &= ~tab[cmd].bitmask;
	}
	outw(val, nand_gpio_base);
}

int snsc_ioa3_nand_ready( void )
{
	return (inw(nand_gpio_base) & IOA_3_NAND_BUSY) ? 1 : 0;
}


/*
 * NOR chip probe, iomap, setup mtd_info structure.
 */
int __init init_nor_device( struct mtd_info **pmtd, struct map_info *map,
			    u_int32_t base, u_int32_t size )
{
	struct mtd_info *mtd;

	printk(KERN_INFO "Flash device: 0x%x[bytes] at 0x%x\n", size, base);

	map->size = size;
	map->map_priv_1 = (unsigned long)ioremap_nocache(base, size);
	if (!map->map_priv_1) {
		printk(KERN_WARNING "Failed to ioremap\n");
		return -EIO;
	}

	mtd = do_map_probe("cfi_probe", map);
	if (!mtd)
		mtd = do_map_probe("jedec_probe", map);

	if (!mtd) {
		printk(KERN_INFO
		       "IOA-3 NOR device at 0x%x is not detected.\n", base);
		return 0;
	}

	mtd->module = THIS_MODULE;

	*pmtd = mtd;

	return 1;
}

/*
 * NAND chip probe, setup mtdinfo structure.
 */
int __init init_nand_device( struct mtd_info **pmtd, void (*hwctrl)(int) )
{
	struct mtd_info *mtd = NULL;
	struct nand_chip *chip = NULL;
	int ret = 0;

	mtd = kmalloc(sizeof(struct mtd_info) + sizeof(struct nand_chip),
		      GFP_KERNEL);
	if (!mtd) {
		printk(KERN_WARNING "Failed to kmalloc\n");
		ret = -ENOMEM;
		goto init_nand_err;
	}
	memset(mtd, 0, sizeof(struct mtd_info) + sizeof(struct nand_chip));

	chip = (struct nand_chip *)(mtd + 1);

	mtd->priv = chip;

	chip->IO_ADDR_R = IO_ADDRESS(nand_fio_base);
	chip->IO_ADDR_W = IO_ADDRESS(nand_fio_base);

	chip->hwcontrol = hwctrl;
	chip->dev_ready = snsc_ioa3_nand_ready;
	chip->cache_page = -1;

	if (nand_scan(mtd)) {
		printk(KERN_INFO "IOA-3 NAND device not detected.\n");
		ret = 0;
		goto init_nand_err;
	}

	chip->data_buf = kmalloc(sizeof(u_char) * (mtd->oobblock + mtd->oobsize), GFP_KERNEL);
	chip->data_cache = kmalloc(sizeof(u_char) * (mtd->oobblock + mtd->oobsize), GFP_KERNEL);

	if (chip->data_buf == NULL || chip->data_cache == NULL) {
		printk(KERN_WARNING "Failed to kmalloc\n");
		ret = -ENOMEM;
		goto init_nand_err;
	}

	*pmtd = mtd;

	return 1;

 init_nand_err:
	if (chip) {
		if (chip->data_buf)
			kfree(chip->data_buf);
		if (chip->data_cache)
			kfree(chip->data_cache);
	}
	if (mtd)
		kfree(mtd);

	return ret;
}


/*
 * init / cleanup routines
 */
static int __init init_snsc_ioa3_flash( void )
{
	int i;
	int err; /* currently not used meaningfully */

	printk(KERN_INFO "SNSC IOA-3 NOR/NAND driver\n");
	if (!ioa_3_probe()) {
		printk("No IOA-3 found\n");
		return -ENODEV;
	}

	err = init_nor_device(&mtdinfo[0], &nor0_map,
			      IOA_3_NOR4_BASE, IOA_3_NOR4_SIZE);

	err = init_nor_device(&mtdinfo[1], &nor1_map,
			      IOA_3_NOR5_BASE, IOA_3_NOR5_SIZE);

	snsc_ioa3_nand_control0(NAND_CTL_CLRNCE);
	snsc_ioa3_nand_control1(NAND_CTL_CLRNCE);
	snsc_ioa3_nand_control0(NAND_CTL_CLRCLE);
	snsc_ioa3_nand_control0(NAND_CTL_CLRALE);

	err = init_nand_device(&mtdinfo[2], &snsc_ioa3_nand_control0);
	if (mtdinfo[2]) {
		mtdinfo[2]->name = "IOA-3 NAND flash 0";
	}

	err = init_nand_device(&mtdinfo[3], &snsc_ioa3_nand_control1);
	if (mtdinfo[3]) {
		mtdinfo[3]->name = "IOA-3 NAND flash 1";
	}

	for ( i = 0; i < IOA_3_MTD_DEVICES; i++ ) {
		if (!mtdinfo[i])
			continue;
		add_mtd_device(mtdinfo[i]);
	}

	return 0;
}

static void __exit cleanup_nor_device( struct mtd_info *mtd, struct map_info *map )
{
	if (mtd) {
		del_mtd_device(mtd);
		map_destroy(mtd);
	}

	if (map->map_priv_1) {
		iounmap((void *)map->map_priv_1);
		(void *)map->map_priv_1 = NULL;
	}

	kfree(mtd);
}

static void __exit cleanup_nand_device( struct mtd_info *mtd )
{
	struct nand_chip *chip = (struct nand_chip *)(mtd + 1);

	if (!mtd)
		return;

	kfree(chip->data_buf);
	kfree(chip->data_cache);

	del_mtd_device(mtd);

	kfree(mtd);
}

static void __exit cleanup_snsc_ioa3_flash( void )
{
	cleanup_nor_device(mtdinfo[0], &nor0_map);
	cleanup_nor_device(mtdinfo[1], &nor1_map);

	cleanup_nand_device(mtdinfo[2]);
	cleanup_nand_device(mtdinfo[3]);
}

module_init(init_snsc_ioa3_flash);
module_exit(cleanup_snsc_ioa3_flash);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Sony NSC IOA-3 NOR/NAND Flash MTD driver");
