/*
 * NOR Flash support routine for MPU-110 Series
 *
 * Copyright 2002,2003 Sony Corporation.
 *
 * This code is based on physmap.c.
 *
 *   Id: physmap.c,v 1.15 2001/10/02 15:05:14 dwmw2 Exp $
 *  
 *   Normal mappings of chips in physical memory
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>
#include <linux/config.h>
#include <linux/ioport.h>
#include <asm/arch/platform.h>
#include <asm/arch/mpu110_series.h>

#ifndef NOR_CHIP_NUM
#error "NOR_CHIP_NUM is not defined."
#endif

#define WINDOW_ADDR_0                                    NOR0_BASE
#define WINDOW_SIZE_0                                    NOR0_SIZE
#define BUS_WIDTH_0                                 NOR0_BUS_WIDTH
#if   NOR_CHIP_NUM >= 2
#define WINDOW_ADDR_1                                    NOR1_BASE
#define WINDOW_SIZE_1                                    NOR1_SIZE
#define BUS_WIDTH_1                                 NOR1_BUS_WIDTH
#elif NOR_CHIP_NUM >= 3
#error "Invalid NOR_CHIP_NUM value."
#endif

#if defined(CONFIG_MTD_REDBOOT_PARTS) && defined(CONFIG_SNSC_MTD_SDM_PARTITION)
#error "both CONFIG_MTD_REDBOOT_PARTS and CONFIG_SNSC_MTD_SDM_PARTITION are defined."
#endif

#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
#include <linux/mtd/sdmparse.h>
#endif /* CONFIG_SNSC_MTD_SDM_PARTITION */

#ifdef CONFIG_MTD_REDBOOT_PARTS
int parse_redboot_partitions(struct mtd_info *master,
			     struct mtd_partition **pparts);
#endif /* CONFIG_MTD_REDBOOT_PARTS */


__u8 mpu110_series_read8(struct map_info *map, unsigned long ofs)
{
	return __raw_readb(map->map_priv_1 + ofs);
}

__u16 mpu110_series_read16(struct map_info *map, unsigned long ofs)
{
	return __raw_readw(map->map_priv_1 + ofs);
}

__u32 mpu110_series_read32(struct map_info *map, unsigned long ofs)
{
	return __raw_readl(map->map_priv_1 + ofs);
}

void mpu110_series_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	/* not much optimized in ARM */
	memcpy_fromio(to, map->map_priv_1 + from, len);
}

void mpu110_series_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	__raw_writeb(d, map->map_priv_1 + adr);
	mb();
}

void mpu110_series_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	__raw_writew(d, map->map_priv_1 + adr);
	mb();
}

void mpu110_series_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	__raw_writel(d, map->map_priv_1 + adr);
	mb();
}

void mpu110_series_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	/* not much optimized in ARM */
	memcpy_toio(map->map_priv_1 + to, from, len);
}

#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
static int sdm_nor_initialized = 0;
#endif

static struct mtd_info *mtd_0 = NULL;
static struct mtd_partition *parsed_parts_0 = NULL;

static struct map_info map_0 = {
#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
	name:      "sdm device NOR 0",
#else
	name:      "Physically mapped NOR Flash 0",
#endif
	size:      WINDOW_SIZE_0,
	buswidth:  BUS_WIDTH_0,
	read8:     mpu110_series_read8,
	read16:    mpu110_series_read16,
	read32:    mpu110_series_read32,
	copy_from: mpu110_series_copy_from,
	write8:    mpu110_series_write8,
	write16:   mpu110_series_write16,
	write32:   mpu110_series_write32,
	copy_to:   mpu110_series_copy_to
};

#if NOR_CHIP_NUM >= 2
static struct mtd_info *mtd_1 = NULL;
static struct mtd_partition *parsed_parts_1 = NULL;

static struct map_info map_1 = {
#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
	name:      "sdm device NOR 1",
#else
	name:      "Physically mapped NOR Flash 1",
#endif
	size:      WINDOW_SIZE_1,
	buswidth:  BUS_WIDTH_1,
	read8:     mpu110_series_read8,
	read16:    mpu110_series_read16,
	read32:    mpu110_series_read32,
	copy_from: mpu110_series_copy_from,
	write8:    mpu110_series_write8,
	write16:   mpu110_series_write16,
	write32:   mpu110_series_write32,
	copy_to:   mpu110_series_copy_to
};
#endif

#ifdef CONFIG_MTD_REDBOOT_PARTS
static int __init map_redboot_partitions( void )
{
	int i, j, ret = 0;

	struct mtd_partition *tmp_parts = NULL;
	int nr_tmpparts = 0;

	int nr_parts_0 = 0, nr_parts_1 = 0;
	int namelen_0 = 0, namelen_1 = 0;
	char *pname = NULL;

	/* parse the first flash.
	   it should have both two flash infomation */
	if (mtd_0 != NULL)
		nr_tmpparts = parse_redboot_partitions(mtd_0, &tmp_parts);

	if (nr_tmpparts > 0) {
		printk(KERN_NOTICE "Found RedBoot partition table.\n");
	} else {
		printk(KERN_NOTICE "No RedBoot partition found.\n");
	}

	/* split the retrieved information */

	/*
	 *  XXX: this code tries to split FIS partitions to NOR 0 and NOR 1
	 *       according to their 'offset' value.
	 *       however, current redboot partition parser rounds down the
	 *       'offset' value by it's master device size. as a result,
	 *       all partition will be mapped on NOR 0 incorrectly.
	 */
	for (i = 0; i < nr_tmpparts; i++) {
		printk(KERN_INFO "[%02d];0x%08x-0x%08x : %s\n",
		       i, tmp_parts[i].offset,
		       tmp_parts[i].offset + tmp_parts[i].size,
		       tmp_parts[i].name);
		if (tmp_parts[i].offset < WINDOW_SIZE_0) {
			nr_parts_0++;
			namelen_0 += strlen(tmp_parts[i].name) + 1;
		} else {
			nr_parts_1++;
			namelen_1 += strlen(tmp_parts[i].name) + 1;
		}
	}

	if ((mtd_0 != NULL) && (nr_parts_0 > 0)) {

		/* allocate partition information */
		parsed_parts_0 =
			kmalloc(sizeof(struct mtd_partition) * nr_parts_0
				+ namelen_0, GFP_KERNEL);
		if (parsed_parts_0 == NULL) {
			printk(KERN_WARNING "%s():%d: kmalloc() failed\n",
			       __FUNCTION__, __LINE__ );
			ret = -ENOMEM;
			goto map_err;
		}
		pname = (char *) &parsed_parts_0[nr_parts_0];

		for (i = 0, j = 0; i < nr_tmpparts; i++) {
			if (tmp_parts[i].offset < WINDOW_SIZE_0) {
				memcpy(&parsed_parts_0[j], &tmp_parts[i],
				       sizeof(struct mtd_partition));
				strcpy(pname, tmp_parts[i].name);
				parsed_parts_0[j].name = pname;
				pname += strlen(pname) + 1;
				j++;
			} 
		}
		add_mtd_partitions(mtd_0, parsed_parts_0, nr_parts_0);

	} else if (mtd_0 != NULL) {
		/* no FIS directory, add whole device */
		add_mtd_device(mtd_0);
	}

#if NOR_CHIP_NUM >= 2
        if ((mtd_1 != NULL) && (nr_parts_1 > 0)) {

		/* allocate partition information */
		parsed_parts_1 =
			kmalloc(sizeof(struct mtd_partition) * nr_parts_1
				+ namelen_1, GFP_KERNEL);
		if (parsed_parts_1 == NULL) {
			printk(KERN_WARNING "%s():%d: kmalloc() failed\n",
			       __FUNCTION__, __LINE__ );
			ret = -ENOMEM;
			goto map_err;
		}
		pname = (char *) &parsed_parts_1[nr_parts_1];

		for (i = 0, j = 0; i < nr_tmpparts; i++) {
			if (tmp_parts[i].offset >= WINDOW_SIZE_0) {
				memcpy(&parsed_parts_1[j], &tmp_parts[i],
				       sizeof(struct mtd_partition));
				strcpy(pname, tmp_parts[i].name);
				parsed_parts_1[j].offset &= (WINDOW_SIZE_1 - 1);
				parsed_parts_1[j].name = pname;
				pname += strlen(pname) + 1;
				j++;
			}
		}
		add_mtd_partitions(mtd_1, parsed_parts_1, nr_parts_1);

	} else if (mtd_1 != NULL) {
		/* no FIS directory, add whole device */
		add_mtd_device(mtd_1);
	}
#endif

	return 0;

 map_err:
	/* parsed_parts_x may not be freed here without unregistering
	   the registered partitions */

	if (tmp_parts != NULL)
		kfree(tmp_parts);
	
	return ret;
}
#endif /* CONFIG_MTD_REDBOOT_PARTS */


static int __init init_mpu110_series_map(unsigned long addr,
					 unsigned long size,
					 struct map_info *map,
					 struct mtd_info **mtd)
{
	printk(KERN_NOTICE "Flash device: %#lx[bytes] at %#lx\n", size, addr);

	map->map_priv_1 = (unsigned long)ioremap_nocache(addr, size);
	if (!map->map_priv_1) {
		printk(KERN_NOTICE "Failed to ioremap\n");
		return -EIO;
	}

	*mtd = do_map_probe("cfi_probe", map);
	if (*mtd == NULL) {
		printk(KERN_NOTICE "cfi probe failed.\n");
		iounmap((void *)map->map_priv_1);
		return -ENXIO;
	}

	(*mtd)->module = THIS_MODULE;
#ifndef CONFIG_MTD_REDBOOT_PARTS
	add_mtd_device(*mtd);
#endif
	return 0;
}

/* 
 *  initialise for two set of flash. 
 */
int __init init_mpu110_series_maps(void)
{
	int	err1;
#if NOR_CHIP_NUM >= 2
	int	err2;
#endif

#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
	if (sdm_nor_initialized)
		return 0;
#endif

	err1 = init_mpu110_series_map(WINDOW_ADDR_0, WINDOW_SIZE_0, 
				      &map_0, &mtd_0);
#if NOR_CHIP_NUM >= 2
	err2 = init_mpu110_series_map(WINDOW_ADDR_1, WINDOW_SIZE_1,
				      &map_1, &mtd_1);
#endif

#ifdef CONFIG_MTD_REDBOOT_PARTS
	if (err1 == 0)
		err1 = map_redboot_partitions();
#if NOR_CHIP_NUM >= 2
	if (err1 == 0)
		err1 = err2;
#endif
#endif

#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
	sdm_register_device(mtd_0);
#if NOR_CHIP_NUM >= 2
	sdm_register_device(mtd_1);
#endif
	sdm_nor_initialized = 1;
#endif

	return err1;
}

#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
int __init sdm_nor_init( void )
{
	return init_mpu110_series_maps();
}
#endif

static void __exit cleanup_mpu110_series_map(struct map_info *map,
					   struct mtd_info *mtd,
                                           struct mtd_partition *parts)
{
	if (mtd != NULL) {
#ifdef CONFIG_MTD_REDBOOT_PARTS
		if (parts != NULL) {
			del_mtd_partitions(mtd);
			kfree(parts);
		} else {
			del_mtd_device(mtd);
		}
#else
		del_mtd_device(mtd);
#endif /* CONFIG_MTD_REDBOOT_PARTS */
		map_destroy(mtd);
		kfree(mtd);

		if (map->map_priv_1) {
			iounmap((void *)map->map_priv_1);
			map->map_priv_1 = 0;
		}
	}
}

static void __exit cleanup_mpu110_series_maps(void)
{
	cleanup_mpu110_series_map(&map_0, mtd_0, parsed_parts_0);
	parsed_parts_0 = NULL;
	mtd_0 = NULL;
#if NOR_CHIP_NUM >= 2
	cleanup_mpu110_series_map(&map_1, mtd_1, parsed_parts_1);
	parsed_parts_1 = NULL;
	mtd_1 = NULL;
#endif

#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
	sdm_nor_initialized = 0;
#endif
}

module_init(init_mpu110_series_maps);
module_exit(cleanup_mpu110_series_maps);


MODULE_LICENSE("GPL");
MODULE_AUTHOR("David Woodhouse <dwmw2@infradead.org>");
MODULE_DESCRIPTION("Generic configurable MTD map driver");
