/*
 * NOR Flash support routine for Sony NSC MPU-210.
 *
 * Copyright 2002,2003 Sony Corporation.
 *
 * This code is based on physmap.c.
 *
 *   Id: physmap.c,v 1.15 2001/10/02 15:05:14 dwmw2 Exp $
 *  
 *   Normal mappings of chips in physical memory
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>
#include <linux/config.h>

#include <asm/it8172/it8172.h>
#include <asm/it8172/snsc_mpu210.h>

#define WINDOW_ADDR_0                             MPU210_NOR0_BASE
#define WINDOW_SIZE_0                             MPU210_NOR0_SIZE
#define BUS_WIDTH_0                                              4

#define WINDOW_ADDR_1                             MPU210_NOR1_BASE
#define WINDOW_SIZE_1                             MPU210_NOR1_SIZE
#define BUS_WIDTH_1                                              4

#if defined(CONFIG_MTD_REDBOOT_PARTS) && defined(CONFIG_SNSC_MTD_SDM_PARTITION)
#error "both CONFIG_MTD_REDBOOT_PARTS and CONFIG_SNSC_MTD_SDM_PARTITION are defined."
#endif

#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
#include <linux/mtd/sdmparse.h>
#endif /* CONFIG_SNSC_MTD_SDM_PARTITION */

#ifdef CONFIG_MTD_REDBOOT_PARTS
int parse_redboot_partitions(struct mtd_info *master,
			     struct mtd_partition **pparts);
#endif /* CONFIG_MTD_REDBOOT_PARTS */


__u8 snsc_mpu210_read8(struct map_info *map, unsigned long ofs)
{
	return __raw_readb(map->map_priv_1 + ofs);
}

__u16 snsc_mpu210_read16(struct map_info *map, unsigned long ofs)
{
	return __raw_readw(map->map_priv_1 + ofs);
}

__u32 snsc_mpu210_read32(struct map_info *map, unsigned long ofs)
{
	return __raw_readl(map->map_priv_1 + ofs);
}

void snsc_mpu210_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	memcpy_fromio(to, map->map_priv_1 + from, len);
}

void snsc_mpu210_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	__raw_writeb(d, map->map_priv_1 + adr);
	mb();
}

void snsc_mpu210_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	__raw_writew(d, map->map_priv_1 + adr);
	mb();
}

void snsc_mpu210_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	__raw_writel(d, map->map_priv_1 + adr);
	mb();
}

void snsc_mpu210_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	memcpy_toio(map->map_priv_1 + to, from, len);
}

static struct mtd_info *mtd_0 = NULL, *mtd_1 = NULL;
static struct mtd_partition *parsed_parts_0 = NULL, *parsed_parts_1 = NULL;

#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
static int sdm_nor_initialized = 0;
#endif

static struct map_info map_0 = {
#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
	name:      "sdm device NOR 0",
#else
	name:      "MPU-210 NOR Flash 0",
#endif
	size:      WINDOW_SIZE_0,
	buswidth:  BUS_WIDTH_0,
	read8:     snsc_mpu210_read8,
	read16:    snsc_mpu210_read16,
	read32:    snsc_mpu210_read32,
	copy_from: snsc_mpu210_copy_from,
	write8:    snsc_mpu210_write8,
	write16:   snsc_mpu210_write16,
	write32:   snsc_mpu210_write32,
	copy_to:   snsc_mpu210_copy_to
};

static struct map_info map_1 = {
#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
	name:      "sdm device NOR 1",
#else
	name:      "MPU-210 NOR flash 1",
#endif
	size:      WINDOW_SIZE_1,
	buswidth:  BUS_WIDTH_1,
	read8:     snsc_mpu210_read8,
	read16:    snsc_mpu210_read16,
	read32:    snsc_mpu210_read32,
	copy_from: snsc_mpu210_copy_from,
	write8:    snsc_mpu210_write8,
	write16:   snsc_mpu210_write16,
	write32:   snsc_mpu210_write32,
	copy_to:   snsc_mpu210_copy_to
};


#ifdef CONFIG_MTD_REDBOOT_PARTS
static int __init map_redboot_partitions(struct mtd_info *mtd,
					 struct mtd_partition **parts)
{
	int nr_parts = 0;

	if (mtd == NULL || parts == NULL)
		return -1;

	nr_parts = parse_redboot_partitions(mtd, parts);
	if (nr_parts > 0) {
		printk(KERN_NOTICE "Found RedBoot partition table.\n");
		add_mtd_partitions(mtd, *parts, nr_parts);
	} else {
		printk(KERN_NOTICE "No RedBoot partition found.\n");
		add_mtd_device(mtd);
	}

	return 0;
}
#endif /* CONFIG_MTD_REDBOOT_PARTS */


static int __init init_snsc_mpu210_map(unsigned long addr,
				       unsigned long size,
				       struct map_info *map,
				       struct mtd_info **mtd)
{
	printk(KERN_NOTICE "Flash device: %#lx[bytes] at %#lx\n", size, addr);

	map->map_priv_1 = (unsigned long)ioremap_nocache(addr, size);
	if (!map->map_priv_1) {
		printk(KERN_NOTICE "Failed to ioremap\n");
		return -EIO;
	}

	*mtd = do_map_probe("cfi_probe", map);
	if (*mtd == NULL) {
		printk(KERN_NOTICE "cfi probe failed.\n");
		iounmap((void *)map->map_priv_1);
		return -ENXIO;
	}

	(*mtd)->module = THIS_MODULE;
#ifndef CONFIG_MTD_REDBOOT_PARTS
	add_mtd_device(*mtd);
#endif
	return 0;
}

/* 
 *  initialise for two set of flash. 
 */
int __init init_snsc_mpu210_maps(void)
{
	int	err1;
	int	err2;

#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
	if (sdm_nor_initialized)
		return 0;
#endif

	err1 = init_snsc_mpu210_map(WINDOW_ADDR_0, WINDOW_SIZE_0,
				    &map_0, &mtd_0);
	err2 = init_snsc_mpu210_map(WINDOW_ADDR_1, WINDOW_SIZE_1,
				    &map_1, &mtd_1);

#ifdef CONFIG_MTD_REDBOOT_PARTS
	if (err1 == 0)
		err1 = map_redboot_partitions(mtd_0, &parsed_parts_0);
	if (err2 == 0)
		err2 = map_redboot_partitions(mtd_1, &parsed_parts_1);
#endif

	if (err1 == 0)
		err1 = err2;

#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
	sdm_register_device(mtd_0);
	sdm_register_device(mtd_1);
	sdm_nor_initialized = 1;
#endif

	return err1;
}

#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
int __init sdm_nor_init( void )
{
	return init_snsc_mpu210_maps();
}
#endif

static void __exit cleanup_snsc_mpu210_map(struct map_info *map,
					   struct mtd_info *mtd,
                                           struct mtd_partition *parts)
{
	if (mtd != NULL) {
#ifdef CONFIG_MTD_REDBOOT_PARTS
		if (parts != NULL) {
			del_mtd_partitions(mtd);
			kfree(parts);
		} else {
			del_mtd_device(mtd);
		}
#else
		del_mtd_device(mtd);
#endif /* CONFIG_MTD_REDBOOT_PARTS */
		map_destroy(mtd);
		kfree(mtd);

		if (map->map_priv_1) {
			iounmap((void *)map->map_priv_1);
			map->map_priv_1 = 0;
		}
	}
}

static void __exit cleanup_snsc_mpu210_maps(void)
{
	cleanup_snsc_mpu210_map(&map_0, mtd_0, parsed_parts_0);
	cleanup_snsc_mpu210_map(&map_1, mtd_1, parsed_parts_1);

	parsed_parts_0 = NULL;
	parsed_parts_1 = NULL;
	mtd_0 = NULL;
	mtd_1 = NULL;

#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
	sdm_nor_initialized = 0;
#endif
}

module_init(init_snsc_mpu210_maps);
module_exit(cleanup_snsc_mpu210_maps);


MODULE_LICENSE("GPL");
MODULE_AUTHOR("David Woodhouse <dwmw2@infradead.org>");
MODULE_DESCRIPTION("Generic configurable MTD map driver");
