/*
 * drivers/mtd/nand/cxd4707_2x-nand.c
 *
 * Based on tx4938ndfmc.c by source@mvista.com
 * Based on nand.c by Steven J. Hill (sjhill@cotw.com)
 * 		      Thomas Gleixner (tglx@linutronix.de)
 *
 *  Overview:
 *   This is a device driver for the NAND flash device connected to
 *   CXD4707 internal NAND flash interface.
 * 
 * Copyright 2004,2005 Sony Corporation.
 * 
 */
/*
 * drivers/mtd/nand/tx4938ndfmc.c
 *
 * Based on spia.c by Steven J. Hill
 *
 *  Overview:
 *   This is a device driver for the NAND flash device connected to
 *   TX4938 internal NAND Memory Controller.
 *   TX4938 NDFMC is almost same as TX4925 NDFMC, but register size are 64 bit.
 *
 * Author: source@mvista.com
 *
 * Copyright 2001-2002 MontaVista Software Inc.
 *
 * Copyright (C) 2000-2001 Toshiba Corporation 
 *
 * 2003 (c) MontaVista Software, Inc. This file is licensed under the
 * terms of the GNU General Public License version 2. This program is
 * licensed "as is" without any warranty of any kind, whether express
 * or implied.
 */
/*
 *  drivers/mtd/nand.c
 *
 *  Overview:
 *   This is the generic MTD driver for NAND flash devices. It should be
 *   capable of working with almost all NAND chips currently available.
 *   
 *	Additional technical information is available on
 *	http://www.linux-mtd.infradead.org/tech/nand.html
 *	
 *  Copyright (C) 2000 Steven J. Hill (sjhill@cotw.com)
 * 		  2002 Thomas Gleixner (tglx@linutronix.de)
 *
 *  10-29-2001  Thomas Gleixner (tglx@linutronix.de)
 * 		- Changed nand_chip structure for controlline function to
 *		support different hardware structures (Access to
 *		controllines ALE,CLE,NCE via hardware specific function. 
 *		- exit out of "failed erase block" changed, to avoid
 *		driver hangup
 *		- init_waitqueue_head added in function nand_scan !!
 *
 *  01-30-2002  Thomas Gleixner (tglx@linutronix.de)
 *		change in nand_writev to block invalid vecs entries
 *
 *  02-11-2002  Thomas Gleixner (tglx@linutronix.de)
 *		- major rewrite to avoid duplicated code
 *		  common nand_write_page function  
 *		  common get_chip function 
 *		- added oob_config structure for out of band layouts
 *		- write_oob changed for partial programming
 *		- read cache for faster access for subsequent reads
 *		from the same page.
 *		- support for different read/write address
 *		- support for device ready/busy line
 *		- read oob for more than one page enabled
 *
 *  02-27-2002	Thomas Gleixner (tglx@linutronix.de)
 *		- command-delay can be programmed
 *		- fixed exit from erase with callback-function enabled
 *
 *  03-21-2002  Thomas Gleixner (tglx@linutronix.de)
 *		- DEBUG improvements provided by Elizabeth Clarke 
 *		(eclarke@aminocom.com)
 *		- added zero check for this->chip_delay
 *
 *  04-03-2002  Thomas Gleixner (tglx@linutronix.de)
 *		- added added hw-driver supplied command and wait functions
 *		- changed blocking for erase (erase suspend enabled)
 *		- check pointers before accessing flash provided by
 *		John Hall (john.hall@optionexist.co.uk)
 *
 *  04-09-2002  Thomas Gleixner (tglx@linutronix.de)
 *		- nand_wait repaired
 *
 *  04-28-2002  Thomas Gleixner (tglx@linutronix.de)	
 *		- OOB config defines moved to nand.h 
 *
 *  08-01-2002  Thomas Gleixner (tglx@linutronix.de)	
 *		- changed my mailaddress, added pointer to tech/nand.html
 *
 *  08-07-2002 	Thomas Gleixner (tglx@linutronix.de)
 *		forced bad block location to byte 5 of OOB, even if
 *		CONFIG_MTD_NAND_ECC_JFFS2 is not set, to prevent
 *		erase /dev/mtdX from erasing bad blocks and destroying
 *		bad block info
 *
 *  08-10-2002 	Thomas Gleixner (tglx@linutronix.de)
 *		Fixed writing tail of data. Thanks to Alice Hennessy
 *		<ahennessy@mvista.com>.
 *
 *  08-10-2002 	Thomas Gleixner (tglx@linutronix.de)
 *		nand_read_ecc and nand_write_page restructured to support
 *		hardware ECC. Thanks to Steven Hein (ssh@sgi.com)
 *		for basic implementation and suggestions.
 *		3 new pointers in nand_chip structure:
 *		calculate_ecc, correct_data, enabled_hwecc 					 
 *		forcing all hw-drivers to support page cache
 *		eccvalid_pos is now mandatory
 *
 *  08-17-2002	tglx: fixed signed/unsigned missmatch in write.c
 *		Thanks to Ken Offer <koffer@arlut.utexas.edu> 	
 *
 *  08-29-2002  tglx: use buffered read/write only for non pagealigned 
 *		access, speed up the aligned path by using the fs-buffer
 *		reset chip removed from nand_select(), implicit done
 *		only, when erase is interrupted
 *		waitfuntion use yield, instead of schedule_timeout
 *		support for 6byte/512byte hardware ECC
 *		read_ecc, write_ecc extended for different oob-layout
 *		selections: Implemented NAND_NONE_OOB, NAND_JFFS2_OOB,
 *		NAND_YAFFS_OOB. fs-driver gives one of these constants
 *		to select the oob-layout fitting the filesystem.
 *		oobdata can be read together with the raw data, when
 *		the fs-driver supplies a big enough buffer.
 *		size = 12 * number of pages to read (256B pagesize)
 *		       24 * number of pages to read (512B pagesize)
 *		the buffer contains 8/16 byte oobdata and 4/8 byte
 *		returncode from calculate_ecc
 *		oobdata can be given from filesystem to program them
 *		in one go together with the raw data. ECC codes are
 *		filled in at the place selected by oobsel.
 *
 *  09-04-2002  tglx: fixed write_verify (John Hall (john.hall@optionexist.co.uk))
 *
 *  11-11-2002  tglx: fixed debug output in nand_write_page 
 *		(John Hall (john.hall@optionexist.co.uk))
 *
 *  11-25-2002  tglx: Moved device ID/ manufacturer ID from nand_ids.h
 *		Splitted device ID and manufacturer ID table. 
 *		Removed CONFIG_MTD_NAND_ECC, as it defaults to ECC_NONE for
 *		mtd->read / mtd->write and is controllable by the fs driver
 *		for mtd->read_ecc / mtd->write_ecc
 *		some minor cleanups
 *
 *  12-05-2000  tglx: Dave Ellis (DGE@sixnetio) provided the fix for
 *		WRITE_VERIFY long time ago. Thanks for remembering me.	
 *	
 * $Id: cxd4707_2x-nand.c,v 1.56 2005/03/09 14:28:01 ichimori Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/config.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/spinlock.h>
#include <linux/module.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#ifdef CONFIG_MTD_SBL
#include "../sbl/sbl.h"
#include "../sbl/ecc.h"
#include <linux/mtd/mtdsbl.h>
#endif	/* CONFIG_MTD_SBL */
#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
#include <linux/mtd/sdmparse.h>
#endif	/* CONFIG_MTD_PARTITIONS */
#include <asm/io.h>
#include <asm/bootinfo.h>

#include <asm/ppc4xx_gpio.h>
#include <asm/ppc4xx_pic.h>
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA
#include <asm/ppc4xx_dma.h>
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA */

#ifdef CONFIG_SNSC_MTD_NAND_GEOMETRY
#include "nand_io.h"	/* For I/O access macros. */
#ifdef CONFIG_SNSC_MTD_NAND_BYTE_SWAP
#ifdef CONFIG_SNSC_MTD_NAND_HOST_BE
#undef NAND_READL
#undef NAND_WRITEL
#define NAND_READL(addr)	in_be32((volatile u32 *)(addr))
#define NAND_WRITEL(addr, n)	out_be32((volatile u32 *)(addr), (n))
#endif	/* CONFIG_SNSC_MTD_NAND_HOST_BE */
#endif	/* CONFIG_SNSC_MTD_NAND_BYTE_SWAP */
#endif	/* CONFIG_SNSC_MTD_NAND_GEOMETRY */

#ifdef CONFIG_SNSC_MPU410
#include <platforms/snsc-mpu410.h>
#endif
#ifdef CONFIG_FRX
#include <platforms/frx.h>
#endif
#ifdef CONFIG_DTGNEXUS
#include <platforms/dtgnexus.h>
#endif


/*
 * CXD4707 NAND flash interface registers
 */
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA
// #define CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
#define CONFIG_MTD_NAND_CXD4707_2X_USE_DMA_READ
#define CONFIG_MTD_NAND_CXD4707_2X_USE_DMA_WRITE

#undef CONFIG_MTD_NAND_CXD4707_2X_ECC_TEST
#ifdef CONFIG_MTD_NAND_CXD4707_2X_ECC_TEST
#include <asm/time.h>	/* For get_tbl() */
#define	CONFIG_MTD_NAND_CXD4707_2X_ECC_ERROR_RATE	10000
#define CONFIG_MTD_NAND_CXD4707_2X_ECC_ERROR_BIT	1
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_ECC_TEST */
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA */

#ifdef CONFIG_CWING

/* C-Wing don't support GPIO controlled chip select inverter */
#ifdef CXD4707_NAND_CHIPS
#undef CXD4707_NAND_CHIPS
#endif	/* CXD4707_NAND_CHIPS */
#define CXD4707_NAND_CHIPS	1

#define EBC0_PB4CR		0xf001a000	/* XXX */
#define EBC0_PB4AP		0x8c840380	/* XXX */

#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA
#define NAND_DMA_CONTROLLER	0
#define NAND_DMA_SID		SID_0
#define NAND_DMA_LOCATION	EXTERNAL_PERIPHERAL
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA */

#else	/* CONFIG_CWING */

#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA
#define NAND_DMA_CONTROLLER	1
#define NAND_DMA_SID		SID_7
#define NAND_DMA_LOCATION	INTERNAL_PERIPHERAL
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA */

#endif	/* CONFIG_CWING */

#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA
#define NAND_DMA_PRIORITY	PRIORITY_MID_LOW

#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
#define DMA_CS(n)		(DMA_CS0 >> (n & 3))
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG */
#define DMA_IN_DMA_REQ(n)	(DMA_IN_DMA_REQ0 >> (n & 3))
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA */


#define CFG_BASE		 0x00000000
#define CS1_BASE		 0x00010000
#define CS1_IMAGE_BASE		 0x00020000


#define NAND_CONFIG		(CFG_BASE + 0x1000)
#define NAND_STATUS		(CFG_BASE + 0x1004)
#define NAND_PIO_CMD		(CFG_BASE + 0x1008)
#define NAND_PIO_ADDR		(CFG_BASE + 0x100c)
#define NAND_PIO_OPEND		(CFG_BASE + 0x1014)
#define NAND_PAGE_INTR		(CFG_BASE + 0x1038)
#define NAND_TIM_PRM		(CFG_BASE + 0x1200)
#define NAND_PIO_DATA		(CFG_BASE + 0x1300)


#define NAND_CONFIG_NDI			0x80000000
#define NAND_CONFIG_FCLR		0x40000000
#define NAND_CONFIG_WECC		0x00020000
#define NAND_CONFIG_RECC		0x00010000
#define NAND_CONFIG_PS			0x00000100
#define NAND_CONFIG_TC			0x00000002
#define NAND_CONFIG_DW			0x00000001

#define NAND_STATUS_WP			0x00000080
#define NAND_STATUS_CERR		0x00000002
#define NAND_STATUS_RDY			0x00000001

#define NAND_PIO_OPEND_PIOE		0x00000001

#define NAND_PAGE_INTR_MBSY		0x00000200
#define NAND_PAGE_INTR_MTE		0x00000100
#define NAND_PAGE_INTR_MASK		(NAND_PAGE_INTR_MBSY | \
					 NAND_PAGE_INTR_MTE)
#define NAND_PAGE_INTR_BSY		0x00000002
#define NAND_PAGE_INTR_TE		0x00000001

#define NAND_TIM_PRM_RTIM_MASK		0x1f000000
#define NAND_TIM_PRM_RTIM_SHIFT		24
#define NAND_TIM_PRM_UTIMW_MASK		0x001f0000
#define NAND_TIM_PRM_UTIMW_SHIFT	16
#define NAND_TIM_PRM_UTIMR_H_MASK	0x00001f00
#define NAND_TIM_PRM_UTIMR_H_SHIFT	8
#define NAND_TIM_PRM_UTIMR_S_MASK	0x0000001f
#define NAND_TIM_PRM_UTIMR_S_SHIFT	0

#ifdef CONFIG_MTD_NAND_CXD4707_2X_TUNING

/* When timing parametar is not defined, use default timing */
#ifndef CXD4707_NAND_TIM_PRM_RTIM
#define CXD4707_NAND_TIM_PRM_RTIM	23
#endif	/* CXD4707_NAND_TIM_PRM_RTIM */
#ifndef CXD4707_NAND_TIM_PRM_UTIMW
#define CXD4707_NAND_TIM_PRM_UTIMW	8
#endif	/* CXD4707_NAND_TIM_PRM_UTIMW */
#ifndef CXD4707_NAND_TIM_PRM_UTIMR_H
#define CXD4707_NAND_TIM_PRM_UTIMR_H	23
#endif	/* CXD4707_NAND_TIM_PRM_UTIMR_H */
#ifndef CXD4707_NAND_TIM_PRM_UTIMR_S
#define CXD4707_NAND_TIM_PRM_UTIMR_S	11
#endif	/* CXD4707_NAND_TIM_PRM_UTIMR_S */

#define CXD4707_NAND_TIM_PRM \
	(((CXD4707_NAND_TIM_PRM_RTIM    << NAND_TIM_PRM_RTIM_SHIFT)    &   \
		NAND_TIM_PRM_RTIM_MASK)                                  | \
	 ((CXD4707_NAND_TIM_PRM_UTIMW   << NAND_TIM_PRM_UTIMW_SHIFT)   &   \
		NAND_TIM_PRM_UTIMW_MASK)                                 | \
	 ((CXD4707_NAND_TIM_PRM_UTIMR_H << NAND_TIM_PRM_UTIMR_H_SHIFT) &   \
		NAND_TIM_PRM_UTIMR_H_MASK)                               | \
	 ((CXD4707_NAND_TIM_PRM_UTIMR_S << NAND_TIM_PRM_UTIMR_S_SHIFT) &   \
		NAND_TIM_PRM_UTIMR_S_MASK))
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_TUNING */


#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA

#define NAND_DMA_CTRL		(CFG_BASE + 0x1018)
#define NAND_DMA_INTR		(CFG_BASE + 0x101c)
#define NAND_PAGE_ADDR		(CFG_BASE + 0x1020)
#define NAND_PAGE_CTRL		(CFG_BASE + 0x1024)
#define NAND_PAGE_ERRP0		(CFG_BASE + 0x1028)
#define NAND_PAGE_ERRP_ND0	(CFG_BASE + 0x1030)
#define NAND_PAGE_ERRP_ND1	(CFG_BASE + 0x1034)
#define NAND_PAGE_ERRP_ND2	(CFG_BASE + 0x1040)
#define NAND_PAGE_ERRP_ND3	(CFG_BASE + 0x1044)
#define NAND_PAGE_ECC_STATUS0	(CFG_BASE + 0x1100)

#define NAND_PAGE_ECC_CODE_0	(CS1_BASE + 0x0800)
#define NAND_PAGE_EXTRA_0	(CS1_BASE + 0x0900)


#define NAND_DMA_CTRL_DMA		0x00000001

#define NAND_DMA_INTR_PCNT		0x04000000
#define NAND_DMA_INTR_PCNTS		0x02000000
#define NAND_DMA_INTR_ANA		0x01000000
#define NAND_DMA_INTR_DRO		0x00020000
#define NAND_DMA_INTR_DREQ		0x00010000
#define NAND_DMA_INTR_MEXT		0x00001000
#define NAND_DMA_INTR_MECC		0x00000800
#define NAND_DMA_INTR_MDT		0x00000400
#define NAND_DMA_INTR_MREQ		0x00000200
#define NAND_DMA_INTR_MTE		0x00000100
#define NAND_DMA_INTR_EXT		0x00000010
#define NAND_DMA_INTR_ECC		0x00000008
#define NAND_DMA_INTR_DT		0x00000004
#define NAND_DMA_INTR_REQ		0x00000002
#define NAND_DMA_INTR_TE		0x00000001

#define NAND_PAGE_CTRL_EPC		0x00000200
#define NAND_PAGE_CTRL_BAC		0x00000100
#define NAND_PAGE_CTRL_DIR		0x00000002
#define NAND_PAGE_CTRL_TS		0x00000001

#define NAND_PAGE_ERRP(n)	(NAND_PAGE_ERRP0 + ((n) << 2))

#define NAND_PAGE_ERRP_EVEN(n) \
		(NAND_PAGE_ERRP_ND0 + \
		 (NAND_PAGE_ERRP_ND2 - NAND_PAGE_ERRP_ND0) * (n))
#define NAND_PAGE_ERRP_ODD(n) \
		(NAND_PAGE_ERRP_ND1 + \
		 (NAND_PAGE_ERRP_ND3 - NAND_PAGE_ERRP_ND1) * (n))

#define ERROR_PAGE(n)			(ffs(n) - 1)

#define NAND_PAGE_ECC_STATUS(n)	(NAND_PAGE_ECC_STATUS0 + ((n) << 2))
#define ECC_STATUS_UCRR			0x02
#define ECC_STATUS_CRR			0x01
#define ECC_STATUS_MASK			(ECC_STATUS_UCRR | ECC_STATUS_CRR)
#define ECC_STATUS_0(n)			(((n) >>  0) & ECC_STATUS_MASK)
#define ECC_STATUS_1(n)			(((n) >>  8) & ECC_STATUS_MASK)
#define ECC_STATUS_2(n)			(((n) >>  4) & ECC_STATUS_MASK)
#define ECC_STATUS_3(n)			(((n) >> 12) & ECC_STATUS_MASK)
#define ECC_STATUS_4(n)			(((n) >> 16) & ECC_STATUS_MASK)
#define ECC_STATUS_5(n)			(((n) >> 24) & ECC_STATUS_MASK)
#define ECC_STATUS_6(n)			(((n) >> 20) & ECC_STATUS_MASK)
#define ECC_STATUS_7(n)			(((n) >> 28) & ECC_STATUS_MASK)

#define NAND_PAGE_ECC_CODE(n)	(NAND_PAGE_ECC_CODE_0 + ((n) << 2))

#define NAND_PAGE_EXTRA(n)	(NAND_PAGE_EXTRA_0 + ((n) << 2))


#define CACHE_ALIGNED(addr) \
		(((unsigned long)(addr) & (L1_CACHE_LINE_SIZE - 1)) == 0)

#ifdef CONFIG_CXD4707_ES_1_1
#define DMA_ALIGNED(addr)	(((unsigned long)(addr) & (16 - 1)) == 0)
#else	/* CONFIG_CXD4707_ES_1_1 */
#define DMA_ALIGNED(addr)	(((unsigned long)(addr) & ( 4 - 1)) == 0)
#endif	/* CONFIG_CXD4707_ES_1_1 */
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA */


static void __inline__ cxd4707_nand_set_config(unsigned long value);
static void __inline__ cxd4707_nand_clr_config(unsigned long value);
static unsigned long __inline__ cxd4707_nand_get_config(void);


/*
 * Macros for low-level register control
 */

#ifdef CONFIG_SNSC_MTD_NAND_GLOBAL_STATE

static spinlock_t nand_global_lock = SPIN_LOCK_UNLOCKED;
static volatile nand_state_t nand_global_state = FL_READY;
static wait_queue_head_t nand_global_wq;

#define CHIP_LOCK  (nand_global_lock)
#define CHIP_STATE (nand_global_state)
#define CHIP_WQ    (nand_global_wq)

#else /* CONFIG_SNSC_MTD_NAND_GLOBAL_STATE */

#define CHIP_LOCK  (this->chip_lock)
#define CHIP_STATE (this->state)
#define CHIP_WQ    (this->wq)

#endif /* CONFIG_SNSC_MTD_NAND_GLOBAL_STATE */

#define nand_select()	this->hwcontrol(NAND_CTL_SETNCE);
#define nand_deselect() this->hwcontrol(NAND_CTL_CLRNCE);

/*
 * NAND low-level MTD interface functions
 */
#ifndef CONFIG_MTD_NAND_CXD4707_2X_USE_DMA_READ
static int nand_read (struct mtd_info *mtd, loff_t from, size_t len,
		      size_t * retlen, u_char * buf);
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_USE_DMA_READ */
static int nand_read_ecc (struct mtd_info *mtd, loff_t from, size_t len,
			  size_t * retlen, u_char * buf, u_char * eccbuf,
			  int oobsel);
static int nand_read_oob (struct mtd_info *mtd, loff_t from, size_t len,
			  size_t * retlen, u_char * buf);
#ifndef CONFIG_MTD_NAND_CXD4707_2X_USE_DMA_WRITE
static int nand_write (struct mtd_info *mtd, loff_t to, size_t len,
		       size_t * retlen, const u_char * buf);
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_USE_DMA_WRITE */
static int nand_write_ecc (struct mtd_info *mtd, loff_t to, size_t len,
			   size_t * retlen, const u_char * buf,
			   u_char * eccbuf, int oobsel);
static int nand_write_oob (struct mtd_info *mtd, loff_t to, size_t len, 
			   size_t * retlen, const u_char *buf);
static int nand_writev (struct mtd_info *mtd, const struct iovec *vecs,
			unsigned long count, loff_t to, size_t * retlen);
static int nand_writev_ecc (struct mtd_info *mtd, const struct iovec *vecs,
			    unsigned long count, loff_t to, size_t * retlen,
			    u_char *eccbuf, int oobsel);
static int nand_erase (struct mtd_info *mtd, struct erase_info *instr);
static void nand_sync (struct mtd_info *mtd);
static int nand_write_page (struct mtd_info *mtd, struct nand_chip *this,
			    int page, int col, 
			    int last, u_char *oob_buf, int oobsel);
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA
static int nand_dma_read(struct mtd_info *mtd, loff_t from,
                         size_t len, size_t *retlen, u_char *buf);
static int nand_dma_read_ecc(struct mtd_info *mtd, loff_t from,
                             size_t len, size_t *retlen,
                             u_char *buf, u_char *eccbuf,
                             int oobsel);
static int nand_dma_read_ecc_compat(struct mtd_info *mtd, loff_t from,
                                    size_t len, size_t *retlen,
                                    u_char *buf, u_char *eccbuf,
                                    int oobsel);
static int nand_dma_write(struct mtd_info *mtd, loff_t to,
                          size_t len, size_t *retlen, const u_char *buf);
static int nand_dma_write_ecc(struct mtd_info *mtd, loff_t to,
                              size_t len, size_t *retlen,
                              const u_char *buf, u_char *eccbuf,
                              int oobsel);
static int nand_dma_read_page(struct mtd_info *mtd, unsigned int page,
                              u_char *buf, u_char *eccbuf);
static int nand_dma_write_page(struct mtd_info *mtd, unsigned int page,
                               const u_char *buf, const u_char *eccbuf);
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA */


/*
 * NAND flash control routine
 */

/* Basic access macro for CXD4707 NAND flash interface registers */
/* Byte swap needed */
#define REG_READL(addr)		readl(addr)
#define REG_WRITEL(n, addr)	writel((n), (addr))

/* Base address for CXD4707 NAND flash interafce registers */
static char *cxd4707_nand;


/* NAND_CONFIG register access routine */
static void __inline__ cxd4707_nand_set_config(unsigned long value)
{
	REG_WRITEL(REG_READL(cxd4707_nand + NAND_CONFIG) | value,
		   cxd4707_nand + NAND_CONFIG);
}

static void __inline__ cxd4707_nand_clr_config(unsigned long value)
{
	REG_WRITEL(REG_READL(cxd4707_nand + NAND_CONFIG) & ~value,
		   cxd4707_nand + NAND_CONFIG);
}

static unsigned long __inline__ cxd4707_nand_get_config(void)
{
	return REG_READL(cxd4707_nand + NAND_CONFIG);
}


/* Set NAND flash to unwritable state */
static __inline__ void
nand_write_protect(void) {
	u32 status;

	/* EMC_PHY_ND_STATUS register */
	status = REG_READL(cxd4707_nand + NAND_STATUS);
	if ((status & NAND_STATUS_WP) != 0) {
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
		printk("STATUS: %08x -> ", status);
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG */
		status &= ~NAND_STATUS_WP;	/* Enable Write Protect */
		REG_WRITEL(status, cxd4707_nand + NAND_STATUS);
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
		status = REG_READL(cxd4707_nand + NAND_STATUS);
		printk("%08x\n", status);
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG */
	}
}

/* Set NAND flash to writable state */
static __inline__ void
nand_write_unprotect(void) {
	u32 status;

	/* EMC_PHY_ND_STATUS register */
	status = REG_READL(cxd4707_nand + NAND_STATUS);
	if ((status & NAND_STATUS_WP) == 0) {
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
		printk("STATUS: %08x -> ", status);
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG */
		status |= NAND_STATUS_WP;	/* Disable Write Protect */
		REG_WRITEL(status, cxd4707_nand + NAND_STATUS);
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
		status = REG_READL(cxd4707_nand + NAND_STATUS);
		printk("%08x\n", status);
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG */
	}
}


static inline int nand_get_status(struct mtd_info *mtd)
{
	struct nand_chip *this = mtd->priv;
	u_char status;

	this->cmdfunc (mtd, NAND_CMD_STATUS, -1, -1);

	if (this->interleaved) {
		u_char status1, status2;
		NAND_READB_2(this->IO_ADDR_R, status1, status2);
		status = (((status1 | status2) & 0x01) |
			  ((status1 & status2) & ~0x01));
	} else {
		status = NAND_READB (this->IO_ADDR_R);
	}

	return status;
}

/*
 *	Get chip for selected access
 */
static inline void nand_get_chip (struct nand_chip *this, struct mtd_info *mtd, int new_state, int *erase_state)
{

	DECLARE_WAITQUEUE (wait, current);

	/* 
	 * Grab the lock and see if the device is available 
	 * For erasing, we keep the spinlock until the
	 * erase command is written. 
	*/
retry:
	spin_lock_bh (&CHIP_LOCK);

	if (CHIP_STATE == FL_READY) {
		CHIP_STATE = new_state;
		if (new_state != FL_ERASING)
			spin_unlock_bh (&CHIP_LOCK);
		return;
	}

#ifndef CONFIG_SNSC /* disable 'erase suspend emulation' because sometimes
		       resetting an undergoing erase caused chip hangup */
	if (CHIP_STATE == FL_ERASING) {
		if (new_state != FL_ERASING) {
			CHIP_STATE = new_state;
			spin_unlock_bh (&CHIP_LOCK);
			nand_select ();	/* select in any case */
			this->cmdfunc(mtd, NAND_CMD_RESET, -1, -1);
			return;
		}
	}
#endif

	set_current_state (TASK_UNINTERRUPTIBLE);
	add_wait_queue (&CHIP_WQ, &wait);
	spin_unlock_bh (&CHIP_LOCK);
	schedule ();
	remove_wait_queue (&CHIP_WQ, &wait);
	goto retry;
}

/*
 *	Nand_page_program function is used for write and writev !
 */
static int nand_write_page (struct mtd_info *mtd, struct nand_chip *this,
			    int page, int col, int last, u_char *oob_buf, int oobsel)
{
	int 	i, status;
	u_char	*oob_data;

	/* pad oob area, if we have no oob buffer from fs-driver */
	if (!oob_buf) {
		oob_data = &this->data_buf[mtd->oobblock];
		for (i = 0; i < mtd->oobsize; i++)
			oob_data[i] = 0xff;
	} else 
		oob_data = oob_buf;
	
	/* Prepad for partial page programming !!! */
	for (i = 0; i < col; i++)
		this->data_poi[i] = 0xff;

	/* Postpad for partial page programming !!! oob is already padded */
	for (i = last; i < mtd->oobblock; i++)
		this->data_poi[i] = 0xff;

	/* Send command to begin auto page programming */
	this->cmdfunc (mtd, NAND_CMD_SEQIN, 0x00, page);

#ifdef CONFIG_SNSC_MTD_NAND_GEOMETRY
#ifdef CONFIG_SNSC_MTD_NAND_B4
	if ((mtd->oobblock % 4) != 0) {
		printk (KERN_WARNING "%s: oobblock (%d) "
			"is not a integer multiple "
			"of data port size (%d)\n",
			__FUNCTION__, mtd->oobblock, 4);
		BUG();
	}
	for (i = 0; i < mtd->oobblock; i += 4) 
		NAND_WRITEL(this->IO_ADDR_W,
			    *((u32 *)(&(this->data_poi[i]))));
#else	/* CONFIG_SNSC_MTD_NAND_B4 */
#error CONFIG_SNSC_MTD_NAND_B2 is not supported.
#endif	/* CONFIG_SNSC_MTD_NAND_B4 */
#else	/* CONFIG_SNSC_MTD_NAND_GEOMETRY */
	for (i = 0; i < mtd->oobblock; i++) 
		writeb ( this->data_poi[i] , this->IO_ADDR_W);
#endif	/* CONFIG_SNSC_MTD_NAND_GEOMETRY */
		
	/* Write out OOB data */
#ifdef CONFIG_SNSC_MTD_NAND_GEOMETRY
#ifdef CONFIG_SNSC_MTD_NAND_B4
	if ((mtd->oobsize % 4) != 0) {
		printk (KERN_WARNING "%s: oobsize (%d) "
		      "is not a integer multiple "
		      "of data port size (%d)\n",
		      __FUNCTION__, mtd->oobsize, 4);
		BUG();
	}
	for (i = 0; i < mtd->oobsize; i += 4) 
		NAND_WRITEL(this->IO_ADDR_W,
		            *((u32 *)(&(oob_data[i]))));
#else	/* CONFIG_SNSC_MTD_NAND_B4 */
#error CONFIG_SNSC_MTD_NAND_B2 is not supported.
#endif	/* CONFIG_SNSC_MTD_NAND_B4 */
#else	/* CONFIG_SNSC_MTD_NAND_GEOMETRY */
	for (i = 0; i <  mtd->oobsize; i++)
		writeb ( oob_data[i] , this->IO_ADDR_W);
#endif	/* CONFIG_SNSC_MTD_NAND_GEOMETRY */

	/* Send command to actually program the data */
	this->cmdfunc (mtd, NAND_CMD_PAGEPROG, -1, -1);

	/* call wait ready function */
	status = this->waitfunc (mtd, this, FL_WRITING);

	/* See if device thinks it succeeded */
	if (status & 0x01) {
		DEBUG (MTD_DEBUG_LEVEL0, "%s: " "Failed write, page 0x%08x, ", __FUNCTION__, page);
		return -EIO;
	}

#ifdef CONFIG_MTD_NAND_VERIFY_WRITE
	/*
	 * The NAND device assumes that it is always writing to
	 * a cleanly erased page. Hence, it performs its internal
	 * write verification only on bits that transitioned from
	 * 1 to 0. The device does NOT verify the whole page on a
	 * byte by byte basis. It is possible that the page was
	 * not completely erased or the page is becoming unusable
	 * due to wear. The read with ECC would catch the error
	 * later when the ECC page check fails, but we would rather
	 * catch it early in the page write stage. Better to write
	 * no data than invalid data.
	 */

	/* Send command to read back the page */
	this->cmdfunc (mtd, NAND_CMD_READ0, col, page);
	/* Loop through and verify the data */
#ifdef CONFIG_SNSC_MTD_NAND_GEOMETRY
#ifdef CONFIG_SNSC_MTD_NAND_B4
	if (((last - col) % 4) != 0) {
		DEBUG(MTD_DEBUG_LEVEL0, "%s: data verify size (%d) "
		      "is not a integer multiple "
		      "of data port size (%d)\n",
		      __FUNCTION__, last - col, 4);
	}
	for (i = col; i < last; i += 4) {
		if (*((u32 *)(&(this->data_poi[i]))) !=
		    NAND_READL (this->IO_ADDR_R)) {
			DEBUG (MTD_DEBUG_LEVEL0, "%s: " "Failed write verify, page 0x%08x ", __FUNCTION__, page);
			return -EIO;
		}
	}
#else	/* CONFIG_SNSC_MTD_NAND_B4 */
#error CONFIG_SNSC_MTD_NAND_B2 is not supported.
#endif	/* CONFIG_SNSC_MTD_NAND_B4 */
#else	/* CONFIG_SNSC_MTD_NAND_GEOMETRY */
	for (i = col; i < last; i++) {
		if (this->data_poi[i] != readb (this->IO_ADDR_R)) {
			DEBUG (MTD_DEBUG_LEVEL0, "%s: " "Failed write verify, page 0x%08x ", __FUNCTION__, page);
			return -EIO;
		}
	}
#endif	/* CONFIG_SNSC_MTD_NAND_GEOMETRY */

	/* check, if we have a fs-supplied oob-buffer */
	if (oob_buf) {
#ifdef CONFIG_SNSC_MTD_NAND_GEOMETRY
#ifdef CONFIG_SNSC_MTD_NAND_B4
		if ((mtd->oobsize % 4) != 0) {
			printk (KERN_WARNING "%s: oobsize (%d) "
			      "is not a integer multiple "
			      "of data port size (%d)\n",
			      __FUNCTION__, mtd->oobsize, 4);
			BUG();
		}
		for (i = 0; i < mtd->oobsize; i += 4) {
			if (*((u32 *)(&(oob_data[i]))) !=
			    NAND_READL (this->IO_ADDR_R)) {
				DEBUG (MTD_DEBUG_LEVEL0, "%s: " "Failed write verify, page 0x%08x ", __FUNCTION__, page);
				return -EIO;
			}
		}
#else	/* CONFIG_SNSC_MTD_NAND_B4 */
#error CONFIG_SNSC_MTD_NAND_B2 is not supported.
#endif	/* CONFIG_SNSC_MTD_NAND_B4 */
#else	/* CONFIG_SNSC_MTD_NAND_GEOMETRY */
		for (i = 0; i < mtd->oobsize; i++) {
			if (oob_data[i] != readb (this->IO_ADDR_R)) {
				DEBUG (MTD_DEBUG_LEVEL0, "%s: " "Failed write verify, page 0x%08x ", __FUNCTION__, page);
				return -EIO;
			}
		}
#endif	/* CONFIG_SNSC_MTD_NAND_GEOMETRY */
	}
#endif
	return 0;
}

#ifndef CONFIG_MTD_NAND_CXD4707_2X_USE_DMA_READ
/*
*	Use NAND read ECC
*/
static int nand_read (struct mtd_info *mtd, loff_t from, size_t len, size_t * retlen, u_char * buf)
{
	return (nand_read_ecc (mtd, from, len, retlen, buf, NULL, 0));
}			   
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_USE_DMA_READ */


/*
 * NAND read with ECC
 */
static int nand_read_ecc (struct mtd_info *mtd, loff_t from, size_t len,
			  size_t * retlen, u_char * buf, u_char * oob_buf, int oobsel)
{
	int j, col, page, end;
	int erase_state = 0;
	int read = 0, oob = 0;
	struct nand_chip *this = mtd->priv;
	u_char *data_poi, *oob_data = oob_buf;

	DEBUG (MTD_DEBUG_LEVEL3, "nand_read_ecc: from = 0x%08x, len = %i\n", (unsigned int) from, (int) len);

	/* Do not allow reads past end of device */
	if ((from + len) > mtd->size) {
		DEBUG (MTD_DEBUG_LEVEL0, "nand_read_ecc: Attempt read beyond end of device\n");
		*retlen = 0;
		return -EINVAL;
	}

	/* Grab the lock and see if the device is available */
	nand_get_chip (this, mtd ,FL_READING, &erase_state);

	/* Select the NAND device */
	nand_select ();

	/* First we calculate the starting page */
	page = from >> this->page_shift;

	/* Get raw starting column */
	col = from & (mtd->oobblock - 1);

	end = mtd->oobblock;

#ifndef CONFIG_SNSC /* fix for consecutive read. we issue read command later */
	/* Send the read command */
	this->cmdfunc (mtd, NAND_CMD_READ0, 0x00, page);
#endif
	
	/* Loop until all data read */
	while (read < len) {
		
#ifdef CONFIG_SNSC /* fix for consecutive page read */
#ifdef CONFIG_SNSC_MTD_NAND_SEQ_ROW_READ
		/* for NAND chips which support sequential row read */
		/* check if this is a first read or first page in a block */
		if ( (read == 0) ||
		     ((page % (mtd->erasesize >> this->page_shift)) == 0) ) {
			this->cmdfunc (mtd, NAND_CMD_READ0, 0x00, page);
			DEBUG(MTD_DEBUG_LEVEL3,
			      "nand_read_ecc: read command issued, "
			      "page=%x\n", page);
		} else {
			DEBUG(MTD_DEBUG_LEVEL3,
			      "nand_read_ecc: read command skipped, "
			      "page=%x\n", page);
			/* apply delay or wait for ready/busy pin */
			if (!this->dev_ready)
				udelay (this->chip_delay);
			else
				while(!this->dev_ready());
		}
#else /* CONFIG_SNSC_MTD_NAND_SEQ_ROW_READ */
		/* issue read command for every page, no sequential row read */
		this->cmdfunc (mtd, NAND_CMD_READ0, 0x00, page);
		DEBUG(MTD_DEBUG_LEVEL3,
		      "nand_read_ecc: read command issued, page=%x\n", page);
		
#endif /* CONFIG_SNSC_MTD_NAND_SEQ_ROW_READ */
#else  /* CONFIG_SNSC */
		/* If we have consequent page reads, apply delay or wait for ready/busy pin */
		if (read) {
			if (!this->dev_ready) 
				udelay (this->chip_delay);
			else
				while (!this->dev_ready());	
		}
#endif /* CONFIG_SNSC */

		/* 
		 * If the read is not page aligned, we have to read into data buffer
		 * due to ecc, else we read into return buffer direct
		 */
		if (!col && (len - read) >= end)  
			data_poi = &buf[read];
		else 
			data_poi = this->data_buf;

		/* get oob area, if we have no oob buffer from fs-driver */
		if (!oob_buf) {
			oob_data = &this->data_buf[end];
			oob = 0;
		} 	
			
		j = 0;

#ifdef CONFIG_SNSC_MTD_NAND_GEOMETRY
#ifdef CONFIG_SNSC_MTD_NAND_B4
		if ((end % 4) != 0) {
			DEBUG(MTD_DEBUG_LEVEL0, "%s: read data size (%d) "
			      "is not a integer multiple "
			      "of data port size (%d)\n",
			      __FUNCTION__, end, 4);
		}
		while (j + 4 <= end) {
			*((u32 *)(&(data_poi[j])))
				= NAND_READL (this->IO_ADDR_R);
			j += 4;
		}
		if (j < end) {
			u8 t0, t1, t2, t3;

			NAND_READB_4(this->IO_ADDR_R, t0, t1, t2, t3);
			data_poi[j++] = t0;
			if (j < end) {
				data_poi[j++] = t1;
				if (j < end)
					data_poi[j++] = t2;
			}
		}
#else	/* CONFIG_SNSC_MTD_NAND_B4 */
#error CONFIG_SNSC_MTD_NAND_B2 is not supported.
#endif	/* CONFIG_SNSC_MTD_NAND_B4 */
#else	/* CONFIG_SNSC_MTD_NAND_GEOMETRY */
		while (j < end)
			data_poi[j++] = readb (this->IO_ADDR_R);
#endif	/* CONFIG_SNSC_MTD_NAND_GEOMETRY */

		/* read oobdata */
#ifdef CONFIG_SNSC_MTD_NAND_GEOMETRY
#ifdef CONFIG_SNSC_MTD_NAND_B4
		if ((mtd->oobsize % 4) != 0) {
			printk (KERN_WARNING "%s: oobsize (%d) "
			      "is not a integer multiple "
			      "of data port size (%d)\n",
			      __FUNCTION__, mtd->oobsize, 4);
			BUG();
		}
		for (j = 0; j <  mtd->oobsize; j += 4)
			*((u32 *)(&(oob_data[oob + j])))
				= NAND_READL (this->IO_ADDR_R);
#else	/* CONFIG_SNSC_MTD_NAND_B4 */
#error CONFIG_SNSC_MTD_NAND_B2 is not supported.
#endif	/* CONFIG_SNSC_MTD_NAND_B4 */
#else	/* CONFIG_SNSC_MTD_NAND_GEOMETRY */
		for (j = 0; j <  mtd->oobsize; j++) 
			oob_data[oob + j] = readb (this->IO_ADDR_R);
#endif	/* CONFIG_SNSC_MTD_NAND_GEOMETRY */
		
		if (col || (len - read) < end) { 
			for (j = col; j < end && read < len; j++)
				buf[read++] = data_poi[j];
		} else		
			read += mtd->oobblock;
		/* For subsequent reads align to page boundary. */
		col = 0;
		/* Increment page address */
		page++;
	}

	/* De-select the NAND device */
	nand_deselect ();

	/* Wake up anyone waiting on the device */
	spin_lock_bh (&CHIP_LOCK);
	CHIP_STATE = FL_READY;
	wake_up (&CHIP_WQ);
	spin_unlock_bh (&CHIP_LOCK);

	*retlen = read;
	return 0;
}

/*
 * NAND read out-of-band
 */
static int nand_read_oob (struct mtd_info *mtd, loff_t from, size_t len, size_t * retlen, u_char * buf)
{
	int col, page;
	int erase_state = 0;
	struct nand_chip *this = mtd->priv;
	int i;

	DEBUG (MTD_DEBUG_LEVEL3, "nand_read_oob: from = 0x%08x, len = %i\n", (unsigned int) from, (int) len);

	/* Shift to get page */
	page = ((int) from) >> this->page_shift;

	/* Mask to get column */
	col = from & (mtd->oobsize - 1);

	/* Initialize return length value */
	*retlen = 0;

	/* Do not allow reads past end of device */
	if ((from + len) > mtd->size) {
		DEBUG (MTD_DEBUG_LEVEL0, "nand_read_oob: Attempt read beyond end of device\n");
		*retlen = 0;
		return -EINVAL;
	}

	/* Grab the lock and see if the device is available */
	nand_get_chip (this, mtd , FL_READING, &erase_state);

	/* Select the NAND device */
	nand_select ();

	/* Send the read command */
#ifdef CONFIG_SNSC_MTD_NAND_GEOMETRY
#ifdef CONFIG_SNSC_MTD_NAND_B4
	/* issue READOOB with 4 byte aligned column */
	this->cmdfunc (mtd, NAND_CMD_READOOB, col & ~3, page);
#else	/* CONFIG_SNSC_MTD_NAND_B4 */
#error CONFIG_SNSC_MTD_NAND_B2 is not supported.
#endif	/* CONFIG_SNSC_MTD_NAND_B4 */
#else	/* CONFIG_SNSC_MTD_NAND_GEOMETRY */
	this->cmdfunc (mtd, NAND_CMD_READOOB, col, page);
#endif	/* CONFIG_SNSC_MTD_NAND_GEOMETRY */
	/* 
	 * Read the data, if we read more than one page
	 * oob data, let the device transfer the data !
	 */
#ifdef CONFIG_SNSC_MTD_NAND_GEOMETRY
#ifdef CONFIG_SNSC_MTD_NAND_B4
	if ((len % 4) != 0) {
		DEBUG(MTD_DEBUG_LEVEL0, "%s: read oob size (%d) "
		      "is not a integer multiple "
		      "of data port size (%d)\n",
		      __FUNCTION__, len, 4);
	}
{
	u_char *p;
	size_t l;

	p = buf;
	l = len;

	if ((col & 3) != 0) {
		/* column is not 4 byte aligned */
		u32 tmp;
		tmp = NAND_READL (this->IO_ADDR_R);
		while (((col & 3) != 0) && (l > 0)) {
			*p++ = *((u_char *)&tmp + (col & 3));
			col++;
			l--;
		}
		/* page overlapped ? */
		if ((col == mtd->oobsize) && (len > 0)) {
			col = 0;
			udelay (this->chip_delay);	/* XXX */
		}
	}

	/* now column is 4 byte alinged */
	while (l >= 4) {
		*(u32 *)p = NAND_READL (this->IO_ADDR_R);
		p += 4;
		col += 4;
		l -= 4;
		/* page overlapped ? */
		if ((col == mtd->oobsize) && (len > 0)) {
			col = 0;
			udelay (this->chip_delay);	/* XXX */
		}
	}

	if (l > 0) {
		/* but length left is smaller than 4 byte */
		u32 tmp;
		tmp = NAND_READL (this->IO_ADDR_R);
		while (l > 0) {
			*p++ = *((u_char *)&tmp + (col & 3));
			col++;
			l--;
		}
	}
}
#else	/* CONFIG_SNSC_MTD_NAND_B4 */
#error CONFIG_SNSC_MTD_NAND_B2 is not supported.
#endif	/* CONFIG_SNSC_MTD_NAND_B4 */
#else	/* CONFIG_SNSC_MTD_NAND_GEOMETRY */
	for (i = 0; i < len; i++) {
		buf[i] = readb (this->IO_ADDR_R);
		if ((col++ & (mtd->oobsize - 1)) == (mtd->oobsize - 1))
			udelay (this->chip_delay);
	}
#endif	/* CONFIG_SNSC_MTD_NAND_GEOMETRY */
	/* De-select the NAND device */
	nand_deselect ();

	/* Wake up anyone waiting on the device */
	spin_lock_bh (&CHIP_LOCK);
	CHIP_STATE = FL_READY;
	wake_up (&CHIP_WQ);
	spin_unlock_bh (&CHIP_LOCK);

	/* Return happy */
	*retlen = len;
	return 0;
}

#ifndef CONFIG_MTD_NAND_CXD4707_2X_USE_DMA_WRITE
/*
*	Use NAND write ECC
*/
static int nand_write (struct mtd_info *mtd, loff_t to, size_t len, size_t * retlen, const u_char * buf)
{
	return (nand_write_ecc (mtd, to, len, retlen, buf, NULL, 0));
}			   
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_USE_DMA_WRITE */
/*
 * NAND write with ECC
 */
static int nand_write_ecc (struct mtd_info *mtd, loff_t to, size_t len,
			   size_t * retlen, const u_char * buf, u_char * eccbuf, int oobsel)
{
	int i, page, col, cnt, ret = 0, oob = 0, written = 0;
	struct nand_chip *this = mtd->priv;

	DEBUG (MTD_DEBUG_LEVEL3, "nand_write_ecc: to = 0x%08x, len = %i\n", (unsigned int) to, (int) len);

	/* Do not allow write past end of device */
	if ((to + len) > mtd->size) {
		DEBUG (MTD_DEBUG_LEVEL0, "nand_write_oob: Attempt to write past end of page\n");
		return -EINVAL;
	}

	/* Shift to get page */
	page = ((int) to) >> this->page_shift;

	/* Get the starting column */
	col = to & (mtd->oobblock - 1);

	/* Grab the lock and see if the device is available */
	nand_get_chip (this, mtd, FL_WRITING, NULL);

	/* Select the NAND device */
	nand_select ();

	/* Set NAND flash to writable state */
	nand_write_unprotect();

	/* Check the WP bit */
	if (!(nand_get_status(mtd) & 0x80))
	{
		DEBUG (MTD_DEBUG_LEVEL0, "nand_write_ecc: Device is write protected!!!\n");
		ret = -EIO;
		goto out;
	}

	/* Loop until all data is written */
	while (written < len) {
		/* 
		 * Check, if we have a full page write, then we can
		 * use the given buffer, else we have to copy
		 */
		if (!col && (len - written) >= mtd->oobblock) {   
			this->data_poi = (u_char*) &buf[written];
			cnt = mtd->oobblock;
		} else {
#ifdef CONFIG_SNSC /* fix for non-page-aligned write */
			for (cnt = 0, i = col;
			     i < (len - written + col) && i < mtd->oobblock;
			     i++, cnt++) {
				this->data_buf[i] = buf[written + i - col];
			}
#else
			cnt = 0;
			for (i = col; i < len && i < mtd->oobblock; i++) {
				this->data_buf[i] = buf[written + i];
				cnt++;
			}
#endif /* CONFIG_SNSC */
			this->data_poi = this->data_buf;		
		}
		/* We use the same function for write and writev !) */

#ifdef CONFIG_SNSC /* fix for non-page-aligned write */
		if (eccbuf) {
			ret = nand_write_page (mtd, this, page, col, col + cnt,
					       &eccbuf[oob], oobsel);
			oob += mtd->oobsize;
		} else 
			ret = nand_write_page (mtd, this, page, col, col + cnt,
					       NULL, oobsel);
#else /* CONFIG_SNSC */
		if (eccbuf) {
			ret = nand_write_page (mtd, this, page, col, cnt ,&eccbuf[oob], oobsel);
			oob += mtd->oobsize;
		} else 
			ret = nand_write_page (mtd, this, page, col, cnt, NULL, oobsel);	
#endif /* CONFIG_SNSC */
		
		if (ret)
			goto out;

		/* Update written bytes count */
		written += cnt;
		/* Next write is aligned */
		col = 0;
		/* Increment page address */
		page++;
	}

out:
	/* Set NAND flash to unwritable state */
	nand_write_protect();

	/* De-select the NAND device */
	nand_deselect ();

	/* Wake up anyone waiting on the device */
	spin_lock_bh (&CHIP_LOCK);
	CHIP_STATE = FL_READY;
	wake_up (&CHIP_WQ);
	spin_unlock_bh (&CHIP_LOCK);

	*retlen = written;
	return ret;
}

/*
 * NAND write out-of-band
 */
static int nand_write_oob (struct mtd_info *mtd, loff_t to, size_t len, size_t * retlen, const u_char * buf)
{
	int i, column, page, status, ret = 0;
	struct nand_chip *this = mtd->priv;

	DEBUG (MTD_DEBUG_LEVEL3, "nand_write_oob: to = 0x%08x, len = %i\n", (unsigned int) to, (int) len);

	/* Shift to get page */
	page = ((int) to) >> this->page_shift;

	/* Mask to get column */
	column = to & (mtd->oobsize - 1);

	/* Initialize return length value */
	*retlen = 0;

	/* Do not allow write past end of page */
	if ((column + len) > mtd->oobsize) {
		DEBUG (MTD_DEBUG_LEVEL0, "nand_write_oob: Attempt to write past end of page\n");
		return -EINVAL;
	}

	/* Grab the lock and see if the device is available */
	nand_get_chip (this, mtd, FL_WRITING, NULL);

	/* Select the NAND device */
	nand_select ();

	/* Set NAND flash to writable state */
	nand_write_unprotect();

	/* Check the WP bit */
	if (!(nand_get_status(mtd) & 0x80))
	{
		DEBUG (MTD_DEBUG_LEVEL0, "nand_write_oob: Device is write protected!!!\n");
		ret = -EIO;
		goto out;
	}

	/* Write out desired data */
	this->cmdfunc (mtd, NAND_CMD_SEQIN, mtd->oobblock, page);
#ifdef CONFIG_SNSC_MTD_NAND_GEOMETRY
#ifdef CONFIG_SNSC_MTD_NAND_B4
	if ((mtd->oobsize % 4) != 0) {
		printk (KERN_WARNING "%s: oobsize (%d) "
		      "is not a integer multiple "
		      "of data port size (%d)\n",
		      __FUNCTION__, mtd->oobsize, 4);
		BUG();
	}
{
	unsigned char *oob_data;
	oob_data = &this->data_buf[mtd->oobblock];
	memset(oob_data, 0xff, mtd->oobsize);
	memcpy(oob_data + column, buf, len);
	for (i = 0; i < mtd->oobsize; i += 4) {	/* write data */ 
		NAND_WRITEL (this->IO_ADDR_W,
		             BYTES_TO_LONG(oob_data[i], oob_data[i + 1],
		                           oob_data[i + 2], oob_data[i + 3]));
	}
}
#else	/* CONFIG_SNSC_MTD_NAND_B4 */
#error CONFIG_SNSC_MTD_NAND_B2 is not supported.
#endif	/* CONFIG_SNSC_MTD_NAND_B4 */
#else	/* CONFIG_SNSC_MTD_NAND_GEOMETRY */
	/* prepad 0xff for partial programming */
	for (i = 0; i < column; i++)
		writeb (0xff, this->IO_ADDR_W);
	/* write data */
	for (i = 0; i < len; i++)
		writeb (buf[i], this->IO_ADDR_W);	
	/* postpad 0xff for partial programming */
	for (i = len + column; i < mtd->oobsize; i++)
		writeb (0xff, this->IO_ADDR_W);
#endif	/* CONFIG_SNSC_MTD_NAND_GEOMETRY */

	/* Send command to program the OOB data */
	this->cmdfunc (mtd, NAND_CMD_PAGEPROG, -1, -1);

	status = this->waitfunc (mtd, this, FL_WRITING);

	/* Set NAND flash to unwritable state */
	nand_write_protect();

	/* See if device thinks it succeeded */
	if (status & 0x01) {
		DEBUG (MTD_DEBUG_LEVEL0, "nand_write_oob: " "Failed write, page 0x%08x\n", page);
		ret = -EIO;
		goto out;
	}
	/* Return happy */
	*retlen = len;

#ifdef CONFIG_MTD_NAND_VERIFY_WRITE
	/* Send command to read back the data */
#ifdef CONFIG_SNSC_MTD_NAND_GEOMETRY
#ifdef CONFIG_SNSC_MTD_NAND_B4
	/* issue READOOB with 4 byte aligned column */
	this->cmdfunc (mtd, NAND_CMD_READOOB, column & ~3, page);
#else	/* CONFIG_SNSC_MTD_NAND_B4 */
#error CONFIG_SNSC_MTD_NAND_B2 is not supported.
#endif	/* CONFIG_SNSC_MTD_NAND_B4 */
#else	/* CONFIG_SNSC_MTD_NAND_GEOMETRY */
	this->cmdfunc (mtd, NAND_CMD_READOOB, column, page);
#endif	/* CONFIG_SNSC_MTD_NAND_GEOMETRY */

	/* Loop through and verify the data */
#ifdef CONFIG_SNSC_MTD_NAND_GEOMETRY
#ifdef CONFIG_SNSC_MTD_NAND_B4
	if ((len % 4) != 0) {
		DEBUG(MTD_DEBUG_LEVEL0, "%s: read oob size (%d) "
		      "is not a integer multiple "
		      "of data port size (%d)\n",
		      __FUNCTION__, len, 4);
	}
{
	const u_char *p;
	size_t l;
	int verify_error = 0;

	p = buf;
	l = len;

	if ((column & 3) != 0) {
		u32 tmp;
		tmp = NAND_READL (this->IO_ADDR_R);
		while ((column & 3) && (l > 0)) {
			if (*p++ != *((u_char *)&tmp + (column & 3))) {
				verify_error = 1;
				goto verify_end;
			}
			column++;
			l--;			
		}
	}
	while (l >= 4) {
		if (*(u32 *)p != NAND_READL (this->IO_ADDR_R)) {
			verify_error = 1;
			goto verify_end;
		}
		p += 4;
		column += 4;
		l -= 4;
	}
	if (l > 0) {
		u32 tmp;
		tmp = NAND_READL (this->IO_ADDR_R);
		while (l > 0) {
			if (*p++ != *((u_char *)&tmp + (column & 3))) {
				verify_error = 1;
				goto verify_end;
			}
			column++;
			l--;
		}
	}

verify_end:
	if (verify_error) {
		DEBUG (MTD_DEBUG_LEVEL0, "nand_write_oob: "
		       "Failed write verify, page 0x%08x\n", page);
		ret = -EIO;
	}
}
#else	/* CONFIG_SNSC_MTD_NAND_B4 */
#error CONFIG_SNSC_MTD_NAND_B2 is not supported.
#endif	/* CONFIG_SNSC_MTD_NAND_B4 */
#else	/* CONFIG_SNSC_MTD_NAND_GEOMETRY */
	for (i = 0; i < len; i++) {
		if (buf[i] != readb (this->IO_ADDR_R)) {
			DEBUG (MTD_DEBUG_LEVEL0, "nand_write_oob: " "Failed write verify, page 0x%08x\n", page);
			ret = -EIO;
			goto out;
		}
	}
#endif	/* CONFIG_SNSC_MTD_NAND_GEOMETRY */
#endif

out:
	/* De-select the NAND device */
	nand_deselect ();

	/* Wake up anyone waiting on the device */
	spin_lock_bh (&CHIP_LOCK);
	CHIP_STATE = FL_READY;
	wake_up (&CHIP_WQ);
	spin_unlock_bh (&CHIP_LOCK);

	return ret;
}


/*
 * NAND write with iovec
 */
static int nand_writev (struct mtd_info *mtd, const struct iovec *vecs, unsigned long count, 
		loff_t to, size_t * retlen)
{
	return (nand_writev_ecc (mtd, vecs, count, to, retlen, NULL, 0));	
}

static int nand_writev_ecc (struct mtd_info *mtd, const struct iovec *vecs, unsigned long count, 
		loff_t to, size_t * retlen, u_char *eccbuf, int oobsel)
{
	int i, page, col, cnt, len, total_len, ret = 0, written = 0;
	struct nand_chip *this = mtd->priv;

	/* Calculate total length of data */
	total_len = 0;
	for (i = 0; i < count; i++)
		total_len += (int) vecs[i].iov_len;

	DEBUG (MTD_DEBUG_LEVEL3,
	       "nand_writev: to = 0x%08x, len = %i, count = %ld\n", (unsigned int) to, (unsigned int) total_len, count);

	/* Do not allow write past end of page */
	if ((to + total_len) > mtd->size) {
		DEBUG (MTD_DEBUG_LEVEL0, "nand_writev: Attempted write past end of device\n");
		return -EINVAL;
	}

	/* Shift to get page */
	page = ((int) to) >> this->page_shift;

	/* Get the starting column */
	col = to & (mtd->oobblock - 1);

	/* Grab the lock and see if the device is available */
	nand_get_chip (this, mtd, FL_WRITING, NULL);

	/* Select the NAND device */
	nand_select ();

	/* Set NAND flash to writable state */
	nand_write_unprotect();

	/* Check the WP bit */
	if (!(nand_get_status(mtd) & 0x80))
	{
		DEBUG (MTD_DEBUG_LEVEL0, "nand_writev: Device is write protected!!!\n");
		ret = -EIO;
		goto out;
	}

	/* Loop until all iovecs' data has been written */
	cnt = col;
	len = 0;
	
	while (count) {
		/* 
		 *  Check, if we write from offset 0 and if the tuple
		 *  gives us not enough data for a full page write. Then we
		 *  can use the iov direct, else we have to copy into
		 *  data_buf.		
		 */
		if (!cnt && (vecs->iov_len - len) >= mtd->oobblock) {
			cnt = mtd->oobblock;
			this->data_poi = (u_char *) vecs->iov_base;
			this->data_poi += len;
			len += mtd->oobblock; 
			/* Check, if we have to switch to the next tuple */
			if (len >= (int) vecs->iov_len) {
				vecs++;
				len = 0;
				count--;
			}
		} else {
			/*
			 * Read data out of each tuple until we have a full page
			 * to write or we've read all the tuples.
		 	*/
			while ((cnt < mtd->oobblock) && count) {
				if (vecs->iov_base != NULL && vecs->iov_len) {
					this->data_buf[cnt++] = ((u_char *) vecs->iov_base)[len++];
				}
				/* Check, if we have to switch to the next tuple */
				if (len >= (int) vecs->iov_len) {
					vecs++;
					len = 0;
					count--;
				}
			}	
			this->data_poi = this->data_buf;	
		}
		
		/* We use the same function for write and writev !) */
		ret = nand_write_page (mtd, this, page, col, cnt, NULL, oobsel);
		if (ret)
			goto out;

		/* Update written bytes count */
		written += (cnt - col);

		/* Reset written byte counter and column */
		col = cnt = 0;

		/* Increment page address */
		page++;
	}

out:
	/* Set NAND flash to unwritable state */
	nand_write_protect();

	/* De-select the NAND device */
	nand_deselect ();

	/* Wake up anyone waiting on the device */
	spin_lock_bh (&CHIP_LOCK);
	CHIP_STATE = FL_READY;
	wake_up (&CHIP_WQ);
	spin_unlock_bh (&CHIP_LOCK);

	*retlen = written;
	return ret;
}

/*
 * NAND erase a block
 */
static int nand_erase (struct mtd_info *mtd, struct erase_info *instr)
{
	int page, len, status, pages_per_block, ret;
	struct nand_chip *this = mtd->priv;
	DECLARE_WAITQUEUE (wait, current);

	DEBUG (MTD_DEBUG_LEVEL3,
	       "nand_erase: start = 0x%08x, len = %i\n", (unsigned int) instr->addr, (unsigned int) instr->len);

	/* Start address must align on block boundary */
	if (instr->addr & (mtd->erasesize - 1)) {
		DEBUG (MTD_DEBUG_LEVEL0, "nand_erase: Unaligned address\n");
		return -EINVAL;
	}

	/* Length must align on block boundary */
	if (instr->len & (mtd->erasesize - 1)) {
		DEBUG (MTD_DEBUG_LEVEL0, "nand_erase: Length not block aligned\n");
		return -EINVAL;
	}

	/* Do not allow erase past end of device */
	if ((instr->len + instr->addr) > mtd->size) {
		DEBUG (MTD_DEBUG_LEVEL0, "nand_erase: Erase past end of device\n");
		return -EINVAL;
	}

	/* Grab the lock and see if the device is available */
	nand_get_chip (this, mtd, FL_ERASING, NULL);

	/* Shift to get first page */
	page = (int) (instr->addr >> this->page_shift);

	/* Calculate pages in each block */
	pages_per_block = mtd->erasesize / mtd->oobblock;

	/* Select the NAND device */
	nand_select ();

	/* Set NAND flash to writable state */
	nand_write_unprotect();

	/* Check the WP bit */
	if (!(nand_get_status(mtd) & 0x80))
	{
		DEBUG (MTD_DEBUG_LEVEL0, "nand_erase: Device is write protected!!!\n");
		instr->state = MTD_ERASE_FAILED;
		goto erase_exit;
	}

	/* Loop through the pages */
	len = instr->len;

	instr->state = MTD_ERASING;

	while (len) {
		/* Check if we have a bad block, we do not erase bad blocks ! */
#ifndef CONFIG_SNSC
		this->cmdfunc (mtd, NAND_CMD_READOOB,
			       this->badblock_pos, page);
		if (readb (this->IO_ADDR_R) != 0xff) {
			printk (KERN_WARNING "nand_erase: attempt to erase a bad block at page 0x%08x\n", page);
#ifndef CONFIG_SNSC_MTD_NAND_FORCE_ERASE
			instr->state = MTD_ERASE_FAILED;
			goto erase_exit;
#endif /* CONFIG_SNSC_MTD_NAND_FORCE_ERASE */
		}
#else /* CONFIG_SNSC */
		{
			/* print bad block flag for debug,
			   retry for single bit failure */
#define BADBLOCK_RETRY 2
			unsigned char badflag = 0;
			int retry = 0;

			do {
#ifdef CONFIG_SNSC_MTD_NAND_GEOMETRY
#ifdef CONFIG_SNSC_MTD_NAND_B4
				/* issue READOOB with 4 byte aligned column */
				u32 tmp;
				this->cmdfunc (mtd, NAND_CMD_READOOB,
					       this->badblock_pos & ~3, page);
				tmp =  NAND_READL(this->IO_ADDR_R);
				badflag = *((u_char *)&tmp +
					    (this->badblock_pos & 3));
#else	/* CONFIG_SNSC_MTD_NAND_B4 */
#error CONFIG_SNSC_MTD_NAND_B2 is not supported.
#endif	/* CONFIG_SNSC_MTD_NAND_B4 */
#else	/* CONFIG_SNSC_MTD_NAND_GEOMETRY */
				this->cmdfunc (mtd, NAND_CMD_READOOB,
					       this->badblock_pos, page);
				badflag = readb(this->IO_ADDR_R);
#endif	/* CONFIG_SNSC_MTD_NAND_GEOMETRY */
			} while ((num_of_bit1_uchar(badflag) < 7) &&
				 (retry++ < BADBLOCK_RETRY));

			if (num_of_bit1_uchar(badflag) < 7) {
				printk(KERN_WARNING
				       "nand_erase: attempt to erase a bad "
				       "block at page 0x%08x, read=%02x\n",
				       page, badflag);
#ifndef CONFIG_SNSC_MTD_NAND_FORCE_ERASE
				instr->state = MTD_ERASE_FAILED;
				goto erase_exit;
#endif /* CONFIG_SNSC_MTD_NAND_FORCE_ERASE */
			}
		}
#endif /* CONFIG_SNSC */

		/* Send commands to erase a page */
		this->cmdfunc (mtd, NAND_CMD_ERASE1, -1, page);
		this->cmdfunc (mtd, NAND_CMD_ERASE2, -1, -1);

		spin_unlock_bh (&CHIP_LOCK);
		status = this->waitfunc (mtd, this, FL_ERASING);

		/* Get spinlock, in case we exit */
		spin_lock_bh (&CHIP_LOCK);
#ifndef CONFIG_SNSC /* status check must be performed
                      only if we were not interrupted, so we do it later */

		/* See if block erase succeeded */
		if (status & 0x01) {
			DEBUG (MTD_DEBUG_LEVEL0, "nand_erase: " "Failed erase, page 0x%08x\n", page);
			instr->state = MTD_ERASE_FAILED;
			goto erase_exit;
		}
#endif /* CONFIG_SNSC */
		
		/* Check, if we were interupted */
		if (CHIP_STATE == FL_ERASING) {
#ifdef CONFIG_SNSC /* we were not interrupted, now do status check */
			/* See if block erase succeeded */
			if (status & 0x01) {
				DEBUG (MTD_DEBUG_LEVEL0, "nand_erase: " "Failed erase, page 0x%08x\n", page);
				instr->state = MTD_ERASE_FAILED;
				goto erase_exit;
			}
#endif
			/* Increment page address and decrement length */
			len -= mtd->erasesize;
			page += pages_per_block;
#ifdef CONFIG_SNSC /* if not interrupted, we soon continue the loop.
		      this avoid re-select the chip, and make DISABLE_LED stuff
		      work correctly */
			continue;
#endif /* CONFIG_SNSC */
		}
#ifdef CONFIG_SNSC /* some one interrupted erase, wait READY state and 
		      retry erase */
erase_retry:
		if (CHIP_STATE == FL_READY) {
			/* interrupted */
			CHIP_STATE = FL_ERASING;
			nand_select();
			continue;
		} else {
			set_current_state (TASK_UNINTERRUPTIBLE);
			add_wait_queue (&CHIP_WQ, &wait);
			spin_unlock_bh (&CHIP_LOCK);
			schedule ();
			remove_wait_queue (&CHIP_WQ, &wait);
			spin_lock_bh (&CHIP_LOCK);
			goto erase_retry;
		}
#else /* CONFIG_SNSC */
		/* Release the spin lock */
		spin_unlock_bh (&CHIP_LOCK);
erase_retry:
		spin_lock_bh (&CHIP_LOCK);
		/* Check the state and sleep if it changed */
		if (CHIP_STATE == FL_ERASING || CHIP_STATE == FL_READY) {
			/* Select the NAND device again, if we were interrupted */
			CHIP_STATE = FL_ERASING;
			nand_select ();
			continue;
		} else {
			set_current_state (TASK_UNINTERRUPTIBLE);
			add_wait_queue (&CHIP_WQ, &wait);
			spin_unlock_bh (&CHIP_LOCK);
			schedule ();
			remove_wait_queue (&CHIP_WQ, &wait);
			goto erase_retry;
		}
#endif /* CONFIG_SNSC */
	}
	instr->state = MTD_ERASE_DONE;

erase_exit:
	/* Set NAND flash to unwritable state */
	nand_write_protect();

	/* De-select the NAND device */
	nand_deselect ();
	spin_unlock_bh (&CHIP_LOCK);

	ret = instr->state == MTD_ERASE_DONE ? 0 : -EIO;;
	/* Do call back function */
	if (!ret && instr->callback)
		instr->callback (instr);

	/* The device is ready */
	spin_lock_bh (&CHIP_LOCK);
	CHIP_STATE = FL_READY;
	spin_unlock_bh (&CHIP_LOCK);

	/* Return more or less happy */
	return ret;
}

/*
 * NAND sync
 */
static void nand_sync (struct mtd_info *mtd)
{
	struct nand_chip *this = mtd->priv;
	DECLARE_WAITQUEUE (wait, current);

	DEBUG (MTD_DEBUG_LEVEL3, "nand_sync: called\n");

retry:
	/* Grab the spinlock */
	spin_lock_bh (&CHIP_LOCK);

	/* See what's going on */
	switch (CHIP_STATE) {
	case FL_READY:
	case FL_SYNCING:
		CHIP_STATE = FL_SYNCING;
		spin_unlock_bh (&CHIP_LOCK);
		break;

	default:
		/* Not an idle state */
		add_wait_queue (&CHIP_WQ, &wait);
		spin_unlock_bh (&CHIP_LOCK);
		schedule ();

		remove_wait_queue (&CHIP_WQ, &wait);
		goto retry;
	}

	/* Lock the device */
	spin_lock_bh (&CHIP_LOCK);

	/* Set the device to be ready again */
	if (CHIP_STATE == FL_SYNCING) {
		CHIP_STATE = FL_READY;
		wake_up (&CHIP_WQ);
	}

	/* Unlock the device */
	spin_unlock_bh (&CHIP_LOCK);
}

/*
 * Scan for the NAND device
 */
static int cxd4707_nand_allow_interleave = 0;

static int __init cxd4707_nand_scan(struct mtd_info *mtd)
{
	int i;
	int nand_maf_id, nand_dev_id;
#ifndef CONFIG_MTD_NAND_CXD4707_2X_PROBE_FORCEx1
	int nand_maf_id2, nand_dev_id2;
#endif /* !CONFIG_MTD_NAND_CXD4707_2X_PROBE_FORCEx1 */
	int interleaved = 0;
	struct nand_chip *this = mtd->priv;

	/* check for proper chip_delay setup, set 20us if not */
	if (!this->chip_delay)
		this->chip_delay = 20;

	/* check, if a user supplied command function given */
	if (this->cmdfunc == NULL)
#ifdef CONFIG_SNSC_MTD_NAND_GEOMETRY
	{
		printk (KERN_WARNING "%s: cmdfunc is not defined\n",
		        __FUNCTION__);
		BUG();
	}
#endif	/* CONFIG_SNSC_MTD_NAND_GEOMETRY */

	if (cxd4707_nand_allow_interleave == 0) {
#ifdef CONFIG_MTD_NAND_CXD4707_2X_PROBE_NBL
		if ((cxd4707_nand_get_config() &
		     (NAND_CONFIG_TC | NAND_CONFIG_DW)) != 0) {
			printk(KERN_INFO "cxd4707_2x-nand: probe mode is "
			       "not compatible to pin multiplex setting\n");
		}
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_PROBE_FORCEx2 */
#ifdef CONFIG_MTD_NAND_CXD4707_2X_PROBE_FORCEx2
		printk(KERN_INFO "cxd4707_2x-nand: probe mode is "
		       "not compatible to pin multiplex setting\n");
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_PROBE_FORCEx2 */
		cxd4707_nand_clr_config(NAND_CONFIG_TC | NAND_CONFIG_DW);
	}
#if defined(CONFIG_MTD_NAND_CXD4707_2X_PROBE_AUTO) || \
    defined(CONFIG_MTD_NAND_CXD4707_2X_PROBE_FORCEx2)
	if (cxd4707_nand_allow_interleave != 0) {
		cxd4707_nand_set_config(NAND_CONFIG_TC);
	}
#endif /* AUTO || FORCEx2 */

	/* Select the device */
	nand_select ();

	/* Send the command for reading device ID */
	this->cmdfunc (mtd, NAND_CMD_READID, 0x00, -1);

	/* Read manufacturer and device IDs */

#ifdef CONFIG_MTD_NAND_CXD4707_2X_PROBE_AUTO
	printk(KERN_INFO "cxd4707_2x-nand: probe mode: AUTO\n");
	if (cxd4707_nand_allow_interleave != 0) {
		NAND_READB_2(this->IO_ADDR_R, nand_maf_id, nand_maf_id2);
		NAND_READB_2(this->IO_ADDR_R, nand_dev_id, nand_dev_id2);
		if ((nand_maf_id == nand_maf_id2) &&
		    (nand_dev_id == nand_dev_id2)) {
			printk(KERN_INFO "--> detected interleaved device.\n");
			interleaved = 1;
		} else {
			printk(KERN_INFO "--> device is not interleaved.\n");
			nand_deselect ();
			cxd4707_nand_clr_config(NAND_CONFIG_TC);
			nand_select ();
			this->cmdfunc (mtd, NAND_CMD_READID, 0x00, -1);
			NAND_READB_2_BYTE (this->IO_ADDR_R,
			                   nand_maf_id, nand_dev_id);
		}
	} else {
		printk(KERN_INFO "--> device is not interleaved.\n");
		NAND_READB_2_BYTE (this->IO_ADDR_R, nand_maf_id, nand_dev_id);
	}
#endif /* CONFIG_MTD_NAND_CXD4707_2X_PROBE_AUTO */
#ifdef CONFIG_MTD_NAND_CXD4707_2X_PROBE_FORCEx1
	printk(KERN_INFO "cxd4707_2x-nand: probe mode: FORCEx1\n");
	NAND_READB_2_BYTE (this->IO_ADDR_R, nand_maf_id, nand_dev_id);
#endif /* CONFIG_MTD_NAND_CXD4707_2X_PROBE_FORCEx1 */
#ifdef CONFIG_MTD_NAND_CXD4707_2X_PROBE_FORCEx2
	printk(KERN_INFO "cxd4707_2x-nand: probe mode: FORCEx2\n");
	interleaved = 1;
	NAND_READB_2(this->IO_ADDR_R, nand_maf_id, nand_maf_id2);
	NAND_READB_2(this->IO_ADDR_R, nand_dev_id, nand_dev_id2);
	if ((nand_maf_id != nand_maf_id2) || (nand_dev_id != nand_dev_id2)) {
		printk(KERN_WARNING "cxd4707_2x-nand: failed to probe "
		       "interleaved chips!!\n");
		nand_maf_id = nand_maf_id2 = 0xffff;
		nand_dev_id = nand_dev_id2 = 0xffff;
	}
#endif /* CONFIG_MTD_NAND_CXD4707_2X_PROBE_FORCEx2 */
#ifdef CONFIG_MTD_NAND_CXD4707_2X_PROBE_NBL
	printk(KERN_INFO "cxd4707_2x-nand: probe mode: NBL\n");
	if (cxd4707_nand_get_config() & NAND_CONFIG_TC) {
		interleaved = 1;
		NAND_READB_2(this->IO_ADDR_R, nand_maf_id, nand_maf_id2);
		NAND_READB_2(this->IO_ADDR_R, nand_dev_id, nand_dev_id2);
		if ((nand_maf_id != nand_maf_id2) ||
		    (nand_dev_id != nand_dev_id2)) {
			printk(KERN_WARNING "cxd4707_2x-nand: failed to probe "
			       "interleaved chips!!\n");
		}
	} else {
		NAND_READB_2_BYTE (this->IO_ADDR_R, nand_maf_id, nand_dev_id);
	}
#endif /* CONFIG_MTD_NAND_CXD4707_2X_PROBE_NBL */

	/* De-select the device */
	nand_deselect ();

	/* Print and store flash device information */
	for (i = 0; nand_flash_ids[i].name != NULL; i++) {
		if (nand_dev_id == nand_flash_ids[i].id && !mtd->size) {
			mtd->name = nand_flash_ids[i].name;
			mtd->erasesize = nand_flash_ids[i].erasesize;
			mtd->size = (1 << nand_flash_ids[i].chipshift);
			mtd->oobblock = nand_flash_ids[i].pagesize;
			mtd->oobsize = nand_flash_ids[i].extrasize;

#ifndef CONFIG_MTD_NAND_CXD4707_2X_PROBE_NBL
			if (mtd->oobblock == 2048) {
				cxd4707_nand_set_config(NAND_CONFIG_PS);
			} else {
				cxd4707_nand_clr_config(NAND_CONFIG_PS);
			}
#endif /* !CONFIG_MTD_NAND_CXD4707_2X_PROBE_NBL */
			if (interleaved) {
				mtd->erasesize <<= 1;
				mtd->size <<= 1;
				mtd->oobblock <<= 1;
				mtd->oobsize <<= 1;
			}

			this->page_shift = ffs(mtd->oobblock) - 1;
			this->interleaved = interleaved;

			/* badblock marker position */
			this->badblock_pos = (mtd->oobblock >> 7) + 1;

			/* Try to identify manufacturer */
			for (i = 0; nand_manuf_ids[i].id != 0x0; i++) {
				if (nand_manuf_ids[i].id == nand_maf_id)
					break;
			}
			printk (KERN_INFO "NAND device: Manufacture ID:"
				" 0x%02x, Chip ID: 0x%02x (%s %s) %s\n", 
				nand_maf_id, nand_dev_id, 
				nand_manuf_ids[i].name , mtd->name,
				((interleaved) ? "(interleaved)" : ""));
			break;
		}
	}

	this->eccmode = NAND_ECC_NONE;

	/* Initialize state, waitqueue and spinlock */
	CHIP_STATE = FL_READY;
	init_waitqueue_head (&this->wq);
	spin_lock_init (&this->chip_lock);

	/* Print warning message for no device */
	if (!mtd->size) {
		printk (KERN_WARNING "No NAND device found!!!\n");
		return 1;
	}

	/* Fill in remaining MTD driver data */
	mtd->type = MTD_NANDFLASH;
	mtd->flags = MTD_CAP_NANDFLASH | MTD_ECC;
	mtd->module = THIS_MODULE;
	mtd->ecctype = MTD_ECC_SW;
	mtd->erase = nand_erase;
	mtd->point = NULL;
	mtd->unpoint = NULL;
#ifdef CONFIG_MTD_NAND_CXD4707_2X_USE_DMA_READ
	mtd->read = nand_dma_read;
	mtd->read_ecc = nand_dma_read_ecc_compat;
#else	/* CONFIG_MTD_NAND_CXD4707_2X_USE_DMA_READ */
	mtd->read = nand_read;
	mtd->read_ecc = nand_read_ecc;
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_USE_DMA_READ */
#ifdef CONFIG_MTD_NAND_CXD4707_2X_USE_DMA_WRITE
	mtd->write = nand_dma_write;
	mtd->write_ecc = nand_dma_write_ecc;
#else	/* CONFIG_MTD_NAND_CXD4707_2X_USE_DMA_WRITE */
	mtd->write = nand_write;
	mtd->write_ecc = nand_write_ecc;
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_USE_DMA_WRITE */
	mtd->read_oob = nand_read_oob;
	mtd->write_oob = nand_write_oob;
	mtd->readv = NULL;
	mtd->writev = nand_writev;
	mtd->writev_ecc = nand_writev_ecc;
	mtd->sync = nand_sync;
	mtd->lock = NULL;
	mtd->unlock = NULL;
	mtd->suspend = NULL;
	mtd->resume = NULL;

#ifdef CONFIG_SNSC_MTD_NAND_GLOBAL_STATE
	init_waitqueue_head(&CHIP_WQ);
#endif /* CONFIG_SNSC_MTD_NAND_GLOBAL_STATE */

	/* Return happy */
	return 0;
}


#ifdef CONFIG_MTD_SBL
static int ecc_set(sbl_dev_t *sbl_dev, const void *data,
		   unsigned char *ecc_info)
{
	mtd_sbl_dev_t *mtd_sbl_dev = sbl_dev->priv;
	struct mtd_info *nand_mtd = mtd_sbl_dev->nand_mtd;
	struct nand_chip *chip = nand_mtd->priv;
	int ecc_write = 1;
	int ecc_size;
	unsigned long ecc;
	unsigned long ecc_even, ecc_odd;
	int i;
#ifdef CONFIG_MTD_SBL_NOT_WRITE_ECC_IF_ALL_FF
	int ecc_all_zero = 1;
#endif /* CONFIG_MTD_SBL_NOT_WRITE_ECC_IF_ALL_FF */
	const unsigned char *p;

	ecc_size = sbl_dev->data_area_size >> 7;

	if (!chip->interleaved) {
		for (i = 0, p = data; i < ecc_size; i += 4, p += 512) {
			ecc_calc((const void *)p, 512, &ecc);
			ecc = (((ecc & 0x3f) << 18) |
			       ((ecc & 0xffffc0) >> 6));
			ecc_info[i    ] = ((ecc >>  0) & 0xff);
			ecc_info[i + 1] = ((ecc >>  8) & 0xff);
			ecc_info[i + 2] = ((ecc >> 16) & 0xff);
			ecc_info[i + 3] = 0x00;
#ifdef CONFIG_MTD_SBL_NOT_WRITE_ECC_IF_ALL_FF
			if (ecc != 0)
				ecc_all_zero = 0;
#endif /* CONFIG_MTD_SBL_NOT_WRITE_ECC_IF_ALL_FF */
		}
	} else {
		for (i = 0, p = data; i < ecc_size; i += 8, p += 1024) {
			ecc_calc_interleaved((const void *)p, 1024,
					     &ecc_even, &ecc_odd);
			ecc_even = (((ecc_even & 0x3f) << 18) |
				    ((ecc_even & 0xffffc0) >> 6));
			ecc_odd =  (((ecc_odd  & 0x3f) << 18) |
				    ((ecc_odd  & 0xffffc0) >> 6));
			ecc_info[i    ] = ((ecc_even >>  0) & 0xff);
			ecc_info[i + 2] = ((ecc_even >>  8) & 0xff);
			ecc_info[i + 4] = ((ecc_even >> 16) & 0xff);
			ecc_info[i + 6] = 0x00;
			ecc_info[i + 1] = ((ecc_odd  >>  0) & 0xff);
			ecc_info[i + 3] = ((ecc_odd  >>  8) & 0xff);
			ecc_info[i + 5] = ((ecc_odd  >> 16) & 0xff);
			ecc_info[i + 7] = 0x00;
#ifdef CONFIG_MTD_SBL_NOT_WRITE_ECC_IF_ALL_FF
			if (ecc_even != 0 || ecc_odd != 0)
				ecc_all_zero = 0;
#endif /* CONFIG_MTD_SBL_NOT_WRITE_ECC_IF_ALL_FF */
		}
	}

#ifdef CONFIG_MTD_SBL_NOT_WRITE_ECC_IF_ALL_FF
	if (ecc_all_zero) {
		int i;
		const unsigned char *data_c = data;

		for (i = 0; i < sbl_dev->data_area_size; i++) {
			if (data_c[i] != 0xff) {
				break;
			}
		}
		if (i == sbl_dev->data_area_size) {
			memset(ecc_info, 0xff, ecc_size);
			ecc_write = 0;
		}
	}
#endif /* CONFIG_MTD_SBL_NOT_WRITE_ECC_IF_ALL_FF */

	return ecc_write;
}

static int ecc_check_correct(sbl_dev_t *sbl_dev, void *data,
			     unsigned char *ecc_info)
{
	mtd_sbl_dev_t *mtd_sbl_dev = sbl_dev->priv;
	struct mtd_info *nand_mtd = mtd_sbl_dev->nand_mtd;
	struct nand_chip *chip = nand_mtd->priv;
	unsigned long ecc1, ecc2;
	unsigned long ecc1_even, ecc1_odd;
	unsigned long ecc2_even, ecc2_odd;
	int ecc_size;
	int res;
	int corrected;
	int i;
	unsigned char *p;

	corrected = 0;

	ecc_size = sbl_dev->data_area_size >> 7;

	if (!chip->interleaved) {
		for (i = 0, p = data; i < ecc_size; i += 4, p += 512) {
			ecc1 = ((ecc_info[i    ] <<  0) |
				(ecc_info[i + 1] <<  8) |
				(ecc_info[i + 2] << 16));
			if (ecc_info[i + 3] != 0x00) {
				printk(KERN_WARNING
				       "ecc_info[%d] must be zero, but"
				       " read %d!\n", i + 3, ecc_info[i + 3]);
			}
			ecc1 = (((ecc1 & 0xfc0000) >> 18) |
				((ecc1 & 0x03ffff) << 6));
			ecc_calc((const void *)p, 512, &ecc2);
			res = ecc_correct((void *)p, 512, ecc1, ecc2, NULL);
			if (res == -EECCCORRECT) {
				printk(KERN_NOTICE "mtd_sbl_dev_read(): "
				       "ECC error, retry (%c)\n",
				       'a' + (i >> 2));
				return -1;
			}
			if (res > 0)
				corrected = 1;
		}
	} else {
		for (i = 0, p = data; i < ecc_size; i += 8, p += 1024) {
			ecc1_even = ((ecc_info[i    ] <<  0) |
				     (ecc_info[i + 2] <<  8) |
				     (ecc_info[i + 4] << 16));
			ecc1_odd =  ((ecc_info[i + 1] <<  0) |
				     (ecc_info[i + 3] <<  8) |
				     (ecc_info[i + 5] << 16));
			if ((ecc_info[i + 6] != 0x00) ||
			    (ecc_info[i + 7] != 0x00)) {
				printk(KERN_WARNING
				       "ecc_info[%d,%d] must be zero, but"
				       " read %d,%d!\n", i + 6, i + 7,
				       ecc_info[i + 6], ecc_info[i + 7]);
			}
			ecc1_even = (((ecc1_even & 0xfc0000) >> 18) |
				     ((ecc1_even & 0x03ffff) << 6));
			ecc1_odd =  (((ecc1_odd  & 0xfc0000) >> 18) |
				     ((ecc1_odd  & 0x03ffff) << 6));
			ecc_calc_interleaved((const void *)p, 1024,
					     &ecc2_even, &ecc2_odd);
			res = ecc_correct_xor_interleaved((void *)p, 1024,
							  ecc1_even ^ ecc2_even,
							  ecc1_odd ^ ecc2_odd,
							  NULL, NULL);
			if (res == -EECCCORRECT) {
				printk(KERN_NOTICE "mtd_sbl_dev_read(): "
				       "ECC error, retry (%c)\n",
				       'a' + (i >> 3));
				return -1;
			}
			if (res > 0)
				corrected = 1;
		}
	}

	/* NOTE: return '1' means that error bit was found and corrected */
	if (corrected)
		return 1;

	return 0;
}

#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA
static int cxd4707_nand_ecc_check_correct(sbl_dev_t *sbl_dev, void *data,
                                          unsigned char *ecc_info)
{
	mtd_sbl_dev_t *mtd_sbl_dev = sbl_dev->priv;
	struct mtd_info *nand_mtd = mtd_sbl_dev->nand_mtd;
	struct nand_chip *chip = nand_mtd->priv;
	unsigned long ecc1;
	unsigned long ecc1_even, ecc1_odd;
	int ecc_size;
	int res;
	int corrected;
	int i;
	unsigned char *p;

	corrected = 0;

	ecc_size = sbl_dev->data_area_size >> 7;

	if (!chip->interleaved) {
		for (i = 0, p = data; i < ecc_size; i += 4, p += 512) {
			ecc1 = ((ecc_info[i    ] <<  0) |
				(ecc_info[i + 1] <<  8) |
				(ecc_info[i + 2] << 16));
			if (ecc_info[i + 3] != 0x00) {
				printk(KERN_WARNING
				       "ecc_info[%d] must be zero, but"
				       " read %d!\n", i + 3, ecc_info[i + 3]);
			}
			ecc1 = (((ecc1 & 0xfc0000) >> 18) |
				((ecc1 & 0x03ffff) << 6));
			res = ecc_correct_xor((void *)p, 512, ecc1, NULL);
			if (res == -EECCCORRECT) {
				printk(KERN_NOTICE "mtd_sbl_dev_read(): "
				       "ECC error, retry (%c)\n",
				       'a' + (i >> 2));
				return -1;
			}
			if (res > 0)
				corrected = 1;
		}
	} else {
		for (i = 0, p = data; i < ecc_size; i += 8, p += 1024) {
			ecc1_even = ((ecc_info[i    ] <<  0) |
				     (ecc_info[i + 1] <<  8) |
				     (ecc_info[i + 2] << 16));
			ecc1_odd =  ((ecc_info[i + 4] <<  0) |
				     (ecc_info[i + 5] <<  8) |
				     (ecc_info[i + 6] << 16));
			if ((ecc_info[i + 3] != 0x00) ||
			    (ecc_info[i + 7] != 0x00)) {
				printk(KERN_WARNING
				       "ecc_info[%d,%d] must be zero, but"
				       " read %d,%d!\n", i + 3, i + 7,
				       ecc_info[i + 3], ecc_info[i + 7]);
			}
			ecc1_even = (((ecc1_even & 0xfc0000) >> 18) |
				     ((ecc1_even & 0x03ffff) << 6));
			ecc1_odd =  (((ecc1_odd  & 0xfc0000) >> 18) |
				     ((ecc1_odd  & 0x03ffff) << 6));
			res = ecc_correct_xor_interleaved((void *)p, 1024,
							  ecc1_even,
							  ecc1_odd,
							  NULL, NULL);
			if (res == -EECCCORRECT) {
				printk(KERN_NOTICE "mtd_sbl_dev_read(): "
				       "ECC error, retry (%c)\n",
				       'a' + (i >> 3));
				return -1;
			}
			if (res > 0)
				corrected = 1;
		}
	}

	/* NOTE: return '1' means that error bit was found and corrected */
	if (corrected)
		return 1;

	return 0;
}
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA */
 
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA
#define SBL_M_MASK \
	(SBL_M_DATA | SBL_M_EXTRA | SBL_M_VERIFY | \
	 SBL_M_ECC | SBL_M_SBLINFO | SBL_M_MOVEBACK)
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA */

static int
mtd_sbl_dev_program_cxd4707(sbl_dev_t *dev, unsigned long block, int page,
			    int mode, const void *data, const void *extra)
{
	int res;
	mtd_sbl_dev_t *mtd_sbl_dev = dev->priv;
	sbl_dev_t *sbl_dev = mtd_sbl_dev->sbl_dev;
	struct mtd_info *nand_mtd = mtd_sbl_dev->nand_mtd;
	unsigned char ecc_info[128];
	int ecc_info_pos;
	size_t ecc_info_size;
	loff_t addr;
	size_t retlen;
	int ecc_write = 0;
	int ecc_written_flag_pos;

	addr = (block * sbl_dev->pages_per_block + page) *
	    sbl_dev->data_area_size;

	ecc_written_flag_pos = (sbl_dev->data_area_size >> 7);

#ifdef CONFIG_MTD_NAND_CXD4707_2X_USE_DMA_WRITE
	if (((mode & (SBL_M_MASK & ~(SBL_M_ECC | SBL_M_SBLINFO))) ==
	     SBL_M_DATA) &&
	    DMA_ALIGNED(data)) {
		res = nand_mtd->read_oob(nand_mtd, addr,
					 sbl_dev->extra_area_size,
					 &retlen, ecc_info);
		if (res < 0) {
			return -ESBLIO;
		}
#ifdef CONFIG_MTD_SBL_CHECK_PAGE_PROGRAMMED
		if (num_of_bit1_uchar(ecc_info[ecc_written_flag_pos]) < 7) {
			printk(KERN_WARNING "mtd_sbl_dev_program(): "
			       "programming a previously programmed page "
			       "(%ld/%d)\n", block, page);
			return -ESBLIO;
		}
#endif /* CONFIG_MTD_SBL_CHECK_PAGE_PROGRAMMED */

		if (mode & SBL_M_SBLINFO) {
			memcpy(ecc_info + sbl_dev->extra_area_size - 8,
			       extra, 8);
		}

		res = nand_dma_write_ecc(nand_mtd, addr,
		                         sbl_dev->data_area_size, &retlen,
		                         data, ecc_info, 1);
		if (res < 0) {
			printk(KERN_NOTICE "mtd_sbl_dev_program(): "
			       "write data error at %ld/%d\n", block, page);
			return -ESBLPROGRAM;
		}

		ecc_info[ecc_written_flag_pos] = 0;
		res = nand_mtd->write_oob(nand_mtd,
					  addr + ecc_written_flag_pos,
					  1, &retlen,
					  &(ecc_info[ecc_written_flag_pos]));
		if (res < 0) {
			printk(KERN_NOTICE "mtd_sbl_dev_program(): "
			       "write ECC written flag error at "
			       "%ld/%d\n", block, page);
			return -ESBLPROGRAM;
		}

		return 0;
	}
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_USE_DMA_WRITE */
	if (mode & SBL_M_DATA) {
#ifdef CONFIG_MTD_SBL_CHECK_PAGE_PROGRAMMED
		res = nand_mtd->read_oob(nand_mtd, addr,
					 sbl_dev->extra_area_size,
					 &retlen, ecc_info);
		if (res < 0) {
			return -ESBLIO;
		}
		if (num_of_bit1_uchar(ecc_info[ecc_written_flag_pos]) < 7) {
			printk(KERN_WARNING "mtd_sbl_dev_program(): "
			       "programming a previously programmed page "
			       "(%ld/%d)\n", block, page);
			return -ESBLIO;
		}
#endif /* CONFIG_MTD_SBL_CHECK_PAGE_PROGRAMMED */

		res = nand_mtd->write(nand_mtd, addr, sbl_dev->data_area_size,
				      &retlen, data);
		if (res < 0) {
			printk(KERN_NOTICE "mtd_sbl_dev_program(): "
			       "verify data error at %ld/%d\n", block, page);
			return -ESBLPROGRAM;
		}
	}
	if (mode & SBL_M_EXTRA) {
		res = nand_mtd->write_oob(nand_mtd, addr,
					  sbl_dev->extra_area_size,
					  &retlen, extra);
		if (res < 0) {
			printk(KERN_NOTICE "mtd_sbl_dev_program(): "
			       "verify extra error at %ld/%d\n", block, page);
			return -ESBLPROGRAM;
		}
	}
	if ((mode & SBL_M_SBLINFO) || (mode & SBL_M_ECC)) {
		memset(ecc_info, 0xff, sbl_dev->extra_area_size);
		if (mode & SBL_M_ECC) {
			ecc_write = ecc_set(sbl_dev, data, ecc_info);
		}
		if (mode & SBL_M_SBLINFO) {
			memcpy(ecc_info + sbl_dev->extra_area_size - 8,
			       extra, 8);
		}

		if (ecc_write || (mode & SBL_M_SBLINFO)) {
			if (ecc_write) {
				ecc_info_pos = 0;
				if (mode & SBL_M_SBLINFO) {
					ecc_info_size =
						sbl_dev->extra_area_size;
				} else {
					ecc_info_size =
						sbl_dev->extra_area_size / 4;
				}
			} else {
				ecc_info_pos = sbl_dev->extra_area_size - 8;
				ecc_info_size = 8;
			}

			res = nand_mtd->write_oob(nand_mtd,
						  addr + ecc_info_pos,
						  ecc_info_size,
						  &retlen,
						  &(ecc_info[ecc_info_pos]));
			if (res < 0) {
				printk(KERN_NOTICE "mtd_sbl_dev_program(): "
				       "verify ECC/sbl_info error at %ld/%d\n",
				       block, page);
				return -ESBLPROGRAM;
			}
			if (ecc_write) {
				ecc_info[ecc_written_flag_pos] = 0;
				res = nand_mtd->write_oob(nand_mtd,
							  addr +
							  ecc_written_flag_pos,
							  1, &retlen,
							  &(ecc_info[ecc_written_flag_pos]));
				if (res < 0) {
					printk(KERN_NOTICE
					       "mtd_sbl_dev_program(): "
					       "verify ECC written flag "
					       "error at %ld/%d\n",
					       block, page);
					return -ESBLPROGRAM;
				}
			}
		}
	}

	return 0;
}

static int
mtd_sbl_dev_read_cxd4707(sbl_dev_t *dev, unsigned long block, int page,
			 int mode, void *data, void *extra)
{
	mtd_sbl_dev_t *mtd_sbl_dev = dev->priv;
	sbl_dev_t *sbl_dev = mtd_sbl_dev->sbl_dev;
	struct mtd_info *nand_mtd = mtd_sbl_dev->nand_mtd;
	unsigned char ecc_info[128];
	int nretry;
	loff_t addr;
	size_t retlen;
	int res;
	int ecc_written_flag_pos;

	res = 0;

	addr = (block * sbl_dev->pages_per_block + page) *
	    sbl_dev->data_area_size;

	ecc_written_flag_pos = (sbl_dev->data_area_size >> 7);

	nretry = MTD_SBL_NRETRY;
retry:
	if (nretry < 0) {
		printk(KERN_NOTICE "mtd_sbl_dev_read(): "
		       "ECC error at %ld/%d\n", block, page);
		return -ESBLECC;
	}
	nretry--;

#ifdef CONFIG_MTD_NAND_CXD4707_2X_USE_DMA_READ
	if (((mode & SBL_M_MASK) == (SBL_M_DATA | SBL_M_ECC)) &&
	    CACHE_ALIGNED(data) &&
	    DMA_ALIGNED(data)) {
		nand_dma_read_ecc(nand_mtd, addr, sbl_dev->data_area_size,
		                  &retlen, data, ecc_info, 1);
		if (num_of_bit1_uchar(ecc_info[ecc_written_flag_pos]) < 7) {
			res = cxd4707_nand_ecc_check_correct(sbl_dev, data,
							     ecc_info);
			if (res < 0)
				goto retry;
		}

		return res;
	}
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_USE_DMA_READ */
	if (mode & SBL_M_DATA) {
		nand_mtd->read(nand_mtd, addr, sbl_dev->data_area_size,
			       &retlen, data);
	}
	if (mode & SBL_M_EXTRA) {
		nand_mtd->read_oob(nand_mtd, addr, sbl_dev->extra_area_size,
				   &retlen, extra);
	}
	if ((mode & SBL_M_ECC) || (mode & SBL_M_SBLINFO)) {
		nand_mtd->read_oob(nand_mtd, addr, sbl_dev->extra_area_size,
				   &retlen, ecc_info);
		if (mode & SBL_M_SBLINFO) {
			memcpy(extra,
			       ecc_info + sbl_dev->extra_area_size - 8, 8);
		}
		if ((mode & SBL_M_ECC) &&
		    num_of_bit1_uchar(ecc_info[ecc_written_flag_pos]) < 7) {
			res = ecc_check_correct(sbl_dev, data, ecc_info);
			if (res < 0)
				goto retry;
		}
	}

	return res;
}
#endif /* CONFIG_MTD_SBL */



/* Chip name */
static const char chip_name[] = "cxd4707_nand";

/* For wait function */
static spinlock_t cxd4707_nand_lock = SPIN_LOCK_UNLOCKED;
static wait_queue_head_t cxd4707_nand_wait_queue;
static volatile int cxd4707_nand_need_interrupt;

#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA
static spinlock_t cxd4707_dmac_lock = SPIN_LOCK_UNLOCKED;
static wait_queue_head_t cxd4707_dmac_wait_queue;
static volatile int cxd4707_dmac_need_interrupt;
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA */

/*
 * MTD structure for CXD4707 NAND flash interface
 */
static struct mtd_info *cxd4707_nand_mtd[CXD4707_NAND_CHIPS];
#ifdef CONFIG_MTD_SBL
static struct mtd_info *cxd4707_nand_mtdsbl[CXD4707_NAND_CHIPS];
#endif	/* CONFIG_MTD_SBL */


/* Initialize routine */
static int __init cxd4707_nand_init(void);
static int __init cxd4707_nand_init_interface(void);
static int __init cxd4707_nand_init_chip(int chip);
/* Clean up routine */
static void __exit cxd4707_nand_cleanup(void);
static void __exit cxd4707_nand_cleanup_interface(void);
static void __exit cxd4707_nand_cleanup_chip(int chip);


#ifdef CONFIG_MTD_SBL

static char *sbl_device_name[] = {
#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
	"sdm device NAND/SBL 0",
#if CXD4707_NAND_CHIPS > 1
	"sdm device NAND/SBL 1",
#endif
#else	/* CONFIG_SNSC_MTD_SDM_PARTITION */
	"NAND/SBL device 0",
#if CXD4707_NAND_CHIPS > 1
	"NAND/SBL device 1",
#endif
#endif /* CONFIG_SNSC_MTD_SDM_PARRITION */
};

static char *mtd_device_name[] = {
	"NAND Flash raw device 0",
#if CXD4707_NAND_CHIPS > 1
	"NAND Flash raw device 1",
#endif
};

#else	/* CONFIG_MTD_SBL */

#ifdef CONFIG_SNSC_MTD_SDM_PARTITION

static char *mtd_device_name[] = {
	"sdm device NAND 0",
#if CXD4707_NAND_CHIPS > 1
	"sdm device NAND 1",
#endif
};

#endif	/* CONFIG_SNSC_MTD_SDM_PARTITION */

#endif	/* CONFIG_MTD_SBL */


/*
 * CXD4707 NAND flash interface control routine
 */

/* Control CS line */
static void cxd4707_nand_hwcontrol0(int cmd)
{
	if (cmd == NAND_CTL_SETNCE) {
#if CXD4707_NAND_CHIPS > 1
#ifdef CONFIG_FRX
		int ret;

		ret = ppc4xx_gpio_set_data_bit(FRX_GPIO_NAND_XCS_PORT,
		                               FRX_GPIO_NAND_XCS_BIT,
		                               FRX_GPIO_NAND_XCS_CS0);
		if (ret != 0) {
			printk("%s: Unable to set GPIO port %d bit %d to %d\n",
			       chip_name,
			       FRX_GPIO_NAND_XCS_PORT,
			       FRX_GPIO_NAND_XCS_BIT,
			       FRX_GPIO_NAND_XCS_CS0);
		}
#endif	/* CONFIG_FRX */
#endif	/* CXD4707_NAND_CHIPS > 1 */
	} else if (cmd == NAND_CTL_CLRNCE) {
		REG_WRITEL(NAND_PIO_OPEND_PIOE, cxd4707_nand + NAND_PIO_OPEND);
	} else {
		printk("%s: hwcontrol: Unsupported cmd: %d\n", chip_name, cmd);
	}
}

#if CXD4707_NAND_CHIPS > 1
static void cxd4707_nand_hwcontrol1(int cmd)
{
	if (cmd == NAND_CTL_SETNCE) {
#ifdef CONFIG_FRX
		int ret;

		ret = ppc4xx_gpio_set_data_bit(FRX_GPIO_NAND_XCS_PORT,
		                               FRX_GPIO_NAND_XCS_BIT,
		                               FRX_GPIO_NAND_XCS_CS1);
		if (ret != 0) {
			printk("%s: Unable to set GPIO port %d bit %d to %d\n",
			       chip_name,
			       FRX_GPIO_NAND_XCS_PORT,
			       FRX_GPIO_NAND_XCS_BIT,
			       FRX_GPIO_NAND_XCS_CS1);
		}
#endif	/* CONFIG_FRX */
	} else if (cmd == NAND_CTL_CLRNCE) {
		REG_WRITEL(NAND_PIO_OPEND_PIOE, cxd4707_nand + NAND_PIO_OPEND);
	} else {
		printk("%s: hwcontrol: Unsupported cmd: %d\n", chip_name, cmd);
	}
}
#endif	/* CXD4707_NAND_CHIPS >= 2 */

static void (*cxd4707_nand_hwcontrols[CXD4707_NAND_CHIPS])(int) = {
	cxd4707_nand_hwcontrol0,
#if CXD4707_NAND_CHIPS > 1
 	cxd4707_nand_hwcontrol1,
#endif	/* CXD4707_NAND_CHIPS > 1 */
};


/* Check busy status */
static int cxd4707_nand_dev_ready(void)
{
	return((REG_READL(cxd4707_nand + NAND_STATUS) & NAND_STATUS_RDY) != 0);
}

#define is_large_page_nand(x) \
(((x)->oobblock == 2048) || ((x)->oobblock == 4096))

/* Send command and address to NAND flash device */
static void cxd4707_nand_cmdfunc(struct mtd_info *mtd, unsigned command,
                                   int column, int page_addr)
{
	struct nand_chip *this = mtd->priv;
	unsigned long NAND_IO_ADDR =
		(unsigned long)(cxd4707_nand + NAND_PIO_CMD);

#ifdef CONFIG_SNSC
	/* Don't issue command, while NAND chip is busy */
	/* Except RESET and READ STATUS command */
	if ((command != NAND_CMD_RESET) &&
	    (command != NAND_CMD_STATUS)) {
		while (this->dev_ready() == 0)	/* Wait until ready */
			;
	}
#endif	/* CONFIG_SNSC */

#ifdef CONFIG_SNSC
	/* support random read for column >= 256 */
	if ((command == NAND_CMD_READ0) &&
	    (mtd->oobblock == 512) &&
	    (column >= 256)) {
		command = NAND_CMD_READ1;
		column -= 256;
	}
#endif /* CONFIG_SNSC */

	/*
	 * Write out the command to the device.
	 */
	if ((command == NAND_CMD_SEQIN) && !is_large_page_nand(mtd)) {
		u8	seek_cmd;
		if (column >= mtd->oobblock) {
			column -= mtd->oobblock;
			seek_cmd = NAND_CMD_READOOB;
		} else if (mtd->oobblock == 512 && column >= 256) {
			column -= 256;
			seek_cmd = NAND_CMD_READ1;
		} else {
			seek_cmd = NAND_CMD_READ0;
		}
		REG_WRITEL(seek_cmd, NAND_IO_ADDR);
	}

	if ((command == NAND_CMD_PAGEPROG) ||
	    (command == NAND_CMD_ERASE2)) {
		/* Start interrupt handler */
		spin_lock(&cxd4707_nand_lock);
		cxd4707_nand_need_interrupt = 1;
		ppc4xx_pic_clear_status(NAND0_INT);
		enable_irq(NAND0_INT);
		spin_unlock(&cxd4707_nand_lock);
	}

	/* fixup for large page nand device */
	if ((command == NAND_CMD_READOOB) &&
	    is_large_page_nand(mtd)) {
		command = NAND_CMD_READ0;
		column += mtd->oobblock;
	}

	REG_WRITEL(command, NAND_IO_ADDR);

	if ((column != -1) || (page_addr != -1)) {
		/*
		 * Write out the address to the device.
		 */
		if (column == -1) {
			column = 0;
		} else if (page_addr == -1) {
			page_addr = 0;
		}
		REG_WRITEL((page_addr << (this->page_shift + 1)) | column,
		       (unsigned long)(cxd4707_nand + NAND_PIO_ADDR));
	}

	if (is_large_page_nand(mtd)) {
		if ((command == NAND_CMD_READ0) ||
		    (command == NAND_CMD_READOOB)) {
			REG_WRITEL(NAND_CMD_READSTART, NAND_IO_ADDR);
		}
	}

	/* 
	 * program and erase have their own busy handlers 
	 * status and sequential in needs no delay
	*/
	switch (command) {
	case NAND_CMD_PAGEPROG:
	case NAND_CMD_ERASE1:
	case NAND_CMD_ERASE2:
	case NAND_CMD_SEQIN:
	case NAND_CMD_STATUS:
		break;
	default:
		/* wait until command is processed */
		while (this->dev_ready() == 0) {
			;
		}
	}
	return;
}


/*
 * Interrupt hander.
 */
static void cxd4707_nand_intr(int i, void *v, struct pt_regs *p)
{
	/* Stop interrupt handler */
	disable_irq(NAND0_INT);

	cxd4707_nand_need_interrupt = 0;

	/* Wake up cxd4707_nand_wait() */
	wake_up(&cxd4707_nand_wait_queue);
}

/*
 * Wait for command done.
 */
static int cxd4707_nand_wait(struct mtd_info *mtd,
                             struct nand_chip *this, int state)
{
	unsigned long t1, t;
	wait_queue_t wait;
	int status;

	t1 = jiffies + 1;
	if (state == FL_ERASING) {
		t1 += (((10 * HZ) + 1000 - 1) / 1000);	/* Max. 10ms */
	} else if (state == FL_WRITING) {
		t1 +=  (((1 * HZ) + 1000 - 1) / 1000);	/* Max.  1ms */
	} else {
		/* Unknown state */
		udelay(1);				/* XXX */
		return 0;
	}

	spin_lock(&cxd4707_nand_lock);

	if (cxd4707_nand_need_interrupt == 0) {
		goto wait_exit;	/* Already interrupted */
	}

	/* wake_event() with timeout */
	init_waitqueue_entry(&wait, current);
	add_wait_queue(&cxd4707_nand_wait_queue, &wait);
	while (1) {
		set_current_state(TASK_INTERRUPTIBLE);
		if (cxd4707_nand_need_interrupt == 0) {
			break;	/* Already interrupted */
		}
		t = jiffies;
		if (t >= t1) {	/* Time out */
			/* Stop interrupt handler */
			disable_irq(NAND0_INT);

			cxd4707_nand_need_interrupt = 0;

			printk("%s: lost interrupt: %d\n", chip_name, state);
			break;
		}
		schedule_timeout(t1 - t);
	}
	current->state = TASK_RUNNING;
	remove_wait_queue(&cxd4707_nand_wait_queue, &wait);

wait_exit:
	status = nand_get_status(mtd);

	spin_unlock(&cxd4707_nand_lock);

	return status;
}


#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA
/*
 * DMA transfer support routine
 */
static int nand_dma_ch, nand_dma_irq;
static volatile unsigned int nand_dma_status;

/*
 * Read error page register for this "page" address
 *   Return value
 *     0: OK
 *     1: NG
 */
static unsigned int
cxd4707_nand_error_page(struct mtd_info *mtd, unsigned int page)
{
	struct nand_chip *this;
	int n, i;
	u32 errp;

	this = mtd->priv;

	if (this->interleaved) {
		if (mtd->oobblock == 4096) {	/* 2048 byte/page x 2 */
			n = page & (64 - 1);	/* 64 page/block */
			if (n >= 32) {
				i = 1;		/* Use NAND_PAGE_ERRP_*(1) */
				n -= 32;
			} else {
				i = 0;		/* Use NAND_PAGE_ERRP_*(0) */
			}
		} else {			/*  512 byte/page x 2 */
			n = page & (32 - 1);	/* 32 page/block */
			i = 0;			/* Use NAND_PAGE_ERRP_*(0) */
		}
		errp = REG_READL(cxd4707_nand + NAND_PAGE_ERRP_EVEN(i)) |
		       REG_READL(cxd4707_nand + NAND_PAGE_ERRP_ODD(i));
	} else {
		if (mtd->oobblock == 2048) {	/* 2048 byte/page x 1 */
			n = page & (64 - 1);	/* 64 page/block */
			if (n >= 32) {
				i = 1;		/* Use NAND_PAGE_ERRP(1) */
				n -= 32;
			} else {
				i = 0;		/* Use NAND_PAGE_ERRP(0) */
			}
		} else {			/*  512 byte/page x 1 */
			n = page & (32 - 1);	/* 32 page/block */
			i = 0;			/* Use NAND_PAGE_ERRP(0) */
		}
		errp = REG_READL(cxd4707_nand + NAND_PAGE_ERRP(i));
	}
	return((errp >> n) & 1);
}

/*
 * Check error page register for this "page" address,
 * and if error exists, return ECC status register.
 *   Return value
 *     0:     OK
 *     Other: ECC status register
 */
static unsigned int
cxd4707_nand_ecc_status(struct mtd_info *mtd, unsigned int page)
{
	u32 errp, status;
	int n;

	errp = cxd4707_nand_error_page(mtd, page);
	if (errp == 0)
		return(0);

	if ((mtd->oobblock == 2048) || (mtd->oobblock == 4096)) {
		/* 2048 byte/page x 1 or 2 */
		n = page & (64 - 1);	/* 64 page/block */
	} else {
		/*  512 byte/page x 1 or 2 */
		n = page & (32 - 1);	/* 32 page/block */
	}
	status = REG_READL(cxd4707_nand + NAND_PAGE_ECC_STATUS(n));
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
	printk("page: %u: ERRP: %08x, ECC_STATUS(%d): %08x\n",
	       page, errp, n, status);
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG */

	return(status);
}

static void
cxd4707_dmac_intr(int i, void *v, struct pt_regs *p)
{
	/* Save and clear DMAC status */
	nand_dma_status = ppc4xx_get_dma_status(NAND_DMA_CONTROLLER);
	ppc4xx_clear_dma_status(nand_dma_ch);

	cxd4707_dmac_need_interrupt = 0;

	/* Wake up cxd4707_dmac_wait() */
	wake_up(&cxd4707_dmac_wait_queue);
}

static int cxd4707_dmac_wait(void) {
	int ret;
	unsigned long t1, t;
	wait_queue_t wait;

	ret = 0;
	t1 = jiffies + (((1 * HZ) + 1000 - 1) / 1000) + 1;	/* Max.  1ms */

	if (cxd4707_dmac_need_interrupt == 0) {
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
		if (nand_dma_status != DMA_CS(nand_dma_ch))
			printk("nand_dma_status: 1: %08x\n", nand_dma_status);
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG */
		return(0);	/* Already interrupted */
	}

	spin_lock(&cxd4707_dmac_lock);

	/* wake_event() with timeout */
	init_waitqueue_entry(&wait, current);
	add_wait_queue(&cxd4707_dmac_wait_queue, &wait);
	while (1) {
		set_current_state(TASK_INTERRUPTIBLE);
		if (cxd4707_dmac_need_interrupt == 0) {
			break;	/* Already interrupted */
		}
		t = jiffies;
		if (t >= t1) {	/* Time out */
			/* Save and clear DMAC status */
			nand_dma_status =
				ppc4xx_get_dma_status(NAND_DMA_CONTROLLER);
			ppc4xx_clear_dma_status(nand_dma_ch);

			cxd4707_dmac_need_interrupt = 0;

			printk("%s: lost interrupt from DMA controller\n",
			       chip_name);

			ret = -1;
			break;
		}
		schedule_timeout(t1 - t);
	}
	current->state = TASK_RUNNING;
	remove_wait_queue(&cxd4707_dmac_wait_queue, &wait);

#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
	if (nand_dma_status != DMA_CS(nand_dma_ch))
		printk("nand_dma_status: 2: %08x\n", nand_dma_status);
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG */
	spin_unlock(&cxd4707_dmac_lock);

	return(ret);
}

static ppc_dma_ch_t nand_dma_normal_setting = {
	.polarity      = 0,
	.buffer_enable = 1,		/* Enable DMA 128byte buffer */
	.tce_enable    = 1,
	.etd_output    = 1,
	.pce           = 0,		/* Disable Parity checking */ /* XXX */
	.pl            = NAND_DMA_LOCATION,
	.pwidth        = PW_32,		/* 32bit width */
	.dai           = 1,
	.sai           = 1,
	.psc           = 0,
	.pwc           = 0,
	.phc           = 0,
	.cp            = NAND_DMA_PRIORITY,			/* XXX */
	.pf            = PREFETCH_2,	/* Prefetch 2 quadwords */
	.sl            = 0,
	.sid           = NAND_DMA_SID,
	.bten          = 1,		/* Enable bursting */
	.bsiz          = BSIZ_8,	/* Burst of 8 */
};

static ppc_dma_ch_t nand_dma_dummy_read_setting = {
	.polarity      = 0,
	.buffer_enable = 0,		/* Disable DMA 128byte buffer */
	.tce_enable    = 1,
	.etd_output    = 1,
	.pce           = 0,		/* Disable Parity checking */ /* XXX */
	.pl            = NAND_DMA_LOCATION,
	.pwidth        = PW_32,		/* 32bit width */
	.dai           = 0,					/* XXX */
	.sai           = 0,					/* XXX */
	.psc           = 0,
	.pwc           = 0,
	.phc           = 0,
	.cp            = NAND_DMA_PRIORITY,			/* XXX */
	.pf            = PREFETCH_2,	/* Prefetch 2 quadwords */
	.sl            = 0,
	.sid           = NAND_DMA_SID,
	.bten          = 0,		/* Disable bursting */
	.bsiz          = BSIZ_8,	/* Burst of 8 */
};

static u32 __attribute((aligned(16))) dummy_read_buffer;

static int
nand_dma_dummy_read(struct mtd_info *mtd)
{
	/* Setup DMA controller */
	if (ppc4xx_init_dma_channel(nand_dma_ch,
	                            &nand_dma_dummy_read_setting) !=
	    DMA_STATUS_GOOD) {
		printk("%s: Unable to setup DMA channel\n", chip_name);
		return -EIO;
	}

	if (ppc4xx_set_dma_mode(nand_dma_ch, DMA_MODE_MM_DEVATSRC) !=
	    DMA_STATUS_GOOD) {
		printk("%s: Unable to set DMA mode\n", chip_name);
		return -EIO;
	}

	if (ppc4xx_enable_dma_interrupt(nand_dma_ch) != DMA_STATUS_GOOD) {
		printk("%s: Unable to use interrupt from DMAC\n", chip_name);
		return -EIO;
	}


	ppc4xx_set_dma_addr2(nand_dma_ch,
	                     NAND0_REG_BASE + NAND_PAGE_EXTRA(7),
	                     virt_to_phys(&dummy_read_buffer));

	ppc4xx_set_dma_count(nand_dma_ch, 4);


	/* semaphore */
	cxd4707_dmac_need_interrupt = 1;

	/* Start DMA controller */
	ppc4xx_enable_dma(nand_dma_ch);
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
#ifdef CONFIG_CWING
	printk("DMAC: %08x %08x %08x %08x %08x\n",
	       mfdcr(DCRN_DMACR0), mfdcr(DCRN_DMACT0), mfdcr(DCRN_DMASA0),
	       mfdcr(DCRN_DMADA0), mfdcr(DCRN_DMA0_SR));
#else	/* CONFIG_CWING */
	printk("DMAC: %08x %08x %08x %08x %08x\n",
	       mfdcr(DCRN_DMACR4), mfdcr(DCRN_DMACT4), mfdcr(DCRN_DMASA4),
	       mfdcr(DCRN_DMADA4), mfdcr(DCRN_DMA1_SR));
#endif	/* CONFIG_CWING */
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG */

	/* Wait interrupt */
	if (cxd4707_dmac_wait() != 0) {
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
		ppc4xx_show_status(NAND_DMA_CONTROLLER);
		ppc4xx_show_dmach(nand_dma_ch);
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG */
	}


	/* Stop DMA controller */
	ppc4xx_disable_dma(nand_dma_ch);

	/* Disable interrupt */
	if (ppc4xx_disable_dma_interrupt(nand_dma_ch) != DMA_STATUS_GOOD) {
		printk("%s: Unable to use interrupt from DMAC\n", chip_name);
		return -EIO;
	}

	return(0);
}

#ifdef CONFIG_MTD_NAND_CXD4707_2X_ECC_TEST
static void
nand_add_bit_error(u_char p[], unsigned int len)
{
	unsigned int i, j;
	u_char c, n;

	/* Locate bit error byte */
	do {
		i = get_tbl() % len;
	} while (p[i] == 0);
	c = p[i];

	/* Locate bit error bit */
	do {
		j = get_tbl() % 8;
		n = 1 << j;
	} while ((c & n) == 0);

	/* Modify bit */
	c &= ~n;
	printk("  %03x: %02x -> %02x\n", i, p[i], c);
	p[i] = c;
}

static int
nand_force_bit_error(struct mtd_info *mtd, unsigned int page)
{
	static int call_count = 0;
	struct nand_chip *this;
	u_char *data_buf, *oob_buf;
	int ret, all_zero, i;

	/* Check call count */
	if (++call_count < CONFIG_MTD_NAND_CXD4707_2X_ECC_ERROR_RATE) {
		return(0);	/* No bit error */
	}


	/* Read page */
	this = mtd->priv;

	data_buf = this->data_buf;
	oob_buf  = data_buf + mtd->oobblock;

	ret = nand_dma_read_page(mtd, page, data_buf, oob_buf);
	if (ret != 0) {
		DEBUG(MTD_DEBUG_LEVEL0,
		      "%s: Failed read for modify/write, page 0x%08x ",
		      __FUNCTION__, page);
		return -EIO;
	}

	/* Is this page all zero ? */
	all_zero = 1;
	for (i = 0; i < mtd->oobblock; i++) {
		if (data_buf[i] != 0) {
			all_zero = 0;
			break;
		}
	}
	if (all_zero) {
		return(0);	/* So, unable to drop bit */
	}


	/* Clear call count */
	call_count = 0;


	/* Add bit error */
	printk("nand_force_bit_error: %d/%d:\n", page / 32, page % 32);
	for (i = 0; i < CONFIG_MTD_NAND_CXD4707_2X_ECC_ERROR_BIT; i++) {
		nand_add_bit_error(data_buf, mtd->oobblock);
	}

	/* Write back */
	this->data_poi = this->data_buf;
	ret = nand_write_page(mtd, this, page, 0, mtd->oobblock,
			      NULL, 0);
	if (ret != 0) {
		DEBUG(MTD_DEBUG_LEVEL0,
		      "%s: Failed overwrite, page 0x%08x ",
		      __FUNCTION__, page);
	}

	return(0);
}
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_ECC_TEST */

static int
nand_dma_write_page(struct mtd_info *mtd, unsigned int page,
                    const u_char *buf, const u_char *eccbuf)
{
	struct nand_chip *this;
	u32 config, dma_intr, page_addr, page_ctrl, dma_ctrl, extra[32];
	int ecc_size, extra_size, i, ret;
#ifdef CONFIG_MTD_NAND_VERIFY_WRITE
	u_char *data_buf, *oob_buf;
#endif	/* CONFIG_MTD_NAND_VERIFY_WRITE */

	this = mtd->priv;


	/* Cache write back before DMA write */
	clean_dcache_range((unsigned long)buf,
	                   (unsigned long)(buf + mtd->oobblock - 1));


	/* Setup DMA controller */
	if (ppc4xx_init_dma_channel(nand_dma_ch, &nand_dma_normal_setting) !=
	    DMA_STATUS_GOOD) {
		printk("%s: Unable to setup DMA channel\n", chip_name);
		return -EIO;
	}

	if (ppc4xx_set_dma_mode(nand_dma_ch, DMA_MODE_MM_DEVATDST) !=
	    DMA_STATUS_GOOD) {
		printk("%s: Unable to set DMA mode\n", chip_name);
		return -EIO;
	}

	if (ppc4xx_enable_dma_interrupt(nand_dma_ch) != DMA_STATUS_GOOD) {
		printk("%s: Unable to use interrupt from DMAC\n", chip_name);
		return -EIO;
	}


	ppc4xx_set_dma_addr2(nand_dma_ch,
	                     virt_to_phys((u_char *)buf),
	                     NAND0_REG_BASE + CS1_IMAGE_BASE);

	ppc4xx_set_dma_count(nand_dma_ch, mtd->oobblock);


	/* Check NAND flash device status */
	while (cxd4707_nand_dev_ready() == 0)
		;

	/* Setup NAND flash controller */
	config = REG_READL(cxd4707_nand + NAND_CONFIG);
	config |= NAND_CONFIG_FCLR;
		/* FCLR: 1 */
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
	printk("CONFIG:    %08x\n", config);
#endif  /* CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG */
	REG_WRITEL(config, cxd4707_nand + NAND_CONFIG);

	/* Write extra data */
	ecc_size = mtd->oobsize / 16;
	extra_size = ecc_size * 3;
	if (eccbuf != NULL) {
		memcpy(extra, eccbuf, mtd->oobsize);
		for (i = 0; i < extra_size; i++) {
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
			printk("NAND_PAGE_EXTRA(%d) <- extra[%d] = %08x\n",
			       i, ecc_size + i, extra[ecc_size + i]);
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG */
			NAND_WRITEL(cxd4707_nand + NAND_PAGE_EXTRA(i),
				    extra[ecc_size + i]);
		}
	} else {
		for (i = 0; i < extra_size; i++) {
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
			printk("NAND_PAGE_EXTRA(%d) <- %08x\n", i, 0xffffffff);
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG */
			NAND_WRITEL(cxd4707_nand + NAND_PAGE_EXTRA(i),
				    0xffffffff);
		}
	}

	dma_intr = NAND_DMA_INTR_ANA  | NAND_DMA_INTR_DREQ  |
	           NAND_DMA_INTR_MEXT | NAND_DMA_INTR_MECC  |
	           NAND_DMA_INTR_EXT  | NAND_DMA_INTR_ECC   |
	           NAND_DMA_INTR_DT   | NAND_DMA_INTR_REQ   |
	           NAND_DMA_INTR_TE;
	/* PCNT: 0, PCNTS: 0, ANA: 1,
	   DRO:  0, DREQ:  1,
	   MEXT: 1, MECC:  1, MDT: 0, MREQ: 0, MTE: 0,
	   EXT:  1, ECC:   1, DT:  1, REQ:  1, TE:  1 */
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
	printk("DMA_INTR:  %08x\n", dma_intr);
#endif  /* CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG */
	REG_WRITEL(dma_intr, cxd4707_nand + NAND_DMA_INTR);

	page_addr = page << (this->page_shift + 1);
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
	printk("PAGE_ADDR: %08x\n", page_addr);
#endif  /* CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG */
	REG_WRITEL(page_addr, cxd4707_nand + NAND_PAGE_ADDR);

	page_ctrl = NAND_PAGE_CTRL_EPC | NAND_PAGE_CTRL_BAC |
	            NAND_PAGE_CTRL_DIR;
		/* EPC: 1, BAC: 1, DIR: 1, TS: 0 */
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
	printk("PAGE_CTRL: %08x\n", page_ctrl);
#endif  /* CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG */
	REG_WRITEL(page_ctrl, cxd4707_nand + NAND_PAGE_CTRL);


	/* semaphore */
	cxd4707_dmac_need_interrupt = 1;

	/* Start DMA controller */
	ppc4xx_enable_dma(nand_dma_ch);
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
#ifdef CONFIG_CWING
	printk("DMAC: %08x %08x %08x %08x %08x\n",
	       mfdcr(DCRN_DMACR0), mfdcr(DCRN_DMACT0), mfdcr(DCRN_DMASA0),
	       mfdcr(DCRN_DMADA0), mfdcr(DCRN_DMA0_SR));
#else	/* CONFIG_CWING */
	printk("DMAC: %08x %08x %08x %08x %08x\n",
	       mfdcr(DCRN_DMACR4), mfdcr(DCRN_DMACT4), mfdcr(DCRN_DMASA4),
	       mfdcr(DCRN_DMADA4), mfdcr(DCRN_DMA1_SR));
#endif	/* CONFIG_CWING */
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG */

	/* Start DMA transfer */
	dma_ctrl = NAND_DMA_CTRL_DMA;
		/* DMA: 1 */
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
	printk("DMA_CTRL:  %08x\n", dma_ctrl);
#endif  /* CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG */
	REG_WRITEL(dma_ctrl, cxd4707_nand + NAND_DMA_CTRL);

	/* Wait interrupt */
	if (cxd4707_dmac_wait() != 0) {
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
		ppc4xx_show_status(NAND_DMA_CONTROLLER);
		ppc4xx_show_dmach(nand_dma_ch);
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG */
	}


	/* Stop DMA controller */
	ppc4xx_disable_dma(nand_dma_ch);


	/* Clear pending internal DMA request */
	if (nand_dma_dummy_read(mtd) != 0) {
		printk("%s: Unable clear pending internal DMA request\n",
		       chip_name);
		return -EIO;
	}


	/* Check write error status */
	ret = 0;
	if (cxd4707_nand_error_page(mtd, page) != 0) {
		ret = -EIO;
	}


	/* Clear DMA controller setting */
	dma_intr = REG_READL(cxd4707_nand + NAND_DMA_INTR);
	dma_intr = NAND_DMA_INTR_DREQ | NAND_DMA_INTR_MEXT |
	           NAND_DMA_INTR_MECC | NAND_DMA_INTR_MDT  |
	           NAND_DMA_INTR_MREQ | NAND_DMA_INTR_EXT  |
	           NAND_DMA_INTR_ECC  | NAND_DMA_INTR_DT   |
	           NAND_DMA_INTR_REQ  | NAND_DMA_INTR_TE;
	REG_WRITEL(dma_intr, cxd4707_nand + NAND_DMA_INTR);

	/* Disable interrupt */
	if (ppc4xx_disable_dma_interrupt(nand_dma_ch) != DMA_STATUS_GOOD) {
		printk("%s: Unable to use interrupt from DMAC\n", chip_name);
		return -EIO;
	}

#ifdef CONFIG_MTD_NAND_VERIFY_WRITE
	data_buf = this->data_buf;
	oob_buf = data_buf + mtd->oobblock;

	ret = nand_dma_read_page(mtd, page, data_buf, oob_buf);
	if (ret != 0) {
		DEBUG(MTD_DEBUG_LEVEL0,
		      "%s: Failed write verify, page 0x%08x ",
		      __FUNCTION__, page);
		return -EIO;
	}

	for (i = 0; i < mtd->oobblock; i++) {
		if (data_buf[i] != buf[i]) {
			DEBUG(MTD_DEBUG_LEVEL0,
			      "%s: Failed write verify, page 0x%08x ",
			      __FUNCTION__, page);
			return -EIO;
		}
	}
	for (i = 0; i < mtd->oobsize / 4; i++) {
		if (oob_buf[i] != 0) {
			DEBUG(MTD_DEBUG_LEVEL0,
			      "%s: Failed write verify, page 0x%08x ",
			      __FUNCTION__, page);
			return -EIO;
		}
	}
	if (eccbuf != NULL) {
		for (i = mtd->oobsize / 4; i < mtd->oobsize; i++) {
			if (oob_buf[i] != eccbuf[i]) {
				DEBUG(MTD_DEBUG_LEVEL0,
				      "%s: Failed write verify, page 0x%08x ",
				      __FUNCTION__, page);
				return -EIO;
			}
		}
	}
#endif	/* CONFIG_MTD_NAND_VERIFY_WRITE */
#ifdef CONFIG_MTD_NAND_CXD4707_2X_ECC_TEST
	nand_force_bit_error(mtd, page);
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_ECC_TEST */

	return(ret);
}

static int
nand_dma_read_page(struct mtd_info *mtd, unsigned int page,
                   u_char *buf, u_char *eccbuf)
{
	struct nand_chip *this;
	u32 config, dma_intr, page_addr, page_ctrl, extra[32];
	int ret, ecc_size, extra_size, i;

	this = mtd->priv;


	/* Check buffer alignment */
	if (!CACHE_ALIGNED(buf)) {
		printk("%s: DMA buffer must be aligned to cache line size\n",
		       chip_name);
		return(-EINVAL);
	}


	/* Setup DMA controller */
	if (ppc4xx_init_dma_channel(nand_dma_ch, &nand_dma_normal_setting) !=
	    DMA_STATUS_GOOD) {
		printk("%s: Unable to setup DMA channel\n", chip_name);
		return -EIO;
	}

	if (ppc4xx_set_dma_mode(nand_dma_ch, DMA_MODE_MM_DEVATSRC) !=
	    DMA_STATUS_GOOD) {
		printk("%s: Unable to set DMA mode\n", chip_name);
		return -EIO;
	}

	if (ppc4xx_enable_dma_interrupt(nand_dma_ch) != DMA_STATUS_GOOD) {
		printk("%s: Unable to use interrupt from DMAC\n", chip_name);
		return -EIO;
	}


	ppc4xx_set_dma_addr2(nand_dma_ch,
	                     NAND0_REG_BASE + CS1_IMAGE_BASE,
	                     virt_to_phys(buf));

	ppc4xx_set_dma_count(nand_dma_ch, mtd->oobblock);


	/* Check NAND flash device status */
	while (cxd4707_nand_dev_ready() == 0)
		;

	/* Setup NAND flash controller */
	config = REG_READL(cxd4707_nand + NAND_CONFIG);
	config |= NAND_CONFIG_FCLR;
		/* FCLR: 1 */
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
	printk("CONFIG:    %08x\n", config);
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG */
	REG_WRITEL(config, cxd4707_nand + NAND_CONFIG);

	dma_intr = NAND_DMA_INTR_PCNT | NAND_DMA_INTR_ANA  |
	           NAND_DMA_INTR_DREQ | NAND_DMA_INTR_MEXT |
	           NAND_DMA_INTR_MECC | NAND_DMA_INTR_EXT  |
	           NAND_DMA_INTR_ECC  | NAND_DMA_INTR_DT   |
	           NAND_DMA_INTR_REQ  | NAND_DMA_INTR_TE;
		/* PCNT: 1, PCNTS: 0, ANA: 1,
		   DRO:  0, DREQ:  1,
		   MEXT: 1, MECC:  1, MDT: 0, MREQ: 0, MTE: 0,
		   EXT:  1, ECC:   1, DT:  1, REQ:  1, TE:  1 */
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
	printk("DMA_INTR:  %08x\n", dma_intr);
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG */
	REG_WRITEL(dma_intr, cxd4707_nand + NAND_DMA_INTR);

	page_addr = page << (this->page_shift + 1);
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
	printk("PAGE_ADDR: %08x\n", page_addr);
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG */
	REG_WRITEL(page_addr, cxd4707_nand + NAND_PAGE_ADDR);

	page_ctrl = NAND_PAGE_CTRL_EPC | NAND_PAGE_CTRL_BAC;
		/* EPC: 1, BAC: 1, DIR: 0, TS: 0 */
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
	printk("PAGE_CTRL: %08x\n", page_ctrl);
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG */
	REG_WRITEL(page_ctrl, cxd4707_nand + NAND_PAGE_CTRL);


	/* semaphore */
	cxd4707_dmac_need_interrupt = 1;

	/* Start DMA controller */
	ppc4xx_enable_dma(nand_dma_ch);
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
#ifdef CONFIG_CWING
	printk("DMAC: %08x %08x %08x %08x %08x\n",
	       mfdcr(DCRN_DMACR0), mfdcr(DCRN_DMACT0), mfdcr(DCRN_DMASA0),
	       mfdcr(DCRN_DMADA0), mfdcr(DCRN_DMA0_SR));
#else	/* CONFIG_CWING */
	printk("DMAC: %08x %08x %08x %08x %08x\n",
	       mfdcr(DCRN_DMACR4), mfdcr(DCRN_DMACT4), mfdcr(DCRN_DMASA4),
	       mfdcr(DCRN_DMADA4), mfdcr(DCRN_DMA1_SR));
#endif	/* CONFIG_CWING */
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG */

	/* Cache invalidate before DMA transfer */
	invalidate_dcache_range((unsigned long)buf,
	                        (unsigned long)(buf + mtd->oobblock - 1));

	/* Start DMA transfer */
	page_ctrl = NAND_PAGE_CTRL_TS;
		/* EPC: 0, BAC: 0: DIR: 0: TS: 1 */
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
	printk("PAGE_CTRL: %08x\n", page_ctrl);
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG */
	REG_WRITEL(page_ctrl, cxd4707_nand + NAND_PAGE_CTRL);

	/* Wait interrupt */
	if (cxd4707_dmac_wait() != 0) {
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
		ppc4xx_show_status(NAND_DMA_CONTROLLER);
		ppc4xx_show_dmach(nand_dma_ch);
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG */
	}


	/* Stop DMA controller */
	ppc4xx_disable_dma(nand_dma_ch);


	/* Cache invalidate after DMA read */
	invalidate_dcache_range((unsigned long)buf,
	                        (unsigned long)(buf + mtd->oobblock - 1));


	/* Read ECC code and extra data, and check ECC status */
	ret = 0;
	if (eccbuf != NULL) {
		ecc_size = mtd->oobsize / 16;
		extra_size = ecc_size * 3;

		/* Read ECC code */
		for (i = 0; i < ecc_size; i++) {
			extra[i] = NAND_READL(cxd4707_nand +
			                      NAND_PAGE_ECC_CODE(i));
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
			printk("extra[%d] <- NAND_PAGE_ECC_CODE(%d) = %08x\n",
			       i, i, extra[i]);
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG */
		}

		/* Read extra data */
		for (i = 0; i < extra_size; i++) {
			extra[ecc_size + i] = NAND_READL(cxd4707_nand +
			                                 NAND_PAGE_EXTRA(i));
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG
			printk("extra[%d] <- NAND_PAGE_EXTRA(%d) = %08x\n",
			       ecc_size + i, i, extra[ecc_size + i]);
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA_DEBUG */
		}

		/* Save ECC code and extra data */
		memcpy(eccbuf, extra, mtd->oobsize);

		/* Check ECC status */
		if (cxd4707_nand_ecc_status(mtd, page) != 0) {
			ret = -EIO;
		}
	}


	/* Clear DMA controller setting */
	dma_intr = REG_READL(cxd4707_nand + NAND_DMA_INTR);
	dma_intr = NAND_DMA_INTR_DREQ | NAND_DMA_INTR_MEXT |
	           NAND_DMA_INTR_MECC | NAND_DMA_INTR_MDT  |
	           NAND_DMA_INTR_MREQ | NAND_DMA_INTR_EXT  |
	           NAND_DMA_INTR_ECC  | NAND_DMA_INTR_DT   |
	           NAND_DMA_INTR_REQ  | NAND_DMA_INTR_TE;
	REG_WRITEL(dma_intr, cxd4707_nand + NAND_DMA_INTR);

	/* Disable interrupt */
	if (ppc4xx_disable_dma_interrupt(nand_dma_ch) != DMA_STATUS_GOOD) {
		printk("%s: Unable to use interrupt from DMAC\n", chip_name);
		return -EIO;
	}

	return(ret);
}


/*
 * NAND DMA write without ECC
 */
static int
nand_dma_write(struct mtd_info *mtd, loff_t to,
               size_t len, size_t *retlen, const u_char *buf)
{
	return(nand_dma_write_ecc(mtd, to, len, retlen, buf, NULL, 0));
}

/*
 * NAND DMA write with ECC
 */
static int
nand_dma_write_ecc(struct mtd_info *mtd, loff_t to,
                   size_t len, size_t *retlen,
                   const u_char *buf, u_char *eccbuf, int oobsel)
{
	struct nand_chip *this;
	unsigned int page;
	int ret, oob;
	size_t written;

	DEBUG(MTD_DEBUG_LEVEL3, "%s: to = 0x%08x, len = %i\n",
	      __FUNCTION__, (unsigned int)to, (int)len);

	/* Do not allow writes past end of device */
	if ((to + len) > mtd->size) {
		DEBUG(MTD_DEBUG_LEVEL0,
		      "%s: Attempt to write beyond end of device\n",
		      __FUNCTION__);
		*retlen = 0;
		return -EINVAL;
	}

	/* Offset and length must be aligned to NAND flash page size  */
	/* Buffer address has also alignment limitation */
	if (((to & (mtd->oobblock - 1)) != 0) ||
	    ((len & (mtd->oobblock - 1)) != 0) ||
	    !DMA_ALIGNED(buf)) {
		return(nand_write_ecc(mtd, to, len, retlen,
		                      buf, eccbuf, oobsel));
	}

	/* Get start page address */
	this = mtd->priv;
	page = to >> this->page_shift;

	/* Parameters are OK */
	ret = 0;
	oob = 0;

	/* Grab the lock and see if the device is available */
	nand_get_chip(this, mtd, FL_WRITING, NULL);

	/* Select the NAND device */
	nand_select();

	/* Set NAND flash to writable state */
	nand_write_unprotect();

	/* Check the WP bit */
	this->cmdfunc (mtd, NAND_CMD_STATUS, -1, -1);
	if ((NAND_READB(this->IO_ADDR_R) & 0x80) == 0) {
		DEBUG(MTD_DEBUG_LEVEL0, "nand_dma_write_ecc: "
		                        "Device is write protected!!!\n");
		ret = -EIO;
		goto out;
	}

	/* Loop until all data is written */
	for (written = 0; written < len; written += mtd->oobblock) {
		ret = nand_dma_write_page(mtd, page, &(buf[written]),
		                          eccbuf ? &(eccbuf[oob]) : NULL);
		if (ret != 0) {
			DEBUG(MTD_DEBUG_LEVEL0,
			      "%s: DMA write error (%d)\n", __FUNCTION__, ret);
			break;
		}

		/* Increment oob position and page address */
		oob += mtd->oobsize;
		page++;
	}

out:
	/* Set NAND flash to unwritable state */
	nand_write_protect();

	/* De-select the NAND device */
	nand_deselect();

	/* Wake up anyone waiting on the device */
	spin_lock_bh(&CHIP_LOCK);
	CHIP_STATE = FL_READY;
	wake_up(&CHIP_WQ);
	spin_unlock_bh(&CHIP_LOCK);

	*retlen = written;
	return(ret);
}

/*
 * NAND DMA read without ECC
 */
static int
nand_dma_read(struct mtd_info *mtd, loff_t from,
              size_t len, size_t *retlen, u_char *buf)
{
	return(nand_dma_read_ecc(mtd, from, len, retlen, buf, NULL, 0));
}

/*
 * NAND DMA read with ECC
 */
static int
nand_dma_read_ecc(struct mtd_info *mtd, loff_t from,
                  size_t len, size_t *retlen,
                  u_char *buf, u_char *oob_buf, int oobsel)
{
	struct nand_chip *this;
	unsigned int page;
	int ret, oob;
	size_t read;

	DEBUG(MTD_DEBUG_LEVEL3, "%s: from = 0x%08x, len = %i\n",
	      __FUNCTION__, (unsigned int)from, (int)len);

	/* Do not allow reads past end of device */
	if ((from + len) > mtd->size) {
		DEBUG(MTD_DEBUG_LEVEL0,
		      "%s: Attempt to read beyond end of device\n",
		      __FUNCTION__);
		*retlen = 0;
		return(-EINVAL);
	}

	/* Offset and length must be aligned to NAND flash page size  */
	/* Buffer address has also alignment limitation */
	if (((from & (mtd->oobblock - 1)) != 0) ||
	    ((len & (mtd->oobblock - 1)) != 0) ||
	    !CACHE_ALIGNED(buf) ||
	    !DMA_ALIGNED(buf)) {
		if (oob_buf != NULL) {
			/*
			 * Unable to obtain DMA mode ECC code
			 * by PIO mode transfer
			 */
			return(-EINVAL);
		}
		return(nand_read_ecc(mtd, from, len, retlen,
		                     buf, oob_buf, oobsel));
	}

	/* Get start page address */
	this = mtd->priv;
	page = from >> this->page_shift;

	/* Parameters are OK */
	ret = 0;
	oob = 0;

	/* Grab the lock and see if the device is available */
	nand_get_chip(this, mtd ,FL_READING, NULL);

	/* Select the NAND device */
	nand_select();

	/* Loop until all data read */
	for (read = 0; read < len; read += mtd->oobblock) {
		ret = nand_dma_read_page(mtd, page, &(buf[read]),
		                         oob_buf ? &(oob_buf[oob]) : NULL);
		if (ret != 0) {
			DEBUG(MTD_DEBUG_LEVEL0,
			      "%s: DMA read error (%d)\n", __FUNCTION__, ret);
			break;
		}

		/* Increment oob position and page address */
		oob += mtd->oobsize;
		page++;
	}

	/* De-select the NAND device */
	nand_deselect();

	/* Wake up anyone waiting on the device */
	spin_lock_bh(&CHIP_LOCK);
	CHIP_STATE = FL_READY;
	wake_up(&CHIP_WQ);
	spin_unlock_bh(&CHIP_LOCK);

	*retlen = read;
	return(ret);
}

/*
 * NAND DMA read with PIO mode ECC code
 */
static int
nand_dma_read_ecc_compat(struct mtd_info *mtd, loff_t from,
                         size_t len, size_t *retlen,
                         u_char *buf, u_char *oob_buf, int oobsel)
{
	struct nand_chip *this;
	unsigned int page;
	int ret, oob;
	size_t read;

	DEBUG(MTD_DEBUG_LEVEL3, "%s: from = 0x%08x, len = %i\n",
	      __FUNCTION__, (unsigned int)from, (int)len);

	/* Do not allow reads past end of device */
	if ((from + len) > mtd->size) {
		DEBUG(MTD_DEBUG_LEVEL0,
		      "%s: Attempt to read beyond end of device\n",
		      __FUNCTION__);
		*retlen = 0;
		return(-EINVAL);
	}

	/* Offset and length must be aligned to NAND flash page size  */
	/* Buffer address has also alignment limitation */
	if (((from & (mtd->oobblock - 1)) != 0) ||
	    ((len & (mtd->oobblock - 1)) != 0) ||
	    !CACHE_ALIGNED(buf) ||
	    !DMA_ALIGNED(buf)) {
		return(nand_read_ecc(mtd, from, len, retlen,
		                     buf, oob_buf, oobsel));
	}

	/* Get start page address */
	this = mtd->priv;
	page = from >> this->page_shift;

	/* Parameters are OK */
	ret = 0;
	oob = 0;

	/* Grab the lock and see if the device is available */
	nand_get_chip(this, mtd ,FL_READING, NULL);

	/* Select the NAND device */
	nand_select();

	/* Loop until all data read */
	for (read = 0; read < len; read += mtd->oobblock) {
		ret = nand_dma_read_page(mtd, page, &(buf[read]), NULL);
		if (ret != 0) {
			DEBUG(MTD_DEBUG_LEVEL0,
			      "%s: DMA read error (%d)\n", __FUNCTION__, ret);
			break;
		}

		if (oob_buf != NULL) {
			/* De-select the NAND device */
			nand_deselect();

			/* Change chip state */
			spin_lock_bh(&CHIP_LOCK);
			CHIP_STATE = FL_READY;
			spin_unlock_bh(&CHIP_LOCK);

			ret = nand_read_oob(mtd, page << this->page_shift,
			                    mtd->oobsize, &len,
			                    &(oob_buf[oob]));

			/* Grab the lock and see if the device is available */
			nand_get_chip(this, mtd ,FL_READING, NULL);

			/* Select the NAND device */
			nand_select();

			if (ret != 0) {
				DEBUG(MTD_DEBUG_LEVEL0,
				      "%s: PIO read error (%d)\n",
				      __FUNCTION__, ret);
				break;
			}
		}

		/* Increment oob position and page address */
		oob += mtd->oobsize;
		page++;
	}

	/* De-select the NAND device */
	nand_deselect();

	/* Wake up anyone waiting on the device */
	spin_lock_bh(&CHIP_LOCK);
	CHIP_STATE = FL_READY;
	wake_up(&CHIP_WQ);
	spin_unlock_bh(&CHIP_LOCK);

	*retlen = read;
	return(ret);
}
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA */


/*
 * Setup routine for NAND flash interface
 */
static int __init cxd4707_nand_init_interface(void)
{
#ifdef CONFIG_CWING
	unsigned int cr, ap;
#else	/* CONFIG_CWING */
	unsigned int sdr0_pfc0;
#endif	/* CONFIG_CWING */
	int err;
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA
	u32 config;
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA */
	u32 page_intr;
#ifdef CONFIG_MTD_NAND_CXD4707_2X_TUNING
	u32 tim_prm;
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_TUNING */

#ifdef CONFIG_CWING
	/*
	 * Change EBC setting
	 */
	cr = mfdcri(DCRN_EBC0, PB4CR);
	if (cr != EBC0_PB4CR) {
		printk("EBC0_PB4CR: %08x -> ", cr);
		mtdcri(DCRN_EBC0, PB4CR, EBC0_PB4CR);
		cr = mfdcri(DCRN_EBC0, PB4CR);
		printk("%08x\n", cr);
	}

	ap = mfdcri(DCRN_EBC0, PB4AP);
	if (ap != EBC0_PB4AP) {
		printk("EBC0_PB4AP: %08x -> ", ap);
		mtdcri(DCRN_EBC0, PB4AP, EBC0_PB4AP);
		ap = mfdcri(DCRN_EBC0, PB4AP);
		printk("%08x\n", ap);
	}

	/* C-Wing has 2-way interleaved NAND Flash chip */
	cxd4707_nand_allow_interleave = 1;
#else	/* CONFIG_CWING */
	/* Check function multiplex setting */
	sdr0_pfc0 = mfdcri(DCRN_SDR0, PFC0);
	if (PFC0_GRP0_FUNC(sdr0_pfc0) == PFC0_GRP0_NAND8) {
		cxd4707_nand_allow_interleave = 0;
		printk("%s: %d-bit width mode\n", chip_name, 8);
	} else if (PFC0_GRP0_FUNC(sdr0_pfc0) == PFC0_GRP0_NAND16) {
		cxd4707_nand_allow_interleave = 1;
		printk("%s: %d-bit width mode\n", chip_name, 16);
	} else {
		printk("%s: NAND flash interface disabled\n", chip_name);
		printk(KERN_DEBUG "%s: sdr0_pfc0: %08x\n",
		       chip_name, sdr0_pfc0);
		return -ENODEV;
	}
#endif	/* CONFIG_CWING */


#if CXD4707_NAND_CHIPS > 1
#ifdef CONFIG_FRX
	/* Register GPIO ports */

	/* NAND_XCS */
	err = ppc4xx_gpio_register_bit(FRX_GPIO_NAND_XCS_PORT,
	                               FRX_GPIO_NAND_XCS_BIT, chip_name);
	if (err != 0) {
		printk("%s: Unable to register GPIO port %d bit %d\n",
		       chip_name,
		       FRX_GPIO_NAND_XCS_PORT,
		       FRX_GPIO_NAND_XCS_BIT);
		return -ENODEV;
	}
	err = ppc4xx_gpio_set_mode_bit(FRX_GPIO_NAND_XCS_PORT,
	                               FRX_GPIO_NAND_XCS_BIT,
	                               FRX_GPIO_NAND_XCS_MODE);
	if (err != 0) {
		printk("%s: Initialization failed on GPIO port %d bit %d\n",
		       chip_name,
		       FRX_GPIO_NAND_XCS_PORT,
		       FRX_GPIO_NAND_XCS_BIT);
		goto exit_1;
	}

	/* TRACE_CBT_CTRL */
	err = ppc4xx_gpio_register_bit(FRX_GPIO_TRACE_CBT_CTRL_PORT,
	                               FRX_GPIO_TRACE_CBT_CTRL_BIT, chip_name);
	if (err != 0) {
		printk("%s: Unable to register GPIO port %d bit %d\n",
		       chip_name,
		       FRX_GPIO_TRACE_CBT_CTRL_PORT,
		       FRX_GPIO_TRACE_CBT_CTRL_BIT);
		goto exit_1;
	}
	err = ppc4xx_gpio_set_mode_bit(FRX_GPIO_TRACE_CBT_CTRL_PORT,
	                               FRX_GPIO_TRACE_CBT_CTRL_BIT,
	                               FRX_GPIO_TRACE_CBT_CTRL_MODE);
	if (err != 0) {
		printk("%s: Initialization failed on GPIO port %d bit %d\n",
		       chip_name,
		       FRX_GPIO_TRACE_CBT_CTRL_PORT,
		       FRX_GPIO_TRACE_CBT_CTRL_BIT);
		goto exit_2;
	}
#endif	/* CONFIG_FRX */
#endif	/* CXD4707_NAND_CHIPS > 1 */


	/* Allocate IRQ */
#ifdef CONFIG_CWING
	err = ppc4xx_pic_setup(NAND0_INT,
	                       IBM4xx_IRQ_SNS_EDG, IBM4xx_IRQ_POL_NEG);
	if (err != 0) {
		printk("%s: ppc4xx_pic_setup: error: %d\n", chip_name, err);
		goto exit_2;
	}
#endif	/* CONFIG_CWING */
	err = ppc4xx_pic_clear_status(NAND0_INT);
	if (err != 0) {
		printk("%s: ppc4xx_pic_clear_status: error: %d\n",
		       chip_name, err);
		goto exit_2;
	}
	err = request_irq(NAND0_INT, cxd4707_nand_intr, 0,
	                  chip_name, &cxd4707_nand);
	if (err != 0) {
		printk("%s: request_irq: error: %d\n", chip_name, err);
		goto exit_2;
	}
	disable_irq(NAND0_INT);


#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA
	/*
	 * DMAC initialize routine
	 */
	nand_dma_ch = ppc4xx_request_dma(NAND_DMA_CONTROLLER);
	printk("nand_dma_ch: %d\n", nand_dma_ch);
	if (nand_dma_ch < 0) {
		printk("%s: Unable to allocate DMA channel"
		       " for NAND Interface.\n", chip_name);
		return -ENODEV;
	}

	nand_dma_irq = ppc4xx_get_dma_irq(nand_dma_ch);
	printk("nand_dma_irq: %d\n", nand_dma_irq);
	if (nand_dma_irq < 0) {
		printk("%s: Unable to allocate IRQ "
		       " for NAND Interface.\n", chip_name);
		ppc4xx_free_dma(nand_dma_ch);
		return -ENODEV;
	}

	if (ppc4xx_pic_clear_status(nand_dma_irq) != 0) {
		printk("%s: Unable to clear irq for DMAC\n", chip_name);
		return -ENODEV;
	}

	if (request_irq(nand_dma_irq, cxd4707_dmac_intr, 0,
	                chip_name, &cxd4707_nand) != 0) {
		printk("%s: Unable to allocate irq for DMAC\n", chip_name);
		return -ENODEV;
	}

#if 0	/* XXX */
	free_irq(nand_dma_irq, &cxd4707_nand);
	ppc4xx_free_dma(nand_dma_ch);
#endif	/* XXX */

#ifndef CONFIG_CWING
	/* Enable burst transfer mode for NAND flash interface */
	mtdcri(DCRN_SDR0, NBURST, 0x00000001);
#endif	/* CONFIG_CWING */
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA */


	/* Remap NAND flash interface registers */
	cxd4707_nand = ioremap64(NAND0_REG_BASE, NAND0_REG_SIZE);
	if (cxd4707_nand == NULL) {
		printk("%s: Unable to remap NAND flash interface registers\n",
		       chip_name);
		goto exit_3;
	}


	/* Setup NAND flash interface registers */
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA
	/* EMC_PHY_ND_CONFIG register */
	config = REG_READL(cxd4707_nand + NAND_CONFIG);
	printk("CONFIG: %08x -> ", config);
	config &= ~NAND_CONFIG_RECC;
	config |= NAND_CONFIG_NDI | NAND_CONFIG_WECC;
	REG_WRITEL(config, cxd4707_nand + NAND_CONFIG);
	config = REG_READL(cxd4707_nand + NAND_CONFIG);
	printk("%08x\n", config);
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA */

	/* EMC_PHY_ND_PAGE_INTR register */
	page_intr = REG_READL(cxd4707_nand + NAND_PAGE_INTR);
	if ((page_intr & NAND_PAGE_INTR_MBSY) != 0) {
		/* Enable interrupt on Busy to Ready status change */
		printk("PAGE_INTR: %08x -> ", page_intr);
		REG_WRITEL((page_intr & ~NAND_PAGE_INTR_MBSY) & NAND_PAGE_INTR_MASK,
		       cxd4707_nand + NAND_PAGE_INTR);
		page_intr = REG_READL(cxd4707_nand + NAND_PAGE_INTR);
		printk("%08x\n", page_intr);
	}

#ifdef CONFIG_MTD_NAND_CXD4707_2X_TUNING
	/* EMC_PHY_ND_TIM_PRM register */
	/* Setup NAND flash access timing */
	tim_prm = REG_READL(cxd4707_nand + NAND_TIM_PRM);
	if (tim_prm != CXD4707_NAND_TIM_PRM) {
		/* Change NAND flash access timing */
		printk("TIM_PRM: %08x -> ", tim_prm);
		REG_WRITEL(CXD4707_NAND_TIM_PRM, cxd4707_nand + NAND_TIM_PRM);
		tim_prm = REG_READL(cxd4707_nand + NAND_TIM_PRM);
		printk("%08x\n", tim_prm);
	} else {
		printk("TIM_PRM: %08x\n", tim_prm);
	}
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_TUNING */


	/* Initialize wait_queue */
	init_waitqueue_head(&cxd4707_nand_wait_queue);
#ifdef CONFIG_MTD_NAND_CXD4707_2X_DMA
	init_waitqueue_head(&cxd4707_dmac_wait_queue);
#endif	/* CONFIG_MTD_NAND_CXD4707_2X_DMA */


	return 0;


exit_3:
	/* Unregister IRQ */
	free_irq(NAND0_INT, &cxd4707_nand);

exit_2:
#if CXD4707_NAND_CHIPS > 1
#ifdef CONFIG_FRX
	/* Unregister TRACE_CBT_CTRL */
	err = ppc4xx_gpio_unregister_bit(FRX_GPIO_TRACE_CBT_CTRL_PORT,
					 FRX_GPIO_TRACE_CBT_CTRL_BIT);
	if (err != 0) {
		printk("%s: Unable to unregister GPIO port %d bit %d\n",
		       chip_name,
		       FRX_GPIO_TRACE_CBT_CTRL_PORT,
		       FRX_GPIO_TRACE_CBT_CTRL_BIT);
	}

exit_1:
	/* Unregister NAND_XCS */
	err = ppc4xx_gpio_unregister_bit(FRX_GPIO_NAND_XCS_PORT,
					 FRX_GPIO_NAND_XCS_BIT);
	if (err != 0) {
		printk("%s: Unable to unregister GPIO port %d bit %d\n",
		       chip_name,
		       FRX_GPIO_NAND_XCS_PORT,
		       FRX_GPIO_NAND_XCS_BIT);
	}
#endif	/* CONFIG_FRX */
#endif	/* CXD4707_NAND_CHIPS > 1 */

	return -ENODEV;
}

/*
 * Setup routine for each NAND flash chip
 */
static int __init cxd4707_nand_init_chip(int chip)
{
	struct nand_chip *this;
#ifdef CONFIG_MTD_SBL
	int	res;
#endif	/* CONFIG_MTD_SBL */

	/* Check NAND flash chip no. */
	if ((chip < 0) || (chip >= CXD4707_NAND_CHIPS)) {
		printk("%s: Invalid NAND flash chip number: %d\n",
		       chip_name, chip);
		return -EINVAL;
	}

	/* Allocate memory for MTD device structure and private data */
	cxd4707_nand_mtd[chip] = kmalloc(sizeof(struct mtd_info) +
	                                 sizeof(struct nand_chip), GFP_KERNEL);
	if (cxd4707_nand_mtd[chip] == NULL) {
		printk("%s: Unable to allocate MTD device structure"
		       " for NAND Interface.\n", chip_name);
		return -ENOMEM;
	}

	/* Initialize structures */
	memset(cxd4707_nand_mtd[chip], 0,
	       sizeof(struct mtd_info) + sizeof(struct nand_chip));

	/* Get pointer to private data */
	this = (struct nand_chip *)(cxd4707_nand_mtd[chip] + 1);

	/* Link the private data with the MTD structure */
	cxd4707_nand_mtd[chip]->priv = this;

	/* Setup struct nand_chip */
	this->IO_ADDR_R     = (unsigned long)(cxd4707_nand + NAND_PIO_DATA);
	this->IO_ADDR_W     = (unsigned long)(cxd4707_nand + NAND_PIO_DATA);
	this->hwcontrol     = cxd4707_nand_hwcontrols[chip];
	this->dev_ready     = cxd4707_nand_dev_ready;
	this->chip_delay    = 12;				/* XXX */
	this->eccmode       = NAND_ECC_NONE;			/* XXX */
	this->cmdfunc       = cxd4707_nand_cmdfunc;
	this->waitfunc      = cxd4707_nand_wait;

	/* Scan to find existance of the device */
	if (cxd4707_nand_scan(cxd4707_nand_mtd[chip]) != 0) {
		kfree(cxd4707_nand_mtd[chip]);
		cxd4707_nand_mtd[chip] = NULL;
		return -ENXIO;
	}

	/* Allocate memory for internal data buffer */
	this->data_buf = kmalloc(cxd4707_nand_mtd[chip]->oobblock +
	                         cxd4707_nand_mtd[chip]->oobsize, GFP_KERNEL);
	if (this->data_buf == NULL) {
		printk("%s: Unable to allocate NAND data buffer"
		       " for NAND Interface.\n", chip_name);
		kfree(cxd4707_nand_mtd[chip]);
		cxd4707_nand_mtd[chip] = NULL;
		return -ENOMEM;
	}

	/* Allocate memory for internal data cache */
	this->data_cache = kmalloc(cxd4707_nand_mtd[chip]->oobblock +
	                           cxd4707_nand_mtd[chip]->oobsize, GFP_KERNEL);
	if (this->data_cache == NULL) {
		printk("%s: Unable to allocate NAND data cache"
		       " for NAND Interface.\n", chip_name);
		kfree (this->data_buf);
		kfree(cxd4707_nand_mtd[chip]);
		cxd4707_nand_mtd[chip] = NULL;
		return -ENOMEM;
	}
	this->cache_page = -1;

	/* Set NAND flash to writable state */
	nand_write_unprotect();
	/* Is NAND flash writable ? */
	if ((nand_get_status(cxd4707_nand_mtd[chip]) & 0x80) == 0) {
		printk(KERN_INFO "%s: NAND flash write protected\n", chip_name);
		cxd4707_nand_mtd[chip]->flags &= ~(MTD_WRITEABLE |
		                                   MTD_ERASEABLE);
	}
	/* Set NAND flash to unwritable state */
	nand_write_protect();

#ifdef CONFIG_MTD_SBL
	res = mtd_sbl_mtd_info_create(cxd4707_nand_mtd[chip],
	                              &cxd4707_nand_mtdsbl[chip],
				      mtd_sbl_dev_program_cxd4707,
				      mtd_sbl_dev_read_cxd4707);
	if (res < 0) {
		kfree (this->data_cache);
		kfree (this->data_buf);
		kfree(cxd4707_nand_mtd[chip]);
		cxd4707_nand_mtd[chip] = NULL;
		return res;
	}
	cxd4707_nand_mtdsbl[chip]->name = sbl_device_name[chip];
#ifdef CONFIG_SNSC_MTD_PROTECT
	cxd4707_nand_mtdsbl[chip]->protect = MTD_PROTECT_PROTECTED;
#endif /* CONFIG_SNSC_MTD_PROTECT */
	add_mtd_device(cxd4707_nand_mtdsbl[chip]);
#endif	/* CONFIG_MTD_SBL */

#if defined(CONFIG_MTD_SBL) || defined(CONFIG_SNSC_MTD_SDM_PARTITION)
	cxd4707_nand_mtd[chip]->name = mtd_device_name[chip];
#ifdef CONFIG_SNSC_MTD_PROTECT
	cxd4707_nand_mtd[chip]->protect = MTD_PROTECT_PROTECTED;
#endif  /* CONFIG_SNSC_MTD_PROTECT */
#endif	/* defined(CONFIG_MTD_SBL) || defined(CONFIG_SNSC_MTD_SDM_PARTITION) */
	add_mtd_device(cxd4707_nand_mtd[chip]);

#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
#ifdef CONFIG_MTD_SBL
	sdm_register_device(cxd4707_nand_mtdsbl[chip]);
#else	/* CONFIG_MTD_SBL */
	sdm_register_device(cxd4707_nand_mtd[chip]);
#endif	/* CONFIG_MTD_SBL */
#endif	/* CONFIG_SNSC_MTD_SDM_PARTITION */

	return 0;
}

/*
 * Main initialization routine
 */
static int __init cxd4707_nand_init(void)
{
	int status, found, i;
#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
	static int cxd4707_nand_initialized = 0;

	if (cxd4707_nand_initialized) {
		return 0;
	}
	cxd4707_nand_initialized = 1;
#endif	/* CONFIG_SNSC_MTD_SDM_PARTITION */

	/* Initialize NAND flash interface */
	status = cxd4707_nand_init_interface();
	if (status != 0) {
		return status;
	}

	/* Initialize NAND flash chips, and register it */
	found = 0;
	for (i = 0; i < CXD4707_NAND_CHIPS; i++) {
		status = cxd4707_nand_init_chip(i);
		if (status == 0) {
			found++;
		}
	}

	if (found == 0) {
		/* No NAND flash chip found */
		cxd4707_nand_cleanup();
		return status;
	}

	/* OK */
	return 0;
}

module_init(cxd4707_nand_init);


#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
int __init sdm_nand_init(void)
{
	return cxd4707_nand_init();
}

#endif	/* CONFIG_SNSC_MTD_SDM_PARTITION */


/*
 * Clean up routine
 */
static void __exit cxd4707_nand_cleanup_interface(void)
{
#if CXD4707_NAND_CHIPS > 1
#ifdef CONFIG_FRX
	int err;
#endif	/* CONFIG_FRX */
#endif	/* CXD4707_NAND_CHIPS > 1 */

	/* Unmap NAND flash interface registers */
	iounmap(cxd4707_nand);

	/* Free IRQ */
	free_irq(NAND0_INT, &cxd4707_nand);

#if CXD4707_NAND_CHIPS > 1
#ifdef CONFIG_FRX
	/* Unregister GPIO ports */

	/* TRACE_CBT_CTRL */
	err = ppc4xx_gpio_unregister_bit(FRX_GPIO_TRACE_CBT_CTRL_PORT,
	                                 FRX_GPIO_TRACE_CBT_CTRL_BIT);
	if (err != 0) {
		printk("%s: Unable to unregister GPIO port %d bit %d\n",
		       chip_name,
		       FRX_GPIO_TRACE_CBT_CTRL_PORT,
		       FRX_GPIO_TRACE_CBT_CTRL_BIT);
	}

	/* NAND_XCS */
	err = ppc4xx_gpio_unregister_bit(FRX_GPIO_NAND_XCS_PORT,
	                                 FRX_GPIO_NAND_XCS_BIT);
	if (err != 0) {
		printk("%s: Unable to unregister GPIO port %d bit %d\n",
		       chip_name,
		       FRX_GPIO_NAND_XCS_PORT,
		       FRX_GPIO_NAND_XCS_BIT);
	}
#endif	/* CONFIG_FRX */
#endif	/* CXD4707_NAND_CHIPS > 1 */
}

static void __exit cxd4707_nand_cleanup_chip(int chip)
{
	struct nand_chip *this;

	/* Check NAND flash chip no. */
	if ((chip < 0) || (chip >= CXD4707_NAND_CHIPS)) {
		printk("%s: Invalid NAND flash chip number: %d\n",
		       chip_name, chip);
		return;
	}

	if (cxd4707_nand_mtd[chip] == NULL) {
		return;
	}
	this = (struct nand_chip *)cxd4707_nand_mtd[chip]->priv;

	/* Unregister the device */
	del_mtd_device(cxd4707_nand_mtd[chip]);
#ifdef CONFIG_MTD_SBL
	del_mtd_device(cxd4707_nand_mtdsbl[chip]);
#endif	/* CONFIG_MTD_SBL */

	/* Free internal data buffer */
	if (this->data_cache != NULL) {
		kfree(this->data_cache);
	}
	if (this->data_buf != NULL) {
		kfree(this->data_buf);
	}

	/* Free the MTD device structure */
	kfree(cxd4707_nand_mtd[chip]);
}

static void __exit cxd4707_nand_cleanup(void)
{
	int i;

	for (i = CXD4707_NAND_CHIPS - 1; i >= 0; i--) {
		/* Unregister NAND flash chips */
		cxd4707_nand_cleanup_chip(i);
	}
	cxd4707_nand_cleanup_interface();
}

module_exit(cxd4707_nand_cleanup);


MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Board-specific glue layer for NAND flash on CXD4707 NAND flash interface");
