/*
 *  drivers/mtd/nand/snsc_mpu110-nand.c
 *
 *  Copyright 2002,2003 Sony Corporation.
 *  Copyright (C) 2000 Steven J. Hill (sjhill@cotw.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/slab.h>
#include <linux/module.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/partitions.h>
#include <linux/delay.h>
#include <asm/io.h>

#include <asm/arch/mpu110_series.h>
#include <asm/arch/gpio.h>
#include <asm/arch/regs.h>
#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
#include <linux/mtd/sdmparse.h>
#endif
#ifdef CONFIG_MTD_SBL
#include <linux/mtd/mtdsbl.h>
#endif

#define DEBUG_NAND_IO 0

#if DEBUG_NAND_IO & 1
char *cmdname[] = {
    "SETNCE",
    "CLRNCE",
    "SETCLE",
    "CLRCLE",
    "SETALE",
    "CLRALE",
};
#  define CMD(x) printk("%s\n", cmdname[(x) - 1]);
#  define READY(x) printk("busy = %d\n", x);
#else
#  define CMD(x) 
#  define READY(x) 
#endif

#define NAND_GPIO_IN   ( (1 << NAND_BUSY) )
#define NAND_GPIO_OUT  ( (1 << NAND_CS1) | (1 << NAND_CS0) | \
                         (1 << NAND_CLE) | (1 << NAND_ALE) )

#ifdef CONFIG_MTD_SBL
extern int mtd_sbl_dev_program_compat(struct sbl_dev_t *dev,
				      unsigned long block, int page, int mode,
				      const void *data, const void *extra);
extern int mtd_sbl_dev_read_compat(struct sbl_dev_t *dev,
				   unsigned long block, int page, int mode,
				   void *data, void *extra);
#endif

/*
 * MTD structure for MPU110 board
 */
static struct mtd_info *mpu110_mtd[2];
#ifdef CONFIG_MTD_SBL
static struct mtd_info *mpu110_mtdsbl[2];
#endif /* CONFIG_MTD_SBL */

#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
static int sdm_nand_initialized = 0;
#endif

/*
 * Values specific to the MPU110 board
 */
static int mpu110_fio_base = NAND_BASE;/* Where the flash mapped */

/*
 * Module stuff
 */
#ifdef MODULE
MODULE_PARM(mpu110_fio_base, "i");
__setup("mpu110_fio_base=", mpu110_fio_base);
#endif

/*
 * Define partitions for flash device
 */
#ifdef CONFIG_MTD_SBL
#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
static char *sbl_device_name[] = {
	"sdm device NAND/SBL 0",
	"sdm device NAND/SBL 1"
};
#else /* CONFIG_SNSC_MTD_SDM_PARTITION */
static char *sbl_device_name[] = {
	"NAND/SBL device 0",
	"NAND/SBL device 1"
};
#endif /* CONFIG_SNSC_MTD_SDM_PARRITION */
static char *mtd_device_name[] = {
	"NAND Flash raw device 0",
	"NAND Flash raw device 1"
};
#elif defined(CONFIG_SNSC_MTD_SDM_PARTITION)
static char *mtd_device_name[] = {
	"sdm device NAND 0",
	"sdm device NAND 1"
};
#else
static struct mtd_partition partition_info_0[] = {
	{
	  name: "MPU-110 NAND Flash partition 0",
	  offset: 0,
	  size:   MTDPART_SIZ_FULL },
	{ name: NULL, offset: 0, size: 0 }
};

static struct mtd_partition partition_info_1[] = {
	{
	  name: "MPU-110 NAND Flash partition 1",
	  offset: 0,
	  size:   MTDPART_SIZ_FULL },
	{ name: NULL, offset: 0, size: 0 }
};

static struct mtd_partition *partition_info[] = {
	partition_info_0,
	partition_info_1
};
#endif

/* 
 *	hardware specific access to control-lines
 */

/* for CS0 */
void mpu110_hwcontrol0(int cmd)
{
	static struct {
		int bitnum;
		int value;
	} tab[] = {
		[NAND_CTL_SETCLE] = { NAND_CLE, 1 },
		[NAND_CTL_CLRCLE] = { NAND_CLE, 0 },
		[NAND_CTL_SETALE] = { NAND_ALE, 1 },
		[NAND_CTL_CLRALE] = { NAND_ALE, 0 },
		[NAND_CTL_SETNCE] = { NAND_CS0, 0 },
		[NAND_CTL_CLRNCE] = { NAND_CS0, 1 }
	};
	dragonball_gpio_set_bit(NAND_GPIO_PORT, tab[cmd].bitnum, tab[cmd].value);
}

/* for CS1 */
void mpu110_hwcontrol1(int cmd)
{
	static struct {
		int bitnum;
		int value;
	} tab[] = {
		[NAND_CTL_SETCLE] = { NAND_CLE, 1 },
		[NAND_CTL_CLRCLE] = { NAND_CLE, 0 },
		[NAND_CTL_SETALE] = { NAND_ALE, 1 },
		[NAND_CTL_CLRALE] = { NAND_ALE, 0 },
		[NAND_CTL_SETNCE] = { NAND_CS1, 0 },
		[NAND_CTL_CLRNCE] = { NAND_CS1, 1 }
	};
	dragonball_gpio_set_bit(NAND_GPIO_PORT, tab[cmd].bitnum, tab[cmd].value);
}

/*
 *	read device ready pin
 */
int mpu110_device_ready(void)
{
	return  dragonball_gpio_get_bit(NAND_GPIO_PORT, NAND_BUSY);
}

void (*hwctrl[])(int)  = {
    mpu110_hwcontrol0,
    mpu110_hwcontrol1,
};

static int __init mpu110_init_chip(int num)
{
	struct nand_chip *this;
	struct mtd_info *mtd;
#ifdef CONFIG_MTD_SBL
	struct mtd_info *mtdsbl;
	int res;
#endif

	/* Allocate memory for MTD device structure and private data */
	mtd = kmalloc (sizeof(struct mtd_info) + sizeof (struct nand_chip),
		       GFP_KERNEL);
	if (mtd == NULL) {
		printk(KERN_WARNING "Unable to allocate "
		       "MPU-110 NAND MTD device structure.\n");
		return -ENOMEM;
	}

	/* Get pointer to private data */
	this = (struct nand_chip *)(mtd+1);

	/* Initialize structures */
	memset((char *)mtd, 0, sizeof(struct mtd_info));
	memset((char *)this, 0, sizeof(struct nand_chip));

	/* Link the private data with the MTD structure */
	mtd->priv = this;

	/* Set address of NAND IO lines */
	this->IO_ADDR_R = IO_ADDRESS(mpu110_fio_base);
	this->IO_ADDR_W = IO_ADDRESS(mpu110_fio_base);
	/* Set address of hardware control function */
	this->hwcontrol = hwctrl[num];
	this->dev_ready = mpu110_device_ready;
        //this->chip_delay = 3*1000; /* max. erase time [us] */

        /* MPU-110 board without fix to busy line does not work */
        if (!mpu110_device_ready()) {
                printk(KERN_WARNING "Error in NAND busy line!  "
		       "Has board fix applied properly?\n");
		kfree(mtd);
		return -ENXIO;
        }

	/* Scan to find existence of the device */
	if (nand_scan(mtd)) {
		kfree(mtd);
		return -ENXIO;
	}

	/* Allocate memory for internal data buffer */
	this->data_buf = kmalloc(sizeof(u_char) *
				 (mtd->oobblock + mtd->oobsize),
				 GFP_KERNEL);
	if (this->data_buf == NULL) {
		printk(KERN_WARNING
		       "Unable to allocate NAND data buffer for MPU-110.\n");
		kfree(mtd);
		return -ENOMEM;
	}

	/* Allocate memory for internal data buffer */
	this->data_cache = kmalloc(sizeof(u_char) *
				   (mtd->oobblock + mtd->oobsize),
				   GFP_KERNEL);
	if (this->data_cache == NULL) {
		printk(KERN_WARNING
		       "Unable to allocate NAND data cache for MPU-110.\n");
		kfree(this->data_buf);
		kfree(mtd);
		return -ENOMEM;
	}
	this->cache_page = -1;

	mpu110_mtd[num] = mtd;

#ifdef CONFIG_MTD_SBL
	res = mtd_sbl_mtd_info_create(mtd, &mtdsbl,
				      mtd_sbl_dev_program_compat,
				      mtd_sbl_dev_read_compat);
	if (res < 0) {
		return res;
	}
	mtdsbl->name = sbl_device_name[num];
	add_mtd_device(mtdsbl);
	mpu110_mtdsbl[num] = mtdsbl;
#endif /* CONFIG_MTD_SBL */

#if defined(CONFIG_MTD_SBL) || defined(CONFIG_SNSC_MTD_SDM_PARTITION)
	mtd->name = mtd_device_name[num];
	add_mtd_device(mtd);
#else
	{
		int n;

		for (n = 0; partition_info[num][n].name != NULL; n++);
		add_mtd_partitions(mtd, partition_info[num], n);
	}
#endif /* CONFIG_MTD_SBL || CONFIG_SNSC_MTD_SDM_PARTITION */


#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
#ifdef CONFIG_MTD_SBL
	sdm_register_device(mtdsbl);
#else
	sdm_register_device(mtd);
#endif /* CONFIG_MTD_SBL */
#endif /* CONFIG_SNSC_MTD_SDM_PARTITION */

	/* Return happy */
	return 0;
}

/* 
 * setup EIM
 */
static void __init mpu110_eim_setup(void)
{
	/* use default value */
        outl(regs.eim[1].u, EIM_BASE + EIM_CS1U);
        outl(regs.eim[1].l, EIM_BASE + EIM_CS1L);
}

/*
 * Main initialization routine
 */
static int err0, err1;          /* stores whether properly initialized */
int __init mpu110_init(void)
{
#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
	if (sdm_nand_initialized)
		return 0;
#endif
	/* initialize EIM for access */
	mpu110_eim_setup();

	if (dragonball_register_gpios(NAND_GPIO_PORT, NAND_GPIO_IN, 
				      GPIO | INPUT, "nand") < 0){
		printk(KERN_ERR "snsc_mpu110-nand: can't register gpio.\n");
		return -1;
	}
	if (dragonball_register_gpios(NAND_GPIO_PORT, NAND_GPIO_OUT,
				      GPIO | OUTPUT | OCR_DATA, "nand") < 0){
		printk(KERN_ERR "snsc_mpu110-nand: can't register gpio.\n");
		return -1;
	}

        mpu110_hwcontrol0(NAND_CTL_CLRNCE);
        mpu110_hwcontrol1(NAND_CTL_CLRNCE);
        mpu110_hwcontrol0(NAND_CTL_CLRCLE);
        mpu110_hwcontrol0(NAND_CTL_CLRALE);

        err0 = mpu110_init_chip(0);
        err1 = mpu110_init_chip(1);

	if (err0 == 0 || err1 == 0) {
#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
		sdm_nand_initialized = 1;
#endif
		return 0;
        } else { 
		return -1;
	}
}
module_init(mpu110_init);

#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
int __init sdm_nand_init( void )
{
	return mpu110_init();
}
#endif

static void __exit mpu110_cleanup_chip (struct mtd_info *mtd)
{
        struct nand_chip *this = (struct nand_chip *) (mtd+1);
        
	/* Unregister partitions */
	del_mtd_partitions(mtd);

	/* Unregister the device */
	del_mtd_device (mtd);

	/* Free internal data buffer */
	kfree (this->data_buf);
	kfree (this->data_cache);

	/* Free the MTD device structure */
	kfree (mtd);

        return;
}

/*
 * Clean up routine
 */
#ifdef MODULE
static void __exit mpu110_cleanup (void)
{
	if (mpu110_mtd[0] != NULL) {
		mpu110_cleanup_chip(mpu110_mtd[0]);
	}
	if (mpu110_mtd[1] != NULL) {
		mpu110_cleanup_chip(mpu110_mtd[1]);
	}

	dragonball_unregister_gpios(NAND_GPIO_PORT,
				    NAND_GPIO_IN | NAND_GPIO_OUT);

#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
	sdm_nand_initialized = 0;
#endif
}
module_exit(mpu110_cleanup);
#endif

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Sony Corporation");
MODULE_DESCRIPTION("Board-specific glue layer for NAND flash on MPU-110 board");
