/*
 *  drivers/mtd/nand/snsc_mpu210-nand.c
 *
 *  Copyright 2002,2003 Sony Corporation.
 *
 *  Derived from drivers/mtd/nand/spia.c
 *       Copyright (C) 2000 Steven J. Hill (sjhill@cotw.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *  Overview:
 *   This is a device driver for the NAND flash device found on the
 *   Sony NSC MPU-210 board which utilizes the Samsung K8F5608U0A part. This is
 *   a 256Mibit (32MiB x 8 bits) NAND flash device.
 */

#include <linux/slab.h>
#include <linux/module.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/partitions.h>
#include <linux/delay.h>
#include <asm/io.h>

#include <asm/it8172/it8172.h>
#include <asm/it8172/snsc_mpu210.h>
#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
#include <linux/mtd/sdmparse.h>
#endif
#ifdef CONFIG_MTD_SBL
#include <linux/mtd/mtdsbl.h>
#endif

#define DEBUG_NAND_IO 0

#if DEBUG_NAND_IO & 1
char *cmdname[] = {
    "SETNCE",
    "CLRNCE",
    "SETCLE",
    "CLRCLE",
    "SETALE",
    "CLRALE",
};
#  define CMD(x) printk("%s\n", cmdname[(x) - 1]);
#  define READY(x) printk("busy = %d\n", x);
#else
#  define CMD(x) 
#  define READY(x) 
#endif


#ifdef CONFIG_MTD_SBL
extern int mtd_sbl_dev_program_compat(struct sbl_dev_t *dev,
				      unsigned long block, int page, int mode,
				      const void *data, const void *extra);
extern int mtd_sbl_dev_read_compat(struct sbl_dev_t *dev,
				   unsigned long block, int page, int mode,
				   void *data, void *extra);
#endif

/*
 * MTD structure for MPU-210 board
 */
static struct mtd_info *mpu210_mtd[CONFIG_MTD_NAND_MPU210_CHIPS];
#ifdef CONFIG_MTD_SBL
static struct mtd_info *mpu210_mtdsbl[CONFIG_MTD_NAND_MPU210_CHIPS];
#endif /* CONFIG_MTD_SBL */

#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
static int sdm_nand_initialized = 0;
#endif

/*
 * Values specific to the MPU-210 board
 */
static int mpu210_io_base = MPU210_NAND_IO_BASE; /* Where the control line mapped */
static int mpu210_fio_base = MPU210_NAND_FIO_BASE;/* Where the flash mapped */

/*
 * Module stuff
 */
#ifdef MODULE
MODULE_PARM(mpu210_io_base, "i");
MODULE_PARM(mpu210_fio_base, "i");

__setup("mpu210_io_base=",mpu210_io_base);
__setup("mpu210_fio_base=",mpu210_fio_base);
#endif

/*
 * Define partitions for flash device
 */
#ifdef CONFIG_MTD_SBL
#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
static char *sbl_device_name[] = {
	"sdm device NAND/SBL 0",
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 2
	"sdm device NAND/SBL 1",
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 3
	"sdm device NAND/SBL 2",
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 4
	"sdm device NAND/SBL 3",
#endif
#endif
#endif
};
#else /* CONFIG_SNSC_MTD_SDM_PARTITION */
static char *sbl_device_name[] = {
	"NAND/SBL device 0",
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 2
	"NAND/SBL device 1",
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 3
	"NAND/SBL device 2",
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 4
	"NAND/SBL device 3",
#endif
#endif
#endif
};
#endif /* CONFIG_SNSC_MTD_SDM_PARRITION */
static char *mtd_device_name[] = {
	"NAND Flash raw device 0",
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 2
	"NAND Flash raw device 1",
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 3
	"NAND Flash raw device 2",
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 4
	"NAND Flash raw device 3",
#endif
#endif
#endif
};
#elif defined(CONFIG_SNSC_MTD_SDM_PARTITION)
static char *mtd_device_name[] = {
	"sdm device NAND 0",
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 2
	"sdm device NAND 1",
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 3
	"sdm device NAND 2",
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 4
	"sdm device NAND 3",
#endif
#endif
#endif
};
#else
static struct mtd_partition partition_info_0[] = {
	{
	  name: "MPU-210 NAND Flash partition 0",
	  offset: 0,
	  size:   MTDPART_SIZ_FULL },
	{ name: NULL, offset: 0, size: 0 }
};

#if CONFIG_MTD_NAND_MPU210_CHIPS >= 2
static struct mtd_partition partition_info_1[] = {
	{
	  name: "MPU-210 NAND Flash partition 1",
	  offset: 0,
	  size:   MTDPART_SIZ_FULL },
	{ name: NULL, offset: 0, size: 0 }
};

#if CONFIG_MTD_NAND_MPU210_CHIPS >= 3
static struct mtd_partition partition_info_2[] = {
	{
	  name: "MPU-210 NAND Flash partition 2",
	  offset: 0,
	  size:   MTDPART_SIZ_FULL },
	{ name: NULL, offset: 0, size: 0 }
};

#if CONFIG_MTD_NAND_MPU210_CHIPS >= 4
static struct mtd_partition partition_info_3[] = {
	{
	  name: "MPU-210 NAND Flash partition 3",
	  offset: 0,
	  size:   MTDPART_SIZ_FULL },
	{ name: NULL, offset: 0, size: 0 }
};
#endif
#endif
#endif

static struct mtd_partition *partition_info[] = {
	partition_info_0
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 2
	, partition_info_1
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 3
	, partition_info_2
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 4
	, partition_info_3
#endif
#endif
#endif
};
#endif

/* 
 *	hardware specific access to control-lines
 */

#define	NAND_IO_TYPE		u16
#define	NAND_IO_IN(port)	(*(volatile NAND_IO_TYPE *)KSEG1ADDR(port))
#define	NAND_IO_OUT(val, port)	do { *(volatile NAND_IO_TYPE *)KSEG1ADDR(port) = (val); } while (0)

/* for CS0 */
void mpu210_hwcontrol0(int cmd)
{
    NAND_IO_TYPE tmp;

    CMD(cmd);
    tmp = NAND_IO_IN(mpu210_io_base);
    switch(cmd){
	case NAND_CTL_SETCLE: 
            tmp &= ~MPU210_FLASH_NAND_CLE;	/* /NAND_CLE */
            break;
	case NAND_CTL_CLRCLE:
            tmp |= MPU210_FLASH_NAND_CLE;	/* /NAND_CLE */
            break;
	case NAND_CTL_SETALE: 
            tmp &= ~MPU210_FLASH_NAND_ALE;	/* /NAND_ALE */
            break;
	case NAND_CTL_CLRALE:
            tmp |= MPU210_FLASH_NAND_ALE;	/* /NAND_ALE */
            break;
	case NAND_CTL_SETNCE: 
            tmp &= ~MPU210_FLASH_NAND_CS0;	/* /NAND_CS0 */
            break;
	case NAND_CTL_CLRNCE:
            tmp |= MPU210_FLASH_NAND_CS0;	/* /NAND_CS0 */
            break;
    }
    NAND_IO_OUT(tmp, mpu210_io_base);
}

#if CONFIG_MTD_NAND_MPU210_CHIPS >= 2
/* for CS1 */
void mpu210_hwcontrol1(int cmd)
{
    NAND_IO_TYPE tmp;

    CMD(cmd);
    tmp = NAND_IO_IN(mpu210_io_base);
    switch(cmd){
	case NAND_CTL_SETCLE: 
            tmp &= ~MPU210_FLASH_NAND_CLE;	/* /NAND_CLE */
            break;
	case NAND_CTL_CLRCLE:
            tmp |= MPU210_FLASH_NAND_CLE;	/* /NAND_CLE */
            break;
	case NAND_CTL_SETALE: 
            tmp &= ~MPU210_FLASH_NAND_ALE;	/* /NAND_ALE */
            break;
	case NAND_CTL_CLRALE:
            tmp |= MPU210_FLASH_NAND_ALE;	/* /NAND_ALE */
            break;
	case NAND_CTL_SETNCE: 
            tmp &= ~MPU210_FLASH_NAND_CS1;	/* /NAND_CS1 */
            break;
	case NAND_CTL_CLRNCE:
            tmp |= MPU210_FLASH_NAND_CS1;	/* /NAND_CS1 */
            break;
    }
    NAND_IO_OUT(tmp, mpu210_io_base);
}

#if CONFIG_MTD_NAND_MPU210_CHIPS >= 3
/* for CS2 */
void mpu210_hwcontrol2(int cmd)
{
    NAND_IO_TYPE tmp;

    CMD(cmd);
    tmp = NAND_IO_IN(mpu210_io_base);
    switch(cmd){
	case NAND_CTL_SETCLE: 
            tmp &= ~MPU210_FLASH_NAND_CLE;	/* /NAND_CLE */
            break;
	case NAND_CTL_CLRCLE:
            tmp |= MPU210_FLASH_NAND_CLE;	/* /NAND_CLE */
            break;
	case NAND_CTL_SETALE: 
            tmp &= ~MPU210_FLASH_NAND_ALE;	/* /NAND_ALE */
            break;
	case NAND_CTL_CLRALE:
            tmp |= MPU210_FLASH_NAND_ALE;	/* /NAND_ALE */
            break;
	case NAND_CTL_SETNCE: 
            tmp &= ~MPU210_FLASH_NAND_CS2;	/* /NAND_CS2 */
            break;
	case NAND_CTL_CLRNCE:
            tmp |= MPU210_FLASH_NAND_CS2;	/* /NAND_CS2 */
            break;
    }
    NAND_IO_OUT(tmp, mpu210_io_base);
}

#if CONFIG_MTD_NAND_MPU210_CHIPS >= 4
/* for CS3 */
void mpu210_hwcontrol3(int cmd)
{
    NAND_IO_TYPE tmp;

    CMD(cmd);
    tmp = NAND_IO_IN(mpu210_io_base);
    switch(cmd){
	case NAND_CTL_SETCLE: 
            tmp &= ~MPU210_FLASH_NAND_CLE;	/* /NAND_CLE */
            break;
	case NAND_CTL_CLRCLE:
            tmp |= MPU210_FLASH_NAND_CLE;	/* /NAND_CLE */
            break;
	case NAND_CTL_SETALE: 
            tmp &= ~MPU210_FLASH_NAND_ALE;	/* /NAND_ALE */
            break;
	case NAND_CTL_CLRALE:
            tmp |= MPU210_FLASH_NAND_ALE;	/* /NAND_ALE */
            break;
	case NAND_CTL_SETNCE: 
            tmp &= ~MPU210_FLASH_NAND_CS3;	/* /NAND_CS3 */
            break;
	case NAND_CTL_CLRNCE:
            tmp |= MPU210_FLASH_NAND_CS3;	/* /NAND_CS3 */
            break;
    }
    NAND_IO_OUT(tmp, mpu210_io_base);
}
#endif
#endif
#endif

/*
 *	read device ready pin
 */
int mpu210_device_ready(void)
{
    /* Check /FLASH_BY */
    if ((NAND_IO_IN(mpu210_io_base) & MPU210_FLASH_NAND_BY) == MPU210_FLASH_NAND_BY) {
        READY(1);
        return 1;                /* ready */
    } else {
        READY(0);
        return 0;                /* not ready */
    }
}

void (*hwctrl[])(int)  = {
    mpu210_hwcontrol0
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 2
    , mpu210_hwcontrol1
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 3
    , mpu210_hwcontrol2
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 4
    , mpu210_hwcontrol3
#endif
#endif
#endif
};

static int __init mpu210_init_chip (int num)
{
	struct nand_chip *this;
	struct mtd_info *mtd;
#ifdef CONFIG_MTD_SBL
	struct mtd_info *mtdsbl;
	int res;
#endif

	/* Allocate memory for MTD device structure and private data */
	mtd = kmalloc (sizeof(struct mtd_info) + sizeof (struct nand_chip),
		       GFP_KERNEL);
	if (mtd == NULL) {
		printk(KERN_WARNING "Unable to allocate "
		       "MPU-210 NAND MTD device structure.\n");
		return -ENOMEM;
	}

	/* Get pointer to private data */
	this = (struct nand_chip *)(mtd+1);

	/* Initialize structures */
	memset((char *)mtd, 0, sizeof(struct mtd_info));
	memset((char *)this, 0, sizeof(struct nand_chip));

	/* Link the private data with the MTD structure */
	mtd->priv = this;

	/* Set address of NAND IO lines */
	this->IO_ADDR_R = KSEG1ADDR(mpu210_fio_base);
	this->IO_ADDR_W = KSEG1ADDR(mpu210_fio_base);
	/* Set address of hardware control function */
	this->hwcontrol = hwctrl[num];
	this->dev_ready = mpu210_device_ready;
	/* 15 us command delay time */
        this->chip_delay = 15;

	/* Scan to find existence of the device */
	if (nand_scan(mtd)) {
		kfree(mtd);
		return -ENXIO;
	}

	/* Allocate memory for internal data buffer */
	this->data_buf = kmalloc(sizeof(u_char) *
				 (mtd->oobblock + mtd->oobsize),
				 GFP_KERNEL);
	if (this->data_buf == NULL) {
		printk(KERN_WARNING
		       "Unable to allocate NAND data buffer for MPU-210.\n");
		kfree(mtd);
		return -ENOMEM;
	}

	/* Allocate memory for internal data buffer */
	this->data_cache = kmalloc(sizeof(u_char) *
				   (mtd->oobblock + mtd->oobsize),
				   GFP_KERNEL);
	if (this->data_cache == NULL) {
		printk(KERN_WARNING
		       "Unable to allocate NAND data cache for MPU-210.\n");
		kfree (this->data_buf);
		kfree (mtd);
		return -ENOMEM;
	}
	this->cache_page = -1;

	mpu210_mtd[num] = mtd;

#ifdef CONFIG_MTD_SBL
	res = mtd_sbl_mtd_info_create(mtd, &mtdsbl,
				      mtd_sbl_dev_program_compat,
				      mtd_sbl_dev_read_compat);
	if (res < 0) {
		return res;
	}
	mtdsbl->name = sbl_device_name[num];
	add_mtd_device(mtdsbl);
	mpu210_mtdsbl[num] = mtdsbl;
#endif /* CONFIG_MTD_SBL */

#if defined(CONFIG_MTD_SBL) || defined(CONFIG_SNSC_MTD_SDM_PARTITION)
	mtd->name = mtd_device_name[num];
	add_mtd_device(mtd);
#else
	{
		int n;

		for (n = 0; partition_info[num][n].name != NULL; n++);
		add_mtd_partitions(mtd, partition_info[num], n);
	}
#endif /* CONFIG_MTD_SBL || CONFIG_SNSC_MTD_SDM_PARTITION */


#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
#ifdef CONFIG_MTD_SBL
	sdm_register_device(mtdsbl);
#else
	sdm_register_device(mtd);
#endif /* CONFIG_MTD_SBL */
#endif /* CONFIG_SNSC_MTD_SDM_PARTITION */

	/* Return happy */
	return 0;
}


/*
 * Main initialization routine
 */
static int err0;		/* stores whether properly initialized */
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 2
static int err1;
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 3
static int err2;
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 4
static int err3;
#endif
#endif
#endif

int __init mpu210_init(void)
{
#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
	if (sdm_nand_initialized)
		return 0;
#endif

        mpu210_hwcontrol0(NAND_CTL_CLRNCE);
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 2
        mpu210_hwcontrol1(NAND_CTL_CLRNCE);
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 3
        mpu210_hwcontrol2(NAND_CTL_CLRNCE);
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 4
        mpu210_hwcontrol3(NAND_CTL_CLRNCE);
#endif
#endif
#endif
        mpu210_hwcontrol0(NAND_CTL_CLRCLE);
        mpu210_hwcontrol0(NAND_CTL_CLRALE);

        err0 = mpu210_init_chip(0);
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 2
        err1 = mpu210_init_chip(1);
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 3
        err2 = mpu210_init_chip(2);
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 4
        err3 = mpu210_init_chip(3);
#endif
#endif
#endif

        if (err0)
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 2
            if (err1)
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 3
                if (err2)
#if CONFIG_MTD_NAND_MPU210_CHIPS >= 4
                    if (err3)
#endif
#endif
#endif
                        return -1;

#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
	sdm_nand_initialized = 1;
#endif
        return 0;
}
module_init(mpu210_init);

#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
int __init sdm_nand_init( void )
{
	return mpu210_init();
}
#endif

static void __exit mpu210_cleanup_chip (struct mtd_info *mtd)
{
        struct nand_chip *this = (struct nand_chip *) (mtd+1);
        
	/* Unregister partitions */
	del_mtd_partitions(mtd);

	/* Unregister the device */
	del_mtd_device (mtd);

	/* Free internal data buffer */
	kfree (this->data_buf);
	kfree (this->data_cache);

	/* Free the MTD device structure */
	kfree (mtd);

        return;
}

/*
 * Clean up routine
 */
#ifdef MODULE
static void __exit mpu210_cleanup (void)
{
	int i;

	for (i = 0; i < CONFIG_MTD_NAND_MPU210_CHIPS; i++) {
		if (mpu210_mtd[i] != NULL)
			mpu210_cleanup_chip(mpu210_mtd[i]);
	}

#ifdef CONFIG_SNSC_MTD_SDM_PARTITION
	sdm_nand_initialized = 0;
#endif
}
module_exit(mpu210_cleanup);
#endif

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Sony NSC");
MODULE_DESCRIPTION("Board-specific glue layer for NAND flash on MPU-210 board");
