/*
 *  mtdsbl.c -- the MTD driver for NAND/SBL
 *
 */
/*
 *  Copyright 2003,2004 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/wrapper.h>
#include <linux/sched.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>
#include <asm/softirq.h>
#include <linux/init.h>
#include <linux/config.h>
#include <linux/interrupt.h>

#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/mtdsbl.h>

#include "sbl.h"
#include "ecc.h"

MODULE_AUTHOR("Sony NSC");
MODULE_DESCRIPTION("MTD driver for NAND/SBL");
MODULE_LICENSE("GPL");

/* ======================================================================= *
 * constant                                                                *
 * ======================================================================= */

#define MTD_SBL_NAME	"MTD/SBL"

/* ======================================================================= *
 * prototype declaration                                                   *
 * ======================================================================= */

static int mtd_sbl_read(struct mtd_info *mtd, loff_t from, size_t len,
			size_t *retlen, u_char *buf);
static int mtd_sbl_write(struct mtd_info *mtd, loff_t to, size_t len,
			 size_t *retlen, const u_char *buf);
static int mtd_sbl_erase(struct mtd_info *mtd, struct erase_info *instr);
#ifdef CONFIG_MTD_SBL_JFFS2_HACK
static int mtd_sbl_writev(struct mtd_info *mtd, const struct iovec *vecs,
			  unsigned long count, loff_t to, size_t *retlen);
static int mtd_sbl_writev_ecc(struct mtd_info *mtd, const struct iovec *vecs,
			      unsigned long count, loff_t to, size_t *retlen,
			      u_char *eccbuf, int oobsel);
static int mtd_sbl_read_ecc(struct mtd_info *mtd, loff_t from, size_t len,
			    size_t *retlen, u_char *buf,
			    u_char *eccbuf, int oobsel);
static int mtd_sbl_write_ecc(struct mtd_info *mtd, loff_t to, size_t len,
			     size_t *retlen, const u_char *buf,
			     u_char *eccbuf, int oobsel);
static int mtd_sbl_read_oob(struct mtd_info *mtd, loff_t from, size_t len,
			    size_t *retlen, u_char *buf);
static int mtd_sbl_write_oob(struct mtd_info *mtd, loff_t to, size_t len,
			     size_t *retlen, const u_char *buf);
#endif /* CONFIG_MTD_SBL_JFFS2_HACK */

static int mtd_sbl_dev_is_bad(sbl_dev_t *dev, unsigned long block, int mode);
static int mtd_sbl_dev_erase(sbl_dev_t *dev, unsigned long block, int mode);

/* ======================================================================= *
 * global variable                                                         *
 * ======================================================================= */

/* ======================================================================= *
 * public function                                                         *
 * ======================================================================= */

EXPORT_SYMBOL(mtd_sbl_mtd_info_create);

int
mtd_sbl_mtd_info_create(struct mtd_info *nand_mtd, struct mtd_info **mtd,
			sbl_dev_program_t dev_program, sbl_dev_read_t dev_read)
{
	int res;
	mtd_sbl_dev_t *mtd_sbl_dev;
	sbl_dev_t *sbl_dev;

	sbl_dev = kmalloc(sizeof(sbl_dev_t), GFP_KERNEL);
	if (sbl_dev == NULL) {
		return -ENOMEM;
	}
	memset(sbl_dev, 0, sizeof(sbl_dev_t));
	sbl_dev->sbl_version	= SBL_CURRENT_VERSION;
	sbl_dev->flags		= (SBL_F_VERIFY_DATA | SBL_F_VERIFY_EXTRA);
	sbl_dev->normal_blocks	= nand_mtd->size / nand_mtd->erasesize;
	sbl_dev->spare_blocks	=
		sbl_dev->normal_blocks * MTD_SBL_SPARE_PERCENT / 100;
	sbl_dev->normal_blocks	-= sbl_dev->spare_blocks;
	sbl_dev->pages_per_block = nand_mtd->erasesize / nand_mtd->oobblock;
	sbl_dev->data_area_size	= nand_mtd->oobblock;
	sbl_dev->extra_area_size = nand_mtd->oobsize;
	sbl_dev->page_buf	= kmalloc(sbl_dev->data_area_size, GFP_KERNEL);
	sbl_dev->copy_page_buf	= kmalloc(sbl_dev->data_area_size, GFP_KERNEL);
	sbl_dev->sbl_info_cache	= kmalloc((sbl_dev->normal_blocks +
					   sbl_dev->spare_blocks) * 4,
					  GFP_KERNEL);
	sbl_dev->is_bad		= mtd_sbl_dev_is_bad;
	sbl_dev->erase		= mtd_sbl_dev_erase;
	sbl_dev->program	= dev_program;
	sbl_dev->read		= dev_read;
	if (sbl_dev->page_buf == NULL || sbl_dev->copy_page_buf == NULL ||
	    sbl_dev->sbl_info_cache == NULL) {
		return -ENOMEM;
	}

	*mtd = kmalloc(sizeof(struct mtd_info), GFP_KERNEL);
	if (*mtd == NULL) {
		return -ENOMEM;
	}

	memset(*mtd, 0, sizeof(struct mtd_info));
	(*mtd)->name		= MTD_SBL_NAME;
	(*mtd)->erasesize	= nand_mtd->erasesize;
	(*mtd)->size		= sbl_dev->normal_blocks * nand_mtd->erasesize;
	(*mtd)->oobblock	= nand_mtd->oobblock;
	(*mtd)->type		= MTD_SBL;
	(*mtd)->flags		= (MTD_CLEAR_BITS | MTD_ERASEABLE);
	(*mtd)->module		= THIS_MODULE;
	(*mtd)->erase		= mtd_sbl_erase;
	(*mtd)->read		= mtd_sbl_read;
	(*mtd)->write		= mtd_sbl_write;
#ifdef CONFIG_MTD_SBL_JFFS2_HACK
	(*mtd)->oobsize		= 16;
	(*mtd)->ecctype		= MTD_ECC_SW;
	(*mtd)->writev		= mtd_sbl_writev;
	(*mtd)->writev_ecc	= mtd_sbl_writev_ecc;
	(*mtd)->read_ecc	= mtd_sbl_read_ecc;
	(*mtd)->write_ecc	= mtd_sbl_write_ecc;
	(*mtd)->read_oob	= mtd_sbl_read_oob;
	(*mtd)->write_oob	= mtd_sbl_write_oob;
#endif /* CONFIG_MTD_SBL_JFFS2_HACK */

	mtd_sbl_dev = kmalloc(sizeof(mtd_sbl_dev_t), GFP_KERNEL);
	if (mtd_sbl_dev == NULL) {
		return -ENOMEM;
	}
	mtd_sbl_dev->sbl_dev	= sbl_dev;
	mtd_sbl_dev->nand_mtd	= nand_mtd;
	mtd_sbl_dev->buf	= kmalloc(sbl_dev->data_area_size, GFP_KERNEL);
	init_MUTEX(&mtd_sbl_dev->sem);
	if (mtd_sbl_dev->buf == NULL) {
		return -ENOMEM;
	}

	sbl_dev->priv		= mtd_sbl_dev;
	(*mtd)->priv		= mtd_sbl_dev;

	res = sbl_dev_init(sbl_dev);
	if (res < 0) {
		return -EINVAL;
	}

	return 0;
}

/* ======================================================================= *
 * MTD driver interface                                                    *
 * ======================================================================= */

static int
mtd_sbl_read(struct mtd_info *mtd, loff_t from, size_t len, size_t *retlen,
	     u_char *buf)
{
	int res;
	mtd_sbl_dev_t *mtd_sbl_dev = mtd->priv;
	sbl_dev_t *sbl_dev = mtd_sbl_dev->sbl_dev;
	unsigned long block;
	int page;
	int nread;
	unsigned char *buf_i;

	*retlen = 0;

	if (from + len > mtd->size) {
		return -EINVAL;
	}

	while (len > 0) {
		nread = (sbl_dev->data_area_size -
			 ((unsigned long)from + *retlen) %
			 sbl_dev->data_area_size);
		if (len < nread) {
			nread = len;
		}

		down(&mtd_sbl_dev->sem);
		if (nread != sbl_dev->data_area_size) {
			buf_i = mtd_sbl_dev->buf;
		} else {
			buf_i = buf + *retlen;
		}
		block = (((unsigned long)from + *retlen) /
			 sbl_dev->data_area_size) /
			sbl_dev->pages_per_block;
		page = (((unsigned long)from + *retlen) /
			sbl_dev->data_area_size) %
			sbl_dev->pages_per_block;
		
		res = sbl_read(sbl_dev, block, page,
			       SBL_M_SPARE | SBL_M_DATA | SBL_M_ECC |
			       SBL_M_MOVEBACK | SBL_M_VERIFY, buf_i);
		if (res == -ESBLNOENT) {
			memset(buf_i, 0xff, sbl_dev->data_area_size);
		} else if (res == -ESBLECC) {
			printk(KERN_WARNING "mtd_sbl_read() ECC error "
			       "at %ld/%d ... ignore\n", block, page);
		} else if (res < 0) {
			printk(KERN_NOTICE "mtd_sbl_read() failed (%d) "
			       "at %ld/%d\n", -res, block, page);
			up(&mtd_sbl_dev->sem);
			return -EIO;
		}
		if (nread != sbl_dev->data_area_size) {
			memcpy(buf + *retlen,
			       buf_i + (((unsigned long)from + *retlen) %
					sbl_dev->data_area_size),
			       nread);
		}
		up(&mtd_sbl_dev->sem);

		*retlen += nread;
		len -= nread;
	}

	return 0;
}

static int
mtd_sbl_write(struct mtd_info *mtd, loff_t to, size_t len, size_t *retlen,
	      const u_char *buf)
{
	int res;
	mtd_sbl_dev_t *mtd_sbl_dev = mtd->priv;
	sbl_dev_t *sbl_dev = mtd_sbl_dev->sbl_dev;
	unsigned long block;
	int page;
	int nwrite;
	const unsigned char *buf_i;
	int nretry;

	*retlen = 0;

	if (to + len > mtd->size) {
		return -EINVAL;
	}

	while (len > 0) {
		nwrite = (sbl_dev->data_area_size -
			  ((unsigned long)to + *retlen) %
			  sbl_dev->data_area_size);
		if (len < nwrite) {
			nwrite = len;
		}

		down(&mtd_sbl_dev->sem);
		if (nwrite != sbl_dev->data_area_size) {
			memset(mtd_sbl_dev->buf, 0xff,
			       sbl_dev->data_area_size);
			memcpy(mtd_sbl_dev->buf +
			       (((unsigned long)to + *retlen) %
				sbl_dev->data_area_size),
			       buf + *retlen, nwrite);
			buf_i = mtd_sbl_dev->buf;
		} else {
			buf_i = buf + *retlen;
		}
		block = (((unsigned long)to + *retlen) /
			 sbl_dev->data_area_size) /
			sbl_dev->pages_per_block;
		page = (((unsigned long)to + *retlen) /
			sbl_dev->data_area_size) %
			sbl_dev->pages_per_block;

		nretry = MTD_SBL_NRETRY;
retry:
		if (nretry < 0) {
			printk(KERN_WARNING "mtd_sbl_write(): retry failed\n");
			up(&mtd_sbl_dev->sem);
			return -EIO;
		}
		nretry--;

		res = sbl_program(sbl_dev, block, page,
				  SBL_M_SPARE | SBL_M_DATA | SBL_M_ECC |
				  SBL_M_VERIFY,
				  buf_i);
		if (res < 0) {
			printk(KERN_NOTICE "mtd_sbl_write() failed (%d) "
			       "at %ld/%d\n", -res, block, page);
			if (res == -ESBLPROGRAM || res == -ESBLECC ||
			    res == -ESBLVERIFY || res == -ESBLINFO) {
				goto retry;
			}
			up(&mtd_sbl_dev->sem);
			return -EIO;
		}
		up(&mtd_sbl_dev->sem);

		*retlen += nwrite;
		len -= nwrite;
	}

	return 0;
}

static int
mtd_sbl_erase(struct mtd_info *mtd, struct erase_info *instr)
{
	int res;
	mtd_sbl_dev_t *mtd_sbl_dev = mtd->priv;
	sbl_dev_t *sbl_dev = mtd_sbl_dev->sbl_dev;
	unsigned long block;
	int nerased_byte;
	int nretry;

	if (instr->addr & (mtd->erasesize - 1)) {
		return -EINVAL;
	}
	if (instr->len & (mtd->erasesize - 1)) {
		return -EINVAL;
	}
	if (instr->len + instr->addr > mtd->size) {
		return -EINVAL;
	}

	nerased_byte = 0;
	while (nerased_byte < instr->len) {
		block = ((instr->addr + nerased_byte) /
			 sbl_dev->data_area_size) /
			sbl_dev->pages_per_block;

		nretry = MTD_SBL_NRETRY;
retry:
		if (nretry < 0) {
			printk(KERN_WARNING "mtd_sbl_erase(): retry failed\n");
			return -EIO;
		}
		nretry--;

		down(&mtd_sbl_dev->sem);
		res = sbl_erase(sbl_dev, block, SBL_M_SPARE);
		up(&mtd_sbl_dev->sem);
		if (res < 0) {
			printk(KERN_NOTICE "mtd_sbl_erase() failed (%d) "
			       "at %ld\n", -res, block);
			if (res == -ESBLINFO) {
				goto retry;
			}
			return -EIO;
		}

		nerased_byte += (sbl_dev->data_area_size *
				 sbl_dev->pages_per_block);
	}
	instr->state = MTD_ERASE_DONE;
	if (instr->callback != NULL) {
		instr->callback(instr);
	}

	return 0;
}

#ifdef CONFIG_MTD_SBL_JFFS2_HACK
static int
mtd_sbl_writev(struct mtd_info *mtd, const struct iovec *vecs,
	       unsigned long count, loff_t to, size_t *retlen)
{
	int res;
	mtd_sbl_dev_t *mtd_sbl_dev = mtd->priv;
	sbl_dev_t *sbl_dev = mtd_sbl_dev->sbl_dev;
	unsigned long block;
	int page;
	int vec_i;
	unsigned long vec_pos;
	unsigned long len;
	unsigned long buf_to;
	int nwrite;
	int ncopy;
	int nretry;

	*retlen = 0;

	len = 0;
	for (vec_i = 0; vec_i < count; vec_i++) {
		len += vecs[vec_i].iov_len;
	}

	if (to + len > mtd->size) {
		return -EINVAL;
	}

	vec_i = 0;
	vec_pos = 0;
	while (len > 0) {
		nwrite = 0;
		buf_to = ((unsigned long)to + *retlen) %
			sbl_dev->data_area_size;

		down(&mtd_sbl_dev->sem);
		if (buf_to > 0) {
			memset(mtd_sbl_dev->buf, 0xff, buf_to);
		}
		while (buf_to + nwrite < sbl_dev->data_area_size &&
		       vec_i < count) {
			ncopy = sbl_dev->data_area_size - (buf_to + nwrite);
			if (vecs[vec_i].iov_len - vec_pos < ncopy) {
				ncopy = vecs[vec_i].iov_len - vec_pos;
			}
			memcpy(mtd_sbl_dev->buf + (buf_to + nwrite),
			       (unsigned char *)vecs[vec_i].iov_base + vec_pos,
			       ncopy);
			nwrite += ncopy;
			vec_pos += ncopy;
			if (vec_pos == vecs[vec_i].iov_len) {
				vec_i++;
				vec_pos = 0;
			}
		}
		if (buf_to + nwrite < sbl_dev->data_area_size) {
			memset(mtd_sbl_dev->buf + (buf_to + nwrite), 0xff,
			       sbl_dev->data_area_size - (buf_to + nwrite));
		}

		block = (((unsigned long)to + *retlen) /
			 sbl_dev->data_area_size) /
			sbl_dev->pages_per_block;
		page = (((unsigned long)to + *retlen) /
			sbl_dev->data_area_size) %
			sbl_dev->pages_per_block;

		nretry = MTD_SBL_NRETRY;
retry:
		if (nretry < 0) {
			printk(KERN_WARNING
			       "mtd_sbl_writev(): retry failed\n");
			up(&mtd_sbl_dev->sem);
			return -EIO;
		}
		nretry--;

		res = sbl_program(sbl_dev, block, page,
				  SBL_M_SPARE | SBL_M_DATA | SBL_M_ECC |
				  SBL_M_VERIFY,
				  mtd_sbl_dev->buf);
		if (res < 0) {
			printk(KERN_NOTICE "mtd_sbl_writev() failed (%d) "
			       "at %ld/%d\n", -res, block, page);
			if (res == -ESBLPROGRAM || res == -ESBLECC ||
			    res == -ESBLVERIFY || res == -ESBLINFO) {
				goto retry;
			}
			up(&mtd_sbl_dev->sem);
			return -EIO;
		}
		up(&mtd_sbl_dev->sem);

		*retlen += nwrite;
		len -= nwrite;
	}

	return 0;
}

static int
mtd_sbl_writev_ecc(struct mtd_info *mtd, const struct iovec *vecs,
		   unsigned long count, loff_t to, size_t *retlen,
		   u_char *eccbuf, int oobsel)
{
	return mtd_sbl_writev(mtd, vecs, count, to, retlen);
}

static int
mtd_sbl_read_ecc(struct mtd_info *mtd, loff_t from, size_t len,
		 size_t *retlen, u_char *buf, u_char *eccbuf, int oobsel)
{
	return mtd_sbl_read(mtd, from, len, retlen, buf);
}

static int
mtd_sbl_write_ecc(struct mtd_info *mtd, loff_t to, size_t len,
		  size_t *retlen, const u_char *buf,
		  u_char *eccbuf, int oobsel)
{
	return mtd_sbl_write(mtd, to, len, retlen, buf);
}

static int
mtd_sbl_read_oob(struct mtd_info *mtd, loff_t from, size_t len,
		 size_t *retlen, u_char *buf)
{
	memset(buf, 0xff, len);
	*retlen = len;

	return 0;
}

static int
mtd_sbl_write_oob(struct mtd_info *mtd, loff_t to, size_t len,
		  size_t *retlen, const u_char *buf)
{
	*retlen = len;

	return 0;
}
#endif /* CONFIG_MTD_SBL_JFFS2_HACK */

/* ======================================================================= *
 * driver access function for SBL                                          *
 * ======================================================================= */

static int
mtd_sbl_dev_is_bad(sbl_dev_t *dev, unsigned long block, int mode)
{
	mtd_sbl_dev_t *mtd_sbl_dev = dev->priv;
	sbl_dev_t *sbl_dev = mtd_sbl_dev->sbl_dev;
	struct mtd_info *nand_mtd = mtd_sbl_dev->nand_mtd;
	struct nand_chip *chip = nand_mtd->priv;
	unsigned char badblk_flag;
	loff_t addr;
	size_t retlen;

	addr = block * sbl_dev->pages_per_block * sbl_dev->data_area_size;

	nand_mtd->read_oob(nand_mtd, addr + chip->badblock_pos, 1, &retlen,
			   &badblk_flag);
	if (num_of_bit1_uchar(badblk_flag) < 7) {
		return 1;
	}

	return 0;
}

static int
mtd_sbl_dev_erase(sbl_dev_t *dev, unsigned long block, int mode)
{
	int res;
	mtd_sbl_dev_t *mtd_sbl_dev = dev->priv;
	sbl_dev_t *sbl_dev = mtd_sbl_dev->sbl_dev;
	struct mtd_info *nand_mtd = mtd_sbl_dev->nand_mtd;
	struct nand_chip *nand = (struct nand_chip *)nand_mtd->priv;
	struct erase_info instr;

	memset(&instr, 0, sizeof(instr));
	instr.mtd = nand_mtd;
	instr.addr = block * sbl_dev->data_area_size *
		sbl_dev->pages_per_block;
	instr.len = sbl_dev->data_area_size * sbl_dev->pages_per_block;
	res = nand_mtd->erase(nand_mtd, &instr);
	if (res < 0) {
		int retlen;
		unsigned char data = 0;
		nand_mtd->write_oob(nand_mtd, instr.addr + nand->badblock_pos,
				    sizeof(data), &retlen, &data);
		return -ESBLERASE;
	}

	return 0;
}
