/*
 * Copyright 2003 Sony Corporation.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "sbl.h"

/* ======================================================================= *
 * constant (internal)                                                     *
 * ======================================================================= */

#define SBL_INFO0_READ_RETRY	5

#define SBL_BYTEORDER_LITTLE	1234
#define SBL_BYTEORDER_BIG	4321

#ifndef SBL_BYTEORDER
#define SBL_BYTEORDER SBL_BYTEORDER_LITTLE
#endif

#ifndef NULL
#define NULL	0
#endif

/* ======================================================================= *
 * Debug print function from dpr package                                   *
 * ======================================================================= */

#define DPL_EMERG	0
#define DPL_ALERT	1
#define DPL_CRIT	2
#define DPL_ERR		3
#define DPL_WARNING	4
#define DPL_NOTICE	5
#define DPL_INFO	6
#define DPL_DEBUG	7

#define DPM_PORPOSE	0x0000000f
#define DPM_ERR		0x00000001
#define DPM_TRACE	0x00000002
#define DPM_LAYER	0x000000f0
#define DPM_APPLICATION	0x00000010
#define DPM_MIDDLEWARE	0x00000020
#define DPM_DRIVER	0x00000040
#define DPM_KERNEL	0x00000080
#define DPM_USERDEF	0xffff0000

#ifdef DEBUG

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL DPL_DEBUG
#endif
#ifndef DEBUG_MASK
#define DEBUG_MASK  DPM_PORPOSE
#endif

#ifndef DPR
#include <stdio.h>
#define DPR(fmt, args...)	fprintf(stderr, fmt, ## args)
#endif

#define DPRL(level, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && DPR(fmt, ## args))
#define DPRM(mask, fmt, args...) \
	(void)(((mask) & DEBUG_MASK) && DPR(fmt, ## args))
#define DPRLM(level, mask, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && ((mask) & DEBUG_MASK) && \
		DPR(fmt, ## args))
#define DPRC(cond, fmt, args...) \
	(void)((cond) && DPR(fmt, ## args))
#define DPRF(fmt, args...)	DPR(__FUNCTION__ ": " fmt, ## args)
#define DPRFL(level, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && DPRF(fmt, ## args))
#define DPRFM(mask, fmt, args...) \
	(void)(((mask) & DEBUG_MASK) && DPRF(fmt, ## args))
#define DPRFLM(level, mask, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && ((mask) & DEBUG_MASK) && \
		DPRF(fmt, ## args))
#define DPRFC(cond, fmt, args...) \
	(void)((cond) && DPRF(fmt, ## args))

#define	DPRFS	DPRFLM(DPL_DEBUG, DPM_TRACE, "start\n")
#define	DPRFE	DPRFLM(DPL_DEBUG, DPM_TRACE, "end\n")
#define DPRFR	DPRFLM(DPL_DEBUG, DPM_TRACE, "return\n")

#else  /* DEBUG */

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL -1
#endif
#ifndef DEBUG_MASK
#define DEBUG_MASK  0
#endif

#ifndef DPR
#define DPR(fmt, args...)
#endif

#define DPRL(level, fmt, args...)
#define DPRM(mask, fmt, args...)
#define DPRLM(level, mask, fmt, args...)
#define DPRC(cond, fmt, args...)
#define DPRF(fmt, args...)
#define DPRFL(level, fmt, args...)
#define DPRFM(mask, fmt, args...)
#define DPRFLM(level, mask, fmt, args...)
#define DPRFC(cond, fmt, args...)

#define	DPRFS
#define	DPRFE
#define DPRFR

#endif /* DEBUG */

/* ======================================================================= *
 * prototype declaration (internal)                                        *
 * ======================================================================= */

/* crc */
static void sbl_make_crc32_table(void);
static unsigned long sbl_crc32(unsigned long crc, unsigned char *buf,
			       unsigned int len);

/* sbl_info */
static int sbl_info0_read(sbl_dev_t *dev, unsigned long pblock,
			  unsigned long *lblock);
static int sbl_info_read(sbl_dev_t *dev, unsigned long pblock,
			 unsigned long *lblock);
static int sbl_info0_program(sbl_dev_t *dev, unsigned long pblock,
			     unsigned long lblock);
static int sbl_info1_program(sbl_dev_t *dev, unsigned long pblock,
			     int invalidate);
static int sbl_info_invalidate(sbl_dev_t *dev, unsigned long pblock);
static int sbl_info_invalidate_if_matched(sbl_dev_t *dev, unsigned long pblock,
					  unsigned long lblock);
static int sbl_info_lblock_is_matched(sbl_dev_t *dev, unsigned long pblock,
				      unsigned long lblock);
static int sbl_info0_magic_distance(unsigned short magic1,
				    unsigned short magic2);

/* erase */
static int sbl_erase_raw(sbl_dev_t *dev, unsigned long pblock);
static int sbl_erase_spare(sbl_dev_t *dev, unsigned long lblock);
static int sbl_erase_block(sbl_dev_t *dev, unsigned long pblock);
static int sbl_erase_and_write_sbl_info(sbl_dev_t *dev, unsigned long pblock,
					unsigned long lblock);
static int sbl_erase_success_or_invalidate(sbl_dev_t *dev,
					   unsigned long pblock,
					   unsigned long lblock);

/* program */
static int sbl_program_raw(sbl_dev_t *dev, unsigned long pblock, int page,
				int mode, const void *data);
static int sbl_program_spare(sbl_dev_t *dev,
				  unsigned long lblock, int page,
				  int mode, const void *data);
static int sbl_program_page_in_current_block(sbl_dev_t *dev,
					     unsigned long lblock,
					     int page, int mode,
					     const void *data,
					     unsigned long *pblock);
static int sbl_program_in_new_block(sbl_dev_t *dev, unsigned long lblock,
				    int page, int mode, const void *data,
				    unsigned long pblock_orig);
static int sbl_invalidate_lblocks(sbl_dev_t *dev, unsigned long lblock,
				  unsigned long pblock_excluded);
static int sbl_program_page(sbl_dev_t *dev, unsigned long pblock, int page,
			    int mode, const void *data);
static int sbl_program_with_copying(sbl_dev_t *dev,
				    unsigned long pblock, int page,
				    int mode, const void *data,
				    unsigned long pblock_orig,
				    unsigned long lblock);

/* copy/move */
static int sbl_move_to_new_block(sbl_dev_t *dev, unsigned long lblock,
				 int mode, unsigned long pblock_orig,
				 unsigned long *pblock_new);
static int sbl_move_to_normal_block(sbl_dev_t *dev, unsigned long lblock,
				    unsigned long pblock_orig, int mode);
static int sbl_copy_block(sbl_dev_t *dev, unsigned long pblock,
			  unsigned long pblock_orig,
			  int mode, unsigned long lblock);

/* read */
static int sbl_read_raw(sbl_dev_t *dev, unsigned long pblock, int page,
			int mode, void *data);
static int sbl_read_spare(sbl_dev_t *dev, unsigned long lblock, int page,
			  int mode, void *data);
static int sbl_read_page(sbl_dev_t *dev, unsigned long pblock, int page,
			 int mode, void *data);

/* verify */
static int sbl_is_same_data(const void *data1, const void *data2,
			    unsigned int data_size);
static int sbl_verify_raw(sbl_dev_t *dev, unsigned long pblock, int page,
			  int mode, const void *data);
static int sbl_verify_data(sbl_dev_t *dev, unsigned long pblock, int page,
			   const void *data);
static int sbl_verify_extra(sbl_dev_t *dev, unsigned long pblock, int page,
			    const void *extra);
static int sbl_verify_sbl_info(sbl_dev_t *dev, unsigned long pblock, int page,
			       const void *sbl_info);

/* misc */
static int sbl_find_invalid_from(sbl_dev_t *dev, unsigned long *pblock,
				 unsigned long pblock_excluded);
static int sbl_find_lblock_next(sbl_dev_t *dev, unsigned long lblock,
				unsigned long *pblock,
				unsigned long pblock_excluded);
static int sbl_data_is_all_ff(const void *data, unsigned int data_size);

/* ======================================================================= *
 * inline/macro function (internal)                                        *
 * ======================================================================= */

/* ------------------------ *
 * byte order               *
 * ------------------------ */

#if SBL_BYTEORDER == SBL_BYTEORDER_BIG
static inline unsigned long
sbl_cpu_to_le32(unsigned long x)
{
	return ((x >> 24) + ((x >> 8) & 0x0000ff00) +
		((x << 8) & 0x00ff0000) + ((x << 24) & 0xff000000));
}

static inline unsigned short
sbl_cpu_to_le16(unsigned short x)
{
	return ((x >> 8) + ((x << 8) & 0xff00));
}
#else /* SBL_BYTEORDER */
#define sbl_cpu_to_le32(x)	(x)
#define sbl_cpu_to_le16(x)	(x)
#endif /* SBL_BYTEORDER */

#define sbl_le32_to_cpu(x)	sbl_cpu_to_le32(x)
#define sbl_le16_to_cpu(x)	sbl_cpu_to_le16(x)

/* ------------------------ *
 * sbl info                 *
 * ------------------------ */

static inline unsigned long
sbl_info0_lblock(void *sbl_info0)
{
	return sbl_le32_to_cpu(*((unsigned long *)sbl_info0));
}

static inline unsigned short
sbl_info0_magic(void *sbl_info0)
{
	return sbl_le16_to_cpu(*((unsigned short *)(sbl_info0 + 4)));
}

static inline unsigned char
sbl_info0_crc(void *sbl_info0)
{
	return *((unsigned char *)(sbl_info0 + 6));
}

static inline unsigned char
sbl_info1_valid_flag(void *sbl_info1)
{
	return *((unsigned char *)(sbl_info1 + 7));
}

static inline int
sbl_info0_crc_is_correct(void *sbl_info0)
{
	return ((unsigned char)sbl_crc32(0, sbl_info0, 6) ==
		sbl_info0_crc(sbl_info0));
}

static inline int
sbl_info1_block_is_valid(void *sbl_info1)
{
	unsigned char valid_flag = sbl_info1_valid_flag(sbl_info1);
	return (valid_flag >= 0xfd || valid_flag == 0xfb ||
		valid_flag == 0xf7 || valid_flag == 0xef ||
		valid_flag == 0xdf || valid_flag == 0xbf ||
		valid_flag == 0x7f);
}

static inline void
sbl_info0_set(void *sbl_info0, unsigned long lblock)
{
	*((unsigned long *)sbl_info0)		= sbl_cpu_to_le32(lblock);
	*((unsigned short *)(sbl_info0 + 4))	= sbl_cpu_to_le16(SBL_MAGIC);
	*((unsigned char *)(sbl_info0 + 6))	= sbl_crc32(0, sbl_info0, 6);
	*((unsigned char *)(sbl_info0 + 7))	= 0xff;
}

static inline void
sbl_info1_set(void *sbl_info1, unsigned char valid_flag)
{
	*((unsigned long *)sbl_info1)		= 0xffffffff;
	*((unsigned short *)(sbl_info1 + 4))	= 0xffff;
	*((unsigned char *)(sbl_info1 + 6))	= 0xff;
	*((unsigned char *)(sbl_info1 + 7))	= valid_flag;
}

/* ------------------------ *
 * wrapper                  *
 * ------------------------ */

static inline int
sbl_dev_is_bad(sbl_dev_t *dev, unsigned long block, int mode)
{
	int res;
	unsigned long *cache = dev->sbl_info_cache;

	if (cache != NULL && cache[block] == SBL_SLBN_BADBLK) {
		return 1;
	}
	res = dev->is_bad(dev, block, mode);
	if (cache != NULL && res == 1) {
		cache[block] = SBL_SLBN_BADBLK;
	}

#ifdef DEBUG
	if (res != 0 && res != 1 && res != -ESBLIO) {
		DPRFL(DPL_WARNING, "Invalid error code (%d)\n", -res);
		res = -ESBLIO;
	}
#endif /* DEBUG */

	return res;
}

static inline int
sbl_dev_erase(sbl_dev_t *dev, unsigned long block, int mode)
{
	int res;
	unsigned long *cache = dev->sbl_info_cache;

	res = dev->erase(dev, block, mode);
	if (cache != NULL && res == -ESBLBADBLK) {
		cache[block] = SBL_SLBN_BADBLK;
	}

#ifdef DEBUG
	if (res != 0 && res != -ESBLIO && res != -ESBLERASE) {
		DPRFL(DPL_WARNING, "Invalid error code (%d)\n", -res);
		res = -ESBLIO;
	}
#endif /* DEBUG */

	return res;
}

static inline int
sbl_dev_program(sbl_dev_t *dev, unsigned long block, int page,
		int mode, const void *data, const void *extra)
{
	int res;
	unsigned long *cache = dev->sbl_info_cache;

	if (cache != NULL && cache[block] == SBL_SLBN_BADBLK) {
		return -ESBLBADBLK;
	}
	res = dev->program(dev, block, page, mode, data, extra);
	if (cache != NULL && res == -ESBLBADBLK) {
		cache[block] = SBL_SLBN_BADBLK;
	}

#ifdef DEBUG
	if (res != 0 && res != -ESBLIO && res != -ESBLPROGRAM &&
	    res != -ESBLBADBLK && res != -ESBLVERIFY) {
		DPRFL(DPL_WARNING, "Invalid error code (%d)\n", -res);
		res = -ESBLIO;
	}
#endif /* DEBUG */

	return res;
}

static inline int
sbl_dev_read(sbl_dev_t *dev, unsigned long block, int page,
	     int mode, void *data, void *extra)
{
	int res;
	unsigned long *cache = dev->sbl_info_cache;

	if (cache != NULL && cache[block] == SBL_SLBN_BADBLK) {
		return -ESBLBADBLK;
	}
	res = dev->read(dev, block, page, mode, data, extra);
	if (cache != NULL && res == -ESBLBADBLK) {
		cache[block] = SBL_SLBN_BADBLK;
	}

#ifdef DEBUG
	if (res != 0 && res != -ESBLIO && res != -ESBLECC &&
	    res != -ESBLBADBLK) {
		DPRFL(DPL_WARNING, "Invalid error code (%d)\n", -res);
		res = -ESBLIO;
	}
#endif /* DEBUG */

	return res;
}

/* ======================================================================= *
 * SBL library interface                                                   *
 * ======================================================================= */

int
sbl_dev_init(sbl_dev_t *dev)
{
	DPRFS;

	if (dev->normal_blocks == 0 ||
	    dev->normal_blocks > SBL_SLBN_BLOCK_MAX ||
	    dev->normal_blocks + dev->spare_blocks > SBL_SLBN_BLOCK_MAX ||
	    dev->normal_blocks + dev->spare_blocks < dev->normal_blocks ||
	    dev->pages_per_block < 2 ||
	    dev->page_buf == NULL ||
	    dev->copy_page_buf == NULL) {
		DPRFR;
		return -ESBLINVAL;
	}

	/* initialize sbl_info_cache */
	if (dev->sbl_info_cache != NULL) {
		unsigned long n;
		unsigned long *p = dev->sbl_info_cache;
		unsigned long nblocks = dev->normal_blocks + dev->spare_blocks;

		for (n = 0; n < nblocks; n++) {
			p[n] = SBL_SLBN_NOTCACHED;
		}
	}

	DPRFE;
	return 0;
}

int
sbl_is_bad(sbl_dev_t *dev, unsigned long block, int mode)
{
	int res;
	DPRFS;

	if (block >= dev->normal_blocks + dev->spare_blocks) {
		DPRFL(DPL_NOTICE, "invalid block number %ld\n", block);
		return -ESBLINVAL;
	}
	if (mode != SBL_M_RAW) {
		DPRFL(DPL_NOTICE, "invalid mode %d\n", mode);
		return -ESBLINVAL;
	}

	res = sbl_dev_is_bad(dev, block, mode);

	DPRFE;
	return res;
}

int
sbl_erase(sbl_dev_t *dev, unsigned long block, int mode)
{
	int res;
	DPRFS;

	switch (mode) {
	case SBL_M_RAW:
		res = sbl_erase_raw(dev, block);
		break;
	case SBL_M_SPARE:
		res = sbl_erase_spare(dev, block);
		break;
	default:
		DPRFL(DPL_NOTICE, "invalid mode %d\n", mode);
		return -ESBLINVAL;
	}

	DPRFE;
	return res;
}

int
sbl_program(sbl_dev_t *dev, unsigned long block, int page, int mode,
	    const void *data)
{
	int res;
	DPRFS;

	switch (mode & SBL_M_ACCESS_MODE_MASK) {
	case SBL_M_RAW:
		res = sbl_program_raw(dev, block, page, mode, data);
		break;
	case SBL_M_SPARE:
		res = sbl_program_spare(dev, block, page, mode, data);
		break;
	default:
		DPRFL(DPL_NOTICE, "invalid mode %d\n", mode);
		return -ESBLINVAL;
	}

	DPRFE;
	return res;
}

int
sbl_read(sbl_dev_t *dev, unsigned long block, int page, int mode, void *data)
{
	int res;
	DPRFS;

	switch (mode & SBL_M_ACCESS_MODE_MASK) {
	case SBL_M_RAW:
		res = sbl_read_raw(dev, block, page, mode, data);
		break;
	case SBL_M_SPARE:
		res = sbl_read_spare(dev, block, page, mode, data);
		break;
	default:
		DPRFL(DPL_NOTICE, "invalid mode %d\n", mode);
		return -ESBLINVAL;
	}

	DPRFE;
	return res;
}

int
sbl_scan(sbl_dev_t *dev, sbl_scan_t *info)
{
	int res;
	unsigned long n;
	unsigned long pblock;
	unsigned long lblock;
	DPRFS;

	/* calculate used_normal_blocks */
	for (n = 0; n < dev->normal_blocks; n++) {
		pblock = dev->normal_blocks - n - 1;
		res = sbl_info_lblock_is_matched(dev, pblock, pblock);
		if (res == 1) {
			break;
		}
		if (res == -ESBLIO) {
			DPRFR;
			return res;
		}
	}
	info->used_normal_blocks = dev->normal_blocks - n;

	/* calculate used_spare_blocks */
	for (n = 0; n < dev->spare_blocks; n++) {
		pblock = dev->normal_blocks + n;
		res = sbl_info_read(dev, pblock, &lblock);
		if (res == 0 && lblock <= SBL_SLBN_BLOCK_MAX) {
			break;
		}
		if (res == -ESBLIO) {
			DPRFR;
			return res;
		}
	}
	info->used_spare_blocks = dev->spare_blocks - n;

	DPRFE;
	return 0;
}

/* ======================================================================= *
 * internal function                                                       *
 * ======================================================================= */

/* ------------------------ *
 * CRC32                    *
 * ------------------------ */

static unsigned long sbl_crc32_table[256];
static int sbl_crc32_table_computed = 0;

static void
sbl_make_crc32_table(void)
{
	unsigned long c;
	int i, j;

	for (i = 0; i < 256; i++) {
		c = i;
		for (j = 0; j < 8; j++) {
			if (c & 1) {
				c = (0xedb88320 ^ (c >> 1));
			} else {
				c = (c >> 1);
			}
		}
		sbl_crc32_table[i] = c;
	}
	sbl_crc32_table_computed = 1;
}

static unsigned long
sbl_crc32(unsigned long crc, unsigned char *buf, unsigned int len)
{
	int i;

	if (!sbl_crc32_table_computed) {
		sbl_make_crc32_table();
	}

	crc ^= 0xffffffff;
	for (i = 0; i < len; i++) {
		crc = (sbl_crc32_table[(crc ^ buf[i]) & 0xff] ^ (crc >> 8));
	}
	return (crc ^ 0xffffffff);
}

/* ------------------------ *
 * sbl info                 *
 * ------------------------ */

/* possible errors: ESBLIO */
static int
sbl_info0_read(sbl_dev_t *dev, unsigned long pblock, unsigned long *lblock)
{
	int res;
	char sbl_info[SBL_INFO_SIZE];
	int i;
	unsigned short magic;
	DPRFS;

	for (i = 0; i < SBL_INFO0_READ_RETRY; i++) {
		res = sbl_dev_read(dev, pblock, 0,
				   SBL_M_RAW | SBL_M_SBLINFO, NULL, sbl_info);
		if (res == -ESBLIO) {
			DPRFR;
			return res;
		}
		if (res == -ESBLBADBLK) {
			*lblock = SBL_SLBN_BADBLK;
			goto ret;
		}
		magic = sbl_info0_magic(sbl_info);
		if (magic == SBL_MAGIC && sbl_info0_crc_is_correct(sbl_info)) {
			*lblock = sbl_info0_lblock(sbl_info);
			goto ret;
		}
		if (sbl_info0_magic_distance(magic, SBL_MAGIC) > 2) {
			break;
		}
	}
	*lblock = SBL_SLBN_INVALID;
ret:
	DPRFE;
	return 0;
}

/* possible errors: ESBLIO */
static int
sbl_info_read(sbl_dev_t *dev, unsigned long pblock, unsigned long *lblock)
{
	int res;
	unsigned long *cache = dev->sbl_info_cache;
	char sbl_info[SBL_INFO_SIZE];
	DPRFS;

	if (cache != NULL && cache[pblock] != SBL_SLBN_NOTCACHED) {
		*lblock = cache[pblock];
		DPRFR;
		return 0;
	}

	/* page 0 */
	res = sbl_info0_read(dev, pblock, lblock);
	if (res == -ESBLIO) {
		DPRFR;
		return res;
	}
	if (*lblock == SBL_SLBN_BADBLK || *lblock == SBL_SLBN_INVALID) {
		goto cache_and_return;
	}

	/* page 1 */
	res = sbl_dev_read(dev, pblock, 1,
			   SBL_M_RAW | SBL_M_SBLINFO, NULL, sbl_info);
	if (res == -ESBLIO) {
		DPRFR;
		return res;
	}
	if (res < 0) {
		*lblock = SBL_SLBN_BADBLK;
		goto cache_and_return;
	}
	if (!sbl_info1_block_is_valid(sbl_info)) {
		*lblock = SBL_SLBN_INVALID;
	}

cache_and_return:
	if (cache != NULL) {
		cache[pblock] = *lblock;
	}

	DPRFE;
	return 0;
}

/* possible errors: ESBLIO, ESBLBADBLK, ESBLPROGRAM, ESBLVERIFY, ESBLINFO */
static int
sbl_info0_program(sbl_dev_t *dev, unsigned long pblock, unsigned long lblock)
{
	int res, res2;
	unsigned long *cache = dev->sbl_info_cache;
	char sbl_info[SBL_INFO_SIZE];
	DPRFS;

	if (cache != NULL) {
		cache[pblock] = SBL_SLBN_NOTCACHED;
	}

	sbl_info0_set(sbl_info, lblock);
	res = sbl_dev_program(dev, pblock, 0,
			      SBL_M_RAW | SBL_M_SBLINFO, NULL, sbl_info);
	if (res < 0) {
		goto err_after_program;
	}

	res = sbl_verify_sbl_info(dev, pblock, 0, sbl_info);
	if (res < 0) {
		goto err_after_program;
	}

	DPRFE;
	return 0;

err_after_program:
	res2 = sbl_info_invalidate_if_matched(dev, pblock, lblock);
	if (res2 < 0) {
		DPRFR;
		return res2;
	}
	DPRFR;
	return res;
}

/* possible errors: ESBLIO, ESBLBADBLK, ESBLPROGRAM, ESBLVERIFY */
static int
sbl_info1_program(sbl_dev_t *dev, unsigned long pblock, int invalidate)
{
	int res;
	unsigned long *cache = dev->sbl_info_cache;
	unsigned char valid_flag;
	char sbl_info[SBL_INFO_SIZE];
	DPRFS;

	if (cache != NULL) {
		cache[pblock] = SBL_SLBN_NOTCACHED;
	}

	if (invalidate) {
		valid_flag = 0x00;
	} else {
		valid_flag = 0xff;
	}
	sbl_info1_set(sbl_info, valid_flag);
	res = sbl_dev_program(dev, pblock, 1,
			      SBL_M_RAW | SBL_M_SBLINFO, NULL, sbl_info);
	if (res < 0) {
		DPRFR;
		return res;
	}

	res = sbl_verify_sbl_info(dev, pblock, 1, sbl_info);
	if (res < 0) {
		DPRFR;
		return res;
	}

	DPRFE;
	return 0;
}

/* possible errors: ESBLIO, ESBLINFO */
static int
sbl_info_invalidate(sbl_dev_t *dev, unsigned long pblock)
{
	int res;
	unsigned long *cache = dev->sbl_info_cache;
	DPRFS;

	if (cache != NULL) {
		cache[pblock] = SBL_SLBN_NOTCACHED;
	}

	res = sbl_info1_program(dev, pblock, 1);
	if (res == -ESBLIO) {
		DPRFR;
		return res;
	}
	if (res < 0) {
		res = sbl_erase_block(dev, pblock);
		if (res == -ESBLIO) {
			DPRFR;
			return res;
		}
		if (res < 0) {
			unsigned long lblock;
			res = sbl_info_read(dev, pblock, &lblock);
			if (res == -ESBLIO) {
				DPRFR;
				return res;
			}
			if (lblock != SBL_SLBN_BADBLK &&
			    lblock != SBL_SLBN_INVALID) {
				DPRFR;
				return -ESBLINFO;
			}
		}
	} else {
		if (cache != NULL) {
			cache[pblock] = SBL_SLBN_INVALID;
		}
	}

	DPRFE;
	return 0;
}

/* possible errors: ESBLIO, ESBLINFO */
static int
sbl_info_invalidate_if_matched(sbl_dev_t *dev, unsigned long pblock,
			       unsigned long lblock)
{
	int res;
	unsigned long *cache = dev->sbl_info_cache;
	DPRFS;

	if (cache != NULL) {
		cache[pblock] = SBL_SLBN_NOTCACHED;
	}

	res = sbl_info_lblock_is_matched(dev, pblock, lblock);
	if (res < 0) {
		DPRFR;
		return res;
	}
	if (res == 1) {
		res = sbl_info_invalidate(dev, pblock);
		if (res < 0) {
			DPRFR;
			return res;
		}
	}

	DPRFE;
	return 0;
}

/* possible errors: ESBLIO */
static int
sbl_info_lblock_is_matched(sbl_dev_t *dev, unsigned long pblock,
			   unsigned long lblock)
{
	int res;
	unsigned long info_lblock;
	DPRFS;

	res = sbl_info_read(dev, pblock, &info_lblock);
	if (res < 0) {
		DPRFR;
		return res;
	}

	DPRFE;
	return (info_lblock == lblock);
}

static int
sbl_info0_magic_distance(unsigned short magic1, unsigned short magic2)
{
	unsigned short magic_xor;
	int distance;
	int i;
	DPRFS;

	magic_xor = (magic1 ^ magic2);
	distance = 0;
	for (i = 0; i < 16; i++) {
		distance += (magic_xor & 1);
		magic_xor >>= 1;
	}

	DPRFE;
	return distance;
}

/* ------------------------ *
 * erase                    *
 * ------------------------ */

/* possible errors: ESBLIO, ESBLERASE, ESBLINVAL */
static int
sbl_erase_raw(sbl_dev_t *dev, unsigned long pblock)
{
	int res;
	DPRFS;

	/* check arguments */
	if (pblock >= dev->normal_blocks + dev->spare_blocks) {
		DPRFL(DPL_NOTICE, "invalid block number %ld\n", pblock);
		return -ESBLINVAL;
	}

	/* erase */
	res = sbl_erase_block(dev, pblock);

	DPRFE;
	return res;
}

/* possible errors: ESBLIO, ESBLINVAL, ESBLNOSPC, ESBLINFO */
static int
sbl_erase_spare(sbl_dev_t *dev, unsigned long lblock)
{
	int res;
	unsigned long pblock;
	unsigned long n;
	DPRFS;

	/* check arguments */
	if (lblock >= dev->normal_blocks) {
		DPRFL(DPL_NOTICE, "invalid block number %ld\n", lblock);
		return -ESBLINVAL;
	}

	/* try to erase the normal block */
	pblock = lblock;
	res = sbl_dev_is_bad(dev, pblock, SBL_M_RAW);
	if (res == -ESBLIO) {
		DPRFR;
		return res;
	}
	if (res == 0) {
		res = sbl_erase_success_or_invalidate(dev, pblock, lblock);
		if (res == -ESBLIO || res == -ESBLINFO) {
			DPRFR;
			return res;
		}
		if (res == 0) {
			DPRFR;
			return 0;
		}
	}

	/* try to erase the spare block */
	for (n = 0; n < dev->spare_blocks; n++) {
		pblock = dev->normal_blocks + dev->spare_blocks - n - 1;

		res = sbl_info_lblock_is_matched(dev, pblock, lblock);
		if (res == -ESBLIO) {
			DPRFR;
			return res;
		}
		if (res == 1) {
			res = sbl_erase_success_or_invalidate(dev,
							      pblock, lblock);
			if (res == -ESBLIO || res == -ESBLINFO) {
				DPRFR;
				return res;
			}
			if (res == 0) {
				DPRFR;
				return 0;
			}
		}
	}

	/* try to allocate new spare block */
	pblock = dev->normal_blocks + dev->spare_blocks - 1;
	while (pblock >= dev->normal_blocks) {
		res = sbl_find_invalid_from(dev, &pblock, SBL_SLBN_INVALID);
		if (res < 0) {
			DPRFR;
			return res;

		}

		res = sbl_erase_success_or_invalidate(dev, pblock, lblock);
		if (res == -ESBLIO || res == -ESBLINFO) {
			DPRFR;
			return res;
		}
		if (res == 0) {
			DPRFR;
			return 0;
		}
		pblock--;
	}

	DPRFE;
	return -ESBLNOSPC;
}

/* possible errors: ESBLIO, ESBLERASE */
static int
sbl_erase_block(sbl_dev_t *dev, unsigned long pblock)
{
	int res;
	unsigned long *cache = dev->sbl_info_cache;
	DPRFS;

	if (cache != NULL) {
		cache[pblock] = SBL_SLBN_NOTCACHED;
	}

	res = sbl_dev_erase(dev, pblock, SBL_M_RAW);
	if (res < 0) {
		DPRFR;
		return res;
	}

	DPRFE;
	return 0;
}

/* possible errors: ESBLIO, ESBLERASE, ESBLINFO */
static int
sbl_erase_and_write_sbl_info(sbl_dev_t *dev, unsigned long pblock,
			     unsigned long lblock)
{
	int res;
	DPRFS;

	res = sbl_erase_block(dev, pblock);
	if (res < 0) {
		DPRFR;
		return res;
	}
	res = sbl_info0_program(dev, pblock, lblock);
	if (res == -ESBLIO || res == -ESBLINFO) {
		DPRFR;
		return res;
	}
	if (res < 0) {
		DPRFR;
		return -ESBLERASE;
	}

	DPRFE;
	return 0;
}

/* possible errors: ESBLIO, ESBLINFO, ESBLERASE */
static int
sbl_erase_success_or_invalidate(sbl_dev_t *dev, unsigned long pblock,
				unsigned long lblock)
{
	int res;
	DPRFS;

	res = sbl_erase_and_write_sbl_info(dev, pblock, lblock);
	if (res == -ESBLIO || res == -ESBLINFO) {
		DPRFR;
		return res;
	}
	if (res < 0) {
		/* invalidate the block */
		res = sbl_info_lblock_is_matched(dev, pblock, lblock);
		if (res == -ESBLIO) {
			DPRFR;
			return res;
		}
		if (res == 1) {
			res = sbl_info_invalidate(dev, pblock);
			if (res < 0) {
				DPRFR;
				return res;
			}
		}

		DPRFR;
		return -ESBLERASE;
	}

	/* invalidate other blocks */
	res = sbl_invalidate_lblocks(dev, lblock, pblock);
	if (res < 0) {
		DPRFR;
		return res;
	}

	DPRFE;
	return 0;
}

/* ------------------------ *
 * program                  *
 * ------------------------ */

/* possible errors: ESBLIO, ESBLBADBLK, ESBLPROGRAM, ESBLVERIFY, ESBLINVAL */
static int
sbl_program_raw(sbl_dev_t *dev, unsigned long pblock, int page,
		int mode, const void *data)
{
	int res;
	unsigned long *cache = dev->sbl_info_cache;
	DPRFS;

	/* check arguments */
	if (pblock >= dev->normal_blocks + dev->spare_blocks) {
		DPRFL(DPL_NOTICE, "invalid block number %ld\n", pblock);
		return -ESBLINVAL;
	}
	if (page < 0 || page >= dev->pages_per_block) {
		DPRFL(DPL_NOTICE, "invalid page %d\n", page);
		return -ESBLINVAL;
	}
	if ((mode & SBL_M_EXTRA) && (mode & SBL_M_ECC)) {
		DPRFL(DPL_NOTICE, "invalid mode 0x%08x\n", mode);
		return -ESBLINVAL;
	}
	if (cache != NULL && cache[pblock] == SBL_SLBN_BADBLK) {
		DPRFR;
		return -ESBLBADBLK;
	}

	if (!(mode & SBL_M_DATA) && !(mode & SBL_M_EXTRA)) {
		mode |= SBL_M_DATA;
	}

	res = sbl_program_page(dev, pblock, page, mode, data);

	DPRFE;
	return res;
}

/* possible errors: ESBLIO, ESBLINVAL, ESBLPROGRAM, ESBLNOSPC,
                    ESBLECC, ESBLINFO */
static int
sbl_program_spare(sbl_dev_t *dev, unsigned long lblock, int page,
		  int mode, const void *data)
{
	int res;
	unsigned long pblock;
	DPRFS;

	/* check arguments */
	if (lblock >= dev->normal_blocks) {
		DPRFL(DPL_NOTICE, "invalid block number %ld\n", lblock);
		return -ESBLINVAL;
	}
	if (page < -1 || page >= dev->pages_per_block) {
		DPRFL(DPL_NOTICE, "invalid page %d\n", page);
		return -ESBLINVAL;
	}
	if (mode & SBL_M_EXTRA) {
		DPRFL(DPL_NOTICE, "invalid mode 0x%08x\n", mode);
		return -ESBLINVAL;
	}

	/* try to program a page/block on the current block */
	res = sbl_program_page_in_current_block(dev, lblock, page, mode, data,
						&pblock);
	if (res == 0) {
		DPRFE;
		return 0;
	}
	if (res == -ESBLIO || res == -ESBLINFO) {
		DPRFR;
		return res;
	}
	if (res == -ESBLNOENT || res == -ESBLBADBLK) {
		pblock = SBL_SLBN_INVALID;
	}

	/* try to program a page/block on new block */
	res = sbl_program_in_new_block(dev, lblock, page, mode, data, pblock);
	if (res < 0) {
		DPRFR;
		return res;
	}

	DPRFE;
	return 0;
}

/* possible errors: ESBLIO, ESBLNOENT, ESBLBADBLK, ESBLPROGRAM, ESBLVERIFY,
                    ESBLINFO */
static int
sbl_program_page_in_current_block(sbl_dev_t *dev, unsigned long lblock,
				  int page, int mode, const void *data,
				  unsigned long *pblock)
{
	int res;
	int i;
	int m_verify;
	DPRFS;

	m_verify = (mode & SBL_M_VERIFY);

	/* find the block */
	*pblock = SBL_SLBN_INVALID;
	res = sbl_find_lblock_next(dev, lblock, pblock, SBL_SLBN_INVALID);
	if (res < 0) {
		DPRFR;
		return res;
	}

	if (page == -1) {
		/* program the whole of a block */
		for (i = 0; i < dev->pages_per_block; i++) {
			res = sbl_program_page(dev, *pblock, i,
					       SBL_M_RAW | SBL_M_DATA |
					       SBL_M_ECC | m_verify,
					       data + dev->data_area_size * i);
			if (res < 0) {
				DPRFR;
				return res;
			}
		}
	} else {
		/* program a page */
		res = sbl_program_page(dev, *pblock, page,
				       SBL_M_RAW | SBL_M_DATA | SBL_M_ECC |
				       m_verify, data);
		if (res < 0) {
			DPRFR;
			return res;
		}
	}

	DPRFE;
	return 0;
}

/* possible errors: ESBLIO, ESBLNOSPC, ESBLECC, ESBLINFO */
static int
sbl_program_in_new_block(sbl_dev_t *dev, unsigned long lblock,
			 int page, int mode, const void *data,
			 unsigned long pblock_orig)
{
	int res;
	unsigned long pblock;
	DPRFS;

	pblock = SBL_SLBN_INVALID;
	while (1) {
		if (pblock == SBL_SLBN_INVALID) {
			pblock = lblock;
		} else {
			if (pblock == lblock) {
				pblock = dev->normal_blocks +
					dev->spare_blocks - 1;
			} else {
				pblock--;
			}
			if (pblock < dev->normal_blocks) {
				DPRFR;
				return -ESBLNOSPC;
			}
		}

		/* find an invalid block */
		res = sbl_find_invalid_from(dev, &pblock, pblock_orig);
		if (res < 0) {
			DPRFR;
			return res;
		}

		/* erase the block */
		res = sbl_erase_block(dev, pblock);
		if (res == -ESBLIO) {
			DPRFR;
			return res;
		}
		if (res < 0) {
			continue;
		}

		/* program the page/block */
		res = sbl_program_with_copying(dev, pblock, page, mode, data,
					       pblock_orig, lblock);
		if (res == -ESBLIO || res == -ESBLECC || res == -ESBLINFO) {
			DPRFR;
			return res;
		}
		if (res < 0) {
			continue;
		}

		/* write sbl_info */
		res = sbl_info0_program(dev, pblock, lblock);
		if (res == -ESBLIO || res == -ESBLINFO) {
			DPRFR;
			return res;
		}
		if (res < 0) {
			continue;
		}
		break;
	}

	/* invalidate other blocks */
	res = sbl_invalidate_lblocks(dev, lblock, pblock);
	if (res < 0) {
		DPRFR;
		return res;
	}

	DPRFE;
	return 0;
}

/* possible errors: ESBIO, ESBLINFO */
static int
sbl_invalidate_lblocks(sbl_dev_t *dev, unsigned long lblock,
		       unsigned long pblock_excluded)
{
	int res;
	unsigned long pblock;
	DPRFS;

	pblock = SBL_SLBN_INVALID;
	while (1) {
		res = sbl_find_lblock_next(dev, lblock, &pblock,
					   pblock_excluded);
		if (res == -ESBLNOENT) {
			break;
		}
		if (res < 0) {
			DPRFR;
			return res;
		}

		res = sbl_info_invalidate(dev, pblock);
		if (res == -ESBLIO || res == -ESBLINFO) {
			DPRFR;
			return res;
		}
	}

	DPRFE;
	return 0;
}

/* possible errors: ESBLIO, ESBLBADBLK, ESBLPROGRAM, ESBLVERIFY */
static int
sbl_program_page(sbl_dev_t *dev, unsigned long pblock, int page,
		 int mode, const void *data)
{
	int res;
	unsigned long *cache = dev->sbl_info_cache;
	int m_ecc;
	DPRFS;

	if (cache != NULL && cache[pblock] == SBL_SLBN_BADBLK) {
		DPRFR;
		return -ESBLBADBLK;
	}

	m_ecc = (mode & SBL_M_ECC);

	if ((mode & SBL_M_DATA) && (mode & SBL_M_EXTRA)) {
		res = sbl_dev_program(dev, pblock, page,
				      SBL_M_RAW | SBL_M_DATA | SBL_M_EXTRA |
				      m_ecc,
				      data, data + dev->data_area_size);
	} else if (mode & SBL_M_EXTRA) {
		res = sbl_dev_program(dev, pblock, page,
				      SBL_M_RAW | SBL_M_EXTRA | m_ecc,
				      NULL, data);
	} else {
		res = sbl_dev_program(dev, pblock, page,
				      SBL_M_RAW | SBL_M_DATA | m_ecc,
				      data, NULL);
	}
	if (res < 0) {
		DPRFR;
		return res;
	}

	if (mode & SBL_M_VERIFY) {
		res = sbl_verify_raw(dev, pblock, page, mode, data);
	}

	DPRFE;
	return res;
}

/* possible errors: ESBLIO, ESBLBADBLK, ESBLPROGRAM, ESBLVERIFY, ESBLECC,
                    ESBLINFO */
static int
sbl_program_with_copying(sbl_dev_t *dev, unsigned long pblock, int page,
			 int mode, const void *data,
			 unsigned long pblock_orig, unsigned long lblock)
{
	int res;
	int i;
	const void *program_data = NULL;
	int m_ecc;
	int m_verify;
	DPRFS;

	m_ecc = (mode & SBL_M_ECC);
	m_verify = (mode & SBL_M_VERIFY);

	if (page == -1) {
		/* program the whole of a block */
		for (i = 0; i < dev->pages_per_block; i++) {
			res = sbl_program_page(dev, pblock, i,
					       SBL_M_RAW | SBL_M_DATA |
					       SBL_M_ECC | m_verify,
					       data + dev->data_area_size * i);
			if (res < 0) {
				DPRFR;
				return res;
			}
		}
	} else if (pblock_orig == SBL_SLBN_INVALID) {
		/* program a page normally */
		res = sbl_program_page(dev, pblock, page,
				       SBL_M_RAW | SBL_M_DATA | SBL_M_ECC |
				       m_verify, data);
		if (res < 0) {
			DPRFR;
			return res;
		}
	} else {
		/* program with copying */
		for (i = 0; i < dev->pages_per_block; i++) {
			if (i == page) {
				program_data = data;
			} else {
				res = sbl_dev_read(dev, pblock_orig, i,
						   SBL_M_RAW | SBL_M_DATA |
						   m_ecc,
						   dev->copy_page_buf, NULL);
				if (res < 0) {
					DPRFR;
					return res;
				}
				program_data = dev->copy_page_buf;
			}
			if (sbl_data_is_all_ff(program_data,
					       dev->data_area_size)) {
				continue;
			}
			res = sbl_program_page(dev, pblock, i,
					       SBL_M_RAW | SBL_M_DATA |
					       SBL_M_ECC | m_verify,
					       program_data);
			if (res < 0) {
				DPRFR;
				return res;
			}
		}
	}

	DPRFE;
	return 0;
}

/* ------------------------ *
 * copy/move                *
 * ------------------------ */

/* possible errors: ESBLIO, ESBLNOSPC, ESBLECC, ESBLINFO */
static int
sbl_move_to_new_block(sbl_dev_t *dev, unsigned long lblock,
		      int mode, unsigned long pblock_orig,
		      unsigned long *pblock_new)
{
	int res;
	unsigned long pblock;
	DPRFS;

	pblock = SBL_SLBN_INVALID;
	while (1) {
		if (pblock == SBL_SLBN_INVALID) {
			pblock = lblock;
		} else {
			if (pblock == lblock) {
				pblock = dev->normal_blocks +
					dev->spare_blocks - 1;
			} else {
				pblock--;
			}
			if (pblock < dev->normal_blocks) {
				DPRFR;
				return -ESBLNOSPC;
			}
		}

		/* find an invalid block */
		res = sbl_find_invalid_from(dev, &pblock, pblock_orig);
		if (res < 0) {
			DPRFR;
			return res;
		}

		/* erase the block */
		res = sbl_erase_block(dev, pblock);
		if (res == -ESBLIO) {
			DPRFR;
			return res;
		}
		if (res < 0) {
			continue;
		}

		/* copy the block */
		res = sbl_copy_block(dev, pblock, pblock_orig, mode, lblock);
		if (res == -ESBLIO || res == -ESBLECC || res == -ESBLINFO) {
			DPRFR;
			return res;
		}
		if (res < 0) {
			continue;
		}
		break;
	}

	/* invalidate other blocks */
	res = sbl_invalidate_lblocks(dev, lblock, pblock);
	if (res < 0) {
		DPRFR;
		return res;
	}

	if (pblock_new != NULL) {
		*pblock_new = pblock;
	}

	DPRFE;
	return 0;
}

/* possible errors: ESBLIO, ESBLNOSPC, ESBLERASE, ESBLBADBLK, ESBLPROGRAM,
                    ESBLVERIFY, ESBLECC, ESBLINFO */
static int
sbl_move_to_normal_block(sbl_dev_t *dev, unsigned long lblock,
			 unsigned long pblock_orig, int mode)
{
	int res;
	unsigned long info_lblock;
	DPRFS;

	if (pblock_orig == lblock) {
		DPRFR;
		return 0;
	}

	res = sbl_info_read(dev, lblock, &info_lblock);
	if (res < 0) {
		DPRFR;
		return res;
	}
	if (info_lblock != SBL_SLBN_INVALID) {
		DPRFR;
		return -ESBLNOSPC;
	}

	res = sbl_dev_is_bad(dev, lblock, SBL_M_RAW);
	if (res == -ESBLIO) {
		DPRFR;
		return res;
	} else if (res) {
		DPRFR;
		return -ESBLNOSPC;
	}

	/* erase the block */
	res = sbl_erase_block(dev, lblock);
	if (res < 0) {
		DPRFR;
		return res;
	}

	/* copy the block */
	res = sbl_copy_block(dev, lblock, pblock_orig, mode, lblock);
	if (res < 0) {
		DPRFR;
		return res;
	}

	/* invalidate other blocks */
	res = sbl_invalidate_lblocks(dev, lblock, lblock);
	if (res < 0) {
		DPRFR;
		return res;
	}

	DPRFE;
	return 0;
}

/* possible errors: ESBLIO, ESBLBADBLK, ESBLPROGRAM, ESBLVERIFY, ESBLECC,
		    ESBLINFO */
static int
sbl_copy_block(sbl_dev_t *dev, unsigned long pblock, unsigned long pblock_orig,
	       int mode, unsigned long lblock)
{
	int res;
	int i;
	const void *program_data = NULL;
	int m_ecc;
	int m_verify;
	DPRFS;

	m_ecc = (mode & SBL_M_ECC);
	m_verify = (mode & SBL_M_VERIFY);

	for (i = 0; i < dev->pages_per_block; i++) {
		res = sbl_dev_read(dev, pblock_orig, i,
				   SBL_M_RAW | SBL_M_DATA | m_ecc,
				   dev->copy_page_buf, NULL);
		if (res < 0) {
			DPRFR;
			return res;
		}
		program_data = dev->copy_page_buf;

		if (sbl_data_is_all_ff(program_data, dev->data_area_size)) {
			continue;
		}
		res = sbl_program_page(dev, pblock, i,
				       SBL_M_RAW | SBL_M_DATA | SBL_M_ECC |
				       m_verify,
				       program_data);
		if (res < 0) {
			DPRFR;
			return res;
		}
	}

	/* write sbl_info */
	res = sbl_info0_program(dev, pblock, lblock);
	if (res < 0) {
		DPRFR;
		return res;
	}

	DPRFE;
	return 0;
}

/* ------------------------ *
 * read                     *
 * ------------------------ */

/* possible errors: ESBLIO, ESBLECC, ESBLBADBLK, ESBLINVAL */
static int
sbl_read_raw(sbl_dev_t *dev, unsigned long pblock, int page,
	     int mode, void *data)
{
	int res;
	unsigned long *cache = dev->sbl_info_cache;
	DPRFS;

	/* check arguments */
	if (pblock >= dev->normal_blocks + dev->spare_blocks) {
		DPRFL(DPL_NOTICE, "invalid block number %ld\n", pblock);
		return -ESBLINVAL;
	}
	if (page < 0 || page >= dev->pages_per_block) {
		DPRFL(DPL_NOTICE, "invalid page %d\n", page);
		return -ESBLINVAL;
	}
	if (cache != NULL && cache[pblock] == SBL_SLBN_BADBLK) {
		DPRFR;
		return -ESBLBADBLK;
	}

	if (!(mode & SBL_M_DATA) && !(mode & SBL_M_EXTRA)) {
		mode |= SBL_M_DATA;
	}

	res = sbl_read_page(dev, pblock, page, mode, data);

	DPRFE;
	return res;
}

/* possible errors: ESBLIO, ESBLNOENT, ESBLECC, ESBLBADBLK, ESBLINVAL */
static int
sbl_read_spare(sbl_dev_t *dev, unsigned long lblock, int page,
	       int mode, void *data)
{
	int res;
	unsigned long pblock;
	unsigned long pblock_new;
	int i;
	int need_copy;
	DPRFS;

	/* check arguments */
	if (lblock >= dev->normal_blocks) {
		DPRFL(DPL_NOTICE, "invalid block number %ld\n", lblock);
		return -ESBLINVAL;
	}
	if (page < -1 || page >= dev->pages_per_block) {
		DPRFL(DPL_NOTICE, "invalid page %d\n", page);
		return -ESBLINVAL;
	}
	if (mode & SBL_M_EXTRA) {
		DPRFL(DPL_NOTICE, "invalid mode 0x%08x\n", mode);
		return -ESBLINVAL;
	}

	if (!(mode & SBL_M_DATA) && !(mode & SBL_M_EXTRA)) {
		mode |= SBL_M_DATA;
	}

	/* find the block */
	pblock = SBL_SLBN_INVALID;
	res = sbl_find_lblock_next(dev, lblock, &pblock, SBL_SLBN_INVALID);
	if (res < 0) {
		DPRFR;
		return res;
	}

	/* read */
	need_copy = 0;
	if (page == -1) {
		for (i = 0; i < dev->pages_per_block; i++) {
			res = sbl_read_page(dev, pblock, i, mode,
					    data + dev->data_area_size * i);
			if (res < 0) {
				DPRFR;
				return res;
			}
			if (res == 1) {
				need_copy = 1;
			}
		}
	} else {
		res = sbl_read_page(dev, pblock, page, mode, data);
		if (res < 0) {
			DPRFR;
			return res;
		}
		if (res == 1) {
			need_copy = 1;
		}
	}

	if (need_copy) {
		res = sbl_move_to_new_block(dev, lblock, mode, pblock,
					    &pblock_new);
		if (res == -ESBLIO) {
			DPRFR;
			return res;
		}
		if (res == 0 && pblock_new != lblock) {
			/* move back */
			res = sbl_move_to_normal_block(dev, lblock, pblock_new,
						       mode);
			if (res == -ESBLIO) {
				DPRFR;
				return res;
			}
		}
	} else if ((mode & SBL_M_MOVEBACK) && pblock != lblock) {
		/* move back */
		res = sbl_move_to_normal_block(dev, lblock, pblock, mode);
		if (res == -ESBLIO) {
			DPRFR;
			return res;
		}
	}

	DPRFE;
	return 0;
}

/* possible errors: ESBLIO, ESBLECC, ESBLBADBLK */
static int
sbl_read_page(sbl_dev_t *dev, unsigned long pblock, int page,
		 int mode, void *data)
{
	int res;
	int m_ecc;
	DPRFS;

	m_ecc = (mode & SBL_M_ECC);

	if ((mode & SBL_M_DATA) && (mode & SBL_M_EXTRA)) {
		res = sbl_dev_read(dev, pblock, page,
				   SBL_M_RAW | SBL_M_DATA | SBL_M_EXTRA |
				   m_ecc,
				   data, data + dev->data_area_size);
	} else if (mode & SBL_M_EXTRA) {
		res = sbl_dev_read(dev, pblock, page,
				   SBL_M_RAW | SBL_M_EXTRA,
				   NULL, data);
	} else {
		res = sbl_dev_read(dev, pblock, page,
				   SBL_M_RAW | SBL_M_DATA | m_ecc,
				   data, NULL);
	}
	if (res < 0) {
		DPRFR;
		return res;
	}

	DPRFE;
	return res;
}

/* ------------------------ *
 * verify                   *
 * ------------------------ */

static int
sbl_is_same_data(const void *data1, const void *data2, unsigned int data_size)
{
	int i;
	DPRFS;

	/* data1 and data2 are 32-bits aligned. */
	for (i = 0; i + 4 <= data_size; i += 4) {
		if (*((const unsigned long *)(data1 + i)) !=
		    *((const unsigned long *)(data2 + i))) {
			DPRFR;
			return 0;
		}
	}
	for (; i < data_size; i++) {
		if (*((const unsigned char *)(data1 + i)) !=
		    *((const unsigned char *)(data2 + i))) {
			DPRFR;
			return 0;
		}
	}

	DPRFE;
	return 1;
}

/* possible errors: ESBLIO, ESBLVERIFY */
static int
sbl_verify_raw(sbl_dev_t *dev, unsigned long pblock, int page,
	       int mode, const void *data)
{
	int res;
	const void *data_extra;
	DPRFS;

	if (mode & SBL_M_DATA) {
		res = sbl_verify_data(dev, pblock, page, data);
		if (res < 0) {
			DPRFR;
			return res;
		}
	}
	if (mode & SBL_M_EXTRA) {
		if (mode & SBL_M_DATA) {
			data_extra = data + dev->data_area_size;
		} else {
			data_extra = data;
		}
		res = sbl_verify_extra(dev, pblock, page, data_extra);
		if (res < 0) {
			DPRFR;
			return res;
		}
	}

	DPRFE;
	return 0;
}

/* possible errors: ESBLIO, ESBLVERIFY */
static int
sbl_verify_data(sbl_dev_t *dev, unsigned long pblock, int page,
		const void *data)
{
	int res;
	DPRFS;

	if (dev->flags & SBL_F_VERIFY_DATA) {
		DPRFR;
		return 0;
	}

	res = sbl_dev_read(dev, pblock, page, SBL_M_RAW | SBL_M_DATA,
			   dev->page_buf, NULL);
	if (res == -ESBLIO) {
		DPRFR;
		return res;
	}
	if (res < 0) {
		DPRFR;
		return -ESBLVERIFY;
	}

	if (!sbl_is_same_data(data, dev->page_buf, dev->data_area_size)) {
		DPRFR;
		return -ESBLVERIFY;
	}

	DPRFE;
	return 0;
}

/* possible errors: ESBLIO, ESBLVERIFY */
static int
sbl_verify_extra(sbl_dev_t *dev, unsigned long pblock, int page,
		 const void *extra)
{
	int res;
	DPRFS;

	if (dev->flags & SBL_F_VERIFY_EXTRA) {
		DPRFR;
		return 0;
	}

	res = sbl_dev_read(dev, pblock, page, SBL_M_RAW | SBL_M_EXTRA,
			   NULL, dev->page_buf);
	if (res == -ESBLIO) {
		DPRFR;
		return res;
	}
	if (res < 0) {
		DPRFR;
		return -ESBLVERIFY;
	}

	if (!sbl_is_same_data(extra, dev->page_buf, dev->extra_area_size)) {
		DPRFR;
		return -ESBLVERIFY;
	}

	DPRFE;
	return 0;
}

/* possible errors: ESBLIO, ESBLVERIFY */
static int
sbl_verify_sbl_info(sbl_dev_t *dev, unsigned long pblock, int page,
		    const void *sbl_info)
{
	int res;
	char buf[SBL_INFO_SIZE];
	DPRFS;

	if (dev->flags & SBL_F_VERIFY_EXTRA) {
		DPRFR;
		return 0;
	}

	res = sbl_dev_read(dev, pblock, page, SBL_M_RAW | SBL_M_SBLINFO,
			   NULL, buf);
	if (res == -ESBLIO) {
		DPRFR;
		return res;
	}
	if (res < 0) {
		DPRFR;
		return -ESBLVERIFY;
	}

	if (!sbl_is_same_data(sbl_info, buf, sizeof(buf))) {
		DPRFR;
		return -ESBLVERIFY;
	}

	DPRFE;
	return 0;
}

/* ------------------------ *
 * misc                     *
 * ------------------------ */

/* possible errors: ESBLIO, ESBLNOSPC */
static int
sbl_find_invalid_from(sbl_dev_t *dev, unsigned long *pblock,
		      unsigned long pblock_excluded)
{
	int res;
	unsigned long info_lblock;
	DPRFS;

	while (1) {
		if (*pblock != pblock_excluded) {
			res = sbl_info_read(dev, *pblock, &info_lblock);
			if (res < 0) {
				DPRFR;
				return res;
			}
			if (info_lblock == SBL_SLBN_INVALID) {
				break;
			}
		}
		if (*pblock < dev->normal_blocks) {
			*pblock = dev->normal_blocks + dev->spare_blocks - 1;
		} else {
			(*pblock)--;
		}
		if (*pblock < dev->normal_blocks) {
			DPRFR;
			return -ESBLNOSPC;
		}
	}

	DPRFE;
	return 0;
}

/* possible errors: ESBLIO, ESBLNOENT */
static int
sbl_find_lblock_next(sbl_dev_t *dev, unsigned long lblock,
		     unsigned long *pblock, unsigned long pblock_excluded)
{
	int res;
	DPRFS;

	while (1) {
		if (*pblock == SBL_SLBN_INVALID) {
			*pblock = lblock;
		} else {
			if (*pblock == lblock) {
				*pblock = dev->normal_blocks +
					dev->spare_blocks - 1;
			} else {
				(*pblock)--;
			}
			if (*pblock < dev->normal_blocks) {
				break;
			}
		}
		if (*pblock == pblock_excluded) {
			continue;
		}

		res = sbl_info_lblock_is_matched(dev, *pblock, lblock);
		if (res == 1) {
			DPRFR;
			return 0;
		}
		if (res == -ESBLIO) {
			DPRFR;
			return res;
		}
	};

	DPRFE;
	return -ESBLNOENT;
}

/* possible errors: none */
static int
sbl_data_is_all_ff(const void *data, unsigned int data_size)
{
	unsigned int i;
	DPRFS;

	for (i = 0; i < data_size; i++) {
		if (*((const unsigned char *)data + i) != 0xff) {
			DPRFR;
			return 0;
		}
	}

	DPRFE;
	return 1;
}
