/*
 *  mpu110_eth.c
 *
 *  SMC91111 setup code for MPU-110
 *
 *  Copyright 2003 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <asm/io.h>
#include <asm/arch/platform.h>
#include <asm/arch/gpio.h>
#include <asm/arch/regs.h>
#include <asm/arch/mpu110_series.h>
#include "mpu110_series_eth.h"
#ifdef CONFIG_MPU110_IOA_3_CS
#include "ioa_3_eth.h"
#endif

/*
 * functions for MPU-110 onboard ether device
 */

/* 
 * returns interrupt number
 */
int mpu110_eth_intnum(void)
{
	return LAN_GPIO_INT;
}

/* 
 * returns base address
 */
int mpu110_eth_baseaddr(void)
{
	return LAN_BASE + LAN_IO_OFFSET;
}

/* 
 * return < 0 if disabled
 */
int mpu110_eth_check_disable(void)
{
	if (inl(MPU110_GPIO1) & MPU110_SW_ETHER_DISABLE) {
		printk("On Board Ethernet is disabled by DIPSW\n");
		return -ENODEV;
	}
	return 0;
}

/* 
 * setup the device
 */
int mpu110_eth_setup(void)
{
	unsigned int tmp;

	tmp = inl(MPU110_GPIO2);
	outl(tmp & ~MPU110_PWR_ETHER, MPU110_GPIO2); // set 0 to turn on

	/* choose CS3 at Syscon: FMCR */
	tmp = inl(SYSCON_BASE + SYSCON_FMCR);
	outl(tmp & ~FMCR_SELECT_CSD1, SYSCON_BASE + SYSCON_FMCR);

	return 0;
}

/* 
 * requests ethernet irq
 */
int mpu110_eth_request_irq(unsigned int irq, 
		       void (*handler)(int, void *, struct pt_regs *),
		       unsigned long irq_flags, const char * devname, 
		       void *dev_id)
{
	int retval;

	retval = dragonball_register_gpio(LAN_GPIO_PORT, LAN_GPIO_BIT,
					  LAN_GPIO_MODE, LAN_GPIO_NAME);
	if (retval >= 0)
		dragonball_gpio_config_intr(LAN_GPIO_PORT, LAN_GPIO_BIT, POSITIVE_LEVEL);
	else
		return retval;
	
	retval = request_irq(irq, handler, irq_flags, devname, dev_id);
	if (retval >= 0) {
		dragonball_gpio_unmask_intr(LAN_GPIO_PORT, LAN_GPIO_BIT);
	}
	
	return retval;
}

/*
 * free ethernet irq 
 */
int mpu110_eth_free_irq(unsigned int irq, void *dev_id)
{
	/* unregister interrupt on CPU GPIO */
	dragonball_gpio_mask_intr(LAN_GPIO_PORT, LAN_GPIO_BIT);
	dragonball_unregister_gpio(LAN_GPIO_PORT, LAN_GPIO_BIT);
	
	free_irq(irq, dev_id);

	return 0;
}

/* 
 * clears interrupt
 */
int mpu110_eth_clear_intr(void *dev_id)
{
	dragonball_gpio_clear_intr(LAN_GPIO_PORT, LAN_GPIO_BIT);
	return 0;
}


/*
 * function table for MPU-110 onboard ether device
 */
static mpu110_series_eth_ops_t mpu110_eth = {
	intnum        : mpu110_eth_intnum,
	baseaddr      : mpu110_eth_baseaddr,
	check_disable : mpu110_eth_check_disable,
	setup         : mpu110_eth_setup,
	request_irq   : mpu110_eth_request_irq,
	free_irq      : mpu110_eth_free_irq,
	clear_intr    : mpu110_eth_clear_intr,
};

/*
 * MPU-110 Ether Devices
 */
mpu110_series_eth_ops_t *mpu110_series_eths[] = {
	&mpu110_eth,  /* First device: Onboard ether */
#ifdef CONFIG_MPU110_IOA_3_CS
	&ioa_3_eth,   /* Second device: IOA-3 ether */
#endif
};
