/*
 *  pci_conf.c -- do PCI configuration cycle
 *
 *  Copyright 2003 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/config.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/pci.h>
#include <linux/pci_ids.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/slab.h>
#include <linux/pci_conf.h>
#ifdef CONFIG_PCI_CONF_EARLY
#include <linux/bootmem.h>
#endif

#define DEBUG		1
#define DEBUG_LEVEL	DPL_NOTICE

#ifdef DEBUG
#define DPR(fmt, args...)	printk(fmt, ## args)
#endif /* DEBUG */

/* ======================================================================= *
 * Debug print function from dpr package                                   *
 * ======================================================================= */

#define DPL_EMERG	0
#define DPL_ALERT	1
#define DPL_CRIT	2
#define DPL_ERR		3
#define DPL_WARNING	4
#define DPL_NOTICE	5
#define DPL_INFO	6
#define DPL_DEBUG	7

#define DPM_PURPOSE	0x0000000f
#define DPM_ERR		0x00000001
#define DPM_TRACE	0x00000002
#define DPM_LAYER	0x000000f0
#define DPM_APPLICATION	0x00000010
#define DPM_MIDDLEWARE	0x00000020
#define DPM_DRIVER	0x00000040
#define DPM_KERNEL	0x00000080
#define DPM_USERDEF	0xffff0000

#ifdef DEBUG

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL DPL_DEBUG
#endif
#ifndef DEBUG_MASK
#define DEBUG_MASK  DPM_PURPOSE
#endif

#ifndef DPR
#include <stdio.h>
#define DPR(fmt, args...)	fprintf(stderr, fmt, ## args)
#endif

#define DPRL(level, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && DPR(fmt, ## args))
#define DPRM(mask, fmt, args...) \
	(void)(((mask) & DEBUG_MASK) && DPR(fmt, ## args))
#define DPRLM(level, mask, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && ((mask) & DEBUG_MASK) && \
		DPR(fmt, ## args))
#define DPRC(cond, fmt, args...) \
	(void)((cond) && DPR(fmt, ## args))
#define DPRF(fmt, args...)	DPR("%s: " fmt, __FUNCTION__, ## args)
#define DPRFL(level, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && DPRF(fmt, ## args))
#define DPRFM(mask, fmt, args...) \
	(void)(((mask) & DEBUG_MASK) && DPRF(fmt, ## args))
#define DPRFLM(level, mask, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && ((mask) & DEBUG_MASK) && \
		DPRF(fmt, ## args))
#define DPRFC(cond, fmt, args...) \
	(void)((cond) && DPRF(fmt, ## args))

#define	DPRFS	DPRFLM(DPL_DEBUG, DPM_TRACE, "start\n")
#define	DPRFE	DPRFLM(DPL_DEBUG, DPM_TRACE, "end\n")
#define DPRFR	DPRFLM(DPL_DEBUG, DPM_TRACE, "return\n")

#else  /* DEBUG */

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL -1
#endif
#ifndef DEBUG_MASK
#define DEBUG_MASK  0
#endif

#ifndef DPR
#define DPR(fmt, args...)
#endif

#define DPRL(level, fmt, args...)
#define DPRM(mask, fmt, args...)
#define DPRLM(level, mask, fmt, args...)
#define DPRC(cond, fmt, args...)
#define DPRF(fmt, args...)
#define DPRFL(level, fmt, args...)
#define DPRFM(mask, fmt, args...)
#define DPRFLM(level, mask, fmt, args...)
#define DPRFC(cond, fmt, args...)

#define	DPRFS
#define	DPRFE
#define DPRFR

#endif /* DEBUG */

/* ======================================================================= *
 * Prototype declaration                                                   *
 * ======================================================================= */

static u32 pciconf_align(u32 val, u32 align);
static u32 pciconf_align_pow2(u32 val);

static int pciconf_dev_to_default_conftype(struct pci_dev *dev,
					   struct pci_ops *ops);

static struct resource *pciconf_malloc_res(const char *name,
					   unsigned long flags);

static int pciconf_conftype_getflg(int conftype);

#ifdef CONFIG_PCI_CONF_SPECIAL_DEVS
static int pciconf_allocate_resource_at(struct resource *resources[], 
			struct resource *new,
			unsigned long addr, unsigned long size, 
			unsigned long align);
#endif
static int pciconf_allocate_resource(struct resource *resources[],
				     struct resource *new,
				     unsigned long size, unsigned long align);

static void pciconf_scan_bus(int *busnum, struct pci_bus *parent_bus,
			     struct pci_ops *ops,
			     pciconf_dev_to_conftype_t dev_to_conftype,
			     u32 *io_size, u32 *mem_size, u32 *pmem_size);
static int pciconf_alloc_tree(int busnum, struct pci_bus *parent_bus,
			      struct pci_ops *ops,
			      pciconf_dev_to_conftype_t dev_to_conftype,
			      struct resource *io_resources[],
			      struct resource *mem_resources[],
			      struct resource *pmem_resources[]);

static void pciconf_add_barsize(struct pci_dev *dev, struct pci_ops *ops,
				size_t num,
				u32 *io_size, u32 *mem_size, u32 *pmem_size,
				int flags);

static int pciconf_alloc_bar(struct pci_dev *dev, struct pci_ops *ops,
			     size_t num,
			     struct resource *io_resources[],
			     struct resource *mem_resources[],
			     struct resource *pmem_resources[],
			     int flags);

static void pciconf_preset_bridge_win(struct pci_dev *dev, struct pci_ops *ops,
				      u32 io_size, u32 mem_size,
				      u32 pmem_size, int flags);

static int pciconf_alloc_bridge_win(struct pci_dev *dev, struct pci_ops *ops,
				    pciconf_dev_to_conftype_t dev_to_conftype,
				    struct resource *io_resources[],
				    struct resource *mem_resources[],
				    struct resource *pmem_resources[],
				    int flags);

static void pciconf_preset_cb_win(struct pci_dev *dev, struct pci_ops *ops,
				  u32 io_size, u32 mem_size, u32 pmem_size);

/* ======================================================================= *
 * Kernel service routine                                                  *
 * ======================================================================= */

int
pciconf_conf(int busnum, struct pci_ops *ops,
	     pciconf_dev_to_conftype_t dev_to_conftype,
	     struct resource *io_resources[], struct resource *mem_resources[])
{
	int res;
	int busnum_max;
	u32 io_size = 0x0;
	u32 mem_size = 0x0;
	u32 pmem_size = 0x0;
	DPRFS;

	busnum_max = busnum;
	pciconf_scan_bus(&busnum_max, NULL, ops, dev_to_conftype,
			 &io_size, &mem_size, &pmem_size);

	res = pciconf_alloc_tree(busnum, NULL, ops, dev_to_conftype,
				 io_resources, mem_resources, mem_resources);
	if (res < 0) {
		DPRFR;
		return res;
	}

	DPRFE;
	return (busnum_max + 1);
}

int
pciconf_special_conftype(struct pci_dev *dev, struct pci_ops *ops,
			 pciconf_special_dev_t *list)
{
	u16 vendor_id;
	u16 device_id;
	pciconf_special_dev_t *p;
	DPRFS;

	ops->read_word(dev, PCI_VENDOR_ID, &vendor_id);
	ops->read_word(dev, PCI_DEVICE_ID, &device_id);
	for (p = list; p->vendor != 0; p++) {
		if ((vendor_id == p->vendor || p->vendor == (u16)PCI_ANY_ID) &&
		    (device_id == p->device || p->device == (u16)PCI_ANY_ID)) {
			DPRFR;
			return p->conftype;
		}
	}

	DPRFE;
	return PCICONF_CONFTYPE_DEFAULT;
}

/* ======================================================================= *
 * Internal function                                                       *
 * ======================================================================= */

static u32
pciconf_align(u32 val, u32 align)
{
	return ((val + (align - 1)) & ~(align - 1));
}

static u32
pciconf_align_pow2(u32 val)
{
	int i;

	for (i = 0; i < 32; i++) {
		if (val <= ((u32)1 << i)) {
			break;
		}
	}

	return (1 << i);
}

static int
pciconf_dev_to_default_conftype(struct pci_dev *dev, struct pci_ops *ops)
{
	int devtype;
	u8  header_type;
	u32 class_revision;
	u16 class_high;
	DPRFS;

	ops->read_byte(dev, PCI_HEADER_TYPE, &header_type);
	ops->read_dword(dev, PCI_CLASS_REVISION, &class_revision);
	class_high = (class_revision >> 16);

	devtype = PCICONF_CONFTYPE_UNKNOWN;
	if ((header_type & 0x7f) == PCI_HEADER_TYPE_NORMAL &&
	    class_high != PCI_CLASS_BRIDGE_PCI) {
		devtype = PCICONF_CONFTYPE_NORMAL;
	} else if ((header_type & 0x7f) == PCI_HEADER_TYPE_BRIDGE &&
		   class_high == PCI_CLASS_BRIDGE_PCI) {
		devtype = PCICONF_CONFTYPE_PCIPCI_BRIDGE;
	} else if ((header_type & 0x7f) == PCI_HEADER_TYPE_CARDBUS &&
		   class_high == PCI_CLASS_BRIDGE_CARDBUS) {
		devtype = PCICONF_CONFTYPE_CARDBUS_BRIDGE;
	}

	DPRFE;
	return devtype;
}

static int
pciconf_conftype_getflg(int conftype)
{
	return conftype & PCICONF_CONFTYPE_FLG_MASK;
}

static int
pciconf_conftype_clrflg(int conftype)
{
	return conftype & ~(PCICONF_CONFTYPE_FLG_MASK);
}

static struct resource *
pciconf_malloc_res(const char *name, unsigned long flags)
{
	struct resource *res;
	DPRFS;

#ifdef CONFIG_PCI_CONF_EARLY
	res = alloc_bootmem(sizeof(struct resource));
	if (res == NULL) {
		DPRFR;
		return NULL;
	}
#else
	res = kmalloc(sizeof(struct resource), GFP_KERNEL);
	if (res == NULL) {
		DPRFR;
		return NULL;
	}
#endif
	memset(res, 0, sizeof(struct resource));
	res->name = name;
	res->flags = flags;

	DPRFE;
	return res;
}

static int
pciconf_allocate_resource(struct resource *resources[], struct resource *new,
			  unsigned long size, unsigned long align)
{
	int res;
	int i;
	DPRFS;
	DPRFL(DPL_DEBUG,"(size=0x%lx, align=0x%lx)\n", size, align);

	if (size == 0) {
		DPRFE;
		return 0;
	}

	for (i = 0;; i++) {
		DPRFL(DPL_DEBUG,"resources[%d]=0x%lx)\n", i, (unsigned int)resources[i]);
		if (resources[i] == NULL) {
			res = -EBUSY;
			goto err;
		}
		res = allocate_resource(resources[i], new, size,
					0, -1, align, NULL, NULL);
		if (res == -EBUSY) {
			continue;
		} else if (res < 0) {
			goto err;
		}
		break;
	}

	DPRFE;
	return 0;

err:
	DPRFR;
	return res;
}

#ifdef CONFIG_PCI_CONF_SPECIAL_DEVS
static int
pciconf_allocate_resource_at(struct resource *resources[], struct resource *new,
			     unsigned long addr, unsigned long size, 
			     unsigned long align)
{
	int res;
	int i;
	DPRFS;

	if (size == 0) {
		DPRFE;
		return 0;
	}

	for (i = 0;; i++) {
		if (resources[i] == NULL) {
			res = -EBUSY;
			goto err;
		}
		/* 
		 * Sanity check.
		 * Specified value might not be correct in Linux.
		 */
		DPRFL(DPL_INFO,
		      "Trying : range(%08lx-%08lx), align(0x%08lx), parent (%08lx-%08lx)\n",
		      addr, addr + size, align,
		      resources[i]->start, resources[i]->end);
		res = allocate_resource(resources[i], new, size,
					addr, addr + size, 
					align, NULL, NULL);
		if (res < 0) {
			DPRL(DPL_ERR,
			     "Specified resource at addr=0x%08lx size=0x%08lx invalid.\n"
			     "  Out of parent range (0x%08lx-0x%08lx)\n"
			     "  or already reserved by another predefined device?\n"
			     "  Trying normal method...\n",
			     addr, size,
			     resources[i]->start, resources[i]->end);
			res =  allocate_resource(resources[i], new, size,
						 0, -1, align, NULL, NULL);
		} 
		if (res == -EBUSY) {
			continue;
		} else if (res < 0) {
			goto err;
		}
		break;
	}

	DPRFE;
	return 0;

err:
	DPRFR;
	return res;
}
#endif

static void
pciconf_scan_bus(int *busnum, struct pci_bus *parent_bus,
		 struct pci_ops *ops,
		 pciconf_dev_to_conftype_t dev_to_conftype,
		 u32 *io_size, u32 *mem_size, u32 *pmem_size)
{
	int res;
	struct pci_bus bus;
	struct pci_dev dev;
	u16 vendor_id;
	u8 header_type;
	int conftype;
	int flags;

	DPRFS;

	bus.parent = parent_bus;
	bus.number = *busnum;

	bus.ops = ops;
	dev.bus = &bus;

	for (dev.devfn = 0; dev.devfn < 0x100; dev.devfn++) {
		res = ops->read_word(&dev, PCI_VENDOR_ID, &vendor_id);
		if (res == -EBUSY) {
			continue;
		} else if (res != 0 || vendor_id == 0xffff) {
			if (PCI_FUNC(dev.devfn) == 0) {
				dev.devfn += 7;
			}
			continue;
		}
		ops->read_byte(&dev, PCI_HEADER_TYPE, &header_type);

		/* calculate conftype */
		conftype = PCICONF_CONFTYPE_DEFAULT;
		if (dev_to_conftype != NULL) {
			conftype = dev_to_conftype(&dev, ops,
						   PCICONF_PHASE_SCAN);
			flags    = pciconf_conftype_getflg(conftype);
			conftype = pciconf_conftype_clrflg(conftype);
		}
		if (conftype == PCICONF_CONFTYPE_DEFAULT) {
			conftype = pciconf_dev_to_default_conftype(&dev, ops);
		}

		/* calculate sizes according to the conftype */
		if (conftype == PCICONF_CONFTYPE_IGNORE) {
			DPRFL(DPL_NOTICE, "skip [%02x:%02x.%x]\n",
			      dev.bus->number, PCI_SLOT(dev.devfn),
			      PCI_FUNC(dev.devfn));
		} else if (conftype == PCICONF_CONFTYPE_NORMAL) {
			pciconf_add_barsize(&dev, ops, 6,
					    io_size, mem_size, pmem_size, flags);
		} else if (conftype == PCICONF_CONFTYPE_PCIPCI_BRIDGE) {
			u32 sub_io_size = 0;
			u32 sub_mem_size = 0;
			u32 sub_pmem_size = 0;

			pciconf_add_barsize(&dev, ops, 2,
					    io_size, mem_size, pmem_size, flags);

			ops->write_byte(&dev, PCI_PRIMARY_BUS, *busnum);
			(*busnum)++;
			ops->write_byte(&dev, PCI_SECONDARY_BUS, *busnum);
			ops->write_byte(&dev, PCI_SUBORDINATE_BUS, 0xff);

			pciconf_scan_bus(busnum, &bus, ops, dev_to_conftype,
					 &sub_io_size, &sub_mem_size,
					 &sub_pmem_size);

			sub_io_size = pciconf_align(sub_io_size, 0x1000);
			sub_mem_size = pciconf_align(sub_mem_size, 0x100000);
			sub_pmem_size = pciconf_align(sub_pmem_size, 0x100000);
			ops->write_byte(&dev, PCI_SUBORDINATE_BUS, *busnum);
			pciconf_preset_bridge_win(&dev, ops, sub_io_size,
						  sub_mem_size, sub_pmem_size, flags);
			*io_size += pciconf_align_pow2(sub_io_size);
			*mem_size += pciconf_align_pow2(sub_mem_size);
			*pmem_size += pciconf_align_pow2(sub_pmem_size);
		} else if (conftype == PCICONF_CONFTYPE_CARDBUS_BRIDGE) {
			pciconf_add_barsize(&dev, ops, 1,
					    io_size, mem_size, pmem_size, flags);

			ops->write_byte(&dev, PCI_CB_PRIMARY_BUS, *busnum);
			(*busnum)++;
			ops->write_byte(&dev, PCI_CB_CARD_BUS, *busnum);
			ops->write_byte(&dev, PCI_CB_SUBORDINATE_BUS, *busnum);
			pciconf_preset_cb_win(&dev, ops, 0, 0, 0);
		} else {
			/* unknown device function */
			DPRFL(DPL_NOTICE,
			      "PCI[%02x:%02x.%x] Unknown PCI header type.\n",
			      dev.bus->number, PCI_SLOT(dev.devfn),
			      PCI_FUNC(dev.devfn));
		}
		if (!(header_type & 0x80) && PCI_FUNC(dev.devfn) == 0) {
			dev.devfn += 7;
		}
	}

	DPRFL(DPL_DEBUG,
	      "\n  Total sum for PCI bus[%02x] : \n"
	      "  io_size  = 0x%08x\n"
	      "  mem_size = 0x%08x\n"
	      "  pmem_size= 0x%08x\n",
	      dev.bus->number, 
	      *io_size, *mem_size, *pmem_size);

	DPRFE;
	return;
}

static int
pciconf_alloc_tree(int busnum, struct pci_bus *parent_bus, struct pci_ops *ops,
		   pciconf_dev_to_conftype_t dev_to_conftype,
		   struct resource *io_resources[],
		   struct resource *mem_resources[],
		   struct resource *pmem_resources[])
{
	int res;
	struct pci_bus bus;
	struct pci_dev dev;
	u16 vendor_id;
	u8  header_type;
	int conftype;
	int flags, pass;
	DPRFS;

	bus.parent = parent_bus;
	bus.number = busnum;

	bus.ops = ops;
	dev.bus = &bus;

	/* First pass allocates predefined regions. */
	pass = 0;
#ifdef CONFIG_PCI_CONF_SPECIAL_DEVS
second_pass:
	DPRFL(DPL_DEBUG, "Pass %d\n", pass);
#endif
	for (dev.devfn = 0; dev.devfn < 0x100; dev.devfn++) {
		res = ops->read_word(&dev, PCI_VENDOR_ID, &vendor_id);
		if (res == -EBUSY) {
			continue;
		} else if (res != 0 || vendor_id == 0xffff) {
			if (PCI_FUNC(dev.devfn) == 0) {
				dev.devfn += 7;
			}
			continue;
		}
		ops->read_byte(&dev, PCI_HEADER_TYPE, &header_type);

		/* calculate conftype */
		conftype = PCICONF_CONFTYPE_DEFAULT;
		if (dev_to_conftype != NULL) {
			conftype = dev_to_conftype(&dev, ops,
						   PCICONF_PHASE_ALLOC);
			flags    = pciconf_conftype_getflg(conftype);
			conftype = pciconf_conftype_clrflg(conftype);
		}
		if (conftype == PCICONF_CONFTYPE_DEFAULT) {
			conftype = pciconf_dev_to_default_conftype(&dev, ops);
		}

#ifdef CONFIG_PCI_CONF_SPECIAL_DEVS
		/* 
		 * Try to configure only if:
		 *   - Preset is set     and Pass == 0
		 *   - Preset is not set and Pass == 1
		 */
		if (!(((flags & PCICONF_CONFTYPE_FLG_PRESET) > 0) ^ (pass > 0)))
			continue;
#endif
		/* allocate resources according to the conftype */
		if (conftype == PCICONF_CONFTYPE_IGNORE) {
			DPRFL(DPL_NOTICE, "skip [%02x:%02x.%x]\n",
			      dev.bus->number, PCI_SLOT(dev.devfn),
			      PCI_FUNC(dev.devfn));
		} else if (conftype == PCICONF_CONFTYPE_NORMAL) {
			res = pciconf_alloc_bar(&dev, ops, 6, io_resources,
						mem_resources, pmem_resources, flags);
			if (res < 0) {
				goto err;
			}
		} else if (conftype == PCICONF_CONFTYPE_PCIPCI_BRIDGE) {
			res = pciconf_alloc_bar(&dev, ops, 2, io_resources,
						mem_resources, pmem_resources, flags);
			if (res < 0) {
				goto err;
			}
			res = pciconf_alloc_bridge_win(&dev, ops,
						       dev_to_conftype,
						       io_resources,
						       mem_resources,
						       pmem_resources,
						       flags);
			if (res < 0) {
				goto err;
			}
		} else if (conftype == PCICONF_CONFTYPE_CARDBUS_BRIDGE) {
			res = pciconf_alloc_bar(&dev, ops, 1, io_resources,
						mem_resources, pmem_resources,
						flags);
			if (res < 0) {
				goto err;
			}
		} else {
			/* unknown device function */
			DPRFL(DPL_NOTICE,
			      "PCI[%02x:%02x.%x] Unknown PCI header type.\n",
			      dev.bus->number, PCI_SLOT(dev.devfn),
			      PCI_FUNC(dev.devfn));
		}
		if (!(header_type & 0x80) && PCI_FUNC(dev.devfn) == 0) {
			dev.devfn += 7;
		}
	}
#ifdef CONFIG_PCI_CONF_SPECIAL_DEVS
	if (pass == 0) {
		pass++;
		goto second_pass;
	}
#endif

	DPRFE;
	return 0;

err:
	DPRFR;
	return res;
}

static void
pciconf_add_barsize(struct pci_dev *dev, struct pci_ops *ops, size_t num,
		      u32 *io_size, u32 *mem_size, u32 *pmem_size,
		      int flags)
{
	int i;
	u32 bar;
	int is_64bits_memory;
#ifdef CONFIG_PCI_CONF_SPECIAL_DEVS
	u32 save_bar;
#endif
	DPRFS;

	for (i = 0; i < num; i++) {
		is_64bits_memory = 0;

#ifdef CONFIG_PCI_CONF_SPECIAL_DEVS
		/* 
		 * Save if predefined BAR address is available, 
		 * it will be restored at end of this function.
		 */
		if (flags & PCICONF_CONFTYPE_FLG_PRESET) 
			ops->read_dword(dev, PCI_BASE_ADDRESS_0 + i * 4, &save_bar);
#endif
		ops->write_dword(dev, PCI_BASE_ADDRESS_0 + i * 4, 0xffffffff);
		ops->read_dword(dev, PCI_BASE_ADDRESS_0 + i * 4, &bar);

		DPRFL(DPL_DEBUG, "bar[%d] = 0x%08x\n", i, bar);

		if (bar == 0) {
			continue;
		}
		if ((bar & 0xffff0000) == 0) {
			bar |= 0xffff0000;	/* for buggy device */
		}
		if (bar & PCI_BASE_ADDRESS_SPACE) {
			/* I/O space */
			*io_size += ~(bar & PCI_BASE_ADDRESS_IO_MASK) + 1;
		} else {
			/* memory space */
			if ((bar & PCI_BASE_ADDRESS_MEM_TYPE_MASK) ==
			    PCI_BASE_ADDRESS_MEM_TYPE_64) {
				is_64bits_memory = 1;
			}
			if (bar & PCI_BASE_ADDRESS_MEM_PREFETCH) {
				/* prefetchable memory space */
				*pmem_size +=
					~(bar & PCI_BASE_ADDRESS_MEM_MASK) + 1;
			} else {
				/* memory space */
				*mem_size +=
					~(bar & PCI_BASE_ADDRESS_MEM_MASK) + 1;
			}
		}
		if (is_64bits_memory) {
			i++;
		}
#ifdef CONFIG_PCI_CONF_SPECIAL_DEVS
		/* Restore prefined BAR address for later use. */
		if (flags & PCICONF_CONFTYPE_FLG_PRESET)
			ops->write_dword(dev, PCI_BASE_ADDRESS_0 + i * 4, save_bar);
#endif
	}

	DPRFE;
	return;
}

static int
pciconf_alloc_bar(struct pci_dev *dev, struct pci_ops *ops, size_t num,
		    struct resource *io_resources[],
		    struct resource *mem_resources[],
		    struct resource *pmem_resources[],
		    int flags)
{
	int res;
	int i;
	u32 bar;
	struct resource *new_res;
	u32 size;
	u16 command;
	int is_64bits_memory;
#ifdef CONFIG_PCI_CONF_SPECIAL_DEVS
	u32 save_bar;
#endif
	DPRFS;

	for (i = 0; i < num; i++) {
		is_64bits_memory = 0;

#ifdef CONFIG_PCI_CONF_SPECIAL_DEVS
		/* 
		 * Prefined address is saved in the BAR itself.
		 */
		if (flags & PCICONF_CONFTYPE_FLG_PRESET) {
			ops->read_dword(dev, PCI_BASE_ADDRESS_0 + i * 4, &save_bar);
			save_bar &= PCI_BASE_ADDRESS_IO_MASK & PCI_BASE_ADDRESS_MEM_MASK;
		}
#endif
		ops->write_dword(dev, PCI_BASE_ADDRESS_0 + i * 4, 0xffffffff);
		ops->read_dword(dev, PCI_BASE_ADDRESS_0 + i * 4, &bar);

		if (bar == 0) {
			continue;
		}
		if ((bar & 0xffff0000) == 0) {
			bar |= 0xffff0000;	/* for buggy device */
		}
		DPRFL(DPL_INFO, "bar[%02x:%02x.%x][%d] = 0x%08x\n",
		      dev->bus->number, PCI_SLOT(dev->devfn),
		      PCI_FUNC(dev->devfn), i, bar);

#ifdef CONFIG_PCI_CONF_SPECIAL_DEVS
		if (flags & PCICONF_CONFTYPE_FLG_PRESET) 
			DPRFL(DPL_NOTICE, "bar[%02x:%02x.%x][%d] = 0x%08x (predefined)\n",
			      dev->bus->number, PCI_SLOT(dev->devfn),
			      PCI_FUNC(dev->devfn), i, save_bar);
#endif
		if (bar & PCI_BASE_ADDRESS_SPACE) {
			/* I/O space */
			size = ~(bar & PCI_BASE_ADDRESS_IO_MASK) + 1;
			new_res = pciconf_malloc_res("PCI I/O resource",
						     IORESOURCE_IO);
			if (new_res == NULL) {
				res = -ENOMEM;
				goto err;
			}

#ifdef CONFIG_PCI_CONF_SPECIAL_DEVS
			if (flags & PCICONF_CONFTYPE_FLG_PRESET) 
				res = pciconf_allocate_resource_at(io_resources,
							new_res, 
							save_bar, size, size);
			else 
#endif
				res = pciconf_allocate_resource(io_resources,
							new_res, size, size);
			if (res < 0) {
				goto err;
			}
			DPRFL(DPL_INFO,
			      "PCI[%02x:%02x.%x] allocated I/O resource "
			      "0x%08lx-0x%08lx\n",
			      dev->bus->number, PCI_SLOT(dev->devfn),
			      PCI_FUNC(dev->devfn),
			      new_res->start, new_res->end);
		} else {
			/* memory space */
			if ((bar & PCI_BASE_ADDRESS_MEM_TYPE_MASK) ==
			    PCI_BASE_ADDRESS_MEM_TYPE_64) {
				is_64bits_memory = 1;
			}
			size = ~(bar & PCI_BASE_ADDRESS_MEM_MASK) + 1;
			new_res = pciconf_malloc_res("PCI memory resource",
						     IORESOURCE_MEM);
			if (new_res == NULL) {
				res = -ENOMEM;
				goto err;
			}
			if (bar & PCI_BASE_ADDRESS_MEM_PREFETCH) {
				/* prefetchable memory space */
				new_res->flags |= IORESOURCE_PREFETCH;

#ifdef CONFIG_PCI_CONF_SPECIAL_DEVS
				if (flags & PCICONF_CONFTYPE_FLG_PRESET) 
					res = pciconf_allocate_resource_at(
						pmem_resources, new_res, 
						save_bar, size, size);
				else
#endif
					res = pciconf_allocate_resource(
						pmem_resources, new_res, 
						size, size);
				
			} else {
				/* memory space */
#ifdef CONFIG_PCI_CONF_SPECIAL_DEVS
				if (flags & PCICONF_CONFTYPE_FLG_PRESET)
					res = pciconf_allocate_resource_at(
						mem_resources, new_res, 
						save_bar, size, size);
				else
#endif
					res = pciconf_allocate_resource(
						mem_resources, new_res,
						size, size);
			}				
			if (res < 0) {
				goto err;
			}
			DPRFL(DPL_DEBUG,
			      "PCI[%02x:%02x.%x] allocated memory resource "
			      "0x%08lx-0x%08lx\n",
			      dev->bus->number, PCI_SLOT(dev->devfn),
			      PCI_FUNC(dev->devfn),
			      new_res->start, new_res->end);
		}
		ops->write_dword(dev, PCI_BASE_ADDRESS_0 + i * 4,
				 new_res->start);
		if (is_64bits_memory) {
			i++;
			if (i < num) {
				ops->write_dword(dev,
						 PCI_BASE_ADDRESS_0 + i * 4,
						 0);
			}
		}
	}

	/* enable I/O and memory access */
	ops->read_word(dev, PCI_COMMAND, &command);
	command |= (PCI_COMMAND_IO | PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER);
	ops->write_word(dev, PCI_COMMAND, command);

	/* set latency timer */
	ops->write_byte(dev, PCI_LATENCY_TIMER, 0x80);

	DPRFE;
	return 0;

err:
	DPRFR;
	return res;
}

static void
pciconf_preset_bridge_win(struct pci_dev *dev, struct pci_ops *ops,
			  u32 io_size, u32 mem_size, u32 pmem_size,
			  int flags)
{
	u16 io_limit16, mem_limit16, pmem_limit16;
	u8  io_limit8;
	DPRFS;

	DPRFL(DPL_DEBUG,"(io_size=0x%x, mem_size=0x%x, pmem_size=0x%x)\n", 
	      io_size, mem_size, pmem_size);

#ifdef CONFIG_PCI_CONF_SPECIAL_DEVS
	if (flags & PCICONF_CONFTYPE_FLG_PRESET)
		return;
#endif
	/* Calculate limit address (by base 0x0) and store in bridge
	   register to pass data. */
	io_limit8    = (io_size   >> 8)  & 0xff;
	io_limit16   = (io_size   >> 16) & 0xffff;
	mem_limit16  = (mem_size  >> 16) & 0xffff;
	pmem_limit16 = (pmem_size >> 16) & 0xffff;

	ops->write_byte(dev, PCI_IO_BASE, 0xff);
	ops->write_word(dev, PCI_IO_BASE_UPPER16, 
			0xffff & PCI_IO_RANGE_MASK);
	ops->write_byte(dev, PCI_IO_LIMIT,
			io_limit8 & PCI_IO_RANGE_MASK);
	ops->write_word(dev, PCI_IO_LIMIT_UPPER16, io_limit16);

	ops->write_word(dev, PCI_MEMORY_BASE,
			0xffff & PCI_MEMORY_RANGE_MASK);
	ops->write_word(dev, PCI_MEMORY_LIMIT,
			mem_limit16 & PCI_MEMORY_RANGE_MASK);

	ops->write_word(dev, PCI_PREF_MEMORY_BASE,
			0xffff & PCI_PREF_RANGE_MASK);
	ops->write_dword(dev, PCI_PREF_BASE_UPPER32, 0);
	ops->write_word(dev, PCI_PREF_MEMORY_LIMIT,
			pmem_limit16 & PCI_PREF_RANGE_MASK);
	ops->write_dword(dev, PCI_PREF_LIMIT_UPPER32, 0);

	DPRFE;
	return;
}

static int
pciconf_alloc_bridge_win(struct pci_dev *dev, struct pci_ops *ops,
			 pciconf_dev_to_conftype_t dev_to_conftype,
			 struct resource *io_resources[],
			 struct resource *mem_resources[],
			 struct resource *pmem_resources[],
			 int flags)
{
	u8 secondary_bus;
	u8 io_limit8;
	u16 io_limit_upper16;
	u16 mem_limit16;
	u16 pmem_limit16;
	u32 io_limit, io_size;
	u32 mem_limit, mem_size;
	u32 pmem_limit, pmem_size;
	int res;
	struct resource *new_io_resources[2];
	struct resource *new_mem_resources[2];
	struct resource *new_pmem_resources[2];
#ifdef CONFIG_PCI_CONF_SPECIAL_DEVS
	u8 io_base8;
	u16 io_base_upper16, mem_base16, pmem_base16;
	u32 io_base_preset, mem_base_preset, pmem_base_preset;
#endif
	DPRFS;

	/* 
	 * Allocate I/O window 
	 */

#ifdef CONFIG_PCI_CONF_SPECIAL_DEVS
	/* read preset value */
	if (flags & PCICONF_CONFTYPE_FLG_PRESET) {
		ops->read_byte(dev, PCI_IO_BASE, &io_base8);
		ops->read_word(dev, PCI_IO_BASE_UPPER16, &io_base_upper16);
		io_base_preset = (((io_base8 & PCI_IO_RANGE_MASK) << 8) |
				  (io_base_upper16 << 16));
	}
#endif
	ops->read_byte(dev, PCI_IO_LIMIT, &io_limit8);
	ops->read_word(dev, PCI_IO_LIMIT_UPPER16, &io_limit_upper16);
	io_limit = (((io_limit8 & PCI_IO_RANGE_MASK) << 8) +
		    (io_limit_upper16 << 16));
	new_io_resources[0] =
		pciconf_malloc_res("PCI-to-PCI bridge I/O window",
				   IORESOURCE_IO);
	if (new_io_resources[0] == NULL) {
		res = -ENOMEM;
		goto err;
	}

	res = 0;
	io_size = io_limit; /* is set at pciconf_preset_bridge_win() */
#ifdef CONFIG_PCI_CONF_SPECIAL_DEVS
	if (flags & PCICONF_CONFTYPE_FLG_PRESET) {
		new_io_resources[0]->start = 0;
		new_io_resources[0]->end   = -1;
		io_size -= io_base_preset;
		res = pciconf_allocate_resource_at(io_resources, new_io_resources[0],
						   io_base_preset, 
						   io_size,
						   1); /* align as specifed */
	} else 
#endif
	{
		DPRFL(DPL_DEBUG,"io_size = 0x%08x\n", io_size);
		if (io_size) {
			res = pciconf_allocate_resource(io_resources, 
							new_io_resources[0],
							io_size,
							pciconf_align_pow2(io_size));
		}
	}

	if (res < 0) {
		goto err;
	}

	if (io_size != 0) {
		/* set window registers */
		ops->write_byte(dev, PCI_IO_BASE,
				(new_io_resources[0]->start >> 8) & PCI_IO_RANGE_MASK);
		ops->write_word(dev, PCI_IO_BASE_UPPER16,
				(new_io_resources[0]->start >> 16));
		ops->write_byte(dev, PCI_IO_LIMIT,
				(((new_io_resources[0]->end - 1) >> 8) &
				 PCI_IO_RANGE_MASK));
		ops->write_word(dev, PCI_IO_LIMIT_UPPER16,
				(new_io_resources[0]->end - 1) >> 16);
	} else {
		/* clear window registers */
		ops->write_byte(dev, PCI_IO_BASE, 0);
		ops->write_word(dev, PCI_IO_BASE_UPPER16, 0);
		ops->write_byte(dev, PCI_IO_LIMIT, 0);
		ops->write_word(dev, PCI_IO_LIMIT_UPPER16, 0);
	}


	/* 
	 * Allocate memory window 
	 */
#ifdef CONFIG_PCI_CONF_SPECIAL_DEVS
	/* Read preset value */
	if (flags & PCICONF_CONFTYPE_FLG_PRESET) {
		ops->read_word(dev, PCI_MEMORY_BASE, &mem_base16);
		mem_base_preset = (mem_base16 & PCI_MEMORY_RANGE_MASK) << 16;
	}
#endif
	ops->read_word(dev, PCI_MEMORY_LIMIT, &mem_limit16);
	mem_limit = ((mem_limit16 & PCI_MEMORY_RANGE_MASK) << 16);
	new_mem_resources[0] =
		pciconf_malloc_res("PCI-to-PCI bridge memory window",
				   IORESOURCE_MEM);
	if (new_mem_resources[0] == NULL) {
		res = -ENOMEM;
		goto err;
	}

	res = 0;
	mem_size = mem_limit; /* is set at pciconf_preset_bridge_win() */
#ifdef CONFIG_PCI_CONF_SPECIAL_DEVS
	if (flags & PCICONF_CONFTYPE_FLG_PRESET) {
		mem_size -= mem_base_preset;
		res = pciconf_allocate_resource_at(mem_resources, new_mem_resources[0],
						   mem_base_preset, 
						   mem_size,
						   1); /* align as specified */
	} else 
#endif
	{
		res = pciconf_allocate_resource(mem_resources, new_mem_resources[0],
						mem_size,
						pciconf_align_pow2(mem_size));
	}

	if (res < 0) {
		goto err;
	}
	if (mem_size != 0) {
		/* set window registers */
		ops->write_word(dev, PCI_MEMORY_BASE,
				((new_mem_resources[0]->start >> 16) &
				 PCI_MEMORY_RANGE_MASK));
		ops->write_word(dev, PCI_MEMORY_LIMIT,
				(((new_mem_resources[0]->end - 1) >> 16) &
				 PCI_MEMORY_RANGE_MASK));
	} else {
		/* clear window registers */
		ops->write_word(dev, PCI_MEMORY_BASE, 0);
		ops->write_word(dev, PCI_MEMORY_LIMIT, 0);
	}

	/* 
	 * Allocate prefetchable memory window 
	 */
#ifdef CONFIG_PCI_CONF_SPECIAL_DEVS
	/* Read preset value */
	if (flags & PCICONF_CONFTYPE_FLG_PRESET) {
		ops->read_word(dev, PCI_PREF_MEMORY_BASE, &pmem_base16);
		pmem_base_preset = (pmem_base16 & PCI_PREF_RANGE_MASK) << 16;
	}
#endif
	ops->read_word(dev, PCI_PREF_MEMORY_LIMIT, &pmem_limit16);
	pmem_limit = ((pmem_limit16 & PCI_PREF_RANGE_MASK) << 16);
	new_pmem_resources[0] =
		pciconf_malloc_res("PCI-to-PCI bridge prefetchable "
				   "memory window",
				   IORESOURCE_MEM | IORESOURCE_PREFETCH);
	if (new_pmem_resources[0] == NULL) {
		res = -ENOMEM;
		goto err;
	}

	pmem_size = pmem_limit; /* is set at pciconf_preset_bridge_win() */
#ifdef CONFIG_PCI_CONF_SPECIAL_DEVS
	if (flags & PCICONF_CONFTYPE_FLG_PRESET) {
		pmem_size -= pmem_base_preset;
		res = pciconf_allocate_resource_at(pmem_resources, new_pmem_resources[0],
						   pmem_base_preset, pmem_size,
						   1); /* align as specified */
	} else 
#endif
	{
		res = pciconf_allocate_resource(pmem_resources, new_pmem_resources[0],
						pmem_size,
						pciconf_align_pow2(pmem_size));
	}

	if (res < 0) {
		goto err;
	}

	if (pmem_size != 0) {
		/* set window registers */
		ops->write_word(dev, PCI_PREF_MEMORY_BASE,
				((new_pmem_resources[0]->start >> 16) &
				 PCI_PREF_RANGE_MASK));
		ops->write_dword(dev, PCI_PREF_BASE_UPPER32, 0);
		ops->write_word(dev, PCI_PREF_MEMORY_LIMIT,
				(((new_pmem_resources[0]->end - 1) >> 16) &
				 PCI_PREF_RANGE_MASK));
		ops->write_dword(dev, PCI_PREF_LIMIT_UPPER32, 0);
	} else {
		/* clear window registers */
		ops->write_word(dev, PCI_PREF_MEMORY_BASE, 0);
		ops->write_dword(dev, PCI_PREF_BASE_UPPER32, 0);
		ops->write_word(dev, PCI_PREF_MEMORY_LIMIT, 0);
		ops->write_dword(dev, PCI_PREF_LIMIT_UPPER32, 0);
	}

	/* allocate resouces on sub bus */
	new_io_resources[1] = NULL;
	new_mem_resources[1] = NULL;
	new_pmem_resources[1] = NULL;
	ops->read_byte(dev, PCI_SECONDARY_BUS, &secondary_bus);
	res = pciconf_alloc_tree(secondary_bus, dev->bus, ops, dev_to_conftype,
				 new_io_resources, new_mem_resources,
				 new_pmem_resources);
	if (res < 0) {
		goto err;
	}

	DPRFE;
	return 0;

err:
	DPRFR;
	return res;
}

static void
pciconf_preset_cb_win(struct pci_dev *dev, struct pci_ops *ops,
		      u32 io_size, u32 mem_size, u32 pmem_size)
{
	DPRFS;

	/**** TODO: use *_size */

	ops->write_dword(dev, PCI_CB_MEMORY_BASE_0, 0xffffffff);
	ops->write_dword(dev, PCI_CB_MEMORY_LIMIT_0, 0);
	ops->write_dword(dev, PCI_CB_MEMORY_BASE_1, 0xffffffff);
	ops->write_dword(dev, PCI_CB_MEMORY_LIMIT_1, 0);
	ops->write_dword(dev, PCI_CB_IO_BASE_0, 0xffffffff);
	ops->write_dword(dev, PCI_CB_IO_LIMIT_0, 0);
	ops->write_dword(dev, PCI_CB_IO_BASE_1, 0xffffffff);
	ops->write_dword(dev, PCI_CB_IO_LIMIT_1, 0);

	DPRFE;
	return;
}
