/*
 * linux/drivers/scsi/scsi_proc.c
 *
 * The functions in this file provide an interface between
 * the PROC file system and the SCSI device drivers
 * It is mainly used for debugging, statistics and to pass 
 * information directly to the lowlevel driver.
 *
 * (c) 1995 Michael Neuffer neuffer@goofy.zdv.uni-mainz.de 
 * Version: 0.99.8   last change: 95/09/13
 * 
 * generic command parser provided by: 
 * Andreas Heilwagen <crashcar@informatik.uni-koblenz.de>
 *
 * generic_proc_info() support of xxxx_info() by:
 * Michael A. Griffith <grif@acm.org>
 */

#include <linux/config.h>	/* for CONFIG_PROC_FS */
#define __NO_VERSION__
#include <linux/module.h>

#include <linux/string.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>
#include <linux/errno.h>
#include <linux/stat.h>
#include <linux/blk.h>
#include <linux/smp_lock.h>

#include <asm/uaccess.h>

#include "scsi.h"
#include "hosts.h"

#ifdef CONFIG_SCSI_PROC_EVENT /* FES patch */
#include <linux/init.h>
#undef SCSI_PROC_EVENT_DEBUG

#ifdef SCSI_PROC_EVENT_DEBUG
#define scsi_event_printk(x...) printk(x)
#else
#define scsi_event_printk(x...) do{}while(0)
#endif
#endif

#ifndef TRUE
#define TRUE  1
#define FALSE 0
#endif

#ifdef CONFIG_PROC_FS

/* generic_proc_info
 * Used if the driver currently has no own support for /proc/scsi
 */
int generic_proc_info(char *buffer, char **start, off_t offset, int length, 
		      const char *(*info) (struct Scsi_Host *),
		      struct Scsi_Host *sh)
{
	int len, pos, begin;

	begin = 0;
	if (info && sh) {
		pos = len = sprintf(buffer, "%s\n", info(sh));
	} else {
		pos = len = sprintf(buffer,
			"The driver does not yet support the proc-fs\n");
	}
	if (pos < offset) {
		len = 0;
		begin = pos;
	}
	*start = buffer + (offset - begin);	/* Start of wanted data */
	len -= (offset - begin);
	if (len > length)
		len = length;

	return (len);
}

/* dispatch_scsi_info is the central dispatcher 
 * It is the interface between the proc-fs and the SCSI subsystem code
 */
static int proc_scsi_read(char *buffer, char **start, off_t offset,
	int length, int *eof, void *data)
{
	struct Scsi_Host *hpnt = data;
	int n;
        int (*host_proc_info)(char *, char **, off_t, int, int, int) = NULL;
        struct Scsi_Host *hpnt2;
        int host_no;

        /*
         * check the host is in the registered list.  /proc
         * sometimes hold pointers which have been already 
         * released (unregistered) scsi_host structure.
         * This situation caused by following steps
         * o a process opens /proc/scsi/xxx entry
         * o u-stor delayed thread  calls scsi_unregister_host() 
         *   for xxx
         * o the process start to read /proc/scsi/xxx
         * To prevent deletion of scsi_host, obtain big kernel
         * lock.
         */
        lock_kernel();
        for (hpnt2 = scsi_hostlist; hpnt2; hpnt2 = hpnt2->next) {
                if (hpnt2 == hpnt) {
                        if (hpnt->hostt) {
                                host_proc_info = hpnt->hostt->proc_info;
                                host_no = hpnt->host_no;
                        }
                        break;
                }
        }

        if (!hpnt2) {
                printk("%s: no entry found\n", __FUNCTION__);
                unlock_kernel();
                return -ENXIO;
        };

	if (host_proc_info == NULL)
		n = generic_proc_info(buffer, start, offset, length,
				      hpnt->hostt->info, hpnt);
	else
		n = ((*host_proc_info)(buffer, start, offset,
                                       length, host_no, 0));
	*eof = (n<length);
        unlock_kernel();
	return n;
}

#define PROC_BLOCK_SIZE (3*1024)     /* 4K page size, but our output routines 
				      * use some slack for overruns 
				      */

static int proc_scsi_write(struct file * file, const char * buf,
                           unsigned long count, void *data)
{
	struct Scsi_Host *hpnt = data;
	ssize_t ret = 0;
	char * page;
	char *start;
        int (*host_proc_info)(char *, char **, off_t, int, int, int) = NULL;
        struct Scsi_Host *hpnt2;
        int host_no;

	if (hpnt->hostt->proc_info == NULL)
		ret = -ENOSYS;

	if (count > PROC_BLOCK_SIZE)
		return -EOVERFLOW;

	if (!(page = (char *) __get_free_page(GFP_KERNEL)))
		return -ENOMEM;
	if(copy_from_user(page, buf, count))
	{
		free_page((ulong) page);
		return -EFAULT;
	}

        lock_kernel();
        for (hpnt2 = scsi_hostlist; hpnt2; hpnt2 = hpnt2->next) {
                if (hpnt2 == hpnt) {
                        if (hpnt->hostt) {
                                host_proc_info = hpnt->hostt->proc_info;
                                host_no = hpnt->host_no;
                        }
                        break;
                }
        }

	if (host_proc_info)
                ret = (*host_proc_info)(page, &start, 0, count, host_no, 1);
        else {
                printk("%s: no entry found\n", __FUNCTION__);
                ret = -ENXIO;
        }
        unlock_kernel();

	free_page((ulong) page);
	return(ret);
}

void build_proc_dir_entries(Scsi_Host_Template * tpnt)
{
	struct Scsi_Host *hpnt;
	char name[10];	/* see scsi_unregister_host() */

	tpnt->proc_dir = proc_mkdir(tpnt->proc_name, proc_scsi);
        if (!tpnt->proc_dir) {
                printk(KERN_ERR "Unable to proc_mkdir in scsi.c/build_proc_dir_entries");
                return;
        }
	tpnt->proc_dir->owner = tpnt->module;

	hpnt = scsi_hostlist;
	while (hpnt) {
		if (tpnt == hpnt->hostt) {
			struct proc_dir_entry *p;
			sprintf(name,"%d",hpnt->host_no);
			p = create_proc_read_entry(name,
					S_IFREG | S_IRUGO | S_IWUSR,
					tpnt->proc_dir,
					proc_scsi_read,
					(void *)hpnt);
			if (!p)
				panic("Not enough memory to register SCSI HBA in /proc/scsi !\n");
			p->write_proc=proc_scsi_write;
			p->owner = tpnt->module;
		}
		hpnt = hpnt->next;
	}
}

/*
 *  parseHandle *parseInit(char *buf, char *cmdList, int cmdNum); 
 *              gets a pointer to a null terminated data buffer
 *              and a list of commands with blanks as delimiter 
 *      in between. 
 *      The commands have to be alphanumerically sorted. 
 *      cmdNum has to contain the number of commands.
 *              On success, a pointer to a handle structure
 *              is returned, NULL on failure
 *
 *      int parseOpt(parseHandle *handle, char **param);
 *              processes the next parameter. On success, the
 *              index of the appropriate command in the cmdList
 *              is returned, starting with zero.
 *              param points to the null terminated parameter string.
 *              On failure, -1 is returned.
 *
 *      The databuffer buf may only contain pairs of commands
 *          options, separated by blanks:
 *              <Command> <Parameter> [<Command> <Parameter>]*
 */

typedef struct {
	char *buf,		/* command buffer  */
	*cmdList,		/* command list    */
	*bufPos,		/* actual position */
	**cmdPos,		/* cmdList index   */
	 cmdNum;		/* cmd number      */
} parseHandle;

inline int parseFree(parseHandle * handle)
{				/* free memory     */
	kfree(handle->cmdPos);
	kfree(handle);

	return -1;
}

parseHandle *parseInit(char *buf, char *cmdList, int cmdNum)
{
	char *ptr;		/* temp pointer    */
	parseHandle *handle;	/* new handle      */

	if (!buf || !cmdList)	/* bad input ?     */
		return NULL;
	handle = (parseHandle *) kmalloc(sizeof(parseHandle), GFP_KERNEL);
	if (!handle)
		return NULL;	/* out of memory   */
	handle->cmdPos = (char **) kmalloc(sizeof(int) * cmdNum, GFP_KERNEL);
	if (!handle->cmdPos) {
		kfree(handle);
		return NULL;	/* out of memory   */
	}
	handle->buf = handle->bufPos = buf;	/* init handle     */
	handle->cmdList = cmdList;
	handle->cmdNum = cmdNum;

	handle->cmdPos[cmdNum = 0] = cmdList;
	for (ptr = cmdList; *ptr; ptr++) {	/* scan command string */
		if (*ptr == ' ') {	/* and insert zeroes   */
			*ptr++ = 0;
			handle->cmdPos[++cmdNum] = ptr++;
		}
	}
	return handle;
}

int parseOpt(parseHandle * handle, char **param)
{
	int cmdIndex = 0, cmdLen = 0;
	char *startPos;

	if (!handle)		/* invalid handle  */
		return (parseFree(handle));
	/* skip spaces     */
	for (; *(handle->bufPos) && *(handle->bufPos) == ' '; handle->bufPos++);
	if (!*(handle->bufPos))
		return (parseFree(handle));	/* end of data     */

	startPos = handle->bufPos;	/* store cmd start */
	for (; handle->cmdPos[cmdIndex][cmdLen] && *(handle->bufPos); handle->bufPos++) {	/* no string end?  */
		for (;;) {
			if (*(handle->bufPos) == handle->cmdPos[cmdIndex][cmdLen])
				break;	/* char matches ?  */
			else if (memcmp(startPos, (char *) (handle->cmdPos[++cmdIndex]), cmdLen))
				return (parseFree(handle));	/* unknown command */

			if (cmdIndex >= handle->cmdNum)
				return (parseFree(handle));	/* unknown command */
		}

		cmdLen++;	/* next char       */
	}

	/* Get param. First skip all blanks, then insert zero after param  */

	for (; *(handle->bufPos) && *(handle->bufPos) == ' '; handle->bufPos++);
	*param = handle->bufPos;

	for (; *(handle->bufPos) && *(handle->bufPos) != ' '; handle->bufPos++);
	*(handle->bufPos++) = 0;

	return (cmdIndex);
}

void proc_print_scsidevice(Scsi_Device * scd, char *buffer, int *size, int len)
{

	int x, y = *size;
	extern const char *const scsi_device_types[MAX_SCSI_DEVICE_CODE];

#if defined(CONFIG_DTV) || defined(CONFIG_DTGNEXUS) || defined(CONFIG_DTG3G)
	y = sprintf(buffer + len,
	     "Host: scsi%d Channel: %02d Id: %02d Lun: %02d Dev: %s\n  Vendor: ",
		    scd->host->host_no, scd->channel, scd->id, scd->lun, scd->device_name); /* FES patch */
#else
	y = sprintf(buffer + len,
	     "Host: scsi%d Channel: %02d Id: %02d Lun: %02d\n  Vendor: ",
		    scd->host->host_no, scd->channel, scd->id, scd->lun);
#endif
	for (x = 0; x < 8; x++) {
		if (scd->vendor[x] >= 0x20)
			y += sprintf(buffer + len + y, "%c", scd->vendor[x]);
		else
			y += sprintf(buffer + len + y, " ");
	}
	y += sprintf(buffer + len + y, " Model: ");
	for (x = 0; x < 16; x++) {
		if (scd->model[x] >= 0x20)
			y += sprintf(buffer + len + y, "%c", scd->model[x]);
		else
			y += sprintf(buffer + len + y, " ");
	}
	y += sprintf(buffer + len + y, " Rev: ");
	for (x = 0; x < 4; x++) {
		if (scd->rev[x] >= 0x20)
			y += sprintf(buffer + len + y, "%c", scd->rev[x]);
		else
			y += sprintf(buffer + len + y, " ");
	}
	y += sprintf(buffer + len + y, "\n");

	y += sprintf(buffer + len + y, "  Type:   %s ",
		     (0 <= scd->type && scd->type < MAX_SCSI_DEVICE_CODE) ?
	       scsi_device_types[(int) scd->type] : "Unknown          ");
	y += sprintf(buffer + len + y, "               ANSI"
		     " SCSI revision: %02x", (scd->scsi_level - 1) ? scd->scsi_level - 1 : 1);
	if (scd->scsi_level == 2)
		y += sprintf(buffer + len + y, " CCS\n");
	else
		y += sprintf(buffer + len + y, "\n");

	*size = y;
	return;
}


#ifdef CONFIG_SCSI_PROC_EVENT /* FES patch */
#include <linux/scsi_proc_event.h>
/**************************************************************************/
/* /proc/scsi/event                                                       */
/**************************************************************************/

#define EVBUF_SIZE 100
#define LINEBUF_SIZE 80

struct proc_evbuf {
	struct list_head buf_list;
	struct proc_evbuf_log {
		char string[80];
	} log[EVBUF_SIZE];
	int head;
	int tail;
	loff_t ppos;
	wait_queue_head_t ev_wait;
	char linebuf[LINEBUF_SIZE];
	int linebuf_size;
	int linebuf_pos;
	struct semaphore ev_sem;
};

static LIST_HEAD(proc_evbuf_list);
static struct semaphore proc_evbuf_list_sem;

/* FES5 patch > */
static scsi_proc_event_handler spe_handler = NULL;

int scsi_proc_event_register(scsi_proc_event_handler handler)
{
	down(&proc_evbuf_list_sem);
	if (spe_handler) {
		up(&proc_evbuf_list_sem);
		return -EINVAL;
	}
	spe_handler = handler;
	up(&proc_evbuf_list_sem);
	return 0;
}

int scsi_proc_event_unregister(scsi_proc_event_handler handler)
{
	down(&proc_evbuf_list_sem);
	if (spe_handler != handler) {
		up(&proc_evbuf_list_sem);
		return -EINVAL;
	}
	spe_handler = NULL;
	up(&proc_evbuf_list_sem);

	return 0;
}
/* < FES5 patch */

static size_t copy_linebuf_to_user(char *buf, size_t len, struct proc_evbuf *evbuf)
{
	int clen;

	clen = evbuf->linebuf_size - evbuf->linebuf_pos;
	if( clen == 0 )
		return 0;
	if( clen > len )
		clen = len;
	__copy_to_user(buf, evbuf->linebuf + evbuf->linebuf_pos, clen);
	evbuf->linebuf_pos += clen;
	return clen;
}

static int fill_linebuf(struct proc_evbuf *evbuf)
{
	int blen, pos;

	blen = evbuf->linebuf_size - evbuf->linebuf_pos;
	if( blen > 0 )
		return blen;
	if( evbuf->head == evbuf->tail )
		return 0;
	pos = evbuf->tail;

	blen = sprintf(evbuf->linebuf, "%s", evbuf->log[pos].string);

	evbuf->tail = (pos + 1) % EVBUF_SIZE;
	evbuf->linebuf_size = blen;
	evbuf->linebuf_pos = 0;

	return blen;
}

static ssize_t scsi_proc_event_read(struct file *file, char *buf, size_t len, loff_t *ppos)
{
	struct proc_evbuf *evbuf;
	size_t ulen;
	int rval;

	evbuf = (struct proc_evbuf *)file->private_data;
	scsi_event_printk("rdchk_event_read: len=%d ppos=%d(%d)\n", len, (int)*ppos, (int)evbuf->ppos);
	if( *ppos != evbuf->ppos )
		return -EIO;
	if( !access_ok(VERIFY_READ, buf, len) )
		return -EFAULT;

	down(&evbuf->ev_sem);
	ulen = copy_linebuf_to_user(buf, len, evbuf);
	while( ulen < len ){
		rval = fill_linebuf(evbuf);
		scsi_event_printk("rdchk_event_read: ulen=%d fill=%d\n", ulen, rval);
		if( rval > 0 )
			ulen += copy_linebuf_to_user(buf + ulen, len - ulen, evbuf);
		else if( ulen > 0 )
			break;
		else {
			up(&evbuf->ev_sem);
			if( file->f_flags & O_NONBLOCK )
				return -EAGAIN;
			scsi_event_printk("rdchk_event_read: sleep\n");
			interruptible_sleep_on(&evbuf->ev_wait);
			scsi_event_printk("rdchk_event_read: wakeup\n");
			if (signal_pending(current))
				return -ERESTARTSYS;
			down(&evbuf->ev_sem);
		}
	}
	*ppos += ulen;
	evbuf->ppos = *ppos;
	up(&evbuf->ev_sem);
	scsi_event_printk("rdchk_event_read: read_size=%d\n", ulen);
	return ulen;
}

static int scsi_proc_event_open(struct inode * inode, struct file * file)
{
	struct proc_evbuf *evbuf;

	evbuf = (struct proc_evbuf *)kmalloc(sizeof(struct proc_evbuf), GFP_KERNEL);
	if( evbuf == NULL )
		return -ENOMEM;
	evbuf->head = evbuf->tail = 0;
	evbuf->ppos = 0;
	evbuf->linebuf_size = evbuf->linebuf_pos = 0;
	init_waitqueue_head(&evbuf->ev_wait);
	sema_init(&evbuf->ev_sem, 1);
	down(&proc_evbuf_list_sem);
	list_add(&evbuf->buf_list, &proc_evbuf_list);
	up(&proc_evbuf_list_sem);
	file->private_data = evbuf;
	return 0;
}

static int scsi_proc_event_release(struct inode * inode, struct file * file)
{
	struct proc_evbuf *evbuf;

	evbuf = (struct proc_evbuf *)file->private_data;
	down(&proc_evbuf_list_sem);
	list_del(&evbuf->buf_list);
	up(&proc_evbuf_list_sem);
	kfree(evbuf);
	return 0;
}

static struct file_operations event_fops = {
	read:       scsi_proc_event_read,
	open:       scsi_proc_event_open,
	release:    scsi_proc_event_release
};

static struct proc_dir_entry *proc_event;

int __init scsi_proc_event_init(struct proc_dir_entry *parent)
{
	sema_init(&proc_evbuf_list_sem, 1);
	proc_event = create_proc_entry("event", S_IRUGO, parent);
	if( proc_event )
		proc_event->proc_fops = &event_fops;
	return 0;
}

void scsi_proc_event_exit(void)
{
	if( proc_event )
		remove_proc_entry("event", NULL);
}

void scsi_proc_event_notify(Scsi_Device * scd, int flag)
{
	struct list_head *list;
	struct proc_evbuf *evbuf;
	int x, y, next;
	char *buffer;

	down(&proc_evbuf_list_sem);

	/* FES5 patch > */
	if (spe_handler) {
		struct scsi_proc_event_data data;
                data.flag = flag;
#if defined(CONFIG_DTV) || defined(CONFIG_DTGNEXUS) || defined(CONFIG_DTG3G)
                memcpy(data.device_name, scd->device_name, sizeof(data.device_name));
#else
                memset(data.device_name, 0, sizeof(data.device_name));
#endif
                data.host_no = scd->host->host_no;
                data.channel = scd->channel;
                data.id      = scd->id;
                data.lun     = scd->lun;
		memcpy(data.vendor, scd->vendor, 8);
		memcpy(data.model, scd->model, 16);

		(*spe_handler)(&data);
	}
	/* < FES5 patch */

	for( list = proc_evbuf_list.next; list != &proc_evbuf_list; list = list->next ){
		evbuf = list_entry(list, struct proc_evbuf, buf_list);
		down(&evbuf->ev_sem);

		next = (evbuf->head + 1) % EVBUF_SIZE;
		if ( evbuf->tail == next )      /* buffer full */
			evbuf->tail = (evbuf->tail + 1) % EVBUF_SIZE;

		buffer = evbuf->log[evbuf->head].string;
#if defined(CONFIG_DTV) || defined(CONFIG_DTGNEXUS) || defined(CONFIG_DTG3G)
		y = sprintf(buffer, "%c,%s,%d,%d,%d,%d,", flag, scd->device_name, scd->host->host_no, scd->channel, scd->id, scd->lun);
#else
		y = sprintf(buffer, "%c,%s,%d,%d,%d,%d,", flag, "", scd->host->host_no, scd->channel, scd->id, scd->lun);
#endif
		for (x = 0; x < 8; x++) {
			if (scd->vendor[x] >= 0x20)
				y += sprintf(buffer + y, "%c", scd->vendor[x]);
			else
				y += sprintf(buffer + y, " ");
		}
		y += sprintf(buffer + y, ",");
		for (x = 0; x < 16; x++) {
			if (scd->model[x] >= 0x20)
				y += sprintf(buffer + y, "%c", scd->model[x]);
			else
				y += sprintf(buffer + y, " ");
		}
		y += sprintf(buffer + y, "\n");

		scsi_event_printk(evbuf->log[evbuf->head].string);

		evbuf->head = next;

		up(&evbuf->ev_sem);
		wake_up(&evbuf->ev_wait);
	}
	up(&proc_evbuf_list_sem);
}
#endif /* CONFIG_SCSI_PROC_EVENT */


#else				/* if !CONFIG_PROC_FS */

void proc_print_scsidevice(Scsi_Device * scd, char *buffer, int *size, int len)
{
}

#endif				/* CONFIG_PROC_FS */

/*
 * Overrides for Emacs so that we get a uniform tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 4
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -4
 * c-argdecl-indent: 4
 * c-label-offset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */
