/*
 * NBL boot config driver SDMWL version
 *
 * $Id: nblcfg_sdmwl.c,v 1.6 2006/07/11 09:20:33 satoru-i Exp $
 */
/* With non GPL files, use following license */
/*
 * Copyright 2001-2006 Sony Corporation.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/* Otherwise with GPL files, use following license */
/*
 * Copyright 2001-2006 Sony Corporation.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License.
 *
 * THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 * WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 * USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * You should have received a copy of the  GNU General Public License along
 * with this program; if not, write  to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/version.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/spinlock.h>
#include <linux/miscdevice.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/sdmwl.h>
#include <linux/nblcfg_ioctl.h>
#include "nblcfg_sdmwl.h"

#ifndef NBLCFG_MINOR
#define NBLCFG_MINOR                243
#endif

#define MAX_LOAD_RETRY        2

static struct nblcfg_data sdmwlbuf SDMWL_CACHELINE_ALIGNED;
static struct nblcfg_data nblconfig SDMWL_CACHELINE_ALIGNED;
static struct sdmwl_desc wldesc;

/* single-open device */
static volatile int nblcfg_open_cnt = 0;
static spinlock_t nblcfg_lock;

static u_int16_t
nblcfg_calc_cksum(struct nblcfg_data *config)
{
        int i;
        u_int16_t sum = 0;

        for (i = 0; i < NBLCONFIG_SIZE - sizeof(u_int16_t); i++) {
                sum += ((u_int8_t *)config)[i];
        }

        return sum;
}

static int
nblcfg_isvalid(void *buf)
{
	struct nblcfg_data *config = (struct nblcfg_data *)buf;

	if (config->magic != NBLCONFIG_MAGIC)
		return 0;

	if (config->cksum != nblcfg_calc_cksum(config))
		return 0;

	return 1;
}

static void
nblcfg_setcksum(void *buf)
{
	struct nblcfg_data *config = (struct nblcfg_data *)buf;

	config->cksum = nblcfg_calc_cksum(config);
}

static int
nblcfg_read_config(void)
{
	int ret;
	int retry = 0;

read_retry:
	ret = sdmwl_read_chunk(&wldesc, 0, &nblconfig);

	if (ret != SDMWL_ERR_OK && ++retry < MAX_LOAD_RETRY)
		goto read_retry;

	if (ret == SDMWL_ERR_EINVALCHUNK) {
		INFO("# no valid nblconfig found, initializing...\n");
		memset(&nblconfig, 0xff, sizeof(struct nblcfg_data));
		nblconfig.magic = NBLCONFIG_MAGIC;
		nblconfig.boot_stat = BOOT_STAT_BOOTDONE;
		nblconfig.kern_sel = KERN_SEL_RECOVERY;
		nblconfig.retry = 0;
	} else if (ret != SDMWL_ERR_OK) {
		ERROR("# read error while reading nblconfig\n");
		return -1;
	}

	return 0;
}

static int 
nblcfg_write_config(void)
{
	if (sdmwl_write_chunk(&wldesc, &nblconfig) != SDMWL_ERR_OK) {
		ERROR("# failed to write nblconfig\n");
		return -1;
	}
	return 0;
}

#ifdef DEBUG
static void
nblcfg_dump_config(void)
{
        int i;
        u_int8_t *buf;

	buf = (u_int8_t *)nblconfig;

        for (i = 0; i < NBLCONFIG_SIZE; i++) {
                if ((i % 16) == 0)
                        PRINT("\n0x%02x: ", i);
                PRINT("%02x ", buf[i]);
        }
        PRINT("\n");
}
#endif

static int 
nblcfg_ioctl(struct inode *inode, struct file *file,
             unsigned int cmd, unsigned long arg)
{
        unsigned int minor = MINOR(inode->i_rdev);
        int rval = 0;
        u_int8_t sel;

        if (minor != NBLCFG_MINOR)
                return -ENODEV;

        switch(cmd) {
        case NBLCFG_IOC_BOOTDONE:
 		nblconfig.boot_stat = BOOT_STAT_BOOTDONE;
                break;
        case NBLCFG_IOC_KERNSEL:
                sel = arg & 0xff;
                if ((sel != KERN_SEL_RECOVERY) && (sel != KERN_SEL_NORMAL)) {
                        rval = -EINVAL;
                        break;
                }
		nblconfig.kern_sel = sel;
                break;
#ifdef DEBUG
        case NBLCFG_IOC_DUMP:
		nblcfg_dump_config();
                break;
#endif
        default:
                rval = -ENOIOCTLCMD;
        }

        return rval;
}

static int 
nblcfg_open(struct inode *inode, struct file *file)
{
	struct sdmwl_chunk_info chinfo;
        unsigned int minor = MINOR(inode->i_rdev);
	int ret;

        if (minor != NBLCFG_MINOR)
                return -ENODEV;

        spin_lock(&nblcfg_lock);

        if (nblcfg_open_cnt) {
                spin_unlock(&nblcfg_lock);
                return -EBUSY;
        }
        nblcfg_open_cnt++;

        spin_unlock(&nblcfg_lock);

	chinfo.chunk_size = NBLCONFIG_CHUNK_SIZE;
	chinfo.color_offset = (unsigned int)&((struct nblcfg_data*)0)->color;
	chinfo.isvalid = nblcfg_isvalid;
	chinfo.setcrc = nblcfg_setcksum;

	ret = sdmwl_open(NBLCFG_SDMPART, &chinfo, &sdmwlbuf, &wldesc);

	if (ret != SDMWL_ERR_OK) {
		ERROR("# sdmwl_open() failed for '%s'.\n", NBLCFG_SDMPART);
                ret = -ENODEV;
		goto err_exit;
	}
	if (nblcfg_read_config() != 0) {
                ret = -EIO;
		goto err_close_exit;
	}

        MOD_INC_USE_COUNT;
        return 0;

err_close_exit:
	sdmwl_close(&wldesc);
err_exit:
	spin_lock(&nblcfg_lock);
	nblcfg_open_cnt--;
	spin_unlock(&nblcfg_lock);

	return ret;
}

static int
nblcfg_release(struct inode *inode, struct file *file)
{
	int ret;

	ret = 0;

	if (nblcfg_write_config() != 0) {
		ret = -EIO;
	}
	sdmwl_close(&wldesc);

	spin_lock(&nblcfg_lock);
	nblcfg_open_cnt--;
	spin_unlock(&nblcfg_lock);

        MOD_DEC_USE_COUNT;
        return ret;
}

static struct file_operations nblcfg_fops =
{
	owner:		THIS_MODULE,
	ioctl:		nblcfg_ioctl,
	open:		nblcfg_open,
	release:	nblcfg_release,
};

static struct miscdevice nblcfg_miscdev =
{
	NBLCFG_MINOR,
	"nblcfg",
	&nblcfg_fops
};

static int __init
nblcfg_sdmwl_init(void)
{
        int rval;

        spin_lock_init(&nblcfg_lock);

        /* register as a misc driver */
        if ((rval = misc_register(&nblcfg_miscdev)) < 0) {
                printk("cannot register nblcfg as a misc driver\n");
                return rval;
        }
	return 0;
}

static void __exit
nblcfg_sdmwl_cleanup(void)
{
        misc_deregister(&nblcfg_miscdev);
}

module_init(nblcfg_sdmwl_init);
module_exit(nblcfg_sdmwl_cleanup);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,17)
MODULE_LICENSE("GPL");
#endif
