/*
 * Copyright 2004 Sony Corporation.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License.
 *
 * THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 * WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 * USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * You should have received a copy of the  GNU General Public License along
 * with this program; if not, write  to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/sched.h>
#include <linux/spinlock.h>
#include <linux/proc_fs.h>
#include <linux/time.h>
#include <linux/sdmwl.h>
#include <linux/sdmlog.h>
#include <asm/uaccess.h>

#define DEBUG
#define DEBUG_LEVEL DPL_NOTICE

#ifdef DEBUG
#define DPR(fmt , args...)	printk(fmt, ## args)
#endif /* DEBUG */

/* ======================================================================= *
 * constant                                                                *
 * ======================================================================= */

#define SDMLOG_PROC_NAME	"sdmlog"

#ifdef CONFIG_SDMLOG_RECORD_SIZE
#define SDMLOG_RECORD_SIZE	CONFIG_SDMLOG_RECORD_SIZE
#else
#define SDMLOG_RECORD_SIZE	4096
#endif

/* ======================================================================= *
 * debug print function from dpr package                                   *
 * ======================================================================= */

#define DPL_EMERG	0
#define DPL_ALERT	1
#define DPL_CRIT	2
#define DPL_ERR		3
#define DPL_WARNING	4
#define DPL_NOTICE	5
#define DPL_INFO	6
#define DPL_DEBUG	7

#define DPM_PURPOSE	0x0000000f
#define DPM_ERR		0x00000001
#define DPM_TRACE	0x00000002
#define DPM_LAYER	0x000000f0
#define DPM_APPLICATION	0x00000010
#define DPM_MIDDLEWARE	0x00000020
#define DPM_DRIVER	0x00000040
#define DPM_KERNEL	0x00000080
#define DPM_USERDEF	0xffff0000

#ifdef DEBUG

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL DPL_DEBUG
#endif
#ifndef DEBUG_MASK
#define DEBUG_MASK  DPM_PURPOSE
#endif

#ifndef DPR
#include <stdio.h>
#define DPR(fmt, args...)	fprintf(stderr, fmt , ## args)
#endif

#define DPRL(level, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && DPR(fmt , ## args))

#define DPRM(mask, fmt, args...) \
	(void)(((mask) & DEBUG_MASK) && DPR(fmt , ## args))
#define DPRLM(level, mask, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && ((mask) & DEBUG_MASK) && \
		DPR(fmt , ## args))
#define DPRC(cond, fmt, args...) \
	(void)((cond) && DPR(fmt , ## args))

#define DPRF(fmt, args...)	DPR("%s: " fmt, __FUNCTION__ , ## args)
#define DPRFL(level, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && DPRF(fmt , ## args))
#define DPRFM(mask, fmt, args...) \
	(void)(((mask) & DEBUG_MASK) && DPRF(fmt , ## args))
#define DPRFLM(level, mask, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && ((mask) & DEBUG_MASK) && \
		DPRF(fmt , ## args))
#define DPRFC(cond, fmt, args...) \
	(void)((cond) && DPRF(fmt , ## args))

#define	DPRFS	DPRFLM(DPL_DEBUG, DPM_TRACE, "start\n")
#define	DPRFE	DPRFLM(DPL_DEBUG, DPM_TRACE, "end\n")
#define DPRFR	DPRFLM(DPL_DEBUG, DPM_TRACE, "return\n")

#else  /* DEBUG */

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL -1
#endif
#ifndef DEBUG_MASK
#define DEBUG_MASK  0
#endif

#ifndef DPR
#define DPR(fmt, args...)
#endif

#define DPRL(level, fmt, args...)
#define DPRM(mask, fmt, args...)
#define DPRLM(level, mask, fmt, args...)
#define DPRC(cond, fmt, args...)
#define DPRF(fmt, args...)
#define DPRFL(level, fmt, args...)
#define DPRFM(mask, fmt, args...)
#define DPRFLM(level, mask, fmt, args...)
#define DPRFC(cond, fmt, args...)

#define	DPRFS
#define	DPRFE
#define DPRFR

#endif /* DEBUG */

/* ======================================================================= *
 * prototype declaration                                                   *
 * ======================================================================= */

static int sdmlog_proc_init(void);
static void sdmlog_proc_exit(void);

static int sdmlog_proc_open(struct inode *inode, struct file *filp);
static int sdmlog_proc_release(struct inode *inode, struct file *filp);
static ssize_t sdmlog_proc_write(struct file *filp, const char *buf,
				 size_t count, loff_t *f_pos);
static ssize_t sdmlog_proc_read(struct file *filp, char *buf,
				size_t count, loff_t *f_pos);
static int sdmlog_proc_ioctl(struct inode *inode, struct file *filp,
			     unsigned int cmd, unsigned long arg);

/* ======================================================================= *
 * global variable                                                         *
 * ======================================================================= */

static struct file_operations sdmlog_proc_fops = {
	.open		= sdmlog_proc_open,
	.release	= sdmlog_proc_release,
	.write		= sdmlog_proc_write,
	.read		= sdmlog_proc_read,
	.ioctl		= sdmlog_proc_ioctl,
};

DECLARE_MUTEX(sdmlog_semlock);
static DECLARE_MUTEX(sdmlog_buf_semlock);
static spinlock_t sdmlog_irqlock = SPIN_LOCK_UNLOCKED;
unsigned long sdmlog_flags_saved;

static char sdmlog_buf[SDMLOG_RECORD_SIZE];

/* ======================================================================= *
 * Linux kernel interface                                                  *
 * ======================================================================= */

/* ------------------------ *
 * init/finalization        *
 * ------------------------ */

static int
sdmlog_proc_init(void)
{
	struct proc_dir_entry *ent;
	DPRFS;

	ent = create_proc_entry(SDMLOG_PROC_NAME, S_IRUGO | S_IWUSR, NULL);
	if (ent == NULL) {
		DPRFL(DPL_WARNING, "cannot create proc file.\n");
		return -ENOMEM;
	}
	ent->proc_fops = &sdmlog_proc_fops;

	DPRFE;
	return 0;
}

static void
sdmlog_proc_exit(void)
{
	DPRFS;

	remove_proc_entry(SDMLOG_PROC_NAME, NULL);

	DPRFE;
	return;
}

module_init(sdmlog_proc_init);
module_exit(sdmlog_proc_exit);

/* ------------------------ *
 * file operations          *
 * ------------------------ */

static int
sdmlog_proc_open(struct inode *inode, struct file *filp)
{
	DPRFS;

	MOD_INC_USE_COUNT;

	DPRFE;
	return 0;
}

static int
sdmlog_proc_release(struct inode *inode, struct file *filp)
{
	DPRFS;

	MOD_DEC_USE_COUNT;

	DPRFE;
	return 0;
}

static ssize_t
sdmlog_proc_write(struct file *filp, const char *buf,
		  size_t count, loff_t *f_pos)
{
	int res;
	size_t ncopy;
	DPRFS;

	down(&sdmlog_buf_semlock);

	ncopy = sizeof(sdmlog_buf) - 1;
	if (count < ncopy) {
		ncopy = count;
	}
	res = copy_from_user(sdmlog_buf, buf, ncopy);
	if (res != 0) {
		up(&sdmlog_buf_semlock);
		DPRFR;
		return -ESDMLOG_INVAL;
	}
	sdmlog_buf[ncopy] = 0;

	res = sdmlog_add(sdmlog_buf);
	if (res < 0) {
		up(&sdmlog_buf_semlock);
		DPRFR;
		return res;
	}

	up(&sdmlog_buf_semlock);

	DPRFE;
	return count;
}

static ssize_t
sdmlog_proc_read(struct file *filp, char *buf,
		 size_t count, loff_t *f_pos)
{
	int res;
	int pos;
	DPRFS;

	res = access_ok(VERIFY_WRITE, buf, count);
	if (!res) {
		DPRFR;
		return -EFAULT;
	}
	if (count == 0) {
		DPRFR;
		return 0;
	}

	pos = *f_pos;
	res = sdmlog_read(buf, count, &pos);
	*f_pos = pos;

	DPRFE;
	return res;
}

static int
sdmlog_proc_ioctl(struct inode *inode, struct file *filp,
		  unsigned int cmd, unsigned long arg)
{
	int res;
	char *argp;
	char c;
	size_t ncopy;
	DPRFS;

	down(&sdmlog_buf_semlock);

	argp = (char *)arg;
	for (ncopy = 0; ncopy < sizeof(sdmlog_buf) - 1; ncopy++) {
		res = get_user(c, argp + ncopy);
		if (res != 0) {
			up(&sdmlog_buf_semlock);
			DPRFR;
			return -ESDMLOG_INVAL;
		}
		if (ncopy > 0 && c == 0) {
			break;
		}
		sdmlog_buf[ncopy] = c;
	}
	sdmlog_buf[ncopy] = 0;

	if (sdmlog_buf[0] == 0 && ncopy > 0) {
		if (strcmp(sdmlog_buf + 1, ":sync") == 0) {
			res = sdmlog_sync();
		} else if (strcmp(sdmlog_buf + 1, ":clear") == 0) {
			res = sdmlog_clear();
		} else if (strcmp(sdmlog_buf + 1, ":show") == 0) {
			res = sdmlog_show();
		} else {
			res = -ESDMLOG_INVAL;
		}
	} else {
		res = sdmlog_add_async(sdmlog_buf);
	}

	up(&sdmlog_buf_semlock);

	DPRFE;
	return res;
}

/* ------------------------ *
 * for sdmlog               *
 * ------------------------ */

void
sdmlog_intr_lock(void)
{
	unsigned long flags;
	DPRFS;

	spin_lock_irqsave(&sdmlog_irqlock, flags);
	sdmlog_flags_saved = flags;

	DPRFE;
	return;
}

void
sdmlog_intr_unlock(void)
{
	unsigned long flags;
	DPRFS;

	flags = sdmlog_flags_saved;
	spin_unlock_irqrestore(&sdmlog_irqlock, flags);

	DPRFE;
	return;
}

uint32_t
sdmlog_gettime(void)
{
	struct timeval tv;
	DPRFS;

	do_gettimeofday(&tv);

	DPRFE;
	return tv.tv_sec;
}

/* ======================================================================= *
 * export sdmlog functions                                                 *
 * ======================================================================= */

EXPORT_SYMBOL(sdmlog_open);
EXPORT_SYMBOL(sdmlog_close);
EXPORT_SYMBOL(sdmlog_add);
EXPORT_SYMBOL(sdmlog_add_async);
EXPORT_SYMBOL(sdmlog_add_intr);
EXPORT_SYMBOL(sdmlog_clear);
EXPORT_SYMBOL(sdmlog_sync);
EXPORT_SYMBOL(sdmlog_show);
EXPORT_SYMBOL(sdmlog_read);
