/*
 * SDM wearleveling library MTD glue layer
 *
 * $Id: sdmwl_mtd.c,v 1.6 2004/05/17 13:26:10 nagahama Exp $
 */
/*
 * Copyright 2004 Sony Corporation.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License.
 *
 * THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 * WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 * USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * You should have received a copy of the  GNU General Public License along
 * with this program; if not, write  to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/version.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/sdmparse.h>
#include <linux/sdmwl.h>
#ifdef CONFIG_SDMLOG
#include <linux/sdmlog.h>
#endif

static int
sdmwl_mtd_open(const char *name, void **priv)
{
	struct mtd_info *mtd;
	int i;

	for (i = 0; i < MAX_MTD_DEVICES; i++) {
		mtd = __get_mtd_device(NULL, i);
		if (mtd == NULL)
			continue;
		if (strncmp(mtd->name, name, strlen(name) + 1) == 0)
			break;
	}
	if (i == MAX_MTD_DEVICES)
		return -1;

	mtd = get_mtd_device(mtd, i);
	if (mtd == NULL)
		return -1;

	*(struct mtd_info **)priv = mtd;
	return 0;
}

static int
sdmwl_mtd_close(void *priv)
{
	struct mtd_info *mtd = (struct mtd_info *)priv;

	if (mtd->sync != NULL)
		mtd->sync(mtd);

	put_mtd_device(mtd);

	return 0;
}

static int
sdmwl_mtd_read(void *priv, void *buf, unsigned int offset, unsigned int length)
{
	struct mtd_info *mtd = (struct mtd_info *)priv;
	int retlen;

	if ((mtd->read(mtd, offset, length, &retlen, buf) != 0) ||
	    (retlen != length)) {
		return -1;
	}

	return 0;
}

static int
sdmwl_mtd_write(void *priv, const void *buf, unsigned int offset,
		unsigned int length)
{
	struct mtd_info *mtd = (struct mtd_info *)priv;
	int retlen;

	if ((mtd->write(mtd, offset, length, &retlen, buf) != 0) ||
	    (retlen != length)) {
		return -1;
	}

	if (mtd->sync != NULL)
		mtd->sync(mtd);

	return 0;
}

static int
sdmwl_mtd_erase(void *priv, unsigned int offset, unsigned int sectors)
{
	struct erase_info instr;
	struct mtd_info *mtd = (struct mtd_info *)priv;

	instr.mtd = mtd;
	instr.addr = offset;
	instr.len = sectors * mtd->erasesize;
	instr.callback = NULL;
	
	if (mtd->erase(mtd, &instr) != 0) {
		return -1;
	}
	if (instr.state != MTD_ERASE_DONE) {
		printk(KERN_WARNING "erase from 0x%x to 0x%x finished, "
		       "but state is not MTD_ERASE_DONE (%d)\n", offset,
		       offset + sectors * mtd->erasesize, instr.state);
		return -1;
	}

	return 0;
}

static int
sdmwl_mtd_query_sector_size(void *priv)
{
	struct mtd_info *mtd = (struct mtd_info *)priv;

	if (mtd == NULL)
		return -1;

	return mtd->erasesize;
}

static int
sdmwl_mtd_query_num_sectors(void *priv)
{
	struct mtd_info *mtd = (struct mtd_info *)priv;

	if (mtd == NULL)
		return -1;

	return mtd->size / mtd->erasesize;
}

static struct sdmwl_dev_ops dev_ops = {
	sdmwl_mtd_open,
	sdmwl_mtd_close,
	sdmwl_mtd_read,
	sdmwl_mtd_write,
	sdmwl_mtd_erase,
	sdmwl_mtd_query_sector_size,
	sdmwl_mtd_query_num_sectors
};

static int __init
sdmwl_mtd_init(void)
{
	int ret;

	if (init_sdmparse() < 0)
		return -1;

	ret = sdmwl_register_dev(&dev_ops);
	if (ret != SDMWL_ERR_OK) {
		printk(KERN_ERR "sdmwl_register_dev() failed : err = %d\n",
		       ret);
		return -1;
	}

#ifdef CONFIG_SDMLOG
	ret = sdmlog_open();
	if (ret < 0) {
		printk(KERN_ERR "sdmlog_open() failed : err = %d\n", ret);
		return -1;
	}
#endif /* CONFIG_SDMLOG */

	return 0;
}

module_init(sdmwl_mtd_init);

#ifdef MODULE
MODULE_AUTHOR("Sony NSC");
MODULE_DESCRIPTION("SDM wearleveling library MTD glue module");
#endif /* MODULE */
