/*
 *  linux/drivers/char/serial_dragonball_console.c
 *
 *  Driver for DragonBall serial ports
 *
 *  Copyright 2002 Sony Corporation.
 *
 *  Original source based on linux/drivers/char/serial_sa1100.c
 *  Based on drivers/char/serial.c, by Linus Torvalds, Theodore Ts'o.
 *
 *  Copyright (C) 2000 Deep Blue Solutions Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/circ_buf.h>
#include <linux/serial.h>
#include <linux/console.h>
#include <linux/sysrq.h>
#include <linux/wait.h>
#include <linux/pm.h>
#include <linux/pm_id.h>
#include <linux/proc_fs.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>
#include <asm/hardware.h>
#include <asm/pgalloc.h>

#if defined(CONFIG_ARCH_DRAGONBALL)
#include <asm/arch/platform.h>
#include <asm/arch/gpio.h>
#include <asm/arch/mpu110_series.h>
#include <asm/arch/irq.h>
#endif

#if defined(CONFIG_SERIAL_DRAGONBALL_CONSOLE) && defined(CONFIG_MAGIC_SYSRQ)
#define SUPPORT_SYSRQ
#endif

#include <linux/serial_core.h>

/* set permutations of:
   0 : disables all debug output
   1 : Print output. (Default)
   2 : More verbose, Dump registers
   4 : Print entry of functions.
   8 : More verbose
 */
#define UART_DEBUG (1)           /* Set this to debug */

#define DEBUG_NAME "UART: "

#if UART_DEBUG & 1
#define PRINTK(format, args...) printk(KERN_INFO DEBUG_NAME "%s():%d  " format , __FUNCTION__ , __LINE__ , ## args)
#else
#define PRINTK(format, args...) 
#endif

#if UART_DEBUG & 2
#define PRINTK2(format, args...) printk(DEBUG_NAME "%s():%d  " format , __FUNCTION__ , __LINE__ , ## args)
#define DUMP_REGS(port) ({                                                     \
    })
#else
#define PRINTK2(format, args...) 
#define DUMP_REGS(port)
#endif

#if UART_DEBUG & 4
#define ENTRY() printk(DEBUG_NAME "%s():%d\n", __FUNCTION__ , __LINE__ )
#else
#define ENTRY() 
#endif

#if UART_DEBUG & 8
#define VERBOSE(format, args...) printk(DEBUG_NAME "%s():%d  " format , __FUNCTION__ , __LINE__ , ## args)
#else
#define VERBOSE(format, args...)
#endif

/* We've been assigned a range on the "Low-density serial ports" major */
#define SERIAL_DRAGONBALL_MAJOR	        204
#define CALLOUT_DRAGONBALL_MAJOR	205
#define MINOR_START		          5

#define NR_PORTS		          2

/* default FIFO settings */
#define TX_LVL   2
#define RX_LVL   24            /* 4,2 */
#define CTS_LVL  24            /* 4,2 */

#define RX_TIMER 5             /* how much jiffy to wait  */

#define TX_MAX   16            /* 4 */

static int irq_active[2];

/* 
   Uart I/O functions
 */
static inline unsigned int uart_in(struct uart_port *port, u_int offset)
{
	return inl(port->mapbase + offset);
}

static inline void uart_out(struct uart_port *port, u_int offset, unsigned int value)
{
	outl(value, port->mapbase + offset);
}

static inline void uart_putc(struct uart_port *port, char c)
{
	outl((unsigned int)(c & 0xff), port->mapbase + UART_TXDATA);
}

static inline unsigned int uart_getc(struct uart_port *port)
{
	return inl(port->mapbase + UART_RXDATA);
}

/* Test which uart */
static inline int get_uart_id(struct uart_port *port)
{
	return (port->mapbase == UART1_BASE) ? 0 : 1;
}


static struct timer_list rx_timer_list[2];
static void rx_timer(unsigned int data);

static inline void uart_init_timer(struct uart_port *port, struct timer_list *tl)
{
	if (port->mapbase == UART1_BASE) {
		init_timer(&tl[0]);
	} else {
		init_timer(&tl[1]);
	}
}

static inline int uart_del_timer(struct uart_port *port, struct timer_list *tl)
{
	if (port->mapbase == UART1_BASE) {
		return del_timer(&tl[0]);
	} else {
		return del_timer(&tl[1]);
	}
}

static inline void uart_add_timer(struct uart_port *port, struct timer_list *tl,
                                  void *func, unsigned long data, unsigned long expire)
{
	if (port->mapbase == UART1_BASE) {
		tl[0].function = func;
		tl[0].data = data;
		tl[0].expires = expire;
		add_timer(&tl[0]);
	} else {
		tl[1].function = func;
		tl[1].data = data;
		tl[1].expires = expire;
		add_timer(&tl[1]);
	}
}
    

#if UART_DEBUG & 1
/* debug /proc entry */
struct proc_dir_entry *puart1, *puart2;
static int proc_dbmx1_uart_dump(char *buf, char **start, off_t fpos,
                                int length, int *eof, void *data)
{
	int len = 0;
	struct uart_port *port = data;
	unsigned int reg;
	unsigned char tmp[80];
	int div=0;

	len += sprintf(buf+len, 
		       "UART Port %s\n", (port->mapbase == UART1_BASE) ? "1" : "2");

/* CR1 */
	reg = uart_in(port,UART_CR1);
	sprintf(tmp , "IDEN(%d)	 ",(reg & UART_ICD_MASK) >> UART_ICD_BIT);
	len += sprintf(buf+len, "UCR1=0x%08x\n", reg);
	len += sprintf(buf+len, "    %s%s%s%s%s%s%s%s%s%s%s%s%s%s\n",
		       (reg & UART_ADEN)      ? "ADEN "   : "", 
		       (reg & UART_ADBR)      ? "ADBR "   : "", 
		       (reg & UART_TRDYEN)    ? "TRDYEN " : "", 
		       (reg & UART_IDEN)      ? (char*)tmp: "", 

		       (reg & UART_RRDYEN)    ? "RRDYEN " : "", 
		       (reg & UART_RDMAEN)    ? "RDMAEN " : "", 

		       (reg & UART_IREN)      ? "IREN " : "", 
		       (reg & UART_TXEMPTYEN) ? "TXEMPTYEN " : "", 
		       (reg & UART_RTSDEN)    ? "RTSDEN " : "", 
		       (reg & UART_SNDBRK)    ? "SNDBRK " : "", 

		       (reg & UART_TDMAEN)    ? "TDMAEN " : "", 
		       (reg & UART_UARTCLKEN) ? "UARTCLKEN " : "", 
		       (reg & UART_DOZE)      ? "DOZE " : "", 
		       (reg & UART_UARTEN)    ? "UARTEN " : "");

/* CR2 */
	reg = uart_in(port,UART_CR2);
	len += sprintf(buf+len, "UCR2=0x%08x\n", reg);
	len += sprintf(buf+len, "    %s%s%s%s%sRTEC(%d) %s%s%s%s%s%s%s%s\n",
		       (reg & UART_ESCI)      ? "ESCI "   : "", 
		       (reg & UART_IRTS)      ? "IRTS "   : "",
		       (reg & UART_CTSC)      ? "CTSC "   : "",
		       (reg & UART_CTS)       ? "CTS "   : "",

		       (reg & UART_ESCEN)     ? "ESCEN "   : "",
		       (reg & UART_RTEC_MASK) >> UART_RTEC_BIT, 
		       (reg & UART_PREN)      ? "PARITY_ENABLE "   : "PARITY_NONE ",

		       (reg & UART_PREN)      ? 
                       ((reg & UART_PROE)      ? "ODD "   : "EVEN ") : "" ,
		       (reg & UART_STPB)      ? "STOP-2 "   : "STOP-1 ",
		       (reg & UART_WS)        ? "8BIT "   : "7BIT ",
		       (reg & UART_RTSEN)     ? "RTSEN "   : "",

		       (reg & UART_TXEN)      ? "TXEN "   : "",
		       (reg & UART_RXEN)      ? "RXEN "   : "",
		       (reg & UART_SRST)      ? "SRST "   : "");


/* CR3 */
	reg = uart_in(port,UART_CR3);
	len += sprintf(buf+len, "UCR3=0x%08x\n", reg);
	len += sprintf(buf+len, "    DPEC(%d) %s%s%s%s%s%s%s%s%s%s%s%s%s%s\n",
		       (reg & UART_DPEC_MASK) >> UART_DPEC_BIT,
		       (reg & UART_DTREN)      ? "DTREN "   : "",
		       (reg & UART_PARERREN)   ? "PARERREN "   : "",

		       (reg & UART_FRAERREN)   ? "FRAERREN "   : "",
		       (reg & UART_DSR)        ? "DSR "   : "",
		       (reg & UART_DCD)        ? "DCD "   : "",
		       (reg & UART_RI)         ? "RI "   : "",

		       (reg & UART_TIMEOUTEN)  ? "TIMEOUTEN "   : "",
		       (reg & UART_RXDSEN)     ? "RXDSEN "   : "",
		       (reg & UART_AIRINTEN)   ? "AIRINTEN "   : "",
		       (reg & UART_AWAKEN)     ? "AWAKEN "   : "",

		       (reg & UART_REF25)      ? "REF25 "   : "",
		       (reg & UART_REF30)      ? "REF30 "   : "",
		       (reg & UART_INVT)       ? "INVT "   : "",
		       (reg & UART_BPEN)       ? "BPEN "   : "");

/* CR4 */
	reg = uart_in(port,UART_CR4);
	len += sprintf(buf+len, "UCR4=0x%08x\n", reg);
	len += sprintf(buf+len, "    CTS trigger = %d chars\n",
		       (reg & UART_CTSTL_MASK) >> UART_CTSTL_BIT);
	len += sprintf(buf+len, "    %s%s%s%s%s%s%s%s%s\n",
		       (reg & UART_INVR)      ? "INVR "   : "",
		       (reg & UART_ENIRI)     ? "ENIRI "   : "",

		       (reg & UART_WKEN)      ? "WKEN "   : "",
		       (reg & UART_REF16)     ? "REF16 "   : "",
		       (reg & UART_IRSC)      ? "IRSC "   : "",

		       (reg & UART_TCEN)      ? "TCEN "   : "",
		       (reg & UART_BKEN)      ? "BKEN "   : "",
		       (reg & UART_OREN)      ? "OREN "   : "",
		       (reg & UART_DREN)      ? "DREN "   : "");

/* FCR */
	reg = uart_in(port,UART_UFCR);
	len += sprintf(buf+len, "UFCR=0x%08x\n", reg);
	len += sprintf(buf+len, "    TX trigger = %d chars\n",
		       (reg & UART_TXTL_MASK) >> UART_TXTL_BIT);
	len += sprintf(buf+len, "    RX trigger = %d chars\n",
		       (reg & UART_RXTL_MASK) >> UART_RXTL_BIT);
	switch (reg & UART_RFDIV_MASK) {
	case   UART_RFDIV_7:
		div = 7;
		break;
	case   UART_RFDIV_6:
		div = 6;
		break;
	case   UART_RFDIV_5:
		div = 5;
		break;
	case   UART_RFDIV_4:
		div = 4;
		break;
	case   UART_RFDIV_3:
		div = 3;
		break;
	case   UART_RFDIV_2:
		div = 2;
		break;
	case   UART_RFDIV_1:
		div = 1;
		break;
	}
	len += sprintf(buf+len, "    RFDIV = %d\n", div);

/* SR1 */
	reg = uart_in(port,UART_SR1);
	len += sprintf(buf+len, "USR1=0x%08x\n", reg);
	len += sprintf(buf+len, "    %s%s%s%s%s%s%s%s%s%s%s\n",
		       (reg & UART_PARITY_ERR)? "PARITY_ERR "   : "parity_err ",
		       (reg & UART_RTSS)      ? "RTSS "   : "rtss ",
		       (reg & UART_TRDY)      ? "TRDY "   : "trdy ",
		       (reg & UART_RTSD)      ? "RTSD "   : "rtsd ",

		       (reg & UART_ESCF)      ? "ESCF "   : "escf ",
		       (reg & UART_FRAMERR)      ? "FRAMERR "   : " framerr ",
		       (reg & UART_RRDY)      ? "RRDY "   : "rrdy ",

		       (reg & UART_TIMEOUT)      ? "TIMEOUT "   : "timeout ",
		       (reg & UART_RXDS)      ? "RXDS "   : "rxds ",
		       (reg & UART_AIRINT)      ? "AIRINT "   : "airint ",
		       (reg & UART_AWAKE)      ? "AWAKE "   : "awake ");

/* SR2 */
	reg = uart_in(port,UART_SR2);
	len += sprintf(buf+len, "USR2=0x%08x\n", reg);
	len += sprintf(buf+len, "    %s%s%s%s%s%s%s%s%s%s%s\n",
		       (reg & UART_ADET)      ? "ADET "   : "adet ",
		       (reg & UART_TXFE)      ? "TXFE "   : "txfe ",
		       (reg & UART_DTRF)      ? "DTRF "   : "dtrf ",
		       (reg & UART_IDLE)      ? "IDLE "   : "idle ",

		       (reg & UART_IRINT)     ? "IRINT "   : "irint ",

		       (reg & UART_WAKE)      ? "WAKE "   : "wake ",
		       (reg & UART_RTSF)      ? "RTSF "   : "rtsf ",

		       (reg & UART_TXDC)      ? "TXDC "   : "txdc ",
		       (reg & UART_BRCD)      ? "BRCD "   : "brcd ",
		       (reg & UART_ORE)       ? "ORE "   : "ore ",
		       (reg & UART_RDR)       ? "RDR "   : "rdr ");

	*eof = 1;
	return len;
}
#endif


/*
 * This is the size of our serial port register set.
 */
#define UART_PORT_SIZE	SZ_4K

static struct tty_driver normal, callout;
static struct tty_struct *dragonball_table[NR_PORTS];
static struct termios *dragonball_termios[NR_PORTS], *dragonball_termios_locked[NR_PORTS];
#ifdef SUPPORT_SYSRQ
static struct console dragonball_console;
#endif

/*
 * interrupts disabled on entry
 */
static void dragonball_stop_tx(struct uart_port *port, u_int from_tty)
{
	unsigned int ucr1;

	ENTRY();
	ucr1 = uart_in(port, UART_CR1);
	uart_out(port, UART_CR1, ucr1 & ~(UART_TXEMPTYEN));
}

/*
 * interrupts may not be disabled on entry
 */
static void dragonball_start_tx(struct uart_port *port, u_int nonempty, u_int from_tty)
{
	unsigned int ucr1;
	ENTRY();

	if (nonempty) {
		ucr1 = uart_in(port, UART_CR1);
		uart_out(port, UART_CR1, ucr1 | (UART_TXEMPTYEN));
	}
}

/*
 * Interrupts enabled
 */
static void dragonball_stop_rx(struct uart_port *port)
{
	unsigned int ucr1;

	ENTRY();

	ucr1 = uart_in(port, UART_CR1);
	uart_out(port, UART_CR1, ucr1 & ~UART_RRDYEN);
}

/*
 * Interrupts disabled
 */
static void dragonball_start_rx(struct uart_port *port)
{
	unsigned int ucr1;

	ENTRY();

	ucr1 = uart_in(port, UART_CR1);
	uart_out(port, UART_CR1, ucr1 | UART_RRDYEN);
}

/*
 * enable RTS delta interrupts
 */
static void dragonball_enable_ms(struct uart_port *port)
{
	ENTRY();

	uart_out(port,
		 UART_CR1,
		 uart_in(port, UART_CR1) | UART_RTSDEN);
}

/* 
 * called within interrupt handler
 */
static int
dragonball_rx_chars(struct uart_info *info)
{
	struct uart_port *port = info->port;
	struct tty_struct *tty = info->tty;
	unsigned int rx;
	int i;

	if (!tty)
		return -ENODEV;

	/* Check to not overflow buffer */
	if (tty->flip.count >= TTY_FLIPBUF_SIZE) {
		tty->flip.tqueue.routine((void *)tty);
		if (tty->flip.count >= TTY_FLIPBUF_SIZE) {
			return 1;
		}
	}

	rx = uart_getc(port);
	i = RX_LVL;
	do {
		if (tty->flip.count >= TTY_FLIPBUF_SIZE) {
			tty->flip.tqueue.routine((void *)tty);
			if (tty->flip.count >= TTY_FLIPBUF_SIZE) {
				return 1;
			}
		}
		*tty->flip.char_buf_ptr = rx & 0xff;
		*tty->flip.flag_buf_ptr = TTY_NORMAL;
		port->icount.rx++;

		if (rx & (UART_PRERR | UART_OVRRUN | UART_FRAMERR)) {
			/*	
			 * For statistics only
			 */
			if (rx & UART_PRERR) {
				port->icount.parity++;
			} else if (rx & UART_OVRRUN) {
				port->icount.overrun++;
			} else if (rx & UART_FRAMERR) {
				port->icount.frame++;
			}

			/*	
			 * Mask off conditions which should be ingored.
			 */
			rx &= port->read_status_mask;

			if (rx & UART_PRERR) {
				*tty->flip.flag_buf_ptr++ = TTY_PARITY;
			} else if (rx & UART_FRAMERR) {
				*tty->flip.flag_buf_ptr++ = TTY_FRAME;
			}
		}
		if ((rx & port->ignore_status_mask) == 0) {
			tty->flip.flag_buf_ptr++;
			tty->flip.char_buf_ptr++;
			tty->flip.count++;
		}
		if ((rx & UART_OVRRUN) &&
		    tty->flip.count < TTY_FLIPBUF_SIZE) {
			/*	
			 * Overrun is special, since it's reported
			 * immediately, and doesn't affect the current
			 * character.	
			 */
			*tty->flip.flag_buf_ptr = TTY_OVERRUN;
			tty->flip.flag_buf_ptr++;
			tty->flip.char_buf_ptr++;
			tty->flip.count++;
		}
		if (--i == 0)
			break;
	} while ((rx = uart_getc(port)) & UART_CHARRDY);

	tty_flip_buffer_push(tty);
	return 0;
}

static void dragonball_tx_chars(struct uart_info *info)
{
	struct uart_port *port = info->port;
	int len;
	ENTRY();

	if (port->x_char) {
		uart_putc(port, port->x_char);
		port->icount.tx++;
		port->x_char = 0;
		return;
	}

	if (info->xmit.head == info->xmit.tail
	    || info->tty->stopped
	    || info->tty->hw_stopped) {
		dragonball_stop_tx(info->port, 0);
		return;
	}

	/* TRDY is automatically cleared if FIFO level goes beyond threshold */
	len = CIRC_CNT(info->xmit.head, info->xmit.tail, UART_XMIT_SIZE);
	len = (len > TX_MAX) ? TX_MAX : len;
	while (len-- > 0) {
		uart_putc(port, info->xmit.buf[info->xmit.tail]);
		info->xmit.tail = (info->xmit.tail + 1) & (UART_XMIT_SIZE - 1);
		port->icount.tx++;
		if (info->xmit.head == info->xmit.tail)
			break;
	}

	if (CIRC_CNT(info->xmit.head, 
		     info->xmit.tail, 
		     UART_XMIT_SIZE) < WAKEUP_CHARS)
		uart_event(info, EVT_WRITE_WAKEUP);

	if (info->xmit.head == info->xmit.tail)
		dragonball_stop_tx(info->port, 0);
}

/* 
   Only has 1 mode. Interrupt cause are TRDY or TXEMPTY.
 */
static void dragonball_tx_int(int irq, void *dev_id, struct pt_regs *regs)
{
	struct uart_info *info = dev_id;

	ENTRY();
	dragonball_tx_chars(info);
}

/* 
 */
static void dragonball_rx_int(int irq, void *dev_id, struct pt_regs *regs)
{
	struct uart_info *info = dev_id;
	struct uart_port *port = info->port;
	unsigned int usr1,usr2;
	int id;

	id = get_uart_id(port);

	uart_del_timer(port, rx_timer_list);

	usr1 = uart_in(port, UART_SR1);
	usr2 = uart_in(port, UART_SR2);

	/* should be RRDY interrupt	 */
	if (usr1 & UART_RRDY || usr2 & UART_RDR) {
		if (dragonball_rx_chars(info)) {
			dragonball_stop_rx(port);
		}
	}

	/* add timer to set DREN */
	uart_add_timer(port, rx_timer_list, rx_timer, (unsigned long)info, jiffies + RX_TIMER);

	uart_out(port, UART_CR4, 
		 (uart_in(port, UART_CR4) & ~UART_DREN));

	return;
}

static void dragonball_rts_int(int irq, void *dev_id, struct pt_regs *regs)
{
	struct uart_info *info = dev_id;
	struct uart_port *port = info->port;

	ENTRY();
    
	/* clear interrupt */
#if 0
	uart_out(port, UART_SR2, UART_RTSF); /* RTS int */
#endif
	uart_out(port, UART_SR1, UART_RTSD); /* RTS diff int */

	port->icount.cts++;
	wake_up_interruptible(&info->delta_msr_wait);

	DUMP_REGS(port);
}

static void dragonball_uartc_int(int irq, void *dev_id, struct pt_regs *regs)
{
	struct uart_info *info = dev_id;
	struct uart_port *port = info->port;
	unsigned int usr1,usr2;

	ENTRY();

	usr1 = uart_in(port, UART_SR1);
	usr2 = uart_in(port, UART_SR2);

	if (usr2 & UART_ORE) {
		PRINTK("Overrun Error\n");
		uart_out(port, UART_SR2, UART_ORE);
	}

	if (usr1 & UART_AWAKE) {
		PRINTK2("Async WAKE\n");
		/* clear int 	*/
		uart_out(port, UART_SR1, UART_AWAKE);

		/* disable asynchronous wake up 	interrupt */
		uart_out(port, 
			 UART_CR3,
			 uart_in(port, UART_CR3) & ~UART_AWAKEN);
	}

	return;
}


static void rx_timer(unsigned int data)
{
	struct uart_info *info = (struct uart_info *)data;
	struct uart_port *port = info->port;
	unsigned long flags;

	ENTRY();

	local_irq_save(flags);

	dragonball_start_rx(port);
	uart_out(port, UART_CR4, 
		 (uart_in(port, UART_CR4) | UART_DREN));

	local_irq_restore(flags);

	return ;
}

/*
 * Return TIOCSER_TEMT when transmitter is empty.
 */
static u_int dragonball_tx_empty(struct uart_port *port)
{
	VERBOSE("\n");

	return (uart_in(port, UART_SR2) & UART_TXFE) ? TIOCSER_TEMT : 0;
}

static u_int dragonball_get_mctrl(struct uart_port *port)
{
	unsigned int status = 0, usr1, ucr2, ucr3;
	usr1 = uart_in(port, UART_SR1);
	ucr2 = uart_in(port, UART_CR2);
	ucr3 = uart_in(port, UART_CR3);

	// ENTRY();                /* this function is called periodically */

	status |= (ucr2 & UART_CTS)  ? TIOCM_RTS : 0;
	status |= (usr1 & UART_RTSS) ? TIOCM_CTS : 0;
	status |= (ucr3 & UART_DCD)  ? TIOCM_CAR : 0; /* only in UART2 */
	status |= (ucr3 & UART_DTREN)? TIOCM_DTR : 0; /* only in UART2 */

	return status;
}

static void dragonball_set_mctrl(struct uart_port *port, u_int mctrl)
{
	unsigned int ucr1, ucr2;

	ENTRY();
	PRINTK2("mctrl = 0x%08x\n", mctrl);

	ucr1 = uart_in(port, UART_CR1);
	ucr2 = uart_in(port, UART_CR2);

	if (mctrl & TIOCM_RTS) {
		ucr2 &= ~UART_CTS;
	} else {
		ucr2 |= UART_CTS;
	}

	if (mctrl & TIOCM_CTS) {
		ucr1 |= UART_RTSDEN;  /* enable RTS delta interrupt */
		ucr2 &= ~UART_IRTS;   /* stop ignoring RTS */
	} else {
		ucr1 &= ~UART_RTSDEN;
		ucr2 |= UART_IRTS;
	}

#if 0  /* never going to support */
	/* Modem control for UART2: untested */
	if (mctrl & TIOCM_CD) {
		ucr3 |= UART_DCD;
	} else {
		ucr3 &= ~UART_DCD;
	}

	if (mctrl & TIOCM_DTR) {
		ucr3 |= UART_DTREN;
	} else {
		ucr3 &= ~UART_DTREN;
	}

	if (mctrl & TIOCM_RI) {
		ucr3 |= UART_RI;
	} else {
		ucr3 &= ~UART_RI;
	}

	if (mctrl & TIOCM_DSR) {
		ucr3 |= UART_DSR;
	} else {
		ucr3 &= ~UART_DSR;
	}
#endif

	uart_out(port, UART_CR1, ucr1);
	uart_out(port, UART_CR2, ucr2);
}

/*
 * Interrupts always disabled.
 */
static void dragonball_break_ctl(struct uart_port *port, int break_state)
{
	unsigned int ucr1;

	ENTRY();

	ucr1 = uart_in(port, UART_CR1);

	if (break_state == -1)
		ucr1 |= UART_SNDBRK;
	else
		ucr1 &= ~UART_SNDBRK;

	uart_out(port, UART_CR1, ucr1);
}

static int set_irq(struct uart_port *port, struct uart_info *info)
{
	int retval;
	const char *tx_names[] = {"Uart1: Tx", "Uart2: Tx"};
	const char *rx_names[] = {"Uart1: Rx", "Uart2: Rx"};
	const char *rts_names[] = {"Uart1: RTS", "Uart2: RTS"};
	const char *uartc_names[] = {"Uart1: UARTC", "Uart2: UARTC"};
	int id;
        
	ENTRY();

	id = get_uart_id(port);

	/*
	 * Allocate the IRQ
    	 */
	/* TX */
	retval = request_irq(port->irq, dragonball_tx_int, 0, tx_names[id], info);
	if (retval < 0) {
		printk(KERN_ERR DEBUG_NAME "Request IRQ for %s failed\n", tx_names[id]);
		retval = -EBUSY;
		goto err_tx;
	}

	/* RX */
	retval = request_irq(port->irq+1, dragonball_rx_int, 0, rx_names[id], info);
	if (retval < 0) {
		printk(KERN_ERR DEBUG_NAME "Request IRQ for %s failed\n", rx_names[id]);
		retval = -EBUSY;
		goto err_rx;
	}

	/* RTS */
	retval = request_irq(port->irq-3, dragonball_rts_int, 0, rts_names[id], info);
	if (retval < 0) {
		printk(KERN_ERR DEBUG_NAME "Request IRQ for %s failed\n", rts_names[id]);
		retval = -EBUSY;
		goto err_rts;
	}

	/* UARTC */
	retval = request_irq(port->irq-1, dragonball_uartc_int, 0, uartc_names[id], info);
	if (retval < 0) {
		printk(KERN_ERR DEBUG_NAME "Request IRQ for %s failed\n", uartc_names[id]);
		retval = -EBUSY;
		goto err_uartc;
	}

	irq_active[id] = 1;

	return 0;

 err_uartc:
	free_irq(port->irq-3, info); /* RTS */
 err_rts:
	free_irq(port->irq+1, info); /* RX */
 err_rx:
	free_irq(port->irq, info);  /* TX */
 err_tx:
	return retval;
}

static void unset_irq(struct uart_port *port, struct uart_info *info)
{
	int id;

	ENTRY();

	uart_del_timer(port, rx_timer_list);
	id = get_uart_id(port);

	free_irq(port->irq, info);  /* TX */
	free_irq(port->irq+1, info); /* RX */
	free_irq(port->irq-3, info); /* RTS */
	free_irq(port->irq-1, info); /* UARTC */

	irq_active[id] = 0;

	return;
}

static int dragonball_startup(struct uart_port *port, struct uart_info *info)
{
	int retval;
	int id;
	ENTRY();

	id = get_uart_id(port);

	/* initialise the old status of the modem signals */
	port->ops->set_mctrl(port, info->mctrl);

	/* Rx FIFO */
	uart_out(port, UART_UFCR, 
		 (uart_in(port, UART_UFCR) & ~UART_RXTL_MASK) | (RX_LVL<<UART_RXTL_BIT));

	/* Tx FIFO */
	uart_out(port, UART_UFCR, 
		 (uart_in(port, UART_UFCR) & ~UART_TXTL_MASK) | (TX_LVL<<UART_TXTL_BIT));

	/* CTS Level setting */
	uart_out(port, UART_CR4, 
		 (uart_in(port, UART_CR4) & ~UART_CTSTL_MASK) | (CTS_LVL << UART_CTSTL_BIT));

	/* Enable clock, enable uart, enable interrupts */
	uart_out(port, UART_CR1, 
		 (UART_UARTCLKEN | UART_UARTEN | UART_RRDYEN ));
	uart_out(port, UART_CR2, 
		 (UART_CTSC | UART_IRTS | UART_NON_PARITY | 
		  UART_STOPBIT_1 | UART_8BIT | UART_TXEN | UART_RXEN));
#if 0
	uart_out(port, UART_CR3, 
		 (uart_in(port, UART_CR3) | UART_TIMEOUTEN));
#endif
	uart_out(port, UART_CR4, 
		 (uart_in(port, UART_CR4) | UART_DREN | UART_OREN));

	if (!irq_active[id]) {
		/* Now set IRQ */
		if ((retval = set_irq(port, info))) {
			return retval;
		}
		/* initialize Rx Timer */
		uart_init_timer(port, rx_timer_list); 
	}

	return 0;
}

static void dragonball_shutdown(struct uart_port *port, struct uart_info *info)
{
	int id;

	ENTRY();

	id = get_uart_id(port);

	/* Put to reset value */
	uart_out(port, UART_CR1, 0x4);
	uart_out(port, UART_CR2, 0x1);
	uart_out(port, UART_CR3, 0x0);
	uart_out(port, UART_CR4, 0x8040);

	/* Free the interrupt */
	if (irq_active[id])
		unset_irq(port, info);
}

static void dragonball_change_speed(struct uart_port *port, u_int cflag, u_int iflag, u_int quot)
{
	unsigned int ucr1, ucr2, ucr3, ucr4, bir, bmr, fcr;
	unsigned long flags;

	ENTRY();
	PRINTK2("cflag = 0x%08x, iflag = 0x%08x, quot = %d\n", cflag, iflag, quot);

	/* for uart options */
	ucr1 = uart_in(port, UART_CR1);
	ucr2 = uart_in(port, UART_CR2);

	/* for baud rate */
	ucr3 = uart_in(port, UART_CR3);
	ucr4 = uart_in(port, UART_CR4);
	bir = uart_in(port, UART_BIR);
	bmr = uart_in(port, UART_BMR);

	/* byte size */
	ucr2 &= ~(UART_WS);      /* clear bits */
	if ((cflag & CSIZE) == CS7) 
		ucr2 |= UART_7BIT;
	else if ((cflag & CSIZE) == CS8) 
		ucr2 |= UART_8BIT;
	else 
		ucr2 |= UART_8BIT;

	/* parity */
	ucr2 &= ~(UART_PREN | UART_PROE); /* clear bits */
	if (cflag & PARENB) {   
		ucr2 |= UART_PREN;  /* parity enable */
		if (cflag & PARODD)
			ucr2 |= UART_PARITY_ODD; /* odd */
		else 
			ucr2 |= UART_PARITY_EVEN; /* even */
	} else {
		ucr2 |= UART_NON_PARITY; /* non parity */
	}

	/* stop bit */
	ucr2 &= ~(UART_STPB);   /* clear bits */
	if (cflag & CSTOPB) {
		ucr2 |= UART_STOPBIT_2;
	} else {
		ucr2 |= UART_STOPBIT_1;
	}

	/* flow control */
	if (cflag & CRTSCTS) {
		ucr2 &= ~UART_IRTS;     /* stop ingoring RTS  */
		ucr2 |= UART_CTSC;      /* enbale CTS */
	} else {
		ucr2 |= UART_IRTS;
		ucr2 &= ~UART_CTSC;
	}

	port->read_status_mask = UART_OVRRUN;
	if (iflag & IGNPAR)
		port->read_status_mask |= UART_FRAMERR | UART_PRERR;
    
	/*
	 * Characteres to ignore
	 */
	port->ignore_status_mask = 0;
	if (iflag & IGNPAR)
		port->ignore_status_mask |= UART_PRERR | UART_FRAMERR;
	if (iflag & IGNBRK && iflag & IGNPAR)
		port->ignore_status_mask |= UART_OVRRUN;
    
	/*
	 * ignore all characters if CREAD is not set
	 */
	if ((cflag & CREAD) == 0)
		port->ignore_status_mask |= UART_CHARRDY;
    
	/* this routine only for:
    	   - UART_REF_CLOCK == 16Mhz 
	   - PERCLK1 == 16Mhz
	*/
	/* PerCLK1 == 16Mhz */
	//fcr = (0x2 << UART_TXTL_BIT) | (0x1 << UART_RXTL_BIT) | UART_RFDIV_1;
	fcr = (TX_LVL << UART_TXTL_BIT) | (RX_LVL << UART_RXTL_BIT) | UART_RFDIV_1;
	ucr4 |= UART_REF16;
	ucr3 &= ~(UART_REF25 | UART_REF30);

	/* 
	   e.g. 
	   115200 baud: 115200*16/16MHz = 0.1152 = 1152/10000 	
	   BIR = 1152 - 1, BMR = 10000 - 1
    	*/
	switch(cflag & CBAUD) {
	case B1200:
		PRINTK2("cflag = B1200\n");
		/* 12/10000 = 3/2500*/
		bir = 3 - 1;
		bmr = 2500 - 1;
		break;
	case B1800:
		PRINTK2("cflag = B1800\n");
		/* 18/10000 = 9/5000 */
		bir = 9 - 1;
		bmr = 5000 - 1;
		break;
	case B2400:
		PRINTK2("cflag = B2400\n");
		/* 24/10000 = 3/1250 */
		bir = 3 - 1;
		bmr = 1250 - 1;
		break;
	case B4800:
		PRINTK2("cflag = B4800\n");
		/* 48/10000 = 3/625 */
		bir = 3 - 1;
		bmr = 625 - 1;
		break;
	case B9600:
		PRINTK2("cflag = B9600\n");
		/* 96/10000 = 6/625 */
		bir = 6 - 1;
		bmr = 625 - 1;
		break;
	case B19200:
		PRINTK2("cflag = B19200\n");
		/* 192/10000 = 12/625 */
		bir = 12 - 1;
		bmr = 625 - 1;
		break;
	case B38400:
		PRINTK2("cflag = B38400\n");
		/* 384/10000 = 96/2500 = 24/625 */
		bir = 24 - 1;
		bmr = 625 - 1;
		break;
	case B57600:
		PRINTK2("cflag = B57600\n");
		/* 576/10000 = 36/625 */
		bir = 36 - 1;
		bmr = 625 - 1;
		break;
	case B115200:
		PRINTK2("cflag = B115200\n");
		/* 1152/10000 = 72/625 */
		bir = 72 - 1;
		bmr = 625 - 1;
		break;
	case B230400:
		PRINTK2("cflag = B230400\n");
		/* 2304/10000 = 144/625 */
		bir = 144 - 1;
		bmr = 625 - 1;
		break;
	case B460800:
		PRINTK2("cflag = B460800\n");
		/* 4608/10000 = 288/625 */
		bir = 288 - 1;
		bmr = 625 - 1;
		break;
	case B500000:
		PRINTK2("cflag = B500000\n");
		/* 5000/10000 = 1/2 */
		bir = 1 - 1;
		bmr = 2 - 1;
		break;
	case B921600:
		PRINTK2("cflag = B921600\n");
		/* 9216/10000 = 576/625 */
		bir = 576 - 1;
		bmr = 625 - 1;
		break;
	case B1000000:
		PRINTK2("cflag = B1000000\n");
		/* 10000/10000 = 1/1 */
		bir = 1 - 1;
		bmr = 1 - 1;
		break;
	default:
		PRINTK("Unexpected baudrate = 0x%08x\n", cflag & CBAUD);
		bir = 1152 - 1;
		bmr = 10000 - 1;
	}

	/* disable interrupts */
	local_irq_save(flags);

	/* disable tx and rx */
	uart_out(port, UART_CR2, uart_in(port, UART_CR2) & ~(UART_TXEN | UART_RXEN));

	/* set values */
	uart_out(port, UART_CR1, ucr1);
	uart_out(port, UART_CR2, ucr2);
	uart_out(port, UART_CR3, ucr3);
	uart_out(port, UART_CR4, ucr4);
	uart_out(port, UART_UFCR, fcr);
	uart_out(port, UART_BIR, bir);
	uart_out(port, UART_BMR, bmr);
	
	/* reenable interrupts */
	local_irq_restore(flags);

	DUMP_REGS(port);
}

static const char *dragonball_type(struct uart_port *port)
{
	ENTRY();

	return port->type == PORT_DRAGONBALL ? "DragonBall MX1" : NULL;
}

/*
 * Release the memory region(s) being used by 'port'.
 */
static void dragonball_release_port(struct uart_port *port)
{
	ENTRY();

	release_mem_region(port->mapbase, UART_PORT_SIZE);
}

/*
 * Request the memory region(s) being used by 'port'.
 */
static int dragonball_request_port(struct uart_port *port)
{
	ENTRY();

	return request_mem_region(port->mapbase, UART_PORT_SIZE,
				  "serial_dragonball") != NULL ? 0 : -EBUSY;
}

/*
 * Configure/autoconfigure the port.
 */
static void dragonball_config_port(struct uart_port *port, int flags)
{
	ENTRY();

	if (flags & UART_CONFIG_TYPE && dragonball_request_port(port) == 0)
		port->type = PORT_DRAGONBALL;
}

/*
 * Verify the new serial_struct (for TIOCSSERIAL).
 */
static int dragonball_verify_port(struct uart_port *port, struct serial_struct *ser)
{
	int ret = 0;

	ENTRY();

	if (ser->type != PORT_UNKNOWN && ser->type != PORT_DRAGONBALL)
		ret = -EINVAL;
	if (ser->irq < 0 || ser->irq >= NR_IRQS)
		ret = -EINVAL;
	return ret;
}

/* 
 * PM function called at PM_SUSPEND
 */
static void dragonball_pm(struct uart_port *port, u_int state, u_int oldstate)
{
	ENTRY();

	switch (state) {
	case 0:
		/* disable asynchronous wake up interrupt */
		uart_out(port,
			 UART_CR3,
			 uart_in(port, UART_CR3) & ~UART_AWAKEN);
		break;
	case 1:
		PRINTK("PM_SUSPEND: UART%d [D1]\n", get_uart_id(port) + 1);

		/* enable asynchronous wake up interrupt */
		uart_out(port, UART_SR1, UART_AWAKE); /* clear first */
		uart_out(port,
			 UART_CR3,
			 uart_in(port, UART_CR3) | UART_AWAKEN);

		PRINTK2("ucr1=0x%08x, ucr2=0x%08x, ucr3=0x%08x, ucr4=0x%08x\n", 
			uart_in(port, UART_CR1), uart_in(port, UART_CR2),
			uart_in(port, UART_CR3), uart_in(port, UART_CR4));
		break;
	default:
		PRINTK2("state=%d, oldstate=%d\n", state, oldstate);
		break;
	}
}

static struct uart_ops dragonball_pops = {
	tx_empty:	dragonball_tx_empty,
	set_mctrl:	dragonball_set_mctrl,
	get_mctrl:	dragonball_get_mctrl,
	stop_tx:	dragonball_stop_tx,
	start_tx:	dragonball_start_tx,
	stop_rx:	dragonball_stop_rx,
	enable_ms:	dragonball_enable_ms,
	break_ctl:	dragonball_break_ctl,
	startup:	dragonball_startup,
	shutdown:	dragonball_shutdown,
	change_speed:	dragonball_change_speed,
	pm:		dragonball_pm,
	type:		dragonball_type,

	release_port:	dragonball_release_port,
	request_port:	dragonball_request_port,
	config_port:	dragonball_config_port,
	verify_port:	dragonball_verify_port,
};

static struct uart_port dragonball_ports[NR_PORTS] = {
	{
		iobase:		UART1_BASE,
		mapbase:	UART1_BASE,
		irq:		IRQ_UART1_MINT_TX, /* irq number for RX = TX+ 1*/
		uartclk:	16*1000*1000, /* UART_REF_CLOCK: 16Mhz */
		fifosize:	32,
		ops:		&dragonball_pops,
		flags:		ASYNC_BOOT_AUTOCONF,
	},
	{
		iobase:		UART2_BASE,
		mapbase:	UART2_BASE,
		irq:		IRQ_UART2_MINT_TX, /* irq number for RX = TX+ 1*/
		uartclk:	16*1000*1000, /* UART_REF_CLOCK: 16Mhz */
		fifosize:	32,
		ops:		&dragonball_pops,
		flags:		ASYNC_BOOT_AUTOCONF,
	},
};


#ifdef CONFIG_SERIAL_DRAGONBALL_CONSOLE

static inline void wait_for_xmitr(struct uart_port *port)
{
	unsigned int status, tmout = 1000000;
  
	do {
		status = uart_in(port, UART_SR1);
		if (--tmout == 0)
			break;
	} while(!(status & UART_TRDY));
}

static void dragonball_console_write(struct console *co, const char *s, u_int count)
{
	struct uart_port *port = dragonball_ports + co->index;
	unsigned int ucr1, ucr2;
	u_int i;

	/* If port is disabled, do nothing */
	ucr1 = uart_in(port, UART_CR1);
	ucr2 = uart_in(port, UART_CR1);
	if (!((ucr1 & UART_UARTEN) && (ucr2 & (UART_TXEN | UART_RXEN))))
		return;

	/*
	 * Now, do each character
	 */
	for (i = 0; i < count; i++) {
		wait_for_xmitr(port);

		/*
		 *	Send the character out.
		 *	If a LF, also do CR...
		 */
		uart_putc( port, s[i]);
		if (s[i] == '\n') {
			wait_for_xmitr(port);
			uart_putc( port, '\r' );
		}
	}

	/*
	 * Wait for the tranceiver to empty
	 */
	wait_for_xmitr(port);
}

static kdev_t dragonball_console_device(struct console *co)
{
	ENTRY();

	return MKDEV(SERIAL_DRAGONBALL_MAJOR, MINOR_START + co->index);
}

#if 0
static int dragonball_console_wait_key(struct console *co)
{
	struct uart_port *port = dragonball_ports + co->index;
	int ch;

	ENTRY();

	/*
	 * Wait for a character
	 */
	while( !(uart_in(port, UART_SR2) & UART_RDR) );
	ch = uart_getc(port) & 0xff;

	return ch;
}
#endif

/*
 * If the port was already initialised (eg, by a boot loader), try to determine
 * the current setup.
 */
static void __init
dragonball_console_get_options(struct uart_port *port, int *baud, int *parity, int *bits)
{
	unsigned int ucr1, ucr2, ucr3, ucr4, bir, bmr;

	ENTRY();

	ucr1 = uart_in(port, UART_CR1);
	ucr2 = uart_in(port, UART_CR2);
	if ((ucr1 & UART_UARTEN) && (ucr2 & (UART_TXEN | UART_RXEN))) {
		/* this port is enabled */

		/* check parity */
		if (ucr2 & UART_PREN) { 
			if ((ucr2 & UART_PROE) == UART_PARITY_ODD) {
				*parity = 'o';  /* odd */
			} else {
				*parity = 'e';  /* even */
			}
		} else {  
			*parity = 'n';      /* non-parity */
		}

		/* check word size */
		if ((ucr2 & UART_WS) == UART_8BIT) {
			*bits = 8;
		} else {
			*bits = 7;
		}

		/* this code suppose UART_REF_CLOCK to be 16 Mhz */
		ucr3 = uart_in(port, UART_CR3);
		ucr4 = uart_in(port, UART_CR4);
		bir = uart_in(port, UART_BIR);
		bmr = uart_in(port, UART_BMR);
		if ((ucr4 & UART_REF16) && !(ucr3 & UART_REF25) && !(ucr3 & UART_REF30)) {
			/* UART_REF_CLOCK == 16Mhz */

			/*
			  Baudrate
			  e.g. 
			  115200 baud: 115200*16/16MHz = 0.1152 = 1152/10000 	
			  BIR = 1152 - 1, BMR = 10000 - 1
			*/
			*baud = 10000*(bir+1)/(bmr+1);
            
		} else if (!(ucr4 & UART_REF16) && (ucr3 & UART_REF25) && !(ucr3 & UART_REF30)) {
			/* UART_REF_CLOCK == 25Mhz */
			printk("%s:%d:  Sorry, no support for UART_REF_CLOCK == 25Mhz\n", __FUNCTION__, __LINE__);
			*baud = 115200;
		} else if (!(ucr4 & UART_REF16) && !(ucr3 & UART_REF25) && (ucr3 & UART_REF30)) {
			/* UART_REF_CLOCK == 30Mhz */
			printk("%s:%d:  Sorry, no support for UART_REF_CLOCK == 30Mhz\n", __FUNCTION__, __LINE__);
			*baud = 115200;
		}
	}
}

static int __init
dragonball_console_setup(struct console *co, char *options)
{
	struct uart_port *port;
	int baud = CONFIG_DRAGONBALL_DEFAULT_BAUDRATE;
	int bits = 8;
	int parity = 'n';
	int flow = 'n';

	ENTRY();

	/*
    	 * Check whether an invalid uart number has been specified, and
	 * if so, search for the first available port that does have
	 * console support.
	 */
	port = uart_get_console(dragonball_ports, NR_PORTS, co);

	if (options)
		uart_parse_options(options, &baud, &parity, &bits, &flow);
	else
		dragonball_console_get_options(port, &baud, &parity, &bits);

	return uart_set_options(port, co, baud, parity, bits, flow);
}

static struct console dragonball_console = {
	name:   "ttyDB",
	write:  dragonball_console_write,
	device: dragonball_console_device,
#if 0
	wait_key:   dragonball_console_wait_key,
#endif
	setup:  dragonball_console_setup,
	flags:  CON_PRINTBUFFER,
	index:  -1,
};

void __init dragonball_console_init(void)
{
	ENTRY();
	register_console(&dragonball_console);
}

#define DRAGONBALL_CONSOLE	&dragonball_console
#else
#define DRAGONBALL_CONSOLE	NULL
#endif /* CONFIG_SERIAL_DRAGONBALL_CONSOLE */

static struct uart_driver dragonball_reg = {
	owner:			THIS_MODULE,
	normal_major:		SERIAL_DRAGONBALL_MAJOR,
#ifdef CONFIG_DEVFS_FS
	normal_name:		"ttyDB%d",
	callout_name:		"cudb%d",
#else
	normal_name:		"ttyDB",
	callout_name:		"cudb",
#endif
	normal_driver:		&normal,
	callout_major:		CALLOUT_DRAGONBALL_MAJOR,
	callout_driver:		&callout,
	table:			dragonball_table,
	termios:		dragonball_termios,
	termios_locked:		dragonball_termios_locked,
	minor:			MINOR_START,
	nr:			NR_PORTS,
	port:			dragonball_ports,
	cons:			DRAGONBALL_CONSOLE,
};

static int __init dragonball_serial_init(void)
{
	int retval;
#if defined (CONFIG_DRAGONBALL_SNSC_MPU110)
	unsigned tmp;
#endif
	ENTRY();

#if defined (CONFIG_DRAGONBALL_SNSC_MPU110)
	/* enable MPU-110 GPIO power for UART */
	tmp = inl(MPU110_GPIO2);
	tmp &= ~(MPU110_PWR_UART1A | MPU110_PWR_UART1B | 
		 MPU110_PWR_UART2A | MPU110_PWR_UART2B);
	outl(tmp, MPU110_GPIO2); // set 0 to turn on
#endif

	/* set the CPU-GPIO */
	retval = dragonball_register_gpios(PORT_C, 
					   (1 << 12) | /* UART1: RXD */
					   (1 << 11) | /* UART1: TXD */
					   (1 << 10) | /* UART1: RTS */
					   (1 <<  9),  /* UART1: CTS */
					   PRIMARY, "uart1");
	if (retval < 0) goto error;

	retval = dragonball_register_gpios(PORT_B, 
					   (1 << 31) | /* UART2: RXD */
					   (1 << 30) | /* UART2: TXD */
					   (1 << 29) | /* UART2: RTS */
					   (1 << 28),  /* UART2: CTS */
					   PRIMARY, "uart2");
	if (retval < 0) goto error;

#if 0
	dragonball_register_gpios(PORT_D, 
				  (1 << 10) | /* UART2: DSR */
				  (1 <<  9) | /* UART2: RI */
				  (1 <<  8) | /* UART2: DCD */
				  (1 <<  7),  /* UART2: DTR */
				  SECONDARY, "uart2(modem)");
	if (retval < 0) 
		goto error;

#endif

	/* initialize anyway at startup */
	uart_out(&dragonball_ports[0],
		 UART_CR1,
		 UART_UARTCLKEN | UART_UARTEN | UART_RRDYEN | UART_TRDYEN);
	uart_out(&dragonball_ports[0],
		 UART_CR2,
		 UART_CTSC | UART_IRTS | UART_NON_PARITY | UART_STOPBIT_1 |
		 UART_8BIT | UART_TXEN | UART_RXEN);

	/* intialize the debug proc device */
#if UART_DEBUG & 1
	puart1 = create_proc_read_entry("dragonball_mx1/uart_1", 0, NULL,
					proc_dbmx1_uart_dump, (void *)&dragonball_ports[0]);
	puart2 = create_proc_read_entry("dragonball_mx1/uart_2", 0, NULL,
					proc_dbmx1_uart_dump, (void *)&dragonball_ports[1]);
#endif

	irq_active[0] = 0;
	irq_active[1] = 0;

	return uart_register_driver(&dragonball_reg);

 error:
	dragonball_unregister_gpios(PORT_C, (1 << 12) | (1 << 11) | (1 << 10) | (1 <<  9));
	dragonball_unregister_gpios(PORT_B, (1 << 31) | (1 << 30) | (1 << 29) | (1 << 28));
#if 0
	dragonball_unregister_gpios(PORT_D, (1 << 10) | (1 <<  9) | (1 <<  8) | (1 <<  7));
#endif
	return retval;
}

static void __exit dragonball_serial_exit(void)
{
	ENTRY();

#if UART_DEBUG & 1
	if (puart1) 
		remove_proc_entry("dragonball_mx1/uart_1", NULL);
	if (puart2) 
		remove_proc_entry("dragonball_mx1/uart_2", NULL);
#endif

	/* reset */
	outl(0, UART1_BASE + UART_CR2);
	outl(0, UART2_BASE + UART_CR2);

	dragonball_unregister_gpios(PORT_C, (1 << 12) | (1 << 11) | (1 << 10) | (1 <<  9));
	dragonball_unregister_gpios(PORT_B, (1 << 31) | (1 << 30) | (1 << 29) | (1 << 28));
#if 0
	dragonball_unregister_gpios(PORT_D, (1 << 10) | (1 <<  9) | (1 <<  8) | (1 <<  7));
#endif

	uart_unregister_driver(&dragonball_reg);
}

module_init(dragonball_serial_init);
module_exit(dragonball_serial_exit);

EXPORT_NO_SYMBOLS;

MODULE_AUTHOR("Sony Corporation");
MODULE_DESCRIPTION("DragonBall MX1 generic serial port driver");
MODULE_LICENSE("GPL");

