/*
 *  cxd9802_i2s.c -- CXD9802 I2S sound driver
 *
 */
/*
 *  Copyright 2002,2003,2004 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/module.h>
#include <linux/version.h>

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/pci.h>
#include <asm/uaccess.h>

#include <linux/sound.h>
#include <linux/soundcard.h>

#include <asm-mips/cxd9802/cxd9802.h>

#define DEBUG
#define DEBUG_LEVEL DPL_NOTICE

#ifdef DEBUG
#define DPR(fmt, args...)	printk(fmt , ## args)
#endif /* DEBUG */

MODULE_AUTHOR("Sony Corporation");
MODULE_DESCRIPTION("CXD9802 I2S driver");
MODULE_LICENSE("GPL");

/* ======================================================================= *
 * debug print function from dpr package                                   *
 * ======================================================================= */

#define DPL_EMERG	0
#define DPL_ALERT	1
#define DPL_CRIT	2
#define DPL_ERR		3
#define DPL_WARNING	4
#define DPL_NOTICE	5
#define DPL_INFO	6
#define DPL_DEBUG	7

#define DPM_PURPOSE	0x0000000f
#define DPM_ERR		0x00000001
#define DPM_TRACE	0x00000002
#define DPM_LAYER	0x000000f0
#define DPM_APPLICATION	0x00000010
#define DPM_MIDDLEWARE	0x00000020
#define DPM_DRIVER	0x00000040
#define DPM_KERNEL	0x00000080
#define DPM_USERDEF	0xffff0000

#ifdef DEBUG

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL DPL_DEBUG
#endif
#ifndef DEBUG_MASK
#define DEBUG_MASK  DPM_PURPOSE
#endif

#ifndef DPR
#include <stdio.h>
#define DPR(fmt, args...)	fprintf(stderr, fmt , ## args)
#endif

#define DPRL(level, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && DPR(fmt , ## args))
#define DPRM(mask, fmt, args...) \
	(void)(((mask) & DEBUG_MASK) && DPR(fmt , ## args))
#define DPRLM(level, mask, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && ((mask) & DEBUG_MASK) && \
		DPR(fmt , ## args))
#define DPRC(cond, fmt, args...) \
	(void)((cond) && DPR(fmt , ## args))
#define DPRF(fmt, args...)	DPR("%s: " fmt, __FUNCTION__ , ## args)
#define DPRFL(level, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && DPRF(fmt , ## args))
#define DPRFM(mask, fmt, args...) \
	(void)(((mask) & DEBUG_MASK) && DPRF(fmt , ## args))
#define DPRFLM(level, mask, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && ((mask) & DEBUG_MASK) && \
		DPRF(fmt , ## args))
#define DPRFC(cond, fmt, args...) \
	(void)((cond) && DPRF(fmt , ## args))

#define	DPRFS	DPRFLM(DPL_DEBUG, DPM_TRACE, "start\n")
#define	DPRFE	DPRFLM(DPL_DEBUG, DPM_TRACE, "end\n")
#define DPRFR	DPRFLM(DPL_DEBUG, DPM_TRACE, "return\n")

#else  /* DEBUG */

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL -1
#endif
#ifndef DEBUG_MASK
#define DEBUG_MASK  0
#endif

#ifndef DPR
#define DPR(fmt, args...)
#endif

#define DPRL(level, fmt, args...)
#define DPRM(mask, fmt, args...)
#define DPRLM(level, mask, fmt, args...)
#define DPRC(cond, fmt, args...)
#define DPRF(fmt, args...)
#define DPRFL(level, fmt, args...)
#define DPRFM(mask, fmt, args...)
#define DPRFLM(level, mask, fmt, args...)
#define DPRFC(cond, fmt, args...)

#define	DPRFS
#define	DPRFE
#define DPRFR

#endif /* DEBUG */

/* ======================================================================= *
 * constant                                                                *
 * ======================================================================= */

#define CXD9802_I2S_DEVNAME		"cxd9802_i2s"
#define CXD9802_I2S_NCLKCHANNELS	2
#define CXD9802_I2S_NDMACHANNELS	3

#define CXD9802_I2S_MAX_DMA_SIZE	32768
#define CXD9802_I2S_DEFAULT_BUF_SIZE	(CXD9802_I2S_MAX_DMA_SIZE * 6)

#define CXD9802_I2S_DEFAULT_SCDESC_NELEM	4

/* DAC/DAC mode */
#define CXD9802_I2S_DACADC_MODE_NODE	0
#define CXD9802_I2S_DACADC_MODE_TX	1
#define CXD9802_I2S_DACADC_MODE_RX	2

/* ------------------------ *
 * config register          *
 * ------------------------ */

/* CXD9802_CONFIG_PCFG */
#define  CXD9802M_CONFIG_I2SMODE	(0x3LL << 58)
#define  CXD9802S_CONFIG_I2SMODE(x)	((u64)(x) << 58)
#define  CXD9802C_CONFIG_I2SMODE_ACLC	CXD9802S_CONFIG_I2SMODE(0)
#define  CXD9802C_CONFIG_I2SMODE_I2S	CXD9802S_CONFIG_I2SMODE(1)
#define  CXD9802C_CONFIG_I2SMODE_INPUT	CXD9802S_CONFIG_I2SMODE(3)
/* CXD9802_CONFIG_MCLKCTRL */
#define  CXD9802M_CONFIG_BAUDEN		(1 << 15)
#define  CXD9802M_CONFIG_MCLKBE(n)	(1 << (4 * (n) + 3))
#define  CXD9802M_CONFIG_MCLKCTRL(n)	(0x7 << (4 * (n)))
#define  CXD9802S_CONFIG_MCLKCTRL(n, x)	((x) << (4 * (n)))
#define  CXD9802C_CONFIG_MCLKCTRL_44100(n)	CXD9802S_CONFIG_MCLKCTRL(n, 0)
#define  CXD9802C_CONFIG_MCLKCTRL_48000(n)	CXD9802S_CONFIG_MCLKCTRL(n, 1)
#define  CXD9802C_CONFIG_MCLKCTRL_32000(n)	CXD9802S_CONFIG_MCLKCTRL(n, 2)
#define  CXD9802C_CONFIG_MCLKCTRL_24000(n)	CXD9802S_CONFIG_MCLKCTRL(n, 3)
#define  CXD9802C_CONFIG_MCLKCTRL_16000(n)	CXD9802S_CONFIG_MCLKCTRL(n, 4)
#define  CXD9802C_CONFIG_MCLKCTRL_8000(n)	CXD9802S_CONFIG_MCLKCTRL(n, 5)
#define  CXD9802C_CONFIG_MCLKCTRL_EXT(n)	CXD9802S_CONFIG_MCLKCTRL(n, 6)

/* ------------------------ *
 * DMA register             *
 * ------------------------ */

/* CXD9802_DMAC_DMCHAR(c, n) */
/* CXD9802_DMAC_DMSAR(c, n) */
/* CXD9802_DMAC_DMDAR(c, n) */
/* CXD9802_DMAC_DMCNTR(c, n) */
/* CXD9802_DMAC_DMSAIR(c, n) */
/* CXD9802_DMAC_DMDAIR(c, n) */
/* CXD9802_DMAC_DMCCR(c, n) */
#define  CXD9802M_DMAC_IMMCHN		(1LL << 29)
#define  CXD9802M_DMAC_USEXFSZ		(1LL << 28)
#define  CXD9802M_DMAC_LE		(1LL << 27)
#define  CXD9802M_DMAC_DBINH		(1LL << 26)
#define  CXD9802M_DMAC_SBINH		(1LL << 25)
#define  CXD9802M_DMAC_CHRST		(1LL << 24)
#define  CXD9802M_DMAC_REVBYTE		(1LL << 23)
#define  CXD9802M_DMAC_ACKPOL		(1LL << 22)
#define  CXD9802M_DMAC_REQPL		(1LL << 21)
#define  CXD9802M_DMAC_EGREQ		(1LL << 20)
#define  CXD9802M_DMAC_CHDN		(1LL << 19)
#define  CXD9802M_DMAC_DNCTL		(0x3LL << 17)
#define  CXD9802M_DMAC_EXTRQ		(1LL << 16)
#define  CXD9802M_DMAC_STLTIMEINTRQD	(0x7LL << 13)
#define  CXD9802M_DMAC_INTENE		(1LL << 12)
#define  CXD9802M_DMAC_INTENC		(1LL << 11)
#define  CXD9802M_DMAC_INTENT		(1LL << 10)
#define  CXD9802M_DMAC_CHNENC		(1LL << 9)
#define  CXD9802M_DMAC_XFACT		(1LL << 8)
#define  CXD9802M_DMAC_SMPCHIN		(1LL << 5)
#define  CXD9802M_DMAC_XFSZ		(0x7LL << 2)
#define  CXD9802S_DMAC_XFSZ(x)		((x) << 2)
#define  CXD9802C_DMAC_XFSZ_1B		CXD9802S_DMAC_XFSZ(0)
#define  CXD9802C_DMAC_XFSZ_2B		CXD9802S_DMAC_XFSZ(1)
#define  CXD9802C_DMAC_XFSZ_4B		CXD9802S_DMAC_XFSZ(2)
#define  CXD9802C_DMAC_XFSZ_1DW		CXD9802S_DMAC_XFSZ(3)
#define  CXD9802C_DMAC_XFSZ_4DW		CXD9802S_DMAC_XFSZ(4)
#define  CXD9802C_DMAC_XFSZ_8DW		CXD9802S_DMAC_XFSZ(5)
#define  CXD9802C_DMAC_XFSZ_16DW	CXD9802S_DMAC_XFSZ(6)
#define  CXD9802C_DMAC_XFSZ_32DW	CXD9802S_DMAC_XFSZ(7)
#define  CXD9802M_DMAC_MEMIO		(1LL << 1)
#define  CXD9802M_DMAC_SNGAD		(1LL << 0)
/* CXD9802_DMAC_DMCSR(c, n) */
#define  CXD9802M_DMAC_WAITC		(0xffffLL << 16)
#define  CXD9802M_DMAC_CHNENS		(1LL << 10)
#define  CXD9802M_DMAC_STLXFER		(1LL << 9)
#define  CXD9802M_DMAC_XFACT		(1LL << 8)
#define  CXD9802M_DMAC_ABCHC		(1LL << 7)
#define  CXD9802M_DMAC_NCHNC		(1LL << 6)
#define  CXD9802M_DMAC_NTRNFC		(1LL << 5)
#define  CXD9802M_DMAC_EXTDN		(1LL << 4)
#define  CXD9802M_DMAC_CFERR		(1LL << 3)
#define  CXD9802M_DMAC_CHERR		(1LL << 2)
#define  CXD9802M_DMAC_DESERR		(1LL << 1)
#define  CXD9802M_DMAC_SORERR		(1LL << 0)
/* CXD9802_DMAC_DMMFDR(c) */
/* CXD9802_DMAC_DMMCR(c) */
#define  CXD9802M_DMAC_EIS(n)		(1LL << (28 + (n)))
#define  CXD9802M_DMAC_DIS(n)		(1LL << (24 + (n)))
#define  CXD9802M_DMAC_FIFVC		(0x7fLL << 14)
#define  CXD9802M_DMAC_FIFWP		(0x7LL << 11)
#define  CXD9802M_DMAC_FIFRP		(0x7LL << 8)
#define  CXD9802M_DMAC_RSFIF		(1LL << 7)
#define  CXD9802M_DMAC_FIFUM(n)		(1LL << (3 + (n)))
#define  CXD9802M_DMAC_RRPT		(1LL << 1)
#define  CXD9802M_DMAC_MSTEN		(1LL << 0)

/* ------------------------ *
 * I2S register             *
 * ------------------------ */

/* CXD9802_I2S_I2SMCR */
#define  CXD9802M_I2S_START_CH(n)	(1 << (n))
/* CXD9802_I2S_I2SCCR */
#define  CXD9802M_I2S_MODE		(1 << 31)
#define  CXD9802C_I2S_MODE_2CH_IO	(0 << 31)
#define  CXD9802C_I2S_MODE_3CH_OUT	(1 << 31)
#define  CXD9802M_I2S_WS_CHECK(n)	(1 << (11 + 16 * (n)))
#define  CXD9802M_I2S_MCLK_CHECK_EN(n)	(1 << (10 + 16 * (n)))
#define  CXD9802M_I2S_CLOCK_DATA(n)	(1 << (9 + 16 * (n)))
#define  CXD9802M_I2S_DATA_SWAP(n)	(1 << (8 + 16 * (n)))
#define  CXD9802M_I2S_MCLK_MODE(n)	(1 << (7 + 16 * (n)))
#define  CXD9802M_I2S_MASTER_MODE(n)	(1 << (6 + 16 * (n)))
#define  CXD9802M_I2S_CLOCK_DELAY(n)	(1 << (5 + 16 * (n)))
#define  CXD9802M_I2S_INVERT_WS(n)	(1 << (4 + 16 * (n)))
#define  CXD9802M_I2S_JUSTIFY(n)	(1 << (3 + 16 * (n)))
#define  CXD9802M_I2S_DATA_SELECT(n)	(0x3 << (1 + 16 * (n)))
#define  CXD9802S_I2S_DATA_SELECT(n, x)	((x) << (1 + 16 * (n)))
#define  CXD9802C_I2S_DATA_SELECT_16(n)	CXD9802S_I2S_DATA_SELECT(n, 0)
#define  CXD9802C_I2S_DATA_SELECT_18(n)	CXD9802S_I2S_DATA_SELECT(n, 1)
#define  CXD9802C_I2S_DATA_SELECT_20(n)	CXD9802S_I2S_DATA_SELECT(n, 2)
#define  CXD9802C_I2S_DATA_SELECT_24(n)	CXD9802S_I2S_DATA_SELECT(n, 3)
#define  CXD9802M_I2S_IO_MODE(n)		(1 << (0 + 16 * (n)))
/* CXD9802_I2S_I2SICTRL */
#define  CXD9802M_I2S_UORUN_INT_MASK(n)	(1 << (20 + (n)))
#define  CXD9802M_I2S_WSERR_INT_MASK(n)	(1 << (18 + (n)))
#define  CXD9802M_I2S_MCLKERR_INT_MASK(n)	(1 << (16 + (n)))
#define  CXD9802M_I2S_UORUN_INT(n)		(1 << (4 + (n)))
#define  CXD9802M_I2S_WSERR_INT(n)		(1 << (2 + (n)))
#define  CXD9802M_I2S_MCLKERR_INT(n)	(1 << (0 + (n)))
/* CXD9802_I2S_I2SCOR */
#define  CXD9802M_I2S_MCLK_CHECKV(n)	(0x1f << (8 + 16 * (n)))
#define  CXD9802S_I2S_MCLK_CHECKV(n, x)	((x) << (8 + 16 * (n)))
#define  CXD9802M_I2S_TIMESLOT(n)	(0x3 << (6 + 16 * (n)))
#define  CXD9802S_I2S_TIMESLOT(n, x)	((x) << (6 + 16 * (n)))
#define  CXD9802M_I2S_MCLK_DIV(n)	(0xf << (0 + 16 * (n)))
#define  CXD9802S_I2S_MCLK_DIV(n, x)	((x) << (0 + 16 * (n)))
#define  CXD9802C_I2S_MCLK_DIV_1(n)	CXD9802S_I2S_MCLK_DIV(n, 0)
#define  CXD9802C_I2S_MCLK_DIV_2(n)	CXD9802S_I2S_MCLK_DIV(n, 2)
#define  CXD9802C_I2S_MCLK_DIV_3(n)	CXD9802S_I2S_MCLK_DIV(n, 3)
#define  CXD9802C_I2S_MCLK_DIV_4(n)	CXD9802S_I2S_MCLK_DIV(n, 4)
#define  CXD9802C_I2S_MCLK_DIV_6(n)	CXD9802S_I2S_MCLK_DIV(n, 5)
#define  CXD9802C_I2S_MCLK_DIV_8(n)	CXD9802S_I2S_MCLK_DIV(n, 6)
#define  CXD9802C_I2S_MCLK_DIV_9(n)	CXD9802S_I2S_MCLK_DIV(n, 7)
#define  CXD9802C_I2S_MCLK_DIV_12(n)	CXD9802S_I2S_MCLK_DIV(n, 8)
#define  CXD9802C_I2S_MCLK_DIV_24(n)	CXD9802S_I2S_MCLK_DIV(n, 9)
/* CXD9802_I2S_I2SCHD(n) */

/* ======================================================================= *
 * type and structure                                                      *
 * ======================================================================= */

typedef struct cxd9802_i2s_clkchannel_t {
	int		 num;
	int		 dsp_unit_num;
	int		 mixer_unit_num;
	int		 dsp_opened;
	int		 mixer_nopened;
	int		 nchannels;			/* globallock */
	int		 format;
	int		 freq;
	int		 pcm_level;
	int		 state;				/* irqlock(i2s) */
#define CXD9802_I2S_STATE_IDLE		0
#define CXD9802_I2S_STATE_TX		1
#define CXD9802_I2S_STATE_RX		2
#define CXD9802_I2S_STATE_CANCEL	3
	int		 waiting_for_idle;
	dma_addr_t	 buf_bus;
	void		*buf_virt;
	size_t		 buf_size;
	int		 buf_head;
	int		 buf_dma_tail;
	int		 buf_tail;
	int		 buf_ndata;			/* irqlock(dma) */
	int		 buf_dma_ndata;			/* irqlock(dma) */
	int		 max_dma_size;
	struct semaphore  lock;
	wait_queue_head_t wait;
} cxd9802_i2s_clkchannel_t;

/* DMA command descripter for simple chain transfer */
typedef struct cxd9802_i2s_dma_scdesc_t {
	u64	dmchar;
	u64	dmsar;
	u64	dmdar;
	u64	dmcntr;
} cxd9802_i2s_dma_scdesc_t;

#define CXD9802_I2S_DMA_SCDESC_SIZE	128	/* 32 double words */

typedef struct cxd9802_i2s_dmachannel_t {
	int	 num;
	int	 running;				/* irqlock(dma) */
	dma_addr_t			 scdesc_bus;
	cxd9802_i2s_dma_scdesc_t	*scdesc;
	int	*dma_sizes;
	size_t	 scdesc_nelem;
	int	 scdesc_head;
	int	 scdesc_tail;
	int	 scdesc_ndata;				/* irqlock(dma) */
} cxd9802_i2s_dmachannel_t;

/* ======================================================================= *
 * prototype declaration                                                   *
 * ======================================================================= */

/* file operation (dsp) */
static int cxd9802_i2s_dsp_open(struct inode *inode, struct file *filp);
static int cxd9802_i2s_dsp_release(struct inode *inode, struct file *filp);
static ssize_t cxd9802_i2s_dsp_read(struct file *filp, char *buf,
				    size_t count, loff_t *f_pos);
static ssize_t cxd9802_i2s_dsp_write(struct file *filp, const char *buf,
				     size_t count, loff_t *f_pos);
static int cxd9802_i2s_dsp_ioctl(struct inode *inode, struct file *filp,
				 unsigned int cmd, unsigned long arg);

/* file operation (mixer) */
static int cxd9802_i2s_mixer_open(struct inode *inode, struct file *filp);
static int cxd9802_i2s_mixer_release(struct inode *inode, struct file *filp);
static int cxd9802_i2s_mixer_ioctl(struct inode *inode, struct file *filp,
				   unsigned int cmd, unsigned long arg);

/* interrupt handler */
static void cxd9802_i2s_intr_handler(int irq, void *dev_id,
				     struct pt_regs *regs);
static void cxd9802_i2s_dma_intr_handler(int irq, void *dev_id,
					 struct pt_regs *regs);

/* ioctl (dsp/mixer) */
static int cxd9802_i2s_ioctl_read_devmask(cxd9802_i2s_clkchannel_t *clkchannel,
					  unsigned long arg);
static int cxd9802_i2s_ioctl_read_stereodevs(
	cxd9802_i2s_clkchannel_t *clkchannel, unsigned long arg);
static int cxd9802_i2s_ioctl_read_recmask(cxd9802_i2s_clkchannel_t *clkchannel,
					  unsigned long arg);
static int cxd9802_i2s_ioctl_read_recsrc(cxd9802_i2s_clkchannel_t *clkchannel,
					 unsigned long arg);
static int cxd9802_i2s_ioctl_write_pcm(cxd9802_i2s_clkchannel_t *clkchannel,
				       unsigned long arg);
static int cxd9802_i2s_ioctl_read_pcm(cxd9802_i2s_clkchannel_t *clkchannel,
				      unsigned long arg);

/* ioctl (dsp) */
static int cxd9802_i2s_ioctl_channels(cxd9802_i2s_clkchannel_t *clkchannel,
				      unsigned long arg);
static int cxd9802_i2s_ioctl_read_channels(
	cxd9802_i2s_clkchannel_t *clkchannel, unsigned long arg);
static int cxd9802_i2s_ioctl_setfmt(cxd9802_i2s_clkchannel_t *clkchannel,
				    unsigned long arg);
static int cxd9802_i2s_ioctl_getfmts(cxd9802_i2s_clkchannel_t *clkchannel,
				     unsigned long arg);
static int cxd9802_i2s_ioctl_speed(cxd9802_i2s_clkchannel_t *clkchannel,
				   unsigned long arg);
static int cxd9802_i2s_ioctl_read_rate(cxd9802_i2s_clkchannel_t *clkchannel,
				       unsigned long arg);
static int cxd9802_i2s_ioctl_read_bits(cxd9802_i2s_clkchannel_t *clkchannel,
				       unsigned long arg);
static int cxd9802_i2s_ioctl_getblksize(cxd9802_i2s_clkchannel_t *clkchannel,
					unsigned long arg);
static int cxd9802_i2s_ioctl_sync(cxd9802_i2s_clkchannel_t *clkchannel,
				  unsigned long arg);
static int cxd9802_i2s_ioctl_post(cxd9802_i2s_clkchannel_t *clkchannel,
				  unsigned long arg);

/* init/finalization */
static int cxd9802_i2s_init(void);
static void cxd9802_i2s_exit(void);

static int cxd9802_i2s_clkchannel_init(cxd9802_i2s_clkchannel_t *clkchannel);
static void cxd9802_i2s_clkchannel_final(cxd9802_i2s_clkchannel_t *clkchannel);
static int cxd9802_i2s_buf_init(cxd9802_i2s_clkchannel_t *clkchannel);
static void cxd9802_i2s_buf_final(cxd9802_i2s_clkchannel_t *clkchannel);

static int cxd9802_i2s_dmachannel_init(cxd9802_i2s_dmachannel_t *dmachannel);
static void cxd9802_i2s_dmachannel_final(cxd9802_i2s_dmachannel_t *dmachannel);

/* set I2S parameters */
static void cxd9802_i2s_set_nchannels(cxd9802_i2s_clkchannel_t *clkchannel,
				      int nchannels);
static void cxd9802_i2s_set_format(cxd9802_i2s_clkchannel_t *clkchannel,
				   int format);
static int cxd9802_i2s_set_freq(cxd9802_i2s_clkchannel_t *clkchannel,
				int freq);

/* txrx operation */
static void cxd9802_i2s_txrx_start_ch(cxd9802_i2s_clkchannel_t *clkchannel);
static void cxd9802_i2s_txrx_stop_ch(cxd9802_i2s_clkchannel_t *clkchannel);
static void cxd9802_i2s_tx_start(cxd9802_i2s_clkchannel_t *clkchannel);
static void cxd9802_i2s_rx_start(cxd9802_i2s_clkchannel_t *clkchannel);
static void cxd9802_i2s_txrx_cancel(cxd9802_i2s_clkchannel_t *clkchannel);

/* DMA operation */
static void cxd9802_i2s_dma_scdesc_add(cxd9802_i2s_dmachannel_t *dmachannel,
				       u64 dmsar, u64 dmdar, u64 dmcntr,
				       int dma_size);
static void cxd9802_i2s_schain_dma_start(int num, int sadinc, int dadinc);
static void cxd9802_i2s_tx_dma_start(cxd9802_i2s_clkchannel_t *clkchannel);
static void cxd9802_i2s_rx_dma_start(cxd9802_i2s_clkchannel_t *clkchannel);

static void cxd9802_i2s_dma_txrx_comp_intr(int dmachannel_num);

/* misc */
static int cxd9802_i2s_wait_for_idle(cxd9802_i2s_clkchannel_t *clkchannel);

/* DAC/ADC dependent */
static int cxd9802_i2s_dacadc_init(void);
static void cxd9802_i2s_dacadc_final(void);
static int cxd9802_i2s_dacadc_set_mode(int channel, int mode);
static int cxd9802_i2s_dacadc_set_nchannels(int channel, int nchannels);
static void cxd9802_i2s_dac_mute(int channel, int on);
static int cxd9802_i2s_dac_set_pcm_level(int channel, int level);

/* ======================================================================= *
 * global variable                                                         *
 * ======================================================================= */

static struct file_operations cxd9802_i2s_dsp_fops = {
	owner:		THIS_MODULE,
	open:		cxd9802_i2s_dsp_open,
	release:	cxd9802_i2s_dsp_release,
	read:		cxd9802_i2s_dsp_read,
	write:		cxd9802_i2s_dsp_write,
	ioctl:		cxd9802_i2s_dsp_ioctl,
};

static struct file_operations cxd9802_i2s_mixer_fops = {
	owner:		THIS_MODULE,
	open:		cxd9802_i2s_mixer_open,
	release:	cxd9802_i2s_mixer_release,
	ioctl:		cxd9802_i2s_mixer_ioctl,
};

static spinlock_t cxd9802_i2s_irqlock = SPIN_LOCK_UNLOCKED;
static struct semaphore cxd9802_i2s_globallock;

static cxd9802_i2s_clkchannel_t
	cxd9802_i2s_clkchannels[CXD9802_I2S_NCLKCHANNELS];

static cxd9802_i2s_dmachannel_t
	cxd9802_i2s_dmachannels[CXD9802_I2S_NDMACHANNELS];

/* ======================================================================= *
 * inline function                                                         *
 * ======================================================================= */

static inline void
cxd9802_i2s_buf_clear(cxd9802_i2s_clkchannel_t *clkchannel)
{
	unsigned long flags;
	DPRFS;

	spin_lock_irqsave(&cxd9802_i2s_irqlock, flags);
	clkchannel->buf_ndata = 0;
	clkchannel->buf_dma_ndata = 0;
	clkchannel->buf_head = 0;
	clkchannel->buf_dma_tail = 0;
	clkchannel->buf_tail = 0;
	spin_unlock_irqrestore(&cxd9802_i2s_irqlock, flags);

	DPRFE;
	return;
}

static inline int
cxd9802_i2s_all_dmachannels_was_completed(cxd9802_i2s_clkchannel_t *clkchannel,
					  cxd9802_i2s_dmachannel_t *dmachannel)
{
	int was_completed;
	DPRFS;

	was_completed = 0;
	if (clkchannel->nchannels == 2) {
		was_completed = 1;
	} else if (clkchannel->nchannels == 4 &&
		   (cxd9802_i2s_dmachannels[0].scdesc_ndata <=
		    dmachannel->scdesc_ndata) &&
		   (cxd9802_i2s_dmachannels[1].scdesc_ndata <=
		    dmachannel->scdesc_ndata)) {
		was_completed = 1;
	} else if (clkchannel->nchannels == 6 &&
		   (cxd9802_i2s_dmachannels[0].scdesc_ndata <=
		    dmachannel->scdesc_ndata) &&
		   (cxd9802_i2s_dmachannels[1].scdesc_ndata <=
		    dmachannel->scdesc_ndata) &&
		   (cxd9802_i2s_dmachannels[2].scdesc_ndata <=
		    dmachannel->scdesc_ndata)) {
		was_completed = 1;
	}

	DPRFE;
	return was_completed;
}

static inline int
cxd9802_i2s_dmachannel_is_full(cxd9802_i2s_clkchannel_t *clkchannel)
{
	int num = clkchannel->num;
	int is_full;
	DPRFS;

	is_full = ((cxd9802_i2s_dmachannels[num].scdesc_ndata ==
		    cxd9802_i2s_dmachannels[num].scdesc_nelem) ||
		   (clkchannel->nchannels >= 4 &&
		    cxd9802_i2s_dmachannels[1].scdesc_ndata ==
		    cxd9802_i2s_dmachannels[1].scdesc_nelem) ||
		   (clkchannel->nchannels == 6 &&
		    cxd9802_i2s_dmachannels[2].scdesc_ndata ==
		    cxd9802_i2s_dmachannels[2].scdesc_nelem));

	DPRFE;
	return is_full;
}

/* ------------------------ *
 * level pack/unpack        *
 * ------------------------ */

static inline int
cxd9802_i2s_level(int left, int right)
{
	return ((left & 0xff) | ((right & 0xff) << 8));
}

static inline int
cxd9802_i2s_level_left(int level)
{
	return (level & 0xff);
}

static inline int
cxd9802_i2s_level_right(int level)
{
	return ((level >> 8) & 0xff);
}

/* ======================================================================= *
 * Linux driver interface                                                  *
 * ======================================================================= */

/* ------------------------ *
 * file operation (dsp)     *
 * ------------------------ */

static int
cxd9802_i2s_dsp_open(struct inode *inode, struct file *filp)
{
	int res;
	int i;
	cxd9802_i2s_clkchannel_t *clkchannel;
	DPRFS;

	/* minor to clkchannel */
	for (i = 0; i < CXD9802_I2S_NCLKCHANNELS; i++) {
		clkchannel = cxd9802_i2s_clkchannels + i;
		if (MINOR(inode->i_rdev) == clkchannel->dsp_unit_num) {
			break;
		}
	}
	if (i == CXD9802_I2S_NCLKCHANNELS) {
		DPRFR;
		return -EINVAL;
	}

	res = down_interruptible(&clkchannel->lock);
	if (res < 0) {
		DPRFR;
		return res;
	}
	if (clkchannel->dsp_opened) {
		up(&clkchannel->lock);
		DPRFR;
		return -EBUSY;
	}

	if (clkchannel->state == CXD9802_I2S_STATE_RX) {
		cxd9802_i2s_txrx_cancel(clkchannel);
	}
	if (clkchannel->state != CXD9802_I2S_STATE_IDLE) {
		res = cxd9802_i2s_wait_for_idle(clkchannel);
		if (res < 0) {
			up(&clkchannel->lock);
			DPRFR;
			return res;
		}
	}

	res = down_interruptible(&cxd9802_i2s_globallock);
	if (res < 0) {
		up(&clkchannel->lock);
		DPRFR;
		return res;
	}
	if (clkchannel->num == 1 &&
	    cxd9802_i2s_clkchannels[0].nchannels > 2) {
		up(&cxd9802_i2s_globallock);
		up(&clkchannel->lock);
		DPRFR;
		return -EBUSY;
	}
	cxd9802_i2s_set_nchannels(clkchannel, 2);
	cxd9802_i2s_set_format(clkchannel, AFMT_S16_LE);
	clkchannel->dsp_opened = 1;
	up(&cxd9802_i2s_globallock);

	up(&clkchannel->lock);

	filp->private_data = clkchannel;

	MOD_INC_USE_COUNT;

	DPRFE;
	return 0;
}

static int
cxd9802_i2s_dsp_release(struct inode *inode, struct file *filp)
{
	cxd9802_i2s_clkchannel_t *clkchannel = filp->private_data;
	DPRFS;

	down(&clkchannel->lock);
	if (clkchannel->state == CXD9802_I2S_STATE_RX) {
		cxd9802_i2s_txrx_cancel(clkchannel);
	}
	if (clkchannel->state != CXD9802_I2S_STATE_IDLE) {
		cxd9802_i2s_wait_for_idle(clkchannel);
		clkchannel->state = CXD9802_I2S_STATE_IDLE;
	}
	up(&clkchannel->lock);

	clkchannel->dsp_opened = 0;

	MOD_DEC_USE_COUNT;

	DPRFE;
	return 0;
}

static ssize_t
cxd9802_i2s_dsp_read(struct file *filp, char *buf, size_t count, loff_t *f_pos)
{
	cxd9802_i2s_clkchannel_t *clkchannel = filp->private_data;
	int res;
	int nread;
	int copy_len;
	unsigned long flags;
	DPRFS;

	if (clkchannel->nchannels > 2) {
		DPRFR;
		return -ENOTTY;
	}

	res = access_ok(VERIFY_WRITE, buf, count);
	if (!res) {
		DPRFR;
		return -EFAULT;
	}

	count = count - count % (clkchannel->nchannels * 2);	/* alignment */

	res = down_interruptible(&clkchannel->lock);
	if (res < 0) {
		DPRFR;
		return res;
	}

	if (clkchannel->state != CXD9802_I2S_STATE_RX &&
	    clkchannel->state != CXD9802_I2S_STATE_IDLE) {
		res = cxd9802_i2s_wait_for_idle(clkchannel);
		if (res < 0) {
			up(&clkchannel->lock);
			DPRFR;
			return res;
		}
	}
	if (clkchannel->state == CXD9802_I2S_STATE_IDLE) {
		cxd9802_i2s_buf_clear(clkchannel);
	}

	res = cxd9802_i2s_dacadc_set_mode(clkchannel->num,
					  CXD9802_I2S_DACADC_MODE_RX);
	if (res < 0) {
		up(&clkchannel->lock);
		DPRFR;
		return res;
	}

	nread = 0;
	while (count > 0) {
		cxd9802_i2s_rx_start(clkchannel);

		res = wait_event_interruptible(clkchannel->wait,
					       clkchannel->buf_ndata > 0);
		if (res < 0) {
			up(&clkchannel->lock);
			DPRFR;
			return res;
		}

		copy_len = clkchannel->buf_tail - clkchannel->buf_head;
		if (copy_len <= 0) {
			copy_len = clkchannel->buf_size - clkchannel->buf_head;
		}
		if (copy_len > count) {
			copy_len = count;
		}

		res = copy_to_user(buf + nread,
				   clkchannel->buf_virt +
				   clkchannel->buf_head,
				   copy_len);
		if (res != 0) {
			up(&clkchannel->lock);
			DPRFR;
			return -EFAULT;
		}

		spin_lock_irqsave(&cxd9802_i2s_irqlock, flags);
		clkchannel->buf_ndata -= copy_len;
		clkchannel->buf_dma_ndata -= copy_len;
		spin_unlock_irqrestore(&cxd9802_i2s_irqlock, flags);

		clkchannel->buf_head += copy_len;
		if (clkchannel->buf_head == clkchannel->buf_size) {
			clkchannel->buf_head = 0;
		}

		nread += copy_len;
		count -= copy_len;
	}
	up(&clkchannel->lock);

	DPRFE;
	return nread;
}

static ssize_t
cxd9802_i2s_dsp_write(struct file *filp, const char *buf, size_t count,
		      loff_t *f_pos)
{
	cxd9802_i2s_clkchannel_t *clkchannel = filp->private_data;
	int res;
	int nwrote;
	int copy_len;
	unsigned long flags;
	DPRFS;

	res = access_ok(VERIFY_READ, buf, count);
	if (!res) {
		DPRFR;
		return -EFAULT;
	}

	count = count - count % (clkchannel->nchannels * 2);	/* alignment */

	res = down_interruptible(&clkchannel->lock);
	if (res < 0) {
		DPRFR;
		return res;
	}

	if (clkchannel->state == CXD9802_I2S_STATE_RX) {
		cxd9802_i2s_txrx_cancel(clkchannel);
	}
	if (clkchannel->state != CXD9802_I2S_STATE_TX &&
	    clkchannel->state != CXD9802_I2S_STATE_IDLE) {
		res = cxd9802_i2s_wait_for_idle(clkchannel);
		if (res < 0) {
			up(&clkchannel->lock);
			DPRFR;
			return res;
		}
	}
	if (clkchannel->state == CXD9802_I2S_STATE_IDLE) {
		cxd9802_i2s_buf_clear(clkchannel);
	}

	res = cxd9802_i2s_dacadc_set_mode(clkchannel->num,
					  CXD9802_I2S_DACADC_MODE_TX);
	if (res < 0) {
		up(&clkchannel->lock);
		DPRFR;
		return res;
	}

	nwrote = 0;
	while (count > 0) {
		res = wait_event_interruptible(clkchannel->wait,
					       clkchannel->buf_ndata <
					       clkchannel->buf_size);
		if (res < 0) {
			up(&clkchannel->lock);
			DPRFR;
			return res;
		}

		copy_len = clkchannel->buf_head - clkchannel->buf_tail;
		if (copy_len <= 0) {
			copy_len = clkchannel->buf_size - clkchannel->buf_tail;
		}
		if (copy_len > count) {
			copy_len = count;
		}

		res = copy_from_user(clkchannel->buf_virt +
				     clkchannel->buf_tail,
				     buf + nwrote, copy_len);
		if (res != 0) {
			up(&clkchannel->lock);
			DPRFR;
			return -EFAULT;
		}

		spin_lock_irqsave(&cxd9802_i2s_irqlock, flags);
		clkchannel->buf_ndata += copy_len;
		clkchannel->buf_dma_ndata += copy_len;
		spin_unlock_irqrestore(&cxd9802_i2s_irqlock, flags);

		clkchannel->buf_tail += copy_len;
		if (clkchannel->buf_tail == clkchannel->buf_size) {
			clkchannel->buf_tail = 0;
		}

		nwrote += copy_len;
		count -= copy_len;

		cxd9802_i2s_tx_start(clkchannel);
	}
	up(&clkchannel->lock);

	DPRFE;
	return nwrote;
}

static int
cxd9802_i2s_dsp_ioctl(struct inode *inode, struct file *filp,
		      unsigned int cmd, unsigned long arg)
{
	cxd9802_i2s_clkchannel_t *clkchannel = filp->private_data;
	int res;
	DPRFS;

	res = down_interruptible(&clkchannel->lock);
	if (res < 0) {
		DPRFR;
		return res;
	}

	switch (cmd) {
	case SOUND_MIXER_READ_DEVMASK:
		res = cxd9802_i2s_ioctl_read_devmask(clkchannel, arg);
		break;
	case SOUND_MIXER_READ_STEREODEVS:
		res = cxd9802_i2s_ioctl_read_stereodevs(clkchannel, arg);
		break;
	case SOUND_MIXER_READ_RECMASK:
		res = cxd9802_i2s_ioctl_read_recmask(clkchannel, arg);
		break;
	case SOUND_MIXER_READ_RECSRC:
		res = cxd9802_i2s_ioctl_read_recsrc(clkchannel, arg);
		break;
	case SOUND_MIXER_WRITE_PCM:
		res = cxd9802_i2s_ioctl_write_pcm(clkchannel, arg);
		break;
	case SOUND_MIXER_READ_PCM:
		res = cxd9802_i2s_ioctl_read_pcm(clkchannel, arg);
		break;
	case SNDCTL_DSP_CHANNELS:
		res = cxd9802_i2s_ioctl_channels(clkchannel, arg);
		break;
	case SOUND_PCM_READ_CHANNELS:
		res = cxd9802_i2s_ioctl_read_channels(clkchannel, arg);
		break;
	case SNDCTL_DSP_SETFMT:
		res = cxd9802_i2s_ioctl_setfmt(clkchannel, arg);
		break;
	case SNDCTL_DSP_GETFMTS:
		res = cxd9802_i2s_ioctl_getfmts(clkchannel, arg);
		break;
	case SNDCTL_DSP_SPEED:
		res = cxd9802_i2s_ioctl_speed(clkchannel, arg);
		break;
	case SOUND_PCM_READ_RATE:
		res = cxd9802_i2s_ioctl_read_rate(clkchannel, arg);
		break;
	case SOUND_PCM_READ_BITS:
		res = cxd9802_i2s_ioctl_read_bits(clkchannel, arg);
		break;
	case SNDCTL_DSP_GETBLKSIZE:
		res = cxd9802_i2s_ioctl_getblksize(clkchannel, arg);
		break;
	case SNDCTL_DSP_SYNC:
		res = cxd9802_i2s_ioctl_sync(clkchannel, arg);
		break;
	case SNDCTL_DSP_POST:
		res = cxd9802_i2s_ioctl_post(clkchannel, arg);
		break;
	default:
		res = -EINVAL;
	}

	up(&clkchannel->lock);

	DPRFE;
	return res;
}

/* ------------------------ *
 * file operation (mixer)   *
 * ------------------------ */

static int
cxd9802_i2s_mixer_open(struct inode *inode, struct file *filp)
{
	int res;
	int i;
	cxd9802_i2s_clkchannel_t *clkchannel;
	DPRFS;

	/* minor to clkchannel */
	for (i = 0; i < CXD9802_I2S_NCLKCHANNELS; i++) {
		clkchannel = cxd9802_i2s_clkchannels + i;
		if (MINOR(inode->i_rdev) == clkchannel->mixer_unit_num) {
			break;
		}
	}
	if (i == CXD9802_I2S_NCLKCHANNELS) {
		DPRFR;
		return -EINVAL;
	}

	res = down_interruptible(&cxd9802_i2s_globallock);
	if (res < 0) {
		up(&clkchannel->lock);
		DPRFR;
		return res;
	}
	if (clkchannel->num == 1 &&
	    cxd9802_i2s_clkchannels[0].nchannels > 2) {
		up(&cxd9802_i2s_globallock);
		DPRFR;
		return -EBUSY;
	}
	clkchannel->mixer_nopened++;
	up(&cxd9802_i2s_globallock);

	filp->private_data = clkchannel;

	MOD_INC_USE_COUNT;

	DPRFE;
	return 0;
}

static int
cxd9802_i2s_mixer_release(struct inode *inode, struct file *filp)
{
	cxd9802_i2s_clkchannel_t *clkchannel = filp->private_data;
	DPRFS;

	down(&cxd9802_i2s_globallock);
	clkchannel->mixer_nopened--;
	up(&cxd9802_i2s_globallock);

	MOD_DEC_USE_COUNT;

	DPRFE;
	return 0;
}

static int
cxd9802_i2s_mixer_ioctl(struct inode *inode, struct file *filp,
			unsigned int cmd, unsigned long arg)
{
	cxd9802_i2s_clkchannel_t *clkchannel = filp->private_data;
	int res;
	DPRFS;

	res = down_interruptible(&clkchannel->lock);
	if (res < 0) {
		DPRFR;
		return res;
	}

	switch (cmd) {
	case SOUND_MIXER_READ_DEVMASK:
		res = cxd9802_i2s_ioctl_read_devmask(clkchannel, arg);
		break;
	case SOUND_MIXER_READ_STEREODEVS:
		res = cxd9802_i2s_ioctl_read_stereodevs(clkchannel, arg);
		break;
	case SOUND_MIXER_READ_RECMASK:
		res = cxd9802_i2s_ioctl_read_recmask(clkchannel, arg);
		break;
	case SOUND_MIXER_READ_RECSRC:
		res = cxd9802_i2s_ioctl_read_recsrc(clkchannel, arg);
		break;
	case SOUND_MIXER_WRITE_PCM:
		res = cxd9802_i2s_ioctl_write_pcm(clkchannel, arg);
		break;
	case SOUND_MIXER_READ_PCM:
		res = cxd9802_i2s_ioctl_read_pcm(clkchannel, arg);
		break;
	default:
		res = -EINVAL;
	}

	up(&clkchannel->lock);

	DPRFE;
	return -EINVAL;
}

/* ======================================================================= *
 * interrupt handler                                                       *
 * ======================================================================= */

static void
cxd9802_i2s_intr_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	cxd9802_i2s_clkchannel_t *clkchannel;
	int i;
	u32 i2sictrl;
	DPRFS;

	i2sictrl = cxd9802_reg_read32(CXD9802_I2S_I2SICTRL);
	DPRFL(DPL_DEBUG, "I2SICTRL=%08x\n", i2sictrl);

	for (i = 0; i < CXD9802_I2S_NCLKCHANNELS; i++) {
		clkchannel = cxd9802_i2s_clkchannels + i;

		if (i2sictrl & CXD9802M_I2S_UORUN_INT(i)) {
			if (!clkchannel->waiting_for_idle &&
			    clkchannel->state != CXD9802_I2S_STATE_IDLE &&
			    clkchannel->state != CXD9802_I2S_STATE_CANCEL) {
				DPRFL(DPL_WARNING, "underrun/overrun (%d)",
				      clkchannel->state);
			}
			cxd9802_i2s_txrx_stop_ch(clkchannel);
			clkchannel->state = CXD9802_I2S_STATE_IDLE;
			wake_up_interruptible(&clkchannel->wait);
		}
	}

	cxd9802_reg_write32(CXD9802_I2S_I2SICTRL, i2sictrl);

	DPRFE;
	return;
}

static void
cxd9802_i2s_dma_intr_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	int dmachannel_num;
	u64 dmcsr;
	DPRFS;

	dmachannel_num = irq - CXD9802_IRQ_DMA1(0);

	/* read and clear interrupt status */
	dmcsr = cxd9802_reg_read64(CXD9802_DMAC_DMCSR(1, dmachannel_num));
	cxd9802_reg_write64(CXD9802_DMAC_DMCSR(1, dmachannel_num), dmcsr);

	if (dmcsr & CXD9802M_DMAC_ABCHC) {
		/**** TODO ****/
		DPRFL(DPL_WARNING, "dma error[%d] %016llx\n",
		      dmachannel_num, dmcsr);
	}
	if (dmcsr & CXD9802M_DMAC_NCHNC) {
		cxd9802_i2s_dmachannels[dmachannel_num].running = 0;
	}
	if (dmcsr & CXD9802M_DMAC_NTRNFC) {
		cxd9802_i2s_dma_txrx_comp_intr(dmachannel_num);
	}

	DPRFE;
	return;
}

/* ======================================================================= *
 * ioctl internal                                                          *
 * ======================================================================= */

/* ------------------------ *
 * ioctl (dsp/mixer)        *
 * ------------------------ */

static int
cxd9802_i2s_ioctl_read_devmask(cxd9802_i2s_clkchannel_t *clkchannel,
			       unsigned long arg)
{
	int res;
	int devmask;
	DPRFS;

	devmask = 0;	/**** TODO ****/

	res = copy_to_user((void *)arg, &devmask, sizeof(devmask));
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}

	DPRFE;
	return 0;
}

static int
cxd9802_i2s_ioctl_read_stereodevs(cxd9802_i2s_clkchannel_t *clkchannel,
				  unsigned long arg)
{
	int res;
	int stereodevs;
	DPRFS;

	stereodevs = 0;		/**** TODO ****/

	res = copy_to_user((void *)arg, &stereodevs, sizeof(stereodevs));
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}

	DPRFE;
	return 0;
}

static int
cxd9802_i2s_ioctl_read_recmask(cxd9802_i2s_clkchannel_t *clkchannel,
			       unsigned long arg)
{
	int res;
	int recmask;
	DPRFS;

	recmask = 0;

	res = copy_to_user((void *)arg, &recmask, sizeof(recmask));
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}

	DPRFE;
	return 0;
}

static int
cxd9802_i2s_ioctl_read_recsrc(cxd9802_i2s_clkchannel_t *clkchannel,
			      unsigned long arg)
{
	int res;
	int recsrc;
	DPRFS;

	recsrc = SOUND_MASK_MIC;

	res = copy_to_user((void *)arg, &recsrc, sizeof(recsrc));
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}

	DPRFE;
	return 0;
}

static int
cxd9802_i2s_ioctl_write_pcm(cxd9802_i2s_clkchannel_t *clkchannel,
			    unsigned long arg)
{
	int res;
	int level;
	DPRFS;

	res = copy_from_user(&level, (void *)arg, sizeof(level));
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}

	res = cxd9802_i2s_dac_set_pcm_level(clkchannel->num, level);
	if (res < 0) {
		DPRFR;
		return res;
	}
	clkchannel->pcm_level = res;

	DPRFE;
	return res;
}

static int
cxd9802_i2s_ioctl_read_pcm(cxd9802_i2s_clkchannel_t *clkchannel,
			   unsigned long arg)
{
	int res;
	DPRFS;

	res = copy_to_user((void *)arg, &clkchannel->pcm_level,
			   sizeof(clkchannel->pcm_level));
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}

	DPRFE;
	return 0;
}

/* ------------------------ *
 * ioctl (dsp)              *
 * ------------------------ */

static int
cxd9802_i2s_ioctl_channels(cxd9802_i2s_clkchannel_t *clkchannel,
			   unsigned long arg)
{
	int res;
	int channels;
	DPRFS;

	if (clkchannel->state != CXD9802_I2S_STATE_IDLE) {
		DPRFR;
		return -EBUSY;
	}

	res = copy_from_user(&channels, (void *)arg, sizeof(channels));
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}

	res = down_interruptible(&cxd9802_i2s_globallock);
	if (res < 0) {
		DPRFR;
		return res;
	}

	if ((channels == 4 || channels == 6) &&
	    (cxd9802_i2s_clkchannels[1].dsp_opened ||
	     cxd9802_i2s_clkchannels[1].mixer_nopened > 0 ||
	     cxd9802_i2s_clkchannels[1].state != CXD9802_I2S_STATE_IDLE)) {
		up(&cxd9802_i2s_globallock);
		DPRFR;
		return -EBUSY;
	}

	if (clkchannel->num != 0 || !(channels == 4 || channels == 6)) {
		channels = 2;
	}

	res = cxd9802_i2s_dacadc_set_nchannels(clkchannel->num, channels);
	if (res < 0) {
		up(&cxd9802_i2s_globallock);
		DPRFR;
		return res;
	}

	cxd9802_i2s_set_nchannels(clkchannel, channels);

	up(&cxd9802_i2s_globallock);

	res = copy_to_user((void *)arg, &channels, sizeof(channels));
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}

	DPRFE;
	return 0;
}

static int
cxd9802_i2s_ioctl_read_channels(cxd9802_i2s_clkchannel_t *clkchannel,
				unsigned long arg)
{
	int res;
	DPRFS;

	res = copy_to_user((void *)arg, &clkchannel->nchannels,
			   sizeof(clkchannel->nchannels));
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}

	DPRFE;
	return 0;
}

static int
cxd9802_i2s_ioctl_setfmt(cxd9802_i2s_clkchannel_t *clkchannel,
			 unsigned long arg)
{
	int res;
	int fmt;
	DPRFS;

	if (clkchannel->state != CXD9802_I2S_STATE_IDLE) {
		DPRFR;
		return -EBUSY;
	}

	res = copy_from_user(&fmt, (void *)arg, sizeof(fmt));
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}

	if (fmt != AFMT_S16_LE || fmt != AFMT_S16_BE) {
		fmt = AFMT_S16_LE;
	}

	cxd9802_i2s_set_format(clkchannel, fmt);

	res = copy_to_user((void *)arg, &fmt, sizeof(fmt));
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}

	DPRFE;
	return 0;
}

static int
cxd9802_i2s_ioctl_getfmts(cxd9802_i2s_clkchannel_t *clkchannel,
			  unsigned long arg)
{
	int res;
	DPRFS;

	res = copy_to_user((void *)arg, &clkchannel->format,
			   sizeof(clkchannel->format));
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}

	DPRFE;
	return 0;
}

static int
cxd9802_i2s_ioctl_speed(cxd9802_i2s_clkchannel_t *clkchannel,
			unsigned long arg)
{
	int res;
	int freq;
	DPRFS;

	if (clkchannel->state != CXD9802_I2S_STATE_IDLE) {
		DPRFR;
		return -EBUSY;
	}

	res = copy_from_user(&freq, (void *)arg, sizeof(freq));
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}

	res = cxd9802_i2s_set_freq(clkchannel, freq);
	if (res < 0) {
		DPRFR;
		return res;
	}
	freq = res;

	res = copy_to_user((void *)arg, &freq, sizeof(freq));
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}

	DPRFE;
	return 0;
}

static int
cxd9802_i2s_ioctl_read_rate(cxd9802_i2s_clkchannel_t *clkchannel,
			    unsigned long arg)
{
	int res;
	DPRFS;

	res = copy_to_user((void *)arg, &clkchannel->freq,
			   sizeof(clkchannel->freq));
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}

	DPRFE;
	return 0;
}

static int
cxd9802_i2s_ioctl_read_bits(cxd9802_i2s_clkchannel_t *clkchannel,
			    unsigned long arg)
{
	int res;
	int bits = 16;	/* 16 bits only */
	DPRFS;

	res = copy_to_user((void *)arg, &bits, sizeof(bits));
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}

	DPRFE;
	return 0;
}

static int
cxd9802_i2s_ioctl_getblksize(cxd9802_i2s_clkchannel_t *clkchannel,
			     unsigned long arg)
{
	int res;
	DPRFS;

	res = copy_to_user((void *)arg, &clkchannel->buf_size,
			   sizeof(clkchannel->buf_size));
	if (res != 0) {
		DPRFR;
		return -EFAULT;
	}

	DPRFE;
	return 0;
}

static int
cxd9802_i2s_ioctl_sync(cxd9802_i2s_clkchannel_t *clkchannel,
		       unsigned long arg)
{
	int res;
	DPRFS;

	res = cxd9802_i2s_wait_for_idle(clkchannel);
	if (res < 0) {
		DPRFR;
		return res;
	}

	DPRFE;
	return 0;
}

static int
cxd9802_i2s_ioctl_post(cxd9802_i2s_clkchannel_t *clkchannel,
		       unsigned long arg)
{
	DPRFS;

	cxd9802_i2s_txrx_cancel(clkchannel);

	DPRFE;
	return 0;
}

/* ======================================================================= *
 * internal function                                                       *
 * ======================================================================= */

/* ------------------------ *
 * init/finalization        *
 * ------------------------ */

static int
cxd9802_i2s_init(void)
{
	int res;
	int i;
	DPRFS;

	init_MUTEX(&cxd9802_i2s_globallock);

	/* set to use I2S pin */
	cxd9802_reg_changebits64(CXD9802_CONFIG_PCFG, CXD9802M_CONFIG_I2SMODE,
				 CXD9802C_CONFIG_I2SMODE_I2S);

	res = cxd9802_i2s_dacadc_init();
	if (res < 0) {
		goto err1;
	}

	for (i = 0; i < CXD9802_I2S_NDMACHANNELS; i++) {
		cxd9802_i2s_dmachannels[i].num = i;
		res = cxd9802_i2s_dmachannel_init(cxd9802_i2s_dmachannels + i);
		if (res < 0) {
			goto err2;
		}
	}

	for (i = 0; i < CXD9802_I2S_NCLKCHANNELS; i++) {
		cxd9802_i2s_clkchannels[i].num = i;
		res = cxd9802_i2s_clkchannel_init(cxd9802_i2s_clkchannels + i);
		if (res < 0) {
			goto err3;
		}
	}

	res = request_irq(CXD9802_IRQ_I2S, cxd9802_i2s_intr_handler,
			  SA_INTERRUPT, CXD9802_I2S_DEVNAME,
			  cxd9802_i2s_clkchannels);
	if (res < 0) {
		goto err4;
	}
	cxd9802_reg_clearbits32_atomic(CXD9802_I2S_I2SICTRL,
				       CXD9802M_I2S_UORUN_INT_MASK(0) |
				       CXD9802M_I2S_UORUN_INT_MASK(1));

	DPRFE;
	return 0;

err4:
	i = CXD9802_I2S_NCLKCHANNELS;
err3:
	for (; i > 0; i--) {
		cxd9802_i2s_clkchannel_final(cxd9802_i2s_clkchannels +
					     (i - 1));
	}
	i = CXD9802_I2S_NDMACHANNELS;
err2:
	for (; i > 0; i--) {
		cxd9802_i2s_dmachannel_final(cxd9802_i2s_dmachannels +
					     (i - 1));
	}
	cxd9802_i2s_dacadc_final();
err1:
	DPRFR;
	return res;
}

static void
cxd9802_i2s_exit(void)
{
	int i;
	DPRFS;

	free_irq(CXD9802_IRQ_I2S, cxd9802_i2s_clkchannels);
	for (i = 0; i < CXD9802_I2S_NCLKCHANNELS; i++) {
		cxd9802_i2s_clkchannel_final(cxd9802_i2s_clkchannels + i);
	}
	for (i = 0; i < CXD9802_I2S_NDMACHANNELS; i++) {
		cxd9802_i2s_dmachannel_final(cxd9802_i2s_dmachannels + i);
	}
	cxd9802_i2s_dacadc_final();

	DPRFE;
	return;
}

module_init(cxd9802_i2s_init);
module_exit(cxd9802_i2s_exit);

/* ------------------------ *
 * clkchannel init/final    *
 * ------------------------ */

static int
cxd9802_i2s_clkchannel_init(cxd9802_i2s_clkchannel_t *clkchannel)
{
	int res;
	DPRFS;

	clkchannel->buf_size = CXD9802_I2S_DEFAULT_BUF_SIZE;
	res = cxd9802_i2s_buf_init(clkchannel);
	if (res < 0) {
		goto err1;
	}

	res = register_sound_dsp(&cxd9802_i2s_dsp_fops, -1);
	if (res < 0) {
		goto err2;
	}
	clkchannel->dsp_unit_num = res;

	res = register_sound_mixer(&cxd9802_i2s_mixer_fops, -1);
	if (res < 0) {
		goto err3;
	}
	clkchannel->mixer_unit_num = res;

	clkchannel->dsp_opened = 0;
	clkchannel->mixer_nopened = 0;
	clkchannel->nchannels = 2;
	clkchannel->format = AFMT_S16_LE;
	clkchannel->freq = 44100;
	clkchannel->pcm_level = cxd9802_i2s_level(100, 100);
	clkchannel->state = CXD9802_I2S_STATE_IDLE;
	clkchannel->waiting_for_idle = 0;
	clkchannel->max_dma_size = CXD9802_I2S_MAX_DMA_SIZE;

	init_MUTEX(&clkchannel->lock);
	init_waitqueue_head(&clkchannel->wait);

	DPRFE;
	return 0;

err3:
	unregister_sound_dsp(clkchannel->dsp_unit_num);
err2:
	cxd9802_i2s_buf_final(clkchannel);
err1:
	DPRFR;
	return res;
}

static void
cxd9802_i2s_clkchannel_final(cxd9802_i2s_clkchannel_t *clkchannel)
{
	DPRFS;

	unregister_sound_mixer(clkchannel->mixer_unit_num);
	unregister_sound_dsp(clkchannel->dsp_unit_num);
	cxd9802_i2s_buf_final(clkchannel);

	DPRFE;
	return;
}

static int
cxd9802_i2s_buf_init(cxd9802_i2s_clkchannel_t *clkchannel)
{
	DPRFS;

	clkchannel->buf_virt = pci_alloc_consistent(NULL, clkchannel->buf_size,
						    &clkchannel->buf_bus);
	if (clkchannel->buf_virt == NULL) {
		DPRFR;
		return -ENOMEM;
	}
	cxd9802_i2s_buf_clear(clkchannel);

	DPRFE;
	return 0;
}

static void
cxd9802_i2s_buf_final(cxd9802_i2s_clkchannel_t *clkchannel)
{
	DPRFS;

	pci_free_consistent(NULL, clkchannel->buf_size, clkchannel->buf_virt,
			    clkchannel->buf_bus);
	clkchannel->buf_virt = NULL;

	DPRFE;
	return;
}

/* ------------------------ *
 * dmachannel init/final    *
 * ------------------------ */

static int
cxd9802_i2s_dmachannel_init(cxd9802_i2s_dmachannel_t *dmachannel)
{
	int res;
	int num = dmachannel->num;
	DPRFS;

	cxd9802_reg_setbits64_atomic(CXD9802_DMAC_DMMCR(1),
				     CXD9802M_DMAC_MSTEN);

	/* reset */
	cxd9802_reg_setbits64_atomic(CXD9802_DMAC_DMCCR(1, num),
				     CXD9802M_DMAC_CHRST);
	cxd9802_reg_clearbits64_atomic(CXD9802_DMAC_DMCCR(1, num),
				       CXD9802M_DMAC_CHRST);

	/* set */
	cxd9802_reg_changebits64_atomic(CXD9802_DMAC_DMCCR(1, num),
					CXD9802M_DMAC_IMMCHN |
					CXD9802M_DMAC_USEXFSZ |
					CXD9802M_DMAC_DBINH |
					CXD9802M_DMAC_SBINH |
					CXD9802M_DMAC_ACKPOL |
					CXD9802M_DMAC_REQPL |
					CXD9802M_DMAC_EGREQ |
					CXD9802M_DMAC_CHDN |
					CXD9802M_DMAC_DNCTL |
					CXD9802M_DMAC_EXTRQ |
					CXD9802M_DMAC_STLTIMEINTRQD |
					CXD9802M_DMAC_INTENE |
					CXD9802M_DMAC_INTENC |
					CXD9802M_DMAC_INTENT |
					CXD9802M_DMAC_SMPCHIN |
					CXD9802M_DMAC_XFSZ |
					CXD9802M_DMAC_SNGAD,
					CXD9802M_DMAC_IMMCHN |
					CXD9802M_DMAC_USEXFSZ |
					CXD9802M_DMAC_EXTRQ |
					CXD9802M_DMAC_SMPCHIN |
					CXD9802C_DMAC_XFSZ_4B);

	/* irq */
	res = request_irq(CXD9802_IRQ_DMA1(num),
			  cxd9802_i2s_dma_intr_handler,
			  SA_INTERRUPT,
			  CXD9802_I2S_DEVNAME,
			  cxd9802_i2s_clkchannels);
	if (res < 0) {
		goto err1;
	}

	dmachannel->scdesc_nelem = CXD9802_I2S_DEFAULT_SCDESC_NELEM;
	dmachannel->scdesc =
		pci_alloc_consistent(NULL,
				     sizeof(cxd9802_i2s_dma_scdesc_t) *
				     dmachannel->scdesc_nelem,
				     &dmachannel->scdesc_bus);
	if (dmachannel->scdesc == NULL) {
		res = -ENOMEM;
		goto err2;
	}
	dmachannel->dma_sizes = kmalloc(sizeof(int) * dmachannel->scdesc_nelem,
					GFP_KERNEL);
	if (dmachannel->dma_sizes == NULL) {
		res = -ENOMEM;
		goto err3;
	}
	dmachannel->running = 0;
	dmachannel->scdesc_ndata = 0;
	dmachannel->scdesc_head = 0;
	dmachannel->scdesc_tail = 0;

	DPRFE;
	return 0;

err3:
	pci_free_consistent(NULL,
			    sizeof(cxd9802_i2s_dma_scdesc_t) *
			    dmachannel->scdesc_nelem,
			    dmachannel->scdesc, dmachannel->scdesc_bus);
err2:
	free_irq(CXD9802_IRQ_DMA1(num), cxd9802_i2s_clkchannels);
err1:
	DPRFR;
	return res;
}

static void
cxd9802_i2s_dmachannel_final(cxd9802_i2s_dmachannel_t *dmachannel)
{
	int num = dmachannel->num;
	DPRFS;

	/**** TODO: stop dma ****/

	kfree(dmachannel->dma_sizes);
	pci_free_consistent(NULL,
			    sizeof(cxd9802_i2s_dma_scdesc_t) *
			    dmachannel->scdesc_nelem,
			    dmachannel->scdesc, dmachannel->scdesc_bus);
	dmachannel->scdesc = NULL;

	free_irq(CXD9802_IRQ_DMA1(num), cxd9802_i2s_clkchannels);

	DPRFE;
	return;
}

/* ------------------------ *
 * set I2S parameters       *
 * ------------------------ */

static void
cxd9802_i2s_set_nchannels(cxd9802_i2s_clkchannel_t *clkchannel, int nchannels)
{
	DPRFS;

	if (nchannels > 2) {
		cxd9802_reg_setbits32_atomic(CXD9802_I2S_I2SCCR,
					     CXD9802M_I2S_MODE);
	} else {
		cxd9802_reg_clearbits32_atomic(CXD9802_I2S_I2SCCR,
					       CXD9802M_I2S_MODE);
	}
	clkchannel->nchannels = nchannels;

	DPRFE;
	return;
}

static void
cxd9802_i2s_set_format(cxd9802_i2s_clkchannel_t *clkchannel, int format)
{
	int num = clkchannel->num;
	DPRFS;

	if (format == AFMT_S16_BE) {
		cxd9802_reg_setbits32_atomic(CXD9802_I2S_I2SCCR,
					     CXD9802M_I2S_DATA_SWAP(num));
	} else {
		cxd9802_reg_clearbits32_atomic(CXD9802_I2S_I2SCCR,
					       CXD9802M_I2S_DATA_SWAP(num));
	}
	clkchannel->format = format;

	DPRFE;
	return;
}

static int
cxd9802_i2s_set_freq(cxd9802_i2s_clkchannel_t *clkchannel, int freq)
{
	int num = clkchannel->num;
	u32 mclkctrl;
	DPRFS;

	cxd9802_reg_clearbits32_atomic(CXD9802_CONFIG_MCLKCTRL,
				       CXD9802M_CONFIG_MCLKBE(num));

	switch (freq) {
	case 44100:
		mclkctrl = CXD9802C_CONFIG_MCLKCTRL_44100(num);
		break;
	case 48000:
		mclkctrl = CXD9802C_CONFIG_MCLKCTRL_48000(num);
		break;
	case 32000:
		mclkctrl = CXD9802C_CONFIG_MCLKCTRL_32000(num);
		break;
	case 24000:
		mclkctrl = CXD9802C_CONFIG_MCLKCTRL_24000(num);
		break;
	case 16000:
		mclkctrl = CXD9802C_CONFIG_MCLKCTRL_16000(num);
		break;
	case 8000:
		mclkctrl = CXD9802C_CONFIG_MCLKCTRL_8000(num);
		break;
	default:
		freq = 44100;
		mclkctrl = CXD9802C_CONFIG_MCLKCTRL_44100(num);
	}

	cxd9802_reg_changebits32_atomic(CXD9802_CONFIG_MCLKCTRL,
					CXD9802M_CONFIG_MCLKCTRL(num),
					mclkctrl);
	clkchannel->freq = freq;

	cxd9802_reg_setbits32_atomic(CXD9802_CONFIG_MCLKCTRL,
				     CXD9802M_CONFIG_MCLKBE(num));

	DPRFE;
	return freq;
}

/* ------------------------ *
 * txrx operation           *
 * ------------------------ */

/* may be called with irqlock */
static void
cxd9802_i2s_txrx_start_ch(cxd9802_i2s_clkchannel_t *clkchannel)
{
	int num = clkchannel->num;
	u32 start_ch;
	DPRFS;

	start_ch = CXD9802M_I2S_START_CH(num);
	if (clkchannel->nchannels >= 4) {
		start_ch |= CXD9802M_I2S_START_CH(1);
	}
	if (clkchannel->nchannels == 6) {
		start_ch |= CXD9802M_I2S_START_CH(2);
	}
	cxd9802_reg_setbits32(CXD9802_I2S_I2SMCR, start_ch);

	DPRFE;
	return;
}

/* may be called with irqlock */
static void
cxd9802_i2s_txrx_stop_ch(cxd9802_i2s_clkchannel_t *clkchannel)
{
	int num = clkchannel->num;
	u32 stop_ch;
	DPRFS;

	stop_ch = CXD9802M_I2S_START_CH(num);
	if (clkchannel->nchannels >= 4) {
		stop_ch |= CXD9802M_I2S_START_CH(1);
	}
	if (clkchannel->nchannels == 6) {
		stop_ch |= CXD9802M_I2S_START_CH(2);
	}
	cxd9802_reg_clearbits32(CXD9802_I2S_I2SMCR, stop_ch);

	DPRFE;
	return;
}

static void
cxd9802_i2s_tx_start(cxd9802_i2s_clkchannel_t *clkchannel)
{
	int num = clkchannel->num;
	unsigned long flags;
	DPRFS;

	spin_lock_irqsave(&cxd9802_i2s_irqlock, flags);

	if (clkchannel->buf_ndata <= 0) {
		spin_unlock_irqrestore(&cxd9802_i2s_irqlock, flags);
		DPRFR;
		return;
	}

	cxd9802_reg_clearbits32(CXD9802_I2S_I2SCCR, CXD9802M_I2S_IO_MODE(num));
	clkchannel->state = CXD9802_I2S_STATE_TX;
	cxd9802_i2s_tx_dma_start(clkchannel);

	spin_unlock_irqrestore(&cxd9802_i2s_irqlock, flags);

	DPRFE;
	return;
}

static void
cxd9802_i2s_rx_start(cxd9802_i2s_clkchannel_t *clkchannel)
{
	int num = clkchannel->num;
	unsigned long flags;
	DPRFS;

	spin_lock_irqsave(&cxd9802_i2s_irqlock, flags);

	if (clkchannel->buf_ndata == clkchannel->buf_size) {
		spin_unlock_irqrestore(&cxd9802_i2s_irqlock, flags);
		DPRFR;
		return;
	}

	cxd9802_reg_setbits32(CXD9802_I2S_I2SCCR, CXD9802M_I2S_IO_MODE(num));
	clkchannel->state = CXD9802_I2S_STATE_RX;
	cxd9802_i2s_rx_dma_start(clkchannel);

	spin_unlock_irqrestore(&cxd9802_i2s_irqlock, flags);

	DPRFE;
	return;
}

static void
cxd9802_i2s_txrx_cancel(cxd9802_i2s_clkchannel_t *clkchannel)
{
	unsigned long flags;
	DPRFS;

	spin_lock_irqsave(&cxd9802_i2s_irqlock, flags);
	if (clkchannel->state == CXD9802_I2S_STATE_TX ||
	    clkchannel->state == CXD9802_I2S_STATE_RX) {
		clkchannel->state = CXD9802_I2S_STATE_CANCEL;
	}
	spin_unlock_irqrestore(&cxd9802_i2s_irqlock, flags);

	DPRFE;
	return;
}

/* ------------------------ *
 * DMA operation            *
 * ------------------------ */

/* should be called with irqlock */
static void
cxd9802_i2s_dma_scdesc_add(cxd9802_i2s_dmachannel_t *dmachannel,
			   u64 dmsar, u64 dmdar, u64 dmcntr, int dma_size)
{
	int tail, tail_prev;
	DPRFS;

	tail = dmachannel->scdesc_tail;
	dmachannel->scdesc[tail].dmchar = 0;
	dmachannel->scdesc[tail].dmsar = dmsar;
	dmachannel->scdesc[tail].dmdar = dmdar;
	dmachannel->scdesc[tail].dmcntr = dmcntr;
	dmachannel->dma_sizes[tail] = dma_size;
	if (dmachannel->scdesc_ndata > 0) {
		tail_prev = tail - 1;
		if (tail_prev < 0) {
			tail_prev = dmachannel->scdesc_nelem - 1;
		}
		dmachannel->scdesc[tail_prev].dmchar =
			dmachannel->scdesc_bus +
			sizeof(cxd9802_i2s_dma_scdesc_t) * tail;
	}
	tail++;
	if (tail == dmachannel->scdesc_nelem) {
		tail = 0;
	}
	dmachannel->scdesc_tail = tail;
	dmachannel->scdesc_ndata++;

	DPRFE;
	return;
}

/* should be called with irqlock */
static void
cxd9802_i2s_schain_dma_start(int num, int sadinc, int dadinc)
{
	DPRFS;

	cxd9802_reg_write64(CXD9802_DMAC_DMCNTR(1, num), 0);
	cxd9802_reg_write64(CXD9802_DMAC_DMSAR(1, num), 0);
	cxd9802_reg_write64(CXD9802_DMAC_DMDAR(1, num), 0);
	cxd9802_reg_write64(CXD9802_DMAC_DMSAIR(1, num), sadinc);
	cxd9802_reg_write64(CXD9802_DMAC_DMDAIR(1, num), dadinc);
	cxd9802_reg_changebits64(CXD9802_DMAC_DMCSR(1, num), 0, 0);
	cxd9802_reg_setbits64(CXD9802_DMAC_DMCCR(1, num),
			      CXD9802M_DMAC_INTENE | CXD9802M_DMAC_INTENC |
			      CXD9802M_DMAC_INTENT);
	cxd9802_reg_write64(CXD9802_DMAC_DMCHAR(1, num),
			    cxd9802_i2s_dmachannels[num].scdesc_bus +
			    sizeof(cxd9802_i2s_dma_scdesc_t) *
			    cxd9802_i2s_dmachannels[num].scdesc_head);

	DPRFE;
	return;
}

/* should be called with irqlock */
static void
cxd9802_i2s_tx_dma_start(cxd9802_i2s_clkchannel_t *clkchannel)
{
	int num = clkchannel->num;
	int dma_size;
	int dmcntr, sadinc;
	int scdesc_is_set;
	DPRFS;

	DPRFL(DPL_DEBUG, "num=%d\n", num);

	scdesc_is_set = 0;
	while (!cxd9802_i2s_dmachannel_is_full(clkchannel)) {
		if (clkchannel->buf_dma_ndata == 0) {
			break;
		}
		dma_size = clkchannel->buf_tail - clkchannel->buf_dma_tail;
		if (dma_size <= 0) {
			dma_size = (clkchannel->buf_size -
				    clkchannel->buf_dma_tail);
		}
		if (dma_size > clkchannel->max_dma_size) {
			dma_size = clkchannel->max_dma_size;
		}
		dma_size = dma_size - dma_size % (clkchannel->nchannels * 2);
		if (dma_size == 0) {
			break;
		}

		dmcntr = dma_size / (clkchannel->nchannels / 2);

		/* set chain DMA descripter */
		cxd9802_i2s_dma_scdesc_add(
			&(cxd9802_i2s_dmachannels[num]),
			clkchannel->buf_bus + clkchannel->buf_dma_tail,
			CXD9802_REG_BASE64 + CXD9802_I2S_I2SCHD(num),
			dmcntr, dma_size);
		if (clkchannel->nchannels >= 4) {
			cxd9802_i2s_dma_scdesc_add(
				&(cxd9802_i2s_dmachannels[1]),
				clkchannel->buf_bus +
				clkchannel->buf_dma_tail + 4,
				CXD9802_REG_BASE64 + CXD9802_I2S_I2SCHD(1),
				dmcntr, dma_size);
		}
		if (clkchannel->nchannels == 6) {
			cxd9802_i2s_dma_scdesc_add(
				&(cxd9802_i2s_dmachannels[2]),
				clkchannel->buf_bus +
				clkchannel->buf_dma_tail + 8,
				CXD9802_REG_BASE64 + CXD9802_I2S_I2SCHD(2),
				dmcntr, dma_size);
		}
		scdesc_is_set = 1;
		clkchannel->buf_dma_tail += dma_size;
		clkchannel->buf_dma_ndata -= dma_size;
		if (clkchannel->buf_dma_tail == clkchannel->buf_size) {
			clkchannel->buf_dma_tail = 0;
		}
	}
	if (!scdesc_is_set) {
		DPRFR;
		return;
	}

	/* start chain DMA */
	sadinc = clkchannel->nchannels * 2;
	if (!cxd9802_i2s_dmachannels[num].running) {
		cxd9802_i2s_schain_dma_start(num, sadinc, 0);
		if (clkchannel->nchannels >= 4) {
			cxd9802_i2s_schain_dma_start(1, sadinc, 0);
		}
		if (clkchannel->nchannels == 6) {
			cxd9802_i2s_schain_dma_start(2, sadinc, 0);
		}
	}
	cxd9802_i2s_dmachannels[num].running = 1;

	cxd9802_i2s_txrx_start_ch(clkchannel);

	DPRFE;
	return;
}

/* should be called with irqlock */
static void
cxd9802_i2s_rx_dma_start(cxd9802_i2s_clkchannel_t *clkchannel)
{
	int num = clkchannel->num;
	int dma_size;
	int dmcntr, dadinc;
	int scdesc_is_set;
	DPRFS;

	DPRFL(DPL_DEBUG, "num=%d\n", num);

	scdesc_is_set = 0;
	while (!cxd9802_i2s_dmachannel_is_full(clkchannel)) {
		if (clkchannel->buf_dma_ndata == clkchannel->buf_size) {
			break;
		}
		dma_size = clkchannel->buf_head - clkchannel->buf_dma_tail;
		if (dma_size <= 0) {
			dma_size = (clkchannel->buf_size -
				    clkchannel->buf_dma_tail);
		}
		if (dma_size > clkchannel->max_dma_size) {
			dma_size = clkchannel->max_dma_size;
		}
		dma_size = dma_size - dma_size % (clkchannel->nchannels * 2);
		if (dma_size == 0) {
			break;
		}

		dmcntr = dma_size / (clkchannel->nchannels / 2);

		/* set chain DMA descripter */
		cxd9802_i2s_dma_scdesc_add(
			&(cxd9802_i2s_dmachannels[num]),
			CXD9802_REG_BASE64 + CXD9802_I2S_I2SCHD(num),
			clkchannel->buf_bus + clkchannel->buf_dma_tail,
			dmcntr, dma_size);
		if (clkchannel->nchannels >= 4) {
			cxd9802_i2s_dma_scdesc_add(
				&(cxd9802_i2s_dmachannels[1]),
				CXD9802_REG_BASE64 + CXD9802_I2S_I2SCHD(1),
				clkchannel->buf_bus +
				clkchannel->buf_dma_tail + 4,
				dmcntr, dma_size);
		}
		if (clkchannel->nchannels == 6) {
			cxd9802_i2s_dma_scdesc_add(
				&(cxd9802_i2s_dmachannels[2]),
				CXD9802_REG_BASE64 + CXD9802_I2S_I2SCHD(2),
				clkchannel->buf_bus +
				clkchannel->buf_dma_tail + 8,
				dmcntr, dma_size);
		}
		scdesc_is_set = 1;
		clkchannel->buf_dma_tail += dma_size;
		clkchannel->buf_dma_ndata += dma_size;
		if (clkchannel->buf_dma_tail == clkchannel->buf_size) {
			clkchannel->buf_dma_tail = 0;
		}
	}
	if (!scdesc_is_set) {
		DPRFR;
		return;
	}

	/* start chain DMA */
	dadinc = clkchannel->nchannels * 2;
	if (!cxd9802_i2s_dmachannels[num].running) {
		cxd9802_i2s_schain_dma_start(num, 0, dadinc);
		if (clkchannel->nchannels >= 4) {
			cxd9802_i2s_schain_dma_start(1, 0, dadinc);
		}
		if (clkchannel->nchannels == 6) {
			cxd9802_i2s_schain_dma_start(2, 0, dadinc);
		}
	}
	cxd9802_i2s_dmachannels[num].running = 1;

	cxd9802_i2s_txrx_start_ch(clkchannel);

	DPRFE;
	return;
}

/* should be called with irqlock */
static void
cxd9802_i2s_dma_txrx_comp_intr(int dmachannel_num)
{
	cxd9802_i2s_clkchannel_t *clkchannel;
	cxd9802_i2s_dmachannel_t *dmachannel;
	u64 dmchar;
	int new_scdesc_head;
	unsigned long flags;
	DPRFS;

	spin_lock_irqsave(&cxd9802_i2s_irqlock, flags);

	if (cxd9802_i2s_clkchannels[0].nchannels > 2) {
		clkchannel = &(cxd9802_i2s_clkchannels[0]);
	} else {
		clkchannel = &(cxd9802_i2s_clkchannels[dmachannel_num]);
	}
	dmachannel = &(cxd9802_i2s_dmachannels[dmachannel_num]);

	/* calc new head */
	dmchar = cxd9802_reg_read64(CXD9802_DMAC_DMCHAR(1, dmachannel_num));
	if (!cxd9802_i2s_dmachannels[dmachannel_num].running) {
		new_scdesc_head = dmachannel->scdesc_tail;
	} else if (dmchar == 0) {
		new_scdesc_head = dmachannel->scdesc_tail - 1;
	} else {
		new_scdesc_head = ((dmchar - dmachannel->scdesc_bus) /
				   sizeof(cxd9802_i2s_dma_scdesc_t)) - 1;
	}
	if (new_scdesc_head == -1) {
		new_scdesc_head = dmachannel->scdesc_nelem - 1;
	}
	if (new_scdesc_head < 0 ||
	    new_scdesc_head >= dmachannel->scdesc_nelem) {
		DPRFL(DPL_CRIT,
		      "wrong dmchar %016llx\n", dmchar);
		new_scdesc_head = dmachannel->scdesc_tail;
	}

	while (dmachannel->scdesc_head != new_scdesc_head) {
		int dma_size;

		dma_size = dmachannel->dma_sizes[dmachannel->scdesc_head];

		dmachannel->scdesc_head++;
		if (dmachannel->scdesc_head == dmachannel->scdesc_nelem) {
			dmachannel->scdesc_head = 0;
		}
		dmachannel->scdesc_ndata--;
		if (cxd9802_i2s_all_dmachannels_was_completed(clkchannel,
							      dmachannel)) {
			if (clkchannel->state == CXD9802_I2S_STATE_TX) {
				clkchannel->buf_ndata -= dma_size;
				clkchannel->buf_head += dma_size;
				if (clkchannel->buf_head ==
				    clkchannel->buf_size) {
					clkchannel->buf_head = 0;
				}
				wake_up_interruptible(&clkchannel->wait);

				cxd9802_i2s_tx_start(clkchannel);
			} else if (clkchannel->state == CXD9802_I2S_STATE_RX) {
				clkchannel->buf_ndata += dma_size;
				clkchannel->buf_tail += dma_size;
				if (clkchannel->buf_tail ==
				    clkchannel->buf_size) {
					clkchannel->buf_tail = 0;
				}
				wake_up_interruptible(&clkchannel->wait);

				cxd9802_i2s_rx_start(clkchannel);
			}
		}
	}

	spin_unlock_irqrestore(&cxd9802_i2s_irqlock, flags);

	DPRFE;
	return;
}

/* ------------------------ *
 * misc                     *
 * ------------------------ */

static int
cxd9802_i2s_wait_for_idle(cxd9802_i2s_clkchannel_t *clkchannel)
{
	int res;
	DPRFS;

	clkchannel->waiting_for_idle = 1;
	res = wait_event_interruptible(clkchannel->wait,
				       clkchannel->state ==
				       CXD9802_I2S_STATE_IDLE);
	clkchannel->waiting_for_idle = 0;
	if (res < 0) {
		DPRFR;
		return res;
	}

	DPRFE;
	return 0;
}

/* ======================================================================= *
 * DAC/ADC dependent                                                       *
 * ======================================================================= */

/* ------------------------ *
 * MPU-230 (IOP-7)          *
 * ------------------------ */

#include <linux/cxd9790_gpio.h>

#define CXD9802_I2S_MPU230_GPIO_PORT		GPIO_PORT_B
#define CXD9802_I2S_MPU230_NDEMPHON_MASK	(1 << 5)
#define CXD9802_I2S_MPU230_NMUTEOFF_MASK	(1 << 7)
#define CXD9802_I2S_MPU230_FREQ_MASK		(0x3 << 2)
#define CXD9802_I2S_MPU230_FREQ_44100		(0 << 2)
#define CXD9802_I2S_MPU230_FREQ_48100		(1 << 2)
#define CXD9802_I2S_MPU230_FREQ_32000		(2 << 2)
#define CXD9802_I2S_MPU230_FREQ_CPU		(3 << 2)

static int
cxd9802_i2s_dacadc_init(void)
{
	int res;
	DPRFS;

	res = cxd9790_gpio_register(CXD9802_I2S_MPU230_GPIO_PORT,
				    CXD9802_I2S_MPU230_NDEMPHON_MASK |
				    CXD9802_I2S_MPU230_NMUTEOFF_MASK |
				    CXD9802_I2S_MPU230_FREQ_MASK,
				    CXD9802_I2S_DEVNAME);
	if (res < 0) {
		goto err;
	}
	cxd9790_gpio_set_mode(CXD9802_I2S_MPU230_GPIO_PORT,
			      CXD9802_I2S_MPU230_NDEMPHON_MASK |
			      CXD9802_I2S_MPU230_NMUTEOFF_MASK |
			      CXD9802_I2S_MPU230_FREQ_MASK,
			      GPIO_MODE_OUT_HIGH);

	/* set IOP-7 */
	cxd9790_gpio_set_data(CXD9802_I2S_MPU230_GPIO_PORT,
			      CXD9802_I2S_MPU230_FREQ_MASK,
			      CXD9802_I2S_MPU230_FREQ_CPU);
	cxd9790_gpio_set_data(CXD9802_I2S_MPU230_GPIO_PORT,
			      CXD9802_I2S_MPU230_NDEMPHON_MASK,
			      CXD9802_I2S_MPU230_NDEMPHON_MASK);

	/* set I2S module */
	cxd9802_reg_write32(CXD9802_I2S_I2SCCR,
			    CXD9802C_I2S_MODE_2CH_IO |
			    CXD9802M_I2S_MCLK_MODE(1) |
			    CXD9802M_I2S_MASTER_MODE(1) |
			    CXD9802M_I2S_CLOCK_DELAY(1) |
			    CXD9802M_I2S_CLOCK_DELAY(1) |
			    CXD9802C_I2S_DATA_SELECT_16(1) |
			    CXD9802M_I2S_IO_MODE(1) |
			    CXD9802M_I2S_MCLK_MODE(0) |
			    CXD9802M_I2S_MASTER_MODE(0) |
			    CXD9802M_I2S_CLOCK_DELAY(0) |
			    CXD9802M_I2S_CLOCK_DELAY(0) |
			    CXD9802C_I2S_DATA_SELECT_16(0) |
			    CXD9802M_I2S_IO_MODE(0));
	cxd9802_reg_write32(CXD9802_I2S_I2SCOR,
			    CXD9802S_I2S_MCLK_CHECKV(0, 0x1f) |
			    CXD9802S_I2S_TIMESLOT(0, 2) |
			    CXD9802C_I2S_MCLK_DIV_8(0) |
			    CXD9802S_I2S_MCLK_CHECKV(1, 0x1f) |
			    CXD9802S_I2S_TIMESLOT(1, 2) |
			    CXD9802C_I2S_MCLK_DIV_8(1));

	cxd9802_i2s_dac_mute(0, 0);

	DPRFE;
	return 0;

err:
	DPRFR;
	return res;
}

static void
cxd9802_i2s_dacadc_final(void)
{
	DPRFS;

	cxd9802_i2s_dac_mute(0, 1);	/* DAC mute */

	cxd9790_gpio_set_mode(CXD9802_I2S_MPU230_GPIO_PORT,
			      CXD9802_I2S_MPU230_NDEMPHON_MASK |
			      CXD9802_I2S_MPU230_NMUTEOFF_MASK |
			      CXD9802_I2S_MPU230_FREQ_MASK,
			      GPIO_MODE_IN);
	cxd9790_gpio_unregister(CXD9802_I2S_MPU230_GPIO_PORT,
				CXD9802_I2S_MPU230_NDEMPHON_MASK |
				CXD9802_I2S_MPU230_NMUTEOFF_MASK |
				CXD9802_I2S_MPU230_FREQ_MASK);

	DPRFE;
	return;
}

static int
cxd9802_i2s_dacadc_set_mode(int channel, int mode)
{
	DPRFS;

	if (mode == CXD9802_I2S_DACADC_MODE_RX) {
		DPRFR;
		return -ENOTTY;
	}

	DPRFE;
	return 0;
}

static int
cxd9802_i2s_dacadc_set_nchannels(int channel, int nchannels)
{
	DPRFS;

	if (nchannels != 2) {
		DPRFR;
		return -ENOTTY;
	}

	DPRFE;
	return 0;
}

static void
cxd9802_i2s_dac_mute(int channel, int on)
{
	DPRFS;

	if (on) {
		cxd9790_gpio_set_data(CXD9802_I2S_MPU230_GPIO_PORT,
				      CXD9802_I2S_MPU230_NMUTEOFF_MASK,
				      CXD9802_I2S_MPU230_NMUTEOFF_MASK);
	} else {
		cxd9790_gpio_set_data(CXD9802_I2S_MPU230_GPIO_PORT,
				      CXD9802_I2S_MPU230_NMUTEOFF_MASK,
				      0);
	}

	DPRFE;
	return;
}

static int
cxd9802_i2s_dac_set_pcm_level(int channel, int level)
{
	DPRFS;

	DPRFE;
	return -ENOTTY;
}
