/*
 * linux/drivers/usb/device/bi/dbmx1.c
 *
 * DragonBall MX1 USB Device Bus Interface Driver
 *
 * Copyright 2003 Sony Corporation
 *
 * Copyright (c) 2000, 2001, 2002 Lineo
 *
 * By: 
 *      Stuart Lynne <sl@lineo.com>, 
 *      Tom Rushworth <tbr@lineo.com>, 
 *      Bruce Balden <balden@lineo.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/netdevice.h>
#include <linux/delay.h>

#include <asm/atomic.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/system.h>
#include <asm/types.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/arch/gpio.h>

#include "../usbd-export.h"
#include "../usbd-build.h"
#include "../usbd-module.h"
#include "../usbd.h"
#include "../usbd-func.h"
#include "../usbd-bus.h"
#include "../usbd-inline.h"
#include "usbd-bi.h"

#include "dbmx1.h"

MODULE_AUTHOR ("Sony Corporation");
MODULE_DESCRIPTION ("DBMX1 USB Device Bus Interface");
MODULE_LICENSE ("GPL");
USBD_MODULE_INFO ("dbmx1_bi 0.1");

#define DBMX1_USBD_DEBUG
#define DBMX1_USBD_USE_MULTI_INTERFACE
#define DBMX1_USBD_USE_ONE_ALTERNATE

#ifdef DBMX1_USBD_DEBUG

/* debug mask */
#define DBG_USB  0x01
#define DBG_CON  0x02
#define DBG_INIT 0x04
#define DBG_FIFO 0x08
#define DBG_INT  0x10
#define DBG_EP   0x20

static int debug = 0x0;
MODULE_PARM(debug, "i");
MODULE_PARM_DESC(debug, "debug mask");

#define PDEBUG(mask, format, arg...) \
          if (mask & debug) \
            printk(KERN_DEBUG "%s: %s: " format "\n" , __FILE__, __FUNCTION__ , ## arg)
#define DUMP(mask, format, ptr, num, arg...) \
	  { \
	    int _i; \
	     __u8 *p; \
	     if (mask & debug) { \
	       printk(KERN_DEBUG "%s: %s: " format , __FILE__, __FUNCTION__ , ## arg); \
	       p = (__u8*)ptr; \
	       for (_i=0; _i < num; _i++) { \
	         printk(" %02x", *p); \
	         *p++; \
	       } \
	       printk("\n"); \
	     } \
	  }

#else

#define PDEBUG(mask, format, arg...)
#define DUMP(mask, format, ptr, num, arg...)

#endif /* DBMX1_USBD_DEBUG */

#define INFO(format, arg...) \
          printk(KERN_INFO __FILE__ ": " format "\n" , ## arg)

#define MIN(a,b) ((a) < (b) ? (a) : (b))
#define MAX(a,b) ((a) > (b) ? (a) : (b))

typedef struct dbmx1_usbdev {
	__u8               is_connected;         /* connection status */
	__u8               is_suspended;         /* current suspend/resume status */
	__u8               vbus;                 /* current VBUS status */
	struct timer_list  vbus_timer;           /* timer for VBUS event */
	spinlock_t         lock;
} dbmx1_usbdev_t;

static struct dbmx1_usbdev *dev;

static struct usb_device_instance *udc_device;

/*
 * ep_endpoints - map physical endpoints to logical endpoints
 */
static struct usb_endpoint_instance *ep_endpoints[UDC_MAX_ENDPOINTS];

static struct urb *ep0_urb;

extern unsigned int udc_interrupts;

static void dbmx1_vbus_timer(unsigned int oldvbus);
static void dbmx1_epn_xmit (unsigned int ep, struct usb_endpoint_instance *endpoint);

/* ********************************************************************************************* */
/* local functions
 */

static int dbmx1_init(void)
{
	/* allocate memory */
	dev = NULL;
	dev = kmalloc(sizeof(dbmx1_usbdev_t), GFP_KERNEL);
	if (!dev) {
		return -1;
	}

	/* initialize status */
	dev->is_connected = 0;
	dev->is_suspended = (GET_USB_STAT() & USBD_STAT_SUSP) ? 1 : 0;
	dev->vbus = IS_VBUS_ACTIVE(0);
	
	/* initialize spin-lock */
	spin_lock_init(&dev->lock);
	
	/* initialize timer */
	init_timer(&dev->vbus_timer);
	dev->vbus_timer.function = (void*)dbmx1_vbus_timer;

	return 0;
}

static void dbmx1_release(void)
{
	/* free memory */
	if (dev) {
		kfree(dev);
	}
}

static int dbmx1_setup_config(void)
{
	int i, cfg, intf, alt ,ep, cnt = 0;
	__u8 data[5];
	__u32 tmp;
	
	struct usb_device_descriptor *dev_desc;
	struct usb_configuration_descriptor *cfg_desc;
	struct usb_interface_instance *iface_inst;
	struct usb_interface_descriptor *iface_desc;
        struct usb_alternate_instance *alt_inst;
	struct usb_endpoint_descriptor *ep_desc;

	dev_desc = usbd_device_device_descriptor (udc_device, 0);
	if (!dev_desc) {
		PDEBUG(DBG_INIT, "cannot get device descriptor");
		return -1;
	}
	
	/* write configuration data */
	for (i = 0; i < 5; i++) {
		if (i == 2) {
			data[i] = EP0_PACKETSIZE;
		} else { 
			data[i] = 0x0;
		}
		WRITE_REG32(USBD_DDAT, (__u32)data[i]);
		cnt++;
		/* wait until busy bit is cleared */
		while (READ_REG32(USBD_DADR) & USBD_DADR_BSY);
	}
	DUMP(DBG_INIT, "[EP0]", data, 5);

	/* search endpoint descriptor in each setting */
	for (cfg = 0; cfg < dev_desc->bNumConfigurations; cfg++) {
		cfg_desc = usbd_device_configuration_descriptor (udc_device, 0, cfg);
		if (!cfg_desc)
			break;

		for (intf = 0; intf < cfg_desc->bNumInterfaces; intf++) {
			iface_inst = usbd_device_interface_instance (udc_device, 0, cfg, intf);
			if (!iface_inst)
				break;

			for (alt = 0; alt < iface_inst->alternates; alt++) {
				alt_inst = usbd_device_alternate_instance (udc_device, 0, cfg, intf, alt);
				if (!alt_inst)
					break;
				iface_desc = alt_inst->interface_descriptor;
				if (!iface_inst)
					break;

				for (ep = 0; ep < alt_inst->endpoints; ep++) {
					int epnum;
					ep_desc = usbd_device_endpoint_descriptor_index (udc_device, 0, cfg, intf, alt, ep);
					if (!ep_desc)
						break;
					epnum = ep_desc->bEndpointAddress & USB_ENDPOINT_NUMBER_MASK;

					/* make configuration data */
					data[0] = ((epnum << 4) | ((cfg + 1) << 2) | (intf << 0)) & 0xff;
#ifdef DBMX1_USBD_USE_ONE_ALTERNATE
					tmp  = ((0x0 & 0x7) << 29);  /* we use only one alternate setting */
#else
					tmp  = ((alt & 0x7) << 29);  /* we use multiple alternate settings */
#endif
					tmp |= (ep_desc->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) << 27;
					tmp |= ((ep_desc->bEndpointAddress & USB_ENDPOINT_DIR_MASK) >> 7) << 26;
					tmp |= 0x3 << 14;
					tmp |= (ep_desc->wMaxPacketSize & 0xff) << 16;
					tmp |= epnum & 0xff;
					data[1] = (tmp >> 24) & 0xff;
					data[2] = (tmp >> 16) & 0xff;
					data[3] = (tmp >> 8) & 0xff;
					data[4] = tmp & 0xff;

					/* write configuration data */
					for (i = 0; i < 5; i++) {
						WRITE_REG32(USBD_DDAT, data[i]);
						
						/* wait until busy bit is cleared */
						while (READ_REG32(USBD_DADR) & USBD_DADR_BSY);
						cnt++;
						if (cnt > 55) {
							INFO("error - too many configurations");
							return -1;
						}
					}
					DUMP(DBG_INIT, "[EP%d] CFG:%d IF:%d ALT:%d :", data , 5, epnum, cfg, intf, alt);
				}
			}
		}
	}

	/* if size of configuration data is less than 55 bytes, zero is padded */
	while (cnt < 55) {
		WRITE_REG32(USBD_DDAT, 0x00000000);
		while (READ_REG32(USBD_DADR) & USBD_DADR_BSY);
		cnt++;
	}

	return 0; 
}

static void dbmx1_suspend(void)
{
	struct usb_endpoint_instance *endpoint = ep_endpoints[EP0];

	PDEBUG(DBG_CON, "called");

	if (!dev->is_suspended) {
		dev->is_suspended = 1;
		endpoint->state = EP0_WAIT_SETUP;
		endpoint->tx_urb = NULL;
		endpoint->rcv_urb = NULL;
		usbd_device_event (udc_device, DEVICE_BUS_INACTIVE, 0);
	}
}

static void dbmx1_resume(void)
{
	PDEBUG(DBG_CON, "called");

	if (dev->is_suspended) {
		dev->is_suspended = 0;
		usbd_device_event (udc_device, DEVICE_BUS_ACTIVITY, 0);
	}
}

static int dbmx1_chk_fifo_err(unsigned int ep)
{
	__u32 status;
  
	status = GET_FIFO_STAT(ep);

	if (status & USBD_EP_FSTAT_ERROR) {
		INFO("error - [EP%d] FIFO Error!", ep);
	} else if (status & USBD_EP_FSTAT_OF) {
		INFO("error - [EP%d] FIFO Overflow!", ep);
	} else if (status & USBD_EP_FSTAT_UF) {
		INFO("error - [EP%d] FIFO Underflow!", ep);
	} else {
		return 0; /* no error */
	}

	/* clear error status */
	CLEAR_FIFO_STAT(ep, status);
	udc_reset_ep(ep);
	
	return -1;
}


/* ********************************************************************************************* */
/* IO
 */

static void dbmx1_enable_pullup(void)
{
	PDEBUG(DBG_CON, "called");
	
	/* deassert USBD_SUSPEND */
	dragonball_gpio_set_bit(PORT_B, 23, 0);

	/* enable pull up D+ line */
	UNSETBIT_REG32(MPU110_GPIO2, MPU110_PWR_USB0);
}

static void dbmx1_disable_pullup(void)
{
	PDEBUG(DBG_CON, "called");

	/* disable pull up D+ line */
	SETBIT_REG32(MPU110_GPIO2, MPU110_PWR_USB0);

	/* assert USBD_SUSPEND */
	dragonball_gpio_set_bit(PORT_B, 23, 1);
}

static __inline__ void dbmx1_read_fifo(unsigned int ep, __u8 *buf, int size)
{
	PDEBUG(DBG_FIFO, "[EP%d] read %d bytes", ep, size);

	/* read data from FIFO */
	while (size-- > 0) {
		*buf++ = READ_FIFO8(ep);
	}
}

static __inline__ void dbmx1_write_fifo(unsigned int ep, __u8 *buf, int size)
{
	PDEBUG(DBG_FIFO, "[EP%d] wrote %d bytes", ep, size);

	/* write (size-1) bytes data to FIFO */
	while (size-- > 1) {
		WRITE_FIFO8(ep, *buf++);
	}

	/* set WFR bit */
	SET_FIFO_WFR(ep);

	/* write last data to FIFO */
	WRITE_FIFO8(ep, *buf);
}


/* ********************************************************************************************* */
/* Control (endpoint zero)
 */

void dbmx1_enable_ep0_xmit(void) 
{
	PDEBUG(DBG_EP, "[EP0]");

	/* start transmitting */
	ENABLE_FIFO_LOW_INTR(EP0);
}

void dbmx1_disable_ep0_xmit(void)
{
	PDEBUG(DBG_EP, "[EP0]");

	/* stop transmitting */
	DISABLE_FIFO_LOW_INTR(EP0);
}

static void dbmx1_ep0_recv(void)
{
	struct usb_endpoint_instance *endpoint = ep_endpoints[EP0];
	struct urb *urb = endpoint->rcv_urb;
	int rcvcnt = 0;
	
	rcvcnt = GET_FIFO_CNT(EP0);

	PDEBUG(DBG_INT, "called");

	/* no data */
	if (rcvcnt <= 0) {
		return;
	}

	/* read packet from FIFO */
	PDEBUG(DBG_USB, "[EP0] read %d bytes", rcvcnt);
	dbmx1_read_fifo(EP0, (__u8*)(urb->buffer + urb->actual_length), rcvcnt);
	
	urb->actual_length += rcvcnt;

	if (urb->actual_length == urb->device_request.wLength) {

		/* process setup packet */
		usbd_recv_setup (ep0_urb);

		/* free urb buffer */
		if (urb->buffer) {
			kfree(urb->buffer);
		}
		endpoint->state = EP0_WAIT_SETUP;
	}
}

static void dbmx1_ep0_xmit(void)
{
	struct usb_endpoint_instance *endpoint = ep_endpoints[EP0];
	struct urb *urb = endpoint->tx_urb;

	PDEBUG(DBG_INT, "called");

	/* How much are we sending this time? (May be zero!)
	   (Note that later call of tx_complete() will add last to sent.) */
	if (endpoint->tx_urb == NULL) {
		endpoint->last = 0;
	} else {
		endpoint->last = MIN(urb->actual_length - endpoint->sent, endpoint->tx_packetSize);
	}

	if (endpoint->last > 0 && urb->buffer) {
		/* write data to FIFO */
		PDEBUG(DBG_USB, "[EP0] wrote %d bytes", endpoint->last);
		dbmx1_write_fifo(EP0, (__u8*)(urb->buffer + endpoint->sent), endpoint->last);
        }

        /* Is this the end of the data state? (We've sent all the data, plus any required ZLP.) */
        if (!endpoint->tx_urb || (endpoint->last < endpoint->tx_packetSize)) {

		/* stop transmitting */
		dbmx1_disable_ep0_xmit();
                endpoint->state = EP0_WAIT_SETUP;

	} else if ((endpoint->last == endpoint->tx_packetSize) && 
		   ((endpoint->last + endpoint->sent) == ep0_urb->actual_length)  &&
		   (ep0_urb->actual_length < ep0_urb->device_request.wLength)) {

		/* send zero length packet */
		PDEBUG(DBG_USB, "[EP0] send zero length packet");
		endpoint->last = 0;
		SET_ZLPS(EP0);

		/* stop transmitting */
		dbmx1_disable_ep0_xmit();
		endpoint->state = EP0_WAIT_SETUP;
	}
	
	usbd_tx_complete_irq (endpoint, 0);
}

static int dbmx1_ep0_setup(void)
{
	struct usb_endpoint_instance *endpoint = ep_endpoints[EP0];
	int rcvcnt;
	
	rcvcnt = GET_FIFO_CNT(EP0);

	/* no data */
	if (rcvcnt <= 0) {
		return 0;
	}

	/* read device request from FIFO */
	dbmx1_read_fifo(EP0, (__u8*)&ep0_urb->device_request, 8);
	DUMP(DBG_USB, "SETUP:", (__u8*)&ep0_urb->device_request, 8);

	/* process setup packet */
	if (usbd_recv_setup (ep0_urb)) {
		if (ep0_urb->device_request.wLength) {
			PDEBUG(DBG_USB, "[EP0] SETUP failed");
			return -1;
		}
		PDEBUG(DBG_USB, "[EP0] SETUP ignored");
		return 0;
	}

	/* check direction */
	if ((ep0_urb->device_request.bmRequestType & USB_REQ_DIRECTION_MASK) == USB_REQ_HOST2DEVICE) {

		/* Control Write - we are receiving more data from the host */
		if (ep0_urb->device_request.wLength) {
			PDEBUG(DBG_USB, "[EP0] HOST2DEVICE: %d bytes", ep0_urb->device_request.wLength);
			endpoint->rcv_urb = ep0_urb;
			endpoint->rcv_urb->actual_length = 0;
			endpoint->state = EP0_DATA_STATE_RECV;
			return 0;
		}
		/* no data transfer */
		PDEBUG(DBG_USB, "[EP0] NODATA");

	} else {

		/* Control Read - we are transmitting data to the host */
		PDEBUG(DBG_USB, "[EP0] DEVICE2HOST: %d bytes", ep0_urb->device_request.wLength);
		
		/* verify that we have non-zero request length */
		if (!ep0_urb->device_request.wLength) {
			PDEBUG(DBG_USB, "[EP0] setup stalling zero wLength");
			return -1;
		}
		
		/* verify that we have non-zero length response */
		if (!ep0_urb->actual_length) {
			PDEBUG(DBG_USB, "[EP0] setup stalling zero response");
			return -1;
		}

		/* verify that we have too long response */
		if (ep0_urb->device_request.wLength < ep0_urb->actual_length) {
			PDEBUG(DBG_USB, "[EP0] setup stalling too long response: %d", ep0_urb->actual_length);
			return -1;
		}

		/* start transmitting */
		endpoint->tx_urb = ep0_urb;
		endpoint->sent = 0;
		endpoint->last = 0;
		endpoint->state = EP0_DATA_STATE_XMIT;
		dbmx1_enable_ep0_xmit();
	}
	return 0;
}


/* ********************************************************************************************* */
/* Bulk OUT (recv)
 */

static void dbmx1_enable_epn_recv(unsigned int ep) 
{
	PDEBUG(DBG_EP, "[EP%d]", ep);

	/* start receiving */
	ENABLE_EOF_EOT_INTR(ep);
}

static void dbmx1_epn_recv(unsigned int ep, struct usb_endpoint_instance *endpoint)
{
	int rcvcnt = GET_FIFO_CNT(ep);

	PDEBUG(DBG_EP, "[EP%d] receive %d bytes", ep, rcvcnt);

	if (endpoint) {
		struct urb *urb = endpoint->rcv_urb;
		if (!urb) {
			urb = first_urb_detached (&endpoint->rdy);
		}
		if (urb) {
			
			/* read packet from FIFO */
			PDEBUG(DBG_USB, "[EP%d] read %d bytes", ep, rcvcnt);			
			dbmx1_read_fifo(ep, urb->buffer + urb->actual_length, rcvcnt);
			
			/* let the upper levels handle the urb */
			usbd_rcv_complete_irq (endpoint, rcvcnt, 0);

		} else {
			
			INFO("error - [EP%d] cannot get urb", ep);
			udc_reset_ep (ep);
			
		}
	}
	else {
		
		INFO("error - [EP%d] endpint is not allocated", ep);
		udc_reset_ep (ep);
		
	}
}


/* ********************************************************************************************* */
/* Bulk IN (tx)
 */

static void dbmx1_enable_epn_xmit(struct usb_endpoint_instance *endpoint)
{
	unsigned int ep = endpoint->endpoint_address & 0x7f;
	__u32 status;
	
	PDEBUG(DBG_EP, "[EP%d]", ep);

	if (!test_bit(EP_IN_PROGRESS, & endpoint->state)) {
		status = GET_EP_INTR(ep);
		if (status & USBD_EP_INTR_EOF) {
			CLEAR_EP_INTR(ep, (status & USBD_EP_INTR_EOF));
		}
		dbmx1_epn_xmit (ep, endpoint);
	}
	
	/* start transmitting */
	ENABLE_EOF_INTR(ep);
}

static void dbmx1_disable_epn_xmit(struct usb_endpoint_instance *endpoint)
{
	unsigned int ep = endpoint->endpoint_address & 0x7f;

	PDEBUG(DBG_EP, "[EP%d]", ep);

	/* start receiving */
	DISABLE_EOF_INTR(ep);
	clear_bit(EP_IN_PROGRESS, &endpoint->state);
}

static void dbmx1_epn_xmit (unsigned int ep, struct usb_endpoint_instance *endpoint)
{
	struct urb *urb = endpoint->tx_urb;

	PDEBUG(DBG_EP, "[EP%d] called", ep);

	set_bit(EP_IN_PROGRESS, &endpoint->state);
	
	if (urb) {
		if ((urb->actual_length - endpoint->sent) > 0) {
			
			/* write packet to FIFO */
			endpoint->last = MIN (urb->actual_length - endpoint->sent, endpoint->tx_packetSize);
			PDEBUG(DBG_USB, "[EP%d] wrote %d bytes", ep, endpoint->last);
			dbmx1_write_fifo(ep, urb->buffer + endpoint->sent, endpoint->last);

		} else {

			/* send zero length packet */
			PDEBUG(DBG_USB, "send zero length packet");
			SET_ZLPS(ep);
			endpoint->last = 0;
		}
	} else {
		/* stop transmitting packet */
		dbmx1_disable_epn_xmit(endpoint);
	}
	
	usbd_tx_complete_irq(endpoint, 0);	
}


/* ********************************************************************************************* */
/* Interrupt Handler
 */

/**
 * int_hndlr - interrupt handler
 *
 */
static void dbmx1_ep0_intr(int irq, void *dev_id, struct pt_regs *regs)
{
	struct usb_endpoint_instance *endpoint = ep_endpoints[EP0];
	__u32 status = GET_EP_INTR(EP0);
#ifdef DBMX1_USBD_DEBUG
	__u32 mask = GET_EP_INTR_MASK(EP0);
	PDEBUG(DBG_INT, "status/mask = 0x%03x/0x%03x", (status & 0x1ff), (mask & 0x1ff));
#endif
	udc_interrupts++;
	
	/* clear EOF interrupt before reading FIFO */
	CLEAR_EP_INTR(EP0, (status & USBD_EP_INTR_EOF));

	if (status & USBD_EP_INTR_FIFO_ERROR) {  /* FIFO error */

		dbmx1_chk_fifo_err(EP0);
		
	} else if (status & USBD_EP_INTR_MDEVREQ) {  /* multiple device request */

		INFO("error - receive multiple device request!");
		udc_stall_ep (EP0);
		udc_reset_ep (EP0);

	} else if (status & USBD_EP_INTR_DEVREQ) {  /* device request */
		
		PDEBUG(DBG_USB, "receive SETUP");

		endpoint->state = EP0_WAIT_SETUP;
		if (dbmx1_ep0_setup() < 0) {
			udc_stall_ep(EP0);
			udc_reset_ep (EP0);
		}
		SET_CMD_OVER();

	} else if (endpoint->state == EP0_DATA_STATE_RECV) {  /* control write transfer */

		dbmx1_ep0_recv();
    
	} else if (endpoint->state == EP0_DATA_STATE_XMIT) {  /* control read transfer */

		dbmx1_ep0_xmit();

	}

	/* clear interrupt except for EOF */
	CLEAR_EP_INTR(EP0, (status & ~USBD_EP_INTR_EOF));
}

static void dbmx1_epn_intr(int irq, void *dev_id, struct pt_regs *regs)
{	
	struct usb_endpoint_instance *endpoint;
	unsigned int ep;
	__u32 status;
#ifdef DBMX1_USBD_DEBUG
	__u32 mask;
#endif
	udc_interrupts++;
	
	ep = irq - IRQ_USBD_INT_0;
	endpoint = ep_endpoints[ep];
	status = GET_EP_INTR(ep);

#ifdef DBMX1_USBD_DEBUG
	mask = GET_EP_INTR_MASK(ep);
	PDEBUG(DBG_INT, "[EP%d] status/mask = 0x%03x/0x%03x", ep, (status & 0x1ff), (mask & 0x1ff));
#endif

	/* clear EOF interrupt before reading FIFO */
	CLEAR_EP_INTR(ep, (status & USBD_EP_INTR_EOF));

	/* check FIFO error */
	if (status & USBD_EP_INTR_FIFO_ERROR) {  /* FIFO error */
		dbmx1_chk_fifo_err(ep);
	}

	if (endpoint->endpoint_address & USB_DIR_IN) { /* IN */
		
		dbmx1_epn_xmit(ep, endpoint);
		
	} else if (endpoint->endpoint_address) { /* OUT */

		if (status & USBD_EP_INTR_EOF) {
			dbmx1_epn_recv(ep, endpoint);
		}
		
	}
  
	/* clear interrupt except for EOF */
	CLEAR_EP_INTR(ep, (status & ~USBD_EP_INTR_EOF));
}

static void dbmx1_udc_intr(int irq, void *dev_id, struct pt_regs *regs)
{
	__u32 status = GET_USB_INTR() & ~GET_USB_INTR_MASK();
#ifdef DBMX1_USBD_DEBUG
	__u32 usb_status = GET_USB_STAT();
	PDEBUG(DBG_INT, "usbintr/usbstat = 0x%08x/0x%08x", status, usb_status);
#endif

	udc_interrupts++;

	/* config change interrupt */
	if (status & USBD_INTR_CFG_CHG) {
 
		/* get configuration */
		udc_device->configuration = GET_CFG() - 1;
		udc_device->interface = GET_INTF();
		udc_device->alternate = GET_ALTSET();
		
		PDEBUG(DBG_USB, "config changed: CFG:%d INTF:%d ALT:%d",
		       udc_device->configuration,
		       udc_device->interface,
		       udc_device->alternate);

		if (udc_device->device_state != STATE_CONFIGURED) {
			usbd_device_event (udc_device, DEVICE_ADDRESS_ASSIGNED, 0);
			usbd_device_event (udc_device, DEVICE_CONFIGURED, 0);
		} else {
			usbd_device_event (udc_device, DEVICE_SET_INTERFACE, 0);
		}
	}
  
	/* active to suspend interrupt */
	if (status & USBD_INTR_SUSP) {
		PDEBUG(DBG_CON, "USB Active to Suspend");
		dbmx1_suspend();
	}

	/* suspend to resume interrupt */
	if (status & USBD_INTR_RES) {
		PDEBUG(DBG_CON, "USB Suspend to Resume");
		dbmx1_resume();
	}

	/* reset start interrupt */
	if (status & USBD_INTR_RESET_START) {
		PDEBUG(DBG_CON, "USB Reset Start signal");
		usbd_device_event (udc_device, DEVICE_RESET, 0);
	}

	/* reset stop interrupt */
	if (status & USBD_INTR_RESET_STOP) {
		PDEBUG(DBG_CON, "USB Reset Stop signal");
	}

	/* wakeup interrupt*/
	if (status & USBD_INTR_WAKEUP) {
		PDEBUG(DBG_CON, "USB Wakeup signal");
	}
	
	/* clear interrupt */
	CLEAR_USB_INTR(status);
}

/* ********************************************************************************************* */


/* ********************************************************************************************* */
/*
 * Start of public functions.
 */

/**
 * udc_start_in_irq - start transmit
 * @eendpoint: endpoint instance
 *
 * Called by bus interface driver to see if we need to start a data transmission.
 */
void udc_start_in_irq (struct usb_endpoint_instance *endpoint)
{
	unsigned long flags;
#ifdef DBMX1_USBD_DEBUG	
	unsigned int ep = endpoint->endpoint_address & 0x7f;
	PDEBUG(DBG_EP, "[EP%d] called", ep);
#endif
	
	spin_lock_irqsave(&dev->lock, flags);
	dbmx1_enable_epn_xmit(endpoint);
	spin_unlock_irqrestore(&dev->lock, flags);
}

/**
 * udc_init - initialize
 *
 * Return non-zero if we cannot see device.
 **/
int udc_init (void)
{
	unsigned long flags;

	PDEBUG(DBG_INIT, "called");

	spin_lock_irqsave(&dev->lock, flags);

	/* set USB divider for 48 MHz USB clock */
	SET_PLL_USBDIV(0x1);
	udelay(100);

	spin_unlock_irqrestore(&dev->lock, flags);

	return 0;
}


/**
 * udc_start_in - start transmit
 * @eendpoint: endpoint instance
 *
 * Called by bus interface driver to see if we need to start a data transmission.
 */
void udc_start_in (struct usb_endpoint_instance *endpoint)
{
#ifdef DBMX1_USBD_DEBUG	
	unsigned int ep = endpoint->endpoint_address & 0x7f;
#endif

	PDEBUG(DBG_EP, "[EP%d] called", ep);

	if (endpoint) {
		udc_start_in_irq(endpoint);
	}
}


/**
 * udc_stall_ep - stall endpoint
 * @ep: physical endpoint
 *
 * Stall the endpoint.
 */
void udc_stall_ep (unsigned int ep)
{
	if (ep < UDC_MAX_ENDPOINTS) {
		PDEBUG(DBG_FIFO, "stall EP%d", ep);
		SET_EP_STALL(ep);
	}
}


/**
 * udc_reset_ep - reset endpoint
 * @ep: physical endpoint
 * reset the endpoint.
 *
 * returns : 0 if ok, -1 otherwise
 */
void udc_reset_ep (unsigned int ep)
{
	struct usb_endpoint_instance *endpoint = ep_endpoints[ep];

	if (ep < UDC_MAX_ENDPOINTS) {
		PDEBUG(DBG_FIFO, "reset EP%d", ep);
		CLEAR_FIFO(ep);
		clear_bit(EP_IN_PROGRESS, &endpoint->state);
	}
}


/**
 * udc_endpoint_halted - is endpoint halted
 * @ep:
 *
 * Return non-zero if endpoint is halted
 */
int udc_endpoint_halted (unsigned int ep)
{
	return 0;
}


/**
 * udc_set_address - set the USB address for this device
 * @address:
 *
 * Called from control endpoint function after it decodes a set address setup packet.
 */
void udc_set_address (unsigned char address)
{
	/* unused */
}

/**
 * udc_serial_init - set a serial number if available
 */
int __init udc_serial_init (struct usb_bus_instance *bus)
{
	PDEBUG(DBG_INIT, "called");
	
	return -EINVAL;
}

/* ********************************************************************************************* */

/**
 * udc_max_endpoints - max physical endpoints 
 *
 * Return number of physical endpoints.
 */
int udc_max_endpoints (void)
{
	return UDC_MAX_ENDPOINTS;
}


/**
 * udc_check_ep - check logical endpoint 
 * @lep:
 *
 * Return physical endpoint number to use for this logical endpoint or zero if not valid.
 */
int udc_check_ep (int logical_endpoint, int packetsize)
{
	PDEBUG(DBG_INIT, "[EP%d]", logical_endpoint);
	
	return (((logical_endpoint & 0xf) >= UDC_MAX_ENDPOINTS) || (packetsize > 64)) ?
		0 : (logical_endpoint & 0xf);
}


/**
 * udc_set_ep - setup endpoint 
 * @ep:
 * @endpoint:
 *
 * Associate a physical endpoint with endpoint_instance
 */
void udc_setup_ep (struct usb_device_instance *device, unsigned int ep,
		   struct usb_endpoint_instance *endpoint)
{
        __u32 tmp = USBD_EP_STAT_FLUSH; /* clear fifo */
	int type = 0;
	int max = 0;

	if (ep < UDC_MAX_ENDPOINTS && endpoint) {

		ep_endpoints[ep] = endpoint;

		/* EP0 */
		if (ep == 0) {
			endpoint->state = EP0_WAIT_SETUP;
			return;
		}
		/* IN */
		else if (endpoint->endpoint_address & USB_DIR_IN) {
			type = endpoint->tx_attributes & USB_ENDPOINT_XFERTYPE_MASK;
			while ((8 << max < endpoint->tx_packetSize) && max <= 3)
				max++;
			tmp |= USBD_EP_STAT_DIR; /* IN */
			tmp |= type << USBD_EP_STAT_TYP_BIT;
			tmp |= max << USBD_EP_STAT_MAX_BIT;
		}
		/* OUT */
		else if (endpoint->endpoint_address) {
			type = endpoint->rcv_attributes & USB_ENDPOINT_XFERTYPE_MASK;
			while ((8 << max < endpoint->rcv_packetSize) && max <= 3)
				max++;
			tmp |= type << USBD_EP_STAT_TYP_BIT;
			tmp |= max << USBD_EP_STAT_MAX_BIT;
			if (!endpoint->rcv_urb) {
				usbd_fill_rcv (device, endpoint, 5);
				endpoint->rcv_urb = first_urb_detached (&endpoint->rdy);
			}
			dbmx1_enable_epn_recv(ep);
		}
		/* error */
		else {
			return;
		}
  
		PDEBUG(DBG_INIT, "[EP%d] setup: EP%d_STAT=0x%02x", ep, ep, tmp);

		/* write endpoint settings */
		WRITE_REG32(USBD_EP_STAT(ep), tmp);
  
		/* set GR/ALRM value */
		SET_FIFO_GR(ep, 0x7);
		SET_FIFO_ALRM(ep, 0x0);
		
		/* set frame mode */
		SET_FRAME_MODE(ep);

		/* enable FIFO error interrupt */
		ENABLE_FIFO_ERROR_INTR(ep);
	}
}


/**
 * udc_disable_ep - disable endpoint
 * @ep:
 *
 * Disable specified endpoint 
 */
void udc_disable_ep (unsigned int ep)
{
	if (ep < UDC_MAX_ENDPOINTS) {
		struct usb_endpoint_instance *endpoint;

		if ((endpoint = ep_endpoints[ep])) {
			ep_endpoints[ep] = NULL;
			usbd_flush_ep (endpoint);
		}
	}
}

/* ********************************************************************************************* */

/**
 * udc_connected - is the USB cable connected
 *
 * Return non-zeron if cable is connected.
 */
int udc_connected ()
{
	return dev->is_connected;
}


/**
 * udc_connect - enable pullup resistor
 *
 * Turn on the USB connection by enabling the pullup resistor.
 */
void udc_connect (void)
{
	PDEBUG(DBG_CON, "called");

	/* enable pull up D+ */
	dbmx1_enable_pullup();
}


/**
 * udc_disconnect - disable pullup resistor
 *
 * Turn off the USB connection by disabling the pullup resistor.
 */
void udc_disconnect (void)
{
	PDEBUG(DBG_CON, "called");

	/* disable pull up D+ */
	dbmx1_disable_pullup();
}

/**
 * udc_int_hndlr_cable - interrupt handler for cable
 */
static void udc_int_hndlr_cable (int irq, void *dev_id, struct pt_regs *regs)
{
	int vbus;

	/* check interrupt status */
	if (dragonball_gpio_intr_status_bit(USBD_VBUS_GPIO_PORT, USBD_VBUS_GPIO_PIN)) {
		dragonball_gpio_clear_intr(USBD_VBUS_GPIO_PORT, USBD_VBUS_GPIO_PIN);
	} else {
		return;
	}

	PDEBUG(DBG_CON, "called");

	/* get current vbus status */
	vbus = dragonball_gpio_get_bit(USBD_VBUS_GPIO_PORT, USBD_VBUS_GPIO_PIN);

	/* delete timer */
	del_timer(&dev->vbus_timer);

	/* set timer to check VBUS line again */
	dev->vbus_timer.data = vbus;
	dev->vbus_timer.expires = jiffies + VBUS_TIMER_EXPIRE;
	add_timer(&dev->vbus_timer);
}

static void dbmx1_vbus_timer(unsigned int oldvbus)
{
	unsigned long flags;
	unsigned int vbus;

	PDEBUG(DBG_CON, "called");

	spin_lock_irqsave(&dev->lock, flags);

	/* get current vbus status */
	vbus = dragonball_gpio_get_bit(USBD_VBUS_GPIO_PORT, USBD_VBUS_GPIO_PIN);

	if (vbus == oldvbus) {

		/* store status to global variable */
		dev->vbus = vbus;
		
		if (IS_VBUS_ACTIVE(vbus)) {
			dev->is_connected = 1;
		} else {
			if (dev->is_connected) {
				dev->is_connected = 0;
				dbmx1_suspend();
			}
		}

		/* set next interrupt trigger mode */
		dragonball_gpio_config_intr(USBD_VBUS_GPIO_PORT, USBD_VBUS_GPIO_PIN, 
					    (vbus ? ICR_NEGATIVE_EDGE : ICR_POSITIVE_EDGE));

	} else {
    
		/* set timer again */
		dev->vbus_timer.data = vbus;
		dev->vbus_timer.expires = jiffies + VBUS_TIMER_EXPIRE;
		add_timer(&dev->vbus_timer);
		
	}
  
	spin_unlock_irqrestore(&dev->lock, flags);
}

static void dbmx1_enable_monitor_vbus(int only_insert)
{
	int vbus;
	unsigned long flags;

	PDEBUG(DBG_CON, "called");

	spin_lock_irqsave(&dev->lock, flags);

	/* set interrupt trigger mode */
	if (only_insert) {
		vbus = IS_VBUS_ACTIVE(0);
	} else {
		vbus = dragonball_gpio_get_bit(USBD_VBUS_GPIO_PORT, USBD_VBUS_GPIO_PIN);
	}
	dragonball_gpio_config_intr(USBD_VBUS_GPIO_PORT, USBD_VBUS_GPIO_PIN, 
				    (vbus ? ICR_NEGATIVE_EDGE : ICR_POSITIVE_EDGE));

	/* delete timer */
	del_timer(&dev->vbus_timer);

	/* initialize timer to monitor VBUS line */
	if (!only_insert) {
		dev->vbus_timer.data = vbus;
		dev->vbus_timer.expires = jiffies + VBUS_TIMER_EXPIRE;
		add_timer(&dev->vbus_timer);
	}

	/* enable VBUS interrupt */
	dragonball_gpio_unmask_intr(USBD_VBUS_GPIO_PORT, USBD_VBUS_GPIO_PIN);

	spin_unlock_irqrestore(&dev->lock, flags);
}

static void dbmx1_disable_monitor_vbus(void)
{
	PDEBUG(DBG_CON, "called");

	/* delete timer */
	del_timer(&dev->vbus_timer);

	/* disable VBUS interrupt */
	dragonball_gpio_mask_intr(USBD_VBUS_GPIO_PORT, USBD_VBUS_GPIO_PIN);
}

/* ********************************************************************************************* */

/**
 * udc_enable_interrupts - enable interrupts
 *
 * Switch on UDC interrupts.
 *
 */
void udc_all_interrupts (struct usb_device_instance *device)
{
	/* unused */
}


/**
 * udc_suspended_interrupts - enable suspended interrupts
 *
 * Switch on only UDC resume interrupt.
 *
 */
void udc_suspended_interrupts (struct usb_device_instance *device)
{
	/* unused */
}


/**
 * udc_disable_interrupts - disable interrupts.
 *
 * switch off interrupts
 */
void udc_disable_interrupts (struct usb_device_instance *device)
{
	int i;

	PDEBUG(DBG_INIT, "called");
	
	/* clear and mask all interrupts */
	CLEAR_USB_INTR(0x800000ff);
	DISABLE_USB_INTR(0x800000ff);
	for (i=0; i < UDC_MAX_ENDPOINTS; i++) {
		CLEAR_EP_INTR(i, 0x000001ff);
		DISABLE_EP_INTR(i, 0x000001ff);
	}
}

/* ********************************************************************************************* */

/**
 * udc_ep0_packetsize - return ep0 packetsize
 */
int udc_ep0_packetsize (void)
{
	return EP0_PACKETSIZE;
}

/**
 * udc_enable - enable the UDC
 *
 * Switch on the UDC
 */
void udc_enable (struct usb_device_instance *device)
{
	unsigned long flags;
	int i, j, max = 0;
	__u32 tmp;
	
	PDEBUG(DBG_INIT, "device: %p", device);

	/* save the device structure pointer */
	udc_device = device;

	/* allocate ep0 urb */
	if (!ep0_urb) {
		if (!(ep0_urb = usbd_alloc_urb (device, device->function_instance_array, 0, 512))) {
			INFO("error - usbd_alloc_urb failed");
			return;
		}
	}

	spin_lock_irqsave(&dev->lock, flags);

	/* reset and enable USBD */
	DISABLE_USB();
	udelay(100);
	RESET_USB();

	/* wait until USBD is enabled */
	while (!IS_USB_ENABLE());

	if (!(READ_REG32(USBD_DADR) & USBD_DADR_CFG)) {
		INFO("error - UDC is not fully reset!");
		return;
	}
	
	/* write configuration data */
	if (dbmx1_setup_config()) {
		INFO("warning - UDC is not filly initialized");
	}

	/* clear and mask all interrupts */
	CLEAR_USB_INTR(0x800000ff);
	DISABLE_USB_INTR(0x800000ff);
	for (i=0; i < UDC_MAX_ENDPOINTS; i++) {
		CLEAR_EP_INTR(i, 0x000001ff);
		DISABLE_EP_INTR(i, 0x000001ff);
	}

	/* wait until reset signal finished */
	while (GET_USB_INTR() & USBD_MASK_RESET_START);

	/* set EP0 max packet size : 32 bytes */
	while ((8 << max < EP0_PACKETSIZE) && max <= 3)
		max++;
	tmp  = max << USBD_EP_STAT_MAX_BIT;
	tmp |= USBD_EP_STAT_FLUSH;

	/* write EP0 settings */
	WRITE_REG32(USBD_EP_STAT(EP0), tmp);
	
	/* set EP0 GR/ALRM value */
	SET_FIFO_GR(EP0, 0x7);
	SET_FIFO_ALRM(EP0, 0x0);

	/* set frame mode */
	SET_FRAME_MODE(EP0);

#ifdef DBMX1_USBD_USE_MULTI_INTERFACE
	/* prepare to setup endpoint */
	udc_device->configuration = 0;
	for (i = 0; i <= 1; i++) {
		for (j = 0; j <=1; j++) {
			if (usbd_device_interface_descriptor (udc_device,  0, 0, i, j)) {
				udc_device->interface = i;
				udc_device->alternate = j;
				usbd_device_event (udc_device, DEVICE_SET_INTERFACE, 0);
			}
		}
	}
#endif

	/* enable the USB for processing */
	ENABLE_USB_PROCESS();

	/* enable USB/EP interrupt */
	ENABLE_EP_INTR(EP0, USBD_EP_MASK_MDEVREQ | USBD_EP_MASK_DEVREQ |
		       USBD_EP_MASK_EOF | USBD_EP_MASK_EOT | USBD_EP_MASK_FIFO_ERROR);
	ENABLE_USB_INTR(USBD_MASK_RES | USBD_MASK_SUSP | USBD_MASK_CFG_CHG |
			USBD_MASK_RESET_START |	USBD_MASK_RESET_STOP);

	spin_unlock_irqrestore(&dev->lock, flags);
}


/**
 * udc_disable - disable the UDC
 *
 * Switch off the UDC
 */
void udc_disable (void)
{
	PDEBUG(DBG_INIT, "called");

	/* reset device pointer */
	udc_device = NULL;

	/* disable USBD */
	DISABLE_USB_PROCESS();
	DISABLE_USB();

	/* free ep0 urb */
	kfree(ep0_urb->buffer);
}


/**
 * udc_startup - allow udc code to do any additional startup
 */
void udc_startup_events (struct usb_device_instance *device)
{
	PDEBUG(DBG_INIT, "called");
	
	usbd_device_event (device, DEVICE_INIT, 0);
	usbd_device_event (device, DEVICE_CREATE, 0);
	usbd_device_event (device, DEVICE_HUB_CONFIGURED, 0);
	usbd_device_event (device, DEVICE_RESET, 0);
}


/* ********************************************************************************************* */

/**
 * udc_name - return name of USB Device Controller
 */
char *udc_name (void)
{
	return UDC_NAME;
}

/**
 * udc_request_udc_irq - request UDC interrupt
 *
 * Return non-zero if not successful.
 */
int udc_request_udc_irq ()
{
	unsigned long flags;
	int i;
	
	PDEBUG(DBG_INIT, "called");

	/* initialize driver */
	if (dbmx1_init() < 0) {
		return -1;
	}

	spin_lock_irqsave(&dev->lock, flags);

	/* setup GPIO */
	dragonball_register_gpio(PORT_B, 23, PRIMARY | OUTPUT, "USB_SUS");
	dragonball_register_gpio(PORT_B, 27, PRIMARY | OUTPUT, "USB_VMO");
	dragonball_register_gpio(PORT_B, 26, PRIMARY | OUTPUT, "USB_VPO");
	dragonball_register_gpio(PORT_B, 21, PRIMARY | OUTPUT, "USB_OE");
	dragonball_register_gpio(PORT_B, 25, PRIMARY | INPUT, "USB_VM");
	dragonball_register_gpio(PORT_B, 24, PRIMARY | INPUT, "USB_VP");
	dragonball_register_gpio(PORT_B, 22, PRIMARY | INPUT, "USB_RCV");

	if (request_irq(IRQ_USBD_INT_0, dbmx1_ep0_intr,
			SA_INTERRUPT, "usbd_ep0", (void*)dev) < 0) {
		INFO("error - EP0 request_irq failed");
		goto error_ep0;
	}
	if (request_irq(IRQ_USBD_INT_1, dbmx1_epn_intr,
			SA_INTERRUPT, "usbd_ep1", (void*)dev) < 0) {
		INFO("error - EP1 request_irq failed");
		goto error_ep1;
	}
	if (request_irq(IRQ_USBD_INT_2, dbmx1_epn_intr,
			SA_INTERRUPT, "usbd_ep2", (void*)dev) < 0) {
		INFO("error - EP2 request_irq failed");
		goto error_ep2;
	}
	if (request_irq(IRQ_USBD_INT_3, dbmx1_epn_intr,
			SA_INTERRUPT, "usbd_ep3", (void*)dev) < 0) {
		INFO("error - EP3 request_irq failed");
		goto error_ep3;
	}
	if (request_irq(IRQ_USBD_INT_4, dbmx1_epn_intr,
			SA_INTERRUPT, "usbd_ep4", (void*)dev) < 0) {
		INFO("error - EP4 request_irq failed");
		goto error_ep4;
	}
	if (request_irq(IRQ_USBD_INT_5, dbmx1_epn_intr,
			SA_INTERRUPT, "usbd_ep5", (void*)dev) < 0) {
		INFO("error - EP5 request_irq failed");
		goto error_ep5;
	}
	if (request_irq(IRQ_USBD_INT_6, dbmx1_udc_intr,
			SA_INTERRUPT, "usbd", (void*)dev) < 0) {
		INFO("error - USBD request_irq failed");
		goto error_usbd;
	}

	spin_unlock_irqrestore(&dev->lock, flags);
	return 0;

	/* error handling */
 error_usbd:
	free_irq(IRQ_USBD_INT_5, (void*)dev);
 error_ep5:
	free_irq(IRQ_USBD_INT_4, (void*)dev);
 error_ep4:
	free_irq(IRQ_USBD_INT_3, (void*)dev);
 error_ep3:
	free_irq(IRQ_USBD_INT_2, (void*)dev);
 error_ep2:
	free_irq(IRQ_USBD_INT_1, (void*)dev);
 error_ep1:
	free_irq(IRQ_USBD_INT_0, (void*)dev);
 error_ep0:
	for (i=21; i<=27; i++) {
		dragonball_unregister_gpio(PORT_B, i);
	}
	spin_unlock_irqrestore(&dev->lock, flags);
	return -1;
}

/**
 * udc_request_cable_irq - request Cable interrupt
 *
 * Return non-zero if not successful.
 */
int udc_request_cable_irq ()
{
	unsigned long flags;

	PDEBUG(DBG_INIT, "called");
	
	spin_lock_irqsave(&dev->lock, flags);

	/* setup GPIO for VBUS */
	dragonball_register_gpio(USBD_VBUS_GPIO_PORT, USBD_VBUS_GPIO_PIN,
			   GPIO | INPUT | TRISTATE, "USB_VBUS");

	/* start monitoring vbus */
	if (request_irq(USBD_VBUS_GPIO_IRQ, udc_int_hndlr_cable,
			SA_SHIRQ, "usbd_vbus", (void*)dev) < 0) {
		INFO("error - VBUS request_irq failed");
		goto error_vbus;
	}
	dbmx1_enable_monitor_vbus(0);

	spin_unlock_irqrestore(&dev->lock, flags);
	return 0;

 error_vbus:
	dbmx1_disable_monitor_vbus();
	free_irq(USBD_VBUS_GPIO_IRQ, (void*)dev);
	dragonball_unregister_gpio(USBD_VBUS_GPIO_PORT, USBD_VBUS_GPIO_PIN);  
	spin_unlock_irqrestore(&dev->lock, flags);
	return -1;
}

/**
 * udc_request_udc_io - request UDC io region
 *
 * Return non-zero if not successful.
 */
int udc_request_io ()
{
	return 0;
}

/**
 * udc_release_release_io - release UDC io region
 */
void udc_release_io ()
{
}

/**
 * udc_release_udc_irq - release UDC irq
 */
void udc_release_udc_irq ()
{
	int i;

	PDEBUG(DBG_INIT, "called");
	
	free_irq(IRQ_USBD_INT_0, (void*)dev);
	free_irq(IRQ_USBD_INT_1, (void*)dev);
	free_irq(IRQ_USBD_INT_2, (void*)dev);
	free_irq(IRQ_USBD_INT_3, (void*)dev);
	free_irq(IRQ_USBD_INT_4, (void*)dev);
	free_irq(IRQ_USBD_INT_5, (void*)dev);
	free_irq(IRQ_USBD_INT_6, (void*)dev);

	/* unregister GPIO */
	for (i=21; i<=27; i++) {
		dragonball_unregister_gpio(PORT_B, i);
	}
}

/**
 * udc_release_cable_irq - release Cable irq
 */
void udc_release_cable_irq ()
{
	PDEBUG(DBG_INIT, "called");
	
	dbmx1_disable_monitor_vbus();
	free_irq(USBD_VBUS_GPIO_IRQ, (void*)dev);

	/* unregister GPIO */
	dragonball_unregister_gpio(USBD_VBUS_GPIO_PORT, USBD_VBUS_GPIO_PIN);

	dbmx1_release();
}

/**
 * udc_regs - dump registers
 *
 * Dump registers with printk
 */
void udc_regs (void)
{
	printk ("\n");
}
