/*
 *  linux/fs/xvfat/file.c
 *  based upon linux/fs/fat/file.c
 *
 *  Written 1992,1993 by Werner Almesberger
 *
 *  regular file handling primitives for fat-based filesystems
 *
 * Copyright 2003 Sony Corporation
 */

#include <linux/sched.h>
#include <linux/locks.h>
#include <linux/fs.h>
#include <linux/xvfat_fs.h>
#include <linux/errno.h>
#include <linux/fcntl.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/pagemap.h>
#include <linux/fat_cvf.h>
#include <linux/fwrq.h>

#include <asm/uaccess.h>
#include <asm/system.h>

#define PRINTK(x)
#define Printk(x) printk x

static ssize_t xfat_file_write(struct file *filp, const char *buf,
				  size_t count, loff_t *ppos);
static int xfat_file_flush(struct file *filp);
static int xfat_file_fsync(struct file *filp, struct dentry *dentry,
			  int datasync);
static int xfat_file_ioctl(struct inode * inode, struct file * filp,
			  unsigned int cmd, unsigned long arg);

struct file_operations xfat_file_operations = {
	llseek:		generic_file_llseek,
	read:		xfat_file_read,
	write:		xfat_file_write,
	mmap:		generic_file_mmap,
	fsync:		xfat_file_fsync,
	flush:		xfat_file_flush,
	ioctl:		xfat_file_ioctl,
};

struct inode_operations xfat_file_inode_operations = {
	truncate:	xfat_truncate,
	setattr:	xfat_notify_change,
};

ssize_t xfat_file_read(
	struct file *filp,
	char *buf,
	size_t count,
	loff_t *ppos)
{
	struct inode *inode = filp->f_dentry->d_inode;
	if (fwrq_check_blkdev_media(inode->i_dev, 1) == 1)
		return -EIO;
	return MSDOS_SB(inode->i_sb)->cvf_format
			->cvf_file_read(filp,buf,count,ppos);
}


int xfat_get_block(struct inode *inode, long iblock, struct buffer_head *bh_result, int create)
{
	struct super_block *sb = inode->i_sb;
	unsigned long phys;

	phys = xfat_bmap(inode, iblock);
	if (phys) {
		bh_result->b_dev = inode->i_dev;
		bh_result->b_blocknr = phys;
		bh_result->b_state |= (1UL << BH_Mapped);
		return 0;
	}
	if (!create)
		return 0;
	if (iblock << sb->s_blocksize_bits != MSDOS_I(inode)->mmu_private) {
		return -EIO;
	}
	if (!(iblock % MSDOS_SB(inode->i_sb)->cluster_size)) {
		fwrq_transaction_t *ts;
		ts = MSDOS_I(inode)->i_transaction;
		if (xfat_add_cluster_ts(inode, ts) < 0)
				return -ENOSPC;
	}
	MSDOS_I(inode)->mmu_private += sb->s_blocksize;
	phys = xfat_bmap(inode, iblock);
	if (!phys) {
		return -EIO;
	}
	bh_result->b_dev = inode->i_dev;
	bh_result->b_blocknr = phys;
	bh_result->b_state |= (1UL << BH_Mapped);
	bh_result->b_state |= (1UL << BH_New);
	return 0;
}

static ssize_t xfat_file_write(
	struct file *filp,
	const char *buf,
	size_t count,
	loff_t *ppos)
{
	struct inode *inode = filp->f_dentry->d_inode;
	struct super_block *sb = inode->i_sb;
	if (fwrq_check_blkdev_media(inode->i_dev, 1) == 1)
		return -EIO;
	return MSDOS_SB(sb)->cvf_format
			->cvf_file_write(filp,buf,count,ppos);
}

ssize_t default_xfat_file_write(
	struct file *filp,
	const char *buf,
	size_t count,
	loff_t *ppos)
{
	struct inode *inode = filp->f_dentry->d_inode;
	int retval;
	fwrq_transaction_t *ts;

	vfat_printk("default_xfat_file_write[%d:%s]: filp=%08x count=%d *ppos=%d\n",
		    current->pid, current->comm, (unsigned)filp, count, *ppos);
	ts = fwrq_transaction_get(inode->i_dev, FWRQ_TS_LOCK);
	if(!ts) return -EIO; /* FES patch */
	if( MSDOS_I(inode)->i_transaction )
		printk("default_xfat_file_write: exist old fwrq transaction\n");
	MSDOS_I(inode)->i_transaction = ts;

	retval = generic_file_write(filp, buf, count, ppos);
	if (retval > 0) {
		inode->i_mtime = inode->i_ctime = CURRENT_TIME;
		MSDOS_I(inode)->i_attrs |= ATTR_ARCH;
		fwrq_add_queue_inode(inode);
	}
	if( MSDOS_I(inode)->i_transaction ){
		if( fwrq_transaction_commit_to_queue(MSDOS_I(inode)->i_transaction) == -1 )
			printk("Warning: unsafety writing order (default_xfat_file_write)\n");
		if( MSDOS_I(inode)->i_transaction != ts ){
			printk("default_xfat_file_write: detect unknown fwrq transaction\n");
		}
		MSDOS_I(inode)->i_transaction = NULL;
		
	}else{
		printk("default_xfat_file_write: broken fwrq transaction\n");
	}
	return retval;
}

void xfat_truncate_ts(struct inode *inode, fwrq_transaction_t *ts)
{
	struct msdos_sb_info *sbi = MSDOS_SB(inode->i_sb);
	int cluster;

	/* Why no return value?  Surely the disk could fail... */
	if (IS_RDONLY (inode))
		return /* -EPERM */;
	if (IS_IMMUTABLE(inode))
		return /* -EPERM */;
	cluster = 1 << sbi->cluster_bits;
	/* 
	 * This protects against truncating a file bigger than it was then
	 * trying to write into the hole.
	 */
	if (MSDOS_I(inode)->mmu_private > inode->i_size)
		MSDOS_I(inode)->mmu_private = inode->i_size;
	MSDOS_I(inode)->i_attrs |= ATTR_ARCH;
	inode->i_ctime = inode->i_mtime = CURRENT_TIME;
	fwrq_transaction_add_tail_inode(ts,inode);
	xfat_free_ts(inode,
		    (inode->i_size + (cluster - 1)) >> sbi->cluster_bits,
		    ts);
}

void xfat_truncate(struct inode *inode)
{
	/* 
	 *  fat_notify_change() 
	 *    vmtruncate()
	 *      fat_truncate()         (i_op->truncate())
	 */
	fwrq_transaction_t *ts;

	ts = fwrq_transaction_get(inode->i_dev, FWRQ_TS_NOLOCK);
	if(!ts) return; /* FES patch */
	fwrq_flush_queue(inode->i_dev, 1);
	xfat_truncate_ts(inode, ts);
	if( fwrq_transaction_commit_to_queue_and_flush(ts, 1) == -1 )
		printk("Warning: unsafety writing order (xfat_truncate: inode=%08x)\n", (unsigned)inode);
}

int xfat_file_flush(struct file *filp)
{
	struct inode *inode = filp->f_dentry->d_inode;
	int rval;

	vfat_printk("xfat_file_flush[%d:%s]: filp=%08x\n",
		    current->pid, current->comm, (unsigned)filp);
	if (fwrq_check_blkdev_media(inode->i_dev, 1) == 1) {
		rval = -EIO;
		goto flush_end;
	}
	rval = fsync_dev(inode->i_dev);
	vfat_printk("xfat_file_flush[%d:%s]: fsync_dev() rval=%d\n",
		    current->pid, current->comm, rval);
	if (rval < 0)
		goto flush_end;
	vfat_printk("xfat_file_flush[%d:%s]: fwrq_check_blkdev_media()\n",
		    current->pid, current->comm);
	if (fwrq_check_blkdev_media(inode->i_dev, 1) == 1) {
		rval = -EIO;
		goto flush_end;
	}
 flush_end:
	vfat_printk("xfat_file_flush[%d:%s]: fwrq_flush_queue() rval=%d\n",
		    current->pid, current->comm, rval);
	if( rval >= 0 )
		fwrq_flush_queue(inode->i_dev, 1);
	vfat_printk("xfat_file_flush[%d:%s]: return=%d\n",
		    current->pid, current->comm, rval);
	return rval;
}

static int
xfat_file_fsync(struct file *filp, struct dentry *dentry, int datasync)
{
	struct inode *inode = filp->f_dentry->d_inode;
	int rval;

	vfat_printk("xfat_file_fsync[%d:%s]: filp=%08x, datasync=%d\n",
		    current->pid, current->comm, (unsigned)filp, datasync);
	if (fwrq_check_blkdev_media(inode->i_dev, 1) == 1) {
		rval = -EIO;
		goto fsync_end;
	}
	rval = file_fsync(filp, dentry, datasync);
	if( rval != 0 )
		goto fsync_end;
	if (fwrq_check_blkdev_media(inode->i_dev, 1) == 1)
		rval = -EIO;
 fsync_end:
	vfat_printk("xfat_file_fsync[%d:%s]: return=%d\n",
		    current->pid, current->comm, rval);
	return rval;
}

static int xfat_file_ioctl(struct inode * inode, struct file * filp,
			  unsigned int cmd, unsigned long arg)
{
	switch (cmd) {
	case XVFAT_IOCTL_GETLONGNAME:
		return xvfat_ioctl_iochar_to_uni16(inode,(struct xvfat_lname_box *)arg);
	case XVFAT_IOCTL_GETMINFO:
		return xvfat_ioctl_get_minfo(inode,(struct xvfat_media_info *)arg);
	default:
		return -EINVAL;
	}
	return 0;
}
