/*
 * linux/fs/xvfat/fwrq.c 
 *
 * Copyright 2003 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; version 2
 * of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/list.h>
#include <linux/locks.h>
#include <linux/smp_lock.h>
#include <linux/slab.h>
#include <linux/blkdev.h>
#include <linux/fwrq.h>
#include <linux/rdchk.h>

#undef FWRQ_DEBUG_FLUSH

#ifdef FWRQ_DEBUG
# ifndef FWRQ_DEBUG_FLUSH
#  define FWRQ_DEBUG_FLUSH
# endif
#endif

#define MEDIA_POLLING_INTERVAL 3

#define calc_bh_rsector(bh)     ((bh)->b_blocknr * ((bh)->b_size >> 9))

#define FWRQ_HDS_MAX	(0x20000 / 512)	/* Up to 128KB cluster */

static struct semaphore dh_alloc_sem;

static struct list_head dev_removable;
static kmem_cache_t *fwrq_head_cachep;
static kmem_cache_t *fwrq_transaction_cachep;
static struct buffer_head *bhs[FWRQ_HDS_MAX];
static spinlock_t ts_list_lock = SPIN_LOCK_UNLOCKED;

#define fwrq_list_entry(list) list_entry((list),struct fwrq_head,fwrqh_list)
#define fwrq_block_list_entry(list) list_entry((list),struct fwrq_head,b.block_list)
#define fwrq_seq_list_entry(list) list_entry((list),struct fwrq_head,b.seq_list)
#define dh_list_entry(list) list_entry((list),struct fwrq_dev_head,dh_list)

static void fwrq_free_block_list(struct list_head *list);

static unsigned get_blkno(struct fwrq_dev_head *dh, unsigned sector)
{
	if( sector < dh->block_offset )
		return 0;
	else
		return 1 + (sector - dh->block_offset) / dh->block_size;
}

static unsigned get_first_sec(struct fwrq_dev_head *dh, unsigned blkno)
{
	if( blkno == 0 )
		return 0;
	else
		return dh->block_offset + (blkno - 1) * dh->block_size;
}

static unsigned get_last_sec(struct fwrq_dev_head *dh, unsigned blkno)
{
	if( blkno == 0 )
		return dh->block_offset - 1;
	else
		return dh->block_offset + blkno * dh->block_size - 1;
}

/***************************************************************************/
#define DH_HASH_SIZE	32
#define dh_hashfn(dev)	((MAJOR(dev)*4+MINOR(dev))%DH_HASH_SIZE)
static struct list_head dh_hash_table[DH_HASH_SIZE];
static spinlock_t dh_lock = SPIN_LOCK_UNLOCKED;

static struct fwrq_dev_head *get_dh_hash(kdev_t dev)
{
	struct list_head *list, *pos;
	struct fwrq_dev_head *dh;

	spin_lock(&dh_lock);
	list = &dh_hash_table[dh_hashfn(dev)];
	if( list_empty(list) ){
		spin_unlock(&dh_lock);
		return NULL;
	}
	for( pos = list->next; pos != list; pos = pos->next ){
		dh = list_entry(pos, struct fwrq_dev_head, dh_hash_list);
		if( dh->dh_dev == dev ){
			spin_unlock(&dh_lock);
			return dh;
		}
	}
	spin_unlock(&dh_lock);
	return NULL;
}

static void add_dh_hash(kdev_t dev, struct fwrq_dev_head *dh)
{
	struct list_head *list;

	spin_lock(&dh_lock);
	list = &dh_hash_table[dh_hashfn(dev)];
	list_add(&dh->dh_hash_list, list);
	spin_unlock(&dh_lock);
}

static void remove_dh_hash(struct fwrq_dev_head *dh)
{
	spin_lock(&dh_lock);
	list_del(&dh->dh_hash_list);
	spin_unlock(&dh_lock);
}

static void free_dh_all(void)
{
	static int __fwrq_finish_queue(struct fwrq_dev_head *dh);
	int i;
	struct list_head *pos;
	struct fwrq_dev_head *ptr;

	for( i=0; i<DH_HASH_SIZE; i++ ){
		while( (pos = dh_hash_table[i].next) != &dh_hash_table[i] ){
			ptr = list_entry(pos, struct fwrq_dev_head,
					 dh_hash_list);
			while(__fwrq_finish_queue(ptr))
				;
		}
	}
}
/***************************************************************************/
#define blktop_hashfn(bn)	(bn%FWRQ_BLKTOP_HASH_SIZE)
static spinlock_t blktop_hash_lock = SPIN_LOCK_UNLOCKED;

static struct fwrq_head *
get_blktop_hash(struct fwrq_dev_head *dh, unsigned blkno)
{
	struct list_head *list, *pos;
	struct fwrq_head *btop;

	spin_lock(&blktop_hash_lock);
	list = &dh->blktop_hash_table[blktop_hashfn(blkno)];
	if( list_empty(list) ){
		spin_unlock(&blktop_hash_lock);
		return NULL;
	}
	for( pos = list->next; pos != list; pos = pos->next ){
		btop = list_entry(pos, struct fwrq_head, b.hash_list);
		if( btop->b.b_blkno == blkno ){
			spin_unlock(&blktop_hash_lock);
			return btop;
		}
	}
	spin_unlock(&blktop_hash_lock);
	return NULL;
}

static void
add_blktop_hash(struct fwrq_dev_head *dh, struct fwrq_head *btop)
{
	struct list_head *list;

	spin_lock(&blktop_hash_lock);
	list = &dh->blktop_hash_table[blktop_hashfn(btop->b.b_blkno)];
	list_add(&btop->b.hash_list, list);
	spin_unlock(&blktop_hash_lock);
}

static void
del_blktop_hash(struct fwrq_head *btop)
{
	spin_lock(&blktop_hash_lock);
	list_del(&btop->b.hash_list);
	spin_unlock(&blktop_hash_lock);
}

#ifdef NEVER
void free_blktop_hash_all(struct fwrq_dev_head *dh)
{
	int i;

	spin_lock(&blktop_hash_lock);
	for( i=0; i<FWRQ_BLKTOP_HASH_SIZE; i++ )
		INIT_LIST_HEAD(&dh->blktop_hash_table[i]);
	spin_unlock(&blktop_hash_lock);
}
#endif

static void print_blktop_hash_table(struct fwrq_dev_head *dh)
{
	int i;
	struct list_head *list, *pos;
	struct fwrq_head *btop;

	spin_lock(&blktop_hash_lock);
	for( i=0; i<FWRQ_BLKTOP_HASH_SIZE; i++ ){
		list = &dh->blktop_hash_table[i];
		for( pos = list->next; pos != list; pos = pos->next ){
			btop = list_entry(pos, struct fwrq_head, b.hash_list);
			printk("blktop_hash[%d]: %08x[%ld] blkno=%ld\n", i,
			       (unsigned)btop, calc_bh_rsector(btop->f_bh), btop->b.b_blkno);
			if( pos->next == pos ){
				printk("  detect infinity loop\n");
				break;
			}
		}
	}
	spin_unlock(&blktop_hash_lock);
}
/***************************************************************************/

static struct fwrq_head *alloc_fwrq_head(void)
{
	struct fwrq_head *ret;
	static unsigned long last_warning;

	ret = kmem_cache_alloc(fwrq_head_cachep, GFP_NOFS);
	if (ret == 0) {
		printk("out of memory for fwrq_head\n");
		if (time_after(jiffies, last_warning + 5*HZ)) {
			printk(KERN_NOTICE "ENOMEM in %s, retrying.\n",
			       __FUNCTION__);
			last_warning = jiffies;
		}
		while (ret == 0) {
			//current->policy |= SCHED_YIELD;
			//schedule();
			yield();
			ret = kmem_cache_alloc(fwrq_head_cachep, GFP_NOFS);
		}
	}
	return ret;
}

static void free_fwrq_head(struct fwrq_head *fh)
{
	kmem_cache_free(fwrq_head_cachep, fh);
}

static fwrq_transaction_t *alloc_fwrq_transaction(void)
{
	fwrq_transaction_t *ret;
	static unsigned long last_warning;

	ret = kmem_cache_alloc(fwrq_transaction_cachep, GFP_NOFS);
	if (ret == 0) {
		printk("out of memory for fwrq_head\n");
		if (time_after(jiffies, last_warning + 5*HZ)) {
			printk(KERN_NOTICE "ENOMEM in %s, retrying.\n",
			       __FUNCTION__);
			last_warning = jiffies;
		}
		while (ret == 0) {
			//current->policy |= SCHED_YIELD;
			//schedule();
			yield();
			ret = kmem_cache_alloc(fwrq_transaction_cachep,
					       GFP_NOFS);
		}
	}
	return ret;
}

static void free_fwrq_transaction(fwrq_transaction_t *ts)
{
	spin_lock(&ts_list_lock);
	list_del(&ts->ts_list);
	spin_unlock(&ts_list_lock);
	kmem_cache_free(fwrq_transaction_cachep, ts);
}

/* This function is called only from fwrq_free_transaction_list */
/* It is corrected to be deleted in the multiple. */
static void free_fwrq_transaction_no_del(fwrq_transaction_t *ts)
{
	kmem_cache_free(fwrq_transaction_cachep, ts);
}

static void set_blktop(struct fwrq_head *fh)
{
	set_bit(FWRQ_ST_BLKTOP, &fh->f_state);
	add_blktop_hash(fh->f_dh, fh);
	fwrq_printk2("set_blktop:   <fh=%08x[%d]\n", fh, calc_bh_rsector(fh->f_bh));
}

static void unset_blktop(struct fwrq_head *btop)
{
	del_blktop_hash(btop);
	clear_bit(FWRQ_ST_BLKTOP, &btop->f_state);
	fwrq_printk2("unset_blktop: >fh=%08x[%d]\n", btop, calc_bh_rsector(btop->f_bh));
}

static struct fwrq_head *
attach_fwrq_head(struct fwrq_dev_head *dh, struct buffer_head *bh)
{
	struct fwrq_head *fh, *free_fh;

	fh = alloc_fwrq_head();

	if( buffer_fwrq(bh) ){
		free_fh = fh;
		fh = bh->b_private;
		free_fwrq_head(free_fh);
	}else{
		fwrq_printk2("attach_fwrq_head: fh=%08x[%d] b_count=%d\n",
			    fh, calc_bh_rsector(bh),
			    atomic_read(&bh->b_count));
		get_bh(bh);
		set_bit(BH_FWRQ, &bh->b_state);
		bh->b_private = fh;
		fh->f_bh = bh;
		fh->f_state = FWRQ_ST_VALID;
		fh->f_dh = dh;
		INIT_LIST_HEAD(&fh->fwrqh_list);
		INIT_LIST_HEAD(&fh->b.block_list);
		INIT_LIST_HEAD(&fh->b.seq_list);
		INIT_LIST_HEAD(&fh->b.hash_list);
		fh->b.b_blkno = get_blkno(dh, calc_bh_rsector(bh));
		fh->b.b_blkcnt = 0;
		INIT_LIST_HEAD(&fh->t.ts_head);
		INIT_LIST_HEAD(&fh->t.ts_tail);
	}
	return fh;
}

static void detach_fwrq_head(struct fwrq_head *fh)
{
	struct buffer_head *bh;

	if( fwrq_buffer_valid(fh) )
		panic("detach_fwrqh: invalid fwrq_head (fh=0x%08x)\n",
		      (unsigned int)fh);

	fwrq_printk2("detach_fwrq_head: fh=%08x[%ld] b_count=%d\n",
		    (unsigned)fh, calc_bh_rsector(fh->f_bh),
		    atomic_read(&fh->f_bh->b_count));

	/* extract fwrq_head from a hash table...  */

	bh = fh->f_bh;
	clear_bit(BH_FWRQ, &bh->b_state);
	bh->b_private = NULL;
	fh->f_bh = NULL;
	fh->f_state = 0;
	free_fwrq_head(fh);
	brelse(bh);
}

static void
__detach_fwrq_block(struct fwrq_head *btop)
{
	struct fwrq_head *p;
	struct list_head *pos;

	while( (pos = btop->fwrqh_list.prev) != &btop->fwrqh_list ) {
		p = fwrq_list_entry(pos);
		mark_buffer_uptodate(p->f_bh, 0);
		list_del(pos);
		detach_fwrq_head(p);
	}
	unset_blktop(btop);
	mark_buffer_uptodate(btop->f_bh, 0);
	list_del(&btop->fwrqh_list);
	detach_fwrq_head(btop);
}

static void
__detach_fwrq_block_seq_list(struct fwrq_head *seq_first)
{
	struct list_head *list;
	struct fwrq_head *btop;

	while( (list = seq_first->b.seq_list.prev)
	       != &seq_first->b.seq_list ) {
		btop = fwrq_seq_list_entry(list);
		list_del(list);
		__detach_fwrq_block(btop);
	}
	list_del(&seq_first->b.seq_list);
	__detach_fwrq_block(seq_first);
}

static struct fwrq_head *
find_fwrq_block(struct fwrq_dev_head *dh, struct fwrq_head *fh)
{
	return get_blktop_hash(dh, fh->b.b_blkno);
}

static struct fwrq_head *
make_new_block(struct fwrq_head *fh)
{
	set_blktop(fh);
	set_bit(FWRQ_ST_BLOCK, &fh->f_state);
	fh->b.b_blkcnt = 1;
	return fh;
}

static void __list_replace(struct list_head *list, struct list_head *newlist)
{
	struct list_head *p;

	if( !list_empty(list) ){
		p = list->prev;
		list_del_init(list);
		list_add(newlist, p);
	}else{
		INIT_LIST_HEAD(newlist);
	}
}

static void print_noblktop_list(struct list_head *list)
{
	struct list_head *pos;
	struct fwrq_head *btop;

	for( pos = list->next; pos != list; pos = pos->next ){
		btop = fwrq_list_entry(pos);
		printk(":%ld", calc_bh_rsector(btop->f_bh));
		if( fwrq_buffer_blktop(btop) ){
			printk("*");
		}
		if( list_empty(pos) ){
			printk(" {btop->fwrqh_list = (empty)}\n");
			return;
		}
	}
}

#ifdef FWRQ_DEBUG
static void print_btop(struct fwrq_head *btop)
{
	printk(" btop: %08x %c [%ld", (unsigned)btop,
	       list_empty(&btop->b.block_list) ? ' ' : '*',
	       calc_bh_rsector(btop->f_bh));
	print_noblktop_list(&btop->fwrqh_list);
	printk("]\n");
}
#endif

static void print_seq_list(struct list_head *list)
{
	struct list_head *pos;
	struct fwrq_head *btop;

	for( pos = list->next; pos != list; pos = pos->next ){
		btop = fwrq_seq_list_entry(pos);
		printk("  seq:   %08x %c [%ld", (unsigned)btop,
		       list_empty(&btop->b.block_list) ? ' ' : '*',
		       calc_bh_rsector(btop->f_bh));
		print_noblktop_list(&btop->fwrqh_list);
		printk("]\n");
		if( list_empty(pos) ){
			printk(" {btop->b.seq_list = (empty)}\n");
			return;
		}
	}
}

static void print_block_list(struct list_head *list)
{
	struct list_head *pos;
	struct fwrq_head *btop;

	for( pos = list->next; pos != list; pos = pos->next ){
		btop = fwrq_block_list_entry(pos);
		printk(" block: %08x %c [%ld", (unsigned)btop,
		       list_empty(&btop->b.block_list)?'#':' ',
		       calc_bh_rsector(btop->f_bh));
		print_noblktop_list(&btop->fwrqh_list);
		printk("]\n");
		print_seq_list(&btop->b.seq_list);
		if( list_empty(pos) ){
			printk(" {btop->b.block_list = (empty)}\n");
			return;
		}
	}
}

static void print_dh_list(struct fwrq_dev_head *dh)
{
	printk("dh list[dev=%04x]: rej=%d\n", dh->dh_dev, dh->reject_count);
	printk("block_queue:\n");
	print_block_list(&dh->block_queue);
	printk("pend_queue:\n");
	print_block_list(&dh->block_pending);
}

#ifdef NEVER
static void print_dh(struct fwrq_dev_head *dh)
{
	int pre;
	printk("dh=0x%08x dev=%d dh_count=%d rej=%d\n",
	       (unsigned)dh, dh->dh_dev, atomic_read(&dh->dh_count),
	       dh->reject_count);
	printk(" SecBytes=%ld blksiz=%ld blkoff=%ld last tslock=[%d:%s]\n",
	       dh->dh_secbytes, dh->block_size, dh->block_offset,
	       dh->dh_tslock_pid, dh->dh_tslock_comm);
	printk(" dh_state=%04x ", dh->dh_state);
	pre = '[';
	if( fwrq_dh_eio(dh) ){printk("%ceio", pre);pre = ',';}
	if( fwrq_dh_mediain(dh) ){printk("%cmediain", pre);pre = ',';}
	if( fwrq_dh_mcheck(dh) ){printk("%cmcheck", pre);pre = ',';}
#ifdef NEVER2
	printk("]  e_count=%d\n", atomic_read(&dh->bioe->e_count));
#else
	printk("]\n");
#endif
}
#endif

static void
refile_transaction_list(struct list_head *head_list,
			struct list_head *tail_list,
			struct fwrq_head *btop)
{
	struct list_head *pos;
	struct fwrq_transaction *ts;

	fwrq_printk2("refile_transaction_list: btop=%08x[%d]\n",
		     btop, calc_bh_rsector(btop->f_bh));
	for( pos=head_list->next; pos != head_list; pos=pos->next ){
		ts = list_entry(pos, struct fwrq_transaction, head_list);
		ts->t_head = btop;
		fwrq_printk2("  refile transaction: %08x (head)\n", ts);
	}
	for( pos=tail_list->next; pos != tail_list; pos=pos->next ){
		ts = list_entry(pos, struct fwrq_transaction, tail_list);
		ts->t_tail = btop;
		fwrq_printk2("  refile transaction: %08x (tail)\n", btop);
	}
}

static void
__insert_fh_before_btop(struct fwrq_head *fh, struct fwrq_head *btop)
{
	__list_replace(&btop->b.block_list, &fh->b.block_list);
	__list_replace(&btop->b.seq_list, &fh->b.seq_list);
	__list_replace(&btop->t.ts_head, &fh->t.ts_head);
	__list_replace(&btop->t.ts_tail, &fh->t.ts_tail);
	list_add_tail(&fh->fwrqh_list, &btop->fwrqh_list);
	fh->b.b_blkno = btop->b.b_blkno;
	fh->b.b_blkcnt = btop->b.b_blkcnt + 1;
	refile_transaction_list(&fh->t.ts_head, &fh->t.ts_tail, fh);
	unset_blktop(btop);
	set_blktop(fh);
}

static struct fwrq_head *
insert_fh_block(struct fwrq_head *fh, struct fwrq_head *btop)
{
	struct fwrq_head *p;
	struct list_head *pos;

	for( pos = btop->fwrqh_list.prev; pos != &btop->fwrqh_list;
	     pos = pos->prev ){
		p = fwrq_list_entry(pos);
		if( calc_bh_rsector(fh->f_bh) > calc_bh_rsector(p->f_bh) ){
			btop->b.b_blkcnt++;
			list_add(&fh->fwrqh_list, &p->fwrqh_list);
			return btop;
		}
		if( calc_bh_rsector(fh->f_bh) == calc_bh_rsector(p->f_bh) ){
			printk("insert_fh_block: doublicate fwrq_head fh=%08x[%ld], fh2=%08x[%ld] (2)\n", (unsigned)fh, calc_bh_rsector(fh->f_bh), (unsigned)p, calc_bh_rsector(p->f_bh));
			print_blktop_hash_table(fh->f_dh);
			while(1);
		}
	}
	if( calc_bh_rsector(fh->f_bh) > calc_bh_rsector(btop->f_bh) ){
		btop->b.b_blkcnt++;
		list_add(&fh->fwrqh_list, &btop->fwrqh_list);
		return btop;
	}else if( calc_bh_rsector(fh->f_bh) == calc_bh_rsector(btop->f_bh) ){
		printk("insert_fh_block: doublicate fwrq_head fh=%08x[%ld], fh2=%08x[%ld] (2)\n", (unsigned)fh, calc_bh_rsector(fh->f_bh), (unsigned)btop, calc_bh_rsector(btop->f_bh));
		print_blktop_hash_table(fh->f_dh);
		while(1);
	}

	__insert_fh_before_btop(fh, btop);
	return fh;
}

static struct fwrq_head *get_btop(struct fwrq_dev_head *dh,
				  struct fwrq_head *fh)
{
	struct fwrq_head *btop;

	btop = find_fwrq_block(dh, fh);
	if( btop ){
		if( !fwrq_buffer_block(fh) ){
			btop = insert_fh_block(fh, btop);
			set_bit(FWRQ_ST_BLOCK, &fh->f_state);
		}
	}
	return btop;
}

static struct fwrq_head *
get_block_first(struct fwrq_head *btop)
{
	struct list_head *list;
	struct fwrq_head *b;

	if( !list_empty(&btop->b.block_list) )
		return btop;

	if( list_empty(&btop->b.seq_list) )
		return btop;

	for( list = btop->b.seq_list.next;
	     list != &btop->b.seq_list;
	     list = list->next ){
		b = fwrq_seq_list_entry(list);
		if( !list_empty(&b->b.block_list) )
			return b;
	}
	printk("get_block_first: floating block sequence list? btop=%08x[%ld]\n",
	       (unsigned)btop, calc_bh_rsector(btop->f_bh));
	print_dh_list(btop->f_dh);
	while(1);
	return NULL;
}

static void
insert_block_list_tail(struct fwrq_head *btop, struct list_head *list)
{
	list_add_tail(&btop->b.block_list, list);
}

static void
btop_set_bit_all_seq(int flag, struct fwrq_head *btop_first)
{
	struct list_head *list;
	struct fwrq_head *btop;

	set_bit(flag, &btop_first->f_state);

	for( list = btop_first->b.seq_list.next;
	     list != &btop_first->b.seq_list;
	     list = list->next ){
		btop = fwrq_seq_list_entry(list);
		set_bit(flag, &btop->f_state);
	}
}

static void
btop_clear_bit_all_seq(int flag, struct fwrq_head *btop_first)
{
	struct list_head *list;
	struct fwrq_head *btop;

	set_bit(flag, &btop_first->f_state);

	for( list = btop_first->b.seq_list.next;
	     list != &btop_first->b.seq_list;
	     list = list->next ){
		btop = fwrq_seq_list_entry(list);
		clear_bit(flag, &btop->f_state);
	}
}

static int
__fwrq_add_queue(struct fwrq_dev_head *dh, struct fwrq_head *fh)
{
	struct fwrq_head *btop;

	btop = get_btop(dh, fh);
	if( !btop ){
		btop = make_new_block(fh);
		insert_block_list_tail(btop, &dh->block_queue);
	}
	return 0;
}

static struct fwrq_dev_head *
get_dh(struct buffer_head *bh)
{
	struct fwrq_dev_head *dh;

	if( bh->b_size != 512 ){
		printk("get_dh: not support b_size != 512 (b_size=%d)\n", bh->b_size);
		panic("uga-\n");
	}

	dh = get_dh_hash(bh->b_dev);
	if( dh == NULL ){
		printk("get_dh: not initiallized fwrq_dev_head\n");
		return NULL;
	}
	return dh;
}

int
fwrq_add_queue(struct buffer_head *bh)
{
	struct fwrq_dev_head *dh;
	struct fwrq_head *fh;
	int rval;

	fwrq_printk("fwrq_add_queue[%d:%s]: bh=%08x[%ld]\n",
		    current->pid, current->comm, (unsigned)bh,
		    calc_bh_rsector(bh));

	dh = get_dh(bh);
	if( !dh )
		return -1;

	down(&dh->dh_work_sem);
	if( fwrq_dh_eio(dh) ) {
		up(&dh->dh_work_sem);
		return 0;
	}

	fh = attach_fwrq_head(dh, bh);
	rval = __fwrq_add_queue(dh, fh);

	up(&dh->dh_work_sem);

#ifdef FWRQ_DEBUG2
	print_dh_list(dh);
#endif

	return rval;
}

static struct buffer_head *last_committed_bh;

#undef FWRQ_DEBUG_BHSYNC
#ifdef FWRQ_DEBUG_BHSYNC

#define BH_TABLE_SIZE 256
static int bh_table_count;
static struct buffer_head *bh_table[BH_TABLE_SIZE];
static void bh_table_init(void)
{
	bh_table_count = 0;
}
static void bh_table_add(struct buffer_head *bh)
{
	if( bh_table_count < BH_TABLE_SIZE ) {
		get_bh(bh);
		bh_table[bh_table_count++] = bh;
	}
}
static void bh_table_print(void)
{
	int i;
	struct buffer_head *bh;

	for (i=0; i<bh_table_count; i++) {
		bh = bh_table[i];
		if (buffer_locked(bh)) {
			printk("print_bh_table[%d:%s]: n=%d/%d bh=[%ld]\n",
			       current->pid, current->comm, i, bh_table_count,
			       calc_bh_rsector(bh));
		}
		brelse(bh);
	}
	bh_table_count = 0;
}
#endif

static struct fwrq_head *
__fwrq_fill_one(struct fwrq_head *btop, unsigned sec, struct fwrq_head *pos)
{
	struct fwrq_head *fh;
	struct buffer_head *bh;

	bh = bread(btop->f_dh->dh_dev, sec, btop->f_dh->dh_secbytes);
	fwrq_printk2("__fwrq_fill_one: bread(%04x,%d,%ld) bh=%08x\n",
		     btop->f_dh->dh_dev, sec,
		     btop->f_dh->dh_secbytes, (unsigned)bh);
	if (bh == NULL) {
		fwrq_printk("__fwrq_fill_one: fail to read sector %d\n", sec);
		return btop;
	}
	fh = attach_fwrq_head(btop->f_dh, bh);
	brelse(bh);
	fwrq_printk2("__fwrq_fill_one: fh=%08x[%ld] b_count=%d\n",
		     (unsigned)fh, calc_bh_rsector(fh->f_bh),
		     atomic_read(&fh->f_bh->b_count));
	if( pos ){
		btop->b.b_blkcnt++;
		list_add_tail(&fh->fwrqh_list, &pos->fwrqh_list);
	} else {
		__insert_fh_before_btop(fh, btop);
		btop = fh;
	}
	return btop;
}

static struct fwrq_head *
__fwrq_fill_block2(struct fwrq_head *btop)
{
	struct list_head *list;
	struct fwrq_head *p;
	struct fwrq_dev_head *dh;
	unsigned fsec, lsec, sec, head, tail;
	int clssiz;

	dh = btop->f_dh;
	fsec = get_first_sec(btop->f_dh, btop->b.b_blkno);
	lsec = get_last_sec(btop->f_dh, btop->b.b_blkno);
	list = &btop->fwrqh_list;

	if( fsec < dh->block_offset )
		clssiz = dh->block_size;
	else
		clssiz = dh->cluster_size;

	fwrq_printk2("__fwrq_fill_block2[%d:%s]: btop=%08x[%d] sec=[%d-%d]\n",
		    current->pid, current->comm,
		    (unsigned)btop,
		    calc_bh_rsector(btop->f_bh),
		    fsec, lsec);

	for( head = fsec; head <= lsec; head+=clssiz ){
		if( head + clssiz - 1 < lsec )
			tail = head + clssiz - 1;
		else
			tail = lsec;
		p = fwrq_list_entry(list);
		sec = calc_bh_rsector(p->f_bh);

		
		if( sec < head || sec > tail ){
			if( list != &btop->fwrqh_list && sec < head )
				printk("__fwrq_fill_block2: internal error: head=%d tail=%d p=%08x[%ld]\n",
					     head, tail, (unsigned)p,
					     calc_bh_rsector(p->f_bh));
			fwrq_printk2("__fwrq_fill_block2: no fill [%ld-%ld] next=[%ld] btop=[%ld]\n",
				     head, tail, sec,
				     calc_bh_rsector(btop->f_bh));
			continue;
		}

		fwrq_printk2("__fwrq_fill_block2: fill [%ld-%ld] next=[%ld] btop=[%ld]\n",
			     head, tail, sec,
			     calc_bh_rsector(btop->f_bh));
		for( sec = head; sec <= tail; sec++ ){
			p = fwrq_list_entry(list);
			fwrq_printk2("__fwrq_fill_block2: sec=%d p=%08x[%ld]\n",
				     sec, (unsigned)p,
				     calc_bh_rsector(p->f_bh));
			if( calc_bh_rsector(p->f_bh) == sec ){
				list = list->next;
			} else if( btop == p &&
				   calc_bh_rsector(p->f_bh) > sec ){
				btop = __fwrq_fill_one(btop, sec, NULL);
			} else {
				btop = __fwrq_fill_one(btop, sec, p);
			}
		}
	}
	return btop;
}

static void
__fwrq_commit_block2(struct fwrq_head *btop)
{
	struct list_head *list, *prev, *next;
	struct fwrq_head *p;
	int n, first;

	n = 0;
	first = 1;
	list = &btop->fwrqh_list;

	do {
		prev = list;
		next = list->next;

		if( n == FWRQ_HDS_MAX ){
			if( last_committed_bh )
				brelse(last_committed_bh);
			last_committed_bh = bhs[n-1];
			get_bh(last_committed_bh);
			ll_rw_block(WRITE, n, bhs);
			n = 0;
		}

		p = fwrq_list_entry(list);
		if (buffer_locked(p->f_bh)) {
			printk("Warning: found locked BH (sector=%ld)\n", calc_bh_rsector(p->f_bh));
			wait_on_buffer(p->f_bh);
		}
#ifdef FWRQ_DEBUG_BHSYNC
		bh_table_add(p->f_bh);
#endif
		bhs[n++] = p->f_bh;
		mark_buffer_dirty(p->f_bh);
		list_del_init(list);
		fwrq_printk2("__fwrq_commit_block2: fh=%08x[%ld] b_count=%d blktop=%d\n",
			     (unsigned)p, calc_bh_rsector(p->f_bh),
			     atomic_read(&p->f_bh->b_count), first);
		if (first) {
			unset_blktop(p);
			first = 0;
		}
		detach_fwrq_head(p);
	} while ( (list = next) != prev );
	if( last_committed_bh )
		brelse(last_committed_bh);
	last_committed_bh = bhs[n-1];
	get_bh(last_committed_bh);
	fwrq_printk("__fwrq_commit_block2: ll_rw_block() [%ld] L=%d D=%d n=%d\n",
		    calc_bh_rsector(bhs[0]), buffer_locked(bhs[0]),
		    buffer_dirty(bhs[0]), n);
	ll_rw_block(WRITE, n, bhs);
}

static void
__fwrq_commit_seq_list2(struct fwrq_head *seq_first)
{
	struct list_head *list, *prev, *next;
	struct fwrq_head *btop;

	list = &seq_first->b.seq_list;
	do {
		prev = list;
		next = list->next;
#ifdef FWRQ_DEBUG2
		printk("prefill: "); print_btop(fwrq_seq_list_entry(list));
#endif
		btop = __fwrq_fill_block2(fwrq_seq_list_entry(list));
#ifdef FWRQ_DEBUG
		printk("commit: "); print_btop(btop);
		printk("list=%08x &btop->b.seq_list=%08x\n",
		       list, &btop->b.seq_list);
#endif
		list_del_init(&btop->b.seq_list);
		__fwrq_commit_block2(btop);
	} while ( (list = next) != prev );
}

#ifdef NEVER
static void
__fwrq_commit_seq_pending_list_partial2(struct fwrq_head *seq_first)
{
	struct list_head *list, *prev, *next;
	struct fwrq_head *btop;

	list = &seq_first->b.seq_list;
	do {
		prev = list;
		next = list->next;
		btop = fwrq_seq_list_entry(list);
		if( !list_empty(&btop->t.ts_head) || list_empty(&btop->t.ts_tail) )
			return;
		__fwrq_commit_block2(btop);
		if ( next == prev ) 
			list_del_init(&btop->b.block_list);
		list_del_init(list);
	} while ( (list = next) != prev );
}
#endif

static void
__fwrq_commit_list2(struct list_head *block_list)
{
	struct fwrq_head *btop_first;
	struct list_head *pos;

	while( (pos = block_list->next) != block_list ){
		btop_first = fwrq_block_list_entry(pos);
		list_del_init(pos);
		__fwrq_commit_seq_list2(btop_first);
	}
}

#ifdef NEVER
static void
__fwrq_commit_pending_list_partial2(struct list_head *pend_list)
{
	struct list_head *list, *next;

	if( list_empty(pend_list) )
		return;
	for( list = pend_list->next, next = list->next;
	     list != pend_list; list = next, next = list->next ){
		__fwrq_commit_seq_pending_list_partial2(
					  fwrq_block_list_entry(list));
	}
}
#endif

static void
__fwrq_flush_queue(struct fwrq_dev_head *dh, int wait)
{
#ifdef FWRQ_DEBUG
	if( !list_empty(&dh->block_pending) )
		fwrq_printk("__fwrq_flush_queue[%d:%s]: block_pending=(%08x,%08x)\n",
			    current->pid, current->comm,
			    dh->block_pending.prev, dh->block_pending.next);
	if( atomic_read(&dh->dh_tslock_sem.count) <= 0 )
		fwrq_printk("__fwrq_flush_queue[%d:%s]: dh_tslock_sem=%d\n",
			    current->pid, current->comm,
			    atomic_read(&dh->dh_tslock_sem.count));
#endif

	down(&dh->dh_work_sem);

	if( fwrq_dh_eio(dh) ){
		up(&dh->dh_work_sem);
		return;
	}
#ifdef FWRQ_DEBUG_FLUSH
	printk("__fwrq_flush_queue: dev=%s wait=%d\n",
	       bdevname(dh->dh_dev), wait);
	print_dh_list(dh);
#endif
#ifdef FWRQ_DEBUG_BHSYNC
	bh_table_init();
#endif

	last_committed_bh = NULL;

	__fwrq_commit_list2(&dh->block_queue);
	/* __fwrq_commit_pending_list_partial2(&dh->block_pending); */ /* not test */


	if( last_committed_bh ){
		if( wait )
			wait_on_buffer(last_committed_bh);
		brelse(last_committed_bh);
	}
#ifdef FWRQ_DEBUG_BHSYNC
	bh_table_print();
#endif
	up(&dh->dh_work_sem);
}

int
fwrq_flush_queue(kdev_t dev, int wait)
{
	struct fwrq_dev_head *dh;

	fwrq_printk("fwrq_flush_queue[%d:%s]: dev=%04x wait=%d\n",
		     current->pid, current->comm, dev, wait);

	dh = get_dh_hash(dev);
	if( dh == NULL ){
		printk("fwrq_commit: not initiallized fwrq_dev_head\n");
		return -1;
	}

	__fwrq_flush_queue(dh, wait);

#ifdef FWRQ_DEBUG
	print_dh_list(dh);
	print_blktop_hash_table(dh);
#endif

	return 0;
}

int
fwrq_invalidate(kdev_t dev)
{
	static void fwrq_detach_transaction_list(struct list_head *list);
	struct fwrq_dev_head *dh;

	fwrq_printk("fwrq_invalidate[%d:%s]: dev=%04x\n",
		     current->pid, current->comm, dev);

	dh = get_dh_hash(dev);
	if( dh == NULL ){
		printk("fwrq_invalidate: not initiallized fwrq_dev_head\n");
		return -1;
	}

#ifdef FWRQ_DEBUG
	print_dh_list(dh);
	print_blktop_hash_table(dh);
	print_transaction_head(dh);
#endif

	fwrq_detach_transaction_list(&dh->transaction_head);
	fwrq_free_block_list(&dh->block_queue);
	fwrq_free_block_list(&dh->block_pending);

#ifdef FWRQ_DEBUG
	print_dh_list(dh);
	print_blktop_hash_table(dh);
	print_transaction_head(dh);
#endif

	fwrq_finish_queue(dev); /* FES patch */

	return 0;
}

/*******************************************************************/
#ifdef FWRQ_DEBUG
static void print_transaction(fwrq_transaction_t *ts)
{
	struct list_head *list, *next;
	struct fwrq_head *btop;

	printk("ts: %08x %c%c%c%c:", (unsigned)ts,
	       fwrq_ts_reject(ts) ? 'R' : '-',
	       fwrq_ts_lock(ts) ? 'L' : '-',
	       fwrq_ts_valid(ts) ? 'V' : '-',
	       fwrq_ts_fixed(ts) ? 'F' : '-');

	if( !ts->t_head ) {
		printk(" (nil)\n");
		return;
	}

	next = &ts->t_head->b.seq_list;
	do {
		list = next;
		btop = fwrq_seq_list_entry(list);
		printk(" [%ld", calc_bh_rsector(btop->f_bh));
		print_noblktop_list(&btop->fwrqh_list);
		printk("]");
		next = list->next;
	} while(list != &ts->t_tail->b.seq_list);
	printk("\n");
}

static void print_transaction_head(struct fwrq_dev_head *dh)
{
	struct list_head *pos;
	fwrq_transaction_t *ts;

	spin_lock(&ts_list_lock);
	for( pos = dh->transaction_head.next; pos != &dh->transaction_head;
	     pos = pos->next ){
		ts = list_entry(pos, fwrq_transaction_t, ts_list);
		print_transaction(ts);
#ifdef NEVER
		printk("ts: %08x %08x[%ld]-%08x[%ld]\n",
		       (unsigned)ts,
		       (unsigned)ts->t_head,
		       calc_bh_rsector(ts->t_head->f_bh),
		       (unsigned)ts->t_tail,
		       calc_bh_rsector(ts->t_tail->f_bh));
#endif
	}
	spin_unlock(&ts_list_lock);
}
#endif

static void ts_lock(struct fwrq_dev_head *dh)
{
	down(&dh->dh_tslock_sem);
	dh->dh_tslock_pid = current->pid;
	strncpy(dh->dh_tslock_comm, current->comm, 16);
}

static void ts_unlock(struct fwrq_dev_head *dh)
{
	dh->dh_tslock_pid = 0;
	dh->dh_tslock_comm[0] = '\0';
	up(&dh->dh_tslock_sem);
}

fwrq_transaction_t *fwrq_transaction_get(kdev_t dev, int flags)
{
	struct fwrq_dev_head *dh;
	fwrq_transaction_t *ts;

	dh = get_dh_hash(dev);
	if( !dh ){
		printk("fwrq_transaction_get: not initialized (dev=%d)\n",
		      dev); /* FES patch */
		return NULL; /* FES patch */
	}
	ts = alloc_fwrq_transaction();
	ts->t_dh = dh;
	ts->t_head = ts->t_tail = NULL;
	INIT_LIST_HEAD(&ts->head_list);
	INIT_LIST_HEAD(&ts->tail_list);
	INIT_LIST_HEAD(&ts->ts_list);
	spin_lock(&ts_list_lock);
	list_add(&ts->ts_list, &dh->transaction_head);
	ts->t_state = 0;
	if( !fwrq_dh_eio(dh) ) {
		if( flags & FWRQ_TS_LOCK )
			set_bit(FWRQ_TSST_LOCK, &ts->t_state);
		set_bit(FWRQ_TSST_VALID, &ts->t_state);
	}
	spin_unlock(&ts_list_lock);
	if( fwrq_ts_lock(ts) )
		ts_lock(dh);
	return ts;
}

static void set_transaction_head(fwrq_transaction_t *transaction,
				 struct fwrq_head *btop)
{
	if( transaction->t_head )
		list_del_init(&transaction->head_list);
	transaction->t_head = btop;
	if( btop )
		list_add(&transaction->head_list, &btop->t.ts_head);
}

static void set_transaction_tail(fwrq_transaction_t *transaction,
				 struct fwrq_head *btop)
{
	if( transaction->t_tail ){
		list_del(&transaction->tail_list);
		INIT_LIST_HEAD(&transaction->tail_list);
	}
	transaction->t_tail = btop;
	if( btop )
		list_add(&transaction->tail_list, &btop->t.ts_tail);
}

static void detach_transaction(fwrq_transaction_t *transaction)
{
	spin_lock(&ts_list_lock);
	set_transaction_head(transaction, NULL);
	set_transaction_tail(transaction, NULL);
	clear_bit(FWRQ_TSST_VALID, &transaction->t_state);
	spin_unlock(&ts_list_lock);
}

static struct fwrq_head *
fwrq_transaction_merge(struct fwrq_head *first, struct fwrq_head *pos,
		       struct fwrq_head *first_b, struct fwrq_head *pos_b,
		       int *reject)
{
	struct list_head *mid, *last, *list;

	*reject = 0;
	if( first == first_b ){
		for( list = &first->b.seq_list; list != &pos->b.seq_list;
		     list = list->next ){
			if( list == &pos_b->b.seq_list ){
				fwrq_printk("fwrq_transaction_merge: reject!! (first:%08x, A:%08x[%ld], B:%08x[%ld])\n",
					    (unsigned)first, (unsigned)pos,
					    calc_bh_rsector(pos->f_bh),
					    (unsigned)pos_b,
					    calc_bh_rsector(pos_b->f_bh));
				if( list == &first->b.seq_list )
					first = fwrq_seq_list_entry(list->next);
				else {
					list_del(&pos_b->b.seq_list);
					list_add_tail(&pos_b->b.seq_list,
						      &first->b.seq_list);
				}
				*reject = 1;
				return first;
			}
		}
	}else{
		last = first_b->b.seq_list.prev;
		mid = first->b.seq_list.prev;
		first->b.seq_list.prev = last;
		last->next = &first->b.seq_list;
		first_b->b.seq_list.prev = mid;
		mid->next = &first_b->b.seq_list;;
	}
	return first;
}

static int transaction_empty(fwrq_transaction_t *transaction)
{
	return !transaction->t_head;
}

#ifdef NEVER
void fwrq_transaction_add_head(fwrq_transaction_t *transaction,
			       struct buffer_head *bh)
{
	struct fwrq_dev_head *dh;
	struct fwrq_head *fh, *btop, *seq_first;
	struct fwrq_head *btop_ts, *seq_first_ts;
	int exist_btop, rej;

	fwrq_printk("fwrq_transaction_add_head: ts=%08x bh=%08x\n",
	       transaction, bh);

	if( !transaction ){
		fwrq_add_queue(bh);
		return;
	}

	dh = get_dh(bh);
	if( !dh )
		return;

	down(&dh->dh_work_sem);

	if( fwrq_dh_eio(dh) || !fwrq_ts_valid(transaction) ){
		up(&dh->dh_work_sem);
		return;
	}

	fh = attach_fwrq_head(dh, bh);
	btop = get_btop(dh,fh);
	if( btop ){
		seq_first = get_block_first(btop);
		exist_btop = 1;
	}else{
		seq_first = btop = make_new_block(fh);
		exist_btop = 0;
	}
	if( transaction_empty(transaction) ){
		set_transaction_head(transaction, btop);
		set_transaction_tail(transaction, btop);
		if( exist_btop ){
			list_del(&seq_first->b.block_list);
			INIT_LIST_HEAD(&seq_first->b.block_list);
		}
	}else{
		btop_ts = transaction->t_head;
		seq_first_ts = get_block_first(btop_ts);
		list_del(&seq_first_ts->b.block_list);
		INIT_LIST_HEAD(&seq_first_ts->b.block_list);
		if( exist_btop && seq_first_ts != seq_first ){
			list_del(&seq_first->b.block_list);
			INIT_LIST_HEAD(&seq_first->b.block_list);
		}

		seq_first = fwrq_transaction_merge(seq_first, btop,
						   seq_first_ts, btop_ts,
						   &rej);
		if( rej )
			set_bit(FWRQ_TSST_REJECT, &transaction->t_state);
		set_transaction_head(transaction, btop);
	}
	btop_set_bit_all_seq(FWRQ_ST_PEND, seq_first);
	insert_block_list_tail(seq_first, &dh->block_pending);

	up(&dh->dh_work_sem);
}
#endif /* NEVER */

void fwrq_transaction_add_tail(fwrq_transaction_t *transaction,
			       struct buffer_head *bh)
{
	struct fwrq_dev_head *dh;
	struct fwrq_head *fh, *btop, *seq_first;
	struct fwrq_head *btop_ts, *seq_first_ts;
	int exist_btop, rej;

	fwrq_printk("fwrq_transaction_add_tail[%d:%s]: ts=%08x bh=%08x[%ld]\n",
		     current->pid, current->comm, 
		    (unsigned)transaction, (unsigned)bh, calc_bh_rsector(bh));

	if( !transaction ){
		fwrq_add_queue(bh);
		return;
	}

	dh = get_dh(bh);
	if( !dh )
		return;

	down(&dh->dh_work_sem);

	if( fwrq_dh_eio(dh) || !fwrq_ts_valid(transaction) ){
		up(&dh->dh_work_sem);
		return;
	}

	fh = attach_fwrq_head(dh, bh);
	btop = get_btop(dh,fh);
	if( btop ){
		seq_first = get_block_first(btop);
		exist_btop = 1;
	}else{
		seq_first = btop = make_new_block(fh);
		exist_btop = 0;
	}

	if( transaction_empty(transaction) ){
		set_transaction_head(transaction, btop);
		set_transaction_tail(transaction, btop);
		if( exist_btop )
			list_del_init(&seq_first->b.block_list);
	}else{
		btop_ts = transaction->t_tail;
		seq_first_ts = get_block_first(btop_ts);
		list_del_init(&seq_first_ts->b.block_list);
		if( exist_btop && seq_first_ts != seq_first )
			list_del_init(&seq_first->b.block_list);

		/*
		printk(" ts merge: %08x[%d],%08x[%d] + %08x[%d],%08x[%d]\n",
		       seq_first_ts, calc_bh_rsector(seq_first_ts->f_bh),
		       btop_ts, calc_bh_rsector(btop_ts->f_bh),
		       seq_first, calc_bh_rsector(seq_first->f_bh),
		       btop_ts, calc_bh_rsector(btop_ts->f_bh));
		*/

		seq_first = fwrq_transaction_merge(seq_first_ts, btop_ts,
						   seq_first, btop, &rej);
		if( rej )
			set_bit(FWRQ_TSST_REJECT, &transaction->t_state);
		set_transaction_tail(transaction, btop);
	}
	/*
	printk(" queuing to block_pending list: %08x[%d] %08x\n",
	       seq_first, calc_bh_rsector(seq_first->f_bh), &dh->block_pending);
	*/
	btop_set_bit_all_seq(FWRQ_ST_PEND, seq_first);
	insert_block_list_tail(seq_first, &dh->block_pending);

	up(&dh->dh_work_sem);
#ifdef FWRQ_DEBUG2
	print_dh_list(dh);
#endif

}

#ifdef NEVER
static struct fwrq_head *get_transaction_pos(fwrq_transaction_t *transaction,
					     int pos)
{
	struct fwrq_head *btop;

	if( pos >= 0 ){
		btop = transaction->t_head;
		while( --pos >= 0 ){
			if( btop == transaction->t_tail )
				return NULL;
			btop = fwrq_seq_list_entry(btop->b.seq_list.next);
		}
	}else{
		btop = transaction->t_tail;
		while( ++pos < 0 ){
			if( btop == transaction->t_head )
				return NULL;
			btop = fwrq_seq_list_entry(btop->b.seq_list.prev);
		}
	}
	return btop;
}

void fwrq_transaction_add_pos(fwrq_transaction_t *transaction,
			      struct buffer_head *bh, int pos)
{
	struct fwrq_dev_head *dh;
	struct fwrq_head *fh, *btop, *seq_first;
	struct fwrq_head *btop_ts, *seq_first_ts;
	int exist_btop, rej;

	fwrq_printk("fwrq_transaction_add_pos[%d:%s]: ts=%08x bh=%08x pos=%d\n",
		     current->pid, current->comm,
		    (unsigned)transaction, (unsigned)bh, pos);

	if( !transaction ){
		fwrq_add_queue(bh);
		return;
	}

	dh = get_dh(bh);
	if( !dh )
		return;

	down(&dh->dh_work_sem);

	if( fwrq_dh_eio(dh) || !fwrq_ts_valid(transaction) ){
		up(&dh->dh_work_sem);
		return;
	}

	fh = attach_fwrq_head(dh, bh);
	btop = get_btop(dh,fh);
	if( btop ){
		seq_first = get_block_first(btop);
		exist_btop = 1;
	}else{
		seq_first = btop = make_new_block(fh);
		exist_btop = 0;
	}
	if( transaction_empty(transaction) ){
		set_transaction_head(transaction, btop);
		set_transaction_tail(transaction, btop);
		if( exist_btop ){
			list_del(&seq_first->b.block_list);
			INIT_LIST_HEAD(&seq_first->b.block_list);
		}
	}else{
		btop_ts = get_transaction_pos(transaction, pos);
		if( !btop_ts ){
			printk("fwrq_transaction_add_pos: out of range (pos=%d)\n", pos);
			goto exit_func;
		}
		seq_first_ts = get_block_first(btop_ts);
		list_del(&seq_first_ts->b.block_list);
		INIT_LIST_HEAD(&seq_first_ts->b.block_list);
		if( exist_btop && seq_first_ts != seq_first ){
			list_del(&seq_first->b.block_list);
			INIT_LIST_HEAD(&seq_first->b.block_list);
		}
		seq_first = fwrq_transaction_merge(seq_first_ts, btop_ts,
						   seq_first, btop, &rej);
		if( rej )
			set_bit(FWRQ_TSST_REJECT, &transaction->t_state);
		if( pos == -1 )
			set_transaction_tail(transaction, btop);
	}
	btop_set_bit_all_seq(FWRQ_ST_PEND, seq_first);
	insert_block_list_tail(seq_first, &dh->block_pending);
 exit_func:
	up(&dh->dh_work_sem);
}
#endif	/* NEVER */

static void __fwrq_transaction_concat(fwrq_transaction_t *transaction,
				      fwrq_transaction_t *append,
				      int append1_force)
{
	struct fwrq_dev_head *dh;
	struct fwrq_head *btop, *seq_first;
	struct fwrq_head *btop_b, *seq_first_b;
	int rej, islock;

	dh = transaction->t_dh;
	down(&dh->dh_work_sem);

	if( fwrq_dh_eio(dh) || !fwrq_ts_valid(transaction) ||
	    !fwrq_ts_valid(append) ){
		up(&dh->dh_work_sem);
		return;
	}

	islock = fwrq_ts_lock(append);

	if( transaction_empty(transaction) ) {
		set_transaction_head(transaction, append->t_head);
		set_transaction_tail(transaction, append->t_tail);
	} else {
		if( !transaction_empty(append) ) {
			seq_first = get_block_first(transaction->t_head);
			btop = transaction->t_tail;
			seq_first_b = get_block_first(append->t_head);
			btop_b = append->t_head;
			list_del(&seq_first->b.block_list);
			INIT_LIST_HEAD(&seq_first->b.block_list);
			list_del(&seq_first_b->b.block_list);
			INIT_LIST_HEAD(&seq_first_b->b.block_list);
			seq_first = fwrq_transaction_merge(seq_first, btop,
							   seq_first_b, btop_b,
							   &rej);
			if( rej ){
				if( append1_force ){
					fwrq_printk("__fwrq_transaction_concat: allow unsafety write order (add B:%08x[%ld])\n", (unsigned)btop_b, calc_bh_rsector(btop_b->f_bh));
				} else {
					fwrq_printk("__fwrq_transaction_concat: detect unsafety write order (add B:%08x[%ld])\n", (unsigned)btop_b, calc_bh_rsector(btop_b->f_bh));
					set_bit(FWRQ_TSST_REJECT,
						&transaction->t_state);
				}
			}
			insert_block_list_tail(seq_first, &dh->block_pending);
			set_transaction_tail(transaction, append->t_tail);
		}
	}
	if( islock )
		set_bit(FWRQ_TSST_LOCK, &transaction->t_state);
	if( fwrq_ts_fixed(append) ) {
		fwrq_printk("__fwrq_transaction_concat: detect unsafety write order (append fixed tranasction ts=%08x)\n", append);
		set_bit(FWRQ_TSST_REJECT, &transaction->t_state);
	}
	if( fwrq_ts_valid(append) )
	    detach_transaction(append);
	free_fwrq_transaction(append);
	up(&dh->dh_work_sem);
}

void fwrq_transaction_concat(fwrq_transaction_t *transaction,
			     fwrq_transaction_t *append)
{
	fwrq_printk("fwrq_transaction_concat[%d:%s]: ts=%08x ap=%08x\n",
		    current->pid, current->comm, 
		    (unsigned)transaction, (unsigned)append);
	__fwrq_transaction_concat(transaction, append, 0);
}

void fwrq_transaction_append1(fwrq_transaction_t *transaction,
				    fwrq_transaction_t *append)
{
	fwrq_printk("fwrq_transaction_append1_force[%d:%s]: ts=%08x ap=%08x\n",
		    current->pid, current->comm, 
		    (unsigned)transaction, (unsigned)append);
	__fwrq_transaction_concat(transaction, append, 1);
}

static int transaction_now(struct fwrq_head *seq_first)
{
	struct fwrq_head *btop;
	struct list_head *list;

	if( !list_empty(&seq_first->t.ts_head) )
		return 1;
	for( list = seq_first->b.seq_list.next; list != &seq_first->b.seq_list;
	     list = list->next ){
		btop = fwrq_seq_list_entry(list);
		if( !list_empty(&btop->t.ts_head) )
			return 1;
	}
	return 0;
}

static int
__fwrq_transaction_commit_to_queue(fwrq_transaction_t *transaction,
				   struct fwrq_dev_head *dh,
				   int *lock)
{
	struct fwrq_head *seq_first;
	int reject, islock;

	down(&dh->dh_work_sem);
	if( fwrq_dh_eio(dh) || !fwrq_ts_valid(transaction) ){
		*lock = 0;
		up(&dh->dh_work_sem);
		return 0;
	}
	islock = fwrq_ts_lock(transaction);
	reject = fwrq_ts_reject(transaction);
	if( !transaction_empty(transaction) ){
		seq_first = get_block_first(transaction->t_head);
		detach_transaction(transaction);
		if( !transaction_now(seq_first) ){
			list_del(&seq_first->b.block_list);
			btop_clear_bit_all_seq(FWRQ_ST_PEND, seq_first);
			insert_block_list_tail(seq_first, &dh->block_queue);
		}
	}else{
		if( fwrq_ts_valid(transaction) )
			detach_transaction(transaction);
	}
	free_fwrq_transaction(transaction);

	up(&dh->dh_work_sem);
	*lock = islock;
	if( reject ){
		dh->reject_count++;
		return -1;
	} else
		return 0;
}

static int
__fwrq_transaction_fix(fwrq_transaction_t *transaction,
		       struct fwrq_dev_head *dh)
{
	struct fwrq_head *seq_first;
	int reject;

	down(&dh->dh_work_sem);
	if( fwrq_dh_eio(dh) || !fwrq_ts_valid(transaction) ){
		up(&dh->dh_work_sem);
		return 0;
	}
	reject = fwrq_ts_reject(transaction);
	set_bit(FWRQ_TSST_FIXED, &transaction->t_state);
	if( !transaction_empty(transaction) ){
		seq_first = get_block_first(transaction->t_head);
		set_transaction_head(transaction, NULL);
		set_transaction_tail(transaction, NULL);
		if( !transaction_now(seq_first) ){
			list_del(&seq_first->b.block_list);
			btop_clear_bit_all_seq(FWRQ_ST_PEND, seq_first);
			insert_block_list_tail(seq_first, &dh->block_queue);
		}
	}
	up(&dh->dh_work_sem);
	if( reject ){
		return -1;
	} else
		return 0;
}

int fwrq_transaction_commit_to_queue(fwrq_transaction_t *transaction)
{
	struct fwrq_dev_head *dh;
	int rval, islock;

	fwrq_printk("fwrq_transaction_commit_to_queue[%d:%s]: ts=%08x\n",
		    current->pid, current->comm, (unsigned)transaction);

	if( !transaction )
		return 0;

	dh = transaction->t_dh;
	rval = __fwrq_transaction_commit_to_queue(transaction, dh, &islock);
	if( islock )
		ts_unlock(dh);
	return rval;
}

int fwrq_transaction_fix(fwrq_transaction_t *transaction)
{
	struct fwrq_dev_head *dh;
	int rval;

	fwrq_printk("fwrq_transaction_fix[%d:%s]: ts=%08x\n",
		    current->pid, current->comm, (unsigned)transaction);

	if( !transaction )
		return 0;

	dh = transaction->t_dh;
	rval = __fwrq_transaction_fix(transaction, dh);
	return rval;
}

int
fwrq_transaction_commit_to_queue_and_flush(fwrq_transaction_t *transaction,
					   int wait)
{
	struct fwrq_dev_head *dh;
	int rval, islock;

	fwrq_printk("fwrq_transaction_commit_to_queue_and_flush[%d:%s]: ts=%08x\n",
		    current->pid, current->comm, (unsigned)transaction);

	if( !transaction )
		return 0;

	dh = transaction->t_dh;
	rval = __fwrq_transaction_commit_to_queue(transaction, dh, &islock);
	__fwrq_flush_queue(dh, wait);

	if( islock )
		ts_unlock(dh);
	return rval;
}

#ifdef NEVER
int fwrq_transaction_length(fwrq_transaction_t *ts)
{
	struct fwrq_dev_head *dh;
	int len;

	fwrq_printk("fwrq_transaction_length[%d:%s]: ts=%08x\n",
		    current->pid, current->comm, (unsigned)ts);

	if( !ts )
		return -1;

	dh = ts->t_dh;
	down(&dh->dh_work_sem);
	if( transaction_empty(ts) ) {
		len = 0;
	} else {
		struct list_head *list, *next;
		struct fwrq_head *btop;

		len = 0;
		next = &ts->t_head->b.seq_list;
		do {
			list = next;
			len++;
			next = list->next;
		} while(list != &ts->t_tail->b.seq_list);
	}
	up(&dh->dh_work_sem);
	return len;
}
#endif

/*******************************************************************/

int fwrq_check_blkdev_media(kdev_t dev, int call_daemon)
{
	struct fwrq_dev_head *dh;
	int chg;

	dh = get_dh_hash(dev);
	if( !dh ) {
		fwrq_printk("fwrq_check_blkdev: not found DH (dev=%04x)\n",
		       dev);
		return 1;
	}
	if( fwrq_dh_eio(dh) )
		return 1;
	chg = rdchk_get_disk_change(dev, call_daemon);
	if( chg == 1 || chg == -2 ) {
		set_bit(FWRQ_DHST_EIO, &dh->dh_state);
		return 1;
	}
	return 0;
}

int fwrq_revalidate_blkdev_media(kdev_t dev)
{
	struct fwrq_dev_head *dh;

	dh = get_dh_hash(dev);
	if( !dh ){
		fwrq_printk("fwrq_check_blkdev: not found DH (dev=%04x)\n",
		       dev);
		return -1;
	}
	if( fwrq_dh_eio(dh) )
		revalidate_blkdev(dev, 0);
	return 0;
}

#ifdef NEVER
void fwrq_print_debug(kdev_t dev)
{
	struct fwrq_dev_head *dh;

	dh = get_dh_hash(dev);
	if( !dh ){
		panic("fwrq_transaction_get: not initialized (dev=%d)\n", dev);
	}
	printk("fwrq_print_debug: dev=%04x\n", dev);
	print_blktop_hash_table(dh);	
	print_dh_list(dh);
	print_dh(dh);
}
#endif

/*******************************************************************/
static void
fwrq_free_block_list(struct list_head *list)
{
	struct list_head *pos;

	while( (pos = list->next) != list ){
		list_del(pos);
		__detach_fwrq_block_seq_list(fwrq_block_list_entry(pos));
	}
}

static void
fwrq_detach_transaction_list(struct list_head *list)
{
	struct list_head *pos;
	struct fwrq_transaction *ts;
	struct fwrq_dev_head *dh;
	int islock;

	while( (pos = list->next) != list ){
		list_del(pos);
		ts = list_entry(pos, struct fwrq_transaction, ts_list);
		dh = ts->t_dh;
		islock = fwrq_ts_lock(ts);
		if( fwrq_ts_valid(ts) )
			detach_transaction(ts);
		if( islock )
			ts_unlock(dh);
	}
}

static void
fwrq_free_transaction_list(struct list_head *list)
{
	struct list_head *pos;
	struct fwrq_transaction *ts;
	struct fwrq_dev_head *dh;
	int islock;

	while( (pos = list->next) != list ){
		list_del(pos);
		ts = list_entry(pos, struct fwrq_transaction, ts_list);
		dh = ts->t_dh;
		islock = fwrq_ts_lock(ts);
		if( fwrq_ts_valid(ts) )
			detach_transaction(ts);
		free_fwrq_transaction_no_del(ts);
		if( islock )
			ts_unlock(dh);
	}
}

void
fwrq_init_queue(kdev_t dev, int blksiz, int blkoff, int clssiz)
{
	struct fwrq_dev_head *dh;
	int i;

	fwrq_printk("fwrq_init_queue: dev=%04x clssiz=%d blksiz=%d blkoff=%d\n",
		    dev, clssiz, blksiz, blkoff); /* FES patch */
	down(&dh_alloc_sem);
	dh = get_dh_hash(dev);
	if( dh == NULL ){
		dh = (struct fwrq_dev_head *)
			kmalloc(sizeof(struct fwrq_dev_head), GFP_KERNEL);
		if( !dh )
			panic("fwrq_init_queue: NOMEM\n");

#ifdef FWRQ_DEBUG
		printk("XVFAT: dev=%s elevator/IO-scheduler=in-order (%d)\n",
		       bdevname(dev), blk_inorder_elevator(dev, 1));
#else
		blk_inorder_elevator(dev, 1);
#endif
		dh->dh_dev = dev;
		dh->dh_secbytes = 512;
		INIT_LIST_HEAD(&dh->block_queue);
		INIT_LIST_HEAD(&dh->block_pending);
		add_dh_hash(dev, dh);
		for( i=0; i<FWRQ_BLKTOP_HASH_SIZE; i++ )
			INIT_LIST_HEAD(&dh->blktop_hash_table[i]);
		INIT_LIST_HEAD(&dh->transaction_head);
		dh->reject_count = 0;
		sema_init(&dh->dh_work_sem, 1);
		sema_init(&dh->dh_tslock_sem, 1);
		dh->dh_state = 0;
		if( rdchk_register_blkdev(dev) ){
			panic("rdchk_register_blkdev: no such device (dev=%s)\n",
			      bdevname(dev));
		}
		rdchk_get_fs(dev);
		list_add(&dh->dh_removable_list, &dev_removable);
		atomic_set(&dh->dh_count, 1);
	} else {
		atomic_inc(&dh->dh_count);
	}
	dh->block_size = blksiz;
	dh->block_offset = blkoff;
	dh->cluster_size = clssiz;
	up(&dh_alloc_sem);
}

static int
__fwrq_finish_queue(struct fwrq_dev_head *dh)
{
	if( atomic_dec_and_test(&dh->dh_count) ){
#ifdef FWRQ_DEBUG
		printk("XVFAT: dev=%s elevator/IO-scheduler=original (%d)\n",
		       bdevname(dh->dh_dev),
		       blk_inorder_elevator(dh->dh_dev, 0));

		print_blktop_hash_table(dh);	
		print_dh_list(dh);
		print_transaction_head(dh);
#ifdef NEVER
		printk("fwrq_finish_queue: last I/O error count=%d dh_state=%d\n",
		       atomic_read(&dh->bioe->e_count),
		       dh->dh_state);
#endif
#else /* not FWRQ_DEBUG */
		blk_inorder_elevator(dh->dh_dev, 0);
#endif

		fwrq_free_transaction_list(&dh->transaction_head);
		fwrq_free_block_list(&dh->block_queue);
		fwrq_free_block_list(&dh->block_pending);
		dh->reject_count = 0;
		list_del(&dh->dh_removable_list);
		rdchk_put_fs(dh->dh_dev);
		dh->dh_state = 0;
		remove_dh_hash(dh);
		kfree(dh);
		return 0;
	} else {
		return 1;
	}
}

void
fwrq_finish_queue(kdev_t dev)
{
	struct fwrq_dev_head *dh;

	fwrq_printk("fwrq_finish_queue: dev=%04x\n", dev); /* FES patch */
	down(&dh_alloc_sem);
	dh = get_dh_hash(dev);
	if( dh == NULL ){
		fwrq_printk("fwrq_finish_queue: not found dev=%04x\n", dev);
		up(&dh_alloc_sem); /* FES patch */
		return;
	}
	__fwrq_finish_queue(dh);
	up(&dh_alloc_sem);
}

int
__init fwrq_init(void)
{
	int i;

	INIT_LIST_HEAD(&dev_removable);

	fwrq_head_cachep = kmem_cache_create("fwrqh", sizeof(struct fwrq_head),
					 0, SLAB_HWCACHE_ALIGN, NULL, NULL);
	if( fwrq_head_cachep == NULL )
		panic("Cannot create fwrqh SLAB cache");

	fwrq_transaction_cachep  = kmem_cache_create("fwrqt",
					 sizeof(fwrq_transaction_t),
					 0, SLAB_HWCACHE_ALIGN, NULL, NULL);
	if( fwrq_transaction_cachep == NULL )
		panic("Cannot create fwrqt SLAB cache");

	for( i=0; i<DH_HASH_SIZE; i++ )
		INIT_LIST_HEAD(&dh_hash_table[i]);
	sema_init(&dh_alloc_sem, 1);
	return 0;
}

void fwrq_exit(void)
{
	down(&dh_alloc_sem);
	free_dh_all();
	up(&dh_alloc_sem);

	fwrq_printk("fwrq_exit: fwrq_head_cachep\n");
	if( fwrq_head_cachep != NULL )
		kmem_cache_destroy(fwrq_head_cachep);
	fwrq_printk("fwrq_exit: fwrq_transaction_cachep\n");
	if( fwrq_transaction_cachep != NULL )
		kmem_cache_destroy(fwrq_transaction_cachep);
}
