/*
 *  linux/fs/xvfat/inode.c
 *  based upon linux/fs/fat/inode.c and linux/fs/xvfat/namei.c
 *
 *  Written 1992,1993 by Werner Almesberger
 *  VFAT extensions by Gordon Chaffee, merged with msdos fs by Henrik Storner
 *  Rewritten for the constant inumbers support by Al Viro
 *
 *  Fixes:
 *
 *  	Max Cohan: Fixed invalid FSINFO offset when info_sector is 0
 *
 *  Copyright 2003 Sony Corporation
 */

#include <linux/module.h>
#include <linux/xvfat_fs.h>
#include <linux/nls.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/bitops.h>
#include <linux/major.h>
#include <linux/blkdev.h>
#include <linux/fs.h>
#include <linux/stat.h>
#include <linux/locks.h>
#include <linux/fat_cvf.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/fwrq.h>

#include <asm/uaccess.h>
#include <asm/unaligned.h>

extern struct cvf_format default_cvf_xvfat;

/* #define FAT_PARANOIA 1 */
#define DEBUG_LEVEL 0
#ifdef FAT_DEBUG
#  define PRINTK(x) printk x
#else
#  define PRINTK(x)
#endif
#if (DEBUG_LEVEL >= 1)
#  define PRINTK1(x) printk x
#else
#  define PRINTK1(x)
#endif

extern int xfat_raw_access(struct super_block *sb, int nr, unsigned new_value,
			  int w);

/*
 * New FAT inode stuff. We do the following:
 *	a) i_ino is constant and has nothing with on-disk location.
 *	b) FAT manages its own cache of directory entries.
 *	c) *This* cache is indexed by on-disk location.
 *	d) inode has an associated directory entry, all right, but
 *		it may be unhashed.
 *	e) currently entries are stored within struct inode. That should
 *		change.
 *	f) we deal with races in the following way:
 *		1. readdir() and lookup() do FAT-dir-cache lookup.
 *		2. rename() unhashes the F-d-c entry and rehashes it in
 *			a new place.
 *		3. unlink() and rmdir() unhash F-d-c entry.
 *		4. fat_write_inode() checks whether the thing is unhashed.
 *			If it is we silently return. If it isn't we do bread(),
 *			check if the location is still valid and retry if it
 *			isn't. Otherwise we do changes.
 *		5. Spinlock is used to protect hash/unhash/location check/lookup
 *		6. fat_clear_inode() unhashes the F-d-c entry.
 *		7. lookup() and readdir() do igrab() if they find a F-d-c entry
 *			and consider negative result as cache miss.
 */

#define FAT_HASH_BITS	8
#define FAT_HASH_SIZE	(1UL << FAT_HASH_BITS)
#define FAT_HASH_MASK	(FAT_HASH_SIZE-1)
static struct list_head xfat_inode_hashtable[FAT_HASH_SIZE];
spinlock_t xfat_inode_lock = SPIN_LOCK_UNLOCKED;

void xfat_hash_init(void)
{
	int i;
	for(i = 0; i < FAT_HASH_SIZE; i++) {
		INIT_LIST_HEAD(&xfat_inode_hashtable[i]);
	}
}

static inline unsigned long xfat_hash(struct super_block *sb, int i_pos)
{
	unsigned long tmp = (unsigned long)i_pos | (unsigned long) sb;
	tmp = tmp + (tmp >> FAT_HASH_BITS) + (tmp >> FAT_HASH_BITS * 2);
	return tmp & FAT_HASH_MASK;
}

void xfat_attach(struct inode *inode, int i_pos)
{
	spin_lock(&xfat_inode_lock);
	MSDOS_I(inode)->i_location = i_pos;
	list_add(&MSDOS_I(inode)->i_fat_hash,
		xfat_inode_hashtable + xfat_hash(inode->i_sb, i_pos));
	spin_unlock(&xfat_inode_lock);
}

void xfat_detach(struct inode *inode)
{
	spin_lock(&xfat_inode_lock);
	MSDOS_I(inode)->i_location = 0;
	list_del(&MSDOS_I(inode)->i_fat_hash);
	INIT_LIST_HEAD(&MSDOS_I(inode)->i_fat_hash);
	spin_unlock(&xfat_inode_lock);
}

struct inode *xfat_iget(struct super_block *sb, int i_pos)
{
	struct list_head *p = xfat_inode_hashtable + xfat_hash(sb, i_pos);
	struct list_head *walk;
	struct msdos_inode_info *i;
	struct inode *inode = NULL;

	spin_lock(&xfat_inode_lock);
	list_for_each(walk, p) {
		i = list_entry(walk, struct msdos_inode_info, i_fat_hash);
		if (i->i_fat_inode->i_sb != sb)
			continue;
		if (i->i_location != i_pos)
			continue;
		inode = igrab(i->i_fat_inode);
		if (inode)
			break;
	}
	spin_unlock(&xfat_inode_lock);
	return inode;
}

static int xfat_fill_inode(struct inode *inode, struct msdos_dir_entry *de);

struct inode *xfat_build_inode(struct super_block *sb,
				struct msdos_dir_entry *de, int ino, int *res)
{
	struct inode *inode;
	*res = 0;
	inode = xfat_iget(sb, ino);
	if (inode)
		goto out;
	inode = new_inode(sb);
	*res = -ENOMEM;
	if (!inode)
		goto out;
	*res = 0;
	inode->i_ino = iunique(sb, MSDOS_ROOT_INO);
	*res = xfat_fill_inode(inode, de);
	if (*res < 0) {
		iput(inode);
		inode = ERR_PTR(*res);
		goto out;
	}
	xfat_attach(inode, ino);
	insert_inode_hash(inode);
out:
	return inode;
}

void xfat_delete_inode(struct inode *inode)
{
	fwrq_transaction_t *ts;

	vfat_printk("xfat_delete_inode[%d:%s]: inode=%08x i_size=%d\n",
		    current->pid, current->comm,
		    (unsigned)inode, (int)inode->i_size);
	if (!is_bad_inode(inode)) {
		lock_kernel();
		if (MSDOS_I(inode)->i_transaction){
			ts = MSDOS_I(inode)->i_transaction;
			MSDOS_I(inode)->i_transaction = NULL;
		} else {
			ts = fwrq_transaction_get(inode->i_dev, FWRQ_TS_LOCK);
		}
		inode->i_size = 0;
		xfat_truncate_ts(inode,ts);
		unlock_kernel();
		if( fwrq_transaction_commit_to_queue_and_flush(ts, 1) == -1 )
			printk("Warning: unsafety writing order (xfat_delete_inode: inode=%08x)\n", (unsigned)inode);
	}
	clear_inode(inode);
}

void xfat_clear_inode(struct inode *inode)
{
	vfat_printk2("xfat_clear_inode: inode=%08x i_size=%d\n",
		    inode, inode->i_size);
	if (is_bad_inode(inode))
		return;
	lock_kernel();
	spin_lock(&xfat_inode_lock);
	xfat_cache_inval_inode(inode);
	list_del(&MSDOS_I(inode)->i_fat_hash);
	spin_unlock(&xfat_inode_lock);
	unlock_kernel();
}

void xfat_put_super(struct super_block *sb)
{
	unsigned cn;

	vfat_printk("xfat_put_super: 0\n");
#ifdef USE_CVF_never
	if (MSDOS_SB(sb)->cvf_format->cvf_version) {
		dec_cvf_format_use_count_by_version(MSDOS_SB(sb)->cvf_format->cvf_version);
		MSDOS_SB(sb)->cvf_format->unmount_cvf(sb);
	}
#endif

	if (fwrq_check_blkdev_media(sb->s_dev, 0) == 0){
		vfat_printk("xfat_put_super: fwrq_check_blkdev_media() == 0\n");
		if (MSDOS_SB(sb)->fat_bits == 32 && MSDOS_SB(sb)->free_clusters != -1)
			xfat_clusters_flush(sb);

		fwrq_flush_queue(sb->s_dev, 1);

		if (MSDOS_SB(sb)->options.clnshutbit &&
		    (MSDOS_SB(sb)->clnshutbit & 1) &&
		    !(sb->s_flags & MS_RDONLY)) {
			if (MSDOS_SB(sb)->fat_bits == 32) {
				cn = xfat_raw_access(sb,1,0,0);
				if( cn == 0 )
					goto set_clnshutbit_err;

				if ( !(cn & 0x08000000) ){
					if( xfat_raw_access(sb,1, cn | 0x08000000,1) == -1 )
						goto set_clnshutbit_err;
				}

			}
			if (MSDOS_SB(sb)->fat_bits == 16) {
				cn = xfat_raw_access(sb,1,0,0);
				if( cn == 0 )
					goto set_clnshutbit_err;
				if ( !(cn & 0x8000) ) {
					if( xfat_raw_access(sb,1, cn | 0x8000, 1) == -1 )
						goto set_clnshutbit_err;
				}

			}
			goto set_clnshutbit_ok;
			set_clnshutbit_err:
			printk("FAT: fail to set ClnShutBit\n");
			set_clnshutbit_ok:
		}
	} else {
		vfat_printk("xfat_put_super: fwrq_check_blkdev_media() == 1\n");
		printk("FAT: Disk change detected on device %s\n",
		       bdevname(sb->s_dev));
		fwrq_revalidate_blkdev_media(sb->s_dev);
	}
	xfat_cache_inval_dev(sb->s_dev);
	set_blocksize (sb->s_dev,BLOCK_SIZE);
	if (MSDOS_SB(sb)->nls_disk) {
		unload_nls(MSDOS_SB(sb)->nls_disk);
		MSDOS_SB(sb)->nls_disk = NULL;
		MSDOS_SB(sb)->options.codepage = 0;
		MSDOS_SB(sb)->options.cp_suffix[0] = '\0';
	}
	if (MSDOS_SB(sb)->nls_io) {
		unload_nls(MSDOS_SB(sb)->nls_io);
		MSDOS_SB(sb)->nls_io = NULL;
	}
	/*
	 * Note: the iocharset option might have been specified
	 * without enabling nls_io, so check for it here.
	 */
	if (MSDOS_SB(sb)->options.iocharset) {
		kfree(MSDOS_SB(sb)->options.iocharset);
		MSDOS_SB(sb)->options.iocharset = NULL;
	}
	fwrq_finish_queue(sb->s_dev);
}

static int simple_getbool(char *s, int *setval)
{
	if (s) {
		if (!strcmp(s,"1") || !strcmp(s,"yes") || !strcmp(s,"true")) {
			*setval = 1;
		} else if (!strcmp(s,"0") || !strcmp(s,"no") || !strcmp(s,"false")) {
			*setval = 0;
		} else {
			return 0;
		}
	} else {
		*setval = 1;
	}
	return 1;
}

static int parse_options(char *options,int *fat, int *debug,
			 struct fat_mount_options *opts,
			 char *cvf_format, char *cvf_options)
{
	char *this_char,*value,save,*savep;
	char *p;
	int ret = 1, len, val, r;

	opts->name_check = 'n';
	opts->conversion = 'b';
	opts->fs_uid = current->uid;
	opts->fs_gid = current->gid;
	opts->fs_umask = current->fs->umask;
	opts->quiet = opts->sys_immutable = opts->dotsOK = opts->showexec = 0;
	opts->codepage = 0;
	opts->cp_suffix[0] = '\0';
	opts->nocase = 0;
	opts->shortname = 0;
	opts->utf8 = 0;
	opts->iocharset = NULL;
 	opts->clnshutbit = 0;
 	opts->old_ioc16 = 0;
#if defined(CONFIG_DTV)
	opts->ignore_crtime = 1;
#else
	opts->ignore_crtime = 0;
#endif
	xfat_tzset(&opts->timezone, NULL);
	*debug = *fat = 0;

	if (!options)
		goto out;
	save = 0;
	savep = NULL;
	for (this_char = strtok(options,","); this_char;
	     this_char = strtok(NULL,",")) {
		if ((value = strchr(this_char,'=')) != NULL) {
			save = *value;
			savep = value;
			*value++ = 0;
		}
		if (!strcmp(this_char,"check") && value) {
			if (value[0] && !value[1] && strchr("rns",*value))
				opts->name_check = *value;
			else if (!strcmp(value,"relaxed"))
				opts->name_check = 'r';
			else if (!strcmp(value,"normal"))
				opts->name_check = 'n';
			else if (!strcmp(value,"strict"))
				opts->name_check = 's';
			else ret = 0;
		}
		else if (!strcmp(this_char,"conv") && value) {
			if (value[0] && !value[1] && strchr("bta",*value))
				opts->conversion = *value;
			else if (!strcmp(value,"binary"))
				opts->conversion = 'b';
			else if (!strcmp(value,"text"))
				opts->conversion = 't';
			else if (!strcmp(value,"auto"))
				opts->conversion = 'a';
			else ret = 0;
		}
		else if (!strcmp(this_char,"dots")) {
			opts->dotsOK = 1;
		}
		else if (!strcmp(this_char,"nocase")) {
			opts->nocase = 1;
		}
		else if (!strcmp(this_char,"nodots")) {
			opts->dotsOK = 0;
		}
		else if (!strcmp(this_char,"showexec")) {
			opts->showexec = 1;
		}
		else if (!strcmp(this_char,"dotsOK") && value) {
			if (!strcmp(value,"yes")) opts->dotsOK = 1;
			else if (!strcmp(value,"no")) opts->dotsOK = 0;
			else ret = 0;
		}
		else if (!strcmp(this_char,"uid")) {
			if (!value || !*value) ret = 0;
			else {
				opts->fs_uid = simple_strtoul(value,&value,0);
				if (*value) ret = 0;
			}
		}
		else if (!strcmp(this_char,"gid")) {
			if (!value || !*value) ret= 0;
			else {
				opts->fs_gid = simple_strtoul(value,&value,0);
				if (*value) ret = 0;
			}
		}
		else if (!strcmp(this_char,"umask")) {
			if (!value || !*value) ret = 0;
			else {
				opts->fs_umask = simple_strtoul(value,&value,8);
				if (*value) ret = 0;
			}
		}
		else if (!strcmp(this_char,"debug")) {
			if (value) ret = 0;
			else *debug = 1;
		}
		else if (!strcmp(this_char,"fat")) {
			if (!value || !*value) ret = 0;
			else {
				*fat = simple_strtoul(value,&value,0);
				if (*value || (*fat != 12 && *fat != 16 &&
					       *fat != 32)) 
					ret = 0;
			}
		}
		else if (!strcmp(this_char,"quiet")) {
			if (value) ret = 0;
			else opts->quiet = 1;
		}
		else if (!strcmp(this_char,"blocksize")) {
			printk("FAT: blocksize option is obsolete, "
			       "not supported now\n");
		}
		else if (!strcmp(this_char,"sys_immutable")) {
			if (value) ret = 0;
			else opts->sys_immutable = 1;
		}
		else if (!strcmp(this_char,"codepage") && value) {
			opts->codepage = simple_strtoul(value,&value,0);
			if (*value) {
				if (strlen(value) < sizeof(opts->cp_suffix))
					strcpy(opts->cp_suffix, value);
				else
					ret = 0;
			}
			if (ret)
				printk ("MSDOS FS: Using codepage %d%s\n",
					opts->codepage, opts->cp_suffix);
		}
		else if (!strcmp(this_char,"iocharset") && value) {
			p = value;
			while (*value && *value != ',')
				value++;
			len = value - p;
			if (len) {
				char *buffer;

				if (opts->iocharset != NULL) {
					kfree(opts->iocharset);
					opts->iocharset = NULL;
				}
				buffer = kmalloc(len + 1, GFP_KERNEL);
				if (buffer != NULL) {
					opts->iocharset = buffer;
					memcpy(buffer, p, len);
					buffer[len] = 0;
					printk("MSDOS FS: IO charset %s\n", buffer);
				} else
					ret = 0;
			}
		}
#ifdef USE_CVF_never
		else if (!strcmp(this_char,"cvf_format")) {
			if (!value)
				return 0;
			strncpy(cvf_format,value,20);
		}
		else if (!strcmp(this_char,"cvf_options")) {
			if (!value)
				return 0;
			strncpy(cvf_options,value,100);
		}
#endif
		else if (!strcmp(this_char,"timezone")) {
			if (!value)
				return 0;
			if( xfat_tzset(&opts->timezone, value) < 0 ) {
				printk("XVFAT: fail to parse mount option timezone=%s\n", value);
				printk("XVFAT: using default timezone\n");
				xfat_tzset(&opts->timezone, NULL);
			}
		}
		else if (!strcmp(this_char,"clnshutbit")) {
			r = simple_getbool(value, &val);
			if (ret) opts->clnshutbit = val;
		}
		else if (!strcmp(this_char, "old_ioc16")) {
			ret = simple_getbool(value, &val);
			if (ret)  {
				printk("XVFAT: old_ioc16 %d->%d\n",
				       opts->old_ioc16, val);
				opts->old_ioc16 = val;
			}

		}
		else if (!strcmp(this_char, "ignore_crtime")) {
			ret = simple_getbool(value, &val);
			if (ret)  {
				printk("XVFAT: ignore_crtime %d->%d\n",
				       opts->ignore_crtime, val);
				opts->ignore_crtime = val;
			}

		}

		if (this_char != options) *(this_char-1) = ',';
		if (value) *savep = save;
		if (ret == 0)
			break;
	}
out:
	return ret;
}

static int xfat_read_root(struct inode *inode)
{
	struct super_block *sb = inode->i_sb;
	struct msdos_sb_info *sbi = MSDOS_SB(sb);
	int nr;

	INIT_LIST_HEAD(&MSDOS_I(inode)->i_fat_hash);
	MSDOS_I(inode)->i_location = 0;
	MSDOS_I(inode)->i_fat_inode = inode;
	inode->i_uid = sbi->options.fs_uid;
	inode->i_gid = sbi->options.fs_gid;
	inode->i_version = ++event;
	inode->i_generation = 0;
	inode->i_mode = (S_IRWXUGO & ~sbi->options.fs_umask) | S_IFDIR;
	inode->i_op = sbi->dir_ops;
	inode->i_fop = &xfat_dir_operations;
	if (sbi->fat_bits == 32) {
		MSDOS_I(inode)->i_start = sbi->root_cluster;
		if ((nr = MSDOS_I(inode)->i_start) != 0) {
			while (nr != -1) {
				inode->i_size += 1 << sbi->cluster_bits;
				if (!(nr = xfat_access(sb, nr, -1))) {
					printk("Directory %ld: bad FAT\n",
					       inode->i_ino);
					return -EIO;
				}
				if(nr == -2) { /* FES patch */
					printk("Device access error\n");
					return -EIO;
				}
			}
		}
	} else {
		MSDOS_I(inode)->i_start = 0;
		inode->i_size = sbi->dir_entries * sizeof(struct msdos_dir_entry);
	}
	inode->i_blksize = 1 << sbi->cluster_bits;
	inode->i_blocks = ((inode->i_size + inode->i_blksize - 1)
			   & ~(inode->i_blksize - 1)) / 512;
	MSDOS_I(inode)->i_logstart = 0;
	MSDOS_I(inode)->mmu_private = inode->i_size;

	MSDOS_I(inode)->i_attrs = 0;
	inode->i_mtime = inode->i_atime = inode->i_ctime = 0;
	MSDOS_I(inode)->i_ctime_ms = 0;
	inode->i_nlink = xfat_subdirs(inode)+2;
	return 0;
}

/*
 * a FAT file handle with fhtype 3 is
 *  0/  i_ino - for fast, reliable lookup if still in the cache
 *  1/  i_generation - to see if i_ino is still valid
 *          bit 0 == 0 iff directory
 *  2/  i_location - if ino has changed, but still in cache
 *  3/  i_logstart - to semi-verify inode found at i_location
 *  4/  parent->i_logstart - maybe used to hunt for the file on disc
 *
 */
struct dentry *xfat_fh_to_dentry(struct super_block *sb, __u32 *fh,
				int len, int fhtype, int parent)
{
	struct inode *inode = NULL;
	struct list_head *lp;
	struct dentry *result;

	if (fhtype != 3)
		return NULL;
	if (len < 5)
		return NULL;
	if (parent)
		return NULL; /* We cannot find the parent,
				It better just *be* there */

	inode = iget(sb, fh[0]);
	if (!inode || is_bad_inode(inode) ||
	    inode->i_generation != fh[1]) {
		if (inode) iput(inode);
		inode = NULL;
	}
	if (!inode) {
		/* try 2 - see if i_location is in F-d-c
		 * require i_logstart to be the same
		 * Will fail if you truncate and then re-write
		 */

		inode = xfat_iget(sb, fh[2]);
		if (inode && MSDOS_I(inode)->i_logstart != fh[3]) {
			iput(inode);
			inode = NULL;
		}
	}
	if (!inode) {
		/* For now, do nothing
		 * What we could do is:
		 * follow the file starting at fh[4], and record
		 * the ".." entry, and the name of the fh[2] entry.
		 * The follow the ".." file finding the next step up.
		 * This way we build a path to the root of
		 * the tree. If this works, we lookup the path and so
		 * get this inode into the cache.
		 * Finally try the fat_iget lookup again
		 * If that fails, then weare totally out of luck
		 * But all that is for another day
		 */
	}
	if (!inode)
		return ERR_PTR(-ESTALE);

	
	/* now to find a dentry.
	 * If possible, get a well-connected one
	 *
	 * Given the way that we found the inode, it *MUST* be
	 * well-connected, but it is easiest to just copy the
	 * code.
	 */
	spin_lock(&dcache_lock);
	for (lp = inode->i_dentry.next; lp != &inode->i_dentry ; lp=lp->next) {
		result = list_entry(lp,struct dentry, d_alias);
		if (! (result->d_flags & DCACHE_NFSD_DISCONNECTED)) {
			dget_locked(result);
			result->d_vfs_flags |= DCACHE_REFERENCED;
			spin_unlock(&dcache_lock);
			iput(inode);
			return result;
		}
	}
	spin_unlock(&dcache_lock);
	result = d_alloc_root(inode);
	if (result == NULL) {
		iput(inode);
		return ERR_PTR(-ENOMEM);
	}
	result->d_flags |= DCACHE_NFSD_DISCONNECTED;
	return result;

		
}

int xfat_dentry_to_fh(struct dentry *de, __u32 *fh, int *lenp, int needparent)
{
	int len = *lenp;
	struct inode *inode =  de->d_inode;
	
	if (len < 5)
		return 255; /* no room */
	*lenp = 5;
	fh[0] = inode->i_ino;
	fh[1] = inode->i_generation;
	fh[2] = MSDOS_I(inode)->i_location;
	fh[3] = MSDOS_I(inode)->i_logstart;
	fh[4] = MSDOS_I(de->d_parent->d_inode)->i_logstart;
	return 3;
}

static struct super_operations xfat_sops = { 
	write_inode:	xfat_write_inode,
	delete_inode:	xfat_delete_inode,
	put_super:	xfat_put_super,
	statfs:		xfat_statfs,
	clear_inode:	xfat_clear_inode,

	read_inode:	make_bad_inode,
	fh_to_dentry:	xfat_fh_to_dentry,
	dentry_to_fh:	xfat_dentry_to_fh,
};


/* fixup for IOC-16 */
static int fixup_blkdev(struct super_block *sb)
{
	struct buffer_head *bh;
	unsigned long page;
	static struct buffer_head *bhs[1];
	int rval, retry, nr;

	if (sb->s_flags & MS_RDONLY)
		return 0;
	rval = -1;
	retry = 1;
	if (MSDOS_SB(sb)->fats > 1)
		nr = MSDOS_SB(sb)->fat_start + MSDOS_SB(sb)->fat_length;
	else
		nr = MSDOS_SB(sb)->fat_start;
	if( sb->s_blocksize > PAGE_SIZE )
		return -1;
	if(!(page = __get_free_page(GFP_KERNEL)))
		return -1;

	bh = bread(sb->s_dev, nr, sb->s_blocksize);
	if (bh == NULL) {
		printk("fixup_blkdev: unable to read sector (nr=%d)\n", nr);
		goto out_err;
	}
	memcpy((void *)page, bh->b_data,  sb->s_blocksize);
 write_retry:
	mark_buffer_dirty(bh);
	bhs[0] = bh;
	ll_rw_block(WRITE, 1, bhs);
	wait_on_buffer(bh);
	if( !buffer_uptodate(bh) ) {
		printk("fixup_blkdev: IO error in writing dummy sector (retry=%d)\n", retry);
		if( retry-- > 0 ) {
			memcpy(bh->b_data, (void *)page, sb->s_blocksize);
			goto write_retry;
		}
		brelse(bh);
		goto out_err;
	}
	mark_buffer_uptodate(bh, 0);
	brelse(bh);
	
	bh = bread(sb->s_dev, nr, sb->s_blocksize);
	if (bh == NULL) {
		printk("FAT: unable to read sector (nr=%d)\n", nr);
		goto out_err;
	}
	if( memcmp((void *)page, bh->b_data, sb->s_blocksize) != 0 ){
		printk("fixup_blkdev: dummy write sector failed (retry=%d)\n", retry);
		if( retry-- > 0 ) {
			memcpy(bh->b_data, (void *)page, sb->s_blocksize);
			goto write_retry;
		}
		brelse(bh);
		goto out_err;
	}
	brelse(bh);
	rval = 0;
 out_err:
	free_page(page);
	return rval;
}

/*
 * Read the super block of an MS-DOS FS.
 *
 * Note that this may be called from xvfat_read_super
 * with some fields already initialized.
 */
struct super_block *
xfat_read_super(struct super_block *sb, void *data, int silent,
		struct inode_operations *fs_dir_inode_ops)
{
	struct inode *root_inode;
	struct buffer_head *bh;
	struct fat_boot_sector *b;
	struct msdos_sb_info *sbi = MSDOS_SB(sb);
	char *p;
	int logical_sector_size, hard_blksize, fat_clusters = 0;
	unsigned int total_sectors, rootdir_sectors;
	int fat32, debug, error, fat, cp;
	struct fat_mount_options opts;
	char buf[50];
	int i;
	char cvf_format[21];
	char cvf_options[101];

	cvf_format[0] = '\0';
	cvf_options[0] = '\0';
	sbi->cvf_format = NULL;
	sbi->private_data = NULL;

	sbi->dir_ops = fs_dir_inode_ops;

	sb->s_maxbytes = MAX_NON_LFS;
	sb->s_op = &xfat_sops;

	hard_blksize = get_hardsect_size(sb->s_dev);
	if (!hard_blksize)
		hard_blksize = 512;

	opts.isvfat = sbi->options.isvfat;
	if (!parse_options((char *) data, &fat, &debug, &opts,
			   cvf_format, cvf_options))
		goto out_fail;
	/* N.B. we should parse directly into the sb structure */
	memcpy(&(sbi->options), &opts, sizeof(struct fat_mount_options));

	xfat_cache_init();

	sb->s_blocksize = hard_blksize;
	set_blocksize(sb->s_dev, hard_blksize);
	bh = bread(sb->s_dev, 0, sb->s_blocksize);
	if (bh == NULL) {
		printk("FAT: unable to read boot sector\n");
		goto out_fail;
	}

/*
 * The DOS3 partition size limit is *not* 32M as many people think.  
 * Instead, it is 64K sectors (with the usual sector size being
 * 512 bytes, leading to a 32M limit).
 * 
 * DOS 3 partition managers got around this problem by faking a 
 * larger sector size, ie treating multiple physical sectors as 
 * a single logical sector.
 * 
 * We can accommodate this scheme by adjusting our cluster size,
 * fat_start, and data_start by an appropriate value.
 *
 * (by Drew Eckhardt)
 */


	b = (struct fat_boot_sector *) bh->b_data;
	logical_sector_size =
		CF_LE_W(get_unaligned((unsigned short *) &b->sector_size));
	if (!logical_sector_size
	    || (logical_sector_size & (logical_sector_size - 1))) {
		printk("FAT: bogus logical sector size %d\n",
		       logical_sector_size);
		brelse(bh);
		goto out_invalid;
	}

	sbi->cluster_size = b->cluster_size;
	if (!sbi->cluster_size
	    || (sbi->cluster_size & (sbi->cluster_size - 1))) {
		printk("FAT: bogus cluster size %d\n", sbi->cluster_size);
		brelse(bh);
		goto out_invalid;
	}

	if (logical_sector_size < hard_blksize) {
		printk("FAT: logical sector size too small for device"
		       " (logical sector size = %d)\n", logical_sector_size);
		brelse(bh);
		goto out_invalid;
	}

	sbi->cluster_bits = ffs(logical_sector_size * sbi->cluster_size) - 1;
	sbi->fats = b->fats;
	sbi->fat_start = CF_LE_W(b->reserved);
	sbi->free_clusters = -1; /* Don't know yet */
	if (!b->fat_length && b->fat32_length) {
		struct fat_boot_fsinfo *fsinfo;
		struct buffer_head *fsinfo_bh;
		int fsinfo_block, fsinfo_offset;

		/* Must be FAT32 */
		fat32 = 1;
		sbi->fat_length = CF_LE_L(b->fat32_length);
		sbi->root_cluster = CF_LE_L(b->root_cluster);

		sbi->fsinfo_sector = CF_LE_W(b->info_sector);
		/* MC - if info_sector is 0, don't multiply by 0 */
		if (sbi->fsinfo_sector == 0)
			sbi->fsinfo_sector = 1;

		fsinfo_block =
			(sbi->fsinfo_sector * logical_sector_size) / hard_blksize;
		fsinfo_offset =
			(sbi->fsinfo_sector * logical_sector_size) % hard_blksize;
		fsinfo_bh = bh;
		if (fsinfo_block != 0) {
			fsinfo_bh = bread(sb->s_dev, fsinfo_block, hard_blksize);
			if (fsinfo_bh == NULL) {
				printk("FAT: bread failed, FSINFO block"
				       " (blocknr = %d)\n", fsinfo_block);
				brelse(bh);
				goto out_invalid;
			}
		}
		fsinfo = (struct fat_boot_fsinfo *)&fsinfo_bh->b_data[fsinfo_offset];
		if (!IS_FSINFO(fsinfo)) {
			printk("FAT: Did not find valid FSINFO signature.\n"
			       "Found signature1 0x%x signature2 0x%x sector=%ld.\n",
			       CF_LE_L(fsinfo->signature1),
			       CF_LE_L(fsinfo->signature2),
			       sbi->fsinfo_sector);
		} else {
#if CONFIG_XVFAT_IGNORE_FAT32_FREECLUSTER
			sbi->free_clusters = -1;/* mark as don't know yet */
#else
			sbi->free_clusters = CF_LE_L(fsinfo->free_clusters);
#endif
		}

		if (fsinfo_block != 0)
			brelse(fsinfo_bh);
	} else {
		fat32 = 0;
		sbi->fat_length = CF_LE_W(b->fat_length);
		sbi->root_cluster = 0;
		sbi->free_clusters = -1; /* Don't know yet */
	}

	sbi->dir_per_block = logical_sector_size / sizeof(struct msdos_dir_entry);
	sbi->dir_per_block_bits = ffs(sbi->dir_per_block) - 1;

	sbi->dir_start = sbi->fat_start + sbi->fats * sbi->fat_length;
	sbi->dir_entries =
		CF_LE_W(get_unaligned((unsigned short *)&b->dir_entries));
	rootdir_sectors = sbi->dir_entries
		* sizeof(struct msdos_dir_entry) / logical_sector_size;
	sbi->data_start = sbi->dir_start + rootdir_sectors;
	total_sectors = CF_LE_W(get_unaligned((unsigned short *)&b->sectors));
	if (total_sectors == 0)
		total_sectors = CF_LE_L(b->total_sect);
	sbi->clusters = (total_sectors - sbi->data_start) / sbi->cluster_size;

	if (sbi->free_clusters != -1 && sbi->free_clusters > sbi->clusters) {
		printk("free cluster wrong, will be fixed later\n");
		sbi->free_clusters = -1;
	}

	error = 0;
	if (!error) {
		sbi->fat_bits = fat32 ? 32 :
			(fat ? fat :
			 (sbi->clusters > MSDOS_FAT12 ? 16 : 12));
		fat_clusters =
			sbi->fat_length * logical_sector_size * 8 / sbi->fat_bits;
		error = !sbi->fats || (sbi->dir_entries & (sbi->dir_per_block - 1))
			|| sbi->clusters + 2 > fat_clusters + MSDOS_MAX_EXTRA
			|| logical_sector_size < 512
			|| PAGE_CACHE_SIZE < logical_sector_size
			|| !b->secs_track || !b->heads;
	}
	brelse(bh);

	if (error)
		goto out_invalid;

	sb->s_blocksize = logical_sector_size;
	sb->s_blocksize_bits = ffs(logical_sector_size) - 1;
	set_blocksize(sb->s_dev, sb->s_blocksize);
	sbi->cvf_format = &default_cvf_xvfat;
#ifdef USE_CVF_never
	if (!strcmp(cvf_format, "none"))
		i = -1;
	else
		i = detect_cvf(sb,cvf_format);
	if (i >= 0)
		error = cvf_formats[i]->mount_cvf(sb, cvf_options);
#else
	i = -1;
#endif
	if (error || debug) {
		/* The MSDOS_CAN_BMAP is obsolete, but left just to remember */
		printk("[MS-DOS FS Rel. 12,FAT %d,check=%c,conv=%c,"
		       "uid=%d,gid=%d,umask=%03o%s]\n",
		       sbi->fat_bits,opts.name_check,
		       opts.conversion,opts.fs_uid,opts.fs_gid,opts.fs_umask,
		       MSDOS_CAN_BMAP(sbi) ? ",bmap" : "");
		printk("[me=0x%x,cs=%d,#f=%d,fs=%d,fl=%ld,ds=%ld,de=%d,data=%ld,"
		       "se=%u,ts=%u,ls=%d,rc=%ld,fc=%u]\n",
		       b->media, sbi->cluster_size, sbi->fats,
		       sbi->fat_start, sbi->fat_length, sbi->dir_start,
		       sbi->dir_entries, sbi->data_start,
		       CF_LE_W(get_unaligned((unsigned short *)&b->sectors)),
		       CF_LE_L(b->total_sect), logical_sector_size,
		       sbi->root_cluster, sbi->free_clusters);
		printk ("hard sector size = %d\n", hard_blksize);
	}
	if (i < 0)
		if (sbi->clusters + 2 > fat_clusters)
			sbi->clusters = fat_clusters - 2;
	if (error)
		goto out_invalid;

	if (sbi->options.old_ioc16) {
		if (fixup_blkdev(sb) < 0)
			goto out_invalid;
	}

	if (sbi->options.clnshutbit) {
		int cn;
		int clnshutbit;
		int retry = 1;

		clnshutbit = 1;
		cn = xfat_raw_access(sb,1,0,0);
		if( cn == -1 ) {
			printk("FAT: fail to read FAT\n");
			goto out_invalid;
		}
	retry:
		switch(sbi->fat_bits){
		case 32:
			if( ((cn | 0x0c000000) & 0x0fffffff) < 0x0ffffff8 ) {
				printk("FAT: broken FAT (FAT[1]=%x)\n", cn);
				goto out_invalid;
			}
			clnshutbit = (cn & 0x08000000) ? 1 : 0;
			if( !(sb->s_flags & MS_RDONLY) && clnshutbit ){
				if( xfat_raw_access(sb,1, cn & ~0x08000000,1) == -1 ){
					if( retry-- > 0 )
						goto retry;
					printk("FAT: fail to clear ClnShutBit\n");
					goto out_invalid;
				}
			}
			break;
		case 16:
			if( (cn | 0xc000) < 0x0000fff8 ) {
				printk("FAT: broken FAT (FAT[1]=%x)\n", cn);
				goto out_invalid;
			}
			clnshutbit = (cn & 0x8000) ? 1 : 0;
			if( !(sb->s_flags & MS_RDONLY) && clnshutbit ){
				if( xfat_raw_access(sb,1, cn & ~0x8000 & 0xffff,1) == -1 ){
					if( retry-- > 0 )
						goto retry;
					printk("FAT: fail to clear ClnShutBit\n") ;
					goto out_invalid;
				}
			}
			break;
		case 12:
			if( cn < 0x00000ff8 ) {
				printk("FAT: broken FAT (FAT[1]=%x)\n", cn);
				goto out_invalid;
			}
			break;
		}
		if( !clnshutbit )
			printk("FAT: %s-fs: not clean shutdown\n",
			       sbi->fat_bits == 32 ? "FAT32" : "FAT16");
		sbi->clnshutbit = clnshutbit;
	}

	sb->s_magic = MSDOS_SUPER_MAGIC;
	/* set up enough so that it can read an inode */
	init_MUTEX(&sbi->fat_lock);
	sbi->prev_free = 0;

	cp = opts.codepage ? opts.codepage : 437;
	sprintf(buf, "cp%d%s", cp, opts.cp_suffix);
	sbi->nls_disk = load_nls(buf);
	if (! sbi->nls_disk) {
		/* Fail only if explicit charset specified */
		if (opts.codepage != 0)
			goto out_fail;
		sbi->options.codepage = 0; /* already 0?? */
		sbi->nls_disk = load_nls_default();
	}

	sbi->nls_io = NULL;
	if (sbi->options.isvfat && !opts.utf8) {
		p = opts.iocharset ? opts.iocharset : CONFIG_NLS_DEFAULT;
		sbi->nls_io = load_nls(p);
		if (! sbi->nls_io)
			/* Fail only if explicit charset specified */
			if (opts.iocharset)
				goto out_unload_nls;
	}
	if (! sbi->nls_io)
		sbi->nls_io = load_nls_default();

	root_inode = new_inode(sb);
	if (!root_inode)
		goto out_unload_nls;
	root_inode->i_ino = MSDOS_ROOT_INO;
	error = xfat_read_root(root_inode);
	if (error < 0)
		goto out_no_root;

	insert_inode_hash(root_inode);
	sb->s_root = d_alloc_root(root_inode);
	if (!sb->s_root)
		goto out_no_root;
#ifdef USE_CVF_never
	if(i >= 0) {
		sbi->cvf_format = cvf_formats[i];
		++cvf_format_use_count[i];
	}
#endif
	return sb;

out_no_root:
	printk("FAT: get root inode failed\n");
	iput(root_inode);
	unload_nls(sbi->nls_io);
out_unload_nls:
	unload_nls(sbi->nls_disk);
	goto out_fail;
out_invalid:
	if (!silent) {
		printk("VFS: Can't find a valid FAT filesystem on dev %s.\n",
			kdevname(sb->s_dev));
	}
out_fail:
	if (opts.iocharset) {
		printk("FAT: freeing iocharset=%s\n", opts.iocharset);
		kfree(opts.iocharset);
	}
	if(sbi->private_data)
		kfree(sbi->private_data);
	sbi->private_data = NULL;
 
	return NULL;
}

int xfat_statfs(struct super_block *sb,struct statfs *buf)
{
	long free;
	unsigned long nr;
	int ret = 0; /* FES patch */
	
	if (MSDOS_SB(sb)->cvf_format &&
	    MSDOS_SB(sb)->cvf_format->cvf_statfs)
		return MSDOS_SB(sb)->cvf_format->cvf_statfs(sb,buf,
						sizeof(struct statfs));
	  
	lock_xfat(sb);
	if (MSDOS_SB(sb)->free_clusters != -1)
		free = MSDOS_SB(sb)->free_clusters;
	else {
		free = 0;
		for (nr = 2; nr < MSDOS_SB(sb)->clusters + 2; nr++) { 
			if (!(ret = xfat_access(sb,nr,-1))) free++;
			if (ret < -1) { /* FES patch */
				printk("%s:fat access error\n", __FUNCTION__);
				free = 0;
				break;
			}
		}
		if (0 <= ret)
			MSDOS_SB(sb)->free_clusters = free;
	}
	unlock_xfat(sb);
	buf->f_type = sb->s_magic;
	buf->f_bsize = 1 << MSDOS_SB(sb)->cluster_bits;
	buf->f_blocks = MSDOS_SB(sb)->clusters;
	buf->f_bfree = free;
	buf->f_bavail = free;
	buf->f_namelen = MSDOS_SB(sb)->options.isvfat ? 260 : 12;
	if (ret < -1)
		return -EIO;
	else
		return 0;
}

static int is_exec(char *extension)
{
	char *exe_extensions = "EXECOMBAT", *walk;

	for (walk = exe_extensions; *walk; walk += 3)
		if (!strncmp(extension, walk, 3))
			return 1;
	return 0;
}

static int xfat_writepage(struct page *page)
{
	if (fwrq_check_blkdev_media(page->mapping->host->i_dev, 0) == 1) {
		SetPageError(page);
		ClearPageUptodate(page);
		UnlockPage(page);
		return -EIO;
	}
	return block_write_full_page(page,xfat_get_block);
}
static int xfat_readpage(struct file *file, struct page *page)
{
	if (fwrq_check_blkdev_media(page->mapping->host->i_dev, 0) == 1) {
		SetPageError(page);
		ClearPageUptodate(page);
		UnlockPage(page);
		return -EIO;
	}
	return block_read_full_page(page,xfat_get_block);
}
static int xfat_prepare_write(struct file *file, struct page *page, unsigned from, unsigned to)
{
	if (fwrq_check_blkdev_media(page->mapping->host->i_dev, 0) == 1)
		return -EIO;
	return cont_prepare_write(page,from,to,xfat_get_block,
		&MSDOS_I(page->mapping->host)->mmu_private);
}
static int _xfat_bmap(struct address_space *mapping, long block)
{
	return generic_block_bmap(mapping,block,xfat_get_block);
}
static int xfat_commit_write(struct file *file, struct page *page,
		unsigned from, unsigned to)
{
	struct inode *inode = page->mapping->host;
	fwrq_transaction_t *ts;
	loff_t i_size;
	int rval;

	i_size = inode->i_size;
	rval = generic_commit_write(file, page, from, to);
	ts = MSDOS_I(inode)->i_transaction;
	if (i_size < inode->i_size)
		fwrq_transaction_add_tail_inode(ts, inode);
	fwrq_transaction_fix(ts);
	return rval;
}

static struct address_space_operations xfat_aops = {
	readpage: xfat_readpage,
	writepage: xfat_writepage,
	sync_page: block_sync_page,
	prepare_write: xfat_prepare_write,
	commit_write: xfat_commit_write,
	bmap: _xfat_bmap
};

/* doesn't deal with root inode */
static int xfat_fill_inode(struct inode *inode, struct msdos_dir_entry *de)
{
	struct super_block *sb = inode->i_sb;
	struct msdos_sb_info *sbi = MSDOS_SB(sb);
	int nr;

	vfat_printk2("xfat_fill_inode: inode=%08x start=%d starthi=%d\n",
		    inode, de->start, de->starthi);
	INIT_LIST_HEAD(&MSDOS_I(inode)->i_fat_hash);
	MSDOS_I(inode)->i_location = 0;
	MSDOS_I(inode)->i_fat_inode = inode;
	inode->i_uid = sbi->options.fs_uid;
	inode->i_gid = sbi->options.fs_gid;
	inode->i_version = ++event;
	inode->i_generation = CURRENT_TIME;
	
	if ((de->attr & ATTR_DIR) && !IS_FREE(de->name)) {
		inode->i_generation &= ~1;
		inode->i_mode = MSDOS_MKMODE(de->attr,S_IRWXUGO &
		    ~sbi->options.fs_umask) | S_IFDIR;
		inode->i_op = sbi->dir_ops;
		inode->i_fop = &xfat_dir_operations;

		MSDOS_I(inode)->i_start = CF_LE_W(de->start);
		if (sbi->fat_bits == 32) {
			MSDOS_I(inode)->i_start |=
				(CF_LE_W(de->starthi) << 16);
		}
		MSDOS_I(inode)->i_logstart = MSDOS_I(inode)->i_start;
		inode->i_nlink = xfat_subdirs(inode);
		    /* includes .., compensating for "self" */
#ifdef DEBUG
		if (!inode->i_nlink) {
			printk("directory %d: i_nlink == 0\n",inode->i_ino);
			inode->i_nlink = 1;
		}
#endif
		if ((nr = MSDOS_I(inode)->i_start) != 0)
			while (nr != -1) {
				inode->i_size += 1 << sbi->cluster_bits;
				if (!(nr = xfat_access(sb, nr, -1))) {
					printk("Directory %ld: bad FAT\n",
					    inode->i_ino);
					return -EIO;
				}
				if(nr == -2) { /* FES patch */
					printk("Device access error\n");
					return -EIO;
				}
			}
		MSDOS_I(inode)->mmu_private = inode->i_size;
	} else { /* not a directory */
		inode->i_generation |= 1;
		inode->i_mode = MSDOS_MKMODE(de->attr,
		    ((sbi->options.showexec &&
		       !is_exec(de->ext))
		    	? S_IRUGO|S_IWUGO : S_IRWXUGO)
		    & ~sbi->options.fs_umask) | S_IFREG;
		MSDOS_I(inode)->i_start = CF_LE_W(de->start);
		if (sbi->fat_bits == 32) {
			MSDOS_I(inode)->i_start |=
				(CF_LE_W(de->starthi) << 16);
		}
		MSDOS_I(inode)->i_logstart = MSDOS_I(inode)->i_start;
		inode->i_size = CF_LE_L(de->size);
	        inode->i_op = &xfat_file_inode_operations;
	        inode->i_fop = &xfat_file_operations;
		inode->i_mapping->a_ops = &xfat_aops;
		MSDOS_I(inode)->mmu_private = inode->i_size;
	}
	if(de->attr & ATTR_SYS)
		if (sbi->options.sys_immutable)
			inode->i_flags |= S_IMMUTABLE;
	MSDOS_I(inode)->i_attrs = de->attr & ATTR_UNUSED;
	/* this is as close to the truth as we can get ... */
	inode->i_blksize = 1 << sbi->cluster_bits;
	inode->i_blocks = ((inode->i_size + inode->i_blksize - 1)
			   & ~(inode->i_blksize - 1)) / 512;
	inode->i_mtime = inode->i_atime =
		xfat_date_dos2unix(&sbi->options.timezone,
				   CF_LE_W(de->time),CF_LE_W(de->date));
	inode->i_ctime =
		(MSDOS_SB(sb)->options.isvfat && !MSDOS_SB(sb)->options.ignore_crtime)
		? xfat_date_dos2unix(&sbi->options.timezone,
				     CF_LE_W(de->ctime),CF_LE_W(de->cdate))
		: inode->i_mtime;
	MSDOS_I(inode)->i_ctime_ms = de->ctime_ms;
	return 0;
}

void fwrq_add_queue_inode(struct inode *inode)
{
	fwrq_printk("fwrq_add_queue_inode[%d:%s]: inode=%08x\n",
		    current->pid, current->comm, (unsigned)inode);
	if( MSDOS_I(inode)->i_data != 0 ){
		panic("fwrq_add_queue_inode: MSDOS_I(inode)->i_data != 0 (%d)\n",
		      MSDOS_I(inode)->i_data);
	}
	mark_inode_dirty(inode);
	if( !(inode->i_state & I_FREEING) )
		write_inode_now(inode, 0);
}

void fwrq_transaction_add_tail_inode(fwrq_transaction_t *transaction,
				     struct inode *inode)
{
	fwrq_transaction_t *keep;

	fwrq_printk("fwrq_transaction_add_tail_inode[%d:%s]: ts=%08x inode=%08x\n",
		    current->pid, current->comm,
		    (unsigned)transaction, (unsigned)inode);
	if( MSDOS_I(inode)->i_data != 0 ){
		panic("fwrq_transaction_add_tail_inode: MSDOS_I(inode)->i_data != 0 (%d)\n",
		      MSDOS_I(inode)->i_data);
	}
	keep = MSDOS_I(inode)->i_transaction;
	MSDOS_I(inode)->i_data = -1;
	MSDOS_I(inode)->i_transaction = transaction;

	mark_inode_dirty(inode);
	if( !(inode->i_state & I_FREEING) )
		write_inode_now(inode, 0);

	if( MSDOS_I(inode)->i_data != -1 ){
		panic("fwrq_transaction_add_tail_inode: MSDOS_I(inode)->i_data != -1 (%d)\n",
		      MSDOS_I(inode)->i_data);
	}
	MSDOS_I(inode)->i_transaction = keep;
	MSDOS_I(inode)->i_data = 0;
}

void xfat_write_inode(struct inode *inode, int wait)
{
	struct super_block *sb = inode->i_sb;
	struct buffer_head *bh;
	struct msdos_dir_entry *raw_entry;
	unsigned int i_pos;

retry:
	i_pos = MSDOS_I(inode)->i_location;
	if (inode->i_ino == MSDOS_ROOT_INO || !i_pos) {
		return;
	}
	lock_kernel();
	if (!(bh = xfat_bread(sb, i_pos >> MSDOS_SB(sb)->dir_per_block_bits))) {
		printk("dev = %s, ino = %d\n", kdevname(inode->i_dev), i_pos);
		xfat_fs_panic(sb, "msdos_write_inode: unable to read i-node block");
		unlock_kernel();
		return;
	}
	spin_lock(&xfat_inode_lock);
	if (i_pos != MSDOS_I(inode)->i_location) {
		spin_unlock(&xfat_inode_lock);
		xfat_brelse(sb, bh);
		unlock_kernel();
		goto retry;
	}

	raw_entry = &((struct msdos_dir_entry *) (bh->b_data))
	    [i_pos & (MSDOS_SB(sb)->dir_per_block - 1)];
	if (S_ISDIR(inode->i_mode)) {
		raw_entry->attr = ATTR_DIR;
		raw_entry->size = 0;
	}
	else {
		raw_entry->attr = ATTR_NONE;
		raw_entry->size = CT_LE_L(inode->i_size);
	}
	raw_entry->attr |= MSDOS_MKATTR(inode->i_mode) |
	    MSDOS_I(inode)->i_attrs;
	raw_entry->start = CT_LE_W(MSDOS_I(inode)->i_logstart);
	raw_entry->starthi = CT_LE_W(MSDOS_I(inode)->i_logstart >> 16);
	xfat_date_unix2dos(&MSDOS_SB(sb)->options.timezone,
			   inode->i_mtime,&raw_entry->time,&raw_entry->date);
	raw_entry->time = CT_LE_W(raw_entry->time);
	raw_entry->date = CT_LE_W(raw_entry->date);
	if (MSDOS_SB(sb)->options.isvfat && !MSDOS_SB(sb)->options.ignore_crtime) {
		xfat_date_unix2dos(&MSDOS_SB(sb)->options.timezone,
				   inode->i_ctime,&raw_entry->ctime,
				   &raw_entry->cdate);
		raw_entry->ctime_ms = MSDOS_I(inode)->i_ctime_ms;
		raw_entry->ctime = CT_LE_W(raw_entry->ctime);
		raw_entry->cdate = CT_LE_W(raw_entry->cdate);
	}
	spin_unlock(&xfat_inode_lock);
	if( MSDOS_I(inode)->i_data == -1 ){
		fwrq_transaction_add_tail(MSDOS_I(inode)->i_transaction, bh);
	}else{
		fwrq_add_queue(bh);
	}
#ifdef NEVER
	if (wait) {
		ll_rw_block (WRITE, 1, &bh);
		wait_on_buffer (bh);
		if (buffer_req(bh) && !buffer_uptodate(bh)) {
			printk ("IO error syncing FAT inode ["
				"%s:%08lx]\n",
				bdevname(inode->i_dev), inode->i_ino);
		}
	}
#endif
	xfat_brelse(sb, bh);
	unlock_kernel();
}


int xfat_notify_change(struct dentry * dentry, struct iattr * attr)
{
	struct super_block *sb = dentry->d_sb;
	struct inode *inode = dentry->d_inode;
	int error;

	vfat_printk("xfat_notify_change[%d:%s]: inode=%08x, attr=%08x\n",
		    current->pid, current->comm,
		    (unsigned)inode, (unsigned)attr);
	if (fwrq_check_blkdev_media(inode->i_dev, 0) == 1)
		return -EIO;
	/* FAT cannot truncate to a longer file */
	if (attr->ia_valid & ATTR_SIZE) {
		if (attr->ia_size > inode->i_size)
			return -EPERM;
	}

	error = inode_change_ok(inode, attr);
	if (error)
		return MSDOS_SB(sb)->options.quiet ? 0 : error;

	if (((attr->ia_valid & ATTR_UID) && 
	     (attr->ia_uid != MSDOS_SB(sb)->options.fs_uid)) ||
	    ((attr->ia_valid & ATTR_GID) && 
	     (attr->ia_gid != MSDOS_SB(sb)->options.fs_gid)) ||
	    ((attr->ia_valid & ATTR_MODE) &&
	     (attr->ia_mode & ~MSDOS_VALID_MODE)))
		error = -EPERM;

	if (error)
		return MSDOS_SB(sb)->options.quiet ? 0 : error;

	error = inode_setattr(inode, attr);
	if (error)
		return error;

	if (S_ISDIR(inode->i_mode))
		inode->i_mode |= S_IXUGO;

	inode->i_mode = ((inode->i_mode & S_IFMT) | ((((inode->i_mode & S_IRWXU
	    & ~MSDOS_SB(sb)->options.fs_umask) | S_IRUSR) >> 6)*S_IXUGO)) &
	    ~MSDOS_SB(sb)->options.fs_umask;
	fsync_dev(inode->i_dev);
	fwrq_flush_queue(inode->i_dev, 1);
	return 0;
}
MODULE_LICENSE("GPL");
