/*
 *  linux/fs/xvfat/namei.c
 *  based upon linux/fs/vfat/namei.c
 *
 *  Written 1992,1993 by Werner Almesberger
 *
 *  Windows95/Windows NT compatible extended MSDOS filesystem
 *    by Gordon Chaffee Copyright (C) 1995.  Send bug reports for the
 *    VFAT filesystem to <chaffee@cs.berkeley.edu>.  Specify
 *    what file operation caused you trouble and if you can duplicate
 *    the problem, send a script that demonstrates it.
 *
 *  Short name translation 1999, 2001 by Wolfram Pienkoss <wp@bszh.de>
 *
 *  Support Multibyte character and cleanup by
 *  				OGAWA Hirofumi <hirofumi@mail.parknet.co.jp>
 *  Copyright 2003 Sony Corporation
 */

#include <linux/module.h>

#include <linux/sched.h>
#include <linux/xvfat_fs.h>
#include <linux/nls.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/ctype.h>
#include <linux/stat.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/ctype.h>
#include <linux/rdchk.h>
#include <asm/uaccess.h>

#define DEBUG_LEVEL 0
#if (DEBUG_LEVEL >= 1)
#  define PRINTK1(x) printk x
#else
#  define PRINTK1(x)
#endif
#if (DEBUG_LEVEL >= 2)
#  define PRINTK2(x) printk x
#else
#  define PRINTK2(x)
#endif
#if (DEBUG_LEVEL >= 3)
#  define PRINTK3(x) printk x
#else
#  define PRINTK3(x)
#endif

#define SAFETY_UNLINK

spinlock_t xvfat_cmpu_lock = SPIN_LOCK_UNLOCKED;
static wchar_t uname_buf1[260],uname_buf2[260];

static int xvfat_hashi(struct dentry *parent, struct qstr *qstr);
static int xvfat_hash(struct dentry *parent, struct qstr *qstr);
static int xvfat_cmpi(struct dentry *dentry, struct qstr *a, struct qstr *b);
static int xvfat_cmp(struct dentry *dentry, struct qstr *a, struct qstr *b);
static int xvfat_cmpiu(struct dentry *dentry, struct qstr *a, struct qstr *b);
static int xvfat_cmpu(struct dentry *dentry, struct qstr *a, struct qstr *b);
static int xvfat_revalidate(struct dentry *dentry, int);
static int xlate_to_uni(const char *name, int len, char *outname,
			int *longlen, int *outlen, int escape, int utf8, 
			struct nls_table *nls);

static struct dentry_operations xvfat_dentry_ops[8] = {
	{
		d_hash:		xvfat_hashi,
		d_compare:	xvfat_cmpi,
	},
	{
		d_revalidate:	xvfat_revalidate,
		d_hash:		xvfat_hashi,
		d_compare:	xvfat_cmpi,
	},
	{
		d_hash:		xvfat_hash,
		d_compare:	xvfat_cmp,
	},
	{
		d_revalidate:	xvfat_revalidate,
		d_hash:		xvfat_hash,
		d_compare:	xvfat_cmp,
	},
	{
		d_hash:		xvfat_hashi,
		d_compare:	xvfat_cmpiu,
	},
	{
		d_revalidate:	xvfat_revalidate,
		d_hash:		xvfat_hashi,
		d_compare:	xvfat_cmpiu,
	},
	{
		d_hash:		xvfat_hash,
		d_compare:	xvfat_cmpu,
	},
	{
		d_revalidate:	xvfat_revalidate,
		d_hash:		xvfat_hash,
		d_compare:	xvfat_cmpu,
	}
};

typedef struct _ms_erase_parm_tab_t {
	int total_sector; /* total sector number in this partition  */
	int erase_size;   /* flash erase size in sector             */
	int erase_offset; /* first aligned sector# from top of this */
                          /* partition                              */
	char * name;      /* friendly name                          */
} ms_erase_parm_tab_t;

static ms_erase_parm_tab_t ms_erase_tab[] = 
{
	{ 0x001dae8f, 256, 143, "1GB PRO"},
	{ 0x000ea217, 256,  93, "512MB PRO"},
	{ 0x00076f5d, 128,  93, "256MB PRO"},
	{ 0x0003dddF,  32,  31, "128MB"},
	{ 0x0001eed9,  32,  25, "64MB"},
	{ 0x0000f76d,  32,  13, "32MB"},
	{ 0x00007b67,  32,   7, "16MB"},
	{ 0x00003da7,  16,   7, "8MB"},
	{ 0x00001ea5,  16,   5, "4MB"},

	{          0,   0,  0, "TERMINATOR"}
};

static int xvfat_revalidate(struct dentry *dentry, int flags)
{
	PRINTK1(("xvfat_revalidate: %s\n", dentry->d_name.name));
	vfat_printk2("xvfat_revalidate: %08x(%s)\n", (unsigned)dentry,
		    dentry->d_name.name);
	spin_lock(&dcache_lock);
	if (dentry->d_time == dentry->d_parent->d_inode->i_version) {
		spin_unlock(&dcache_lock);
		return 1;
	}
	spin_unlock(&dcache_lock);
	return 0;
}

static int simple_getbool(char *s, int *setval)
{
	if (s) {
		if (!strcmp(s,"1") || !strcmp(s,"yes") || !strcmp(s,"true")) {
			*setval = 1;
		} else if (!strcmp(s,"0") || !strcmp(s,"no") || !strcmp(s,"false")) {
			*setval = 0;
		} else {
			return 0;
		}
	} else {
		*setval = 1;
	}
	return 1;
}

static int atoi(const char *s)
{
	int i=0;

	while (isdigit(*s))
		i = i*10 + *s++ - '0';
	return i;
}

static int simple_getuint(char *s, int *setval)
{
	if (s) {
		*setval = atoi(s);
		return 1;
	} else {
		return 0;
	}
}

static int parse_options(char *options,	struct fat_mount_options *opts)
{
	char *this_char,*value,save,*savep;
	int ret, val;

	opts->unicode_xlate = opts->posixfs = 0;
	opts->numtail = 1;
	opts->utf8 = 0;
	opts->shortname = VFAT_SFN_DISPLAY_LOWER | VFAT_SFN_CREATE_WIN95;
	/* for backward compatible */
	if (opts->nocase) {
		opts->nocase = 0;
		opts->shortname = VFAT_SFN_DISPLAY_WIN95
		  		| VFAT_SFN_CREATE_WIN95;
	}

	opts->compare_unicode = 1;
	opts->fwrq_block = 0;
	opts->fwrq_block_offset = 0;

	if (!options) return 1;
	save = 0;
	savep = NULL;
	ret = 1;
	for (this_char = strtok(options,","); this_char; this_char = strtok(NULL,",")) {
		if ((value = strchr(this_char,'=')) != NULL) {
			save = *value;
			savep = value;
			*value++ = 0;
		}
		if (!strcmp(this_char,"utf8")) {
			ret = simple_getbool(value, &val);
			if (ret) opts->utf8 = val;
		} else if (!strcmp(this_char,"uni_xlate")) {
			ret = simple_getbool(value, &val);
			if (ret) opts->unicode_xlate = val;
		} else if (!strcmp(this_char,"posix")) {
			ret = simple_getbool(value, &val);
			if (ret) opts->posixfs = val;
		} else if (!strcmp(this_char,"nonumtail")) {
			ret = simple_getbool(value, &val);
			if (ret) {
				opts->numtail = !val;
			}
		} else if (!strcmp(this_char, "shortname")) {
			if (!strcmp(value, "lower"))
				opts->shortname = VFAT_SFN_DISPLAY_LOWER
						| VFAT_SFN_CREATE_WIN95;
			else if (!strcmp(value, "win95"))
				opts->shortname = VFAT_SFN_DISPLAY_WIN95
						| VFAT_SFN_CREATE_WIN95;
			else if (!strcmp(value, "winnt"))
				opts->shortname = VFAT_SFN_DISPLAY_WINNT
						| VFAT_SFN_CREATE_WINNT;
			else if (!strcmp(value, "mixed"))
				opts->shortname = VFAT_SFN_DISPLAY_WINNT
						| VFAT_SFN_CREATE_WIN95;
			else
				ret = 0;
		} else if (!strcmp(this_char,"comp_uni")) {
			ret = simple_getbool(value, &val);
			if (ret) opts->compare_unicode = val;
		} else if (!strcmp(this_char, "fblksiz")) {
			ret = simple_getuint(value, &val);
			printk("XVFAT: fblksize %d->%d\n",
			       opts->fwrq_block, val);
			if (ret) opts->fwrq_block = val;

		} else if (!strcmp(this_char, "fblkoff")) {
			ret = simple_getuint(value, &val);
			printk("XVFAT: fblkoff %d->%d\n",
			       opts->fwrq_block_offset, val);
			if (ret) opts->fwrq_block_offset = val;
		} 
		if (this_char != options)
			*(this_char-1) = ',';
		if (value) {
			*savep = save;
		}
		if (ret == 0) {
			return 0;
		}
	}
	if (opts->unicode_xlate) {
		opts->utf8 = 0;
	}
	return 1;
}

static inline unsigned char
xvfat_tolower(struct nls_table *t, unsigned char c)
{
	unsigned char nc = t->charset2lower[c];

	return nc ? nc : c;
}

static inline unsigned char
xvfat_toupper(struct nls_table *t, unsigned char c)
{
	unsigned char nc = t->charset2upper[c];

	return nc ? nc : c;
}

static int
xvfat_strnicmp(struct nls_table *t, const unsigned char *s1,
					const unsigned char *s2, int len)
{
	while(len--)
		if (xvfat_tolower(t, *s1++) != xvfat_tolower(t, *s2++))
			return 1;

	return 0;
}

static wchar_t xvfat_uni_tolower(wchar_t u)
{
	if (u >= 0x100)
		return u;
	return (wchar_t)tolower((unsigned char)u);
}

int xvfat_uni_strnicmp(wchar_t *u1, wchar_t *u2, int len)
{
	while(len--)
		if (xvfat_uni_tolower(*u1++) != xvfat_uni_tolower(*u2++))
			return 1;
	return 0;
}

/*
 * Compute the hash for the vfat name corresponding to the dentry.
 * Note: if the name is invalid, we leave the hash code unchanged so
 * that the existing dentry can be used. The vfat fs routines will
 * return ENOENT or EINVAL as appropriate.
 */
static int xvfat_hash(struct dentry *dentry, struct qstr *qstr)
{
	const char *name;
	int len;

	len = qstr->len;
	name = qstr->name;
	while (len && name[len-1] == '.')
		len--;

	qstr->hash = full_name_hash(name, len);

	return 0;
}

/*
 * Compute the hash for the vfat name corresponding to the dentry.
 * Note: if the name is invalid, we leave the hash code unchanged so
 * that the existing dentry can be used. The vfat fs routines will
 * return ENOENT or EINVAL as appropriate.
 */
static int xvfat_hashi(struct dentry *dentry, struct qstr *qstr)
{
	struct nls_table *t = MSDOS_SB(dentry->d_inode->i_sb)->nls_io;
	const char *name;
	int len;
	unsigned long hash;

	vfat_printk2("xvfat_hashi: dentry=%08x(%s) qstr=%s(%d)\n",
		     (unsigned)dentry, dentry->d_name.name,
		    qstr->name, qstr->len);
	len = qstr->len;
	name = qstr->name;
	while (len && name[len-1] == '.')
		len--;

	hash = init_name_hash();
	while (len--)
		hash = partial_name_hash(xvfat_tolower(t, *name++), hash);
	qstr->hash = end_name_hash(hash);

	return 0;
}

/*
 * Case insensitive compare of two vfat names.
 */
static int xvfat_cmpi(struct dentry *dentry, struct qstr *a, struct qstr *b)
{
	struct nls_table *t = MSDOS_SB(dentry->d_inode->i_sb)->nls_io;
	int alen, blen;

	vfat_printk2("xvfat_cmpi: dentry=%08x(%s) a=%s(%d) b=%s(%d)",
		    (unsigned)dentry, dentry->d_name.name,
		    a->name, a->len, b->name, b->len);
	/* A filename cannot end in '.' or we treat it like it has none */
	alen = a->len;
	blen = b->len;
	while (alen && a->name[alen-1] == '.')
		alen--;
	while (blen && b->name[blen-1] == '.')
		blen--;
	if (alen == blen) {
		if (xvfat_strnicmp(t, a->name, b->name, alen) == 0){
			vfat_printk2("..match\n");
			return 0;
		}
	}
	vfat_printk2("..unmatch\n");
	return 1;
}

static int xvfat_cmpiu(struct dentry *dentry, struct qstr *a, struct qstr *b)
{
	struct nls_table *t = MSDOS_SB(dentry->d_inode->i_sb)->nls_io;
	int alen, blen, aulen, aunilen, bulen, bunilen;
	int utf8 = MSDOS_SB(dentry->d_inode->i_sb)->options.utf8;
	int uni_xlate = MSDOS_SB(dentry->d_inode->i_sb)->options.unicode_xlate;
	int matched= -1;

	vfat_printk2("xvfat_cmpiu: dentry=%08x(%s) a=%s(%d) b=%s(%d)",
		    (unsigned)dentry, dentry->d_name.name,
		    a->name, a->len, b->name, b->len);
	/* A filename cannot end in '.' or we treat it like it has none */
	alen = a->len;
	blen = b->len;
	while (alen && a->name[alen-1] == '.')
		alen--;
	while (blen && b->name[blen-1] == '.')
		blen--;

	spin_lock(&xvfat_cmpu_lock);
	if (xlate_to_uni(a->name, alen, (char *)uname_buf1, &aulen, &aunilen,
			 uni_xlate, utf8, t) < 0)
		goto end;
	if (xlate_to_uni(b->name, blen, (char *)uname_buf2, &bulen, &bunilen,
			 uni_xlate, utf8, t) < 0)
		goto end;

	if (aulen == bulen) {
		if (xvfat_uni_strnicmp(uname_buf1, uname_buf2, aulen) == 0){
			vfat_printk2("..match\n");
			matched = 0;
			goto end;
		}
	}
	matched = 1;
 end:
	spin_unlock(&xvfat_cmpu_lock);
	vfat_printk2("..unmatch\n");
	return matched;
}

/*
 * Case sensitive compare of two vfat names.
 */
static int xvfat_cmp(struct dentry *dentry, struct qstr *a, struct qstr *b)
{
	int alen, blen;

	/* A filename cannot end in '.' or we treat it like it has none */
	alen = a->len;
	blen = b->len;
	while (alen && a->name[alen-1] == '.')
		alen--;
	while (blen && b->name[blen-1] == '.')
		blen--;
	if (alen == blen) {
		if (strncmp(a->name, b->name, alen) == 0)
			return 0;
	}
	return 1;
}

static int xvfat_cmpu(struct dentry *dentry, struct qstr *a, struct qstr *b)
{
	struct nls_table *t = MSDOS_SB(dentry->d_inode->i_sb)->nls_io;
	int alen, blen, aulen, aunilen, bulen, bunilen;
	int utf8 = MSDOS_SB(dentry->d_inode->i_sb)->options.utf8;
	int uni_xlate = MSDOS_SB(dentry->d_inode->i_sb)->options.unicode_xlate;
	int matched= -1;

	vfat_printk2("xvfat_cmpu: dentry=%08x(%s) a=%s(%d) b=%s(%d)",
		    (unsigned)dentry, dentry->d_name.name,
		    a->name, a->len, b->name, b->len);
	/* A filename cannot end in '.' or we treat it like it has none */
	alen = a->len;
	blen = b->len;
	while (alen && a->name[alen-1] == '.')
		alen--;
	while (blen && b->name[blen-1] == '.')
		blen--;

	spin_lock(&xvfat_cmpu_lock);
	if (xlate_to_uni(a->name, alen, (char *)uname_buf1, &aulen, &aunilen,
			 uni_xlate, utf8, t) < 0)
		goto end;
	if (xlate_to_uni(b->name, blen, (char *)uname_buf2, &bulen, &bunilen,
			 uni_xlate, utf8, t) < 0)
		goto end;

	if (aulen == bulen) {
		if (strncmp((char *)uname_buf1, (char *)uname_buf2,
			    aulen * sizeof(wchar_t)) == 0){
			vfat_printk2("..match\n");
			matched = 0;
			goto end;
		}
	}
	matched = 1;
 end:
	spin_unlock(&xvfat_cmpu_lock);
	vfat_printk2("..unmatch\n");
	return matched;
}

#ifdef DEBUG

static void dump_fat(struct super_block *sb,int start)
{
	printk("[");
	while (start) {
		printk("%d ",start);
		start = xfat_access(sb,start,-1);
		if (!start) {
			printk("ERROR");
			break;
		}
		if (start == -1) break;
		if (start == -2) break; /* FES patch */
	}
	printk("]\n");
}

static void dump_de(struct msdos_dir_entry *de)
{
	int i;
	unsigned char *p = (unsigned char *) de;
	printk("[");

	for (i = 0; i < 32; i++, p++) {
		printk("%02x ", *p);
	}
	printk("]\n");
}

#endif

/* MS-DOS "device special files" */

static const char *reserved3_names[] = {
	"con     ", "prn     ", "nul     ", "aux     ", NULL
};

static const char *reserved4_names[] = {
	"com1    ", "com2    ", "com3    ", "com4    ", "com5    ",
	"com6    ", "com7    ", "com8    ", "com9    ",
	"lpt1    ", "lpt2    ", "lpt3    ", "lpt4    ", "lpt5    ",
	"lpt6    ", "lpt7    ", "lpt8    ", "lpt9    ",
	NULL };


/* Characters that are undesirable in an MS-DOS file name */

static wchar_t bad_chars[] = {
	/*  '*'     '?'     '<'    '>'      '|'     '"'     ':'     '/' */
	0x002A, 0x003F, 0x003C, 0x003E, 0x007C, 0x0022, 0x003A, 0x002F,
	/*  '\' */
	0x005C, 0,
};
#define IS_BADCHAR(uni)	(xvfat_unistrchr(bad_chars, (uni)) != NULL)

static wchar_t replace_chars[] = {
	/*  '['     ']'    ';'     ','     '+'      '=' */
	0x005B, 0x005D, 0x003B, 0x002C, 0x002B, 0x003D, 0,
};
#define IS_REPLACECHAR(uni)	(xvfat_unistrchr(replace_chars, (uni)) != NULL)

static wchar_t skip_chars[] = {
	/*  '.'     ' ' */
	0x002E, 0x0020, 0,
};
#define IS_SKIPCHAR(uni) \
	((wchar_t)(uni) == skip_chars[0] || (wchar_t)(uni) == skip_chars[1])

static inline wchar_t *xvfat_unistrchr(const wchar_t *s, const wchar_t c)
{
	for(; *s != c; ++s)
		if (*s == 0)
			return NULL;
	return (wchar_t *) s;
}

static inline int xvfat_is_used_badchars(const wchar_t *s, int len)
{
	int i;
	
	for (i = 0; i < len; i++)
		if (s[i] < 0x0020 || IS_BADCHAR(s[i]))
			return -EINVAL;
	return 0;
}

/* Checks the validity of a long MS-DOS filename */
/* Returns negative number on error, 0 for a normal
 * return, and 1 for . or .. */

static int xvfat_valid_longname(const char *name, int len, int xlate)
{
	const char **reserved, *walk;
	int baselen;

	if (len && name[len-1] == ' ') return -EINVAL;
	if (len >= 256) return -EINVAL;
 	if (len < 3) return 0;

	for (walk = name; *walk != 0 && *walk != '.'; walk++);
	baselen = walk - name;

	if (baselen == 3) {
		for (reserved = reserved3_names; *reserved; reserved++) {
			if (!strnicmp(name,*reserved,baselen))
				return -EINVAL;
		}
	} else if (baselen == 4) {
		for (reserved = reserved4_names; *reserved; reserved++) {
			if (!strnicmp(name,*reserved,baselen))
				return -EINVAL;
		}
	}
	return 0;
}

static int xvfat_find_form(struct inode *dir,char *name)
{
	struct msdos_dir_entry *de;
	struct buffer_head *bh = NULL;
	int ino,res;

	res=xfat_scan(dir,name,&bh,&de,&ino);
	xfat_brelse(dir->i_sb, bh);
	if (res<0)
		return -ENOENT;
	return 0;
}

/* 
 * 1) Valid characters for the 8.3 format alias are any combination of
 * letters, uppercase alphabets, digits, any of the
 * following special characters:
 *     $ % ' ` - @ { } ~ ! # ( ) & _ ^
 * In this case Longfilename is not stored in disk.
 *
 * WinNT's Extension:
 * File name and extension name is contain uppercase/lowercase
 * only. And it is expressed by CASE_LOWER_BASE and CASE_LOWER_EXT.
 *     
 * 2) File name is 8.3 format, but it contain the uppercase and
 * lowercase char, muliti bytes char, etc. In this case numtail is not
 * added, but Longfilename is stored.
 * 
 * 3) When the one except for the above, or the following special
 * character are contained:
 *        .   [ ] ; , + =
 * numtail is added, and Longfilename must be stored in disk .
 */
struct shortname_info {
	unsigned char lower:1,
		      upper:1,
		      valid:1;
};
#define INIT_SHORTNAME_INFO(x)	do {		\
	(x)->lower = 1;				\
	(x)->upper = 1;				\
	(x)->valid = 1;				\
} while (0);

static inline unsigned char
shortname_info_to_lcase(struct shortname_info *base,
			struct shortname_info *ext)
{
	unsigned char lcase = 0;

	if (base->valid && ext->valid) {
		if (!base->upper && base->lower && (ext->lower || ext->upper))
			lcase |= CASE_LOWER_BASE;
		if (!ext->upper && ext->lower && (base->lower || base->upper))
			lcase |= CASE_LOWER_EXT;
	}

	return lcase;
}

static inline int to_shortname_char(struct nls_table *nls,
				    char *buf, int buf_size, wchar_t *src,
				    struct shortname_info *info)
{
	int len;

	if (IS_SKIPCHAR(*src)) {
		info->valid = 0;
		return 0;
	}
	if (IS_REPLACECHAR(*src)) {
		info->valid = 0;
		buf[0] = '_';
		return 1;
	}
	
	len = nls->uni2char(*src, buf, buf_size);
	if (len <= 0) {
		info->valid = 0;
		buf[0] = '_';
		len = 1;
	} else if (len == 1) {
		unsigned char prev = buf[0];

		if (buf[0] >= 0x7F) {
			info->lower = 0;
			info->upper = 0;
		}

		buf[0] = xvfat_toupper(nls, buf[0]);
		if (isalpha(buf[0])) {
			if (buf[0] == prev)
				info->lower = 0;
			else
				info->upper = 0;
		}
	} else {
		info->lower = 0;
		info->upper = 0;
	}
	
	return len;
}

/*
 * Given a valid longname, create a unique shortname.  Make sure the
 * shortname does not exist
 * Returns negative number on error, 0 for a normal
 * return, and 1 for valid shortname
 */
static int xvfat_create_shortname(struct inode *dir, struct nls_table *nls,
				 wchar_t *uname, int ulen,
				 char *name_res, unsigned char *lcase)
{
	wchar_t *ip, *ext_start, *end, *name_start;
	unsigned char base[9], ext[4], buf[8], *p;
	unsigned char charbuf[NLS_MAX_CHARSET_SIZE];
	int chl, chi;
	int sz = 0, extlen, baselen, i, numtail_baselen, numtail2_baselen;
	int is_shortname;
	struct shortname_info base_info, ext_info;
	unsigned short opt_shortname = MSDOS_SB(dir->i_sb)->options.shortname;

	is_shortname = 1;
	INIT_SHORTNAME_INFO(&base_info);
	INIT_SHORTNAME_INFO(&ext_info);

	/* Now, we need to create a shortname from the long name */
	ext_start = end = &uname[ulen];
	while (--ext_start >= uname) {
		if (*ext_start == 0x002E) { /* is '.' */
			if (ext_start == end - 1) {
				sz = ulen;
				ext_start = NULL;
			}
			break;
		}
	}

	if (ext_start == uname - 1) {
		sz = ulen;
		ext_start = NULL;
	} else if (ext_start) {
		/*
		 * Names which start with a dot could be just
		 * an extension eg. "...test".  In this case Win95
		 * uses the extension as the name and sets no extension.
		 */
		name_start = &uname[0];
		while (name_start < ext_start) {
			if (!IS_SKIPCHAR(*name_start))
				break;
			name_start++;
		}
		if (name_start != ext_start) {
			sz = ext_start - uname;
			ext_start++;
		} else {
			sz = ulen;
			ext_start=NULL;
		}
	}

	numtail_baselen = 6;
	numtail2_baselen = 2;
	for (baselen = i = 0, p = base, ip = uname; i < sz; i++, ip++) {
		chl = to_shortname_char(nls, charbuf, sizeof(charbuf),
					ip, &base_info);
		if (chl == 0)
			continue;

		if (baselen < 2 && (baselen + chl) > 2)
			numtail2_baselen = baselen;
		if (baselen < 6 && (baselen + chl) > 6)
			numtail_baselen = baselen;
		for (chi = 0; chi < chl; chi++){
			*p++ = charbuf[chi];
			baselen++;
			if (baselen >= 8)
				break;
		}
		if (baselen >= 8) {
			if ((chi < chl - 1) || (ip + 1) - uname < sz)
				is_shortname = 0;
			break;
		}
	}
	if (baselen == 0) {
		return -EINVAL;
	}

	extlen = 0;
	if (ext_start) {
		for (p = ext, ip = ext_start; extlen < 3 && ip < end; ip++) {
			chl = to_shortname_char(nls, charbuf, sizeof(charbuf),
						ip, &ext_info);
			if (chl == 0)
				continue;

			if ((extlen + chl) > 3) {
				is_shortname = 0;
				break;
			}
			for (chi = 0; chi < chl; chi++) {
				*p++ = charbuf[chi];
				extlen++;
			}
			if (extlen >= 3) {
				if (ip + 1 != end)
					is_shortname = 0;
				break;
			}
		}
	}
	ext[extlen] = '\0';
	base[baselen] = '\0';

	/* Yes, it can happen. ".\xe5" would do it. */
	if (base[0] == DELETED_FLAG)
		base[0] = 0x05;

	/* OK, at this point we know that base is not longer than 8 symbols,
	 * ext is not longer than 3, base is nonempty, both don't contain
	 * any bad symbols (lowercase transformed to uppercase).
	 */

	memset(name_res, ' ', MSDOS_NAME);
	memcpy(name_res, base, baselen);
	memcpy(name_res + 8, ext, extlen);
	*lcase = 0;
	if (is_shortname && base_info.valid && ext_info.valid) {
		if (xvfat_find_form(dir, name_res) == 0)
			return -EEXIST;

		if (opt_shortname & VFAT_SFN_CREATE_WIN95) {
			return (base_info.upper && ext_info.upper);
		} else if (opt_shortname & VFAT_SFN_CREATE_WINNT) {
			if ((base_info.upper || base_info.lower)
			    && (ext_info.upper || ext_info.lower)) {
				*lcase = shortname_info_to_lcase(&base_info,
								 &ext_info);
				return 1;
			}
			return 0;
		} else {
			BUG();
		}
	}
	
	if (MSDOS_SB(dir->i_sb)->options.numtail == 0)
		if (xvfat_find_form(dir, name_res) < 0)
			return 0;

	/*
	 * Try to find a unique extension.  This used to
	 * iterate through all possibilities sequentially,
	 * but that gave extremely bad performance.  Windows
	 * only tries a few cases before using random
	 * values for part of the base.
	 */

	if (baselen>6) {
		baselen = numtail_baselen;
		name_res[7] = ' ';
	}
	name_res[baselen] = '~';
	for (i = 1; i < 10; i++) {
		name_res[baselen+1] = i + '0';
		if (xvfat_find_form(dir, name_res) < 0)
			return 0;
	}

	i = jiffies & 0xffff;
	sz = (jiffies >> 16) & 0x7;
	if (baselen>2) {
		baselen = numtail2_baselen;
		name_res[7] = ' ';
	}
	name_res[baselen+4] = '~';
	name_res[baselen+5] = '1' + sz;
	while (1) {
		sprintf(buf, "%04X", i);
		memcpy(&name_res[baselen], buf, 4);
		if (xvfat_find_form(dir, name_res) < 0)
			break;
		i -= 11;
	}
	return 0;
}

/* Translate a string, including coded sequences into Unicode */
static int
xlate_to_uni(const char *name, int len, char *outname, int *longlen, int *outlen,
	     int escape, int utf8, struct nls_table *nls)
{
	const unsigned char *ip;
	unsigned char nc;
	char *op;
	unsigned int ec;
	int i, k, fill;
	int charlen;

	if (utf8) {
		*outlen = utf8_mbstowcs((__u16 *) outname, name, PAGE_SIZE);
		if (name[len-1] == '.')
			*outlen-=2;
		op = &outname[*outlen * sizeof(__u16)];
	} else {
		if (name[len-1] == '.') 
			len--;
		if (nls) {
			for (i = 0, ip = name, op = outname, *outlen = 0;
			     i < len && *outlen <= 260; *outlen += 1)
			{
				if (escape && (*ip == ':')) {
					if (i > len - 5)
						return -EINVAL;
					ec = 0;
					for (k = 1; k < 5; k++) {
						nc = ip[k];
						ec <<= 4;
						if (nc >= '0' && nc <= '9') {
							ec |= nc - '0';
							continue;
						}
						if (nc >= 'a' && nc <= 'f') {
							ec |= nc - ('a' - 10);
							continue;
						}
						if (nc >= 'A' && nc <= 'F') {
							ec |= nc - ('A' - 10);
							continue;
						}
						return -EINVAL;
					}
					*op++ = ec & 0xFF;
					*op++ = ec >> 8;
					ip += 5;
					i += 5;
				} else {
					if ((charlen = nls->char2uni(ip, len-i, (wchar_t *)op)) < 0)
						return -EINVAL;
					ip += charlen;
					i += charlen;
					op += 2;
				}
			}
		} else {
			for (i = 0, ip = name, op = outname, *outlen = 0;
			     i < len && *outlen <= 260; i++, *outlen += 1)
			{
				*op++ = *ip++;
				*op++ = 0;
			}
		}
	}
	if (*outlen > 260)
		return -ENAMETOOLONG;

	*longlen = *outlen;
	if (*outlen % 13) {
		*op++ = 0;
		*op++ = 0;
		*outlen += 1;
		if (*outlen % 13) {
			fill = 13 - (*outlen % 13);
			for (i = 0; i < fill; i++) {
				*op++ = 0xff;
				*op++ = 0xff;
			}
			*outlen += fill;
		}
	}

	return 0;
}

int
xvfat_xlate_to_uni(const char *name, int len, char *outname, int *longlen,
		  int *outlen, int escape, int utf8, struct nls_table *nls)
{
	return xlate_to_uni(name, len, outname, longlen, outlen, escape,
				 utf8, nls);
}

static int
xvfat_fill_slots(struct inode *dir, struct msdos_dir_slot *ds, const char *name,
		int len, int *slots, int is_dir, int uni_xlate)
{
	struct nls_table *nls_io, *nls_disk;
	wchar_t *uname;
	struct msdos_dir_slot *ps;
	struct msdos_dir_entry *de;
	unsigned long page;
	unsigned char cksum, lcase;
	char *uniname, msdos_name[MSDOS_NAME];
	int res, utf8, slot, ulen, unilen, i;
	loff_t offset;

	*slots = 0;
	utf8 = MSDOS_SB(dir->i_sb)->options.utf8;
	nls_io = MSDOS_SB(dir->i_sb)->nls_io;
	nls_disk = MSDOS_SB(dir->i_sb)->nls_disk;

	if (name[len-1] == '.')
		len--;
	if(!(page = __get_free_page(GFP_KERNEL)))
		return -ENOMEM;

	uniname = (char *) page;
	res = xlate_to_uni(name, len, uniname, &ulen, &unilen, uni_xlate,
			   utf8, nls_io);
	if (res < 0)
		goto out_free;

	uname = (wchar_t *) page;
	res = xvfat_is_used_badchars(uname, ulen);
	if (res < 0)
		goto out_free;

	res = xvfat_create_shortname(dir, nls_disk, uname, ulen,
				    msdos_name, &lcase);
	if (res < 0)
		goto out_free;
	else if (res == 1) {
		de = (struct msdos_dir_entry *)ds;
		res = 0;
		goto shortname;
	}

	/* build the entry of long file name */
	*slots = unilen / 13;
	for (cksum = i = 0; i < 11; i++) {
		cksum = (((cksum&1)<<7)|((cksum&0xfe)>>1)) + msdos_name[i];
	}
	PRINTK3(("xvfat_fill_slots 3: slots=%d\n",*slots));

	for (ps = ds, slot = *slots; slot > 0; slot--, ps++) {
		ps->id = slot;
		ps->attr = ATTR_EXT;
		ps->reserved = 0;
		ps->alias_checksum = cksum;
		ps->start = 0;
		offset = (slot - 1) * 13;
		fatwchar_to16(ps->name0_4, uname + offset, 5);
		fatwchar_to16(ps->name5_10, uname + offset + 5, 6);
		fatwchar_to16(ps->name11_12, uname + offset + 11, 2);
	}
	ds[0].id |= 0x40;
	de = (struct msdos_dir_entry *) ps;

shortname:
	PRINTK3(("xvfat_fill_slots 9\n"));
	/* build the entry of 8.3 alias name */
	(*slots)++;
	strncpy(de->name, msdos_name, MSDOS_NAME);
	de->attr = is_dir ? ATTR_DIR : ATTR_ARCH;
	de->lcase = lcase;
	de->adate = de->cdate = de->date = 0;
	de->ctime_ms = de->ctime = de->time = 0;
	de->start = 0;
	de->starthi = 0;
	de->size = 0;

out_free:
	free_page(page);
	return res;
}

/* We can't get "." or ".." here - VFS takes care of those cases */

static int xvfat_build_slots(struct inode *dir, const char *name, int len,
			    struct msdos_dir_slot *ds, int *slots, int is_dir)
{
	int res, xlate;

	xlate = MSDOS_SB(dir->i_sb)->options.unicode_xlate;
	res = xvfat_valid_longname(name, len, xlate);
	if (res < 0)
		return res;

	return xvfat_fill_slots(dir, ds, name, len, slots, is_dir, xlate);
}

static int xvfat_add_entry_ts(struct inode *dir,struct qstr* qname,
			       int is_dir, struct vfat_slot_info *sinfo_out,
			       struct buffer_head **bh,
			       struct msdos_dir_entry **de,
			       fwrq_transaction_t *ts)
{
	struct super_block *sb = dir->i_sb;
	struct msdos_dir_slot *dir_slots;
	loff_t offset;
	int slots, slot;
	int res, len;
	struct msdos_dir_entry *dummy_de;
	struct buffer_head *dummy_bh;
	int dummy_ino;
	loff_t dummy;

	vfat_printk("xvfat_add_entry_ts[%d:%s]: dir=%08x, ts=%08x\n",
		    current->pid, current->comm,
		    (unsigned)dir, (unsigned)ts);
	dir_slots = (struct msdos_dir_slot *)
	       kmalloc(sizeof(struct msdos_dir_slot) * MSDOS_SLOTS, GFP_KERNEL);
	if (dir_slots == NULL)
		return -ENOMEM;

	len = qname->len;
	while (len && qname->name[len-1] == '.')
		len--;
	res = xfat_search_long(dir, qname->name, len,
			      (MSDOS_SB(sb)->options.name_check != 's')
			      || !MSDOS_SB(sb)->options.posixfs,
			      &dummy, &dummy);
	if (res > 0) /* found */
		res = -EEXIST;
	if (res)
		goto cleanup;

	res = xvfat_build_slots(dir, qname->name, len,
			       dir_slots, &slots, is_dir);
	if (res < 0)
		goto cleanup;

	/* build the empty directory entry of number of slots */
	offset = xfat_add_entries_ts(dir, slots, &dummy_bh, &dummy_de,
				    &dummy_ino, ts);
	if (offset < 0) {
		res = offset;
		goto cleanup;
	}
	xfat_brelse(sb, dummy_bh);

	/* Now create the new entry */
	*bh = NULL;
	for (slot = 0; slot < slots; slot++) {
		if (xfat_get_entry(dir, &offset, bh, de, &sinfo_out->ino) < 0) {
			res = -EIO;
			goto cleanup;
		}
		memcpy(*de, dir_slots + slot, sizeof(struct msdos_dir_slot));
#define calc_bh_rsector(bh)     ((bh)->b_blocknr * ((bh)->b_size >> 9))
		vfat_printk2("xvfat_add_entry_ts: sec=%d+%d nr=%d\n",
			    calc_bh_rsector(*bh), (*bh)->b_size >> 9,
			    (*bh)->b_blocknr);
		fwrq_transaction_add_tail(ts, *bh);
	}

	res = 0;
	/* update timestamp */
	dir->i_ctime = dir->i_mtime = dir->i_atime = CURRENT_TIME;
	fwrq_add_queue_inode(dir);

	xfat_date_unix2dos(&MSDOS_SB(sb)->options.timezone,
			   dir->i_mtime, &(*de)->time, &(*de)->date);
	(*de)->time = CT_LE_W((*de)->time);
	(*de)->ctime = (*de)->time;
	(*de)->date = CT_LE_W((*de)->date);
	(*de)->adate = (*de)->cdate = (*de)->date;

	fwrq_add_queue(*bh);

	/* slots can't be less than 1 */
	sinfo_out->long_slots = slots - 1;
	sinfo_out->longname_offset =
		offset - sizeof(struct msdos_dir_slot) * slots;

cleanup:
	kfree(dir_slots);
	return res;
}

static int xvfat_find(struct inode *dir,struct qstr* qname,
	struct vfat_slot_info *sinfo, struct buffer_head **last_bh,
	struct msdos_dir_entry **last_de)
{
	struct super_block *sb = dir->i_sb;
	loff_t offset;
	int res,len;

	len = qname->len;
	while (len && qname->name[len-1] == '.') 
		len--;
	res = xfat_search_long(dir, qname->name, len,
			(MSDOS_SB(sb)->options.name_check != 's'),
			&offset,&sinfo->longname_offset);
	if (res>0) {
		sinfo->long_slots = res-1;
		if (xfat_get_entry(dir,&offset,last_bh,last_de,&sinfo->ino)>=0)
			return 0;
		res = -EIO;
	} 
	return res ? res : -ENOENT;
}

struct dentry *xvfat_lookup(struct inode *dir,struct dentry *dentry)
{
	int res;
	struct vfat_slot_info sinfo;
	struct inode *inode;
	struct dentry *alias;
	struct buffer_head *bh = NULL;
	struct msdos_dir_entry *de;
	int table;
	
	PRINTK2(("xvfat_lookup: name=%s, len=%d\n", 
		 dentry->d_name.name, dentry->d_name.len));
	vfat_printk2("xvfat_lookup: %08x name=%s, len=%d\n", 
		    (unsigned)dentry, dentry->d_name.name, dentry->d_name.len);

	if (fwrq_check_blkdev_media(dir->i_dev, 1) == 1)
		return (struct dentry *)-EIO;

	table = (MSDOS_SB(dir->i_sb)->options.name_check == 's') ? 2 : 0;
	if (MSDOS_SB(dir->i_sb)->options.compare_unicode)
		table += 4;
	dentry->d_op = &xvfat_dentry_ops[table];

	inode = NULL;
	res = xvfat_find(dir,&dentry->d_name,&sinfo,&bh,&de);
	if (res < 0) {
		table++;
		goto error;
	}
	inode = xfat_build_inode(dir->i_sb, de, sinfo.ino, &res);
	xfat_brelse(dir->i_sb, bh);
	if (res)
		return ERR_PTR(res);
	alias = d_find_alias(inode);
	if (alias) {
		if (d_invalidate(alias)==0)
			dput(alias);
		else {
			iput(inode);
			return alias;
		}
		
	}
error:
	dentry->d_op = &xvfat_dentry_ops[table];
	dentry->d_time = dentry->d_parent->d_inode->i_version;
	d_add(dentry,inode);
	return NULL;
}

int xvfat_create(struct inode *dir,struct dentry* dentry,int mode)
{
	struct super_block *sb = dir->i_sb;
	struct inode *inode = NULL;
	struct buffer_head *bh = NULL;
	struct msdos_dir_entry *de;
	struct vfat_slot_info sinfo;
	int res;
	fwrq_transaction_t *ts;

	vfat_printk("xvfat_create[%d:%s]: dir=%08x\n",
		    current->pid, current->comm, (unsigned)dir);
	if (fwrq_check_blkdev_media(dir->i_dev, 1) == 1)
		return -EIO;

	ts = fwrq_transaction_get(dir->i_dev, FWRQ_TS_LOCK);
	if(!ts) return -EIO; /* FES patch */
	res = xvfat_add_entry_ts(dir, &dentry->d_name, 0, &sinfo, &bh, &de, ts);
	if( fwrq_transaction_commit_to_queue(ts) == -1 )
		printk("Warning: unsafety writing order (xvfat_create: dir=%08x dentry=%08x)\n", (unsigned)dir, (unsigned)dentry);
	if (res < 0) {
		fwrq_flush_queue(dir->i_dev, 1);
		return res;
	}
	inode = xfat_build_inode(sb, de, sinfo.ino, &res);
	xfat_brelse(sb, bh);
	if (!inode) {
		fwrq_flush_queue(dir->i_dev, 1);
		return res;
	}
	inode->i_mtime = inode->i_atime = inode->i_ctime = CURRENT_TIME;
	fwrq_add_queue_inode(inode);
	inode->i_version = ++event;
	dir->i_version = event;
	dentry->d_time = dentry->d_parent->d_inode->i_version;
	d_instantiate(dentry,inode);
	return 0;
}

static void xvfat_remove_entry_ts(struct inode *dir,
				 struct vfat_slot_info *sinfo,
				 struct buffer_head *bh,
				 struct msdos_dir_entry *de,
				 fwrq_transaction_t *ts)
{
	struct super_block *sb = dir->i_sb;
	loff_t offset;
	int i,ino;

	/* remove the shortname */
	dir->i_mtime = CURRENT_TIME;
	dir->i_atime = CURRENT_TIME;
	dir->i_version = ++event;
	fwrq_add_queue_inode(dir);
	de->name[0] = DELETED_FLAG;
	fwrq_transaction_add_tail(ts, bh);

	/* remove the longname */
	offset = sinfo->longname_offset; de = NULL;
	for (i = sinfo->long_slots; i > 0; --i) {
		if (xfat_get_entry(dir, &offset, &bh, &de, &ino) < 0)
			continue;
		de->name[0] = DELETED_FLAG;
		de->attr = 0;
		fwrq_transaction_add_tail(ts, bh);
	}
	if (bh) xfat_brelse(sb, bh);
}

int xvfat_rmdir(struct inode *dir,struct dentry* dentry)
{
	int res;
	struct vfat_slot_info sinfo;
	struct buffer_head *bh = NULL;
	struct msdos_dir_entry *de;
	fwrq_transaction_t *ts;

	vfat_printk("xvfat_rmdir[%d:%s]: dir=%08x dentry=%08x name=%s\n",
		    current->pid, current->comm,
		    (unsigned)dir, (unsigned)dentry, dentry->d_name.name);

	if (fwrq_check_blkdev_media(dir->i_dev, 1) == 1)
		return -EIO;
	res = xfat_dir_empty(dentry->d_inode);
	if (res)
		return res;

	res = xvfat_find(dir,&dentry->d_name,&sinfo, &bh, &de);
	if (res<0)
		return res;
	dentry->d_inode->i_nlink = 0;
	dentry->d_inode->i_mtime = CURRENT_TIME;
	dentry->d_inode->i_atime = CURRENT_TIME;
	xfat_detach(dentry->d_inode);
	ts = fwrq_transaction_get(dentry->d_inode->i_dev, FWRQ_TS_LOCK);
	if(!ts) return -EIO; /* FES patch */
	if( MSDOS_I(dentry->d_inode)->i_transaction )
		printk("xvfat_rmdir: exist old fwrq transaction\n");
	MSDOS_I(dentry->d_inode)->i_transaction = ts;
	fwrq_transaction_add_tail_inode(ts, dentry->d_inode);
	/* releases bh */
	xvfat_remove_entry_ts(dir,&sinfo,bh,de,ts);
	dir->i_nlink--;
#ifdef SAFETY_UNLINK
	if( fwrq_transaction_commit_to_queue_and_flush(ts, 1) == -1 )
		printk("Warning: unsafety writing order (xfat_delete_inode: inode=%08x)\n", (unsigned)(dentry->d_inode));
	MSDOS_I(dentry->d_inode)->i_transaction = NULL;
	fsync_dev(dentry->d_inode->i_dev);
	fwrq_flush_queue(dentry->d_inode->i_dev, 1);
#else
	if( atomic_read(&dentry->d_count) != 1 ){
		if( fwrq_transaction_commit_to_queue_and_flush(ts, 1) == -1 )
			printk("Warning: unsafety writing order (xfat_delete_inode: inode=%08x)\n", dentry->d_inode);
		MSDOS_I(dentry->d_inode)->i_transaction = NULL;
		fsync_dev(dentry->d_inode->i_dev);
		fwrq_flush_queue(dentry->d_inode->i_dev, 1);
	}
#endif
	return 0;
}

int xvfat_unlink(struct inode *dir, struct dentry* dentry)
{
	int res;
	struct vfat_slot_info sinfo;
	struct buffer_head *bh = NULL;
	struct msdos_dir_entry *de;
	fwrq_transaction_t *ts;

	vfat_printk("xvfat_unlink[%d:%s]: dir=%08x dentry=%08x name=%s\n",
		    current->pid, current->comm, (unsigned)dir,
		    (unsigned)dentry, dentry->d_name.name);
	PRINTK1(("xvfat_unlink: %s\n", dentry->d_name.name));
	if (fwrq_check_blkdev_media(dentry->d_inode->i_dev, 1) == 1)
		return -EIO;

	res = xvfat_find(dir,&dentry->d_name,&sinfo,&bh,&de);
	if (res < 0)
		return res;
	dentry->d_inode->i_nlink = 0;
	dentry->d_inode->i_mtime = CURRENT_TIME;
	dentry->d_inode->i_atime = CURRENT_TIME;
	xfat_detach(dentry->d_inode);
	ts = fwrq_transaction_get(dentry->d_inode->i_dev, FWRQ_TS_LOCK);
	if(!ts) return -EIO; /* FES patch */
	fwrq_flush_queue(dentry->d_inode->i_dev, 1);
	if( MSDOS_I(dentry->d_inode)->i_transaction )
		printk("xvfat_rmdir: exist old fwrq transaction\n");
	MSDOS_I(dentry->d_inode)->i_transaction = ts;
	fwrq_add_queue_inode(dentry->d_inode);
	xvfat_remove_entry_ts(dir,&sinfo,bh,de,ts);
	vfat_printk("xvfat_unlink[%d:%s]: res=%d d_count=%d done.\n",
		    current->pid, current->comm,
		    res, atomic_read(&dentry->d_count));
#ifdef SAFETY_UNLINK
	if( fwrq_transaction_commit_to_queue_and_flush(ts, 1) == -1 )
		printk("Warning: unsafety writing order (xfat_delete_inode: inode=%08x)\n", (unsigned)dentry->d_inode);
	MSDOS_I(dentry->d_inode)->i_transaction = NULL;
	fsync_dev(dentry->d_inode->i_dev);
	fwrq_flush_queue(dentry->d_inode->i_dev, 1);
#else
	if( atomic_read(&dentry->d_count) != 1 ){
		if( fwrq_transaction_commit_to_queue_and_flush(ts, 1) == -1 )
			printk("Warning: unsafety writing order (xfat_delete_inode: inode=%08x)\n", dentry->d_inode);
		MSDOS_I(dentry->d_inode)->i_transaction = NULL;
		fsync_dev(dentry->d_inode->i_dev);
		fwrq_flush_queue(dentry->d_inode->i_dev, 1);
	}
#endif

	return res;
}


int xvfat_mkdir(struct inode *dir,struct dentry* dentry,int mode)
{
	struct super_block *sb = dir->i_sb;
	struct inode *inode = NULL;
	struct vfat_slot_info sinfo;
	struct buffer_head *bh = NULL;
	struct msdos_dir_entry *de;
	int res;
	fwrq_transaction_t *ts, *ts2;

	vfat_printk("xvfat_mkdir[%d:%s]: dir=%08x dentry=%08x name=%s\n",
		    current->pid, current->comm, (unsigned)dir,
		    (unsigned)dentry, dentry->d_name.name);
	if (fwrq_check_blkdev_media(dir->i_dev, 1) == 1)
		return -EIO;
	ts2 = fwrq_transaction_get(dir->i_dev, FWRQ_TS_LOCK);
	if(!ts2) return -EIO; /* FES patch */
	res = xvfat_add_entry_ts(dir, &dentry->d_name, 1, &sinfo, &bh, &de, ts2);
	if (res < 0){
		if( fwrq_transaction_commit_to_queue(ts2) == -1 )
			printk("Warning: unsafety writing order (xvfat_mkdir: dir=%08x dentry=%08x name=%s) (0)\n", (unsigned)dir, (unsigned)dentry, dentry->d_name.name);
		fwrq_flush_queue(dir->i_dev, 1);
		return res;
	}
	inode = xfat_build_inode(sb, de, sinfo.ino, &res);
	if (!inode){
		if( fwrq_transaction_commit_to_queue_and_flush(ts2, 1) == -1 )
			printk("Warning: unsafety writing order (xvfat_mkdir: dir=%08x dentry=%08x name=%s)(1)\n", (unsigned)dir, (unsigned)dentry, dentry->d_name.name);
		goto out;
	}
	inode->i_mtime = inode->i_atime = inode->i_ctime = CURRENT_TIME;
	fwrq_add_queue_inode(inode);
	inode->i_version = ++event;
	dir->i_version = event;
	dir->i_nlink++;
	inode->i_nlink = 2; /* no need to mark them dirty */
	ts = fwrq_transaction_get(dir->i_dev, FWRQ_TS_NOLOCK);
	if(!ts) goto mkdir_failed; /* FES patch */
	res = xfat_new_dir_ts(inode, dir, 1, ts);
	fwrq_transaction_append1(ts, ts2);
	if (res < 0)
		goto mkdir_failed;
	if( fwrq_transaction_commit_to_queue_and_flush(ts, 1) == -1 )
		printk("Warning: unsafety writing order (xvfat_mkdir: dir=%08x dentry=%08x name=%s)(2)\n", (unsigned)dir, (unsigned)dentry, dentry->d_name.name);
	fsync_dev(dir->i_dev);
	fwrq_flush_queue(dir->i_dev, 1);
	dentry->d_time = dentry->d_parent->d_inode->i_version;
	d_instantiate(dentry,inode);
out:
	xfat_brelse(sb, bh);
	return res;

mkdir_failed:
	inode->i_nlink = 0;
	inode->i_mtime = CURRENT_TIME;
	inode->i_atime = CURRENT_TIME;
	xfat_detach(inode);
	fwrq_transaction_add_tail_inode(ts, inode);
	/* releases bh */
	xvfat_remove_entry_ts(dir,&sinfo,bh,de,ts);
	if( fwrq_transaction_commit_to_queue(ts) == -1 )
		printk("Warning: unsafety writing order (xvfat_mkdir: dir=%08x dentry=%08x)(2)\n", (unsigned)dir, (unsigned)dentry);
	/* fsync_dev(dir->i_dev); */
	fwrq_flush_queue(dir->i_dev, 1);
	iput(inode);
	dir->i_nlink--;
	return res;
}
 
int xvfat_rename(struct inode *old_dir,struct dentry *old_dentry,
		struct inode *new_dir,struct dentry *new_dentry)
{
	struct super_block *sb = old_dir->i_sb;
	struct buffer_head *old_bh,*new_bh,*dotdot_bh;
	struct msdos_dir_entry *old_de,*new_de,*dotdot_de;
	int dotdot_ino;
	struct inode *old_inode, *new_inode;
	int res, is_dir;
	struct vfat_slot_info old_sinfo,sinfo;
	fwrq_transaction_t *ts, *ts2;

	vfat_printk("xvfat_rename[%d:%s]: (old: %08x, %08x, %s), (new: %08x, %08x, %s)\n",
		    current->pid, current->comm,
		    (unsigned)old_dir, (unsigned)old_dentry,
		    old_dentry->d_name.name,
		    (unsigned)new_dir, (unsigned)new_dentry,
		    new_dentry->d_name.name);
	if (fwrq_check_blkdev_media(old_dir->i_dev, 1) == 1)
		return -EIO;
	ts2 = NULL;
	old_bh = new_bh = dotdot_bh = NULL;
	old_inode = old_dentry->d_inode;
	new_inode = new_dentry->d_inode;
	res = xvfat_find(old_dir,&old_dentry->d_name,&old_sinfo,&old_bh,&old_de);
	PRINTK3(("xvfat_rename 2\n"));
	if (res < 0) goto rename_done;

	is_dir = S_ISDIR(old_inode->i_mode);

	if (is_dir && (res = xfat_scan(old_inode,MSDOS_DOTDOT,&dotdot_bh,
				&dotdot_de,&dotdot_ino)) < 0)
		goto rename_done;

	if (new_dentry->d_inode) {
		vfat_printk("xvfat_rename: exist new_dentry->d_inode=%d\n",
			    (unsigned)new_dentry->d_inode);
		res = xvfat_find(new_dir,&new_dentry->d_name,&sinfo,&new_bh,
				&new_de);
		if (res < 0 || MSDOS_I(new_inode)->i_location != sinfo.ino) {
			/* WTF??? Cry and fail. */
			printk(KERN_WARNING "xvfat_rename: fs corrupted\n");
			goto rename_done;
		}

		if (is_dir) {
			res = xfat_dir_empty(new_inode);
			if (res)
				goto rename_done;
		}
		xfat_detach(new_inode);
	} else {
		vfat_printk("xvfat_rename: add_entry new\n");
		ts2 = fwrq_transaction_get(old_dir->i_dev, FWRQ_TS_LOCK);
		if(!ts2) return -EIO; /* FES patch */
		res = xvfat_add_entry_ts(new_dir,&new_dentry->d_name,is_dir,
					&sinfo, &new_bh, &new_de, ts2);
		if (res < 0){
			if( fwrq_transaction_commit_to_queue(ts2) == -1 )
				printk("Warning: unsafety writing order (xvfat_rename: %08x %08x)(0)\n", (unsigned)old_dir, (unsigned)new_dir);
			/* fsync_dev(new_dir->i_dev); */
			fwrq_flush_queue(new_dir->i_dev, 1);
			goto rename_done;
		}
	}

	new_dir->i_version = ++event;

	/* releases old_bh */
	ts = fwrq_transaction_get(old_dir->i_dev,
				  ts2 ? FWRQ_TS_NOLOCK : FWRQ_TS_LOCK);
	if(!ts) return -EIO; /* FES patch */
	xvfat_remove_entry_ts(old_dir,&old_sinfo,old_bh,old_de, ts);
	if( ts2 )
		fwrq_transaction_concat(ts, ts2);
	old_bh=NULL;
	xfat_detach(old_inode);
	xfat_attach(old_inode, sinfo.ino);
	fwrq_add_queue_inode(old_inode);

	old_dir->i_version = ++event;
	old_dir->i_ctime = old_dir->i_mtime = CURRENT_TIME;
	fwrq_add_queue_inode(old_dir);
	if (new_inode) {
		new_inode->i_nlink--;
		new_inode->i_ctime=CURRENT_TIME;
	}

	if (is_dir) {
		int start = MSDOS_I(new_dir)->i_logstart;
		dotdot_de->start = CT_LE_W(start);
		dotdot_de->starthi = CT_LE_W(start>>16);
		fwrq_add_queue(dotdot_bh);
		old_dir->i_nlink--;
		if (new_inode) {
			new_inode->i_nlink--;
		} else {
			new_dir->i_nlink++;
			fwrq_add_queue_inode(new_dir);
		}
	}
	if( fwrq_transaction_commit_to_queue_and_flush(ts, 1) == -1 )
		printk("Warning: unsafety writing order (xvfat_rename: %08x %08x)(0)\n", (unsigned)old_dir, (unsigned)new_dir);
	fsync_dev(new_dir->i_dev);
	fwrq_flush_queue(new_dir->i_dev, 1);

rename_done:
	xfat_brelse(sb, dotdot_bh);
	xfat_brelse(sb, old_bh);
	xfat_brelse(sb, new_bh);
	return res;

}


/* Public inode operations for the VFAT fs */
struct inode_operations xvfat_dir_inode_operations = {
	create:		xvfat_create,
	lookup:		xvfat_lookup,
	unlink:		xvfat_unlink,
	mkdir:		xvfat_mkdir,
	rmdir:		xvfat_rmdir,
	rename:		xvfat_rename,
	setattr:	xfat_notify_change,
};

struct super_block *xvfat_read_super(struct super_block *sb,void *data,
				    int silent)
{
	struct super_block *res;
	int table;

	MSDOS_SB(sb)->options.isvfat = 1;

	res = xfat_read_super(sb, data, silent, &xvfat_dir_inode_operations);
	if (res == NULL)
		return NULL;

	if (parse_options((char *) data, &(MSDOS_SB(sb)->options))) {
		MSDOS_SB(sb)->options.dotsOK = 0;
		if (MSDOS_SB(sb)->options.posixfs) {
			MSDOS_SB(sb)->options.name_check = 's';
		}
		if (MSDOS_SB(sb)->options.name_check != 's') {
			table = 0;
		} else {
			table = 2;
		}
		if (MSDOS_SB(sb)->options.compare_unicode)
			table += 4;
		sb->s_root->d_op = &xvfat_dentry_ops[table];
	}

#ifdef NEVER
	printk("XVFAT: s_blksize=%ld cluster_size=%d clusters=%ld\n",
	       sb->s_blocksize,
	       MSDOS_SB(sb)->cluster_size, MSDOS_SB(sb)->clusters);
	printk("         fat=%d,%ld fblksiz=%d fblkoff=%d\n",
	       MSDOS_SB(sb)->fat_start, MSDOS_SB(sb)->fat_length,
	       MSDOS_SB(sb)->options.fwrq_block,
	       MSDOS_SB(sb)->options.fwrq_block_offset);
	printk("         dir_start=%ld, dir_entries=%d, data_start=%ld\n",
	       MSDOS_SB(sb)->dir_start, MSDOS_SB(sb)->dir_entries,
	       MSDOS_SB(sb)->data_start);
	printk("         root_cluster=%ld, fat_bits=%d\n",
	       MSDOS_SB(sb)->root_cluster, MSDOS_SB(sb)->fat_bits);
#endif
	{
		unsigned block, block_offset;
		ms_erase_parm_tab_t *tab;
		unsigned int total_sectors;

		total_sectors = MSDOS_SB(sb)->cluster_size * MSDOS_SB(sb)->clusters + 
			MSDOS_SB(sb)->data_start;
		if ( MSDOS_SB(sb)->options.fwrq_block == 0 &&
		     MSDOS_SB(sb)->options.fwrq_block_offset == 0) {
			for ( tab = ms_erase_tab; tab->total_sector; tab++) {
				if (total_sectors == tab->total_sector) {
					MSDOS_SB(sb)->options.fwrq_block = tab->erase_size;
					MSDOS_SB(sb)->options.fwrq_block_offset = tab->erase_offset;
					printk("XVFAT: Using %s MemoryStick parameter\n",
					       tab->name);
				}
			}
		}

		block = MSDOS_SB(sb)->options.fwrq_block;
		block_offset = MSDOS_SB(sb)->options.fwrq_block_offset;

		if ( block == 0 && block_offset == 0 ){
			if( sb->s_blocksize == 512 &&
			    MSDOS_SB(sb)->cluster_size == 64 &&
			    MSDOS_SB(sb)->data_start == 93 ){
				block = 128;
				block_offset = 93;
			}
		}

		if ( block== 0)
			block = MSDOS_SB(sb)->cluster_size;
		if ( block_offset == 0)
			block_offset = MSDOS_SB(sb)->data_start %
				MSDOS_SB(sb)->cluster_size;
		MSDOS_SB(sb)->options.fwrq_block = block;
		MSDOS_SB(sb)->options.fwrq_block_offset = block_offset;

		printk("XVFAT: fblksiz=%d fblkoff=%d cluster_size=%d\n",
		       MSDOS_SB(sb)->options.fwrq_block,
		       MSDOS_SB(sb)->options.fwrq_block_offset,
		       MSDOS_SB(sb)->cluster_size);
		fwrq_init_queue(sb->s_dev,
				MSDOS_SB(sb)->options.fwrq_block,
				MSDOS_SB(sb)->options.fwrq_block_offset,
				MSDOS_SB(sb)->cluster_size);
	}
	return res;
}

/* ioctl */
int
xvfat_ioctl_iochar_to_uni16(struct inode *inode, struct xvfat_lname_box *u_lbox)
{
	struct xvfat_lname_box lbox;
	unsigned long page1, page2;
	wchar_t *uname;
	unsigned char *fname;
	int ulen, ulen_dent;
	int rval;

	if (copy_from_user(&lbox, u_lbox, sizeof(lbox)))
		return -EFAULT;

	lbox.error = XVFAT_LNAME_ERR_NONE;
	if( lbox.io_name_len > NAME_MAX || lbox.io_name_len <= 0 )
		return -EINVAL;
	if( lbox.u_name_alloc_size <= 0 )
		return -EINVAL;

	if(!(page1 = __get_free_page(GFP_KERNEL)))
		return -ENOMEM;
	if(!(page2 = __get_free_page(GFP_KERNEL))){
		free_page(page1);
		return -ENOMEM;
	}

	uname = (wchar_t *) page1;
	fname = (char *) page2;
	ulen = 0;

	if (copy_from_user(fname, lbox.io_name, lbox.io_name_len)){
		rval = -EFAULT;
		goto out_free;
	}
	rval = xlate_to_uni(fname, lbox.io_name_len, (char *)uname,
			    &ulen, &ulen_dent, 0, 0, 
			    MSDOS_SB(inode->i_sb)->nls_io);
	if (rval < 0) {
		if (rval == -EINVAL) {
			lbox.error = XVFAT_LNAME_ERR_BADCHAR;
			rval = 0;
		}
		uname[0] = 0;
		ulen = 1;
		goto out_free;
	}

	if (ulen == 0) {
		lbox.error = XVFAT_LNAME_ERR_BADCHAR;
		uname[0] = 0;
	}

	rval = xvfat_is_used_badchars(uname, ulen);
	if (rval < 0){
		rval = 0;
		lbox.error = XVFAT_LNAME_ERR_BADCHAR;
	}

	ulen++;		/* add a NUL character */

	if ( ulen > lbox.u_name_alloc_size ){
		ulen = lbox.u_name_alloc_size;
		uname[ulen-1] = 0;
		lbox.error = XVFAT_LNAME_ERR_TOOLONG;
	}
out_free:
	if( rval >= 0 ) {
		lbox.u_name_len = ulen - 1;
		rval = 0;
	}
	if (copy_to_user(lbox.u_name, uname, ulen * 2))
		rval = -EFAULT;
	if (copy_to_user(u_lbox, &lbox, sizeof(lbox)))
		rval = -EFAULT;

	free_page(page1);
	free_page(page2);
	return rval;
}

int
xvfat_ioctl_get_minfo(struct inode *inode, struct xvfat_media_info *u_minfo)
{
	struct xvfat_media_info minfo;
	int rval;

	minfo.capability = 0;

	rval = rdchk_get_disk_change(inode->i_dev, 1);
	if( rval < 0 ){
		minfo.change = 0;
	} else {
		minfo.change = rval;
		minfo.capability |= XVFAT_MINFO_CFLAG_CHANGE;
	}

	rval = rdchk_get_disk_insert(inode->i_dev, 0);
	if( rval < 0 ){
		minfo.insert = 0;
	} else {
		minfo.insert = rval;
		minfo.capability |= XVFAT_MINFO_CFLAG_INSERT;
	}

	rval = rdchk_get_disk_state_change_count(inode->i_dev, &minfo.state_change_count, 0);
	if( rval < 0 ){
		minfo.state_change_count = 0;
	}

	if (copy_to_user(u_minfo, &minfo, sizeof(minfo)))
		return -EFAULT;
	return 0;
}
