/*
 * linux/fs/xvfat/rdchk.c
 * Copyright 2003 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; version 2
 * of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/list.h>
#include <linux/locks.h>
#include <linux/smp_lock.h>
#include <linux/slab.h>
#include <linux/blkdev.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#if defined(CONFIG_USB_STORAGE_SCHED_OPTION)
#include <linux/usb.h>
#endif
#include <linux/rdchk.h>

#define MEDIA_POLLING_INTERVAL (HZ*5)

#define POLL_SUPPORT
#ifdef POLL_SUPPORT
#include <linux/poll.h>
#endif

#define RDCHK_S_CHANGE 0
#define RDCHK_S_INSERT 1
#define RDCHK_S_PERIOD 2
#define RDCHK_S_DELAYINV 3

#define RDCHK_S_CHANGE_BIT (1 << (RDCHK_S_CHANGE))
#define RDCHK_S_INSERT_BIT (1 << (RDCHK_S_INSERT))
#define RDCHK_S_PERIOD_BIT (1 << (RDCHK_S_PERIOD))
#define RDCHK_S_DELAYINV_BIT (1 << (RDCHK_S_DELAYINV))

struct rd_info {
	struct list_head dev_list;
	struct list_head wait_list;
	wait_queue_head_t r_wait_head;
	kdev_t r_dev;
	int r_state;
	int r_capability;
	atomic_t r_count;
	struct blkdev_io_error *bioe;
	char name[32];
	unsigned int r_state_change_count;
};

static pid_t rdchkd_running;
static DECLARE_WAIT_QUEUE_HEAD(rdchkd_wait);
static LIST_HEAD(rd_info_list);
static LIST_HEAD(rd_info_wait_list);
static atomic_t rdchk_count = ATOMIC_INIT(1);	/* alway polling */
static struct rw_semaphore rdchk_info_list_sem;
static struct rw_semaphore rdchk_running_sem;
static struct semaphore rdchk_finish_sem;
static void rdchk_proc_notify(struct rd_info *rdi);
static int diff_state(int statea, int stateb, int capa, int capb);

/**************************************************************************/
int get_disk_change_status(kdev_t dev)
{
	const struct block_device_operations * bdops;
	bdops = get_blkfops_noload(dev);
	if (bdops == NULL)
		return 0;
	if (bdops->check_media_change == NULL)
		return 0;
	if (!bdops->check_media_change(dev))
		return 0;
	return 1;
}

int get_disk_insert_status(kdev_t dev)
{
	const struct block_device_operations * bdops;
	bdops = get_blkfops_noload(dev);
	if (bdops == NULL)
		return 0;
	if (bdops->check_media_insert == NULL)
		return 0;
	if (!bdops->check_media_insert(dev))
		return 0;
	return 1;
}

int revalidate_blkdev(kdev_t dev, int flag)
{
	const struct block_device_operations * bdops;
	bdops = get_blkfops_noload(dev);
	if (bdops == NULL)
		return -1;
	if (bdops->revalidate && flag)
		bdops->revalidate(dev);
	rdchk_notify_revalidate(dev);
	return 0;
}

/**************************************************************************/

static void rdi_set_state(struct rd_info *rdi, int type, int val)
{
	int state_save = rdi->r_state;

	if( val )
		set_bit(type, &rdi->r_state);
	else
		clear_bit(type, &rdi->r_state);
#if defined(CONFIG_DTV) || defined(CONFIG_DTGNEXUS) || defined(CONFIG_DTG3G)
	if (state_save != rdi->r_state)
		rdi->r_state_change_count++;
#endif
}

#define rdi_state(rdi, type)  ((rdi)->r_state&(1<<(type)))
#define rdi_capability(rdi, type)  ((rdi)->r_capability&(1<<(type)))

static int
check_blkdev_capability(kdev_t dev)
{
	const struct block_device_operations *bdops;
	int cap;
	
	bdops = get_blkfops_noload(dev);
	if( !bdops )
		return -1;
	cap = 0;
	if( bdops->check_media_change )
		cap |= RDCHK_S_CHANGE_BIT;
	if( bdops->check_media_insert )
		cap |= RDCHK_S_INSERT_BIT;
	return cap;
}

static struct rd_info *find_rdinfo(kdev_t dev)
{
	struct list_head *pos;
	struct rd_info *e;

	for( pos = rd_info_list.next; pos != &rd_info_list; pos = pos->next ){
		e = list_entry(pos, struct rd_info, dev_list);
		if( e->r_dev == dev )
			return e;
	}
	return NULL;
}

static void wait_rdchkd_status_update(struct rd_info *rdi)
{
	static void check_status_no_rdchkd(struct rd_info *rdi);
	struct list_head *pos, *tmp;
	int registered;

	registered = 0;
	down_read(&rdchk_running_sem);
	if( !rdchkd_running ) {
		up_read(&rdchk_running_sem);
		check_status_no_rdchkd(rdi);
		return;
	}
	list_for_each_safe(pos, tmp, &rd_info_wait_list) {
		if( pos == &rdi->wait_list ) {
			registered = 1;
			break;
		}
	}
	if( !registered ) {
		list_add(&rdi->wait_list, &rd_info_wait_list);
		rdchk_printk("wait_rdchkd_status_update: wake_up rdchkd_wait\n");
		wake_up_interruptible(&rdchkd_wait);
	}
	up_read(&rdchk_running_sem);
	rdchk_printk("wait_rdchkd_status_update: wait on r_wait_head\n");
	wait_event(rdi->r_wait_head, list_empty(&rdi->wait_list));
	rdchk_printk("wait_rdchkd_status_update: wake_up.\n");
}

static void check_status(struct rd_info *rdi)
{
	int cap;

	cap = check_blkdev_capability(rdi->r_dev);
	if( cap == -1 )
		rdi->r_capability = 0;
	else
		rdi->r_capability = cap;
	if( !rdi_capability(rdi, RDCHK_S_CHANGE) )
		return;
	if( rdi_state(rdi, RDCHK_S_CHANGE) )
		return;
        if( atomic_read(&rdi->bioe->e_count) > 0 )
		wait_rdchkd_status_update(rdi);
}

static void check_status_no_rdchkd(struct rd_info *rdi)
{
	if( rdi_capability(rdi, RDCHK_S_CHANGE) &&
	    !rdi_state(rdi, RDCHK_S_CHANGE) &&
	    atomic_read(&rdi->bioe->e_count) > 0 ) {
		rdi_set_state(rdi, RDCHK_S_CHANGE, 1);
		rdchk_delay_invalidate(rdi->r_dev);
		rdchk_proc_notify(rdi);
	}
}

int rdchk_get_disk_change(kdev_t dev, int call_daemon)
{
	struct rd_info *rdi;
	int rval;

	down_read(&rdchk_info_list_sem);
	rdi = find_rdinfo(dev);
	if( !rdi ) {
		rval = -2;
		goto end_func;
	}
	if( call_daemon ) {
		check_status(rdi);
	} else {
		check_status_no_rdchkd(rdi);
		/* not use rdchkd daemon */
	}
	if( !rdi_capability(rdi, RDCHK_S_CHANGE) ) {
		rval = -1;
		goto end_func;
	}
	if( rdi_state(rdi, RDCHK_S_CHANGE) )
		rval = 1;
	else
		rval = 0;
 end_func:
	up_read(&rdchk_info_list_sem);
	return rval;
}

int rdchk_get_disk_insert(kdev_t dev, int call_daemon)
{
	struct rd_info *rdi;
	int rval;

	down_read(&rdchk_info_list_sem);
	rdi = find_rdinfo(dev);
	if( !rdi ) {
		rval = -2;
		goto end_func;
	}
	if( call_daemon ) {
		check_status(rdi);
	}
	if( !rdi_capability(rdi, RDCHK_S_INSERT) ) {
		rval = -1;
		goto end_func;
	}
	if( rdi_state(rdi, RDCHK_S_INSERT) )
		rval = 1;
	else
		rval = 0;
 end_func:
	up_read(&rdchk_info_list_sem);
	return rval;
}

int rdchk_get_disk_state_change_count(kdev_t dev, unsigned int *ret, int call_daemon)
{
	struct rd_info *rdi;
	int rval;

	down_read(&rdchk_info_list_sem);
	rdi = find_rdinfo(dev);
	if( !rdi ) {
		rval = -2;
		goto end_func;
	}
	if( call_daemon ) {
		check_status(rdi);
	}
	*ret = rdi->r_state_change_count;
	rval = 1;
 end_func:
	up_read(&rdchk_info_list_sem);
	return rval;
}

static void rdchk_notify_revalidate_internal(kdev_t dev)
{
	struct rd_info *rdi;

	down_read(&rdchk_info_list_sem);
	rdi = find_rdinfo(dev);
	up_read(&rdchk_info_list_sem);
	rdchk_printk("rdchk_notify_revalidate: dev=%s rdi=%08x\n",
		   bdevname(dev), (unsigned)rdi);
	if( rdi ){
		/*
		if( rdi_state(rdi, RDCHK_S_CHANGE) ){
			rdi_set_state(rdi, RDCHK_S_CHANGE, 0);
			rdchk_proc_notify(rdi);
		}
		*/
		wake_up_interruptible(&rdchkd_wait);
	}
}

int rdchk_set_period_check(kdev_t dev, int enable)
{
	struct rd_info *rdi;
	int rval;

	down_read(&rdchk_info_list_sem);
	rdi = find_rdinfo(dev);
	if( rdi == NULL ) {
		rval = -EINVAL;
	} else {
		rdi_set_state(rdi, RDCHK_S_PERIOD, enable);
		rval = 0;
	}
	up_read(&rdchk_info_list_sem);
	return rval;
}

int rdchk_delay_invalidate(kdev_t dev)
{
	struct rd_info *rdi;
	int rval;

	down_read(&rdchk_info_list_sem);
	rdi = find_rdinfo(dev);
	if( rdi == NULL ) {
		rval = -EINVAL;
	} else {
		rdi_set_state(rdi, RDCHK_S_DELAYINV, 1);
		wake_up_interruptible(&rdchkd_wait);
		rval = 0;
	}
	up_read(&rdchk_info_list_sem);
	return rval;
}

void rdchk_get_fs(kdev_t dev)
{
	struct rd_info *rdi;

	down_read(&rdchk_info_list_sem);
	rdi = find_rdinfo(dev);
	if( rdi == NULL )
		panic("rdchk_get_fs: not found device (dev=%s)\n",
		      bdevname(dev));
	atomic_inc(&rdi->r_count);
	if( rdi_capability(rdi, RDCHK_S_CHANGE) &&
	    rdi_state(rdi, RDCHK_S_CHANGE) &&
	    atomic_read(&rdi->bioe->e_count) == 0 ) {
		/* When mount file system, parhaps media not changed   */
		rdi_set_state(rdi, RDCHK_S_CHANGE, 0);
		rdchk_proc_notify(rdi);
	}
	up_read(&rdchk_info_list_sem);
}

void rdchk_put_fs(kdev_t dev)
{
	struct rd_info *rdi;

	down_read(&rdchk_info_list_sem);
	rdi = find_rdinfo(dev);
	if( rdi == NULL )
		panic("rdchk_put_fs: not found device (dev=%s)\n",
		      bdevname(dev));
	atomic_dec(&rdi->r_count);
	up_read(&rdchk_info_list_sem);
}

int rdchk_register_blkdev(kdev_t dev)
{
	struct rd_info *rdi;
	int cap;

	cap = check_blkdev_capability(dev);
	if( cap == -1 )
		return -ENODEV;

	down_read(&rdchk_info_list_sem);
	rdi = find_rdinfo(dev);
	if( rdi ){
		int prev_cap, prev_state;

		prev_cap = rdi->r_capability;
		prev_state = rdi->r_state;

		if( cap & RDCHK_S_CHANGE_BIT ) {
			if( atomic_read(&rdi->r_count) == 0 ){
				rdi_set_state(rdi, RDCHK_S_CHANGE, 0);
				if( cap & RDCHK_S_INSERT_BIT )
					rdi_set_state(rdi, RDCHK_S_INSERT, 1);
			}
		}
		rdi->r_capability = cap;
		up_read(&rdchk_info_list_sem);
		if( diff_state(rdi->r_state, prev_state, rdi->r_capability,
			       prev_cap) )
			rdchk_proc_notify(rdi);
	} else {
		up_read(&rdchk_info_list_sem);
		rdi = (struct rd_info *)
			kmalloc(sizeof(struct rd_info), GFP_KERNEL);
		if( !rdi )
			panic("rdchk_register_blkdev: NOMEM\n");
		memset(rdi, 0, sizeof(struct rd_info));
		rdi->r_dev = dev;
		strncpy(rdi->name, bdevname(dev), 32);
		rdi->name[31] = '\0';
		rdi->r_capability = cap;
		if( MAJOR(dev) == 2 ) {	/* Floppy block device */
			rdi->r_state = 0;
			printk("rdchk_register_blkdev: not polling floppy drive (dev:%s)\n", bdevname(dev));
		} else
			rdi->r_state = RDCHK_S_PERIOD_BIT;
		atomic_set(&rdi->r_count, 0);
		rdi->bioe = bioe_register(dev);
		if( !rdi->bioe )
			panic("rdchk_register_blkdev: NOMEM\n");
		init_waitqueue_head(&rdi->r_wait_head);
		INIT_LIST_HEAD(&rdi->wait_list);
		down_write(&rdchk_info_list_sem);
		list_add(&rdi->dev_list, &rd_info_list);
		up_write(&rdchk_info_list_sem);
	}
	return 0;
}

static void rdchk_unregister(struct rd_info *rdi)
{
	list_del(&rdi->dev_list);
	if( !list_empty(&rdi->wait_list) )
		printk("rdchk_unregister: rdi->wait_list isn't empty list\n");
	if( atomic_read(&rdi->r_count) != 0 )
		printk("rdchk_unregister: rdi->r_count != 0 (r_count=%d)\n",
		       atomic_read(&rdi->r_count));
	list_del_init(&rdi->wait_list);
	bioe_unregister(rdi->r_dev);
	kfree(rdi);
}

static void rdchk_unregister_all(void)
{
	struct list_head *list;
	struct rd_info *rdi;

	down_write(&rdchk_info_list_sem);
	while ( (list = rd_info_list.next) != &rd_info_list ) {
		rdi = list_entry(list, struct rd_info, dev_list);
		rdchk_unregister(rdi);
	}
	up_write(&rdchk_info_list_sem);
}

void rdchk_unregister_blkdev(kdev_t dev)
{
	struct rd_info *rdi;

	down_read(&rdchk_info_list_sem);
	rdi = find_rdinfo(dev);
	up_read(&rdchk_info_list_sem);
	if( rdi == NULL )
		return;
	down_write(&rdchk_info_list_sem);
	rdchk_unregister(rdi);
	up_write(&rdchk_info_list_sem);
}

/**************************************************************************/
static void rdchk_invalidate_device(kdev_t dev, int mount_count)
{
	if( mount_count )
		fwrq_invalidate(dev);
	rdchk_printk("rdchkd: invalidate_device %s\n", bdevname(dev));
	invalidate_device(dev,0);
}

static int diff_state(int statea, int stateb, int capa, int capb)
{
	int mask;

	mask = (RDCHK_S_CHANGE_BIT|RDCHK_S_INSERT_BIT);
	if( (statea & mask) != (stateb & mask) ||
	    (capa & mask) != (capb & mask) )
		return 1;
	else
		return 0;
}

static void rdchk_one(struct rd_info *rdi)
{
	int chg, ins, cap, prev_cap, prev_state, e_count;

	chg = ins = 0;
	prev_cap = rdi->r_capability;
	prev_state = rdi->r_state;
	cap = check_blkdev_capability(rdi->r_dev);
	if( cap == -1 )
		cap = 0;
	rdi->r_capability = cap;
	lock_kernel();
	if( rdi_state(rdi, RDCHK_S_DELAYINV) ){
		rdi_set_state(rdi, RDCHK_S_DELAYINV, 0);
		rdchk_invalidate_device(rdi->r_dev,
					atomic_read(&rdi->r_count));
	}
	e_count = atomic_read(&rdi->bioe->e_count);
	if( rdi_capability(rdi, RDCHK_S_CHANGE) )
		chg = get_disk_change_status(rdi->r_dev);
	if( rdi_capability(rdi, RDCHK_S_INSERT) ) {
		ins =  get_disk_insert_status(rdi->r_dev);
		rdi_set_state(rdi, RDCHK_S_INSERT, ins);
	}
	rdchk_printk("rdchk_one: chg=%d ins=%d\n", chg, ins);
	if( !rdi_capability(rdi, RDCHK_S_CHANGE) ){
		if( diff_state(rdi->r_state, prev_state,
			       rdi->r_capability, prev_cap) )
			rdchk_proc_notify(rdi);
		rdchk_printk("rdchk_one: dev=%s  chg_cap=%d ins_cap=%d chg=%d ins=%d\n",
			     bdevname(rdi->r_dev),
			     rdi_capability(rdi, RDCHK_S_CHANGE),
			     rdi_capability(rdi, RDCHK_S_INSERT),
			     chg,
			     ins);
		unlock_kernel(); /* FES patch */
		return;
	}
	if( chg == 1 ){
		if( !rdi_state(rdi, RDCHK_S_CHANGE) ){
			rdchk_invalidate_device(rdi->r_dev, 
						atomic_read(&rdi->r_count));
			printk("rdchkd: ejected, device %s\n",
			       bdevname(rdi->r_dev));
		}
#ifdef NEVER
		else if( atomic_read(&rdi->r_count) == 0 ){
			rdchk_invalidate_device(rdi->r_dev,
						atomic_read(&rdi->r_count));
			revalidate_blkdev(rdi->r_dev, 1);
		}
#endif
		rdi_set_state(rdi, RDCHK_S_CHANGE, 1);
	} else {
		if( rdi_state(rdi, RDCHK_S_CHANGE) &&
		    atomic_read(&rdi->r_count) == 0 ){	/* mount count == 0 */
			rdi_set_state(rdi, RDCHK_S_CHANGE, 0);
			rdchk_invalidate_device(rdi->r_dev,
						atomic_read(&rdi->r_count));
			revalidate_blkdev(rdi->r_dev, 1);
		}
	}
	if( e_count > 0 ){
		/* clear I/O error count */
		if( atomic_add_negative(-e_count, &rdi->bioe->e_count) )
			BUG();
	}
	unlock_kernel();

	if( diff_state(rdi->r_state, prev_state, rdi->r_capability, prev_cap) )
		rdchk_proc_notify(rdi);

	rdchk_printk("rdchk_one: dev=%s  chg_cap=%d ins_cap=%d chg=%d(%d) ins=%d(%d)\n",
		      bdevname(rdi->r_dev),
		      rdi_capability(rdi, RDCHK_S_CHANGE),
		      rdi_capability(rdi, RDCHK_S_INSERT),
		      rdi_state(rdi, RDCHK_S_CHANGE), chg,
		      rdi_state(rdi, RDCHK_S_INSERT), ins);
}

static int rdchkd(void *startup)
{
	struct task_struct *tsk = current;
	struct rd_info *rdi;
	struct list_head *list;
	int interval;

	daemonize();
	strcpy(tsk->comm, "rdchkd");
	rdchkd_running = 1;

	/* avoid getting signals */
	spin_lock_irq(&tsk->sigmask_lock);
	flush_signals(tsk);
	sigfillset(&tsk->blocked);
	recalc_sigpending(tsk);
	spin_unlock_irq(&tsk->sigmask_lock);

#if defined(CONFIG_USB_STORAGE_SCHED_OPTION)
	if (usb_stor_set_default_schedule_priority() < 0) {
                        printk("%s: set schedule & priority failed(%d:%d). Continue to run\n",
                               __FUNCTION__, 
                               CONFIG_USB_STORAGE_SCHED_POLICY, CONFIG_USB_STORAGE_SCHED_PRIORITY);
	}
#endif 
	complete((struct completion *)startup);

	interval = MEDIA_POLLING_INTERVAL;

	rdchk_printk("rdchkd: start daemon [%d:%s]\n",
		     current->pid, current->comm);

	for (;;) {
		if (atomic_read(&rdchk_count) == 0){
			rdchk_printk2("rdchkd: Sleep.\n");
			wait_event_interruptible(rdchkd_wait,
						 atomic_read(&rdchk_count));
		} else {
			rdchk_printk2("rdchkd: Sleep interval=%d.\n", interval);
			interruptible_sleep_on_timeout(&rdchkd_wait,
						       interval);
		}
		down_read(&rdchk_running_sem);
		if ( !list_empty(&rd_info_wait_list) ) {
			while( (list = rd_info_wait_list.next)
			       != &rd_info_wait_list ) {
				rdi = list_entry(list, struct rd_info,
						 wait_list);
				rdchk_printk("rdchkd: %s\n", bdevname(rdi->r_dev));
				rdchk_one(rdi);
				list_del_init(list);
				wake_up(&rdi->r_wait_head);
			}
			up_read(&rdchk_running_sem);
			continue;
		}
		if( !rdchkd_running ) {
			up_read(&rdchk_running_sem);
			break;
		}
		up_read(&rdchk_running_sem);
		rdchk_printk2("rdchkd: check .. rd_info_list\n");
		down_read(&rdchk_info_list_sem);
		for( list = rd_info_list.next; list != &rd_info_list;
		     list = list->next ){
			rdi = list_entry(list, struct rd_info, dev_list);
			if( rdi_state(rdi, RDCHK_S_PERIOD) )
				rdchk_one(rdi);
		}
		up_read(&rdchk_info_list_sem);
	}
	up(&rdchk_finish_sem);
	return 0;
}

/**************************************************************************/
#ifdef CONFIG_PROC_FS

#define EVBUF_SIZE 100
#define LINEBUF_SIZE 80

struct proc_evbuf {
	struct list_head buf_list;
	struct proc_evbuf_log {
		int r_state;
		int r_capability;
		char name[32];
#if defined(CONFIG_DTV) || defined(CONFIG_DTGNEXUS) || defined(CONFIG_DTG3G)
		unsigned int r_state_change_count;
#endif
	} log[EVBUF_SIZE];
	int head;
	int tail;
	loff_t ppos;
	wait_queue_head_t ev_wait;
	char linebuf[LINEBUF_SIZE];
	int linebuf_size;
	int linebuf_pos;
	struct semaphore ev_sem;
};

static LIST_HEAD(proc_evbuf_list);
static struct semaphore proc_evbuf_list_sem;

static int
rdchk_status_read_proc(char *page, char **start, off_t off, int count,
		       int *eof, void *data)
{
	struct list_head *list;
	struct rd_info *rdi;
	int len = 0;
	int limit = PAGE_SIZE - 80;
	int chg, ins;
	down_read(&rdchk_info_list_sem);
	for( list = rd_info_list.next; list != &rd_info_list;
	     list = list->next ){
		rdi = list_entry(list, struct rd_info, dev_list);
		if( rdi_capability(rdi, RDCHK_S_CHANGE) ) {
			chg = rdi_state(rdi, RDCHK_S_CHANGE) ? 1 : 0;
		} else {
			chg = -1;
		}
		if( rdi_capability(rdi, RDCHK_S_INSERT) ) {
			ins = rdi_state(rdi, RDCHK_S_INSERT) ? 1 : 0;
		} else {
			ins = -1;
		}
#if defined(CONFIG_DTV) || defined(CONFIG_DTGNEXUS) || defined(CONFIG_DTG3G)
		len += sprintf(page+len, "%s %d %d %u\n", rdi->name, ins, chg, rdi->r_state_change_count);
#else
		len += sprintf(page+len, "%s %d %d \n", rdi->name, ins, chg);
#endif
		if( len >= limit ){
			len = -ENOSPC;
			break;
		}
	}
	up_read(&rdchk_info_list_sem);
	*eof = 1;
	return len;
}
#if defined(CONFIG_DTV) || defined(CONFIG_DTGNEXUS) || defined(CONFIG_DTG3G)
static void add_evbuf(struct proc_evbuf *evbuf, char *name, int state,
		      int capability, unsigned int change_count)

#else
static void add_evbuf(struct proc_evbuf *evbuf, char *name, int state,
		      int capability)
#endif
{
	int next;

	next = (evbuf->head + 1) % EVBUF_SIZE;
	if ( evbuf->tail == next )	/* buffer full */
		evbuf->tail = (evbuf->tail + 1) % EVBUF_SIZE;
	evbuf->log[evbuf->head].r_state = state;
	evbuf->log[evbuf->head].r_capability = capability;
#if defined(CONFIG_DTV) || defined(CONFIG_DTGNEXUS) || defined(CONFIG_DTG3G)
	evbuf->log[evbuf->head].r_state_change_count = change_count;
#endif
	strncpy(evbuf->log[evbuf->head].name, name, 32);
	evbuf->head = next;
}

static int rdchk_event_open(struct inode * inode, struct file * file)
{
	struct proc_evbuf *evbuf;

	evbuf = (struct proc_evbuf *)
		kmalloc(sizeof(struct proc_evbuf), GFP_KERNEL);
	if( evbuf == NULL )
		return -ENOMEM;
	evbuf->head = evbuf->tail = 0;
	evbuf->ppos = 0;
	evbuf->linebuf_size = evbuf->linebuf_pos = 0;
	init_waitqueue_head(&evbuf->ev_wait);
	sema_init(&evbuf->ev_sem, 1);
	down(&proc_evbuf_list_sem);
	list_add(&evbuf->buf_list, &proc_evbuf_list);
	up(&proc_evbuf_list_sem);
	file->private_data = evbuf;
	return 0;
}

static int rdchk_event_release(struct inode * inode, struct file * file)
{
	struct proc_evbuf *evbuf;

	evbuf = (struct proc_evbuf *)file->private_data;
	down(&proc_evbuf_list_sem);
	list_del(&evbuf->buf_list);
	up(&proc_evbuf_list_sem);
	kfree(evbuf);
	return 0;
}

static size_t
copy_linebuf_to_user(char *buf, size_t len, struct proc_evbuf *evbuf)
{
	int clen;

	clen = evbuf->linebuf_size - evbuf->linebuf_pos;
	if( clen == 0 )
		return 0;
	if( clen > len )
		clen = len;
	__copy_to_user(buf, evbuf->linebuf + evbuf->linebuf_pos, clen);
	evbuf->linebuf_pos += clen;
	return clen;
}

static int
fill_linebuf(struct proc_evbuf *evbuf)
{
	int blen, pos, chg, ins;

	blen = evbuf->linebuf_size - evbuf->linebuf_pos;
	if( blen > 0 )
		return blen;
	if( evbuf->head == evbuf->tail )
		return 0;
	pos = evbuf->tail;
	if( evbuf->log[pos].r_capability & RDCHK_S_CHANGE_BIT )
		chg = (evbuf->log[pos].r_state & RDCHK_S_CHANGE_BIT) ? 1 : 0;
	else
		chg = -1;
	if( evbuf->log[pos].r_capability & RDCHK_S_INSERT_BIT )
		ins = (evbuf->log[pos].r_state & RDCHK_S_INSERT_BIT) ? 1 : 0;
	else
		ins = -1;
#if defined(CONFIG_DTV) || defined(CONFIG_DTGNEXUS) || defined(CONFIG_DTG3G)
	blen = sprintf(evbuf->linebuf, "%s %d %d %u\n",
		       evbuf->log[pos].name, ins, chg, evbuf->log[pos].r_state_change_count);
#else
	blen = sprintf(evbuf->linebuf, "%s %d %d\n",
		       evbuf->log[pos].name, ins, chg);
#endif
	evbuf->tail = (pos + 1) % EVBUF_SIZE;
	evbuf->linebuf_size = blen;
	evbuf->linebuf_pos = 0;
	return blen;
}

static ssize_t
rdchk_event_read(struct file *file, char *buf, size_t len, loff_t *ppos)
{
	struct proc_evbuf *evbuf;
	size_t ulen;
	int rval;

	evbuf = (struct proc_evbuf *)file->private_data;
	rdchk_printk2("rdchk_event_read: len=%d ppos=%d(%d)\n",
		      len, (int)*ppos, (int)evbuf->ppos);
	if( *ppos != evbuf->ppos )
		return -EIO;
	if( !access_ok(VERIFY_READ, buf, len) )
		return -EFAULT;

	down(&evbuf->ev_sem);
	ulen = copy_linebuf_to_user(buf, len, evbuf);
	while( ulen < len ){
		rval = fill_linebuf(evbuf);
		rdchk_printk2("rdchk_event_read: ulen=%d fill=%d\n",
			      ulen, rval);
		if( rval > 0 )
			ulen += copy_linebuf_to_user(buf + ulen, len - ulen,
						     evbuf);
		else if( ulen > 0 )
			break;
		else {
			up(&evbuf->ev_sem);
			if( file->f_flags & O_NONBLOCK )
				return -EAGAIN;
			rdchk_printk2("rdchk_event_read: sleep\n");
			interruptible_sleep_on(&evbuf->ev_wait);
			rdchk_printk2("rdchk_event_read: wakeup\n");
			if (signal_pending(current))
				return -ERESTARTSYS;
			down(&evbuf->ev_sem);
		}
	}
	*ppos += ulen;
	evbuf->ppos = *ppos;
	up(&evbuf->ev_sem);
	rdchk_printk2("rdchk_event_read: read_size=%d\n", ulen);
	return ulen;
}

#ifdef POLL_SUPPORT
static unsigned int rdchk_event_poll(struct file *file, poll_table * wait)
{
	struct proc_evbuf *evbuf;

	evbuf = (struct proc_evbuf *)file->private_data;

	poll_wait(file, &evbuf->ev_wait, wait);
	if (evbuf->head != evbuf->tail)
		return POLLIN | POLLRDNORM;
	return 0;
}
#endif

static struct file_operations event_fops = 
{
	read:		rdchk_event_read,
	open:		rdchk_event_open,
	release:	rdchk_event_release,
#ifdef POLL_SUPPORT
	poll:		rdchk_event_poll,
#endif
};

static struct proc_dir_entry *proc_rd_root;
static struct proc_dir_entry *proc_status;
static struct proc_dir_entry *proc_event;

static int __init rdchk_proc_init(void)
{
	sema_init(&proc_evbuf_list_sem, 1);
	proc_rd_root = proc_mkdir("removabledisk", 0);
	if( proc_rd_root ){
		proc_status = create_proc_read_entry("status", S_IRUGO,
						     proc_rd_root,
						     rdchk_status_read_proc,
						     NULL);
		proc_event = create_proc_entry("event", S_IRUGO,
					       proc_rd_root);
		if( proc_event )
			proc_event->proc_fops = &event_fops;
	}
	return 0;
}

static void rdchk_proc_exit(void)
{
	if( proc_rd_root ){
		remove_proc_entry("status", proc_rd_root);
		remove_proc_entry("event", proc_rd_root);
		remove_proc_entry("removabledisk", NULL);
	}
}

static void rdchk_proc_notify(struct rd_info *rdi)
{
	struct list_head *list;
	struct proc_evbuf *evbuf;

	rdchk_printk("rdchk_proc_notify: dev=%s  chg_cap=%d ins_cap=%d chg=%d ins=%d\n",
		     bdevname(rdi->r_dev),
		     rdi_capability(rdi, RDCHK_S_CHANGE),
		     rdi_capability(rdi, RDCHK_S_INSERT),
		     rdi_state(rdi, RDCHK_S_CHANGE),
		     rdi_state(rdi, RDCHK_S_INSERT));

	down(&proc_evbuf_list_sem);
	for( list = proc_evbuf_list.next; list != &proc_evbuf_list;
	     list = list->next ){
		evbuf = list_entry(list, struct proc_evbuf, buf_list);
		down(&evbuf->ev_sem);
#if defined(CONFIG_DTV) || defined(CONFIG_DTGNEXUS) || defined(CONFIG_DTG3G)
		add_evbuf(evbuf, rdi->name, rdi->r_state, rdi->r_capability, rdi->r_state_change_count);
#else
		add_evbuf(evbuf, rdi->name, rdi->r_state, rdi->r_capability);
#endif
		up(&evbuf->ev_sem);
		wake_up(&evbuf->ev_wait);
	}
	up(&proc_evbuf_list_sem);
}
#else  /* !CONFIG_PROC_FS */
static void rdchk_proc_notify(struct rd_info *rdi)
{
	rdchk_printk("rdchk_proc_notify: dev=%s  chg_cap=%d ins_cap=%d chg=%d ins=%d\n",
		     bdevname(rdi->r_dev),
		     rdi_capability(rdi, RDCHK_S_CHANGE),
		     rdi_capability(rdi, RDCHK_S_INSERT),
		     rdi_state(rdi, RDCHK_S_CHANGE),
		     rdi_state(rdi, RDCHK_S_INSERT));
}
#endif /* !CONFIG_PROC_FS */

/**************************************************************************/

static int __init
rdchkd_init(void)
{
	static struct completion startup __initdata = COMPLETION_INITIALIZER(startup);

	kernel_thread(rdchkd, &startup, 0);
	wait_for_completion(&startup);
	rdchk_notify_revalidate_func = rdchk_notify_revalidate_internal;
	return 0;
}

static void
rdchkd_finish(void)
{
	rdchk_notify_revalidate_func = NULL;
	down_write(&rdchk_running_sem);
	rdchkd_running = 0;
	up_write(&rdchk_running_sem);
	rdchk_printk("rdchk_init[%d:%s]: wake up rdchkd\n",
		     current->pid, current->comm);
	wake_up_interruptible(&rdchkd_wait);
	down(&rdchk_finish_sem);
	return;
}

int __init rdchk_init(void)
{
	rdchk_printk("rdchk_init[%d:%s]: start\n",
		     current->pid, current->comm);
	init_rwsem(&rdchk_info_list_sem);
	init_rwsem(&rdchk_running_sem);
	sema_init(&rdchk_finish_sem, 0);
#ifdef CONFIG_PROC_FS
	rdchk_proc_init();
#endif
	rdchkd_init();
	return 0;
}

void __exit rdchk_exit(void)
{
	rdchk_printk("rdchk_exit[%d:%s]:\n",
		     current->pid, current->comm);
	rdchk_unregister_all();
	rdchk_printk("rdchk_exit[%d:%s]: stop daemon\n",
		     current->pid, current->comm);
	rdchkd_finish();	
#ifdef CONFIG_PROC_FS
	rdchk_printk("rdchk_exit[%d:%s]: stop proc FS service\n",
		     current->pid, current->comm);
	rdchk_proc_exit();
#endif
	rdchk_printk("rdchk_exit[%d:%s]: end\n",
		     current->pid, current->comm);
}

