/*
 *  LIDS - Trusted Path Execution (TPE) 
 *
 *  Copyright 2001-2004 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/module.h>
#include <linux/lids.h>
#include <linux/fs.h>

#include <asm/mman.h>

#define MODNAME_MAX	256

/*
 * Switch TPE mode off as the
 * initial state.
 */
int lids_tpe = 0;

int lids_exec_tpe_permission(struct linux_binprm *bprm)
{
	int error = 0;
	struct dentry *dentry;

	if (current->p_pptr->pid == 0)
		return 0;

	if (!bprm || !bprm->file)
		return 0;

	if (!(lids_load && lids_local_load))
		return 0;

	if (!lids_tpe)
		return 0;

	dentry = bprm->file->f_dentry;

	/* Check if binary is protected */

	if (!lids_check_base(dentry, LIDS_APPEND)) {
		lids_security_alert("TPE violation! "
				    "Attempt to exec unprotected program "
				    "%s (dev %d:%d inode %ld)",
				    dentry->d_iname,
				    MAJOR(dentry->d_inode->i_dev),
				    MINOR(dentry->d_inode->i_dev),
				    dentry->d_inode->i_ino);
		error = -EACCES;
	}

	return error;
}

int lids_mmap_tpe_permission(struct file *file, unsigned long prot,
			     unsigned long flags)
{
	int error = 0;
	struct dentry *dentry;

	if (!(lids_load && lids_local_load))
		return 0;

	if (!lids_tpe)
		return 0;

	if (!file)
		return 0;

	if (!(prot & PROT_EXEC))
		return 0;

	dentry = file->f_dentry;

	/* Check if the mmap'ed file is protected */

	if (!lids_check_base(dentry, LIDS_APPEND)) {
		lids_security_alert("TPE violation! "
				    "Attempt to mmap %s (dev %d:%d inode %ld)",
				    dentry->d_iname,
				    MAJOR(dentry->d_inode->i_dev),
				    MINOR(dentry->d_inode->i_dev),
				    dentry->d_inode->i_ino);
		error = -EACCES;
	}

	return error;
}

int lids_module_tpe_permission(struct module *mod)
{
	char *modpath, *mark, *p, *q;
	int error = 0, n, namelen;
	struct dentry *dentry;
	struct module_symbol *sym;
	struct nameidata nd;

	if (!(lids_load && lids_local_load))
		return 0;

	if (!lids_tpe)
		return 0;

	namelen = strlen(mod->name);
	if (unlikely(namelen > MODNAME_MAX)) {
		error = -EINVAL;
		goto out1;
	} 

	mark = (char *) kmalloc(sizeof(char) * (MODNAME_MAX + 12), GFP_KERNEL);
	if (mark == NULL) {
		error = -ENOMEM;
		goto out1;
	} 
	sprintf(mark, "__insmod_%s_O", mod->name);

	modpath = (char *) kmalloc(sizeof(char) * (PATH_MAX + 1), GFP_KERNEL);
	if (modpath == NULL) {
		error = -ENOMEM;
		goto out2;
	}

	/* Get the module path */

	sym = mod->syms;
	for (n = 0; n < mod->nsyms; n++) {
		if (strstr(sym->name, mark)) {
			if (unlikely(strlen(sym->name + 11 + namelen) > PATH_MAX)) {
				error = -EINVAL;
				goto out3;
			} 	
			strcpy(modpath, sym->name + 11 + namelen);
			p = modpath;
			q = p;
			while ((p = strstr(p, mod->name))) {
				q = p;
				p++;
			}
			*(q + namelen + 2) = '\0';
			break;
		}
		sym++;
	}

	if (unlikely(strstr(modpath, mod->name) == NULL)) {
		error = -ENOENT;
		goto out3;
	}

	if (path_lookup(modpath, LOOKUP_FOLLOW | LOOKUP_POSITIVE, &nd)) {
		error = -ENOENT;
		goto out3;
	}

	dentry = nd.dentry;

	if (IS_ERR(dentry) || !dentry) {
		error = -ENOENT;
		goto out4;
	}

	if (!dentry->d_inode ||
	    !dentry->d_inode->i_dev ||
	    !dentry->d_inode->i_ino) {
		error = -ENOENT;
		goto out4;
	}

	/* Check if the module path is protected */

	if (!lids_check_base(dentry, LIDS_APPEND)) {
		lids_security_alert("TPE violation! "
				    "Attempt to load unprotected module %s "
				    "(dev %d:%d inode %ld)",
				    dentry->d_iname,
				    MAJOR(dentry->d_inode->i_dev),
				    MINOR(dentry->d_inode->i_dev),
				    dentry->d_inode->i_ino);
		error = -EPERM;
	}
out4:
	path_release(&nd);
out3:
	kfree(modpath);
out2:
	kfree(mark);
out1:
	return error;
}
